/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef __ALD_TIMAGE_H__
#define __ALD_TIMAGE_H__ 1

#ifndef __ASSEMBLY__


#ifndef __SYSBASE_H__
 #include "sysbase.h"
#endif

#ifndef __SYSBASEIO_H__
 #include "sysbaseio.h"
#endif

#ifndef __BIXFORMAT_H__
#include "bixformat.h"
#endif

//typedef struct TPen_wd *TPen;

  /* ************** pxm_LD ************** */
//#define pmxLD_NoMask		1
//#define pmxLD_USE_XSHM		(1<<1)

#define pxmLD_NoMask		(1<<8)
#define pxmLD_USE_XSHM		(1<<9)
#define pxmLD_OnlyInfo		(1<<10)
#define pxmLD_NoERR		(1<<11)
#define pxmLD_NoImgInfoStr	(1<<12)
#define pxmLD_FastZoom		(1<<13)

  /* pxm_LD    scale only for jpeg */
#define imgLD_scale_mask	0x70000
#define imgLD_scale_1		0x00000
#define imgLD_scale_2		0x10000
#define imgLD_scale_4		0x20000
#define imgLD_scale_8		0x30000


  /* ************** out_fu ************** */
#define fuWR_NoMask		1
#define fuWR_gray		(1<<1)
#define fuWR_use_pallette	(1<<2)
#define compress_level_mask	0x0000ff00

  /* ************** memfu *************** */
#define memfu_data_free		1
//#define memfu_mask_data_free	(1<<1)
#define memfu_cmap_free		(1<<2)


  /* ************** stat **************** */
/* used in TPixMap */
#define pmxStat_XPixAkt		(1<<1)
#define pmxStat_ImgAkt		(1<<2)
#define pmxStat_XMaskPixAkt	(1<<3)
#define pmxStat_MaskImgAkt	(1<<4)
#define pmxStat_SharedMemImg	(1<<8)
#define pmxStat_NoBixImgSwitch	(1<<9)
#define pmxStat_HandleMask	(1<<16)

#define pmxStat_has_alpha	(1<<28)
#define pmxStat_has_alpha_tgl	(1<<29)
#define pmxStat_ckh_alpha	(3<<28)

#define AlgTImageExImg AlgTTC1 \
  struct TImageExImg *next;

typedef struct TImageExImg {
  AlgTImageExImg
} TImageExImg;

#define AlgTImage AlgTTC \
  dword  pxm_LD; \
  int  width; \
  int  height; \
  dword memfu; \
  int stat; \
  union { \
    byte *data; \
    word *data16; \
    dword *data32; \
  }; \
  dword  bytes_per_line; \
  byte *mask_data; \
  TImageExImg *ExImg; \
  byte *rgb565_alpha_data; \
  int (*img_malloc)(struct TImage *self); \
  struct TColMap *cmap; \
  struct TPen *Pen; \
  void *struct_draw; \
  TglobPath *in_gp; \
  FILE *in_fp; \
  int in_filetyp; \
  dword  out_fu; \
  TglobPath *out_gp; \
  FILE *out_fp; \
  char *img_info; \
  struct TtmpIO *tIO;

typedef struct TImage {
  AlgTImage
} TImage;


typedef int (*Timg_malloc)(TImage *self);
typedef int (*TImageRD_pg)(TImage *self);


extern TImage *TImage_Create(int size, int wtyp);
#define dupfu_init_imgmem	0x01
#define dupfu_copy_img		0x02
#define dupfu_copy_alpha	0x04
#define dupfu_copy_mask		0x08
#define dupfu_copy_all		0x0f
extern TImage *TImage_Create_dup_struct(TImage *org, int fu);
extern TImage *TImage_Create_wht(int size, int wtyp, int width, int height, dword pxm_LD);
extern bool TImage__Destroy(TImage *self);

extern bool TImage_SetMem(TImage *self);



/* ________________________________________________________ */
/* pxm_LD   pxm_4b_24bpp | pxmLD_NoMask */
/* !!! *fn or *gp !!! */
extern TImage *TImage_Loader(const char *fn, TglobPath *gp, dword pxm_LD);
extern TImage *TImage_Loader_F(const char *fn, TglobPath *gp, dword pxm_LD, FILE *F);

extern TImage *TImage_hLoader(const char *fn, TglobPath *gp, dword pxm_LD,
			int type_size, int wtyp, Timg_malloc imal, FILE *F);

extern bool alpha_chk_mask(TImage *self, int alpha_mask);
extern bool alpha_chk_tgl(TImage *self);
extern void alpha_bm_stat(TImage *self);
extern void alpha2bitmask(TImage *self);
extern void bitmask2alpha(TImage *self);
extern bool alpha_akt(TImage *self);

extern void alpha_data8_get(TImage *self);
extern void alpha_data8_set(TImage *self);
extern void alpha_data8_clr(TImage *self);

extern TImage *TImage_rgb888to565(TImage *self);
extern TImage *TImage_rgb888to565m(TImage *self); // + alpha2bitmask(img)
extern void TImage_rgb888move565(TImage *img888, TImage *img565);
extern TImage *TImage_rgb565to888(TImage *self);
extern TImage *TImage_rgb565to888m(TImage *self); // + alpha2bitmask(img)
extern void TImage_ColorToAlphaData(TImage *self, dword Color);


/* ________________________________________________________ */
typedef struct {
  const char *pattern;
  int filetyp;
  bool (*ChkType)(TglobPath *gp, const char *fn, FILE *F);
  int (*ioproc)(TImage *self);
  bool (*LibError)();
} ThImageLoader;

extern void ThImageRD_Reg(int n, ...);
extern void ThImageWR_Reg(int n, ...);

extern dword TImageRD_AllFiletyp;  //filetyp_PNG,filetyp_JPEG,...
extern dword TImageWR_AllFiletyp;

extern char *TImageIO_GetAllFilePattern(bool wr);

extern FILE *TImageIO_GetStream(const char *fn, TglobPath *gp, dword pxm_LD, 
				ThImageLoader **pphLD, bool wr);

extern const ThImageLoader hImageRD_bmp;
extern const ThImageLoader hImageRD_gif;
extern const ThImageLoader hImageRD_jpeg;
extern const ThImageLoader hImageRD_png;
extern const ThImageLoader hImageRD_pnm;
extern const ThImageLoader hImageRD_xpm;

extern const ThImageLoader hImageWR_jpeg;
extern const ThImageLoader hImageWR_png;
extern const ThImageLoader hImageWR_pnm;
extern const ThImageLoader hImageWR_xpm;

extern bool TImage_ChkPattern(TglobPath *gp, const char *fn, const char *Pattern);


extern int read_xpm(TImage *self);
extern int read_png(TImage *self);
extern int read_jpeg(TImage *self);
extern int read_bmp(TImage *self);
extern int read_pnm(TImage *self);
extern int read_gif(TImage *self);

typedef struct TImageExGif {
  AlgTImageExImg
// Global
  int GlobColorCount;
  dword *GlobColorTab;
  dword BackGroundColor;
// Local
  int left;
  int top;
  int width;
  int height;
  bool Interlace, ncb1, ncb2, ncb3;

  int ColorCount;
  dword *ColorTab;
  int LzwLg;
  byte *LzwData;
// 0xF9 Graphic Control Extension.
  int DisposalMethod;
  dword DelayTime;	// n * 10ms
  dword TransparentIdx;
// 0xFE Comment Extension
  char *Comment;
} TImageExGif;

extern int DecodeExGif(TImage *self, TImageExGif *ExGif);

/* TImage_RD result */
#define IMGRD_ERROR 0
#define IMGWR_ERROR 0
#define IMGRD_OK 1
#define IMGWR_OK 1
#define IMGRD_ERR_HEAD 2
#define IMGWR_ERR_HEAD 2
#define IMGRD_ERR_FORMAT 3
#define IMGWR_ERR_FORMAT 3
#define IMGRD_ERR_MEM 4
#define IMG_ERR_NOSTREAM 4
#define IMGRD_ERR_STREAM 5
#define IMG_ERR_NOLIB 6

/* ****** image_load.c ****** */

#define filetyp_XPM		1
#define filetyp_PNG		(1<<1)
#define filetyp_JPEG		(1<<2)
#define filetyp_BMP		(1<<3)
#define filetyp_PNM		(1<<4)
#define filetyp_GIF		(1<<5)

/****** png.c ******/
extern bool LoadLib_PNG();


/* ****** image_getcolorcount.c ****** */
#define pxgcc_RGB_ALL	0
#define pxgcc_RGB_565	1
#define pxgcc_RGB_666	2
//#define pxgcc_RGB_777	3
#define pxgcc_RGB_Mask	0x07

#define pxgcc_ColorCount_AddMask	(1<<8)


typedef struct TColMap {
  AlgTTC0
  int pxgcc;
  int ColorCount;
  int arr_size;
  byte *bit_arr;
  ic32 *col_map;
  int quant_count;
//  int quant_max_count;
  ic32 *quant_col_map;
  int default_quant_count;
  ic32 *default_quant_col_map;
} TColMap;

extern TColMap *TColMap_Create();

extern int TImage_GetColorCount(TImage *self, int pxgcc);
extern bool TImage_GetColorMap(TImage *self);

/* ****** image_cutcolormap.c ****** */
extern int TImage_CutColorMap(TImage *self, int NewColors, int sortfu);

/* ****** colormap_sort.c ****** */
extern int CompareColor(ic32 a, ic32 b, dword sortfu);
extern void ColorMap_Sort(ic32 *col, int num, dword sortfu);

/* ****** image_remapcolor.c ****** */
extern bool TImage_ReMapColor(TImage *self);


//*********
extern int GetColorByName(const char *s);

/*__________________________ TImage WR _______________________ */
/*
typedef int (*TImageWR_pg)(TImage *self);

extern TImageWR_pg GetImageFileWriter(char *fn, int *filetyp);

extern int TImage_WR_gp(TImage *self, dword out_fu, TImageWR_pg WR_pg);
extern int TImage_WR(TImage *self, char *fn, dword out_fu, TImageWR_pg WR_pg);
*/

//fuWR = fuWR_gray fuWR_NoMask fuWR_use_pallette
//compress_level = 1..10  0=default
extern int TImage_Write_PNG(TImage *self, const char *fn, int fuWR, 
			    dword compress_level);


extern int write_xpm(TImage *self);

/* PNG out_fu 
  fuWR_NoMask	b0
  fuWR_gray	b1
  fuWR_use_pallette  b2 (use pallette ColorCount <= 256)
  compress_level b8..11  
	0=default
	1..10  
*/
extern int write_png(TImage *self);


/* PNG out_fu 
  fuWR_gray
  quality_level b8..15  
	95=default
	1..100  
*/
extern int write_jpeg(TImage *self);


/* PNM out_fu b8=1 text plain */
#define fuWR_txtplain		(1<<8)
extern int write_pnm(TImage *self);

/*___________________________ Resize ______________________________________*/

#define rsz_fu_Fast_rsz		0x010
#define rsz_fu_TmpNotFree	0x100
#define rsz_fu_HasMMX		0x200

#define rszERR_Zoom		1
#define rszERR_LE2x2Pixel	2
#define rszERR_ZoomGT255	3

typedef struct Timg_rsz { // 32Bit     64Bit
  word fu;		//asm 00
  byte src_typ;		//asm 02
  byte dest_typ;	//asm 03

  int src_w;		//asm 04
  int src_h;		//asm 08
  void *src_dat;	//asm 0c	10
  byte *src_mask;	//asm 10	18
  dword src_mask_color;	//asm 14	20
  int to_w;		//asm 18	24
  int to_h;		//asm 1c	28
  word src_off_x;	//asm 20	2c
  word src_off_y;	//asm 22	2e
  word src_lg_x;	//asm 24	30
  word src_lg_y;	//asm 26	32

  word dest_off_x;	//asm 28	34
  word dest_off_y;	//asm 2a	36
  word dest_lg_x;	//asm 2c	38
  word dest_lg_y;	//asm 2e	3a

  void *dest_dat;	//asm 30	40
  byte *dest_mask;	//asm 34	48
  dword dest_mask_color;//asm 38	50

  byte src_line_eoff;	//asm 3c	54
  byte dest_line_eoff;	//asm 3d	55
  word nc_3c;		//asm 3e	56
  void *rsztemp0;	//		58
} Timg_rsz;

cHIDDEN int ImgDat_Resize_XX_11(Timg_rsz *rsz);
cHIDDEN int ImgDat_Copy(Timg_rsz *rsz);


/*______________________ image_resize.c ________________________________*/
extern void TImage_SetBase_rsz(TImage *self, Timg_rsz *rsz);
extern void TImage_SetDestBase_rsz(TImage *self, Timg_rsz *rsz);
extern TImage *TImage_dup_new_size(TImage *self, int new_width, int new_height);

typedef struct  {
  int x;
  int y;
  int width;
  int height;
} Toff_rsz;

extern TImage *TImage_dup_new_size_off(TImage *self, int new_width, int new_height, Toff_rsz *srcoff);

/*______________________ image_move.c __________________________________*/
// fu = dupfu_copy_mask
#define IMGCOPY_MASK		1	//only copy visible data
#define IMGCOPY_MASK_BLEND	(1<<1)
#if 1
// if w=0 or h=0 use src->width src->height
extern void image_copy(TImage *dest, int dest_x, int dest_y,
		       TImage *src, int x, int y, int w, int h, int fu);
#else
extern void image_copy(TImage *src, TImage *dest, int x, int y, int w, int h, 
					int dest_x, int dest_y, int fu);
#endif

/* ______________________ image_line.c __________________________________ */
extern int TImage_DrawLineRGB(TImage *self, dword RGB, int x1, int y1, 
						       int x2, int y2);

extern bool TImage_DrawLineChk(TImage *self, int x1, int y1, int x2, int y2);
extern int TImage_pv_DrawLineRGB(TImage *self, dword RGB, int x1, int y1, 
							  int x2, int y2);

extern int TImage_DrawFLineARGB(TImage *self, dword ARGB, float x1, float Y1, 
					float x2, float y2, float lw, int fu);

extern bool TImage_DrawFLineChk(TImage *self, float x1, float Y1, 
						float x2, float y2);
extern int TImage_pv_DrawFLineARGB(TImage *self, dword ARGB, float x1, float Y1, 
					float x2, float y2, float lw, int fu);

extern void TImage_ClsRGB(TImage *self, dword RGB);


/* ______________________ pen ___________________________________________ */
extern void TImagePen_Create(TImage *img);
extern void TImagePenCairo_Create(TImage *img);  //use libcairo
extern void TImagePen_Free(TImage *img);

extern void TImagePen_Color(TImage *img, dword Color);
extern void TImagePen_LineWidth(TImage *img, float width);
extern void TImagePen_MoveTo(TImage *img, float x, float y);
extern void TImagePen_LineTo(TImage *img, float x, float y);
extern void TImagePen_Line(TImage *img, float x1, float y1, float x2, float y2);

//TImagePen_MoveTo, TImagePen_RectPoint, TImagePen_RectPoint, ..., TImagePen_Flush
extern void TImagePen_RectPoint(TImage *img, float x, float y);

extern void TImagePen_Flush(TImage *img);


#ifdef _OS_WIN32
/* ______________________ misc W32 ______________________________________ */
#ifndef _WINGDI_H
#include <wingdi.h>
#endif
extern bool DIB_chk_error(BITMAPINFOHEADER *p);
extern bool DIB_to_image(TImage *img, BITMAPINFOHEADER *p, byte *src_data);
extern bool image_to_DIB(TImage *img, BITMAPINFOHEADER *p, byte *dest_data);
    // False to chk appError(EMSG_FromLastBuf, NULL);

#endif /* _OS_WIN32 */


#endif /* ! __ASSEMBLY__ */

#endif /* __ALD_TIMAGE_H__ */
