/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef	__SYSBASEIO_H__
#define	__SYSBASEIO_H__    1

#include <sys/types.h>
//#include <sys/stat.h>
//#include <dirent.h>
//#include <time.h>
//#include <unistd.h>

#ifndef	__FILEMSG_H__
 #include "filemsg.h"
#endif


// ******************** File *******************************
#define gpi_Local pathi_Local
//#define gpi_RA_file pathi_RA_file
#ifdef _OS_WIN32
#define gpi_RA_SMB pathi_RA_SMB
#endif
#define gpi_dest pathi_dest  /* use (RxD/TxD Ftp:) xcopy */

typedef struct TglobPath {
  char *RootArt;
  char *URL;
  char *path;
  union {
    char *name;
    const char *name_const;
  };
  char *user;
  char *passwd;
  int fu;  //gpi_Local, ...
} TglobPath;

typedef struct TgpList {
  struct TgpList *next;
  TglobPath gp;
} TgpList;

extern TglobPath *pctogp(const char *fn);
extern void pcsetgp(TglobPath *gp, const char *fn);
extern void gp_free(TglobPath *gp);
#define gpFree gp_free
extern void gp_pc_free(TglobPath *gp);
extern void gp_NULL(TglobPath *gp);
extern bool gp_test_local_drive(TglobPath *gp);
//extern bool gp_test_ftp(TglobPath *gp);
//extern bool test_ftp(const char *fn);

extern void gpList_free(TgpList *gpList);

#define gptfu_noPasswd		1
#define gptfu_PasswdTxt		(1<<1)
#define gptfu_noName		(1<<2)
#define gptfu_PathBack		(1<<3)
#define gptfu_add_slash		(1<<4)
#define gptfu_del_slash		(1<<5)
#define gptfu_noRootArt		(1<<6)
#define gptfu_RootPath		(1<<7) /* ftp://url/tmp/text.txt -> ftp://url/ */
extern char *gptopc(TglobPath *gp, int fu);
extern char *pctogptopc(const char *s, int fu);
extern char *gptofnpc(TglobPath *gp, int fu, const char *fn);

extern TglobPath *gp_dup(TglobPath *gp_org);
extern void REGP2 gp_dup_gpdata(TglobPath *gp, TglobPath *gp_org);
extern void REGP2 gp_dup_gpdata_path(TglobPath *gp, TglobPath *gp_org);
extern void REGP2 gp_dup_gpdata_url(TglobPath *gp, TglobPath *gp_org);

extern const char *pcsetgp_url(TglobPath *gp, const char *url, int URL_lg);
extern const char *pcsetgp_path(TglobPath *gp, const char *path);
extern void REGP2 gp_set_name(TglobPath *gp, const char *name);

// ******************** File *******************************
#define IOfkt_CanWrite		1
#define IOfkt_CanInsert		(1<<1)
#define IOfkt_CanSave		(1<<2)
#define IOfkt_CanSaveAs		(1<<3)

// -----------------------------

extern bool TestIsFile(const char *fn);	//only local  else use file_chk_exist
extern bool TestIsPath(const char *fn);

extern void fd_write_va(int fd, const char *msg, ...);

/* ---------------------- File Msg ------------------------------------- */
struct TFileMsg;

extern void TTC_netError(TTC0 *ttc, TOnFileMsg ev, int ErrType, const char *msg, ...);

#define NetErrFlag_NoDev	0x01
#define NetErrFlag_NoConnect	NetErrFlag_NoDev
#define NetErrFlag_NoLoader	(1<<1)
#define NetErrFlag_NewLogin	(1<<2)

typedef struct {
  int sys_errno;
  char *err;
  const char *const_err;
  int errFlag;
  char *new_user;
  char *new_passwd;
  char *fn;
  const char *bef;
//  int pv_bef;
} TNetErrRet;

cHIDDEN charOS *path_to_local_OS_err(const char *path, TNetErrRet *ErrRet);

#define NetOpenMod_Read 1
#define NetOpenMod_Write 2

#define ioflag_NoPipe    1

#define AlgTNet0    AlgTTC \
  void *owner; \
  int omod; \
  bool ende, error, run, pv_flag0; \
  FileHandle iofd; \
  int ioflag; \
  TNetErrRet *ErrRet; \
  void (*OnFileMsg)(struct TFileMsg *fmsg); \
  uint64_t ref_us; \
  uint64_t next_us; \
  uint64_t next_us_add; \
  TDataIO_Info ioinfo; \
  TFileMsg fmsg; \
  TglobPath gp; 
//  word FREE_LOCK, FREE_LOCK_Destroy; 

#define AlgTNet    AlgTNet0 \
  bool (*set_gp)(struct TNet *self, TglobPath *gp); \
  void *v_nc0; \
  void *v_nc1; \
  int i_nc0;

#define AlgTNetFtp   AlgTNet0 \
  bool (*set_gp)(struct TNetFtp *self, TglobPath *gp); \
  struct TFtpClientCMD *ftpcmd; \
  struct TdData *shd; \
  int retCode;

typedef struct TNet {
  AlgTNet
} TNet;

typedef struct TNetFtp {
  AlgTNetFtp
} TNetFtp;

#define AlgTNetLineBuf  TLineBuf *lb; \
  bool (*open)(struct TNetLineBuf *self, int omod); \
  bool (*wait)(struct TNetLineBuf *self); \
  bool (*stop)(struct TNetLineBuf *self); \
  void (*OnDataAdd)(struct TNetLineBuf *self, bool end);

typedef struct TNetLineBuf {
  AlgTNet
  AlgTNetLineBuf
} TNetLineBuf;

typedef struct TNetFtpLineBuf {
  AlgTNetFtp
  AlgTNetLineBuf
} TNetFtpLineBuf;


/* ---------------------- File Stream ---------------------------------- */

extern void net_err_init(TNetErrRet *ErrRet);
extern void net_err_clr(TNetErrRet *ErrRet);
//extern const char *net_err_get(TNetErrRet *ErrRet);

#define AlgNetLoader  bool (*ChkType)(TglobPath *gp); \
  union { \
    TNetLineBuf *(*lb)(void *owner, TglobPath *gp); \
  } Create;

typedef struct ThNetLoader {
  AlgNetLoader
} ThNetLoader;

//extern void *GetNet_hLD_g(ThNetLoader **hLD, TglobPath *gp, int *retst);

typedef struct ThNetStreamLoader {
  AlgNetLoader     // bool (*ChkType)(TglobPath *gp);
  FILE *(*stream_open)(TglobPath *gp, const char *opentype);
} ThNetStreamLoader;

extern const ThNetStreamLoader hNetStreamLd_ftp;

extern void ThNetStreamLoader_Reg(int n, ...);

extern FILE *net_stream_open(const char *filename, const char *opentype);
extern FILE *net_stream_open_gp(TglobPath *gp, const char *opentype);
extern FILE *net_stream_open_gp_noERR(TglobPath *gp, const char *opentype);
extern FILE *file_stream_open(const char *filename, const char *opentype);
extern FILE *file_stream_open_gp(TglobPath *gp, const char *opentype);
extern bool net_stream_gp_test(TglobPath *gp);

/* ---------------------- File TLineBuf --------------------------------- */

typedef struct ThNetLineBufLoader {
  AlgNetLoader     // bool (*ChkType)(TglobPath *gp);
} ThNetLineBufLoader;

extern const ThNetLineBufLoader hNetLineBufLd_ftp;

extern void ThNetLineBufLoader_Reg(int n, ...);

extern TNetLineBuf *TNetLineBuf_Create(void *owner, TglobPath *gp);
extern bool net_LineBuf_gp_test(TglobPath *gp);

/* ---------------------- File Chmod ------------------------------------ */
extern bool file_chmod(const char *fn, TNetErrRet *ErrRet, int mode);
extern bool file_chmod_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode);
extern bool file_chown(const char *fn, TNetErrRet *ErrRet, int uid, int gid);
extern bool file_chown_gp(TglobPath *gp, TNetErrRet *ErrRet, int uid, int gid);
extern bool file_utime(const char *fn, TNetErrRet *ErrRet, time_t mtime);
extern bool file_utime_gp(TglobPath *gp, TNetErrRet *ErrRet, time_t mtime);
extern bool file_rename(const char *fn, TNetErrRet *ErrRet, const char *new_name);
extern bool file_rename_gp(TglobPath *gp, TNetErrRet *ErrRet, const char *new_name);
extern bool file_mv(const char *fn, TNetErrRet *ErrRet, const char *new_name);
extern bool file_mv_gp(TglobPath *gp, TNetErrRet *ErrRet, TglobPath *new_gp);
extern bool file_unlink(const char *fn, TNetErrRet *ErrRet);
extern bool file_unlink_gp(TglobPath *gp, TNetErrRet *ErrRet);
extern bool file_rmdir(const char *fn, TNetErrRet *ErrRet);
extern bool file_rmdir_gp(TglobPath *gp, TNetErrRet *ErrRet);
extern bool file_mkdir(const char *fn, TNetErrRet *ErrRet, int mode);
extern bool file_mkdir_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode);

#define MkDirError	False
#define MkDirOk		True
#define MkDirExists	2
extern int  file_mkdir_parents(const char *fn, TNetErrRet *ErrRet, int mode);

extern bool file_open_rd(const char *fn, TNetErrRet *ErrRet, FileHandle *fd);
extern bool file_open_rd_gp(TglobPath *gp, TNetErrRet *ErrRet, FileHandle *fd);
extern bool file_open_wr(const char *fn, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode);
extern bool file_open_wr_gp(TglobPath *gp, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode);
extern bool file_open_rdwr(const char *fn, TNetErrRet *ErrRet, FileHandle *fd);
extern bool file_open_rdwr_gp(TglobPath *gp, TNetErrRet *ErrRet, FileHandle *fd);
extern bool file_close(FileHandle fd, TNetErrRet *ErrRet);
extern bool file_read(FileHandle fd, TNetErrRet *ErrRet, void *buf, 
			size_t bufsize, size_t *retsize);
extern bool file_read_full(FileHandle fd, TNetErrRet *ErrRet, void *buf, 
			size_t bufsize, size_t *retsize);
extern bool file_write(FileHandle fd, TNetErrRet *ErrRet, const void *buf, 
			size_t bufsize, size_t *retsize);
extern bool file_seek(FileHandle fd, TNetErrRet *ErrRet, int64_t Offset);

extern bool file_open_rd_OS(const charOS *fn, TNetErrRet *ErrRet, FileHandle *fd);
extern bool file_open_wr_OS(const charOS *fn, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode);
extern bool file_open_rdwr_OS(const charOS *fn, TNetErrRet *ErrRet, FileHandle *fd);
extern bool file_mkdir_OS(const charOS *fn, TNetErrRet *ErrRet, int mode);
extern bool file_unlink_OS(const charOS *fn, TNetErrRet *ErrRet);
extern bool file_chmod_OS(const charOS *fn, TNetErrRet *ErrRet, int mode);
extern bool file_utime_OS(const charOS *fn, TNetErrRet *ErrRet, time_t atime, time_t mtime);

/* ---------------------- File appError --------------------------------- */
extern void net_err_app_4(TNetErrRet *ErrRet, const char *fn, const char *bef, int err_level);
extern void net_err_app(TNetErrRet *ErrRet, int err_level);

extern bool file_mkdir_ae(const char *fn, int mode, int err_level);
extern int  file_mkdir_parents_ae(const char *fn, int mode, int err_level);
extern bool file_open_rd_ae(const char *fn, FileHandle *fd, int err_level);
extern bool file_open_wr_ae(const char *fn, FileHandle *fd, mode_t mode, int err_level);
extern bool file_read_ae(FileHandle fd, void *buf, size_t bufsize, 
				size_t *retsize, const char *fn, int err_level);
extern bool file_write_ae(FileHandle fd, void *buf, size_t bufsize, 
				size_t *retsize, const char *fn, int err_level);
extern bool file_open_rdwr_ae(const char *fn, FileHandle *fd, int err_level);



typedef struct ThNetToolsLoader {
  AlgNetLoader     // bool (*ChkType)(TglobPath *gp);
  bool (*chmod)  (TglobPath *gp, TNetErrRet *ErrRet, int mode);
  bool (*utime)  (TglobPath *gp, TNetErrRet *ErrRet, time_t mtime);
  bool (*rename) (TglobPath *gp, TNetErrRet *ErrRet, const char *new_name);
  bool (*mv)     (TglobPath *gp, TNetErrRet *ErrRet, TglobPath *new_gp);
  bool (*unlink) (TglobPath *gp, TNetErrRet *ErrRet);
  bool (*rmdir)  (TglobPath *gp, TNetErrRet *ErrRet);
  bool (*mkdir)  (TglobPath *gp, TNetErrRet *ErrRet, int mode);
} ThNetToolsLoader;

extern const ThNetToolsLoader hNetToolsLd_ftp;

extern void NetLoaderReg_Tools(int n, ...);

extern bool net_chmod(const char *fn, TNetErrRet *ErrRet, int mode);
extern bool net_chmod_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode);
extern bool net_chmod_gp_test(TglobPath *gp);

extern bool net_utime(const char *fn, TNetErrRet *ErrRet, time_t mtime);
extern bool net_utime_gp(TglobPath *gp, TNetErrRet *ErrRet, time_t mtime);
extern bool net_utime_gp_test(TglobPath *gp);

/* net_rename(/tmp/test/xx, NULL, xx1) 
 /tmp/test/xx -> xx1 --> /tmp/test/xx1
*/
extern bool net_rename_gp(TglobPath *gp, TNetErrRet *ErrRet, const char *new_name);
extern bool net_rename(const char *fn, TNetErrRet *ErrRet, const char *new_name);
extern bool net_rename_gp_test(TglobPath *gp);

/* net_mv(/tmp/test/xx, NULL, /tmp/test/xx1) 
 /tmp/test/xx -> /tmp/test/xx1 --> /tmp/test/xx1
*/
extern bool net_mv_gp(TglobPath *gp, TNetErrRet *ErrRet, TglobPath *new_gp);
extern bool net_mv(const char *fn, TNetErrRet *ErrRet, const char *new_fn);
extern bool net_mv_gp_test(TglobPath *gp, TglobPath *new_gp);

extern bool net_unlink_gp(TglobPath *gp, TNetErrRet *ErrRet);
extern bool net_unlink(const char *fn, TNetErrRet *ErrRet);
extern bool net_unlink_gp_test(TglobPath *gp);

extern bool net_rmdir_gp(TglobPath *gp, TNetErrRet *ErrRet);
extern bool net_rmdir(const char *fn, TNetErrRet *ErrRet);
extern bool net_rmdir_gp_test(TglobPath *gp);

extern bool net_mkdir_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode);
extern bool net_mkdir(const char *fn, TNetErrRet *ErrRet, int mode);
extern bool net_mkdir_gp_test(TglobPath *gp);

extern bool net_file_new_va_gp(TglobPath *gp, TNetErrRet *ErrRet, const char *msg, ...);
extern bool net_file_new_va_gp_test(TglobPath *gp);

/* ---------------------- ftp ------------------------------------------- */
extern bool ftp_chk_RootArt(TglobPath *gp);
extern bool ftp_chk_RootArt_URL(TglobPath *gp);

/* ---------------------- Tstat64 --------------------------------------- */
typedef struct TExFileDat {
  byte type;
  byte fu;
  word lg;
  char dat[];
} TExFileDat;

// ****** Tstat64.st ******
#define st_Dir		1
#define st_Link		(1<<1)
//#define st_ReadOnly	(1<<2)
#define st_Hidden	(1<<3)
#define st_Reg		(1<<4)

#define st_has_ACL	(1<<8)
#define st_has_ACL_attr	(1<<9)

// ****** Tstat64.FileType ******
#define stft_Folder     1
#define stft_Link       2
#define stft_Chardev    3
#define stft_Blockdev   4
#define stft_FIFO       5
#define stft_Socket     6

#define stft_None       16
#define stft_Txt        17
#define stft_Exec       18
#define stft_SO         19
#define stft_ar         20
#define stft_gzip       21
#define stft_Shell      22
#define stft_Pascal     23
#define stft_C          24
#define stft_CXX        25
#define stft_HTML       26
#define stft_Binary     27
#define stft_Make       28
#define stft_Font_A     29
#define stft_Font_Fix   30
#define stft_desktop    31
#define stft_ASM        32
#define stft_IMAGE      33
#define stft_Win32Exe   34
#define stft_sh_perl    35
#define stft_PDF        36
#define stft_ALDapj     37
#define stft_audio      38
#define stft_video      39
#define stft_office     40
#define stft_XML        41
#define stft_rpm        42
#define stft_ListEnde   43

#define stft_ErrNoFile  0xfff0
#define stft_NoFound    0xffff

/* Tstat_mime.sub_mime */
#define mTxt_UTF8	1
#define mTxt_Unicode	2
#define mTxt_UnicodeBE	3
#define mTxt_troff	4

#define mTgz_bzip2	1
#define mTgz_zip	2

#define mImg_png	1


#define mAudio_x_wav	1
#define mAudio_mpeg	2

#define mVideo_x_msvideo	1
#define mVideo_mpeg		2
#define mVideo_mp4		3
#define mVideo_3gpp		4
#define mVideo_x_shockwave_flash 5
#define mVideo_x_flv		6
#define mVideo_quicktime	7

#define moffice_x_doc		0
#define moffice_oOO_writer	1
#define moffice_oOO_writer_v1	2
#define moffice_oOO_base_v1	3

#define mShell_spec	1

typedef struct {
  byte *data;
  size_t lg;
//  size_t BufSize;
  TglobPath *gp;
  const char *fname;    //only name (no path)
  int stft;
  int sub_mime;
} THeadInfo;



#ifdef _OS_WIN32
typedef struct Tstat64 {
  dev_t dev;			//(8B) Device
  mode_t mode;			//(4B) File mode
  short nlink;		//(4B) Link count
  short uid;			//(4B) User ID of the file's owner
  short gid;			//(4B) Group ID of the file's group
  dev_t rdev;			/* Device number, if device.  */
  int64_t size;		//(8B) Size of file, in bytes
  time_t atime;			//(4B) Time of last access
  time_t mtime;			//(4B) Time of last modification
  time_t ctime;  		//(4B) Time of last status change
  dword dwFileAttributes;
  void *mime_id;
  word sub_mime;
  short LinkSize;
  short st;
  word FileType;
  word nc_w;
  word StructSize;

  TExFileDat ExDat;
} Tstat64;
#else
typedef struct Tstat64 {
  dev_t dev;			//(8B) Device
//    unsigned int __pad1;
//    __ino_t __st_ino;			/* 32bit file serial number.	*/
  mode_t mode;			//(4B) File mode
  nlink_t nlink;		//(4B) Link count
  uid_t uid;			//(4B) User ID of the file's owner
  gid_t gid;			//(4B) Group ID of the file's group
  __dev_t rdev;			/* Device number, if device.  */
//    unsigned int __pad2;
  __off64_t size;		//(8B) Size of file, in bytes
//    __blksize_t st_blksize;		/* Optimal block size for I/O.  */
//    __blkcnt64_t st_blocks;		/* Number 512-byte blocks allocated. */
  time_t atime;			//(4B) Time of last access
//    unsigned long int st_atimensec;	/* Nscecs of last access.  */
  time_t mtime;			//(4B) Time of last modification
//    unsigned long int st_mtimensec;	/* Nsecs of last modification.  */
  time_t ctime;  		//(4B) Time of last status change
//    unsigned long int st_ctimensec;	/* Nsecs of last status change.  */
//    __ino64_t st_ino;			/* File serial number.		*/
  void *mime_id;
  word sub_mime;
  word LinkSize;
  word st;
  word FileType;	//stft_Folder,stft_Chardev,...
  word nc_w;
  word StructSize;

  TExFileDat ExDat;
} Tstat64;
#endif

/* ---------------------- File Stat ---------------------------------- */
#define fsFollowLink		1
#define fsNoFileName		(1<<1)
#define fsNoLinkName		(1<<2)
#define fsGetFileType		(1<<3)
#define fsGetFileTypeHead	(1<<4)
#define fsGetFileMime		(1<<5)
#define fsLinkDirekt		(1<<6)
//#define fsOnlyPath		(1<<7)
#define fsAdd_UID_GID_mem	(1<<8)
#define fsAdd_Type_mem		(1<<9)
#define fsSingleReScan		(1<<14)  /* Only FTP (use STAT) */
#define fsForceReScan		(1<<15)  /* Only FTP */

#define fsGetFileAll	fsFollowLink|fsGetFileType|fsGetFileTypeHead|fsGetFileMime
#define fsGetFileAll_UGT fsFollowLink|fsGetFileType|fsGetFileTypeHead|fsGetFileMime|fsAdd_UID_GID_mem|fsAdd_Type_mem

#define fs_UID_GID_MaxMem	20
#define fs_Type_MaxMem		36

extern Tstat64 *file_stat_gp(TglobPath *gp, int fsfu);
extern Tstat64 *file_stat_fn(const char *FileName, int fu);

extern Tstat64 *file_stat_gp_net(TglobPath *gp, TNetErrRet *ErrRet, int fsfu);

extern Tstat64 *file_stat_A(const char *pathfn, const char *fname, int fu);

extern Tstat64 *tstat_dup(Tstat64 *st);

/* ---------------- in unistd.h ------------------------------
mode	R_OK	4		Test for read permission
mode	W_OK	2		Test for write permission
mode	X_OK	1		Test for execute permission
mode	F_OK	0		Test for existence
*/
extern bool file_chk_access_st(Tstat64 *st, int mode);

/* ---------------------- Net Stat ----------------------------------- */
typedef struct ThNetStatLoader {
  AlgNetLoader     // bool (*ChkType)(TglobPath *gp);
  Tstat64 *(*stat)(TglobPath *gp, TNetErrRet *ErrRet, int fu);
  bool (*chk_exist)(TglobPath *gp, int findAtr);
  bool (*chk_open)(TglobPath *gp, TNetErrRet *ErrRet, int fu);
} ThNetStatLoader;

extern const ThNetStatLoader hNetStatLd_ftp;

extern void ThNetStatLoader_Reg(int n, ...);

extern Tstat64 *net_stat_gp(TglobPath *gp, TNetErrRet *ErrRet, int fsfu);	//file:// ftp://
extern Tstat64 *net_stat(const char *fn, TNetErrRet *ErrRet, int fsfu);	//file:// ftp://

extern bool net_chk_exist_gp(TglobPath *gp, int findAtr);
extern bool net_chk_exist(const char *fn, int findAtr);

extern bool net_chk_open(const char *fn, TNetErrRet *ErrRet, int fu);
extern bool net_chk_open_gp(TglobPath *gp, TNetErrRet *ErrRet, int fu);

typedef void (*TNetCache_DelEv)(TglobPath *gp, void *obj);

/* ---------------------- Tstat64 ExFileDat -------------------------- */
#define exfdt_FileName	1
#define exfdt_LinkName	2
#define exfdt_owner	3
#define exfdt_group	4
#define exfdt_type	5
#define exfdt_user	16

#define exfdf_lastdat	1
extern TExFileDat *Get_TExFileDat(Tstat64 *st, int exfdt);
extern const char *Get_ExFileDat_pc(Tstat64 *st, int exfdt);

/* ---------------------- Pattern ------------------------------------ */
typedef struct TPattern{
  AlgTTC0
  char **ppa;
  word *ft_arr;
  int findAtr;
  int fu_case; //default FU_CASE_8859_1
  bool (*match_pattern)(struct TPattern *patt, const char *fn, Tstat64 *st);
  word ft_buf[8];
} TPattern;

// TFileIO_DefLoad.fu  bit 0..3  EOL funktion
#define fio_EOL_Mask		0x000f
#define fio_Res_Mask		0x007f
#define fio_EOLdefault		0
#define fio_EOLunix		1
#define fio_EOLwindos		2
#define fio_EOL_0		3
#define fio_EOLnone		4
#define fio_EOF			(1<<4)
#define fio_EOL_mix		(1<<7)
// TFileIO_DefLoad.fu  bit 8..15  read status
// UTF8 =      BOM = 0xbf,0xbb,0xef
// Unicode =   BOM = 0xfe,0xff    little-endian
// UnicodeBE = BOM = 0xff,0xfe    big-endian
#define fio_txtTyp_Mask		0x300
#define fio_txtTyp_Mask_BOM	0x700
#define fio_txtTyp_ANSI		0x000
#define fio_txtTyp_UTF8		0x100
#define fio_txtTyp_Unicode	0x200
#define fio_txtTyp_UnicodeBE	0x300
#define fio_has_BOM		0x400

#define fio_has_0x00_char	(1<<11)
#define fio_has_0x80_char	(1<<12)
#define fio_has_UTF8_char	(1<<13)
#define fio_err_UTF8_char	(1<<14)
#define fio_has_gt_0xff_char	(1<<15)

// TFileIO_DefLoad.fu  bit 16..23  read(write) funktion
#define fio_Chk_BOM		(1<<16)
//#define fio_SrcIsASCII		(1<<17)
#define fio_force_ASCII		(1<<18)
#define fio_add_EOL_0		(1<<19)

// TFileIO_DefLoad.fu  bit 24..31  Satus for TLines,...
//#define fio_pcl_Free		(1<<24)
#define fio_Strg_C		(1<<25)

#define fio_GetLgPlusEOL	(1<<28)


/* ------ gp_Path.c ------ */
extern void gp_NameToPath(TglobPath *gp);  // /tmp/ xxx --> /tmp/xxx/ NULL
extern bool gp_PathToName(TglobPath *gp);  // / = result False
extern void gp_NameToPath_AddName(TglobPath *gp, const char *name);

/* ------ gp_CryptMini16c_enc.c ------ */
extern void gp_CryptMini16c_enc(TglobPath *gp);

/* filename no passwd */
extern int gptfu_DisplayFileStr;    //gptfu_noPasswd|gptfu_noRootArt
extern char REGP2 *get_DisplayFileStr(const char *fn, TglobPath *gp);
extern char REGP2 *get_DisplayPathStr(const char *fn, TglobPath *gp);

extern bool file_chk_exist_gp(TglobPath *gp, int findAtr);
extern bool file_chk_exist(const char *fn, int findAtr);


/* ------ file_str_xxx.c ------ */
#define fstrSize_space 0x01
/* fu = 0    12.3kB
   fu = 1    12.3 kB */
extern char *file_str_size(uint64_t n, int fu);
extern int  file_str_size_pc(char *pc, uint64_t n, int fu);
extern void file_str_size_ppc(char **ppc, uint64_t n, int fu);

// DateFormat 0=d.m.Y 1=m.d.Y H:M 2=Y.m.d H:M
extern int GlobDateFormat;
// DateFormat +16 = + H:M  +32 = + H:M:S
#define DateFormat_pHM  0x10
#define DateFormat_pHMS 0x20
extern void REGP3 file_str_time_ppc_fu(char **ppc, time_t mtime, int df);

extern void file_str_time_ppc(char **ppc, time_t mtime);
extern void file_str_time_HM_ppc(char **ppc, time_t mtime);
extern void file_str_time_HMS_ppc(char **ppc, time_t mtime);

extern int  file_str_time_pc(char *pc, time_t mtime);
extern int  file_str_time_HM_pc(char *pc, time_t mtime);
extern int  file_str_time_HMS_pc(char *pc, time_t mtime);

extern char *file_str_time(time_t mtime);
extern char *file_str_time_HM(time_t mtime);
extern char *file_str_time_HMS(time_t mtime);

#define StrModT_Art	1
#define StrModT_PlusACL	(1<<1)
extern void file_str_mode_ppc(char **ppc, Tstat64 *st, int StrModT);
extern int  file_str_mode_pc(char *pc, Tstat64 *st, int StrModT);
extern char *file_str_mode(Tstat64 *st, int StrModT);


/* findAtr */
#define findALL 		0
#define findDIR 		1
#define findCHR 		(1<<1)
#define findBLK 		(1<<2)
#define findREG 		(1<<3)
#define findFIFO		(1<<4)
#define findLNK 		(1<<5)
#define findSOCK		(1<<6)

#define findNoHidden		(1<<8)
#define findIsRootPath		(1<<9)
#define findFollowLink		(1<<10)
#define findDIR_NoPattern	(1<<11)
/* file_chk_exist_gp */
#define findGP_fsOnlyPath	(1<<16)

extern bool TPattern_Match(TPattern *patt, const char *s, Tstat64 *st);
extern bool File_stat_Match(int findAtr, TPattern *patt, const char *s, Tstat64 *st);
extern bool file_stat_mode_match(int findAtr, int st_mode);
#ifdef _OS_WIN32
extern char *NT4_GetLongPathName(char *spn);
#endif

extern TPattern *TPattern_Create(int findAtr, const char *pattern, word *FileType_arr);
extern void TPattern_set(TPattern **ppatt, const char *pattern, word *FileType_arr);
//extern void TPattern_set_pattern(TPattern *self, const char *pattern);
//extern void TPattern_set_FileType(TPattern *self, word *FileType_arr);

extern const char *GetArtExt_pc(Tstat64 *st);

extern word get_file_stft_ext(const char *fn, word *SubId, dword *MinSize);
//extern bool ChkHeadIsBash(byte *line, int lg);

typedef struct {
  void (*get_head_mime)(THeadInfo *sif);
  void (*get_head_mime_mem)(byte *data, size_t size, TglobPath *gp, word *stft, word *sub_mime);
  bool (*get_head_mime_gp_st_fn)(Tstat64 *st, const char *pathfn, const char *fname);
} ThGetFileHeadInfo;

extern ThGetFileHeadInfo hGetFileHeadInfo_buildin;
extern ThGetFileHeadInfo *hGetFileHeadInfo;


typedef struct {
  int type;
  void *id;
  char *path;
  char *fn;
  char *exec;
  char *icon;
  char *mimetype;
  char *pattern;
} TMimeDat;

#define MimeOnChange_all	1
#define MimeOnChange_item	2
#define MimeOnChange_rescan_id	3
typedef struct TMime {
  AlgTTC
  void (*stat_set_mime_id)(struct TMime *self, Tstat64 *st, const char *fn);
  struct TPic *(*stat_mime_getpic16)(struct TMime *self, Tstat64 *st);
  struct TPic *(*stat_mime_getpic32)(struct TMime *self, Tstat64 *st);
  void (*OnChange)(struct TMime *self, int bef);
} TMime;

extern TMime *TMime_Create();
extern TMime *TMime_X_Create();
extern TMime *app_mime;

extern const char *TMime_get_fn_mime_overwrite(TMime *self);

extern char *stat_get_bi_mimetype(Tstat64 *st);

// char *mimetype2app(char* mimetype);
extern bool mimetype2app_getdat(const char* mimetype, TMimeDat *md);
extern bool mimetype2app_chk_mime(const char *mime, const char *mt);
extern bool TMime_type2app_getdat(TMime *self, const char* mimetype, TMimeDat *md);

extern bool TMime_type2lnk_getdat(TMime *self, const char* mimetype, TMimeDat *md);
extern bool mimetype2lnk_getdat(const char* mimetype, TMimeDat *md);

extern char *TMime_stat_get_mimetype(TMime *self, Tstat64 *st, const char *fn);
extern char *stat_get_mimetype(Tstat64 *st, const char *fn);

extern char *libmagic_get_mime(const char *fn);

extern struct TMimeObj *TMimeEnum(struct TMimeObj *mo);
extern struct TMimeObj *TMimeObjEnum(struct TMimeObj *mo);
extern char *TMimeObjPath(struct TMimeObj *mo);
extern struct TMimeDat_app *TMimeObjAppEnum(struct TMimeObj *mo, struct TMimeDat_app *mda);
extern const char *TMimeDatAppName(struct TMimeDat_app *mda);



extern char *path_rel2abs(const char *rel, const char *basepath);
extern char *path_abs2rel(const char *path, const char *basepath);

/* path_absolut = path_rel2abs -> ppc_dotdot_korr -> link korr 
  result NULL == Path not exists */
extern char *path_absolut(const char *path);

// ******************* Misc ***********************

extern char *get_pid_cmd_name(int pid);

typedef struct {
  int   pid;		//process id
  char  *tcomm;		//filename of the executable
  char  state[4];	//state (R is running, S is sleeping, D is sleeping in an
			// uninterruptible wait, Z is zombie, T is traced or stopped)
  int   ppid;		//process id of the parent process
  int   pgrp;		//pgrp of the process
  int   sid;		//session id
  int   tty_nr;		//tty the process uses
  int   tty_pgrp;	//pgrp of the tty
  int   flags;		//task flags
  int   min_flt;	//number of minor faults
  int   cmin_flt;	//number of minor faults with child's
  int   maj_flt;	//number of major faults
  int   cmaj_flt;	//number of major faults with child's
  dword utime;		//user mode jiffies
  dword stime;		//kernel mode jiffies
  dword cutime;		//user mode jiffies with child's
  dword cstime;		//kernel mode jiffies with child's
  int   priority;	//priority level
  int   nice;		//nice level
  int   num_threads;	//number of threads
  int   it_real_value;	//(obsolete, always 0)
  dword start_time;	//time the process started after system boot
  dword vsize;		//virtual memory size
  dword rss;		//resident set memory size
} Tpid_stat;

#define PIDSTAT_NO_NAME 1

extern bool get_pid_stat(Tpid_stat *pid_stat, const char *fn, int fkt);
extern void pid_stat_clr(Tpid_stat *pid_stat);

// ******************* TpcIni ***********************
typedef struct TGruppeDat {
  struct TGruppeDat *next;
  char *name;
  int lgname;
  char *org;
  char *dat;
} TGruppeDat;

typedef struct TGruppe {
  struct TGruppe *next;
  char *name;
  int lgname;
  struct TGruppeDat *Dat;
} TGruppe;

typedef struct TpcIni {
  AlgTTC
  struct TGruppe *Gruppe0;
  char *FileName;
  struct TGruppe *lastGp;
  struct TGruppeDat *lastGpD;
  struct TGruppe **npGru;
  struct TGruppeDat **npGruD;
  bool cCase;
  bool Changed;
  bool Exist;
  bool BOM_UTF8;
} TpcIni;

extern TpcIni *TpcIni_Create(const char *FileName);
extern TpcIni *TpcIni_Create_struct();
extern void TpcIni_load(TpcIni *self, FILE *fd);

extern void TpcIni_ChangeGruppeDat(TpcIni *ti, int lg, char *pc, TGruppeDat *Gpd);
extern void TpcIni_AddGruppeDat(TpcIni *ti, int lg, char *pc);
extern void TpcIni_AddGruppe(TpcIni *ti, int lg, const char *pc);
//extern void TpcIni_AddItem(TpcIni *ti, char *pc, int lg, bool MaxLg);
extern TGruppe *TpcIni_FindGruppe(TpcIni *ti, const char *Gruppe);
extern TGruppeDat *TpcIni_FindGruppeDat(TpcIni *ti, const char *Gruppe, const char *Funkt);

extern void TpcIni_Save(TpcIni *self);
extern void TpcIni_SaveAndFree(TpcIni *self);

extern const char *TpcIni_RD(TpcIni *ti, const char *Gruppe, const char *Funkt);
extern const char *TpcIni_RD_pc(TpcIni *ti, const char *Gruppe, 
				const char *Funkt, const char *Default);
//#define TpcIni_RDstr TpcIni_RD_pc
extern char *TpcIni_RDstrdup(TpcIni *ti, const char *Gruppe, const char *Funkt, const char *Default);
extern void TpcIni_WR(TpcIni *self, const char *Gruppe, const char *Funkt, const char *Dat);
//#define TpcIni_WRstr TpcIni_WR

extern int TpcIni_RD_int(TpcIni *ti, const char *Gruppe, const char *Funkt, 
				int Default);
extern int TpcIni_RD_int_mm(TpcIni *ti, const char *Gruppe, const char *Funkt, 
				int Default, int min, int max);

extern float  TpcIni_RD_float(TpcIni *ti, const char *Gruppe, const char *Funkt, float Default);
extern double TpcIni_RD_double(TpcIni *ti, const char *Gruppe, const char *Funkt, double Default);
extern bool   TpcIni_RD_boolean(TpcIni *ti, const char *Gruppe, const char *Funkt, bool Default);

extern void TpcIni_WR_int(TpcIni *ti, const char *Gruppe, const char *Funkt, int Dat);
extern void TpcIni_WR_float(TpcIni *ti, const char *Gruppe, const char *Funkt, float Dat, int nk);
extern void TpcIni_WR_double(TpcIni *ti, const char *Gruppe, const char *Funkt, double Dat, int nk);
extern void TpcIni_WR_boolean(TpcIni *ti, const char *Gruppe, const char *Funkt, bool Dat);

extern void TpcIni_DelGruppeDat(TpcIni *ti, const char *Gruppe);

#define pcIni_PushPath 1
extern void TpcIni_push_stack(TpcIni *self, const char *Gruppe, const char *Funkt, 
				const char *Data, int Max, int Mode);

typedef void (*TpcIniOnGetData)(TpcIni *self, const char *Data, void *owner);

extern void TpcIni_get_stack(TpcIni *self, const char *Gruppe, const char *Funkt, 
				TpcIniOnGetData Callback, void *owner, int Max, int Mode);



/* ************* dotdesktop.c ************* */
#define ddkt_Application	1
#define ddkt_FSDevice		2
#define ddkt_Directory		3
#define ddkt_ALDeskBuildin	4
#define ddkt_Actions		4
#define ddkt_Mime		5

extern int dotdesktop_get_ddkt(const char *dType);

typedef struct {
  const char *type;
  const char *name;
  const char *icon;
  const char *exec;
  const char *tryexec;
  int ddkt;
  bool StartupNotify;
  bool Terminal;
} TDesktopHead;

extern bool dotdesktop_get_head(TpcIni *ini, TDesktopHead *ddkh);

// ******************* lines ***********************

/*********************************************************************
* OnLineChange only by TLines_Add/Ins, TLines_Paste, TLines_DelSel
*
*********************************************************************/

typedef struct {
  AlgTlgs
  char *c;
  word fu;		//EOL_mk
  byte hl_mk;		//used in highlight (highl_*.c)
  byte nc3;
} TColS;

typedef struct {
  int items;
  TColS *cs[];
} TColS_arr;

// ****** TFileS.fu ******
#define fls_Select	1

typedef struct {
  AlgTlgs
  Tstat64 *st;
  void *obj;
  word fu;
  byte hl_mk;
  byte nc3;
} TFileS;

typedef struct {
  int items;
  TFileS *cs[];
} TFileS_arr;

#define csType_TColS	0
#define csType_TFileS	1

#define AlgTLines  AlgTTC \
  int Count; \
  void *Owner; \
  int csType; \
  void (*OnLineChange)(struct TCSelf *self, int idx); \
  void (*OnLineDel)(struct TLines*, TColS*, int); \
  void (*OnLineDelc)(struct TLines*, TColS*, int); \
  void (*OnCount)(struct TLines*); \
  TCompareFunction CompareFunction; \
  char *FileName; \
  int FindXa; \
  int FindYa; \
  int FindXe; \
  int FindYe; \
  int DefaultEOL; \
  int load_fio_fu; \
  int org_load_fio_fu; \
  int stat_fu; \
  Tstat64 *st; \
  void *pv_io; \
  int sort_fu; \
  bool EvStop; \
  bool FindNoCase; \
  bool FindBack; \
  bool Changed;

#define TCSelf TLines
typedef struct TLines {
  AlgTLines
  TOnFileMsg OnFileMsg;
  TColS_arr *lines;
} TLines;
#undef TCSelf

#define sortfu_Case	1

extern TLines *TLines_Create();
extern bool TLines__Destroy(TLines *self);
extern int REGP1 TLines_GetLineCount(TLines *self);

extern void TLines_Clear(TLines *self);
extern void TLines_Clear_c(TLines *tls);	//clear TColS->c
extern void TLines_Sort(TLines *tls);
extern void TLines_ResLinec(TLines *tls, TColS * pcs, int idx);
//extern void TLines_ResLine(TLines *tls, int idx);
extern void TLines_Del(TLines *tls, int idx);
extern void TLines_DelN(TLines *tls, int idx, int n);
extern void  *TLines_Add_(TLines *self);
extern TColS *TLines_nAdd(TLines *self, const char *s, int lg);
extern TColS *TLines_Add(TLines *self, const char *s);
extern TColS *TLines_Add_lgs(TLines *self, Tlgs *lgs, int eol_fu);
extern TColS *TLines_nIns(TLines *self, int idx, const char *s, int lg);
extern TColS *TLines_Ins(TLines *self, int idx, const char *s);
extern void TLines_Paste_lgs(TLines *tls, int *lineNr, int *charPos, Tlgs *lgs, int maxLineLen);
extern void TLines_Paste(TLines *tls, int *lineNr, int *charPos, const char *s);

extern void TLines_ChkMax_xy(TLines *tls, int *xa, int *ya, int *xe, int *ye);
extern void TLines_GetSel_Tlgs(TLines *tls, Tlgs *lgs, int xa, int ya, int xe, int ye, int fu);
extern void TLines_DelSel(TLines *tls, int *xa, int *ya, int xe, int ye);
extern dword TLines_GetSel_Size(TLines *tls, int xa, int ya, int xe, int ye, int fu);
extern dword TLines_GetSize(TLines *tls, int fu);

extern bool TLines_FindS(TLines *tls, const char *s);

extern const char *TLines_Get_pc(TLines *self, int idx);
extern TColS *TLines_Get_pcs(TLines *self, int idx);

extern int REGP2 GetDefaultEOL_fu(int DefaultEOL, int OrgEOL);
extern const char *GetDefaultEOL_s(int DefaultEOL, int OrgEOL);
extern int GetDefaultEOL_lg(int DefaultEOL, int OrgEOL);


extern bool TLines_LoadFromFile(TLines *tls, const char *FileName);
extern bool TLines_SaveToFile(TLines *tls, const char *FileName);

extern void TLines_DoCount(TLines *tls);
extern void TLines_DoLineChange(TLines *tls, int idx);

// *********************** undo - redo *****************************************
#define ur_InsKey	1
#define ur_InsStr	(1<<1)
#define ur_DelSel	(1<<2)
#define ur_DelChar	(1<<4)
#define ur_DelCharA	(1<<5)
#define ur_Nop		(1<<6)
#define ur_plugin	(1<<7)
#define ur_OverKey	(1<<8)
#define ur_keycode	(1<<12)
#define ur_Undo		(1<<14)
#define ur_Redo		(1<<15)
#define ur_DupTlgs	(1<<16)
#define ur_LgsInKeyS	(1<<17)   //lgs=NULL c[0]=lg  c[1]..c[15]=data

typedef union {
    char c[16];		//c[8]=old char
    const char cc[16];
    byte b[16];
    word w[8];
    dword d[4];
} TKeyS;

typedef struct {
  int fu;
  TKeyS KeyS;
  Tlgs lgs;
  int xa, ya, xe, ye;
} TUnRedoLine;

typedef struct TUnRedo {
  AlgTTC0
  int count;
  int idx;
  int max_items;
  TUnRedoLine lines[];
} TUnRedo;

extern TUnRedo *TUnRedo_Create(int max_items);
extern void TUnRedo_AddLine(TUnRedo *ur, int fu, TKeyS *KeyS, Tlgs *lgs, int xa, int ya, int xe, int ye);
extern void TUnRedo_Clr(TUnRedo *ur);
extern TUnRedoLine *TUnRedo_GetUndo(TUnRedo *ur);
extern TUnRedoLine *TUnRedo_GetRedo(TUnRedo *ur);
extern void TUnRedo_Undo_Old(TUnRedo *ur);
extern void TUnRedo_Redo_Old(TUnRedo *ur);
extern void TUnRedo_GetStat(TUnRedo *ur, bool *undo, bool *redo);

// *********************** TcZ ****************************************
extern const Tss Default_TzC;
extern int REGP3 GetTzC(const Tss *ss, dword c);


// *********************** Passwd ***********************************
extern void NetPasswdSave(TglobPath *gp);
extern void NetPasswdSetup_gp(TglobPath *gp);

extern char *get_ftpURL(TglobPath *gp);

// *********************** pwd grp ***********************************

#ifdef _OS_WIN32
#define UGL_NAME_LENGTH 64
typedef struct TUserGroupList {
  struct TUserGroupList *next;
  dword lg;
  byte sid[20];	//XP lg=12
  char Name[UGL_NAME_LENGTH];
  char DomainName[UGL_NAME_LENGTH];
} TUserGroupList;

extern TUserGroupList *ul_getby_uid(PSID *puid);
extern TUserGroupList *gl_getby_gid(PSID *pgid);

#else /* _OS_WIN32 */
typedef struct TUserList {
  struct TUserList *next;
  int uid;
  char name[fs_UID_GID_MaxMem];
} TUserList;

typedef struct TGroupList {
  struct TGroupList *next;
  int gid;
  char name[fs_UID_GID_MaxMem];
} TGroupList;

extern TUserList *ul_getby_uid(int uid);
extern TUserList *ul_getby_name(const char *name);
extern TUserList *ul_get_next(TUserList *ul);

extern TGroupList *gl_getby_gid(int gid);
extern TGroupList *gl_getby_name(const char *name);
extern TGroupList *gl_get_next(TGroupList *gl);
#endif /* ! _OS_WIN32 */

extern void ul_gl_free();
extern void ul_gl_init();


// *********************** notify mount ***********************************
#define nymobef_umount 0
#define nymobef_mount  1
typedef struct {
  int bef;
  void *owner;
  char *mnt_fsname;		/* Device or server for filesystem.  */
  char *mnt_dir;		/* Directory mounted on.  */
  char *mnt_type;		/* Type of filesystem: ufs, nfs, etc.  */
} TNotifyMount_data;

typedef struct TNotifyMount_wd *Tnymo_wd;
typedef void (*ThNotifyMountEv)(Tnymo_wd wd, TNotifyMount_data *ny_data);

#define nymofu_replay 1

extern Tnymo_wd TNotifyMount_Create(void *owner, ThNotifyMountEv hnymoEv);
extern void TNotifyMount_Replay(Tnymo_wd wd);
extern void TNotifyMount_ReScan();
extern bool TNotifyMount_Stop(bool Stop);

extern bool TNotifyMount_ChkMountPointFree(const char *fn);
extern char *TNotifyMount_GetMountPoint(const char *dev);

#endif /* __SYSBASEIO_H__ */
