/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef	__TDIRLIST_H__
#define	__TDIRLIST_H__    1

#ifndef	__FILEMSG_H__
#include "filemsg.h"
#endif

#define dlbef_ExDat		0x20

#define dlbef_New_st		1
#define dlbef_Add		2
#define dlbef_Del		3
#define dlbef_Change		4
#define dlbef_WrClose		5
#define dlbef_DelSelf		6

#define dlbef_NeedSort		13
#define dlbef_Sync		14
#define dlbef_ListEnd		15
#define dlbef_Clr		24
#define dlbef_ClrEnd		25
#define dlbef_Open		26
#define dlbef_Close		28
#define dlbef_Start		30
#define dlbef_Destroy		31

#define dlbefex			dlbef_ExDat
#define dlbef_BefMask		0x3f

#define dlbef_ScanRun		(1<<8)
#define dlbef_RescanRun		(1<<9)
#define dlbef_NotifyRun		(1<<10)


#define dlobef_Rescan		(1<<9)
#define dlobef_Notify		(1<<10)
#define dlobef_level_mask	0xf000
#define dlobef_level_0		0x1000
#define dlobef_level_1		0x2000
#define dlobef_level_2		0x3000
#define dlobef_level_2_Rescan	(dlobef_Rescan | dlobef_level_2)
//dlobef_level_0 only scanfu = 0
//dlobef_level_1 only scanfu = fsFollowLink
//dlobef_level_2 only scanfu = fsFollowLink|fsGetFileType

#define dlmemfu_gpNotFree	1
#define dlmemfu_LOCK		(1<<1)
#define dlmemfu_goDestroy	(1<<2)

#define dflnymk_DelMk		1
#define dflnymk_Rescan		(1<<1)
#define dflnymk_Del		(1<<16)
#define dflnymk_Change		(1<<17)
#define dflnymk_Add		(1<<18)
#define dflnymk_WrClose		(1<<19)

#define chstat_Add	1
#define chstat_Del	(1<<2)
#define chstat_mtime	(1<<4)
#define chstat_size	(1<<5)
#define chstat_mode	(1<<6)
#define chstat_nlink	(1<<7)
#define chstat_uid	(1<<8)
#define chstat_gid	(1<<9)

#define chstat_FileType	(1<<12)
#define chstat_mime_id	(1<<13)
#define chstat_sub_mime	(1<<14)

#ifdef _OS_WIN32
#define chstat_dwFileAttributes	(1<<16)
#endif

// ****** TDirFileList.fu ******
#define dfl_Select	1

typedef struct TDirFileList {
  struct TDirFileList *next;
  const char *fn;
  int	tag;
  void	*tagp;
  int	nymk;
  int	fu;
  void	*obj;
  dword chstat;
  Tstat64 st[0];
} TDirFileList;

// ______________________ TDirListLoader _________________________________
typedef struct TDirListLoader {
  AlgTTC;
  void *owner;
  bool (*open)(struct TDirListLoader *self);
  int  (*load)(struct TDirListLoader *self);
  void (*close)(struct TDirListLoader *self);
  TOnFileMsg OnFileMsg;
  int obef;
  int memfu;	//b0=bp not free
  int scanfu;   //fsFollowLink|fsGetFileType|fsGetFileTypeHead
  TDirFileList **ppdfl;
  TDirFileList **add_ppdfl;
  TDirFileList *rescan_dfl;
  TglobPath *gp;
  TPattern *patt;
  dword appFu_OR;
  TNetCache_DelEv NetCache_DelEv;
} TDirListLoader;

//#define dlnymk_RescanUsed	1
#define dlnymk_MsgOpen		(1<<8)

typedef struct TDirListNotify {
  AlgTTC;
  struct TDirList *dl;
  void (*start)(struct TDirListNotify *self);
  void (*ReScan)(struct TDirListNotify *self);
  TOnFileMsg OnFileMsg;
  TPattern *patt;
  int nymk;
  bool UseRescan;
#ifdef TEST_NOTIFY_TIME
  bool en_test_notify_us;
#endif
} TDirListNotify;

//extern void TDirListINotify__AppDestroy();

typedef struct ThDirListLoader {
  bool (*ChkType)(TglobPath *gp);
  TDirListLoader *(*create)(void *owner, TDirFileList **ppdfl, TOnFileMsg OnFileMsg);
  dword appFu_OR;
} ThDirListLoader;

typedef struct ThDirListNotify {
  bool (*ChkType)(TglobPath *gp);
  TDirListNotify *(*create)(struct TDirList *dl, TOnFileMsg OnFileMsg);
  dword nc;
} ThDirListNotify;

#define dlrunstat_NotOpen	-1
#define dlrunstat_Create	0
#define dlrunstat_Open		1
#define dlrunstat_ScanRun	2
#define dlrunstat_ScanEnd	3
#define dlrunstat_ReScanRun	4
#define dlrunstat_ReScanEnd	5
#define dlrunstat_NotifyRun	6

typedef struct TDirList {
  AlgTTC;
  int runstat;
  TOnFileMsg OnFileMsg;
  TDirFileList *dfl;
  TDirListLoader *ldr;
  TDirListNotify *dlny;
  int bef;   //dlbef_OpenRun|dlbef_RescanRun|dlbef_NotifyRun
  int memfu;	//b0=bp not free
  TglobPath gp;
  struct TDirListObjNy *ObjNy;	//use in TDirListObjNy
  struct TDirList *next;	//use in TDirListObjNy
  TDirFileList *last_dfl;	//private
  TPattern *patt;
  dword appFu_OR;
  TNetCache_DelEv NetCache_DelEv;
} TDirList;


#define dlObjNyfu_pattNotFree	(1<<1)


typedef struct TDirListObjNy {
  AlgTTC
  int fu;
  int count;
  int findAtr;
  dword chstat_Filter;
  TPattern *patt;
  TOnFileMsg OnFileMsg;
  TDirFileList *dfl;
  void *owner;
  struct TDirListObjNy *next;
  TDirList *dl;
  TDirFileList **ppdfl;		//private
  TDirFileList *last_get_dfl;	//private
  dword chstat;
  struct TDirListObjNy **ppObjNy;
  int *pTempWait;
} TDirListObjNy;

extern const ThDirListLoader hdirlist_ld_File;
extern const ThDirListLoader hdirlist_ld_File_pth;
extern const ThDirListLoader hdirlist_ld_ftp;

extern const ThDirListNotify hdirlist_ny_File;
extern const ThDirListNotify hdirlist_nyi_File;
extern const ThDirListNotify hdirlist_ny_Ftp;

extern void ThDirList_Reg(int n, ...);
extern void ThDirList_ny_Reg(int n, ...);

extern TDirList *TDirList_Create(const char *fn);
extern TDirList *TDirList_Create_name(const char *fn);	 //path = path+name

extern TDirList *TDirList_Create_gp(TglobPath *gp);
extern TDirList *TDirList_Create_gp_name(TglobPath *gp); //path = path+name
extern bool TDirList_Open(TDirList *self, int bef);
extern bool TDirList_Open_pp(TDirList *self, int bef, ThDirListLoader **hddl, ThDirListNotify **hdlny);
extern void TDirList_Open_Wait(TDirList *self, int runstat);
extern void TDirFileList_Sort(TDirFileList **ppdfl, int fu);
extern void TDirList_ReScan(TDirList *self);
extern void TDirList_ReScanMime(TDirList *self);

extern TDirListObjNy *TDirListObjNy_Create(void *owner);

extern bool TDirListObjNy_Create_gp(TDirListObjNy **ppObjNy, void *owner, 
					TglobPath *gp, TOnFileMsg OnFileMsg);

extern bool TDirListObjNy_Net_Create_gp(TDirListObjNy **ppObjNy, void *owner, 
					TglobPath *gp, TOnFileMsg OnFileMsg);

extern bool TDirListObjNy_Open(TDirListObjNy *self, const char *fn);
extern bool TDirListObjNy_Open_gp(TDirListObjNy *self, TglobPath *gp);
extern TDirList *TDirListObjNy_Get_dl_gp(TglobPath *gp);
extern void TDirListObjNy_Open_Wait(TDirListObjNy *self);
extern void TDirListObjNy_ReScan(TDirListObjNy *self);
extern void TDirListObjNy_ReScanMime();

#define dlsortFu_DOWN	1
#define dlsortFu_Name	0
#define dlsortFu_Size	2
#define dlsortFu_Type	4
#define dlsortFu_Date	6
#define dlsortFu_Case	(1<<8)
extern bool TDirFileList_ChkSortOk(int fu);
extern void TDirListObjNy_Sort(TDirListObjNy *self, int fu);

extern TglobPath *TDirListObjNy_Get_gp(TDirListObjNy *self);
extern TDirList *TDirListObjNy_Enum_TDirList(TDirList *dl);

extern TDirListObjNy *tfilemsg_get_obj_ny(TFileMsg *fmsg);

/* TDirListObjNy_GetDflSingle + Tfree */
extern TDirFileList *TDirListObjNy_GetDflSingle(TDirListObjNy *self, const char *fn);

/* ****************** tdirlist_misc.c ********************* */
extern TDirFileList *dfl_create(Tstat64 *AHstat);
extern void TDirListLoader_Clr_fmsg(TFileMsg *fmsg);
extern void TDirListLoader_fmsg_dfl(TDirListLoader *self, TFileMsg *fmsg, int bef, TDirFileList *dfl);
extern void TDirListLoader_New_st(TDirListLoader *self, Tstat64 *st, TFileMsg *fmsg);
extern TDirFileList *dfl_dup(TDirFileList *org_dfl);
extern bool chk_file_ok(Tstat64 *st, const char *path);
extern bool chk_pattern_match(TPattern *patt, const char *fn, Tstat64 *st);

#endif /* __TDIRLIST_H__ */
