/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef __ALD_TLINEEDIT_H__
#define __ALD_TLINEEDIT_H__    1

//******************* highlight ****************************
#define wtyp_highlight	0x11000

// ****** TColS.hl_mk (Byte) ******
#define hl_comline	1
#define hl_strline	(1<<1)
#define hl_defline	(1<<2)

// ****** ChangeLine (fu) ******
#define clfu_all	0
#define clfu_use_yaye	1


typedef struct ThlFCol {
  TFont *Font;
  int ForeColor, BackColor;
} ThlFCol;

#define hl_keyw_MAX 31

typedef struct Thl_keyw {
  unsigned char lg;
  unsigned char sc[hl_keyw_MAX];
} Thl_keyw;

typedef struct Thl_keyw_arr {
  ThlFCol fc;
  int font_art;
  int items;
  Thl_keyw keyw[];
} Thl_keyw_arr;

typedef struct Thighlight {
  AlgTTC
  TWins *wins;
  TLines *tls;		//zB.: TLineEdit->tls
  int hl_Type;
  void (*ChangeLine)(struct Thighlight*, struct TLines*, int , int , int); 
  void (*SetCaret)(struct Thighlight*, struct TLines*, int , int , int); 
  void (*MouseOver)(struct Thighlight*, struct TLines*, int , int , int); 
  bool (*hlCtl)(struct Thighlight *self, TWinsCtl *wctl);
  void (*prog_free5)(struct Thighlight*, struct TLines*, int , int , int); 
  TFont *Font;
  TFont *bold_Font;
  TFont *kursiv_Font;
  TFont *boldkursiv_Font;
  ThlFCol fc_comment;
  ThlFCol fc_char;
  ThlFCol fc_str;
  ThlFCol fc_nr;
  ThlFCol fc_def;
  ThlFCol fc_free5;
  ThlFCol fc_free6;
  ThlFCol fc_free7;

  Thl_keyw_arr *fc0_Dat;
  Thl_keyw_arr *fc1_Dat;
  Thl_keyw_arr *fc2_Dat;
  Thl_keyw_arr *fc3_Dat;
  int OldCaretX, OldCaretY;
  int mod;
  bool CharCase;
} Thighlight;


extern Thighlight *Thighlight_C_Create(TWins *wins, TLines *tls);
extern Thighlight *Thighlight_CXX_Create(TWins *wins, TLines *tls);
extern Thighlight *Thighlight_Bash_Create(TWins *wins, TLines *tls);
extern Thighlight *Thighlight_Spec_Create(TWins *wins, TLines *tls);
extern Thighlight *Thighlight_Make_Create(TWins *wins, TLines *tls);
extern Thighlight *Thighlight_ASMi386gcc_Create(TWins *wins, TLines *tls);
extern Thighlight *Thighlight_Pas_Create(TWins *wins, TLines *tls);

extern void hl_Set_keyword(Thighlight *self, Thl_keyw_arr **ppkarr, const char *ds, int font_art, int ForeColor, int BackColor);

extern void highlight_C(Thighlight *hi, TLines *tls, int edAnf, int edEnd, int fu);
extern void highlight_SetCaret_C(Thighlight *hl, TLines *tls, int lx, int ly, int fu);


//******************* TLineEditBase ***************************
#define LEBfu_SelRect		1
#define LEBfu_SelRectXoffAkt	(1<<1)

#define AlgLineEditBase AlgScrollWin \
  TLines *tls; \
  int CaretCharIdx; \
  int CaretCharIdy; \
  int VisibleLines; \
  int SelAx; \
  int SelAy; \
  int LineOffset; \
  dword LEBfu; \
  word SelXoffA, SelXoffE; \
  bool OldX; \
  bool ReadOnly; \
  bool EvStop; \
  bool fio_Strg_C_on;

#define TCSelf TLineEditBase
typedef struct TLineEditBase {
  AlgLineEditBase
} TLineEditBase;
#undef TCSelf

extern int   REGP1 TLineEditBase_GetLineFontHeight(TLineEditBase *self);
extern void  REGP1 TLineEditBase_Res_vds(TLineEditBase *self);
extern TColS REGP2 *TLineEditBase_Set_vds(TLineEditBase *self, int idy);
extern bool  REGP1 TLineEditBase_ChkIsSel(TLineEditBase *self);

extern int   TLineEditBase_PaintLine(TLineEditBase *self, int idy, int y);
extern bool  TLineEditBase_GetSel_lgs(TLineEditBase *self, Tlgs *lgs, int fu);
extern void  TLineEditBase_Strg_C(TLineEditBase *self);

//******************* TLineEdit *******************************

#define LEfu_VSBnoVisibleLines  1


#define AlgLineEdit AlgLineEditBase \
  int LEfu; \
  void (*HSB_DoCount)(struct TWins*, int); \
  void (*VSB_DoCount)(struct TWins*, int); \
  void (*OnChange)(struct TLineEdit*, int, TKeyS *KeyS, Tlgs*, int, int, int, int); \
  void (*OnSetCaret)(struct TLineEdit*, TColS*); \
  TPlugIn_KeyDown *PlugIn_KeyDown; \
  const Tss *TzC; \
  int OldCaretX; \
  int MaxStrX; \
  TUnRedo *UnRedo; \
  Thighlight *highlight; \
  bool Disable_keyplugin_Global; \
  bool nc_b5; \
  bool nc_b6; \
  bool nc_b7; \
  byte user_b0; \
  byte user_b1; \
  byte user_b2; \
  byte user_b3;

#define TCSelf TLineEdit
typedef struct TLineEdit {
  AlgLineEdit
} TLineEdit;
#undef TCSelf

extern TLineEdit *TLineEdit_Create(TWins *Parent);
extern TLineEdit *TLineEdit_CreateA(TWins *Parent, int Left, int Top, int Width, int Height, int Align);
extern bool TLineEdit__Destroy(TLineEdit *self);

extern int  TLineEdit_GetLineCount(TLineEdit *self);

extern bool TLineEdit_Undo(TLineEdit *le, TUnRedoLine *url);
extern bool TLineEdit_Redo(TLineEdit *le, TUnRedoLine *url);

extern void TLineEdit_Clr(TLineEdit *le);
extern void TLineEdit_Clr_c(TLineEdit *le);

extern bool TLineEdit_LoadFromFile(TLineEdit *le, char *FileName);
extern bool TLineEdit_SaveToFile(TLineEdit *le, char *FileName);
extern void TLineEdit_AllAtrNew(TLineEdit *le);

extern void TLineEdit_SetMouseZ_Wheel(TLineEdit *self, int Value);

#define LE_ChkCaret_Ymiddle	1
extern void TLineEdit_ChkCaret(TLineEdit *le, int fu);
extern void TLineEdit_GetChar_xy(TLineEdit *le, int idx, int idy, int *ret_x, int *ret_y, int *ret_h);
extern void TLineEdit_GetCaretIdxy(TLineEdit *le, int x, int y, int *idx, int *idy);
extern void TLineEdit_SetCaret(TLineEdit *le);

extern const char *tlineedit_get_pc(TLineEdit *le, int y);
extern TColS *tlineedit_get_pcs(TLineEdit *le, int y);

extern void TLineEdit_SetSel(TLineEdit *le, int xa, int ya, int xe, int ye);
//extern bool TLineEdit_ChkIsSel(TLineEdit *le);
#define TLineEdit_ChkIsSel(a) TLineEditBase_ChkIsSel((TLineEditBase*)a)

//extern bool TLineEdit_GetSel_lgs(TLineEdit *le, Tlgs *lgs, int fu);

//extern void TLineEdit_Strg_C(TLineEdit *le);
extern void TLineEdit_Strg_X(TLineEdit *le);
extern void TLineEdit_Strg_V(TLineEdit *le);
extern void TLineEdit_DoChange(TLineEdit *le, int bef, TKeyS *KeyS, Tlgs *Ins_lgs);

extern void TLineEdit_AddLine_lgs(TLineEdit *le, Tlgs *lgs);
extern void TLineEdit_AddLine(TLineEdit *le, const char *s);
extern void TLineEdit_Paste(TLineEdit *le, int *lineNr, int *charPos, const char *s);
extern void TLineEdit_AddPaste(TLineEdit *le, const char *s);
extern void TLineEdit_Add(TLineEdit *le, const char *s);
extern void TLineEdit_SetCaretXY(TLineEdit *le, int x, int y);

/* TLineEdit PlugIn */

extern TPlugIn_KeyDown *TLineEdit_keyplugin_Global;

extern void TLineEdit_AddKeyDownPlugIn(TLineEdit *le, TPlugIn_KeyDown *plug);
extern TPlugIn_KeyDown *TLineEdit_GetKeyDownPlugIn(TLineEdit *le);
extern TPlugIn_KeyDown *TLineEdit_keyplugin_cursor_Create();
extern TPlugIn_KeyDown *TLineEdit_keyplugin_edit_Create();
extern TPlugIn_KeyDown *TLineEdit_keyplugin_strg_u_i_Create();

#define TCSelf TLabelMultiLine
typedef struct TLabelMultiLine {
  AlgLineEditBase
  Tlgs lgs;
} TLabelMultiLine;
#undef TCSelf

extern TLabelMultiLine *TLabelMultiLine_Create(TWins *Parent);

#endif /* __ALD_TLINEEDIT_H__ */
