/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef	__ALD_TPIXMAP_H__
#define	__ALD_TPIXMAP_H__    1


#ifndef	__ALD_TIMAGE_H__
#include "image.h"
#endif

#ifndef _OS_WIN32
#define USE_XSHM	1
#include <X11/extensions/shape.h>
#endif

#ifdef USE_XSHM
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

typedef struct TPixMap {
  AlgTImage
  Pixmap pmxID;
  Pixmap maskID;
  TVDrawString *vds;
  XImage *Ximg;
  GC gc;
  GC line_gc;
  GC mask_gc;
  int LineColor;
  Display *dpy;
  XGCValues gcv;
#ifdef USE_XSHM
  XShmSegmentInfo shminfo;
#endif
} TPixMap;

extern TPixMap *TPixMap_Create_depth(int width, int height, dword mem_depth);
extern TPixMap *TPixMap_Create_whLD(int width, int height, dword pxm_LD);
extern TPixMap *TPixMap_Create_XID(Pixmap pixID, Pixmap maskID);
extern bool TPixMap__Destroy(TPixMap *self);
extern int TPixMap_img_malloc(TPixMap *self);
extern void TPixMap_CreateXimg(TPixMap *pxm);
extern void TPixMap_SetXShm(TPixMap *pxm);

extern void TPixMap_XGetImage(TPixMap *pxm);
extern void TPixMap_XPutImage(TPixMap *pxm);
extern void TPixMap_XPutImage_dpy(Display *dpy, TPixMap *pxm);


extern void TPixMap_SetFont(TPixMap *pxm, TFont *font);
extern void TPixMap_SetFontName(TPixMap *pxm, const char *name);
extern void TPixMap_SetFontColor(TPixMap *pxm, dword ForeColor, dword BackColor);
extern void TPixMap_DrawString(TPixMap *pxm, const char *s, int x, int y);

extern void TPixMap_ClsRGB(TPixMap *pxm, dword Color);
extern void TPixMap_SetMaskRGB(TPixMap *pxm, dword Color);
extern void TPixMap_PaintToWin(TPixMap *pxm, Window WinID, int left, int top);
extern void TPixMap_ToWinBackground(TPixMap *pxm, Window WinID);
extern void TPixMap_WinCombineMask(TPixMap *pxm, Window WinID);

extern TPixMap *TPixMap_Loader(const char *fn, TglobPath *gp, dword pxm_LD);
extern TPixMap *TPixMap_Loader_F(const char *fn, TglobPath *gp, dword pxm_LD, FILE *F);

/* ****** tpixmap_load.c tpixmap_load_xxx.c ****** */

/* ****** tpxm_load.c ****** */
#define TpxmLoad_Init TimgLoad_Init
#define TpxmLoad_Info TimgLoad_Info
#define TpxmLoad_Data TimgLoad_Data


/* ****** tpixmap_resize.c ****** */
extern TPixMap *TPixMap_dup_new_size(TPixMap *org_pxm, int new_width, int new_height);
extern void TPixMap_resize(TPixMap **ppxm, int new_width, int new_height);

/* ****** tpixmap_shift.c ****** */
extern void TPixMap_shift(TPixMap *pxm, int xo, int yo);

/* ****** tpixmap_line.c ****** */
// void TPixMap_SetLineColor(TPixMap *pxm, int RGB);
extern void TPixMap_DrawLineRGB(TPixMap *pxm, int RGB, int x1, int y1, int x2, int y2);
extern void TPixMap_DrawLinePixRGB(TPixMap *pxm, int RGB, int x1, int y1, int x2, int y2);
extern int TPixMap_DrawLineImgRGB(TPixMap *pxm, dword RGB, int x1, int y1, int x2, int y2);

/* ****** tpixmap_getcolorcount.c ****** */
extern int TPixMap_GetColorCount(TPixMap *pxm, int fu);


/* ****** selection_tpixmap.c ****** */
extern void SetSelection_TPixMap(TPixMap *pxm);
extern TPixMap *GetSelection_TPixMap();

#endif /* __ALD_TPIXMAP_H__ */
