/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef __TSR232_H__
#define __TSR232_H__

#ifndef _OS_WIN32
#include <termios.h>
#endif

#define RS232TIMEEXACT	0
#define RS232MAXBUF	1024

#define rs232ERR_RxD_OVERRUN	1
#define rs232ERR_TIMEOUT_REP	16
#define rs232ERR_TIMEOUT	17
#define rs232ERR_RxD_CRC_REP	18
#define rs232ERR_RxD_CRC	19
#define rs232ERR_TxD_CRC_REP	20
#define rs232ERR_TxD_CRC	21

#define AlgTrs232    AlgTTC \
  void  *tc_private; \
  byte  *RxDbuf; \
  dword RxDsize; \
  dword RxDidx; \
  dword RxDscr; \
  byte  *TxDbuf; \
  dword TxDsize; \
  size_t TxDDataLg; \
  int   TxDRepCount; \
  int   TxDRepMaxCount; \
  dword TimeOutCount; \
  bool (*open)(struct TCSelf *self, dword com, dword baud, dword rsflag); \
  bool (*close)(struct TCSelf *self); \
  void (*OnTimeOutTimer)(struct TCSelf *self); \
  void (*OnTxD)(struct TCSelf *self, const byte *data, size_t lg); \
  bool (*OnRxD)(struct TCSelf *self, byte *data, size_t lg); \
  bool (*OnERROR)(struct TCSelf *self, int rs232Err, ic32 *nc); \
  FileHandle fd; \
  dword rsflag; \
  uint64_t usTxD, usRxD; \
  uint64_t GesTxD_Byte; \
  uint64_t GesRxD_Byte;


/* ------------- Trs232 ---------------- 
Trs232.OnRxD		result true --> reset RxD buf
Trs232.OnERROR		result true = no ERROR msg
 ------------- Trs232 ---------------- */
#define TCSelf Trs232
typedef struct Trs232 {
  AlgTrs232
} Trs232;
#undef TCSelf

typedef struct Thrs232Filter {
  bool (*init)(Trs232 *rs232);

} Thrs232Filter;

Trs232 *Trs232_Create();

#define rsWLEN8		0x00
#define rsWLEN5		0x01
#define rsWLEN6		0x02
#define rsWLEN7		0x03
#define rsPARITY	0x04
#define rsPARITYODD	0x08
#define rsSTOP2BIT	0x10
#define rsRTSCTS	0x20

extern bool rs232_open(Trs232 *self, dword com, dword baud, dword rsflag);
extern bool rs232_close(Trs232 *self);
extern bool Trs232_IsOpen(Trs232 *self);

extern bool Trs232_TxD(Trs232 *self, const byte *data, size_t lg);
extern bool Trs232_TxDBuf(Trs232 *self);
extern bool Trs232_TxDBuf_Rep(Trs232 *self); /* rs232.usTxD no update */

extern bool Trs232_InitTimeOut(Trs232 *self, dword ms);
extern void Trs232_ClrTimeOut(Trs232 *self);

#ifdef IS_CLASS_LIB
extern bool Trs232__Destroy(Trs232 *self);
extern void *Trs232_Create_struct_init(int tcsize, int wtyp);
extern bool Trs232_open(Trs232 *self, dword com, dword baud, dword rsflag);
extern bool Trs232_close(Trs232 *self);
#endif

#endif /* __TSR232_H__ */
