/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef __APPX_TTY_H__
#define __APPX_TTY_H__    1

#ifndef _OS_WIN32
#include <termios.h>
#endif

/* ttyscan.c */
#define tty_normal	0
#define tty_esc		1
#define tty_square	2
#define tty_nonstd	3
#define tty_percent	4
#define tty_palette	5
#define tty_getnums	6
#define tty_gotpars	7
#define tty_funckey	8
#define tty_hash	9
#define tty_setG0	10
#define tty_setG1	11
#define tty_esc_O	12

#define term_linux	0
#define term_xterm	1

//*********************** misc Key ***********************************
#define yXK_BackSpace		0xFF08	/* back space, back char */
#define yXK_Tab			0xFF09
//#define yXK_Linefeed		0xFF0A	/* Linefeed, LF */
#define yXK_Clear		0xFF0B
#define yXK_Return		0xFF0D	/* Return, enter */
//#define yXK_Pause		0xFF13	/* Pause, hold */
//#define yXK_Scroll_Lock		0xFF14
//#define yXK_Sys_Req		0xFF15
//#define yXK_Escape		0xFF1B

#define yXK_Home		0xFF50
#define yXK_Left		0xFF51	/* Move left, left arrow */
#define yXK_Up			0xFF52	/* Move up, up arrow */
#define yXK_Right		0xFF53	/* Move right, right arrow */
#define yXK_Down		0xFF54	/* Move down, down arrow */
#define yXK_Page_Up		0xFF55
#define yXK_Page_Down		0xFF56
#define yXK_Insert		0xFF63	/* Insert, insert here */
#define yXK_Delete		0xFFFF	/* Delete, rubout */

#define yXK_F1			0xFFBE
#define yXK_F2			0xFFBF
#define yXK_F3			0xFFC0
#define yXK_F4			0xFFC1
#define yXK_F5			0xFFC2
#define yXK_F6			0xFFC3
#define yXK_F7			0xFFC4
#define yXK_F8			0xFFC5
#define yXK_F9			0xFFC6
#define yXK_F10			0xFFC7
#define yXK_F11			0xFFC8
#define yXK_F12			0xFFC9

typedef struct {
  char res[8];
  int reslg;
  int btty_XK;
  int state;
  int laststate;
  int nums[8];
  int numIdx;
  int time_ms;
  int termart;
  bool err;
  bool ques;
} TTermscan;

extern TTermscan *TTermscan_Create();
extern void TTermscan_Destroy(TTermscan *term);
extern bool term_scan(byte c, TTermscan *term);

#define mCOLOR_BLACK	0
#define mCOLOR_RED	1
#define mCOLOR_GREEN	2
#define mCOLOR_YELLOW	3
#define mCOLOR_BLUE	4
#define mCOLOR_CYAN	5
#define mCOLOR_MAGENTA	6
#define mCOLOR_WHITE	7
#define mCOLOR_DEFAULT	9
#define mCOLOR_NONE	-1
#define mCOLOR_REFRESH	-2

typedef struct {
  int fdout;
  int termart;
  int forecolor;
  int backcolor;
} TTermStr;

extern TTermStr *TTermStr_Create(int fdout, int termart);
extern void TTermStr_Destroy(TTermStr *termstr);
extern void TTermStr_ColOut(TTermStr *termstr, const char *str, int fcol, int bcol);
extern void TTermStr_lf(TTermStr *termstr);
extern void TTermStr_DelChars(TTermStr *termstr, int z);
extern void TTermStr_EraseLine(TTermStr *termstr);

//*********************** TTY ****************************
#define MaxTermLines	5

typedef struct Ttty {
  AlgTTC
  char *KeyLine;
  int KeyLineSize;
  int KeyIdx;
  int MaxKeyIdx;
  int LastLineIdx;
  int MaxLastLineIdx;
  bool InputStop, nc1, nc2, nc3;
  int fdin;
  int fdout;

  void *owner;
  TTermscan *term;
  TTermStr *termstr;
  char *linehead;
  void (*OnReturn)(struct Ttty*, char *line, int lg);
  void (*OnKey)(struct Ttty*);
#ifndef _OS_WIN32
  struct termios org_termios;
#endif
  char *LastKeyLines[MaxTermLines+1];
} Ttty;

/* apptty.c */
extern Ttty *Ttty_Create(int fdin, int fdout, const char *linehead);
extern void Ttty_NewLine(Ttty *tty, char *wrs, int fu);
/* apptty_line.c */
extern void Ttty_LineKey(Ttty *tty);

#endif /* __APPX_TTY_H__ */
