/*
 Copyright (C) 2010 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "appXbase.h"
#include "image.h"


bool DIB_to_image(TImage *img, BITMAPINFOHEADER *p, byte *src_data) {
  byte *dest0;
  int bytes_per_line;

	if (!img || DIB_chk_error(p) || !src_data) return False;

	img->width = p->biWidth;
	img->height = p->biHeight;
	TImage_SetMem(img);
	dest0 = img->data;

	if (((img->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) &&
	    (p->biBitCount == 16)) {

	    bytes_per_line = ((p->biWidth * 2) + 3) & 0xfffffffc;
	    src_data += bytes_per_line * (img->height - 1);
	    for (int y=img->height;y>0;y--) {
		word *dest = (word*)dest0;
		word *src =  (word*)src_data;
		for (int x=img->width;x>0;x--) {
		    *dest = *src;
		    dest++; src++;
		}
		dest0 += img->bytes_per_line;
		src_data -= bytes_per_line;
	    }

	} else if (((img->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) &&
		   (p->biBitCount == 24)) {
	    bytes_per_line = ((p->biWidth * 3) + 3) & 0xfffffffc;
	    src_data += bytes_per_line * (img->height - 1);
	    for (int y=img->height;y>0;y--) {
		byte *dest = dest0;
		byte *src =  src_data;
		for (int x=img->width;x>0;x--) {
		    *dest = *src; dest++; src++;
		    *dest = *src; dest++; src++;
		    *dest = *src; dest++;
		    src += 2;
		}
		dest0 += img->bytes_per_line;
		src_data -= bytes_per_line;
	    }

	} else if (((img->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) &&
		   (p->biBitCount == 32)) {
	    bytes_per_line = p->biWidth * 4;
	    src_data += bytes_per_line * (img->height - 1);
	    for (int y=img->height;y>0;y--) {
		word *dest = (word*)dest0;
		dword *src =  (dword*)src_data;
		for (int x=img->width;x>0;x--) {
		    *dest = ((*src & 0xf80000)>>8)|((*src & 0xfc00)>>5)|((*src & 0xf8)>>3);
		    dest++; src++;
		}
		dest0 += img->bytes_per_line;
		src_data -= bytes_per_line;
	    }
	
	} else if (p->biBitCount == 16) {
	    bytes_per_line = ((p->biWidth * 2) + 3) & 0xfffffffc;
	    src_data += bytes_per_line * (img->height - 1);
	    for (int y=img->height;y>0;y--) {
		dword *dest = (dword*)dest0;
		word  *src =  (word*)src_data;
		for (int x=img->width;x>0;x--) {
		    *dest = ((*src & 0x7c00)<<9)|((*src & 0x3e0)<<6)|((*src & 0x1f)<<3);
		    dest++; src++;
		}
		dest0 += img->bytes_per_line;
		src_data -= bytes_per_line;
	    }

	} else if (p->biBitCount == 24) {
	    bytes_per_line = ((p->biWidth * 3) + 3) & 0xfffffffc;
	    src_data += bytes_per_line * (img->height - 1);
	    for (int y=img->height;y>0;y--) {
		dword *dest = (dword*)dest0;
		byte  *src =  src_data;
		for (int x=img->width;x>0;x--) {
		    *dest = src[0]|(src[1]<<8)|(src[2]<<16);
		    dest++; src += 3;
		}
		dest0 += img->bytes_per_line;
		src_data -= bytes_per_line;
	    }

	} else if (p->biBitCount == 32) {
	    bytes_per_line = p->biWidth * 4;
	    src_data += bytes_per_line * (img->height - 1);
	    for (int y=img->height;y>0;y--) {
		dword *dest = (dword*)dest0;
		dword *src =  (dword*)src_data;
		for (int x=img->width;x>0;x--) {
		    *dest = *src;
		    dest++; src++;
		}
		dest0 += img->bytes_per_line;
		src_data -= bytes_per_line;
	    }

	} else return False;
	return True;
}
