/*
 Copyright (C) 2010 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "appXbase.h"

Pixmap XCreatePixmap (Display *dpy, Drawable d, 
		unsigned int width, unsigned int height, unsigned int depth) {
  BITMAPINFO *pbmi;
  HBITMAP hbitmap = 0;
  int bytes_per_line;
  LPBYTE lpBits;
  HDC hdc;
  TBITMAP_Dat *bmDat;
  int m;

	if ((depth != 1) && (depth != 16) && (depth != 24) && (depth != 32)) return 0;

	bytes_per_line = Get_bytes_per_line(depth, width);

	pbmi = xcalloc(sizeof(BITMAPINFO) + (256 * sizeof(RGBQUAD)));
	pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	pbmi->bmiHeader.biWidth = width;
	pbmi->bmiHeader.biHeight = height;
	pbmi->bmiHeader.biPlanes = 1;
	if (depth == 24) pbmi->bmiHeader.biBitCount = 32;
	  else pbmi->bmiHeader.biBitCount = depth;
	if (depth == 1) {
	    pbmi->bmiHeader.biCompression = BI_RGB;
	    pbmi->bmiHeader.biSizeImage = 0;
	} else {
	    pbmi->bmiHeader.biCompression = BI_BITFIELDS;
	    pbmi->bmiHeader.biSizeImage = bytes_per_line * height;
	}

	int *bf = (int*)&pbmi->bmiColors[0].rgbBlue;
	if (depth == 1) {
	    bf[0] = 0x00000000;
	    bf[1] = 0x00FFFFFF;
	} else if (depth == 16) {
	    bf[0] = 0xF800;
	    bf[1] = 0x07E0;
	    bf[2] = 0x001F;
	} else {
	    bf[0] = 0x00FF0000;
	    bf[1] = 0x0000FF00;
	    bf[2] = 0x000000FF;
	}

	hdc = CreateCompatibleDC(0);
	hbitmap = CreateDIBSection(hdc, pbmi, DIB_RGB_COLORS, (void*)&lpBits, 0, 0);
	DeleteDC(hdc);

	if (!hbitmap) goto ende;

	bmDat = xcalloc(sizeof(TBITMAP_Dat));
	bmDat->hbitmap = hbitmap;
	bmDat->depth = depth;

	m = GetObject(hbitmap, sizeof(DIBSECTION), &bmDat->ui.DIBS);
	if (!m) {
	    xfree(bmDat);
	    goto ende;
	}

	bmDat->ui_type = m;
	if (depth == 1) {
	    bmDat->bytes_per_line = bmDat->ui.DIBS.dsBmih.biSizeImage / bmDat->ui.DIBS.dsBmih.biHeight;
	} else {
	    bmDat->bytes_per_line = bytes_per_line;
	}
/*
 printf("h=0x%x %d %ld %ld %ld %ld=%ld\n", (int)hbitmap, bytes_per_line, 
		bmDat->ui.DIBS.dsBm.bmType, bmDat->ui.DIBS.dsBm.bmWidth, 
		bmDat->ui.DIBS.dsBm.bmHeight, 
		bmDat->ui.DIBS.dsBmih.biSizeImage, pbmi->bmiHeader.biSizeImage);
*/
	Add_BITMAP_Dat(bmDat);
ende:
	xfree(pbmi);
	return hbitmap;
}

