/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include <unistd.h>
#include <stdarg.h>
#include <X11/keysym.h>


#ifdef Test_Root_WinId
Window TestRootWinId = 0;
Window OrgRootWinId = 0;
#endif

extern void (*FAppRun)();

TappXDat appXDat;
dword appXDat_Default_Color = 0xc0c0c0;

Tptr_arr *PaintOnEvEnd_List;
int PaintOnEvEnd_Count;
XMotionEvent MotionEvent;
dword Hz30;

bool (*On_wmError)(Display *dpy, XErrorEvent *e);

char *display_name;

TdpyMonitors dpyMonitors;

#ifndef USE_NANO_X
static int wmErrorHandler(Display *dpy, XErrorEvent *e) {
  char  number[16], msg[80], req[80];

	if (AppToAtExit) return 0;
	if (On_wmError && On_wmError(dpy, e)) return 0;

	XGetErrorText(dpy, e->error_code, msg, sizeof(msg));
	inttopc(number, e->request_code);
	XGetErrorDatabaseText(dpy, "XRequest", number, "<unknown>", req, sizeof(req));

	printf("ERROR X11: %s: %s(0x%lx): %s\n", app_name, req, e->resourceid, msg);
	return 0;
}
#endif /* USE_NANO_X */

static int CatchFatal(Display *dpy) {
	printf("ERROR: X Server msg. CatchFatal\n");
//	appDat.gQUIT_Ev = True;
	return 0;
}

static const char *AH_AtomsTab[] = {
  "WM_DELETE_WINDOW",
  "WM_STATE",
  "WM_PROTOCOLS",
  "_MOTIF_WM_HINTS",
  "ALD_selection",
  "NONE",
  "CLIPBOARD",
  "TARGETS",
  "INCR",
  "UTF8_STRING",
  "text/plain",
  "text/uri-list",
  "image/png",
  "application/x-kde-cutselection",
  "application/x-kde-urilist",
  "_NET_WM_STATE",
  "_NET_WM_USER_TIME",
  "_NET_WM_STATE_ABOVE",
#ifdef FORM_NET_WM_PID
  "_NET_WM_PID",
#endif
  NULL
};

#ifdef APP_INI_DAT
static const char DefaultIniDat[] = {
  "[Global]\n"
  "; Clipbord_Mode 0=X11 1=QT3\n"
  "Clipbord_Mode=1\n"
  "Font=-adobe-helvetica-medium-r-normal-*-12-*-75-75-*-*-*-1\n"
  "FontBold=-adobe-helvetica-bold-r-normal-*-12-*-75-75-*-*-*-1\n"
  "; DateFormat 0=d-m-y 1=m-d-y 2=y-m-d\n"
  "DateFormat=0\n"
  "FileSystemUTF8=0\n"
  "MouseZ_Wheel=3\n"
  "\n"
  "[Icon]\n"
  "IconPaths=\""ALDccpPREFIX"/share/icons/\" \"/usr/share/icons/hicolor/\"\n"
  "\n"
  "[App]\n"
  ";dotdesktop=/usr/share/applications /opt/kde3/share/applications/kde\n"
  ";dotdesktop_mimelnk=/usr/share/mimelnk\n"
  "\n"
  "[Dlg_File]\n"
  "Az0_ItemE_Dbl=True\n"
  "Az0_PathE_Dbl=False\n"
  "Az1_ItemE_Dbl=True\n"
  "Az1_PathE_Dbl=False\n"
};
#endif

static const byte XC_Dat[18] = {
    XC_X_cursor,
    XC_arrow,
    XC_bottom_left_corner,
    XC_bottom_right_corner,
    XC_bottom_side,
    XC_dot,
    XC_fleur,
    XC_left_side,
    XC_pirate,
    XC_right_side,
    XC_sb_h_double_arrow,
    XC_sb_v_double_arrow,
    XC_top_left_arrow,
    XC_top_left_corner,
    XC_top_right_corner,
    XC_top_side,
    XC_watch,
    XC_xterm};

/*____________________________________________________________________________*/
static void InitStartForm() {
  TWins *wins;
  Tobjmsg msg;
  XSetWindowAttributes XAttr;

	Fill0(&msg, sizeof(msg));
	msg.msg = msgVisibleFormShow;

	wins = appXDat.Wins;
	while (wins) {
	    if (wins->wtyp & wtyp_Form) Send_MSG(&msg, wins);
	    wins = wins->next;
	}

	Fill0(&XAttr, sizeof(XAttr));
	XAttr.cursor = appXDat.XCursors[XC_watch];
	appXDat.SelHandleWin = XCreateWindow(appXDat.dpy, 
			appXDat.Wins->WinID, 
			0, 0, 1, 1,
			0, CopyFromParent, InputOutput,
			CopyFromParent, CWCursor, &XAttr);

}

static void Send_AtExit(TWins *wins) {
  TWinsCtl wctl;

	while (wins) {
	    if (wins->next_child) Send_AtExit(wins->next_child);
	    wctl.bef = wctlb_AtExit;
	    wctl.flag = 0;
	    WinsCtl(wins, &wctl);
	    wins = wins->next;
	}
}

static void XAppRun() {
  Display *dpy;
// printf("%s(%d): %dms\n", FL, GetTime_ms() - appDat.app_start_ms);
	InitStartForm();
//	XSync(appXDat.dpy, 0);
	AppSetWait0_single();
	AppHandleMessage();
	if (appXDat.OnAppStart) appXDat.OnAppStart();
// printf("%s(%d): %dms\n", FL, GetTime_ms() - appDat.app_start_ms);
	AppSetWait1_single();
NotClose:
	while (AppHandleMessageRun());

	if (appXDat.OnAppClose) appXDat.OnAppClose();
	if (!appDat.gQUIT_Ev) goto NotClose;
	appDat.AppExit = True;
	appError_ev = NULL;

	dpy = appXDat.dpy;
	Send_AtExit(appXDat.Wins);
	appXDat.dpy = NULL;
	app_atexit();
	XCloseDisplay(dpy);
}

/*____________________________________________________________________________*/
void initXapp(int fu, int argc, char **argv) {
  Display *dpy;
  int i;
  const char *pc;
  Atom *at;
  const char **ats;
  long gcm;
  XGCValues gcv;

	initapp(fu, argc, argv);
	appDat.X11DispatchEvent = X11DispatchEvent;
	appDat.Wins_PaintOnEvEnd = Wins_PaintOnEvEnd;
	FAppRun = XAppRun;

#ifdef APP_INI_DAT

	if (!(fu & app_NoIni)) {
	    appDat.inifile_global = pcadd(appDat.ALDini_dir, "/global.ini");
	    if (euidaccess(appDat.inifile_global, F_OK)) {
		FileHandle fd;
		file_open_wr_ae(appDat.inifile_global, &fd, 0640, EMSG_ERR_EXIT);
		file_write(fd, NULL, DefaultIniDat, sizeof(DefaultIniDat)-1, NULL);
		file_close(fd, NULL);
	    }
	}
#endif

	dpy = XOpenDisplay(display_name);
	if (!dpy) appError(EMSG_ERR_EXIT, "Cannot open display %s", XDisplayName(display_name));
	appDat.x_fd = XConnectionNumber(dpy);
	fd_set_count(appDat.x_fd);
	appXDat.dpy = dpy;

// printf("%s(%d): %d %s\n", FL, DefaultScreen(dpy), DisplayString(dpy));

	appXDat.RootW = DefaultRootWindow(dpy);
	if (!appXDat.RootW) 
	  appError(EMSG_ERR_EXIT, "Screen %d is not a valid screen", appXDat.RootW);

	appXDat.default_screen = DefaultScreen(dpy);

	MonitorsInit();

// printf("%d %d %d \n", appXDat.Root_width, appXDat.Root_height, appXDat.RootW_depth);
	appXDat.bixformat = B16_RGB565;
	if (appXDat.RootW_depth > 16) appXDat.bixformat = B32_RGB888;

	at = &appXDat.XA_WM_DELETE_WINDOW;
	ats = AH_AtomsTab;
	while(*ats) {
	    *at = XInternAtom(dpy, *ats, 0);
	    at++; ats++;
	}
	appXDat.CLIPBOARD_ATOM = XA_PRIMARY;

	appXeventDefaultInit();

	XSetIOErrorHandler(CatchFatal);
	XSetErrorHandler(wmErrorHandler);

	appXDat.Default_Color = appXDat_Default_Color;
	appXDat.Default_GC = CreateGC_RGB(appXDat.Default_Color);
	appXDat.Paint_Color = appXDat_Default_Color;
	appXDat.Paint_GC = CreateGC_RGB(appXDat.Paint_Color);

	appXDat.bordH_GC = CreateGC_RGB(0xe0e0e0);
	appXDat.bordL_GC = CreateGC_RGB(0x808080);
	appXDat.GC_FFFFFF = CreateGC_RGB(0xffffff);
	appXDat.GC_C0C0C0 = CreateGC_RGB(0xc0c0c0);
	appXDat.GC_404040 = CreateGC_RGB(0x404040);
	appXDat.GC_000000 = CreateGC_RGB(0x000000);
	appXDat.Select_GC_Back = CreateGC_RGB(0x00007f);

	Fill0(&gcv, sizeof(XGCValues));
#ifdef USE_X11GC_plane_mask
	gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth;
	gcv.plane_mask = AllPlanes;
#else
	gcm = GCFunction | GCGraphicsExposures | GCLineWidth;
#endif
	gcv.function = GXcopy;
	gcv.fill_style = FillSolid;
	appXDat.IconGC = CreateGC(gcm, &gcv);

#ifdef USE_X11GC_plane_mask
	gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth |
	      GCForeground | GCFillStyle;
#else
	gcm = GCFunction | GCGraphicsExposures | GCLineWidth | GCForeground | GCFillStyle;
#endif
	gcv.function  = GXnand;
	gcv.foreground = GetColor_RGB(0xffff80);
//	gcv.function  = GXequiv;
//	gcv.foreground = GetColor_RGB(0x7fffff);
	appXDat.Select_GC = CreateGC(gcm, &gcv);

	i = 0x00000102;
	appXDat.Focus_bitmap = XCreateBitmapFromData(dpy, appXDat.RootW, (char*)&i, 2, 2);
	gcv.stipple = appXDat.Focus_bitmap;
	gcv.function = GXxor;
	gcv.line_width = 1;
	gcv.foreground = GetColor_RGB(0x000000);
	gcv.background = GetColor_RGB(0xc0c0c0);
	gcv.fill_style = FillOpaqueStippled;
	gcv.subwindow_mode = IncludeInferiors;
	gcm |= GCStipple | GCSubwindowMode | GCBackground;
	appXDat.FocusLine_GC = CreateGC(gcm, &gcv);

	gcv.function = GXandInverted;
	appXDat.GC_000000_Dot = CreateGC(gcm, &gcv);

	gcv.function = GXcopy;
	gcv.foreground = GetColor_RGB(0xffffff);
	appXDat.SB_GC = CreateGC(gcm, &gcv);

	for (i=0;i<18;i++) {
	  appXDat.XCursors[XC_Dat[i]] = XCreateFontCursor(dpy, XC_Dat[i]);  //in X11/cursorfont.h
	}
	appXDat.Menue_Cursor = appXDat.XCursors[XC_arrow];

#ifdef APP_INI_DAT
	if (fu & app_NoIni) {
	    appXDat.Default_Font = MakeFont("-*-helvetica-medium-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
	    appXDat.Default_bold_Font = MakeFont("-*-helvetica-bold-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
	} else {
	    TpcIni *ini = TpcIni_Create(appDat.inifile_global);
	    pc = TpcIni_RD(ini, "Global", "Font");
	    appXDat.Default_Font = MakeFont(pc, "-*-helvetica-medium-r-*-*-12-*-75-75-*-*-*-*", NULL);
	    appXDat.Default_bold_Font = MakeFont(TpcIni_RD_pc(ini, "Global", "FontBold", pc), pc, NULL);

	    if (TpcIni_RD_int(ini, "Global", "Clipbord_Mode", 0) == 1) 
		appXDat.CLIPBOARD_ATOM = appXDat.XA_CLIPBOARD;

	    appXDat.UTF8on = TpcIni_RD_boolean(ini, "Global", "UTF8", 0);
	    appXDat.FileSystemUTF8on = TpcIni_RD_boolean(ini, "Global", "FileSystemUTF8", 0);
	    appXDat.Default_MouseZ_Wheel = TpcIni_RD_int(ini, "Global", "MouseZ_Wheel", 3);

	    GlobDateFormat = TpcIni_RD_int(ini, "Global", "DateFormat", 0);
	    Tfree(ini);
	}
#else
	appXDat.Default_Font = MakeFont("-*-helvetica-medium-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
	appXDat.Default_bold_Font = MakeFont("-*-helvetica-bold-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
#endif
	SetFont_Default(appXDat.Default_Font, True);
	SetFont_Default(appXDat.Default_bold_Font, True);
	set_ptr_arr(&PaintOnEvEnd_List, 254);
}

/* ******************* X11 Event's ******************** */


int Send_MSG(Tobjmsg *msg, TWins *wins) {
	if (!msg || !ttc_chk_struct(wins) || !wins->EvProc) return 0;
	msg->wins = wins;
	return wins->EvProc(msg);
}

int Exec_Win_Ev(Tobjmsg *msg, Window win) {
  TWins *wins;
	wins = GetTWins(win); 
	return Send_MSG(msg, wins);
}

/* *********************************************** */
void Send_MSG_MotionEvent() {
  Tobjmsg msg;
  Window window;

	if (!MotionEvent.window) return;
	msg.msg = msgXEv;
	msg.f = msg.m = 0;
	msg.p = &MotionEvent;
	appXDat.last_KeyState = MotionEvent.state;
	window = MotionEvent.window;
	MotionEvent.window = 0;
	Exec_Win_Ev(&msg, window);
}

void X11DispatchEvent() {
  Tobjmsg msg;
  XEvent Event;
  Window win;

	if (appXDat.dpy && (XPending(appXDat.dpy) > 0)) {
loop:
	    XNextEvent(appXDat.dpy, &Event);
	    msg.msg = msgXEv;
	    msg.f = msg.m = 0;
	    msg.p = &Event;
	    win = exec_XEventHandle((void*)&msg);
	    if (win) {
		msg.wins = GetTWins(win);
		if (msg.wins && msg.wins->EvProc) msg.wins->EvProc(&msg);
	    }
	    if (!appDat.gTimerNewEv && 
		appXDat.dpy && (XPending(appXDat.dpy) > 0)) goto loop;

	    AppSetWait1_single();
	}

	if (MotionEvent.window) {
	    if (Hz30 != appDat.Hz30) {
		Hz30 = appDat.Hz30;
		Send_MSG_MotionEvent();
	    }
	    AppSetWait1_single();
	}
}

/* void handle_wait_X11Event() {
	if (appDat.X11DispatchEvent)  appDat.X11DispatchEvent();
	if (appDat.Wins_PaintOnEvEnd) appDat.Wins_PaintOnEvEnd();
} */


void Wins_PaintOnEvEnd() {
  int count, i;
  void **val;
  Tobjmsg msg;

	if (PaintOnEvEnd_Count > 0) {
	    msg.msg = msgPaint;
	    count = PaintOnEvEnd_Count;
	    PaintOnEvEnd_Count = 0;
	    val = pndup(PaintOnEvEnd_List->val, sizeof(void*) * count);
	    
	    for (i=0; i < count; i++) {
		msg.wins = val[i];
		if (ttc_chk_struct(msg.wins) && msg.wins->EvProc)
					msg.wins->EvProc(&msg);
	    }
	    xfree(val);
	    AppSetWait1_single();
	}

	if (appXDat.InfoWin && cpTime(appXDat.InfoWin_us, appDat.app_us)) {
	    SetVisible(appXDat.InfoWin, True);
	}
}


void MapWinID(Window WinID, int fu) {  //fu 0=UnMap 1=Map 2=MapRaised
  Display *dpy = appXDat.dpy;

	if (!fu) {
	    XUnmapWindow(dpy, WinID);
	} else if (fu == 1) {
	    XMapWindow(dpy, WinID);
	} else {
	    XMapRaised(dpy, WinID);
	}
}

void AppHandleWinMessages() {
	if (appDat.X11DispatchEvent)  appDat.X11DispatchEvent();
	if (appDat.Wins_PaintOnEvEnd) appDat.Wins_PaintOnEvEnd();
}



