/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

bool FontExists(const char *fontname) {
  int L;
  char **FontNameList;
	FontNameList = XListFonts(appXDat.dpy, fontname, 1, &L);
	XFreeFontNames(FontNameList);
	return (L > 0);
}

char *GetFullFontName(const char *fontname) {
  int L;
  char **FontNameList;
  char *res = NULL;
	if (!fontname || !*fontname) return NULL;
	FontNameList = XListFonts(appXDat.dpy, fontname, 1, &L);
	if (L > 0) res = pcdup(*FontNameList);
	XFreeFontNames(FontNameList);
	return res;
}

static TFontStructCache *get_fsco(TFont *font) {
  int i;
  TFontStructCache *fsc;
	if (!font || !font->xfont) return NULL;
	fsc = appXDat.FontStructCache;
	for (i=FSCsitze;i>0;i--) {
	    if (fsc->xfont == font->xfont) return fsc;
	    fsc++;
	}
	return NULL;
}

void SetFont_Default(TFont *font, bool def) {
  TFontStructCache *fsco = get_fsco(font);
	if (!fsco) return;
	fsco->fu &= ~tf_isDefaultFont;
	font->xfu &= ~tf_isDefaultFont;
	if (def) {
	    fsco->fu |= tf_isDefaultFont;
	    font->xfu |= tf_isDefaultFont;
	}
}




void FreeFont(TFont *font) {
  TFontStructCache *fsc = get_fsco(font);

	if (!fsc) {
	    printf("ERROR -- can t del old_Font font ??????????????\n");
	} else {
	    fsc->count--;
	    if (fsc->count <= 0) {
		if (fsc->fu & tf_isDefaultFont) {
		    fsc->count = 1;
		} else {
// printf("%s(%d) FreeFont:  count=%d %s\n", FL, fsc->count, fsc->name);
		    if (appXDat.dpy) XFreeFont(appXDat.dpy, fsc->xfont);
		    xfree(fsc->name);
		    fsc->xfont = NULL;
		}
	    }
	}
}

void TFont_Destroy(TFont *font) {
	if (font && !(font->xfu & tf_isDefaultFont)) {
	    FreeFont(font);
	    FreeGC(font->g.FontGC);
	    FreeGC(font->gtmp.FontGC);
	    xfree(font);
	}
}

static void MakeTmpFontGC(TFont *font, TFontGC *g, long ForeColor, long BackColor) {
  long gcm;
  XGCValues gcv;

	if (!font || !font->xfont || !g) return;
	Fill0(&gcv, sizeof(XGCValues));
	if (g->FontGC) {
	    gcm = 0;
	    if (g->ForeColor != ForeColor) {
		g->ForeColor = ForeColor;
		gcv.foreground = GetColor_RGB(g->ForeColor);
		gcm = GCForeground;
	    }
	    if (g->BackColor != BackColor) {
		g->BackColor = BackColor;
		gcv.background = GetColor_RGB(g->BackColor);
		gcm |= GCBackground;
	    }
	    if (gcm) ChangeGC(g->FontGC, gcm, &gcv);
	} else {
	    g->ForeColor = ForeColor;
	    g->BackColor = BackColor;
#ifdef USE_X11GC_plane_mask
	    gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth | 
		  GCForeground | GCBackground | GCFillStyle | GCFont;
	    gcv.plane_mask = AllPlanes;
#else
	    gcm = GCFunction | GCGraphicsExposures | GCLineWidth | 
		  GCForeground | GCBackground | GCFillStyle | GCFont;
#endif
	    gcv.function  = GXcopy;
	    gcv.font = font->xfont->fid;
	    gcv.fill_style = FillSolid;
	    gcv.foreground = GetColor_RGB(g->ForeColor);
	    gcv.background = GetColor_RGB(g->BackColor);
	    g->FontGC = CreateGC(gcm, &gcv);
	}
}

void ChangeFontGC(TFont *font, int ForeColor, int BackColor) {
	if (!(font->xfu & tf_isDefaultFont)) MakeTmpFontGC(font, &font->g, ForeColor, BackColor);
}

GC GetTmpFontGC(TFont *font, int ForeColor, int BackColor) {

	if (!font || !font->xfont) return NULL;
	if ((font->g.ForeColor == ForeColor) && (font->g.BackColor == BackColor)) {
	    return font->g.FontGC;
	}
	MakeTmpFontGC(font, &font->gtmp, ForeColor, BackColor);
	return font->gtmp.FontGC;
}

TFont *MakeFont(const char *fontname, const char *defaultfont, TFont *old_Font) {
  int i;
  int ForeColor, BackColor;
  char *fontpc = NULL;
  XFontStruct *pxfont = NULL;
  TFontStructCache *fsc;
  TFont *F;

	if (!fontname) fontname = "fixed";
	fontpc = GetFullFontName(fontname);
	if (!fontpc) {
	    printf("WARNING -- can t get font \"%s\", trying \"%s\"\n", fontname, defaultfont);
	    fontpc = GetFullFontName(defaultfont);
	}

	if (!fontpc) {
#ifdef _OS_WIN32
	    fontpc = pcdup("-*-system-*-*-*-*-*-*-*-*-*-*-*-*");
#else
	    fontpc = pcdup("fixed");
#endif
	    printf("WARNING -- can t get font \"%s\", trying \"%s\"\n", defaultfont, fontpc);
	}

	pxfont = NULL;

// printf("%s(%d) MakeFont: fontname=%s\n", FL, fontname);
// printf("%s(%d) MakeFont:          %s\n", FL, fontpc);

	fsc = appXDat.FontStructCache;
	for (i=0;i<FSCsitze;i++) {
	    if (fsc->xfont && pccmp(fsc->name, fontpc)) {
		xfree(fontpc);
		fontpc = fsc->name;
		fsc->count++;
		pxfont = fsc->xfont;
// printf("%s(%d) FontCache: count=%d %s\n", FL, fsc->count, fontpc);
		goto fsc_ok;
	    }
	    fsc++;
	}

	if (!pxfont) {
	  fsc = appXDat.FontStructCache;
	  for (i=0;i<FSCsitze;i++) {
	    if (!fsc->xfont) {
// printf("%s(%d) MakeFont: Add %d    %s\n", FL, i, fontpc);
	      fsc->name = fontpc;
	      fsc->count = 1;
	      pxfont = XLoadQueryFont(appXDat.dpy, fontpc);
// printf("%s(%d) MakeFont:  0x%08x   %s\n", FL, (int)pxfont, fontpc);
	      fsc->xfont = pxfont;
	      goto fsc_ok;
	    }
	    fsc++;
	  }
	}

fsc_ok:
	if (!pxfont) {
	  printf("ERROR -- can t get font \"%s\"\n", fontpc);
	  exit(EXIT_FAILURE);
	}

	ForeColor = 0x000000;
	BackColor = appXDat.Default_Color;
	if (old_Font) {
	    ForeColor = old_Font->g.ForeColor;
	    BackColor = old_Font->g.BackColor;
	    TFont_Destroy(old_Font);
	}

	F = xcalloc(sizeof(TFont));
	F->xfont = pxfont;
	F->height = pxfont->ascent + pxfont->descent;
	F->name = fontpc;
	MakeTmpFontGC(F, &F->g, ForeColor, BackColor);
// printf("%s(%d): 0x%x font=0x%x 0x%x 0x%x\n", FL, (int)F, (int)F->xfont, (int)F->xfont->ext_data, (int)F->xfont->per_char);
	return F;
}



//**************** X11 ****************

#ifdef _OS_WIN32
#define Utf8ToX11Unicode(a,b,c,d) Utf8ToWChars(a,b,c,d, 0)
#else
#define Utf8ToX11Unicode(a,b,c,d) Utf8ToWChars(a,b,c,d, UC_WCharBE)
#endif

int TextnWidth(XFontStruct *xfont, const char *pc, int lg, bool UTF8on) {
  int L, result;
  word  *pwc;
  word  wcs[256];

	if ((lg <= 0) || !pc) return 0;

	if (!UTF8on) return XTextWidth(xfont, pc, lg);

	if (lg < 256) {
	    L = 256;
	    pwc =  wcs;
	} else {
	    L = lg + 1;
	    pwc = xmalloc(L*2);
	}

	L = Utf8ToX11Unicode(pwc, L, pc, lg);
	if (!L) return 0;
	result = XTextWidth16(xfont, (XChar2b*)pwc, L);
	if (pwc !=  wcs) xfree(pwc);
	return result;
}

int TextWidth(XFontStruct *xfont, const char *pc, bool UTF8on) {
	return TextnWidth(xfont, pc, pclen(pc), UTF8on);
}

static void DrawString(int x, int y, int lg, const char *pc, TVDrawString *vds) {
  int L, xo, yo;
  GC gc;
  word  *pwc;
  word  wcs[256];

	if ((lg <= 0) || !pc || !vds) return;

	xo = vds->Xoff + x;
#ifdef _OS_WIN32
	yo = vds->Yoff + y;
#else
	yo = vds->Yoff + vds->TmpFont->xfont->ascent + y;
#endif
	gc = GetTmpFontGC(vds->TmpFont, vds->TmpForeColor, vds->TmpBackColor);

	if (!vds->UTF8on) {
	    if (vds->Image)
		XDrawImageString(appXDat.dpy, vds->WinID, gc, xo, yo, pc, lg);
	    else
		XDrawString(appXDat.dpy, vds->WinID, gc, xo, yo, pc, lg);
	    return;
	}

	if (lg < 256) {
	    L = 256;
	    pwc =  wcs;
	} else {
	    L = lg + 1;
	    pwc = xmalloc(L*2);
	}
	L = Utf8ToX11Unicode(pwc, L, pc, lg);
	if (L > 0) {
	  if (vds->Image)
	    XDrawImageString16(appXDat.dpy, vds->WinID, gc, xo, yo, (XChar2b*)pwc, L);
	  else
	    XDrawString16(appXDat.dpy, vds->WinID, gc, xo, yo, (XChar2b*)pwc, L);
	}
	if (pwc !=  wcs) xfree(pwc);
}

static int DrawStringSeg(int x, int y, int lgs, const char *pcs, TVDrawString *vds) {
  int i, lg, w, x0, xn, xt, yt;
  const char *pcsa;
  XFontStruct *xfont;

	if (!pcs) return x;
	xfont = vds->TmpFont->xfont;
	pcsa = pcs;
	if (vds->TabSpaceOn && (vds->TabSpaceWidth > 0)) {
	  for (i=0; i<lgs; i++) {
	    if (*pcs == 0x09) {
	      w = pcs - pcsa;
	      x0 = x;
	      if (w > 0) x0 = x + TextnWidth(xfont, pcsa, w, vds->UTF8on);
	      xn = (x0 / vds->TabSpaceWidth + 1) * vds->TabSpaceWidth;
	      if (!vds->OnlyLg && (xn >= -vds->Xoff)) {
		if ((w > 0) && (x0 >= -vds->Xoff)) {
		  DrawString(x, y, w, pcsa, vds);
		}
		xt = x0 + vds->Xoff;
		yt = y + vds->Yoff;
		FillRectRGB(vds->WinID, vds->TmpBackColor, xt, yt, xn - xt, vds->TmpFont->height);
		if (vds->TabSpaceDot) {
		  yt += xfont->ascent;
		  DrawLine(vds->WinID, appXDat.GC_000000, xt, yt, xt + 1, yt);
		  DrawLine(vds->WinID, appXDat.GC_000000, xt, yt - 1, xt, yt);
		}
	      }
	      x = xn;
	      pcsa = pcs + 1;
	    }
	    pcs++;
	  }
	  lg = pcs - pcsa;
	} else lg = lgs;

	if (lg > 0) {
	    x0 = x + TextnWidth(xfont, pcsa, lg, vds->UTF8on);
	    if (!vds->OnlyLg && (x0 >= -vds->Xoff)) {
		DrawString(x, y, lg, pcsa, vds);
	    }
	    x = x0;
	}
	return x;
}

int DrawColorString(int x, int y, TVDrawString *vds) {
  int x0, yo, L, it, lg, yh, xa, xe, a, e;
  TStrColorDat *scd;
  char *pc;
  bool UTF8on;

	if (!vds) return x;
	L = vds->lg;
	pc = vds->s;
	x0 = x;

	if (vds->scda && vds->scda->items) {
	    UTF8on = vds->UTF8on;
	    it = vds->scda->items;
	    scd = vds->scda->scd;
	    while ((it > 0) && (L > 0) && scd->s) {
		lg = scd->s - pc;

		if (lg > L) lg = L;
		if (lg > 0) {
		    vds->UTF8on = UTF8on;
		    vds->TmpFont = vds->Font;
		    vds->TmpForeColor = vds->ForeColor;
		    vds->TmpBackColor = vds->BackColor;
		    x = DrawStringSeg(x, y, lg, pc, vds);
		    L -= lg;
		    pc += lg;
		}
		lg = scd->lg;
		if (lg > L) lg = L;
		if (lg > 0) {
		    if (scd->fu & scdfu_override_utf8) {
			vds->UTF8on = (scd->fu & scdfu_UTF8on) != 0;
		    }
		    vds->TmpFont = scd->Font;
		    vds->TmpForeColor = scd->ForeColor;
		    vds->TmpBackColor = scd->BackColor;
		    x = DrawStringSeg(x, y, lg, pc, vds);
		    L -= lg;
		    pc += lg;
		}
		it--; scd++;
	    }
	    vds->UTF8on = UTF8on;
	}

	if (L > 0) {
	    vds->TmpFont = vds->Font;
	    vds->TmpForeColor = vds->ForeColor;
	    vds->TmpBackColor = vds->BackColor;
	    x = DrawStringSeg(x, y, L, pc, vds);
// printf("%dx%d %d %s %x %x %x\n", x, y, L, pc, vds->WinID, vds->TmpForeColor, vds->TmpBackColor);
	}
	if (vds->OnlyLg) return x;

	yo = y + vds->Yoff;
	if (vds->Wins && vds->ClsToEnd) {
	  yh = vds->TmpFont->height;
	  xa = x + vds->Xoff;
	  if (xa < 0) xa = 0;
	  e = vds->Wins->width - xa;
	  if (e > 0) {
	    if (vds->SelToEnd) {
		Fill_Rect(vds->WinID, appXDat.Select_GC_Back, xa, yo, e, yh);
	    } else {
		FillRectRGB(vds->WinID, vds->BackColor, xa, yo, e, yh);
	    }
	  }
	}

	if ((vds->SelA != vds->SelE) || vds->SelToEnd) {
	    xe = x + vds->Xoff;
	    xa = 0;
	    if (vds->SelToEnd != 2) {
		a = vds->SelA;
		e = vds->SelE;
		if (a > e) SWAP_INT(a, e); //{e = a; a = vds->SelE;}

		xa = GetChar_XPos(a, vds) + x0;
		if (!vds->SelToEnd) xe = GetChar_XPos(e, vds) + x0;
	    }
	    if (xe > xa)
		Fill_Rect(vds->WinID, appXDat.Select_GC, xa, yo, xe-xa, vds->TmpFont->height);
	} 

	return x;
}

int TextWidthColorString(TVDrawString *vds) {
  int res;
  bool OnlyLg = vds->OnlyLg;

	vds->OnlyLg = True;
	res = DrawColorString(0, 0, vds);
	vds->OnlyLg = OnlyLg;
	return res;
}

void DrawLineString(const char *s, int x, int y, int *max_w, int *max_h, TVDrawString *vds) {
  int mx, lg, m;
  const char *pc;

	*max_w = 0;
	*max_h = vds->Font->height;
	if (!vds || !s) return;
	vds->TmpFont = vds->Font;
	vds->TmpForeColor = vds->ForeColor;
	vds->TmpBackColor = vds->BackColor;

	mx = x;
	pc = s;
	while (*pc) {
	    if ((*pc == 10) || ((*pc == 13) && (pc[1] == 10))) {
		lg = pc - s;
		if (lg > 0) {
		    m = DrawStringSeg(x, y, lg, s, vds);
		    if (m > mx) mx = m;
		}
		if (*pc == 13) pc++;
		s = pc+1;
		y += vds->Font->height;
		*max_h += vds->Font->height;
	    }
	    pc++;
	}

	lg = pclen(s);
	if (lg > 0) {
	    m = DrawStringSeg(x, y, lg, s, vds);
	    if (m > mx) mx = m;
	}

	*max_w = mx;
}
