/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

//#include "appXbase.h"
#include "appbase.h"
#include "image.h"
#include "font_FT_pv.h"
#include <math.h>

static bool chk_cache(TFont_FT_pv *self, int *width, int *rows) {
  TFCache *fcache_glyph;
  FT_GlyphSlot  slot;
  int w, m, sbpl;

	fcache_glyph = self->fcache_glyph;
	if (fcache_glyph && fcache_glyph->data) {
	    w = fcache_glyph->bitmap_width;
	    *rows = fcache_glyph->bitmap_rows;
	    self->sb0 = fcache_glyph->data;
	} else {
	    slot  = self->face->glyph;
	    w     = slot->bitmap.width;
	    *rows = slot->bitmap.rows;
	    self->sb0 = slot->bitmap.buffer;
	}
	sbpl = w;
	if ((self->xo + w) > self->paint_img->width) w = self->paint_img->width - self->xo;
	if (self->fixes_size) {
	    sbpl = ((sbpl + 31) & ~31) >> 3;
	    self->fixes_o = 8;
	    if (self->xo < 0) {
		m = -self->xo;
		self->sb0 += (m & ~0x07) >> 3;
		self->fixes_o = self->xo & 0x07;
		if (!self->fixes_o) self->fixes_o = 8;
		self->xo = 0;
	    }
	    self->fixes_o_shift = 8 - self->fixes_o;
	} else {
	    if (self->xo < 0) {
		w += self->xo;
		self->sb0 -= self->xo;
		self->xo = 0;
	    }
	}
	self->sbpl = sbpl;
	*width = w;
	return w >= 0;
}

static void REGP3 glyph_argb16(TFont_FT_pv *self, TImage *img) {
  int x, y, yl, sbpl, width, rows;
  byte *sb0, *sb, *db0;
  word *dw;
  dword A, A2, R, G, B;

	if (!chk_cache(self, &width, &rows)) return;
	sbpl = self->sbpl;
	sb0 = self->sb0;

	yl = self->yo;
	db0 = img->data + (yl * img->bytes_per_line);
	for (y=rows;y>0;y--) {
	    if ((dword)yl < (dword)img->height) {
		sb = sb0;
		dw = ((word*)db0) + self->xo;
		for (x=width;x>0;x--) {
		    A = *sb;
		    if (A >= 0xf8) {
			*dw = self->col;
		    } else if (A >= 0x08) {
			A2 = 0x100 - A;
			R = ((self->R * A) + ((*dw & 0xf800) * A2)) & 0xf80000;
			G = ((self->G * A) + ((*dw & 0x07e0) * A2)) & 0x07e000;
			B = ((self->B * A) + ((*dw & 0x001f) * A2)) & 0x001f00;
			*dw = (R | G | B) >> 8;
		    }
		    sb++; dw++;
		}
	    }
	    sb0 += sbpl;
	    db0 += img->bytes_per_line;
	    yl++;
	}
}

static void REGP3 glyph_fixed16(TFont_FT_pv *self, TImage *img) {
  int x, y, sbpl, dbpl, bc, width, rows;
  byte *sb0, *sb, *db0;
  word *dw;
  dword sc, yl, dmaxH;

	if (!chk_cache(self, &width, &rows)) return;
	sbpl = self->sbpl;
	sb0 = self->sb0;

	dmaxH = (dword)img->height;
	dbpl = img->bytes_per_line;
	yl = (dword)self->yo;

	db0 = img->data + (yl * dbpl);
	for (y=rows;y>0;y--) {
	    if (yl < dmaxH) {
		sb = sb0;
		dw = ((word*)db0) + self->xo;
		bc = self->fixes_o;
		sc = *sb; sb++;
		if (self->fixes_o_shift) sc <<= self->fixes_o_shift;
		for (x=width;x>0;x--) {
		    if (!bc) {sc = *sb; sb++; bc = 8;}
		    if (sc & 0x80) *dw = self->col;
		    sc <<= 1; bc--; dw++;
		}
	    }
	    sb0 += sbpl; db0 += dbpl; yl++;
	}
}

static void REGP3 glyph_argb32(TFont_FT_pv *self, TImage *img) {
  int x, y, yl, sbpl, width, rows;
  byte *sb0, *sb, *db0;
  dword *dw;
  dword A, A2, R, G, B;

	if (!chk_cache(self, &width, &rows)) return;
	sbpl = self->sbpl;
	sb0 = self->sb0;

	yl = self->yo;
	db0 = img->data + (yl * img->bytes_per_line);
	for (y=rows;y>0;y--) {
	    if ((dword)yl < (dword)img->height) {
		sb = sb0;
		dw = ((dword*)db0) + self->xo;
		for (x=width;x>0;x--) {
		    A = *sb;
		    if (A >= 0xf8) {
			*dw = self->col;
		    } else if (A >= 0x08) {
			A2 = 0x100 - A;
			R = ((self->R * A) + ((*dw & 0xff0000) * A2)) & 0xff000000;
			G = ((self->G * A) + ((*dw & 0x00ff00) * A2)) & 0x00ff0000;
			B = ((self->B * A) + ((*dw & 0x0000ff) * A2)) & 0x0000ff00;
			*dw = (R | G | B) >> 8;
		    }
		    sb++; dw++;
		}
	    }
	    sb0 += sbpl;
	    db0 += img->bytes_per_line;
	    yl++;
	}
}

static void REGP3 glyph_fixed32(TFont_FT_pv *self, TImage *img) {
  int x, y, sbpl, dbpl, bc, width, rows;
  byte *sb0, *sb, *db0;
  dword *dw;
  dword sc, yl, dmaxH, col;

	if (!chk_cache(self, &width, &rows)) return;
	sbpl = self->sbpl;
	sb0 = self->sb0;

	col = self->col;
	dmaxH = (dword)img->height;
	dbpl = img->bytes_per_line;
	yl = (dword)self->yo;

	db0 = img->data + (yl * dbpl);
	for (y=rows;y>0;y--) {
	    if (yl < dmaxH) {
		sb = sb0;
		dw = ((dword*)db0) + self->xo;
		bc = self->fixes_o;
		sc = *sb; sb++;
		if (self->fixes_o_shift) sc <<= self->fixes_o_shift;
		for (x=width;x>0;x--) {
		    if (!bc) {sc = *sb; sb++; bc = 8;}
		    if (sc & 0x80) *dw = col;
		    sc <<= 1; bc--; dw++;
		}
	    }
	    sb0 += sbpl; db0 += dbpl; yl++;
	}
}

static void Add_GlyphToCache(TFont_FT_pv *self, int glyph_index) {
  TFCache *fcache;
  FT_GlyphSlot slot;
  int i;

	slot = self->face->glyph;
	fcache = &self->fcache[glyph_index];
	fcache->advance = slot->advance.x;
	if ((fcache->advance & 0x3f) == 0) {
	    fcache->bitmap_width = slot->bitmap.width;
	    fcache->bitmap_rows = slot->bitmap.rows;
	    fcache->bitmap_left = slot->bitmap_left;
	    fcache->bitmap_top  = slot->bitmap_top;
	    if (self->fixes_size) {
		i = (((fcache->bitmap_width + 31) & ~31) >> 3);
	    } else {
		i = fcache->bitmap_width;
	    }
	    fcache->data = pndup(slot->bitmap.buffer, i * fcache->bitmap_rows);
	}
}

static bool Load_Glyph(TFont_FT_pv *self, dword c) {
  int glyph_index;
  bool ok;

	self->fcache_glyph = NULL;
	glyph_index = FT_Get_Char_Index(self->face, c);
	ok = self->fcache && (self->fcache_max > glyph_index);
	if (ok && (self->fcache[glyph_index].advance >= 0)) {
	    self->fcache_glyph = &self->fcache[glyph_index];
	    if (self->fcache_glyph->data) return true;
	}

	if (FT_Load_Glyph(self->face, glyph_index, FT_LOAD_RENDER)) return false;
	if (self->fcache && (self->fcache_max > glyph_index) && 
	    (self->fcache[glyph_index].advance < 0))
	{
	    Add_GlyphToCache(self, glyph_index);
	}
	return true;
}

static bool TFont_FT_ToImage(TFont_FT_pv *self, TImage *img, int x, int y, 
			const char *s, int slen) {
  int i, L;
  FT_GlyphSlot  slot;
  TFCache *fcache_glyph;
  bool rgb565;
  word  *pwc;
  word  wcs[256];

	if (!self || !img) return false;

	if ((img->width <= 0) || 
	    (img->height <= 0) ||
	    (x >= img->width) ||
	    ((y + self->font_height) < 0) || 
	    (y >= (img->height + self->font_height))) return false;

	if (slen <= 0) return true;
	self->paint_img = img;

	rgb565 = (img->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp;
	if (rgb565) {
	    self->to_img = glyph_argb16;
	    if (self->fixes_size) self->to_img = glyph_fixed16;
	    self->R = (self->Color >> 8) & 0xf800;
	    self->G = (self->Color >> 3) & 0x07e0;
	    self->B = (self->Color >> 3) & 0x001f;
	    self->col = self->R | self->G | self->B;
	} else {
	    self->to_img = glyph_argb32;
	    if (self->fixes_size) self->to_img = glyph_fixed32;
	    self->col = self->Color;
	    self->R = self->col & 0xff0000;
	    self->G = self->col & 0x00ff00;
	    self->B = self->col & 0xff;
	}

	slot = self->face->glyph;
	if (self->UTF8on) {
	    if (slen < 256) {
		L = 256;
		pwc =  wcs;
	    } else {
		L = slen + 1;
		pwc = xmalloc(L*2);
	    }
	    L = Utf8ToWChars(pwc, L, s, slen, 0);
	    if (L <= 0) return true;
	    word *cpc = pwc;
	    if (self->use_angle) {
		self->fcache_glyph = NULL;
		for (i=0;i<L;i++) {
		    FT_Set_Transform(self->face, &self->matrix, &self->pen);
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			self->xo = x  + slot->bitmap_left;
			self->yo = y - slot->bitmap_top;
			if (self->yo < img->height) {
			    self->to_img(self, img);
			}
			self->pen.x += slot->advance.x;
			self->pen.y += slot->advance.y;
		    }
		    cpc++;
		}
	    } else {
		for (i=0;i<L;i++) {
#if 1
		    if (Load_Glyph(self, *cpc)) {
			int bitmap_left, bitmap_top, advance;
			fcache_glyph = self->fcache_glyph;
			if (fcache_glyph && fcache_glyph->data) {
			    bitmap_left = fcache_glyph->bitmap_left;
			    bitmap_top = fcache_glyph->bitmap_top;
			    advance = fcache_glyph->advance;
			} else {
			    bitmap_left = slot->bitmap_left;
			    bitmap_top = slot->bitmap_top;
			    advance = slot->advance.x;
			}
			self->xo = x + (self->pen.x / 64) + bitmap_left;
			self->yo = y + (self->pen.y / 64) - bitmap_top;
			if (self->yo < img->height) {
			    self->to_img(self, img);
			}
			self->pen.x += advance;
		    }
#else
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			self->xo = x + (self->pen.x / 64) + slot->bitmap_left;
			self->yo = y + (self->pen.y / 64) - slot->bitmap_top;
			if (self->yo < img->height) {
			    self->to_img(self, img);
			}
			self->pen.x += slot->advance.x;
//			self->pen.y += slot->advance.y;
		    }
#endif
		    cpc++;
		}
	    }
	    if (pwc !=  wcs) xfree(pwc);

	} else {
	    const byte *cpc = (const byte*)s;
	    if (self->use_angle) {
		self->fcache_glyph = NULL;
		for (i=0;i<slen;i++) {
		    FT_Set_Transform(self->face, &self->matrix, &self->pen);
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			self->xo = x  + slot->bitmap_left;
			self->yo = y - slot->bitmap_top;
			if (self->yo < img->height) {
			    self->to_img(self, img);
			}
			self->pen.x += slot->advance.x;
			self->pen.y += slot->advance.y;
		    }
		    cpc++;
		}
	    } else {
		for (i=0;i<slen;i++) {
#if 1
		    if (Load_Glyph(self, *cpc)) {
			int bitmap_left, bitmap_top, advance;
			fcache_glyph = self->fcache_glyph;
			if (fcache_glyph && fcache_glyph->data) {
			    bitmap_left = fcache_glyph->bitmap_left;
			    bitmap_top = fcache_glyph->bitmap_top;
			    advance = fcache_glyph->advance;
			} else {
			    bitmap_left = slot->bitmap_left;
			    bitmap_top = slot->bitmap_top;
			    advance = slot->advance.x;
			}
			self->xo = x + (self->pen.x / 64) + bitmap_left;
			self->yo = y + (self->pen.y / 64) - bitmap_top;
			if (self->yo < img->height) {
			    self->to_img(self, img);
			}
			self->pen.x += advance;
		    }
#else
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
// printf("%s(%d): %d %d %d %d\n", FL, slot->bitmap_left, slot->bitmap_top, slot->bitmap.width, slot->bitmap.rows);
			self->xo = x + (self->pen.x / 64) + slot->bitmap_left;
			self->yo = y + (self->pen.y / 64) - slot->bitmap_top;
			if (self->yo < img->height) {
			    self->to_img(self, img);
			}
			self->pen.x += slot->advance.x;
//			self->pen.y += slot->advance.y;
		    }
#endif
		    cpc++;
		}
	    }
	}
	return true;
}

static int GetAdvance(TFont_FT_pv *self, dword c) {
  int glyph_index, advance;
  bool ok;

	glyph_index = FT_Get_Char_Index(self->face, c);
	ok = self->fcache && (self->fcache_max > glyph_index);
	if (ok && (self->fcache[glyph_index].advance >= 0))
	    return self->fcache[glyph_index].advance;

	advance = 0;
	if (!FT_Load_Glyph(self->face, glyph_index, FT_LOAD_RENDER))
	    advance = self->face->glyph->advance.x;

	if (ok) self->fcache[glyph_index].advance = advance;
	return advance;
}

static bool TFont_FT_GetVector(TFont_FT_pv *self, float *x, float *y, 
			const char *s, int slen) {
  int i, L;
  FT_GlyphSlot  slot;
  FT_Vector pen;
  word  *pwc;
  word  wcs[256];

	if (!self) return false;
	if (slen <= 0) slen = pclen(s);
	if (slen <= 0) return true;

	pen.x = 0;
	pen.y = 0;
	if (x) pen.x = (FT_Fixed)(*x * 64);
	if (y) pen.y = (FT_Fixed)(*y * 64);
	if (self->use_angle) FT_Set_Transform(self->face, &self->matrix, &pen);

	slot = self->face->glyph;
	if (self->UTF8on) {
	    if (slen < 256) {
		L = 256;
		pwc =  wcs;
	    } else {
		L = slen + 1;
		pwc = xmalloc(L*2);
	    }
	    L = Utf8ToWChars(pwc, L, s, slen, 0);
	    if (L <= 0) return true;
	    word *cpc = pwc;
	    if (self->use_angle) {
		for (i=0;i<L;i++) {
		    FT_Set_Transform(self->face, &self->matrix, &pen);
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			pen.x += slot->advance.x;
			pen.y += slot->advance.y;
		    }
		    cpc++;
		}
	    } else {
		for (i=0;i<L;i++) {
#if 1
		    pen.x += GetAdvance(self, *cpc);
#else
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			pen.x += slot->advance.x;
			pen.y += slot->advance.y;
		    }
#endif
		    cpc++;
		}
	    }
	    if (pwc !=  wcs) xfree(pwc);

	} else {
	    const char *cpc = s;
	    if (self->use_angle) {
		for (i=0;i<slen;i++) {
		    FT_Set_Transform(self->face, &self->matrix, &pen);
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			pen.x += slot->advance.x;
			pen.y += slot->advance.y;
		    }
		    cpc++;
		}
	    } else {
		for (i=0;i<slen;i++) {
#if 1
		    pen.x += GetAdvance(self, *cpc);
#else
		    if (!FT_Load_Char(self->face, *cpc, FT_LOAD_RENDER)) {
			pen.x += slot->advance.x;
			pen.y += slot->advance.y;
		    }
#endif
		    cpc++;
		}
	    }
	}
	if (self->use_angle) FT_Set_Transform(self->face, &self->matrix, &self->pen);

	if (x) *x = (float)pen.x / 64.0;
	if (y) *y = (float)pen.y / 64.0;
	return true;
}

static int TFont_FT_GetWidth(TFont_FT_pv *self, const char *s, int slen) {
  float x;
	x = 0;
	if (!TFont_FT_GetVector(self, &x, NULL, s, slen)) return 0;
	return (int) x;
}

static void SetMatrix(TFont_FT_pv *self, float angle) {
  float gw;
	self->angle = angle;
	self->use_angle = ((angle > 0.01) || (angle < -0.01));
	gw = angle * M_PI_2 / 90;
	self->matrix.xx = (FT_Fixed)( cosf(gw) * 0x10000);
	self->matrix.xy = (FT_Fixed)(-sinf(gw) * 0x10000);
	self->matrix.yx = (FT_Fixed)( sinf(gw) * 0x10000);
	self->matrix.yy = (FT_Fixed)( cosf(gw) * 0x10000);
}

static bool TFont_FT_SetColor(TFont_FT_pv *self, dword Color) {
	self->Color = Color;
	return true;
}

static bool TFont_FT_SetBackColor(TFont_FT_pv *self, dword Color) {
	self->BackColor = Color;
	return true;
}

static bool TFont_FT_SetUTF8(TFont_FT_pv *self, bool UTF8on) {
	self->UTF8on = UTF8on;
	return true;
}

static bool TFont_FT_SetAngle(TFont_FT_pv *self, float angle) {
//	if (!self || self->fixes_size) return False;
	if (self->fixes_size) return False;
	if (self->angle != angle) {
	    SetMatrix(self, angle);
	}
	return true;
}

static bool TFont_FT_ClrOrigin(TFont_FT_pv *self) {
//	if (!self) return False;
	self->pen.x = 0;
	self->pen.y = 0;
	FT_Set_Transform(self->face, &self->matrix, &self->pen);
	return true;
}

static bool TFont_FT_SetOrigin(TFont_FT_pv *self, float x, float y) {
//	if (!self) return False;
	self->pen.x = (FT_Fixed)(x * 64);
	self->pen.y = (FT_Fixed)(y * 64);
	FT_Set_Transform(self->face, &self->matrix, &self->pen);
	return true;
}

static bool TFont_FT_GetOrigin(TFont_FT_pv *self, float *x, float *y) {
//	if (!self) return False;
	if (x) *x = (float)self->pen.x / 64.0;
	if (y) *y = (float)self->pen.y / 64.0;
	return true;
}

static bool TFont_FT_GetMetrics(TFont_FT_pv *self, int *height, int *ascent) {
	if (height) *height = self->metrics_height;
	if (ascent) *ascent = self->metrics_ascent;
	return true;
}

static long TFont_FT_GetGlyphs(TFont_FT_pv *self) {
	return self->glyphs;
}
/*_____________________________ Destroy _________________________________*/
static bool TFont_FT_pv__Destroy(TFont_FT_pv *self) {
  int i;
  TFCache *fcache;

	fcache = self->fcache;
	if (fcache) {
	    for (i=self->fcache_max;i>0;i--) {
		if (fcache->data) xfree(fcache->data);
		fcache++;
	    }
	    xfree(self->fcache);
	}
	if (self->face) FT_Done_Face(self->face);
	if (self->library) FT_Done_FreeType(self->library);
	return False;
}
/*____________________________ Create ___________________________________*/
void *TFont_FT_pv_Create(const char *FontFileName, int font_height, int font_dpi) {
  TFont_FT_pv *self;
  FT_Error err;
  int i;
  TFCache *fcache;

	self = ttc_create(sizeof(TFont_FT_pv), wtyp_Font_FT);
	self->destroy = (TDESTROY)TFont_FT_pv__Destroy;

	err = FT_Init_FreeType(&self->library);  /* initialize library */
	if (err) {
	    printf("%s(%d): ERROR: FT library initialization\n", FL);
end_err:
	    TFont_FT_pv__Destroy(self);
	    xfree(self);
	    return NULL;
	}

	err = FT_New_Face(self->library, FontFileName, 0, &self->face);  
	if (err == FT_Err_Unknown_File_Format) {
	    printf("%s(%d): ERROR: FT font format is unsupported %s\n", FL, FontFileName);
	    goto end_err;
	} else if (err) {
	    printf("%s(%d): ERROR: FT file could not opened %s\n", FL, FontFileName);
	    goto end_err;
	}



	self->fixes_size = FT_IS_SCALABLE(self->face) == 0;

//	if (!self->fixes_size) {
	    if (font_height <= 0) {
		font_height = 12;
		if (self->fixes_size) font_height = self->face->available_sizes->y_ppem / 64;
	    }
	    self->font_height = font_height;
	    if (font_dpi > 0) {
		self->font_dpi = font_dpi;
		err = FT_Set_Char_Size(self->face, font_height * 64, 0, font_dpi, 0);
	    } else {
		err = FT_Set_Pixel_Sizes(self->face, 0, font_height);
	    }
	    if (err) {
		printf("%s(%d): ERROR: FT Set Pixel Sizes\n", FL);
    //	    goto end_err;
	    }
// printf("%s(%d): glyphs=%ld flags=0x%04lx %ld %ld\n", FL, self->face->num_glyphs, self->face->face_flags, self->face->size->metrics.height / 64, self->face->size->metrics.ascender / 64);
/*	} else {
// FT_Load_Char(self->face, 0x20, FT_LOAD_RENDER);
//		err = FT_Set_Pixel_Sizes(self->face, 0, 0);
//	    if (err) printf("%s(%d): ERROR: FT Set Pixel Sizes\n", FL);
 printf("%s(%d): glyphs=%ld flags=0x%04lx %d %d\n", FL, self->face->num_glyphs, self->face->face_flags, self->face->available_sizes->y_ppem / 64, self->face->ascender);
	} */

	self->glyphs = self->face->num_glyphs;
	self->metrics_height = self->face->size->metrics.height / 64;
	self->metrics_ascent = self->face->size->metrics.ascender / 64;

	SetMatrix(self, 0.0);



#if TFCACHE_MAX == 0
	self->fcache_max = 0;
#else
#if TFCACHE_MAX < 128
# error "TFCACHE_MAX < 128"
#endif
	self->fcache_max = self->glyphs;
	if (self->fcache_max < 0) self->fcache_max = 0;
	if (self->fcache_max > TFCACHE_MAX) self->fcache_max = TFCACHE_MAX;
	if (self->fcache_max > 0) {
	    self->fcache = xmalloc(sizeof(TFCache) * self->fcache_max);
	    fcache = self->fcache;
	    for (i=self->fcache_max;i>0;i--) {fcache->advance = -1; fcache++;}
	    for (i=0x20;i<0x80;i++) Load_Glyph(self, i);
#if TFCACHE_MAX > 191
	    for (i=0xa0;i<0x100;i++) Load_Glyph(self, i);
#endif
	}
#endif
	self->SetColor = TFont_FT_SetColor;
	self->SetBackColor = TFont_FT_SetBackColor;
	self->SetUTF8 = TFont_FT_SetUTF8;
	self->ToImage = TFont_FT_ToImage;
	self->SetAngle = TFont_FT_SetAngle;
	self->ClrOrigin = TFont_FT_ClrOrigin;
	self->SetOrigin = TFont_FT_SetOrigin;
	self->GetOrigin = TFont_FT_GetOrigin;
	self->GetWidth = TFont_FT_GetWidth;
	self->GetVector = TFont_FT_GetVector;
	self->GetMetrics = TFont_FT_GetMetrics;
	self->GetGlyphs = TFont_FT_GetGlyphs;

	return self;
}
