/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "appXbase.h"

//*********************** FomtArt ************************************
typedef struct TScanXLD_Art {
  const char *a, *b;
} TScanXLD_Art;

static const TScanXLD_Art ScanAllArtTab[] = {
 {"*", "*"},
 {NULL, NULL}};

static const TScanXLD_Art ScanMediumTab[] = {
#ifndef _OS_WIN32
 {"regular", "r"},
#endif
 {"medium", "r"},
 {"normal", "r"},
 {NULL, NULL}};

static const TScanXLD_Art ScanMediumKTab[] = {
#ifndef _OS_WIN32
 {"regular", "i"},
#endif
 {"medium", "o"},
 {"medium", "i"},
 {"normal", "i"},
 {NULL, NULL}};

static const TScanXLD_Art ScanBoldTab[] = {
 {"bold", "r"},
#ifndef _OS_WIN32
 {"demibold", "r"},
#endif
 {NULL, NULL}};

static const TScanXLD_Art ScanBoldKTab[] = {
 {"bold", "o"},
 {"bold", "i"},
#ifndef _OS_WIN32
 {"demibold", "i"},
#endif
 {NULL, NULL}};

/*
FontNameRegistry-Foundry-FamilyName-WeightName-Slant-
SetwidthName-AddStyleName-PixelSize-PointSize-
ResolutionX-ResolutionY-Spacing-AverageWidth-
CharSetRegistry-CharSetCoding

3  r=Roman (no slant) i=Italic (slant left) o=Oblique (slant left)
   ri=Reverse italic (slant right)
   ro=Reverse oblique (slant right) 

10 m = monospaced (fixed pitch), 
   p = proportional (variable pitch), 
   c = cell

*/

//-adobe-times-medium-r-normal--18-180-75-75-p-94-iso8859-1
//-misc-fixed-bold-r-normal-*-13-120-75-75-c-70-iso10646-1
// 0    1     2    3 4      5 6  7   8  9  1 1  1        1
//                                         0 1  2        3
bool GetFontSXTypeIdx(int idx, const char *sx, const char **ret_pc, int *ret_lg) {
  int i, lg;
  const char *cpc;
//  Result := False;
	*ret_pc = NULL;
	*ret_lg = 0;
	cpc = sx;
	if (!cpc) return False;
	i = -1;
	while (*cpc) {
	    if (*cpc == '-') {i++; if (i == idx) break;}
	    cpc++;
	}
	if ((i == idx) && (*cpc == '-')) {
	    cpc++;
	    *ret_pc = cpc;
	    lg = 0;
	    while (*cpc && (*cpc != '-')) {cpc++; lg++;}
	    *ret_lg = lg;
	    return True;
	}
	return False;
}

bool SetFontSXTypeIdx(int idx, const char *sx, const char *sins, char **ret_sx) {
  int lg = 0;
  const char *cpc;
  char *s;

	*ret_sx = NULL;
	if (!GetFontSXTypeIdx(idx, sx, &cpc, &lg)) return False;
	s = pcndup(sx, cpc-sx);
	cpc += lg;
	if (sins) ppcadd(&s, sins);
	ppcadd(&s, cpc);
	*ret_sx = s;
	return True;
}

bool SetFontSXVTypeIdx(int idx, char **sx, const char *sins) {
  char *s;
	if (!SetFontSXTypeIdx(idx, *sx, sins, &s)) return False;
	xfree(*sx);
	*sx = s;
	return True;
}

bool ChkScaleFont(const char *sx) {
  int L;
  const char *tsx;
	GetFontSXTypeIdx(8, sx, &tsx, &L);
	return ((L == 1) && (*tsx == '0'));
}

bool ChkFixedFont(const char *sx) {
  int L;
  const char *tsx;
	GetFontSXTypeIdx(10, sx, &tsx, &L);
	return ((L == 1) && ((*tsx == 'm') || (*tsx == 'c')));
}

static bool chk_pxld(const TScanXLD_Art *pxld, const char *a, const char *b) {
	while (pxld->a) {
	    if (pccmp(pxld->a, a) && pccmp(pxld->b, b)) return True;
	    pxld++;
	}
	return False;
}

int GetFont_Art(const char *sx) {
  int L;
  const char *tsx;
  char wname[16];
  char slant[4];

	if (!GetFontSXTypeIdx(2, sx, &tsx, &L)) return 0;
	if (L > 15) return 0;
	pcnmove(wname, tsx, L);
	if (!GetFontSXTypeIdx(3, sx, &tsx, &L)) return 0;
	if (L > 3) return 0;
	pcnmove(slant, tsx, L);
	if (chk_pxld(ScanMediumTab, wname, slant)) return FONT_Medium;
	if (chk_pxld(ScanMediumKTab,wname, slant)) return FONT_MediumK;
	if (chk_pxld(ScanBoldTab,   wname, slant)) return FONT_Bold;
	if (chk_pxld(ScanBoldKTab,  wname, slant)) return FONT_BoldK;
	return 0;
}

int GetFont_Size(const char *sx) {
  int L;
  const char *tsx;
	GetFontSXTypeIdx(6, sx, &tsx, &L);
	return pctoint(tsx);
}

char *GetFont_NextSize(const char *sx, int size) {
  int i, L, min, max, min_i, max_i, idx, L2, v;
  char *tsx, *pc, **FontNameList;
  const char *cpc;

	if (ChkScaleFont(sx) || !sx) return NULL;
	SetFontSXTypeIdx(6, sx, "*", &tsx);
	SetFontSXVTypeIdx(7, &tsx, "*");
	SetFontSXVTypeIdx(11, &tsx, "*");
	FontNameList = XListFonts(appXDat.dpy, tsx, 99999, &L);
	xfree(tsx);
	if (!FontNameList) return NULL;

	min = -99999;
	max = 99999;
	max_i = min_i = 0;
	idx = -1;
	for (i=0;i<L;i++) {
	    pc = FontNameList[i];
	    GetFontSXTypeIdx(11, pc, &cpc, &L2); //chk scale Font
	    if ((L2 != 1) || (*cpc != '0')) {
		v = GetFont_Size(pc) - size;
		if (v > 0) {
		    if (v < max) {max = v; max_i = i;}
		} else {
		    if (v > min) {min = v; min_i = i;}
		}
	    }
	}
	if ((max + min) >= 0) {
	    if (min != -99999) idx = min_i;
	} else {
	    if (max != 99999) idx = max_i;
	}
//  printf("%s(%d): -> %d %d  size=%d idx=%d\n", FL, min, max, size, idx);
	pc = NULL;
	if (idx >= 0) pc = pcdup(FontNameList[idx]);
	XFreeFontNames(FontNameList);
	return pc;
}

//********************************************************************
static bool sxSet23(const char *sx, const char *a, const char *b, char **ret_sx) {
	*ret_sx = NULL;
	if (!sx) return False;
	SetFontSXTypeIdx(2, sx, a, ret_sx);
	SetFontSXVTypeIdx(3, ret_sx, b);
// printf(">>%s\n", *ret_sx);
	return True;
}

static bool Test_sxSet23(const char *sx, const char *a, const char *b, char **ret_sx) {
  char *tsx;
  bool res = False;
	if (!sxSet23(sx, a, b, &tsx)) return False;
	res = FontExists(tsx);
	if (ret_sx && res) *ret_sx = tsx; else xfree(tsx);
	return res;
}

static bool GetXLD_Art(const char *sx, const TScanXLD_Art *pxld, char **ret_sx, int fu) {
  int L, size;
  char *tsx, *pc;
  const char *cpc;
  const TScanXLD_Art *px = pxld;

	while (px->a) {
	    if (Test_sxSet23(sx, px->a, px->b, ret_sx)) {
		return True;
	    }
	    px++;
	}
	if ((fu & FONT_NoResize) || ChkScaleFont(sx)) return False;
  //find next font size 
	GetFontSXTypeIdx(6, sx, &cpc, &L);
	size = pctoint(cpc);

	pc = NULL;
	SetFontSXTypeIdx(6, sx, "*", &tsx);
	px = pxld;
	while (px->a) {
	    if (Test_sxSet23(tsx, px->a, px->b, &pc)) break;
	    px++;
	}
	xfree(tsx);

	tsx = GetFont_NextSize(pc, size);
	xfree(pc);
	if (!tsx) return False;

	if (ret_sx) *ret_sx = tsx; else  xfree(tsx);
	return True;
}

bool GetFontArtName(const char *sx, int font_art, char **ret_sx) {
  const TScanXLD_Art *pxld = ScanAllArtTab;
  char *tsx = pcdup(sx);
  bool res;

	if (!ChkFixedFont(tsx)) SetFontSXVTypeIdx(11, &tsx, "*");

	switch (font_art & FONT_SMask) {
	case FONT_Medium:  pxld = ScanMediumTab; break;
	case FONT_MediumK: pxld = ScanMediumKTab; break;
	case FONT_Bold:    pxld = ScanBoldTab; break;
	case FONT_BoldK:   pxld = ScanBoldKTab; break;
	}
	res = GetXLD_Art(tsx, pxld, ret_sx, font_art);
	xfree(tsx);
	return res;
}

bool MakeFontArt(const char *FontName, TFont **ppFont, int font_art) {
  TFont *font;
  char *pc;
	if (!ppFont) return False;
	font = *ppFont;
	if (!FontName && font) FontName = font->name;
	if (!FontName) return False;
	if (!GetFontArtName(FontName, font_art, &pc)) {
	    *ppFont = MakeFont(FontName, NULL, font);
	    return False;
	}
	*ppFont = MakeFont(pc, FontName, font);
	xfree(pc);
	return True;
}
