/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include <X11/extensions/Xinerama.h>
#include <X11/extensions/Xrandr.h>
#include <dlfcn.h>

static int dpyMonitorsStat;

static void *LibXineramaHandle;

Bool (*dl_XineramaIsActive)(Display *dpy);
XineramaScreenInfo *(*dl_XineramaQueryScreens)(Display *dpy, int *num);

/*
static void *LibXrandrHandle;

XRRScreenResources *(*dl_XRRGetScreenResourcesCurrent)(Display *dpy, Window window);
void(*dl_XRRFreeScreenResources)(XRRScreenResources *resources);

XRROutputInfo *(*dl_XRRGetOutputInfo)(Display *dpy, XRRScreenResources *resources, RROutput output);
void (*dl_XRRFreeOutputInfo)(XRROutputInfo *outputInfo);

XRRCrtcInfo *(*dl_XRRGetCrtcInfo)(Display *dpy, XRRScreenResources *resources, RRCrtc crtc);
void(*dl_XRRFreeCrtcInfo)(XRRCrtcInfo *crtcInfo);
*/

void MonitorsInit() {
  Display *dpy;
  TMonitor *mon;
  XineramaScreenInfo *ScreenInfo, *si;
  int i, count;
//  dword t = GetTime_us();

	dpy = appXDat.dpy;
#ifdef Test_Root_WinId
	if (TestRootWinId) {
	    OrgRootWinId = appXDat.RootW;
	    appXDat.RootW = TestRootWinId;
	    appXDat.RootW_depth = DefaultDepth(dpy, appXDat.default_screen);
	    XGetWindowAttributes(dpy, appXDat.RootW, &appXDat.Root_attr);
	    appXDat.Root_width = appXDat.Root_attr.width;
	    appXDat.Root_height = appXDat.Root_attr.height;
	} else {
#endif
	    appXDat.Root_width = DisplayWidth(dpy, appXDat.default_screen);
	    appXDat.Root_height = DisplayHeight(dpy, appXDat.default_screen);
	    appXDat.RootW_depth = DefaultDepth(dpy, appXDat.default_screen);
	    XGetWindowAttributes(dpy, appXDat.RootW, &appXDat.Root_attr);
#ifdef Test_Root_WinId
	}
#endif
	dpyMonitors.count = 1;
	dpyMonitors.PrimaryIdx = 0;
	mon = dpyMonitors.mon;
//	mon->mon_id = 0;
	mon->left = 0;
	mon->top = 0;
	mon->right = appXDat.Root_width;
	mon->bottom = appXDat.Root_height;
#ifdef Test_Root_WinId
	if (TestRootWinId) goto noXinerama;
#endif

	if (dpyMonitorsStat == 2) goto noXinerama;

	if (!dpyMonitorsStat) {
	    dpyMonitorsStat = 1;

/* 	    LibXrandrHandle = dlopen("libXrandr.so.2", RTLD_LAZY);
	    if (LibXrandrHandle) {
		dl_XRRGetScreenResourcesCurrent = load_lib(LibXrandrHandle, "XRRGetScreenResourcesCurrent");
		dl_XRRFreeScreenResources       = load_lib(LibXrandrHandle, "XRRFreeScreenResources");
		dl_XRRGetOutputInfo  = load_lib(LibXrandrHandle, "XRRGetOutputInfo");
		dl_XRRFreeOutputInfo = load_lib(LibXrandrHandle, "XRRFreeOutputInfo");
		dl_XRRGetCrtcInfo    = load_lib(LibXrandrHandle, "XRRGetCrtcInfo");
		dl_XRRFreeCrtcInfo   = load_lib(LibXrandrHandle, "XRRFreeCrtcInfo");
		if (!dl_XRRGetScreenResourcesCurrent) {
		    dlclose(LibXrandrHandle);
		    LibXrandrHandle = NULL;
		}
	    } */

	    LibXineramaHandle = dlopen("libXinerama.so.1", RTLD_LAZY);
	    if (LibXineramaHandle) {
		dl_XineramaIsActive     = load_lib(LibXineramaHandle, "XineramaIsActive");
		dl_XineramaQueryScreens = load_lib(LibXineramaHandle, "XineramaQueryScreens");
		if (!dl_XineramaIsActive || !dl_XineramaQueryScreens) goto noXinerama;
	    }
	}


	if (!dl_XineramaQueryScreens || !dl_XineramaIsActive(dpy)) {
noXinerama:
	    if (LibXineramaHandle) {
//		dlclose(LibXineramaHandle);  ->Crash in XServer (libXcursor.so.1.0)
		LibXineramaHandle = NULL;
	    }
	    dl_XineramaQueryScreens = NULL;
	    dpyMonitors.stat = 0;
	    dpyMonitorsStat = 2;
	    return;
	}

	count = 0;
	ScreenInfo = dl_XineramaQueryScreens(dpy, &count);
// printf("%s(%d): %d 0x%x\n", FL, count, (int)ScreenInfo);
	if (!ScreenInfo || (count <= 0)) goto noXinerama;

	dpyMonitors.stat = dpyM_Active;
	if (count > MAX_MONITORS) count = MAX_MONITORS;
	if (dpyMonitors.count != count) dpyMonitors.stat |= dpyM_Change;
	dpyMonitors.count = count;

	si = ScreenInfo;
	for (i=0;i<count;i++) {
	    mon = &dpyMonitors.mon[i];
	    mon->num = i;
	    if ((mon->left != si->x_org) ||
		(mon->top  != si->y_org) ||
		(mon->right != (si->x_org + si->width)) ||
		(mon->bottom != (si->y_org + si->height))) {
		mon->left = si->x_org;
		mon->top  = si->y_org;
		mon->right = si->x_org + si->width;
		mon->bottom = si->y_org + si->height;
		dpyMonitors.stat |= dpyM_Change;
// printf("%s(%d): %d %d %d %d\n", FL, si->x_org, si->y_org, si->width, si->height);
	    }
	    si++;
	}
	XFree(ScreenInfo);

	if (dpyMonitors.stat & dpyM_Change) {
	    int width = 0;
	    int height = 0;
	    for (i=0;i<count;i++) {
		mon = &dpyMonitors.mon[i];
		if (width < mon->right)   width  = mon->right;
		if (height < mon->bottom) height = mon->bottom;
	    }
	    if (width > 0)  appXDat.Root_width  = width;
	    if (height > 0) appXDat.Root_height = height;
	    if (dpyMonitors.OnChange) dpyMonitors.OnChange();
	}
// printf("%uus\n", GetCurTimeDiv_us(t));

}
