/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "appXbase.h"
#include <stdarg.h>
#include "tpixmap.h"

#define DebugAzAtom 0

static int XmaxSel;
static int AbsMaxSel;
static Tlgs SelTxt_tmp;

void SelClr() {
	Tlgs_Clr(&appXDat.SelTxt);
	Tlgs_Clr(&appXDat.SelTxt_URL);
	ppTfree(&appXDat.Selpxm);
}

void SelRetClr() {
	Tlgs_Clr(&appXDat.SelRet_lgs);
}

/*_______________________ Selection write _______________________________*/
typedef union {
  char *s;
  const char *s_const;
  byte *b;
  Atom *a;
} TwrData;


static dword wr_incr_ms;

static int    wrCount;
static int    wrFormat;
static Atom   wrTarget, wrProperty;
static Window wrRequestor;
static TwrData wrData;


static void wrChangeProperty(int lg) {
  byte *dat = wrData.b;

	if (!dat || (wrCount <= 0) || (lg <= 0)) {
	    lg = 0;
	    dat = NULL;
	} else {
	    if (lg > wrCount) lg = wrCount;
	}
	wrCount -= lg;
	wrData.b += lg;
	lg = lg >> (wrFormat >> 4);

	XChangeProperty(appXDat.dpy, wrRequestor, wrProperty,
			wrTarget, wrFormat, PropModeReplace, dat, lg);
}


static void INCR_PropertyNotify(XPropertyEvent *ev) {
  int lg;

	if (ev->state != PropertyDelete) return;
	lg = wrCount;
	if (lg > XmaxSel) lg = XmaxSel;
	wrChangeProperty(lg);
	if (lg <= 0) {
	    appXDat.PropEvFilter[evFilter_sel_wr].Filter_PropEv = NULL;
	    XSelectInput(appXDat.dpy, wrRequestor, NoEventMask);
	}
}

static void chk_max(int *pmax, int lg, const char *nAtom, XSelectionRequestEvent *ev) {
  int max = *pmax;
  Atom a;
  const char *msg;

	if ((lg <= max) || (max <= 0)) return;
	a = XInternAtom(appXDat.dpy, nAtom, 1);
	if (!a || (a != ev->property)) return;
	
	if (lg > AbsMaxSel) {
	    *pmax = 0;
	    msg = "ERROR";
	} else {
	    *pmax = AbsMaxSel;
	    msg = "WARNING";
	}
	printf("%s(%d): %s: %s \"INCR\" not work\n", FL, msg, nAtom);
}

static Window REGP1 HandleSelectionClear(TXev_msg *msg) {  //29
#if DebugAzAtom
 printf("%s(%d): HandleSelectionClear\n", FL);
#endif
	appXDat.PropEvFilter[evFilter_sel_wr].Filter_PropEv = NULL;
	Tfree(appXDat.Selpxm);
	appXDat.Selpxm = NULL;
	SelRetClr(); //Tlgs_Clr(&appXDat.SelRet_lgs);
	appXDat.SELECT_AtomsCount = 0;
	return 0;
}

static Window REGP1 HandleSelectionRequest(TXev_msg *msg) {  //30
  int i, pbLg;
  bool ok, incr_on;
  XSelectionRequestEvent *ev = (XSelectionRequestEvent*)msg->ev;
  XSelectionEvent evt;
#if DebugAzAtom
 printf("%s(%d): HandleSelectionRequest\n", FL);
 AzAtom(FL, "Sel-Atoms selection", ev->selection);
 AzAtom(FL, "Sel-Atoms target", ev->target);
 AzAtom(FL, "Sel-Atoms property", ev->property);
 fflush(stdout);
#endif
// printf("%s(%d): SelWin=0x%x owner=0x%x requestor=0x%x\n", FL, (int)appXDat.SelHandleWin, (int)ev->owner, (int)ev->requestor);

	evt.type = SelectionNotify;
	evt.serial = 0;
	evt.send_event = 1;
	evt.display = appXDat.dpy;
	evt.requestor = ev->requestor;
	evt.selection = ev->selection;
	evt.target = ev->target;
	evt.property = ev->property;
	evt.time = ev->time;

	wrProperty = ev->property;
	wrRequestor = ev->requestor;
	wrTarget = ev->target;

	wrFormat = 8;
	SelTxt_tmp = appXDat.SelTxt;
	wrCount = SelTxt_tmp.lg;
	wrData.s = SelTxt_tmp.s;
	appXDat.PropEvFilter[evFilter_sel_wr].Filter_PropEv = NULL;

	incr_on = False;
	ok = True;

	if (ev->target == appXDat.XA_TARGETS) {
	    wrData.a = &appXDat.SELECT_Atoms[0];
	    wrCount = appXDat.SELECT_AtomsCount << 2; //32 and 64Bit 
	    wrFormat = 32;
	    wrTarget = XA_ATOM;

	} else if (ev->target == appXDat.XA_KDE_CUT) {
	    wrData.s_const = "0";
	    if (appXDat.SelCut) wrData.s_const = "1";
	    wrCount = 2;

#ifdef DRAG_AND_DROP
	} else if (appXDat.dnd && ev->selection == appXDat.dnd->XA_XdndSelection) {
// printf("%s(%d): \n", FL);
	    if (appXDat.dnd->SelectionReq)
		  ok = appXDat.dnd->SelectionReq(ev, &wrData.s, &wrCount, &wrFormat);
#endif
	} else {
// AzAtom(FL, "SelReq-Atoms", ev->target);
	  for (i=1; i < appXDat.SELECT_AtomsCount; i++) {
	    if (ev->target == appXDat.SELECT_Atoms[i]) {

	      if (ev->target == XA_PIXMAP) {
		if (appXDat.Selpxm && appXDat.Selpxm->pmxID) {
		  wrCount = 4; //32/64Bit ???
		  wrFormat = 32;
		  wrData.s = (char*)&appXDat.Selpxm->pmxID;
		} else ok = False;

	      } else {

		if (ev->target == appXDat.XA_URL_STRING) {
		    SelTxt_tmp = appXDat.SelTxt_URL;
		    wrCount = SelTxt_tmp.lg;
		    wrData.s = SelTxt_tmp.s;
		}

		int max = XmaxSel;
		pbLg = appXDat.SelTxt.lg;
//		chk_max(&max, pbLg, "GDK_SELECTION", ev);
		chk_max(&max, pbLg, "CLIP_TEMPORARY", ev); // Motif clients
		ok = max > 0;
		if (ok && (pbLg > max)) {
		    wr_incr_ms = GetTime_ms() + 1000;
		    wrCount = 4;  //32/64Bit ???
		    wrFormat = 32;
		    wrData.s = (char*)&pbLg;
		    wrTarget = appXDat.XA_INCR;
		    incr_on = True;
		}
	      }
	      break;
	    }
	  }
	}

	if (!ok) {
	    wrCount = 0;
	    wrData.s = NULL;
	    wrTarget = appXDat.XA_NONE;
	}

	wrChangeProperty(wrCount);

	XSendEvent(appXDat.dpy, ev->requestor, False, NoEventMask, (XEvent*)&evt);

	if (incr_on && ok) {
	    wrCount = SelTxt_tmp.lg;
	    wrData.s = SelTxt_tmp.s;
	    wrFormat = 8;
	    wrTarget = ev->target;

	    appXDat.PropEvFilter[evFilter_sel_wr].FilterAtom = ev->property;
	    appXDat.PropEvFilter[evFilter_sel_wr].Filter_PropEv = INCR_PropertyNotify;
	    XSelectInput(appXDat.dpy, ev->requestor, PropertyChangeMask);
	}
	return 0;
}

/*_______________________ Selection read ________________________________*/
static Atom Sel_INCR_Atom;
static int Sel_INCR_lg;
static char *Sel_INCR_data;
static bool rd_sel_ok;

static void rd_reset(int fu) {
	XSelectInput(appXDat.dpy, appXDat.SelHandleWin, NoEventMask);
	appXDat.PropEvFilter[evFilter_sel_rd].Filter_PropEv = NULL;
	if (fu & 0x01) appXDat.SelWait = False;
	if (fu & 0x02) rd_sel_ok = False;
	Sel_INCR_lg = 0;
	if (!rd_sel_ok) SelRetClr(); //Tlgs_Clr(&appXDat.SelRet_lgs);
}

static unsigned long SelRetFormatToByteS(int SelRetFormat) {
	if (SelRetFormat == 16) return 2;
	if (SelRetFormat == 32) return sizeof(long);
	return 1;
}

static void rd_INCR_PropertyNotify(XPropertyEvent *ev) {
  unsigned long Offset, rlg, lg, bytes_left;
  int SelRetFormat;
  Atom SelRetAtom;
  byte *rs;

	if (!appXDat.SelWait || (Sel_INCR_lg <= 0)) goto ev_close;
	if (ev->state != PropertyNewValue) return;

	Offset = 0;
loop:
	if (XGetWindowProperty(appXDat.dpy, appXDat.SelHandleWin, 
		appXDat.XA_ALD_selection, Offset, XmaxSel, 0, 
		AnyPropertyType, &SelRetAtom, &SelRetFormat,
		&rlg, &bytes_left, &rs)) return;

	if (SelRetAtom != Sel_INCR_Atom) return;
	if (rs) {
	    appXDat.SelRetAtom = SelRetAtom;
	    appXDat.SelRetFormat = SelRetFormat;
    // printf("XGetWindowProperty %d=\"%s\" %d off=%d rlg=%d %d\n", (int)SelRetAtom, XGetAtomName(appXDat.dpy, SelRetAtom), appXDat.SelRetFormat, (int)Offset, (int)rlg, (int)bytes_left);

//	    lg = rlg << (appXDat.SelRetFormat >> 4);
	    lg = rlg * SelRetFormatToByteS(SelRetFormat);
	    if (lg > Sel_INCR_lg) lg = Sel_INCR_lg;
	    Sel_INCR_lg -= lg;
	    ppcnmove(&Sel_INCR_data, (char*)rs, lg);
	    XFree(rs);
/*
 unsigned long XX = Offset;
	    XX += rlg / SelRetFormatToByteS(SelRetFormat);
 printf("%s(%d): %ld\n", FL, XX);
*/
	    Offset += (rlg * appXDat.SelRetFormat) >> 5;  //32/64Bit ???
 if (bytes_left) printf("%s(%d): %ld\n", FL, Offset);
/*
	    Offset += rlg / SelRetFormatToByteS(SelRetFormat);
 if (bytes_left) printf("%s(%d): %ld\n", FL, Offset);
*/
	    if (bytes_left) goto loop;
// printf("%s(%d) Sel INCR lg=%d\n", FL, Sel_INCR_lg);
	    XDeleteProperty(appXDat.dpy, appXDat.SelHandleWin, appXDat.XA_ALD_selection);
	    if (Sel_INCR_lg > 0) return;
	    rd_sel_ok = True;
	}
ev_close:
	rd_reset(1);
}

static Window REGP1 HandleSelectionNotify(TXev_msg *msg) {  //31
  unsigned long bytes_left, Offset, lg, rlg;
  unsigned char *rs;
  XSelectionEvent *ev = (XSelectionEvent*)msg->ev;
#if DebugAzAtom
 printf("HandleSelectionNotify %d req=%x s=%d t=%d p=%d\n", appXDat.SelWait, (int)ev->requestor, (int)ev->selection, (int)ev->target, (int)ev->property);
 AzAtom(FL, "HandleSelectionNotify selection", ev->selection);
 AzAtom(FL, "HandleSelectionNotify target", ev->target);
#endif

	if (ev->property == appXDat.XA_NONE) {
	    rd_reset(3);
	    return 0;
	}
#if DebugAzAtom
 AzAtom(FL, "HandleSelectionNotify property", ev->property);
#endif

	Sel_INCR_Atom = ev->target;
	Offset = 0;
	while (appXDat.SelWait) {
	    if (XGetWindowProperty(appXDat.dpy, ev->requestor, appXDat.XA_ALD_selection,
		    Offset, XmaxSel, 0, AnyPropertyType,
		    &appXDat.SelRetAtom, &appXDat.SelRetFormat,
		    &rlg, &bytes_left, &rs) != 0) break;
#if DebugAzAtom
 if (appXDat.SelRetAtom) {
   char *pcAtom = XGetAtomName(appXDat.dpy, appXDat.SelRetAtom);
   printf("XGetWindowProperty %d=\"%s\" %d off=%d rlg=%d %d\n", (int)appXDat.SelRetAtom, pcAtom, appXDat.SelRetFormat, (int)Offset, (int)rlg, (int)bytes_left);
   XFree(pcAtom);
 } else {
   printf("XGetWindowProperty %d=\"NULL\" %d off=%d rlg=%d %d\n", (int)appXDat.SelRetAtom, appXDat.SelRetFormat, (int)Offset, (int)rlg, (int)bytes_left);
 }
#endif
	    if (!rs) {
		rd_reset(3);
		return 0;
	    }

	    if (appXDat.SelRetAtom == appXDat.XA_INCR) {
		Sel_INCR_lg =  *(int*)rs;
		XFree(rs);
#if DebugAzAtom
 printf("%s(%d) Sel INCR lg=%d\n", FL, Sel_INCR_lg);
#endif
		if (Sel_INCR_lg <= 0) {
		    rd_reset(3);
		    return 0;
		}

		Sel_INCR_data = xmalloc(Sel_INCR_lg + 1);
		appXDat.SelRet_lgs.s = Sel_INCR_data;
		appXDat.SelRet_lgs.lg = Sel_INCR_lg;

		appXDat.PropEvFilter[evFilter_sel_rd].FilterAtom = appXDat.XA_ALD_selection;
		appXDat.PropEvFilter[evFilter_sel_rd].Filter_PropEv = rd_INCR_PropertyNotify;

		XSelectInput(appXDat.dpy, appXDat.SelHandleWin, PropertyChangeMask);
		XDeleteProperty(appXDat.dpy, appXDat.SelHandleWin, appXDat.XA_ALD_selection);
		return 0;

	    } else {
//		lg = (rlg << (appXDat.SelRetFormat >> 4));
		lg = rlg * SelRetFormatToByteS(appXDat.SelRetFormat);
		Tlgs_Add_s(&appXDat.SelRet_lgs, (char*)rs, lg);
		XFree(rs);

		Offset += (rlg * appXDat.SelRetFormat) >> 5; //32/64Bit ???
 if (bytes_left) printf("%s(%d): %ld\n", FL, Offset);

/*
		Offset += rlg / SelRetFormatToByteS(appXDat.SelRetFormat);
 if (bytes_left) printf("%s(%d): %ld\n", FL, Offset);
*/
		if (!bytes_left) break;
	    }
	}
	rd_sel_ok = appXDat.SelWait;
#if DebugAzAtom
 printf("%s(%d) %d lg=%d\n", FL, appXDat.SelWait, appXDat.SelRet_lgs.lg);
#endif
	return 0;
}


void SetSelection(int count, ...) {
  va_list ap;
	if (count <= 0) {
	    appXDat.SELECT_AtomsCount = 0;
	    return;
	}
	if (!get_XEventHandle(SelectionRequest)) SelectionInit();

	appXDat.SELECT_AtomsCount = count+1;
	appXDat.SELECT_Atoms[0] = appXDat.XA_TARGETS;

	va_start(ap, count);
	if (count > MaxSELECT_Atoms-1) count = MaxSELECT_Atoms-1;
	for (int i = 1; (i <= count); i++) {
	    appXDat.SELECT_Atoms[i] = va_arg(ap, long);
	}
	va_end(ap);
	XSetSelectionOwner(appXDat.dpy, appXDat.CLIPBOARD_ATOM, appXDat.SelHandleWin, CurrentTime);
}

static void Wait_GetPasteDaten(int ms) {
  dword t = GetTime_ms() + ms;

	AppHandleWinMessages();
	while (appXDat.SelWait && !rd_sel_ok && !cpTime(t, GetTime_ms())) {
	    fd_in_wait_sigle(appDat.x_fd, 10000);
	    AppHandleWinMessages();
	}
}


bool GetPasteDaten(Atom SrcAtom, Atom XA_SelAtom) {

	if (appXDat.SelWait) return False;
	if (!get_XEventHandle(SelectionNotify)) SelectionInit();
	rd_reset(3);
	appXDat.SelWait = True;
// AzAtom(FL, "GetPasteDaten", SrcAtom);
// AzAtom(FL, "GetPasteDaten", XA_SelAtom);

	XConvertSelection(appXDat.dpy, SrcAtom, XA_SelAtom, appXDat.XA_ALD_selection, appXDat.SelHandleWin, CurrentTime);

	Wait_GetPasteDaten(60);  //0,06sec
	if (appXDat.SelWait && !rd_sel_ok) {
	    appSetCursorWait(NULL, True);
	    Wait_GetPasteDaten(10000); //10,0sec
	    appSetCursorWait(NULL, False);
	}
	rd_reset(1);
	return (appXDat.SelRet_lgs.lg > 0);
}

#if DebugAzAtom
void printf_XA_TARGETS() {
  int i;
  Atom *pA;

	pA = (Atom*)&appXDat.SelRet_lgs.s[0];
	i = appXDat.SelRet_lgs.lg / sizeof(long);
	printf("printf_XA_TARGETS: %d\n", i);
	AzAtom(FL, "Sel-Atoms: RetType", appXDat.SelRetAtom);
	for (; i > 0; i--) {
	    AzAtom(FL, "Sel-Atoms", *pA);
	    pA++;
	}
}
#endif
bool Get_XA_TARGETS() {
  bool res = (GetPasteDaten(appXDat.CLIPBOARD_ATOM, appXDat.XA_TARGETS) != 0);
	res = res && (appXDat.SelRetFormat == 32) &&
          	     ((appXDat.SelRetAtom == appXDat.XA_TARGETS) || 
		      (appXDat.SelRetAtom == XA_ATOM));
#if DebugAzAtom
  if (res) printf_XA_TARGETS();
#endif
	return res;
}

Atom get_nSelAtom(Atom a) {
  Atom *pA = (Atom*)appXDat.SelRet_lgs.s;
  int i = appXDat.SelRet_lgs.lg / sizeof(long);
	for (; i > 0; i--) {
	    if (*pA == a) return a;
	    pA++;
	}
	return 0;
}

bool Chk_GetSelection(int count, ...) {
  va_list ap;
  Atom a = 0;

	if ((count <= 0) || !Get_XA_TARGETS()) return False;

	va_start(ap, count);
	while (!a && count) {
	    Atom an = va_arg(ap, long);
	    a = get_nSelAtom(an);
	    count--;
	}
	va_end(ap);
	rd_reset(3);
	return (a != 0);
}

/*____________________________ Init _____________________________________*/
void SelectionInit() {
  int a, m;

	a = m = XExtendedMaxRequestSize(appXDat.dpy);
	if (m > 0x40000) {
	    m = 0x40000;
	} else if (m) {
	    m -= 32;
	    a = m;
	} else {
	    a = m = XMaxRequestSize(appXDat.dpy) - 32;
	}
	AbsMaxSel = a;
	XmaxSel = m;

	set_XEventHandle(SelectionClear, HandleSelectionClear);
	set_XEventHandle(SelectionNotify, HandleSelectionNotify);
	set_XEventHandle(SelectionRequest, HandleSelectionRequest);
}
