/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "sysbase.h"
#include "jpeg_pv.h"

#if LD_LIB_JPEG

#define dl_jpeg_std_error jpeg_std_error

#else /* LD_LIB_JPEG */

#include <dlfcn.h>

void *jpegLibHandle = NULL;

bool jpeglib_RD_load = False;
bool jpeglib_WR_load = False;
bool jpeglib_Err = False;

struct jpeg_error_mgr *(*dl_jpeg_std_error)(struct jpeg_error_mgr * err);
void (*dl_jpeg_CreateDecompress)(j_decompress_ptr cinfo,
					int version, size_t structsize);
#endif /* LD_LIB_JPEG == 0 */

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};

static void __attribute__ ((noreturn)) jpeg_error_exit(j_common_ptr cinfo) {
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

void REGP2 jpeg_error_init(j_common_ptr cinfo, my_error_ptr jerr) {
	cinfo->err = dl_jpeg_std_error(&jerr->pub);
	jerr->pub.error_exit = jpeg_error_exit;
}

#if LD_LIB_JPEG == 0

static void CHK_JPEG_Ver(const char *name) {
  struct jpeg_decompress_struct *cinfo;
  struct my_error_mgr jerr;

	jpegLibHandle = dlopen(name, RTLD_LAZY);
	if (!jpegLibHandle) return;

	dl_jpeg_std_error = load_lib(jpegLibHandle, "jpeg_std_error");
	dl_jpeg_CreateDecompress = load_lib(jpegLibHandle, "jpeg_CreateDecompress");

	if (!dl_jpeg_std_error || !dl_jpeg_CreateDecompress) goto err;
	cinfo = xmalloc(sizeof(struct jpeg_decompress_struct));
	jpeg_error_init((j_common_ptr)cinfo, &jerr);
	if (setjmp(jerr.setjmp_buffer)) {
	    xfree(cinfo);
err:
	    dlclose(jpegLibHandle);
	    jpegLibHandle = NULL;
	    return;
	}
	dl_jpeg_CreateDecompress(cinfo, JPEG_LIB_VERSION, 
				 sizeof(struct jpeg_decompress_struct));
	xfree(cinfo);
}

typedef struct {
  const char *so;
  int ver;
} TTabJPEG;

static const char *TabJPEG[] = {
  "libjpeg.so.62",
  NULL
};

bool LoadLib_JPEG() {
  const char **tab;
  char *pc;

	if (jpeglib_Err) return False;
	if (jpegLibHandle) return True;

	tab = TabJPEG;
	while (!jpegLibHandle && *tab) {
	    CHK_JPEG_Ver(*tab);
	    tab++;
	}

	jpeglib_Err = (jpegLibHandle == NULL);
	if (jpeglib_Err) {
	    pc = pcdup("can't open ");
	    tab = TabJPEG;
	    int i = 3;
	    while (*tab) {
		ppcadd(&pc, *tab);
		tab++;
		i--;
		if (*tab) {
		    ppcadd(&pc, " or ");
		    if (!i) {
			i = 3;
			ppcadd(&pc, "\n               ");
		    }
		}
	    }
	    appError(EMSG_ERR2, pc);
	    xfree(pc);
	}

	return !jpeglib_Err;
}
#endif /* LD_LIB_JPEG == 0 */
