/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "sysbase.h"
#include "image.h"
#include <dlfcn.h>

void *(*dl_png_create_read_struct) (const char *user_png_ver,
		void *error_ptr, void *error_fn, void *warn_fn);
void *(*dl_png_destroy_read_struct) (void **png_ptr_ptr,
		void **info_ptr_ptr, void **end_info_ptr_ptr);
void *(*dl_png_create_info_struct) (void *png_ptr);

void (*dl_png_init_io) (void *png_ptr, FILE *fp);

void *pngLibHandle = NULL;
//int pngVer;
const char *pngVerStr;
bool pnglib_RD_load = False;
bool pnglib_WR_load = False;
bool pnglib_Err = False;


static void CHK_PNG_Ver(const char *name) {
  void *png_ptr;

	pngLibHandle = dlopen(name, RTLD_LAZY);
	if (!pngLibHandle) return;

	dl_png_create_read_struct = load_lib(pngLibHandle, "png_create_read_struct");
	dl_png_destroy_read_struct = load_lib(pngLibHandle, "png_destroy_read_struct");
	dl_png_create_info_struct = load_lib(pngLibHandle, "png_create_info_struct");
	dl_png_init_io = load_lib(pngLibHandle, "png_init_io");
	
	if (dl_png_create_read_struct && dl_png_destroy_read_struct &&
	    dl_png_create_info_struct && dl_png_init_io) {
	    png_ptr = dl_png_create_read_struct(pngVerStr, NULL, NULL, NULL);
	} else png_ptr = NULL;

	if (png_ptr) {
	    dl_png_destroy_read_struct(&png_ptr, NULL, NULL);
	} else {
	    dlclose(pngLibHandle);
	    pngLibHandle = NULL;
	}
}

typedef struct {
  const char *so;
  const char *ver;
} TTabPNG;

static const TTabPNG TabPNG[] = {
  {"libpng12.so.0",  "1.2.0"},
  {"libpng.so.3",    "1.2.0"},
  {"libpng14.so.14", "1.4.0"},
  {"libpng15.so.15", "1.5.0"},
  {"libpng.so.2",    "1.0.5"},
  {NULL, NULL}
};

bool LoadLib_PNG() {
  const TTabPNG *tab;
  char *pc;
//  char buf[32];

	if (pnglib_Err) return False;
	if (pngLibHandle) return True;
/*
	tab = TabPNG;
	while (tab->so) {
	    pc = buf;
	    ppcmove(&pc, "/usr/lib/");
	    ppcmove(&pc, tab->so);
	    if (file_chk_exist(buf, findREG|findLNK)) {
		pngVerStr = tab->ver;
		CHK_PNG_Ver(tab->so);
		break;
	    }
	    tab++;
	}
*/
	tab = TabPNG;
	while (!pngLibHandle && tab->so) {
	    pngVerStr = tab->ver;
	    CHK_PNG_Ver(tab->so);
	    tab++;
	}

	pnglib_Err = (pngLibHandle == NULL);
	if (pnglib_Err) {
	    pc = pcdup("can't open ");
	    tab = TabPNG;
	    int i = 3;
	    while (tab->so) {
		ppcadd(&pc, tab->so);
		tab++;
		i--;
		if (tab->so) {
		    ppcadd(&pc, " or ");
		    if (!i) {
			i = 3;
			ppcadd(&pc, "\n               ");
		    }
		}
	    }
	    appError(EMSG_ERR2, pc);
	    xfree(pc);
	}

	return !pnglib_Err;
}
