/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

#define USE_WR_PALETTE 1


#if LD_LIB_PNG

#include "png.h"

//extern int pngVer;
extern const char *pngVerStr;

#define dl_png_create_write_struct png_create_write_struct
#define dl_png_destroy_write_struct png_destroy_write_struct
#define dl_png_create_info_struct png_create_info_struct
#define dl_png_init_io png_init_io
#define dl_png_set_IHDR png_set_IHDR
//#define dl_png_malloc png_malloc
//#define dl_png_set_sBIT png_set_sBIT
#define dl_png_set_compression_level png_set_compression_level
#define dl_png_write_info png_write_info
#define dl_png_set_packing png_set_packing
#define dl_png_write_row png_write_row
//#define dl_png_write_rows png_write_rows
#define dl_png_write_end png_write_end
//#define dl_png_free png_free

#else /* LD_LIB_PNG */

typedef struct {
   byte red;   /* for use in red green blue files */
   byte green;
   byte blue;
   byte gray;  /* for use in grayscale files */
   byte alpha; /* for alpha channel files */
} png_color_8;

void *(*dl_png_create_write_struct) (const char *user_png_ver,
		void *error_ptr, void *error_fn, void *warn_fn);
void  (*dl_png_destroy_write_struct) (void **png_ptr_ptr, void **info_ptr_ptr);
extern void *(*dl_png_create_info_struct) (void *png_ptr);
extern void  (*dl_png_init_io) (void *png_ptr, FILE *fp);
dword (*dl_png_set_IHDR) (void *png_ptr, void *info_ptr,
		dword width, dword height,
		int bit_depth, int color_type, int interlace_method,
		int compression_method, int filter_method);
//void *(*dl_png_malloc) (void *png_ptr, dword size);
void  (*png_set_PLTE) (void *png_ptr, void *info_ptr, void *palette, int num_palette);
//void  (*dl_png_set_sBIT) (void *png_ptr, void *info_ptr, png_color_8 *sig_bit);
//void  (*png_set_shift) (void *png_ptr, png_color_8 *sig_bit);
//void  (*png_set_gAMA) (void *png_ptr, void *info_ptr, double file_gamma);
void  (*dl_png_set_compression_level) (void *png_ptr, int level);
void  (*dl_png_write_info) (void *png_ptr, void *info_ptr);
void  (*dl_png_set_packing) (void *png_ptr);
void  (*dl_png_write_row) (void *png_ptr, void *row);
//void  (*dl_png_write_rows) (void *png_ptr, void *row, int num_rows);
void  (*dl_png_write_end) (void *png_ptr, void *info_ptr);
//void  (*dl_png_free) (void *png_ptr, void *palette);

extern void *pngLibHandle;
//extern int pngVer;
extern char *pngVerStr;
extern bool pnglib_RD_load;
extern bool pnglib_WR_load;
extern bool pnglib_Err;

static void WriteLib_PNG_Proc() {
	dl_png_create_write_struct = load_lib(pngLibHandle, "png_create_write_struct");
	dl_png_destroy_write_struct = load_lib(pngLibHandle, "png_destroy_write_struct");
	dl_png_set_IHDR = load_lib(pngLibHandle, "png_set_IHDR");
//	dl_png_malloc = load_lib(pngLibHandle, "png_malloc");
	png_set_PLTE = load_lib(pngLibHandle, "png_set_PLTE");
//	dl_png_set_sBIT = load_lib(pngLibHandle, "png_set_sBIT");
//	png_set_shift = load_lib(pngLibHandle, "png_set_shift");
	dl_png_set_compression_level = load_lib(pngLibHandle, "png_set_compression_level");
	dl_png_write_info = load_lib(pngLibHandle, "png_write_info");
	dl_png_set_packing = load_lib(pngLibHandle, "png_set_packing");
	dl_png_write_row = load_lib(pngLibHandle, "png_write_row");
//	dl_png_write_rows = load_lib(pngLibHandle, "png_write_rows");
	dl_png_write_end = load_lib(pngLibHandle, "png_write_end");
//	dl_png_free = load_lib(pngLibHandle, "png_free");
}

static bool LoadLib_PNG_write() {
	if (pnglib_WR_load) return True;
	if (!pngLibHandle && !LoadLib_PNG()) return False;
	WriteLib_PNG_Proc();
	pnglib_WR_load = True;
	return True;
} 

//#define PNG_MAX_PALETTE_LENGTH    256

#endif /* LD_LIB_PNG == 0 */

#if USE_WR_PALETTE
static int getcolhash(int c) {
	return ((c >> 12) & 0xf00) + ((c >> 8) & 0xf0) + ((c >> 4) & 0x0f);
}

static int write_png_sort_color(int a, int b, int fu) {
  int res = getcolhash(a) - getcolhash(b);
	if (!res) res = (a & 0x00ffffff) - (b & 0x00ffffff);
	return res;
}
#endif

int write_png(TImage *self) {
#if LD_LIB_PNG
  png_struct *png_ptr;
  png_info *info_ptr = NULL;
  png_bytep row_pointer;
#else /* LD_LIB_PNG */
  void *png_ptr;
  void *info_ptr = NULL;
  void *row_pointer;
#endif /* LD_LIB_PNG */
  int i;
#if USE_WR_PALETTE
  byte *palette = NULL;
  int ColCount;
#endif
  bool gray;
  bool ALPHA;

	gray = (self->out_fu & fuWR_gray) != 0;

#if LD_LIB_PNG == 0
	if (!LoadLib_PNG_write()) return IMG_ERR_NOLIB;
#endif

	png_ptr = dl_png_create_write_struct(pngVerStr, NULL, NULL, NULL);
	if (!png_ptr) return IMGWR_ERROR;

	info_ptr = dl_png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		dl_png_destroy_write_struct(&png_ptr, &info_ptr);
		return IMGWR_ERR_HEAD;
	}

	dl_png_init_io(png_ptr, self->out_fp);
  int compression_level = ((self->out_fu >> 8) & 0x0f) - 1;
	if (compression_level >= 0) {
	    if (compression_level > 9) compression_level = 9;
	    dl_png_set_compression_level(png_ptr, compression_level);
	}

	ALPHA = !(self->out_fu & fuWR_NoMask) && alpha_akt(self);

#if USE_WR_PALETTE
	ColCount = 0;
	if (!ALPHA && (self->out_fu & fuWR_use_pallette)) {
	    ColCount = TImage_GetColorCount(self, pxgcc_RGB_ALL);
	    if (ColCount > 256) ColCount = 0;
	}
#endif
	i = 2; //PNG_COLOR_MASK_COLOR
	if (gray) i = 0; //PNG_COLOR_TYPE_GRAY
	if (ALPHA) {
	    i |= 4; //PNG_COLOR_MASK_ALPHA
	} 
#if USE_WR_PALETTE
	if (ColCount) i |= 1;  //PNG_COLOR_MASK_PALETTE
#endif

	dl_png_set_IHDR(png_ptr, info_ptr, self->width, self->height, 8, i, 0, 0, 0);

#if USE_WR_PALETTE
	if (ColCount) {
	    TImage_GetColorMap(self);
	    i = 0;
	    if (self->mask_data) i++;
	    QuickSort_PL((void*)self->cmap->col_map, i, ColCount-1, 0, 
			(TCompareFunction)write_png_sort_color);
	    palette = xcalloc(ColCount * 3);
	    byte *db = palette;
	    ic32 *col_map = self->cmap->col_map;
	    for (i=0;i<ColCount;i++) {
		*db++ = col_map[i].b[2];
		*db++ = col_map[i].b[1];
		*db++ = col_map[i].b[0];
	    }
	    png_set_PLTE(png_ptr, info_ptr, (void*)palette, ColCount);
	}
#endif

//	png_set_sBIT(png_ptr, info_ptr, &sig_bit);
//	png_set_shift(png_ptr, &sig_bit);

	dl_png_write_info(png_ptr, info_ptr);
	dl_png_set_packing(png_ptr);

	row_pointer = xmalloc(self->width * 4);

  bool rgb565 = (self->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp;

#if USE_WR_PALETTE
	if (ColCount) {
  int *hash444 = xcalloc(sizeof(int) * 4096);
  int *pint = (int*)self->cmap->col_map;
  int col;
  int mk = -1;

	    for (int y=0;y<ColCount;y++) {
		col = getcolhash(*pint);
		if (col != mk) {
		    hash444[col] = y;
		    mk = col;
		}
		pint++;
	    }

  byte *sb0 = self->data;
	    for (int y=self->height;y;y--) {
  byte *sb, *db;
  dword c, *cal;
		sb = sb0;
		db = row_pointer;
		for (int x=self->width;x;x--) {
		    if (rgb565) {
			c = *((word*)sb);
			col = ((c & 0xf800)<<8)|((c & 0x7e0)<<5)|((c & 0x1f)<<3);
			sb += 2;
		    } else {
			col = *((dword*)sb) & 0x00ffffff;
			sb += 4;
		    }
		    i = hash444[getcolhash(col)];
		    cal = (dword*)self->cmap->col_map + i;
		    while ((i<ColCount) && (*cal != col)) {cal++; i++;}
		    *db = i;
		    db++;
		}
		dl_png_write_row(png_ptr, row_pointer);
		sb0 += self->bytes_per_line;
	    }
	    xfree(hash444);
	    goto ende;
	}
#endif
  dword A = 255;
  byte *sb0 = self->data;
  byte *sbA = self->rgb565_alpha_data;
	for (int y=self->height;y;y--) {
  byte *sb, *db;
	    sb = sb0;
	    db = row_pointer;
	    for (int x=self->width;x;x--) {
  dword R, G, B;
		if (rgb565) {
		    dword c = *((word*)sb);
		    sb += 2;
		    B = (c<<3) & 0xf8; G = (c>>3) & 0xfc; R = (c>>8) & 0xf8;
		    if (sbA) A = *sbA++;
		} else {
		    B = sb[0]; G = sb[1]; R = sb[2]; A = sb[3];
		    sb += 4;
		}

		if (gray) {
		    G = R + G + B;
		    *db++ = ((G * 0x55a) >> 12);  //div 3
		} else {
		    *db++ = R; *db++ = G; *db++ = B;
		}
		if (ALPHA) *db++ = A;
	    }
	    dl_png_write_row(png_ptr, row_pointer);
	    sb0 += self->bytes_per_line;
	}
#if USE_WR_PALETTE
ende:
#endif
	dl_png_write_end(png_ptr, info_ptr);
	dl_png_destroy_write_struct(&png_ptr, &info_ptr);
#if USE_WR_PALETTE
	xfree(palette);
#endif
	xfree(row_pointer);
	return IMGWR_OK;
}

static bool ChkType(TglobPath *gp, const char *fn, FILE *F) {
	return TImage_ChkPattern(gp, fn, hImageWR_png.pattern);
}

static bool GetLibErr() {
#if LD_LIB_PNG
	return False;
#else
	return pnglib_Err;
#endif
}

const ThImageLoader hImageWR_png = {
  "png",
  filetyp_PNG,
  ChkType,
  write_png,
  GetLibErr
};
