/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

typedef struct {
  int ox;
  int dox;
  int lgx;
  dword cxa;
  dword cxe;
} TpxTab;


typedef struct {
  Timg_rsz *rsz;
  TpxTab *pxTab;
  dword ca;
  dword ce;
  dword cxn;
  dword cyn;
  dword cxyn;
  int lc;

  int loff_x;

  union {
    void *srcp;
    byte *src8;
    word *src16;
    dword *src32;
  };
  union {
    void *destp;
    byte *dest8;
    word *dest16;
    dword *dest32;
  };

  void *src_dat;
  void *dest_dat;
  int yoff_src;
  int yoff_dest;

  int kw;
  int kh;
  int kmul_h;
  int kmul_w;
  
  int loc_0_loop;
  int loc_1_loop;
  int loc_N_loop;
  TpxTab *pxTab0;
  int loc_K_Off;
  int xoff;
  int yoff;
  int rgb_byte;
  word src_off_x;
  word src_off_y;
} Tz;

static void math_cxa_cxe(Tz *z, int x) {
  int kw0, a, b, o, w;

	w = z->rsz->src_w;
	kw0 = z->kw * x;
	a = (z->src_off_x << 16) + kw0 + z->loc_K_Off;

	o = 0x10000 - (a & 0xffff);
	o += z->loc_K_Off;
	o *= z->kmul_w;
	z->ca = o >> 12;

	b = a + z->kw - z->loc_K_Off - z->loc_K_Off;
	o = b & 0xffff;
	o += z->loc_K_Off;
	o *= z->kmul_w;
	z->ce = o >> 12;

	a >>= 16;
	if (a >= w) goto err;
	b >>= 16;
	b -= a;
	if (a+b >= w) goto err;
	z->xoff = a;
	z->lc = b;
	return;
err:
	z->xoff = z->rsz->src_w - 1;
	z->lc = 0;
}

static void math_cya_cye(Tz *z, int y) {
  int kh0, a, b, o, h;

	h = z->rsz->src_h;
	kh0 = z->kh * y;
	a = (z->src_off_y << 16) + kh0 + z->loc_K_Off;

	o = 0x10000 - (a & 0xffff);
	o += z->loc_K_Off;
	o *= z->kmul_h;
	z->ca = o >> 12;

	b = a + z->kh - z->loc_K_Off - z->loc_K_Off;
	o = b & 0xffff;
	o += z->loc_K_Off;
	o *= z->kmul_h;
	z->ce = o >> 12;

	a >>= 16;
	if (a >= h) goto err;
	b >>= 16;
	b -= a;
	if (a+b >= h) goto err;
	z->yoff = a;
	z->lc = b;

	z->yoff_src = (z->rsz->src_line_eoff + z->rsz->src_w) * a;
	z->src8 = (byte*)z->src_dat + (z->yoff_src * z->rgb_byte);
	z->yoff_dest = (z->rsz->dest_line_eoff + z->rsz->to_w) * y;
	z->dest8 = (byte*)z->dest_dat + (z->yoff_dest * z->rgb_byte);
	return;
err:
	z->yoff = z->rsz->src_h - 1;
	z->lc = 0;
	z->yoff_src = (z->rsz->src_line_eoff + z->rsz->src_w) * z->yoff;
	z->src8 = (byte*)z->src_dat + (z->yoff_src * z->rgb_byte);
	z->yoff_dest = (z->rsz->dest_line_eoff + z->rsz->to_w) * y;
	z->dest8 = (byte*)z->dest_dat + (z->yoff_dest * z->rgb_byte);
}

/* ___________________ 16bpp rsz _______________________ */
static void rszcpu16_X1Y1(Tz *z, int lx) {
  word *src, *dest;
  TpxTab *pxTab;

	if (lx <= 0) return;
	src = z->src16;
	dest = z->dest16;

	pxTab = z->pxTab;
	while (lx > 0) {
	    dest[pxTab->dox] = src[pxTab->ox];
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void rsz16(Tz *z) {
  int y, ye, x;

	z->loc_K_Off = 0x0fff;
	if (z->loc_K_Off > (z->kh >> 3)) z->loc_K_Off = z->kh >> 3;

//	if (z->rsz->fu & rsz_fu_Fast_rsz) {

//	} else {
	    x = z->loc_0_loop + z->loc_1_loop + z->loc_N_loop;
	    y = z->rsz->dest_off_y;
	    ye = z->rsz->dest_lg_y;
	    while (y < ye) {
		math_cya_cye(z, y);
		z->pxTab = z->pxTab0;
		rszcpu16_X1Y1(z, x);
		y++;
	    }
//	}
}

/* ___________________ 24bpp rsz _______________________ */

typedef struct {
  dword b;
  dword g;
  dword r;
  dword a;

  dword ca, cn, ce;
  union {
    void *srcp;
    byte *src8;
    word *src16;
    dword *src32;
  };
  int lgx;
} Tzx;

#if defined (__arm__)

// #define ATR_INLINE __attribute__ ((noinline))
 #define ATR_INLINE inline
 #include "rsz_arm_32.c"

#else /* alg CPU i386 */

 #define ATR_INLINE __attribute__ ((noinline))
// #define ATR_INLINE inline
 #include "rsz_i386_32.c"

#endif /* alg CPU i386 */

static void __attribute__ ((noinline)) rszcpu24_X1Y1(Tz *z, int lx) {
  dword *src, *dest;
  TpxTab *pxTab;

	if (lx <= 0) return;
	src = z->src32;
	dest = z->dest32;

	pxTab = z->pxTab;
	while (lx > 0) {
	    dest[pxTab->dox] = src[pxTab->ox];
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu24_XNY1(Tz *z, int lx) {
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	zx.cn = z->cxn;
	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = z->src32 + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    zx.lgx = pxTab->lgx;
	    rszcpu24_XN0(&zx);

	    RSZ_zxToResult_24
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu32_XNY1(Tz *z, int lx) {
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	zx.cn = z->cxn;
	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = z->src32 + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    zx.lgx = pxTab->lgx;
	    rszcpu32_XN0(&zx);

	    RSZ_zxToResult_32
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu24_X2Y2(Tz *z, int lx) {
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = z->src32 + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    rszcpu24_X20(&zx);

	    zx.src32 += z->loff_x;
	    zx.ca = (pxTab->cxa * z->ce) >> 16;
	    zx.ce = (pxTab->cxe * z->ce) >> 16;
	    rszcpu24_X2(&zx);

	    RSZ_zxToResult_24
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu32_X2Y2(Tz *z, int lx) {
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = z->src32 + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    rszcpu32_X20(&zx);

	    zx.src32 += z->loff_x;
	    zx.ca = (pxTab->cxa * z->ce) >> 16;
	    zx.ce = (pxTab->cxe * z->ce) >> 16;
	    rszcpu32_X2(&zx);

	    RSZ_zxToResult_32
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu24_X2YN(Tz *z, int lx) {
  dword *src, *dest;
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	src = z->src32;
	dest = z->dest32;

	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = src + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    rszcpu24_X20(&zx);

	    int lc = z->lc - 1;
	    if (lc > 0) {
		zx.ca = (pxTab->cxa * z->cyn) >> 16;
		zx.ce = (pxTab->cxe * z->cyn) >> 16;
ny_loop:
		zx.src32 += z->loff_x;
		rszcpu24_X2(&zx);
		lc--;
		if (lc > 0) goto ny_loop;
	    }

	    zx.src32 += z->loff_x;
	    zx.ca = (pxTab->cxa * z->ce) >> 16;
	    zx.ce = (pxTab->cxe * z->ce) >> 16;
	    rszcpu24_X2(&zx);

	    dest[pxTab->dox] = RSZ_zxToResult_24
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu32_X2YN(Tz *z, int lx) {
  dword *src, *dest;
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	src = z->src32;
	dest = z->dest32;

	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = src + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    rszcpu32_X20(&zx);

	    int lc = z->lc - 1;
	    if (lc > 0) {
		zx.ca = (pxTab->cxa * z->cyn) >> 16;
		zx.ce = (pxTab->cxe * z->cyn) >> 16;
ny_loop:
		zx.src32 += z->loff_x;
		rszcpu32_X2(&zx);
		lc--;
		if (lc > 0) goto ny_loop;
	    }

	    zx.src32 += z->loff_x;
	    zx.ca = (pxTab->cxa * z->ce) >> 16;
	    zx.ce = (pxTab->cxe * z->ce) >> 16;
	    rszcpu32_X2(&zx);

	    dest[pxTab->dox] = RSZ_zxToResult_32
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rszcpu24_XNYN(Tz *z, int lx) {
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = z->src32 + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.cn = (z->cxn * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    zx.lgx = pxTab->lgx;
	    rszcpu24_XN0(&zx);

	    int lc = z->lc - 1;
	    if (lc > 0) {
		zx.ca = (pxTab->cxa * z->cyn) >> 16;
		zx.cn = z->cxyn;
		zx.ce = (pxTab->cxe * z->cyn) >> 16;
ny_loop:
		zx.src32 += z->loff_x;
		rszcpu24_XN(&zx);
		lc--;
		if (lc > 0) goto ny_loop;
	    }

	    zx.src32 += z->loff_x;
	    zx.ca = (pxTab->cxa * z->ce) >> 16;
	    zx.cn = (z->cxn * z->ce) >> 16;
	    zx.ce = (pxTab->cxe * z->ce) >> 16;
	    rszcpu24_XN(&zx);

	    RSZ_zxToResult_24
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}


static void __attribute__ ((noinline)) rszcpu32_XNYN(Tz *z, int lx) {
  TpxTab *pxTab;
  Tzx zx;

	if (lx <= 0) return;
	pxTab = z->pxTab;
	while (lx > 0) {
	    zx.src32 = z->src32 + pxTab->ox;
	    zx.ca = (pxTab->cxa * z->ca) >> 16;
	    zx.cn = (z->cxn * z->ca) >> 16;
	    zx.ce = (pxTab->cxe * z->ca) >> 16;
	    zx.lgx = pxTab->lgx;
	    rszcpu32_XN0(&zx);

	    int lc = z->lc - 1;
	    if (lc > 0) {
		zx.ca = (pxTab->cxa * z->cyn) >> 16;
		zx.cn = z->cxyn;
		zx.ce = (pxTab->cxe * z->cyn) >> 16;
ny_loop:
		zx.src32 += z->loff_x;
		rszcpu32_XN(&zx);
		lc--;
		if (lc > 0) goto ny_loop;
	    }

	    zx.src32 += z->loff_x;
	    zx.ca = (pxTab->cxa * z->ce) >> 16;
	    zx.cn = (z->cxn * z->ce) >> 16;
	    zx.ce = (pxTab->cxe * z->ce) >> 16;
	    rszcpu32_XN(&zx);

	    RSZ_zxToResult_32
	    pxTab++;
	    lx--;
	}
	z->pxTab = pxTab;
}

static void __attribute__ ((noinline)) rsz24(Tz *z) {
  int y, ye, x;

	z->loc_K_Off = 0x0fff;
	if (z->loc_K_Off > (z->kh >> 3)) z->loc_K_Off = z->kh >> 3;

	y = z->rsz->dest_off_y;
	ye = z->rsz->dest_lg_y;
	if (z->rsz->fu & rsz_fu_Fast_rsz) {
	    x = z->loc_0_loop + z->loc_1_loop + z->loc_N_loop;
	    while (y < ye) {
		math_cya_cye(z, y);
		z->pxTab = z->pxTab0;
		rszcpu24_X1Y1(z, x);
		y++;
	    }

	} else if (z->rsz->src_typ == pxm_4b_24bpp8a) {
// printf("%s(%d): pxm_4b_24bpp8a\n", FL);

	    while (y < ye) {
		z->pxTab = z->pxTab0;
		math_cya_cye(z, y);

		if (!z->lc) {
		    rszcpu24_X1Y1(z, z->loc_0_loop);
		    rszcpu32_X2Y1(z, z->loc_1_loop);
		    rszcpu32_XNY1(z, z->loc_N_loop);

		} else if (z->lc == 1) {
		    rszcpu32_X1Y2(z, z->loc_0_loop);
		    rszcpu32_X2Y2(z, z->loc_1_loop);
		    rszcpu32_XNYN(z, z->loc_N_loop);

		} else {
		    rszcpu32_X1Y2(z, z->loc_0_loop);   //TODO X1XN
		    rszcpu32_X2YN(z, z->loc_1_loop);
		    rszcpu32_XNYN(z, z->loc_N_loop);
		}
		y++;
	    }
	} else {
// printf("%s(%d): pxm_4b_24bpp\n", FL);

	    while (y < ye) {
		z->pxTab = z->pxTab0;
		math_cya_cye(z, y);

		if (!z->lc) {
		    rszcpu24_X1Y1(z, z->loc_0_loop);
		    if (z->loc_1_loop > 0) rszcpu24_X2Y1(z, z->loc_1_loop);
		    rszcpu24_XNY1(z, z->loc_N_loop);

		} else if (z->lc == 1) {
		    rszcpu24_X1Y2(z, z->loc_0_loop);
		    rszcpu24_X2Y2(z, z->loc_1_loop);
		    rszcpu24_XNYN(z, z->loc_N_loop);

		} else {
		    rszcpu24_X1Y2(z, z->loc_0_loop);   //TODO X1XN
		    rszcpu24_X2YN(z, z->loc_1_loop);
		    rszcpu24_XNYN(z, z->loc_N_loop);
		}
		y++;
	    }
	}
}


int ImgDat_Resize_XX_11(Timg_rsz *rsz) {
  Tz z;

	if ((rsz->dest_lg_x < 2) || (rsz->dest_lg_y < 2)) return rszERR_LE2x2Pixel;

	z.kw = (rsz->src_lg_x << 16) / rsz->dest_lg_x;
	if (z.kw > 0xffffff) return rszERR_ZoomGT255;

	z.kh = (rsz->src_lg_y << 16) / rsz->dest_lg_y;
	if (z.kh > 0xffffff) return rszERR_ZoomGT255;

	if ((z.kw == 0x00010000) && (z.kh == 0x00010000)) return ImgDat_Copy(rsz);

	if (rsz->src_typ == pxm_2b_16bpp) {
	    z.rgb_byte = 2;
	    z.kmul_h = 0x1007ffff / z.kh;
	    z.kmul_w = 0x1007ffff / z.kw;
	    z.loff_x = rsz->src_line_eoff + rsz->src_w;

	} else if ((rsz->src_typ == pxm_4b_24bpp) || 
		   (rsz->src_typ == pxm_4b_24bpp8a)) {
	    z.rgb_byte = 4;
	    z.kmul_h = 0x10000000 / z.kh;
	    z.kmul_w = 0x10000000 / z.kw;
	    z.loff_x = rsz->src_line_eoff + rsz->src_w;

	} else return rszERR_Zoom;

	z.src_dat = rsz->src_dat;
	z.dest_dat = rsz->dest_dat;

	z.src_off_x = rsz->src_off_x;
	if (z.src_off_x > rsz->src_w) z.src_off_x = rsz->src_w;
	z.src_off_y = rsz->src_off_y;
	if (z.src_off_y >= rsz->src_h) z.src_off_y = rsz->src_h;

	if (!rsz->rsztemp0) {
	    rsz->rsztemp0 = xmalloc(sizeof(TpxTab) * rsz->to_w);
	}

	z.loc_0_loop = 0;
	z.loc_1_loop = 0;
	z.loc_N_loop = 0;
	z.rsz = rsz;
	z.cxn = z.kmul_w << 4;
	z.cyn = z.kmul_h << 4;
	z.cxyn = (z.cxn * z.cyn) >> 16;
	z.pxTab0 = rsz->rsztemp0;
	z.pxTab = rsz->rsztemp0;
	z.loc_K_Off = 0x0fff;
	if (z.loc_K_Off > (z.kw >> 3)) z.loc_K_Off = z.kw >> 3;

	for (int i=rsz->dest_off_x;i<rsz->dest_lg_x;i++) {
	    math_cxa_cxe(&z, i);
	    if (!z.lc) {
		z.pxTab->ox = z.xoff;
		z.pxTab->dox = i;
		z.pxTab->lgx = z.lc - 1;
		z.pxTab->cxa = z.ca;
		z.pxTab->cxe = z.ce;
		z.loc_0_loop++;
		z.pxTab++;
	    }
	}

	for (int i=rsz->dest_off_x;i<rsz->dest_lg_x;i++) {
	    math_cxa_cxe(&z, i);
	    if (z.lc == 1) {
		z.pxTab->ox = z.xoff;
		z.pxTab->dox = i;
		z.pxTab->lgx = z.lc - 1;
		z.pxTab->cxa = z.ca;
		z.pxTab->cxe = z.ce;
		z.loc_1_loop++;
		z.pxTab++;
	    }
	}

	for (int i=rsz->dest_off_x;i<rsz->dest_lg_x;i++) {
	    math_cxa_cxe(&z, i);
	    if (z.lc > 1) {
		z.pxTab->ox = z.xoff;
		z.pxTab->dox = i;
		z.pxTab->lgx = z.lc - 1;
		z.pxTab->cxa = z.ca;
		z.pxTab->cxe = z.ce;
		z.loc_N_loop++;
		z.pxTab++;
	    }
	}
// printf("%s(%d): w=%d 0=%d 1=%d N=%d\n", FL, rsz->dest_lg_x, z.loc_0_loop, z.loc_1_loop, z.loc_N_loop);

	if (rsz->src_typ == pxm_2b_16bpp) {
	    rsz16(&z);

	} else if ((rsz->src_typ == pxm_4b_24bpp) || 
		   (rsz->src_typ == pxm_4b_24bpp8a)) {
	    rsz24(&z);
	};

	if (!(rsz->fu & rsz_fu_TmpNotFree)) ppfree(&rsz->rsztemp0);
	return 0;
}
