/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "sysbase.h"
#include "image.h"


/*
used for Win32 or for speed    (m1870MHz)
	XAllocNamedColor(..,  "red", ..) = ~225us
	XAllocNamedColor(..,  "LightGreen", ..) = ~230us
	XAllocNamedColor(..,  "#234567", ..) = ~125us
	GetColorByName("red") = ~2us
	GetColorByName("LightGreen") = ~6us
	GetColorByName("#234567") = ~1us
*/


static const byte Tab3ColorNameDat[] = {
  "tan" "\xd2\xb4\x8c"
  "red" "\xff\x00\x00"
  "\x00"
};

static const byte Tab4ColorNameDat[] = {
  "snow" "\xff\xfa\xfa"
  "navy" "\x00\x00\x80"
  "blue" "\x00\x00\xff"
  "cyan" "\x00\xff\xff"
  "gold" "\xff\xd7\x00"
  "peru" "\xcd\x85\x3f"
  "pink" "\xff\xc0\xcb"
  "plum" "\xdd\xa0\xdd"
  "\x00"
};

static const byte Tab4Color4Dat[] = {
  "tan" "\xff\xa5\x4f" "\xee\x9a\x49" "\xcd\x85\x3f" "\x8b\x5a\x24"
  "red" "\xff\x00\x00" "\xee\x00\x00" "\xcd\x00\x00" "\x8b\x00\x00"
  "\x00"
};

static const byte Tab5ColorNameDat[] = {
  "linen" "\xfa\xf0\xe6"
  "ivory" "\xff\xff\xf0"
  "azure" "\xf0\xff\xff"
  "white" "\xff\xff\xff"
  "black" "\x00\x00\x00"
  "green" "\x00\xff\x00"
  "khaki" "\xf0\xe6\x8c"
  "beige" "\xf5\xf5\xdc"
  "wheat" "\xf5\xde\xb3"
  "brown" "\xa5\x2a\x2a"
  "coral" "\xff\x7f\x50"
  "\x00"
};

static const byte Tab5Color4Dat[] = {
  "snow" "\xff\xfa\xfa" "\xee\xe9\xe9" "\xcd\xc9\xc9" "\x8b\x89\x89"
  "blue" "\x00\x00\xff" "\x00\x00\xee" "\x00\x00\xcd" "\x00\x00\x8b"
  "cyan" "\x00\xff\xff" "\x00\xee\xee" "\x00\xcd\xcd" "\x00\x8b\x8b"
  "gold" "\xff\xd7\x00" "\xee\xc9\x00" "\xcd\xad\x00" "\x8b\x75\x00"
  "pink" "\xff\xb5\xc5" "\xee\xa9\xb8" "\xcd\x91\x9e" "\x8b\x63\x6c"
  "plum" "\xff\xbb\xff" "\xee\xae\xee" "\xcd\x96\xcd" "\x8b\x66\x8b"
  "\x00"
}; 

static const byte Tab6ColorNameDat[] = {
  "bisque" "\xff\xe4\xc4"
  "yellow" "\xff\xff\x00"
  "sienna" "\xa0\x52\x2d"
  "salmon" "\xfa\x80\x72"
  "orange" "\xff\xa5\x00"
  "tomato" "\xff\x63\x47"
  "maroon" "\xb0\x30\x60"
  "violet" "\xee\x82\xee"
  "orchid" "\xda\x70\xd6"
  "purple" "\xa0\x20\xf0"
  "\x00"
}; 

static const byte Tab6Color4Dat[] = {
  "ivory" "\xff\xff\xf0" "\xee\xee\xe0" "\xcd\xcd\xc1" "\x8b\x8b\x83"
  "azure" "\xf0\xff\xff" "\xe0\xee\xee" "\xc1\xcd\xcd" "\x83\x8b\x8b"
  "green" "\x00\xff\x00" "\x00\xee\x00" "\x00\xcd\x00" "\x00\x8b\x00"
  "khaki" "\xff\xf6\x8f" "\xee\xe6\x85" "\xcd\xc6\x73" "\x8b\x86\x5e"
  "wheat" "\xff\xe7\xba" "\xee\xd8\xae" "\xcd\xba\x96" "\x8b\x7e\x66"
  "brown" "\xff\x40\x40" "\xee\x3b\x3b" "\xcd\x33\x33" "\x8b\x23\x23"
  "coral" "\xff\x72\x56" "\xee\x6a\x50" "\xcd\x5b\x45" "\x8b\x3e\x2f"
  "\x00"
};

static const byte Tab7ColorNameDat[] = {
  "OldLace" "\xfd\xf5\xe6"
  "DimGray" "\x69\x69\x69"
  "SkyBlue" "\x87\xce\xeb"
  "HotPink" "\xff\x69\xb4"
  "magenta" "\xff\x00\xff"
  "thistle" "\xd8\xbf\xd8"
  "DarkRed" "\x8b\x00\x00"
  "\x00"
};

static const byte Tab7Color4Dat[] = {
  "bisque" "\xff\xe4\xc4" "\xee\xd5\xb7" "\xcd\xb7\x9e" "\x8b\x7d\x6b"
  "yellow" "\xff\xff\x00" "\xee\xee\x00" "\xcd\xcd\x00" "\x8b\x8b\x00"
  "sienna" "\xff\x82\x47" "\xee\x79\x42" "\xcd\x68\x39" "\x8b\x47\x26"
  "salmon" "\xff\x8c\x69" "\xee\x82\x62" "\xcd\x70\x54" "\x8b\x4c\x39"
  "orange" "\xff\xa5\x00" "\xee\x9a\x00" "\xcd\x85\x00" "\x8b\x5a\x00"
  "tomato" "\xff\x63\x47" "\xee\x5c\x42" "\xcd\x4f\x39" "\x8b\x36\x26"
  "maroon" "\xff\x34\xb3" "\xee\x30\xa7" "\xcd\x29\x90" "\x8b\x1c\x62"
  "orchid" "\xff\x83\xfa" "\xee\x7a\xe9" "\xcd\x69\xc9" "\x8b\x47\x89"
  "purple" "\x9b\x30\xff" "\x91\x2c\xee" "\x7d\x26\xcd" "\x55\x1a\x8b"
  "\x00"
}; 

static const byte Tab8ColorNameDat[] = {
  "moccasin" "\xff\xe4\xb5"
  "cornsilk" "\xff\xf8\xdc"
  "seashell" "\xff\xf5\xee"
  "honeydew" "\xf0\xff\xf0"
  "lavender" "\xe6\xe6\xfa"
  "NavyBlue" "\x00\x00\x80"
  "SeaGreen" "\x2e\x8b\x57"
  "DeepPink" "\xff\x14\x93"
  "DarkGrey" "\xa9\xa9\xa9"
  "DarkBlue" "\x00\x00\x8b"
  "DarkCyan" "\x00\x8b\x8b"
  "\x00"
}; 

static const byte Tab8Color4Dat[] = {
  "SkyBlue" "\x87\xce\xff" "\x7e\xc0\xee" "\x6c\xa6\xcd" "\x4a\x70\x8b"
  "HotPink" "\xff\x6e\xb4" "\xee\x6a\xa7" "\xcd\x60\x90" "\x8b\x3a\x62"
  "magenta" "\xff\x00\xff" "\xee\x00\xee" "\xcd\x00\xcd" "\x8b\x00\x8b"
  "thistle" "\xff\xe1\xff" "\xee\xd2\xee" "\xcd\xb5\xcd" "\x8b\x7b\x8b"
  "\x00"
}; 

static const byte Tab9ColorNameDat[] = {
  "gainsboro" "\xdc\xdc\xdc"
  "PeachPuff" "\xff\xda\xb9"
  "MintCream" "\xf5\xff\xfa"
  "AliceBlue" "\xf0\xf8\xff"
  "MistyRose" "\xff\xe4\xe1"
  "SlateGray" "\x70\x80\x90"
  "LightGrey" "\xd3\xd3\xd3"
  "SlateBlue" "\x6a\x5a\xcd"
  "RoyalBlue" "\x41\x69\xe1"
  "SteelBlue" "\x46\x82\xb4"
  "LightBlue" "\xad\xd8\xe6"
  "turquoise" "\x40\xe0\xd0"
  "LightCyan" "\xe0\xff\xff"
  "CadetBlue" "\x5f\x9e\xa0"
  "DarkGreen" "\x00\x64\x00"
  "PaleGreen" "\x98\xfb\x98"
  "LawnGreen" "\x7c\xfc\x00"
  "LimeGreen" "\x32\xcd\x32"
  "OliveDrab" "\x6b\x8e\x23"
  "DarkKhaki" "\xbd\xb7\x6b"
  "goldenrod" "\xda\xa5\x20"
  "RosyBrown" "\xbc\x8f\x8f"
  "IndianRed" "\xcd\x5c\x5c"
  "burlywood" "\xde\xb8\x87"
  "chocolate" "\xd2\x69\x1e"
  "firebrick" "\xb2\x22\x22"
  "OrangeRed" "\xff\x45\x00"
  "LightPink" "\xff\xb6\xc1"
  "VioletRed" "\xd0\x20\x90"
  "\x00"
}; 

static const byte Tab9Color4Dat[] = {
  "seashell" "\xff\xf5\xee" "\xee\xe5\xde" "\xcd\xc5\xbf" "\x8b\x86\x82"
  "cornsilk" "\xff\xf8\xdc" "\xee\xe8\xcd" "\xcd\xc8\xb1" "\x8b\x88\x78"
  "honeydew" "\xf0\xff\xf0" "\xe0\xee\xe0" "\xc1\xcd\xc1" "\x83\x8b\x83"
  "SeaGreen" "\x54\xff\x9f" "\x4e\xee\x94" "\x43\xcd\x80" "\x2e\x8b\x57"
  "DeepPink" "\xff\x14\x93" "\xee\x12\x89" "\xcd\x10\x76" "\x8b\x0a\x50"
  "\x00"
};

static const byte Tab10ColorNameDat[] = {
  "GhostWhite" "\xf8\xf8\xff"
  "WhiteSmoke" "\xf5\xf5\xf5"
  "PapayaWhip" "\xff\xef\xd5"
  "MediumBlue" "\x00\x00\xcd"
  "DodgerBlue" "\x1e\x90\xff"
  "PowderBlue" "\xb0\xe0\xe6"
  "aquamarine" "\x7f\xff\xd4"
  "chartreuse" "\x7f\xff\x00"
  "SandyBrown" "\xf4\xa4\x60"
  "DarkSalmon" "\xe9\x96\x7a"
  "DarkOrange" "\xff\x8c\x00"
  "LightCoral" "\xf0\x80\x80"
  "DarkOrchid" "\x99\x32\xcc"
  "DarkViolet" "\x94\x00\xd3"
  "BlueViolet" "\x8a\x2b\xe2"
  "LightGreen" "\x90\xee\x90"
  "\x00"
};

static const byte Tab10Color4Dat[] = {
  "PeachPuff" "\xff\xda\xb9" "\xee\xcb\xad" "\xcd\xaf\x95" "\x8b\x77\x65"
  "MistyRose" "\xff\xe4\xe1" "\xee\xd5\xd2" "\xcd\xb7\xb5" "\x8b\x7d\x7b"
  "SlateBlue" "\x83\x6f\xff" "\x7a\x67\xee" "\x69\x59\xcd" "\x47\x3c\x8b"
  "RoyalBlue" "\x48\x76\xff" "\x43\x6e\xee" "\x3a\x5f\xcd" "\x27\x40\x8b"
  "SteelBlue" "\x63\xb8\xff" "\x5c\xac\xee" "\x4f\x94\xcd" "\x36\x64\x8b"
  "SlateGray" "\xc6\xe2\xff" "\xb9\xd3\xee" "\x9f\xb6\xcd" "\x6c\x7b\x8b"
  "LightBlue" "\xbf\xef\xff" "\xb2\xdf\xee" "\x9a\xc0\xcd" "\x68\x83\x8b"
  "LightCyan" "\xe0\xff\xff" "\xd1\xee\xee" "\xb4\xcd\xcd" "\x7a\x8b\x8b"
  "CadetBlue" "\x98\xf5\xff" "\x8e\xe5\xee" "\x7a\xc5\xcd" "\x53\x86\x8b"
  "turquoise" "\x00\xf5\xff" "\x00\xe5\xee" "\x00\xc5\xcd" "\x00\x86\x8b"
  "PaleGreen" "\x9a\xff\x9a" "\x90\xee\x90" "\x7c\xcd\x7c" "\x54\x8b\x54"
  "OliveDrab" "\xc0\xff\x3e" "\xb3\xee\x3a" "\x9a\xcd\x32" "\x69\x8b\x22"
  "goldenrod" "\xff\xc1\x25" "\xee\xb4\x22" "\xcd\x9b\x1d" "\x8b\x69\x14"
  "RosyBrown" "\xff\xc1\xc1" "\xee\xb4\xb4" "\xcd\x9b\x9b" "\x8b\x69\x69"
  "IndianRed" "\xff\x6a\x6a" "\xee\x63\x63" "\xcd\x55\x55" "\x8b\x3a\x3a"
  "burlywood" "\xff\xd3\x9b" "\xee\xc5\x91" "\xcd\xaa\x7d" "\x8b\x73\x55"
  "chocolate" "\xff\x7f\x24" "\xee\x76\x21" "\xcd\x66\x1d" "\x8b\x45\x13"
  "firebrick" "\xff\x30\x30" "\xee\x2c\x2c" "\xcd\x26\x26" "\x8b\x1a\x1a"
  "OrangeRed" "\xff\x45\x00" "\xee\x40\x00" "\xcd\x37\x00" "\x8b\x25\x00"
  "LightPink" "\xff\xae\xb9" "\xee\xa2\xad" "\xcd\x8c\x95" "\x8b\x5f\x65"
  "VioletRed" "\xff\x3e\x96" "\xee\x3a\x8c" "\xcd\x32\x78" "\x8b\x22\x52"
  "\x00"
};

static const byte Tab11ColorNameDat[] = {
  "FloralWhite" "\xff\xfa\xf0"
  "NavajoWhite" "\xff\xde\xad"
  "DeepSkyBlue" "\x00\xbf\xff"
  "SpringGreen" "\x00\xff\x7f"
  "GreenYellow" "\xad\xff\x2f"
  "YellowGreen" "\x9a\xcd\x32"
  "ForestGreen" "\x22\x8b\x22"
  "LightYellow" "\xff\xff\xe0"
  "SaddleBrown" "\x8b\x45\x13"
  "LightSalmon" "\xff\xa0\x7a"
  "DarkMagenta" "\x8b\x00\x8b"
  "\x00"
};

static const byte Tab11Color4Dat[] = {
  "DodgerBlue" "\x1e\x90\xff" "\x1c\x86\xee" "\x18\x74\xcd" "\x10\x4e\x8b"
  "aquamarine" "\x7f\xff\xd4" "\x76\xee\xc6" "\x66\xcd\xaa" "\x45\x8b\x74"
  "chartreuse" "\x7f\xff\x00" "\x76\xee\x00" "\x66\xcd\x00" "\x45\x8b\x00"
  "DarkOrange" "\xff\x7f\x00" "\xee\x76\x00" "\xcd\x66\x00" "\x8b\x45\x00"
  "DarkOrchid" "\xbf\x3e\xff" "\xb2\x3a\xee" "\x9a\x32\xcd" "\x68\x22\x8b"
  "\x00"
};

static const byte Tab12ColorNameDat[] = {
  "AntiqueWhite" "\xfa\xeb\xd7"
  "LemonChiffon" "\xff\xfa\xcd"
  "MidnightBlue" "\x19\x19\x70"
  "LightSkyBlue" "\x87\xce\xfa"
  "DarkSeaGreen" "\x8f\xbc\x8f"
  "MediumOrchid" "\xba\x55\xd3"
  "MediumPurple" "\x93\x70\xdb"
  "\x00"
};

static const byte Tab12Color4Dat[] = {
  "NavajoWhite" "\xff\xde\xad" "\xee\xcf\xa1" "\xcd\xb3\x8b" "\x8b\x79\x5e"
  "DeepSkyBlue" "\x00\xbf\xff" "\x00\xb2\xee" "\x00\x9a\xcd" "\x00\x68\x8b"
  "SpringGreen" "\x00\xff\x7f" "\x00\xee\x76" "\x00\xcd\x66" "\x00\x8b\x45"
  "LightYellow" "\xff\xff\xe0" "\xee\xee\xd1" "\xcd\xcd\xb4" "\x8b\x8b\x7a"
  "LightSalmon" "\xff\xa0\x7a" "\xee\x95\x72" "\xcd\x81\x62" "\x8b\x57\x42"
  "\x00"
};

static const byte Tab13ColorNameDat[] = {
  "LavenderBlush" "\xff\xf0\xf5"
  "DarkSlateBlue" "\x48\x3d\x8b"
  "PaleTurquoise" "\xaf\xee\xee"
  "DarkTurquoise" "\x00\xce\xd1"
  "LightSeaGreen" "\x20\xb2\xaa"
  "PaleGoldenrod" "\xee\xe8\xaa"
  "DarkGoldenrod" "\xb8\x86\x0b"
  "PaleVioletRed" "\xdb\x70\x93"
  "\x00"
};

static const byte Tab13Color4Dat[] = {
  "AntiqueWhite" "\xff\xef\xdb" "\xee\xdf\xcc" "\xcd\xc0\xb0" "\x8b\x83\x78"
  "LemonChiffon" "\xff\xfa\xcd" "\xee\xe9\xbf" "\xcd\xc9\xa5" "\x8b\x89\x70"
  "LightSkyBlue" "\xb0\xe2\xff" "\xa4\xd3\xee" "\x8d\xb6\xcd" "\x60\x7b\x8b"
  "DarkSeaGreen" "\xc1\xff\xc1" "\xb4\xee\xb4" "\x9b\xcd\x9b" "\x69\x8b\x69"
  "MediumOrchid" "\xe0\x66\xff" "\xd1\x5f\xee" "\xb4\x52\xcd" "\x7a\x37\x8b"
  "MediumPurple" "\xab\x82\xff" "\x9f\x79\xee" "\x89\x68\xcd" "\x5d\x47\x8b"
  "\x00"
};

static const byte Tab14ColorNameDat[] = {
  "BlanchedAlmond" "\xff\xeb\xcd"
  "LightSlateGray" "\x77\x88\x99"
  "CornflowerBlue" "\x64\x95\xed"
  "LightSlateBlue" "\x84\x70\xff"
  "LightSteelBlue" "\xb0\xc4\xde"
  "DarkOliveGreen" "\x55\x6b\x2f"
  "MediumSeaGreen" "\x3c\xb3\x71"
  "LightGoldenrod" "\xee\xdd\x82"
  "\x00"
};

static const byte Tab14Color4Dat[] = {
  "LavenderBlush" "\xff\xf0\xf5" "\xee\xe0\xe5" "\xcd\xc1\xc5" "\x8b\x83\x86"
  "PaleTurquoise" "\xbb\xff\xff" "\xae\xee\xee" "\x96\xcd\xcd" "\x66\x8b\x8b"
  "DarkSlateGray" "\x97\xff\xff" "\x8d\xee\xee" "\x79\xcd\xcd" "\x52\x8b\x8b"
  "DarkGoldenrod" "\xff\xb9\x0f" "\xee\xad\x0e" "\xcd\x95\x0c" "\x8b\x65\x08"
  "PaleVioletRed" "\xff\x82\xab" "\xee\x79\x9f" "\xcd\x68\x89" "\x8b\x47\x5d"
  "\x00"
};

static const byte Tab15ColorNameDat[] = {
  "MediumSlateBlue" "\x7b\x68\xee"
  "MediumTurquoise" "\x48\xd1\xcc"
  "MediumVioletRed" "\xc7\x15\x85"
  "\x00"
};

static const byte Tab15Color4Dat[] = {
  "LightSteelBlue" "\xca\xe1\xff" "\xbc\xd2\xee" "\xa2\xb5\xcd" "\x6e\x7b\x8b"
  "DarkOliveGreen" "\xca\xff\x70" "\xbc\xee\x68" "\xa2\xcd\x5a" "\x6e\x8b\x3d"
  "LightGoldenrod" "\xff\xec\x8b" "\xee\xdc\x82" "\xcd\xbe\x70" "\x8b\x81\x4c"
  "\x00"
};

static const byte Tab16ColorNameDat[] = {
  "MediumAquamarine" "\x66\xcd\xaa"
  "\x00"
};

static const byte Tab17ColorNameDat[] = {
  "MediumSpringGreen" "\x00\xfa\x9a"
  "\x00"
};

static const byte Tab20ColorNameDat[] = {
  "LightGoldenrodYellow" "\xfa\xfa\xd2"
  "\x00"
};

static const byte *tab_name[21] = {
  NULL, 	     NULL,		NULL,		   Tab3ColorNameDat, 
  Tab4ColorNameDat,  Tab5ColorNameDat,  Tab6ColorNameDat,  Tab7ColorNameDat,
  Tab8ColorNameDat,  Tab9ColorNameDat,  Tab10ColorNameDat, Tab11ColorNameDat,
  Tab12ColorNameDat, Tab13ColorNameDat, Tab14ColorNameDat, Tab15ColorNameDat,
  Tab16ColorNameDat, Tab17ColorNameDat, NULL,		   NULL,
  Tab20ColorNameDat
};

static const byte *tab_4name[21] = {
  NULL, 	     NULL,		NULL,		   NULL, 
  Tab4Color4Dat,     Tab5Color4Dat,	Tab6Color4Dat,     Tab7Color4Dat,
  Tab8Color4Dat,     Tab9Color4Dat,	Tab10Color4Dat,    Tab11Color4Dat,
  Tab12Color4Dat,    Tab13Color4Dat,	Tab14Color4Dat,    Tab15Color4Dat,
  NULL,		     NULL,		NULL,		   NULL,
  NULL
};
 
int GetColorByName(const char *s) {
  int m;

	if (*s == '#') {
	    s++;
	    return pctohex(s);
	}

	if (!pcncasecmpASCII(s, "gray", 4)) {
	    s += 4;
	    if (*s && (*s >= '0') && (*s <= '9')) m = (255 * pctoint(s) / 100) & 0xff;
	    else m = 0xbe;
	    return m | (m<<8) | (m<<16);
	}
  int lg = 0;
  const byte *pc = (const byte*)s;
  const byte *cnd;

	while (*pc && (((*pc >= '0') && (*pc <= '9')) ||
		       ((*pc >= 'A') && (*pc <= 'Z')) ||
		       ((*pc >= 'a') && (*pc <= 'z')))) {
	    lg++; pc++;
	}
	if (lg > 20) return 0;

	if (tab_4name[lg] && (s[lg-1] >= '1') && (s[lg-1] <= '4')) {
	    cnd = tab_4name[lg];
	    while (cnd[0]) {
		if (!pcncasecmpASCII(s, (const char*)cnd, lg-1)) {
		    m = lg-1 + ((s[lg-1] & 0x0f)-1) * 3;
		    return ((cnd[m] << 16) | (cnd[m+1] << 8) | cnd[m+2]);
		}
		cnd += lg+11;
	    }
	}

	cnd = tab_name[lg];
	if (!cnd) return 0;

	while (cnd[0]) {
	    if (!pcncasecmpASCII(s, (const char*)cnd, lg)) {
		return ((cnd[lg] << 16) | (cnd[lg+1] << 8) | cnd[lg+2]);
	    }
	    cnd += lg+3;
	}
	return 0;
}

