/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

static int TImage_img_malloc(TImage *self) {

	if ((self->width <= 0)||(self->height <= 0)) return False;
	if (!(self->pxm_LD & pxmLD_bppMask)) self->pxm_LD |= pxm_4b_24bpp;
	if (self->memfu & memfu_data_free) ppfree(&self->data);
	ppfree(&self->mask_data);

	if ((self->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
	    self->bytes_per_line = ((self->width + 1) >> 1) << 2;  //w*4
	} else if (((self->pxm_LD & pxmLD_bppMask) == pxm_4b_24bpp) ||
		   ((self->pxm_LD & pxmLD_bppMask) == pxm_4b_24bpp8a)) {
	    self->bytes_per_line = self->width << 2;  //w*4
	} else return False;

	self->data = xcalloc(self->bytes_per_line * self->height);
	self->memfu = memfu_data_free;
	self->stat |= pmxStat_ImgAkt;

	if (self->stat & pmxStat_HandleMask) {
	    self->mask_data = xcalloc(((self->width+7) >> 3) * self->height);
	    self->stat &= ~pmxStat_XMaskPixAkt;
	    self->stat |= pmxStat_MaskImgAkt;
	}
	return True;
}

bool TImage_SetMem(TImage *self) {
	if (!self->img_malloc) return False;
	return self->img_malloc(self);
}

/*____________________________ Free __________________________________*/
bool TImage__Destroy(TImage *self) {
	if (self->memfu & memfu_data_free) xfree(self->data);
	xfree(self->img_info);
	xfree(self->mask_data);
	xfree(self->rgb565_alpha_data);
	if (self->memfu & memfu_cmap_free) Tfree(self->cmap);
	Tfree(self->Pen);
	Tfree(self->ExImg);
	return False;
}

/*____________________________ Create ________________________________*/
TImage *TImage_Create(int size, int wtyp) {
  TImage *self;

	if (size <= 0) {
	    size = sizeof(TImage);
	    wtyp = wtyp_TImage;
	}
	self = ttc_create(size, wtyp);
	self->destroy = (TDESTROY)TImage__Destroy;
	self->img_malloc = TImage_img_malloc;
	return self;
}

TImage *TImage_Create_dup_struct(TImage *org, int fu) {
  TImage *self;

	if (!org) return TImage_Create(0, 0);
	self = ttc_create(org->wtcsize, org->wtyp);
	self->destroy = org->destroy;
	self->width  = org->width;
	self->height = org->height;
	self->pxm_LD = org->pxm_LD;
	self->stat = org->stat & pmxStat_ckh_alpha;
	self->img_malloc = org->img_malloc;
	if (fu & (dupfu_init_imgmem | dupfu_copy_img)) {
	    self->img_malloc(self);
	    if (fu & dupfu_copy_img) {
		if (self->bytes_per_line != org->bytes_per_line) {
		    int lg = self->bytes_per_line;
		    if (self->bytes_per_line > org->bytes_per_line) lg = org->bytes_per_line;
		    byte *pd = self->data;
		    byte *ps = org->data;
		    for (int i=org->height;i>0;i--) {
			pnmove(pd, ps, lg);
			ps += org->bytes_per_line;
			pd += self->bytes_per_line;
		    }
		} else {
		    pnmove(self->data, org->data, org->bytes_per_line * org->height);
		}
	    }
	    if (org->rgb565_alpha_data && (fu & dupfu_copy_alpha)) {
		self->rgb565_alpha_data = pndup(org->rgb565_alpha_data, org->width * org->height);
	    }
	}
	if (org->mask_data && (fu & dupfu_copy_mask)) {
	    self->mask_data = pndup(org->mask_data, ((org->width+7) >> 3) * org->height);
	}
	return self;
}

TImage *TImage_Create_wht(int size, int wtyp, int width, int height, dword pxm_LD) {
  TImage *self = TImage_Create(size, wtyp);

	self->pxm_LD = pxm_LD;
	self->width = width; 
	self->height = height;
	if (self->img_malloc) self->img_malloc(self);
	return self;
}
