/*
 Copyright (C) 2010 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

typedef struct {
  ic32 mincols;
  ic32 maxcols;
  ic32 *cols;
  dword colslg;
  int sortfu;
} TBlk;

typedef struct {
  int idx;
  int sortfu;
  TBlk *blk[];
} TQBlk;

static dword get_median(dword a, dword b, dword c) {
	if (a < b) {
	  if (b < c) return b;
	  if (a < c) return c;
	  return a;
	}
	if (a < c) return a;
	if (b < c) return c;
	return b;
}

static ic32 *sort_median(ic32 *L, ic32 *R, dword median, int sf) {
  ic32 k;
	while (True) {
	    while (L->b[sf] < median) ++L;
	    R--;
	    while (R->b[sf] > median) --R;
	    if (L >= R) return L;
	    k = *L; *L = *R; *R = k;
	    L++;
	}
}

static void sort_ptn(ic32 *L, ic32 *M, ic32 *R, int sf) {
  ic32 *nx;
  dword m;

	while (R - L > 3) {
	    m = get_median(L->b[sf], (L+((R-L)/ 2))->b[sf], (R-1)->b[sf]);
	    nx = sort_median(L, R, m, sf);
	    if (nx <= M) L = nx; else R = nx;
	}
}

static void TBlk_init(TBlk *self, ic32 *cols, dword colslg) {
	self->cols = cols;
	self->colslg = colslg;
	self->mincols.dw = 0x00ffffff;
	self->maxcols.dw = 0;
}

static void TBlk_SetMinMax(TBlk *self) {
  ic32 min, max, *pt;
	pt = self->cols;
	if (!self->sortfu) {
	    min.dw = max.dw = pt->dw;
	    for(int i=1; i < self->colslg; i++) {
		for(int j=0; j<3; j++) {
		    byte c = pt->b[j];
		    if (min.b[j] > c) min.b[j] = c;
		    if (max.b[j] < c) max.b[j] = c;
		}
		pt++;
	    }
	    self->mincols.dw = min.dw;
	    self->maxcols.dw = max.dw;
	} else {
	    self->mincols.dw = pt[0].dw;
  int maxidx = self->colslg-1;
	    if (maxidx < 0) maxidx = 0;
	    self->maxcols.dw = pt[maxidx].dw;
	}
}

static int MaxDiffIdx(ic32 mincols, ic32 maxcols) {
  int m0 = maxcols.b[0] - mincols.b[0];
  int m1 = maxcols.b[1] - mincols.b[1];
  int m2 = maxcols.b[2] - mincols.b[2];
  int idx = 0;

	if (m0 < m1) {m0 = m1; idx = 1;}
	if (m0 < m2) idx = 2;
	return idx;
}

static int MaxDiff(ic32 mincols, ic32 maxcols) {
  int i = MaxDiffIdx(mincols, maxcols);
	return maxcols.b[i] - mincols.b[i];
}

static void TQBlk_SetMaxTop(TQBlk *qblk) {
  int maxidx, diff, m;
	if (qblk->idx < 2) return;
	maxidx = diff = 0;
	for (int i=0;i<qblk->idx;i++) {
	    m = MaxDiff(qblk->blk[i]->mincols, qblk->blk[i]->maxcols);
	    if (diff < m) {
		diff = m;
		maxidx = i;
	    }
	}
	if (maxidx < (qblk->idx-1)) {
	    TBlk *k = qblk->blk[qblk->idx-1]; 
	    qblk->blk[qblk->idx-1] = qblk->blk[maxidx];
	    qblk->blk[maxidx] = k;
	}
}


int TImage_CutColorMap(TImage *self, int NewColors, int sortfu) {
  TQBlk *qblk;
  TBlk *blk0;
  ic32 *col_map;
  TColMap *cmap = self->cmap;

	if (!cmap) return 0;
	ppfree(&cmap->quant_col_map);
	cmap->quant_count = 0;
//	cmap->quant_max_count = 0;
	if (!cmap->col_map || (cmap->ColorCount <= NewColors)) return 0;

	NewColors -= cmap->default_quant_count;
	if (NewColors <= 0) {
	    cmap->quant_count = cmap->default_quant_count;
	    cmap->quant_col_map = xmalloc(sizeof(ic32)*cmap->quant_count);
	    pnmove(cmap->quant_col_map, cmap->default_quant_col_map, sizeof(ic32)*cmap->quant_count);
	    return cmap->quant_count;
	}


	qblk = xcalloc(sizeof(TQBlk)+(sizeof(TBlk*)*NewColors));
	qblk->sortfu = sortfu;
	col_map = pndup(cmap->col_map, sizeof(ic32)*cmap->ColorCount);
	if (sortfu) ColorMap_Sort(col_map, cmap->ColorCount, sortfu);

	for (int i=0;i<NewColors;i++) {
	    qblk->blk[i] = xcalloc(sizeof(TBlk));
	    qblk->blk[i]->sortfu = sortfu;
	}
	blk0 = qblk->blk[0];
	TBlk_init(blk0, col_map, cmap->ColorCount);
	TBlk_SetMinMax(blk0);
	qblk->idx++;

	while (qblk->idx < NewColors) {
	    blk0 = qblk->blk[qblk->idx-1];
	    if (blk0->colslg < 2) break;
  ic32 *L = blk0->cols;
  ic32 *M = L + ((blk0->colslg)>>1);
  ic32 *R = L + blk0->colslg;

	    sort_ptn(L, M, R, MaxDiffIdx(blk0->mincols, blk0->maxcols));

	    TBlk_init(blk0, L, M-L);
	    TBlk_SetMinMax(blk0);

	    blk0 = qblk->blk[qblk->idx];
	    qblk->idx++;
	    TBlk_init(blk0, M, R-M);
	    TBlk_SetMinMax(blk0);

	    TQBlk_SetMaxTop(qblk);
	}

  int idx = qblk->idx;
  int cidx = 0;
//	cmap->quant_max_count = idx+4;
  ic32 *quant_map = xmalloc(sizeof(ic32)*(idx + cmap->default_quant_count));
	cmap->quant_col_map = quant_map;
	while (idx > 0) {
	    idx--;
	    blk0 = qblk->blk[idx];
  int colslg = blk0->colslg;
  dword r, g, b;
  ic32 *cols = blk0->cols;
            if (colslg > 0) {
		r = g = b = 0;
		for (int i=colslg;i;i--) {
		    b += cols->b[0];
		    g += cols->b[1];
		    r += cols->b[2];
		    cols++;
		}
		quant_map->b[0] = b / colslg;
		quant_map->b[1] = g / colslg;
		quant_map->b[2] = r / colslg;
		quant_map++;
		cidx++;
// printf("%4d   %3u %3u %3u\n", blk0->colslg, b, g, r);
	    }
	}
	cmap->quant_count = cidx + cmap->default_quant_count;

	if (cmap->default_quant_count > 0) {
	    pnmove(quant_map, cmap->default_quant_col_map, sizeof(ic32)*cmap->default_quant_count);
	}

	for (int i=0;i<NewColors;i++) xfree(qblk->blk[i]);
	xfree(col_map);
	xfree(qblk);

	return cidx;
}
