/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"
#include "pen_pv.h"
#include <math.h>

#define VKSIZE 16

typedef byte Tvk[VKSIZE * 6];

typedef struct {
  int x;
  int y;
} Txy;

typedef struct {
  float x;
  float y;
} TFloatPoint;

#define MAX_RECT_POINT 8

#define bixQ_ERROR 0x014000

#define OPT_SIZE 1
#define OPT_RGB_MACRO 0

/* lflag */
#define RI_X1_NEG	1
#define RI_X1_oldNEG	(1<<1)
#define RI_Y1_NEG	(1<<2)
#define RI_Y1_oldNEG	(1<<3)

#define RI_X2_NEG	(1<<4)
#define RI_X2_oldNEG	(1<<5)
#define RI_Y2_NEG	(1<<6)
#define RI_Y2_oldNEG	(1<<7)

#define W_X1_ge45	(1<<8)
#define W_X1_old_ge45	(1<<9)
#define W_X2_ge45	(1<<10)
#define W_X2_old_ge45	(1<<11)
#define LINE_TOP	(1<<12)
#define LINE_BOTTOM	(1<<13)
#define CORNER_X1	(1<<14)
#define CORNER_X2	(1<<15)
#define CORNER		(1<<16)


#define TCSelf TPenALD
typedef struct TPenALD {
  AlgTPen
  void (*lineN_to)(struct TCSelf *self, float X1, float Y1);
  void REGP3 (*bix_rgb) (struct TPenALD *self, int k);
  void REGP3 (*bix_argb)(struct TPenALD *self, int k, dword A1);
  void REGP3 (*bixn_rgb) (struct TPenALD *self, int n, int k);
  void REGP3 (*bixn_argb)(struct TPenALD *self, int n, int k, dword A1);
  int bpp, bpl;
  dword Color565;
  float X1, Y1, X2, Y2;
  float lxg, alxg, lyg, alyg, lxyg;
  float lx, alx, ly, aly, lxy, alxy, slxy;
  float kx_xy, ky_xy;
  int a1add, o1add, a2add, o2add;
  int a1lg, o1lg, a2lg, o2lg, aNlg;
  int max4p;
  int rectpIdx;
  TFloatPoint fp4[MAX_RECT_POINT];
  dword max;
  int yc1, yc2;
  float lx1, ly1, alx1, aly1, kx1, nX1, oX1;
  float lx2, ly2, alx2, aly2, kx2, nX2, oX2;
  dword A, lflag;
  bool rix, rgb565, ncb2, ncb3;
  union {
    byte  *pb;
    word  *pw;
    dword *pd;
  };
  Tvk vk;
} TPenALD;
#undef TCSelf

/* ---------------------- rgb 565 --------------------------------------- */
static void REGP3 bix_rgb16(TPenALD *self, int k) {
    word *pw = self->pw + k;
    *pw = self->Color565;
}

static void REGP3 bixn_rgb16(TPenALD *self, int n, int k) {
    word *pw = self->pw + k;
    word col = self->Color565;
    while (n-- > 0) *pw++ = col;
}

static void REGP3 bix_argb16(TPenALD *self, int k, dword A1) {
    word *pw = self->pw + k;
    dword A2 = 0x100 - A1;
    dword R = (((self->Color565 & 0xf800) * A1) + ((*pw & 0xf800) * A2)) & 0xf80000;
    dword G = (((self->Color565 & 0x07e0) * A1) + ((*pw & 0x07e0) * A2)) & 0x07e000;
    dword B = (((self->Color565 & 0x001f) * A1) + ((*pw & 0x001f) * A2)) & 0x001f00;
    *pw = (R | G | B) >> 8;
}

static void REGP3 bixn_argb16(TPenALD *self, int n, int k, dword A1) {
    word *pw = self->pw + k;
    dword A2 = 0x100 - A1;
    while (n-- > 0) {
	dword R = (((self->Color565 & 0xf800) * A1) + ((*pw & 0xf800) * A2)) & 0xf80000;
	dword G = (((self->Color565 & 0x07e0) * A1) + ((*pw & 0x07e0) * A2)) & 0x07e000;
	dword B = (((self->Color565 & 0x001f) * A1) + ((*pw & 0x001f) * A2)) & 0x001f00;
	*pw = (R | G | B) >> 8;
	pw++;
    }
}

#if OPT_SIZE == 0
static void line16_to(TPenALD *self, float X1, float Y1) {
}
#endif

/* ---------------------- rgb 32 --------------------------------------- */
static void REGP3 bix_rgb32(TPenALD *self, int k) {
    dword *pw = self->pd + k;
    *pw = self->Color;
}

static void REGP3 bixn_rgb32(TPenALD *self, int n, int k) {
    dword *pw = self->pd + k;
    dword col = self->Color;
    while (n-- > 0) *pw++ = col;
}

static void REGP3 bix_argb32(TPenALD *self, int k, dword A1) {
    dword *pw = self->pd + k;
    dword A2 = 0x100 - A1;
    dword R = (((self->Color & 0xff0000) * A1) + ((*pw & 0xff0000) * A2)) & 0xff000000;
    dword G = (((self->Color & 0x00ff00) * A1) + ((*pw & 0x00ff00) * A2)) & 0x00ff0000;
    dword B = (((self->Color & 0x0000ff) * A1) + ((*pw & 0x0000ff) * A2)) & 0x0000ff00;
    *pw = ((R | G | B) >> 8) | (*pw & 0xff000000);
}

static void REGP3 bixn_argb32(TPenALD *self, int n, int k, dword A1) {
    dword *pw = self->pd + k;
    dword A2 = 0x100 - A1;
    while (n-- > 0) {
	dword R = (((self->Color & 0xff0000) * A1) + ((*pw & 0xff0000) * A2)) & 0xff000000;
	dword G = (((self->Color & 0x00ff00) * A1) + ((*pw & 0x00ff00) * A2)) & 0x00ff0000;
	dword B = (((self->Color & 0x0000ff) * A1) + ((*pw & 0x0000ff) * A2)) & 0x0000ff00;
	*pw = ((R | G | B) >> 8) | (*pw & 0xff000000);
	pw++;
    }
}

#if OPT_RGB_MACRO
#define MAC_RGB32 \
    { \
    dword *pw = wx0 + k; \
    dword A1 = ak >> 16; \
    dword A2 = 0x100 - A1; \
    R = (((col & 0xff0000) * A1) + ((*pw & 0xff0000) * A2)) & 0xff000000; \
    G = (((col & 0x00ff00) * A1) + ((*pw & 0x00ff00) * A2)) & 0x00ff0000; \
    B = (((col & 0x0000ff) * A1) + ((*pw & 0x0000ff) * A2)) & 0x0000ff00; \
    *pw = ((R | G | B) >> 8) | (*pw & 0xff000000); \
    }
#endif

static void line32_to(TPenALD *self, float X1, float Y1) {
  int i_xo, i_yo, koff, w, h;
  byte *pbk;
  int k, xoi, yoi, bpl, ri, ricount, bixcount;
#if OPT_RGB_MACRO
  dword *wx0;
  dword col, R, G, B;
#endif
  dword max, bixlg, ak;
  float fkoff;
  bool startloop;

	bpl = self->bpl; //self->img->bytes_per_line >> 2;
	w = self->img->width;
	h = self->img->height;
	max = w * h;
	self->max = max;
#if OPT_RGB_MACRO
	wx0 = (dword*)self->img->data;
	col = self->Color;
#else
	self->pb = self->img->data;
#endif
	i_xo  = (int)(X1  * 0x10000);
	if (self->lxg < 0) i_xo -= 0x10000;
	i_yo  = (int)(Y1  * 0x10000);
	if (self->lyg < 0) i_yo -= 0x10000;
// printf("x=0x%04x y=0x%04x lx=%.1f ly=%.1f\n", i_xo, i_yo, ld->lx, ld->ly);
	
	if (self->alxg >= self->alyg) {
	    ri = 1;
	    if (self->lxg < 0) ri = -1;
	    fkoff = self->lyg / self->alxg;
	    ricount = (i_xo >> 16);
	    bixcount = 0;
	    bixlg = i_xo & 0xffff;
	    startloop = bixlg > 0x01ff;
	    if (!startloop) goto start_x;
	    if (ri > 0) bixlg = (0x10000 - bixlg); else ricount++;
	    koff = (int)(fkoff * bixlg);
loop_x:
		if ((dword)ricount < (dword)w) {
		    yoi = (i_yo >> 16) - 1;
		    k = ricount + (yoi * bpl);
		    pbk = self->vk + ((i_yo >> 12) & 0x0f);
#if OPT_SIZE
		    for (int i=4;i>0;i--) {
			ak = *pbk;
			if (ak && ((dword)k < max)) {
			    ak = (ak * bixlg) >> 16;
			    self->bix_argb(self, k, ak);
			}
			k += bpl; pbk += VKSIZE;
		    }
#else
		    ak = *pbk;
		    if (ak && ((dword)k < max)) {
			ak *= bixlg;
			MAC_RGB32
		    }
		    k += bpl; pbk += VKSIZE;
		    if ((dword)k < max) {
			ak = *pbk * bixlg;
			MAC_RGB32
		    }

		    k += bpl; pbk += VKSIZE;
		    if ((dword)k < max) {
			ak = *pbk * bixlg;
			MAC_RGB32
		    }
		    k += bpl; pbk += VKSIZE;
		    ak = *pbk;
		    if (ak && ((dword)k < max)) {
			ak *= bixlg;
			MAC_RGB32
		    }
#endif
		}
		ricount += ri; i_yo += koff; bixcount -= bixlg;
//	    }
	    if (bixcount > 0xffff) goto loop_x;

	    if (startloop) {
		startloop = false;
start_x:
		bixcount += (int)(self->alx * 0x10000);
		bixlg = 0x10000;
		koff = (int)(fkoff * 0x10000);
		if (bixcount >= 0x10000) goto loop_x;
	    }

	    if (bixcount > 0x01ff) {
		bixlg = bixcount;
		goto loop_x;
	    }

	} else {
	    ri = 1;
	    if (self->lyg < 0) ri = -1;
	    fkoff = self->lxg / self->alyg;
	    ricount = (i_yo >> 16);
	    bixcount = 0;
	    bixlg = i_yo & 0xffff;
	    startloop = bixlg > 0x01ff;
	    if (!startloop) goto start_y;
	    if (ri > 0) bixlg = (0x10000 - bixlg); else ricount++;
	    koff = (int)(fkoff * bixlg);

//	    while (bixcount > 0xffff) {
loop_y:
		if ((dword)ricount < (dword)h) {
		    xoi = (i_xo >> 16) - 1;
		    k = xoi + (ricount * bpl);
		    pbk = self->vk + ((i_xo >> 12) & 0x0f);
#if OPT_SIZE
		    for (int i=4;i>0;i--) {
			if (*pbk && ((dword)k < max)) {
			    ak = (*pbk * bixlg) >> 16;
			    self->bix_argb(self, k, ak);
			}
			k += 1; pbk += VKSIZE;
		    }
#else
		    ak = *pbk;
		    if (ak && ((dword)xoi < (dword)w)) {
			ak *= bixlg;
			MAC_RGB32
		    }
		    k += 1; xoi++; pbk += VKSIZE;
		    if ((dword)xoi < (dword)w) {
			ak = *pbk * bixlg;
			MAC_RGB32
		    }
		    k += 1; xoi++; pbk += VKSIZE;
		    if ((dword)xoi < (dword)w) {
			ak = *pbk * bixlg;
			MAC_RGB32
		    }
		    k += 1; xoi++; pbk += VKSIZE;
		    ak = *pbk;
		    if (ak && ((dword)xoi < (dword)w)) {
			ak *= bixlg;
			MAC_RGB32
		    }
#endif
		}
		ricount += ri; i_xo += koff; bixcount -= bixlg;
//	    }
	    if (bixcount > 0xffff) goto loop_y;

	    if (startloop) {
		startloop = false;
start_y:
		bixcount += (int)(self->aly * 0x10000);
		bixlg = 0x10000;
		koff = (int)(fkoff * 0x10000);
		if (bixcount >= 0x10000) goto loop_y;
	    }
	    if (bixcount > 0x01ff) {
		bixlg = bixcount;
		goto loop_y;
	    }
	}
}

static void __attribute__ ((noinline)) pixel05korr(TPenALD *self) {
  float tf;

	if ((self->kx_xy >= 0) && (self->ky_xy >= 0)) { //bixel0 seg x+ y+
	    if (self->kx_xy >= self->ky_xy) {
		tf = (self->kx_xy - self->ky_xy) * 0.5;
		self->Y1 -= tf; self->Y2 -= tf;
	    } else {
		tf = (self->ky_xy - self->kx_xy) * 0.5;
		self->X1 -= tf; self->X2 -= tf;
	    }
	} else if ((self->kx_xy < 0)&&(self->ky_xy >= 0)) { //bixel0 seg x- y+
	    if (-self->kx_xy >= self->ky_xy) {
		tf = (self->kx_xy + self->ky_xy) * 0.5;
		self->Y1 += tf; self->Y2 += tf;
	    } else {
		tf = (self->ky_xy + self->kx_xy) * 0.5;
		self->X1 += tf; self->X2 += tf;
	    }
	} else if ((self->kx_xy >= 0)&&(self->ky_xy < 0)) { //bixel0 seg x+ y-
	    if (self->kx_xy >= -self->ky_xy) {
		tf = (self->kx_xy + self->ky_xy) * 0.5;
		self->Y1 += tf; self->Y2 += tf;
	    } else {
		tf = (self->ky_xy + self->kx_xy) * 0.5;
		self->X1 += tf; self->X2 += tf;
	    }
	} else {		 			//bixel0 seg x- y-
	    if (self->kx_xy < self->ky_xy) {
		tf = (self->kx_xy - self->ky_xy) * 0.5;
		self->Y1 -= tf; self->Y2 -= tf;
	    } else {
		tf = (self->ky_xy - self->kx_xy) * 0.5;
		self->X1 -= tf; self->X2 -= tf;
	    }
	}
}

static void __attribute__ ((noinline)) xy_min_max_korr(TPenALD *self) {
  float lw05, w, h;

	if ((self->alxg < 0.01) && (self->alyg < 0.01)) return;

	lw05 = self->line_width / 2;
	if (lw05 < 1) lw05 = 1;
	if (self->X1 < -(lw05 + 1)) {
	    int cl = (int)(-(self->X1 + lw05) * self->lxyg / self->lxg);
	    self->X1 = self->X1 + ((float)cl * self->kx_xy);
	    self->Y1 = self->Y1 + ((float)cl * self->ky_xy);
	}
	if (self->Y1 < -(lw05 + 1)) {
	    int cl = (int)(-(self->Y1 + lw05) * self->lxyg / self->lyg);
	    self->X1 = self->X1 + ((float)cl * self->kx_xy);
	    self->Y1 = self->Y1 + ((float)cl * self->ky_xy);
	}

	if (self->X2 < -(lw05 + 1)) {
	    int cl = (int)(-(self->X2 + lw05) * self->lxyg / self->lxg);
	    self->X2 = self->X2 + ((float)cl * self->kx_xy);
	    self->Y2 = self->Y2 + ((float)cl * self->ky_xy);
	}
	if (self->Y2 < -(lw05 + 1)) {
	    int cl = (int)(-(self->Y2 + lw05) * self->lxyg / self->lyg);
	    self->X2 = self->X2 + ((float)cl * self->kx_xy);
	    self->Y2 = self->Y2 + ((float)cl * self->ky_xy);
	}

	w = self->img->width;
	h = self->img->height;
	if (self->X1 > (w + lw05)) {
	    int cl = (int)((w + lw05 - self->X1) * self->lxyg / self->lxg);
	    self->X1 = self->X1 + ((float)cl * self->kx_xy);
	    self->Y1 = self->Y1 + ((float)cl * self->ky_xy);
	}
	if (self->Y1 > (h + lw05)) {
	    int cl = (int)((h + lw05 - self->Y1) * self->lxyg / self->lyg);
	    self->X1 = self->X1 + ((float)cl * self->kx_xy);
	    self->Y1 = self->Y1 + ((float)cl * self->ky_xy);
	}

	if (self->X2 > (w + lw05)) {
	    int cl = (int)((w + lw05 - self->X2) * self->lxyg / self->lxg);
	    self->X2 = self->X2 + ((float)cl * self->kx_xy);
	    self->Y2 = self->Y2 + ((float)cl * self->ky_xy);
	}
	if (self->Y2 > (h + lw05)) {
	    int cl = (int)((h + lw05 - self->Y2) * self->lxyg / self->lyg);
	    self->X2 = self->X2 + ((float)cl * self->kx_xy);
	    self->Y2 = self->Y2 + ((float)cl * self->ky_xy);
	}
}

static void __attribute__ ((noinline)) line_x(TPenALD *self, int X1, int X2, int Y, dword ak) { //X1,X2 24.8
  int l, k, t;

	if (!ak || (Y < 0) || (Y >= self->img->height)) return;
	if (X1 > X2) SWAP_INT(X1, X2);
	if (X1 < 0) X1 = 0;
	if (X2 >= (self->img->width << 8)) X2 = (self->img->width - 1) << 8;
	l = X2 - X1;
	if (l < 8) return;
	k = (self->bpl * Y) + (X1 >> 8);
	t = X1 & 0xff;
	if (t) {
	    t = 0x100 - t;
	    if (t > l) t = l;
	    l -= t;
	    t = (ak * t) >> 8;
	    self->bix_argb(self, k, t);
	    k++;
	}
	if (l > 0xff) {
	    t = l >> 8;
	    if (ak >= 0xff) {
		self->bixn_rgb(self, t, k);
	    } else {
		self->bixn_argb(self, t, k, ak);
	    }
	    k += t;
	    l &= 0xff;
	}
	if (l > 0) {
	    t = (ak * l) >> 8;
	    self->bix_argb(self, k, t);
	}
}

static void __attribute__ ((noinline)) fill_rect(TPenALD *self, int X1, int X2, int Y1, int Y2) { //X1,X2 24.8
  int iyc, l, t;
  dword A;

	if (Y1 > Y2) SWAP_INT(Y1, Y2);
	if (Y1 < 0) Y1 = 0;
	if (Y2 >= (self->img->height << 8)) Y2 = (self->img->height - 1) << 8;
	l = Y2 - Y1;
	if (l < 8) return;
	A = ((self->Color >> 24) & 0xff) + 1;

	iyc = (Y1 >> 8);
	t = Y1 & 0xff;
	if (t) {
	    t = 0x100 - t;
	    if (t > l) t = l;
	    l -= t;
	    t = (A * t) >> 8;
	    line_x(self, X1, X2, iyc, t);
	    iyc++;
	}
	while (l > 0xff) {
	    line_x(self, X1, X2, iyc, A);
	    l -= 0x100;
	    iyc++;
	}
	if (l > 7) {
	    t = (A * l) >> 8;
	    line_x(self, X1, X2, iyc, t);
	}
}

static void __attribute__ ((noinline)) line_x_w(TPenALD *self, int Y) { //X1,X2 24.8
  int X1, nX1, oX1, X2, l, k, bixlg, nbixlg, obixlg, bixQ, gNlg, aNlg, oNlg;
  int ti, ti2;
  int a1lg, o1lg, a1add, o1add, max1lg, gn1lg, go1lg;
  int a2lg, o2lg, a2add, o2add, max2lg;
  int x2np, x2op;
  dword mpx, maxX, td, lflag;

	if ((Y < 0) || (Y >= self->img->height)) return;
	lflag = self->lflag;
	if (!(lflag & (CORNER_X1|CORNER_X2))) goto NoCorner;
	if (!self->yc1) lflag &= ~CORNER_X1;
	if (!self->yc2) lflag &= ~CORNER_X2;
	if (!(lflag & (CORNER_X1|CORNER_X2))) goto NoCorner;

// printf("%s(%d): Y=%d fl=%04x X1=%.2f/%.2f X2=%.2f/%.2f yc1=%03x\n", FL, Y, lflag, self->nX1, self->oX1, self->nX2, self->oX2, self->yc1);
	o1lg = 0;
	o1add = 0;
	X1 = (int)(self->nX1 * 0x100);
	if (lflag & CORNER_X1) {
	    ti = (int)(self->oX1 * 0x100);
	    if (X1 > ti) X1 = ti;
//	    if (!(lflag & RI_X1_NEG)) X1 = (int)(self->oX1 * 0x100);
	    o1lg  = self->o1lg;
	    o1add = self->o1add;
	}
	o2lg = 0;
	o2add = 0;
	X2 = (int)(self->nX2 * 0x100);
	if (lflag & CORNER_X2) {
	    ti = (int)(self->oX2 * 0x100);
	    if (X2 < ti) X2 = ti;
	    o2lg  = self->o2lg;
	    o2add = self->o2add;
	}
// printf("%s(%d): X1=%03x X2=%03x\n", FL, X1, X2);

	if ((X1 > X2) || (X2 < 0))  return;
	maxX = self->img->width << 8;
	if (X1 > (int)maxX) return;

	a1lg  = self->a1lg;
	a1add = self->a1add;
	a2lg  = self->a2lg;
	a2add = self->a2add;

	max1lg = a1lg;
	if (max1lg < o1lg) max1lg = o1lg;
	max2lg = a2lg;
	if (max2lg < o2lg) max2lg = o2lg;

	if ((X1 < 0) && ((X1 + max1lg) <= 0)) {
	    X1 += max1lg;
	    a1lg = 0;
	    o1lg = 0;
	    max1lg = 0;
	    if ((X2 - max2lg) >= 0) X1 = 0;
	}
	if ((X2 > (int)maxX) && ((X2 - max2lg) > (int)maxX)) {
	    X2 -= max2lg;
	    a2lg = 0;
	    o2lg = 0;
	    max2lg = 0;
	    if ((X1 + max1lg) <= (int)maxX) X2 = (int)maxX;
	}
	l = X2 - X1;
	if (l < 8) return;

	gNlg = l - max1lg - max2lg;
	if (gNlg < 0) aNlg = 0;
	aNlg = l - a1lg - a2lg;
	if (aNlg < 0) {
 if (aNlg < -8) printf("%s(%d): ERROR: y=%3d lf=%03x l=%06x a1lg=%04x aNlg=%04x a2lg=%04x  yc=%03x/%03x\n", FL, Y, lflag, l, a1lg, aNlg, a2lg, self->yc1, self->yc2);
	   aNlg = 0;
	}
	oNlg = l - o1lg - o2lg;
	if (oNlg < 0) {
 if (oNlg < -8) printf("%s(%d): ERROR: y=%3d lf=%03x l=%06x o1lg=%04x oNlg=%04x o2lg=%04x  yc=%03x/%03x\n", FL, Y, lflag, l, o1lg, oNlg, o2lg, self->yc1, self->yc2);
	   oNlg = 0;
	}

	nX1 = X1;
	oX1 = X1;
	x2np = X2 - a2lg;
	x2op = X2 - o2lg;
	gn1lg = 0;
	go1lg = 0;

// printf("%s(%d): Y=%d fl=%04x X1=%05x(%5.2f/%5.2f) X2=%05x(%5.2f/%5.2f) yc1=%03x yc2=%03x\n", FL, Y, lflag, X1, self->nX1, self->oX1, X2, self->nX2, self->oX2, self->yc1, self->yc2);

	mpx = (dword)(X1 >> 8);
	k = (self->bpl * Y) + (X1 >> 8);
	bixlg = 0x100 - (X1 & 0xff);
loop0:
	bixQ = 0;
	if (bixlg > l) bixlg = l;
	nbixlg = bixlg;
	obixlg = bixlg;

	if (a1lg > 0) {
	    ti = a1lg;
	    if (ti > nbixlg) ti = nbixlg;
	    ti2 = (ti * a1add) >> 8;
	    if (gn1lg <= 0) {
		bixQ = ti2;
	    } else {
		bixQ = ((ti2 + ((gn1lg * a1add) >> 7)) * ti) >> 8;
	    }
// printf("%s(%d):a1  bixQ=%04x ti=%03x\n", FL, bixQ, ti);
	    nX1 += ti;
	    nbixlg -= ti;
	    a1lg -= ti;
	}
	if (nbixlg > 0) {
	    if (nX1 < x2np) {
		ti = x2np - nX1;
		if (ti > nbixlg) ti = nbixlg;
		if (lflag & CORNER_X1) {
		    bixQ += ti * (0x100 - self->yc1);
		} else {
		    bixQ += ti << 8;
		}
// printf("%s(%d):a1k bixQ=%04x ti=%03x\n", FL, bixQ, ti);
		nX1 += ti;
		nbixlg -= ti;
	    }
	    if ((nbixlg > 0) && (a2lg > 0)) {
		ti = a2lg;
		if (ti > nbixlg) ti = nbixlg;
		a2lg -= ti;
		ti2 = (ti * a2add) >> 8;
		if (a2lg <= 0) {
		    bixQ += ti2;
		} else {
		    bixQ += ((ti2 + ((a2lg * a2add) >> 7)) * ti) >> 8;
		}
// printf("%s(%d):a2  bixQ=%04x ti=%03x\n", FL, bixQ, ti);
	    }
	}

	if (lflag & CORNER_X1) {
	    if (o1lg > 0) {
		ti = o1lg;
		if (ti > obixlg) ti = obixlg;
		ti2 = (ti * o1add) >> 8;
		if (go1lg <= 0) {
		    bixQ += ti2;
		} else {
		    bixQ += ((ti2 + ((go1lg * o1add) >> 7)) * ti) >> 8;
		}
// printf("%s(%d):o1  bixQ=%04x ti=%03x\n", FL, bixQ, ti);
		oX1 += ti;
		obixlg -= ti;
		o1lg -= ti;
	    }
	    if (obixlg > 0) {
		if (oX1 < x2op) {
		    ti = x2op - oX1;
		    if (ti > obixlg) ti = obixlg;
		    bixQ += ti * self->yc1;
// printf("%s(%d):o1k bixQ=%04x ti=%03x\n", FL, bixQ, ti);
		    oX1 += ti;
		    obixlg -= ti;
		}
		if ((obixlg > 0) && (o2lg > 0)) {
		    ti = o2lg;
		    if (ti > obixlg) ti = obixlg;
		    o2lg -= ti;
		    ti2 = (ti * o2add) >> 8;
		    if (o2lg <= 0) {
			bixQ += ti2;
		    } else {
			bixQ += ((ti2 + ((o2lg * o2add) >> 7)) * ti) >> 8;
		    }
// printf("%s(%d):o2  bixQ=%04x ti=%03x\n", FL, bixQ, ti);
		}
	    }
	}

	if (mpx < maxX) {
	    if (bixQ > 0x10000) {
// printf("%s(%d): bixQ=%04x\n", FL, bixQ);
		if (bixQ > bixQ_ERROR) printf("%s(%d): ERROR: bixQ=%05x a1lg=%04x a2lg=%04x\n", FL, bixQ, a1lg, a2lg);
		bixQ = 0x10000;
	    }
// printf("%s(%d):>bixQ=%04x X1=%03x/%03x\n", FL, bixQ, nX1, oX1);
	    td = (self->A * bixQ) >> 16;
	    self->bix_argb(self, k, td);
	}

	k++;
	mpx++;
	l -= bixlg;
	if (l < 8) return;
	gn1lg += bixlg;
	go1lg += bixlg;

	bixlg = 0x100;
	if ((o1lg > 0) || (a1lg > 0)) goto loop0;
	if (!(lflag & CORNER_X2)) goto NoCorner_a1lg_E;
//	if (a1lg > 0) goto loop0;

	ti = (x2np >> 8) - mpx;
	ti2 = (x2op >> 8) - mpx;
	if (ti > ti2) ti = ti2;
// printf("%s(%d): ti=%d\n", FL, ti);
	if (ti > 0) {
	    if (self->A >= 0xff) {
		self->bixn_rgb(self, ti, k);
	    } else {
		self->bixn_argb(self, ti, k, self->A);
	    }
	    mpx += ti;
	    k += ti;
	    l -= ti << 8;
	}
	if (l < 8) return;

/*	if (((lflag & RI_X2_NEG) == 0) == ((lflag & RI_X2_oldNEG) == 0)) {
 printf("%s(%d): lflag=%04x\n", FL, lflag);
	    
	} */

	nX1 = mpx << 8;
	oX1 = nX1;
X2Corner_loopE:
	bixlg = 0x100;
	if (bixlg > l) bixlg = l;
	nbixlg = bixlg;
	obixlg = bixlg;
	bixQ = 0;

	if (nX1 < x2np) {
	    ti = x2np - nX1;
	    if (ti > nbixlg) ti = nbixlg;
	    bixQ += ti * (0x100 - self->yc2);
// printf("%s(%d):a1 bixQ=%04x ti=%03x yc2=%03x\n", FL, bixQ, ti, self->yc2);
	    nX1 += ti;
	    nbixlg -= ti;
	}
	if (nbixlg > 0) {
	    ti = a2lg;
	    if (ti > nbixlg) ti = nbixlg;
	    a2lg -= ti;
	    ti2 = (ti * a2add) >> 8;
	    if (a2lg > 0) {
		bixQ += ((((a2lg * a2add) >> 7) + ti2) * ti) >> 8;
	    } else {
		bixQ += ti2;
	    }
// printf("%s(%d):a2k bixQ=%04x ti=%03x a2lg=%03x\n", FL, bixQ, ti, a2lg);
	    nX1 += ti;
	}

	if (oX1 < x2op) {
	    ti = x2op - oX1;
	    if (ti > obixlg) ti = obixlg;
	    bixQ += ti * self->yc2;
// printf("%s(%d):o bixQ=%04x ti=%03x\n", FL, bixQ, ti);
	    oX1 += ti;
	    obixlg -= ti;
	}

	if (obixlg > 0) {
	    ti = o2lg;
	    if (ti > obixlg) ti = obixlg;
	    o2lg -= ti;
	    ti2 = (ti * o2add) >> 8;
	    if (o2lg > 0) {
		bixQ += ((((o2lg * o2add) >> 7) + ti2) * ti) >> 8;
	    } else {
		bixQ += ti2;
	    }
// printf("%s(%d):o bixQ=%04x ti=%03x o2lg=%03x\n", FL, bixQ, ti, o2lg);
	    oX1 += ti;
	}

	if (mpx < maxX) {
	    if (bixQ > 0x10000) {
// printf("%s(%d): bixQ=%04x\n", FL, bixQ);
		if (bixQ > bixQ_ERROR) printf("%s(%d): ERROR: bixQ=%05x a1lg=%04x a2lg=%04x\n", FL, bixQ, a1lg, a2lg);
		bixQ = 0x10000;
	    }
// printf("%s(%d):>bixQ=%04x X1=%03x/%03x\n", FL, bixQ, nX1, oX1);
	    td = (self->A * bixQ) >> 16;
	    self->bix_argb(self, k, td);
	}

	k++;
	mpx++;
	l -= bixlg;
	if (l > 8) goto X2Corner_loopE;
	return;

NoCorner:
// printf("%s(%d): Y=%d X1=%.2f/%.2f X2=%.2f/%.2f yc1=%03x\n", FL, Y, self->nX1, self->oX1, self->nX2, self->oX2, self->yc1);
	X1 = (int)(self->nX1 * 0x100);
	X2 = (int)(self->nX2 * 0x100);
	if ((X1 > X2) || (X2 < 0))  return;
	maxX = self->img->width << 8;
	if (X1 > (int)maxX) return;

	a1lg  = self->a1lg;
	a1add = self->a1add;
	a2lg  = self->a2lg;
	a2add = self->a2add;
	if ((X1 < 0) && ((X1 + a1lg) <= 0)) {
	    X1 += a1lg;
	    a1lg = 0;
	    if ((X2 - a2lg) >= 0) X1 = 0;
	}
	if ((X2 > (int)maxX) && ((X2 - a2lg) > (int)maxX)) {
	    X2 -= a2lg;
	    a2lg = 0;
	    if ((X1 + a1lg) <= (int)maxX) X2 = (int)maxX;
	}
	l = X2 - X1;
	if (l < 8) return;

	nX1 = X1;
	x2np = X2 - a2lg;
	gn1lg = 0;

	mpx = (dword)(X1 >> 8);
	k = (self->bpl * Y) + (X1 >> 8);
	bixlg = 0x100 - (X1 & 0xff);
NoCorner_loop0:
	bixQ = 0;
	if (bixlg > l) bixlg = l;
	nbixlg = bixlg;

	if (a1lg > 0) {
// printf("%s(%d): bixQ=%04x gn1lg=%03x a1add=%03x\n", FL, bixQ, gn1lg, a1add);
	    ti = a1lg;
	    if (ti > nbixlg) ti = nbixlg;
	    ti2 = (ti * a1add) >> 8;
	    if (gn1lg <= 0) {
		bixQ = ti2;
	    } else {
		bixQ = ((ti2 + ((gn1lg * a1add) >> 7)) * ti) >> 8;
	    }
// printf("%s(%d): bixQ=%04x ti=%03x a1lg=%03x a1add=%03x %04x\n", FL, bixQ, ti, a1lg, a1add, (ti * a1add) >> 8);
	    nX1 += ti;
	    nbixlg -= ti;
	    a1lg -= ti;
	}

	if (nbixlg > 0) {
	    if (nX1 < x2np) {
		ti = x2np - nX1;
		if (ti > nbixlg) ti = nbixlg;
		bixQ += ti << 8;
		nX1 += ti;
// printf("%s(%d): bixQ=%04x ti=%03x\n", FL, bixQ, ti);
		nbixlg -= ti;
	    }
	    if ((nbixlg > 0) && (a2lg > 0)) {
		ti = a2lg;
		if (ti > nbixlg) ti = nbixlg;
		a2lg -= ti;
		ti2 = (ti * a2add) >> 8;
		if (a2lg <= 0) {
		    bixQ += ti2;
		} else {
		    bixQ += ((ti2 + ((a2lg * a2add) >> 7)) * ti) >> 8;
		}
// printf("%s(%d): bixQ=%04x ti=%03x a2lg=%03x a2add=%03x %04x\n", FL, bixQ, ti, a2lg, a2add, (a2lg * a2add) >> 7);
	    }
	}

	if (mpx < maxX) {
// printf("%s(%d):>bixQ=%04x X1=%03x\n", FL, bixQ, X1);
	    if (bixQ > 0x10000) {
		if (bixQ > bixQ_ERROR) printf("%s(%d): ERROR: bixQ=%05x a1lg=%04x a2lg=%04x\n", FL, bixQ, a1lg, a2lg);
		bixQ = 0x10000;
	    }
	    td = (self->A * bixQ) >> 16;
	    self->bix_argb(self, k, td);
	}
	k++;
	mpx++;
	l -= bixlg;
	if (l < 8) return;
	gn1lg += bixlg;
// return;
	if (a1lg > 0) {
	    bixlg = 0x100;
	    goto NoCorner_loop0;
	}

NoCorner_a1lg_E:
	ti = (x2np >> 8) - mpx;
	if (ti > 0) {
	    if (self->A >= 0xff) {
		self->bixn_rgb(self, ti, k);
	    } else {
		self->bixn_argb(self, ti, k, self->A);
	    }
	    mpx += ti;
	    k += ti;
	    l -= ti << 8;
	}
// printf("%s(%d): mpx=%04x nX1=%04x l=%04x\n", FL, mpx, nX1, l);
	if (l < 8) return;
//	if (lflag & CORNER_X2) goto X2EndCorner;

	bixlg = 0x100;
	bixQ = 0;
	if (bixlg > l) bixlg = l;
	nbixlg = bixlg;

	nX1 = mpx << 8;
	if (nX1 < x2np) {
	    ti = x2np - nX1;
	    if (ti > nbixlg) ti = nbixlg;
	    bixQ += ti << 8;
// printf("%s(%d): bixQ=%04x ti=%03x\n", FL, bixQ, ti);
	    nbixlg -= ti;
	}

	if ((nbixlg > 0) && (a2lg > 0)) {
	    ti = a2lg;
	    if (ti > nbixlg) ti = nbixlg;
	    a2lg -= ti;
// printf("%s(%d): bixQ=%04x ti=%03x a2lg=%03x a2add=%03x %04x\n", FL, bixQ, ti, a2lg, a2add, (a2lg * a2add) >> 7);
	    ti2 = (ti * a2add) >> 8;
	    if (a2lg > 0) {
		bixQ += ((((a2lg * a2add) >> 7) + ti2) * ti) >> 8;
	    } else {
		bixQ += ti2;
	    }
	}
NoCorner_loopE:
	if (mpx < maxX) {
	    if (bixQ > 0x10000) {
		if (bixQ > bixQ_ERROR) printf("%s(%d): ERROR: bixQ=%05x a2add=%04x a2lg=%04x\n", FL, bixQ, a2add, a2lg);
		bixQ = 0x10000;
	    }
// printf("%s(%d):>bixQ=%04x A=%03x\n", FL, bixQ, self->A);
	    td = (self->A * bixQ) >> 16;
	    self->bix_argb(self, k, td);
	}
	k++;
	mpx++;
	if (a2lg >= 0x100) {
	    bixlg = 0x100;
	    a2lg -= 0x100;
	    bixQ = ((a2lg * a2add) >> 7) + a2add;
// printf("%s(%d): bixQ=%04x a2lg=%03x a2add=%03x %04x\n", FL, bixQ, a2lg, a2add, (a2lg * a2add) >> 7);
	    goto NoCorner_loopE;
	}
	if (a2lg >= 8) {
//	    ti = (a2lg * a2lg) >> 8;
	    bixQ = (a2lg * a2add) >> 8;
// printf("%s(%d): bixQ=%04x a2lg=%03x a2add=%03x %04x\n", FL, bixQ, a2lg, a2add, (a2lg * a2add) >> 7);
	    a2lg = 0;
	    goto NoCorner_loopE;
	}
}

static void next4p(TPenALD *self, int *a, int *b) {
	*a = *b;
	*b = *b + 1;
	if (*b >= self->max4p) *b = 0; 
}

static void prev4p(TPenALD *self, int *a, int *b) {
	*a = *b;
	*b = *b - 1;
	if (*b < 0) *b = self->max4p - 1;
}

static void Set_a1lg(TPenALD *self, float yh) {
  float t;
	self->o1lg = self->a1lg;
	if (self->aly1 < 0.01) {
	    t = self->alx1;
	} else {
	    t = self->alx1 * yh / self->aly1;
	    if (t > self->alx1) t = self->alx1;
	}
	self->a1lg = (int)((t * 0x100) + 0.5);
}

static void Set_a2lg(TPenALD *self, float yh) {
  float t;
	self->o2lg = self->a2lg;
	if (self->aly2 < 0.01) {
	    t = self->alx2;
	} else {
	    t = self->alx2 * yh / self->aly2;
	    if (t > self->alx2) t = self->alx2;
	}
	self->a2lg = (int)((t * 0x100) + 0.5);
}

static void Setlxy1(TPenALD *self, int idx1, int idx2) {
  float t;

	self->lflag &= ~(RI_X1_oldNEG|RI_Y1_oldNEG|W_X1_old_ge45);
	self->lflag |= (self->lflag & (RI_X1_NEG|RI_Y1_NEG|W_X1_ge45)) << 1;
	self->lflag &= ~(RI_X1_NEG|RI_Y1_NEG|W_X1_ge45);

	self->lx1 = self->fp4[idx1].x - self->fp4[idx2].x;
	self->alx1 = self->lx1;
	if (self->alx1 < 0) {
	    self->alx1 = -self->alx1;
	    self->lflag |= RI_X1_NEG;
	}
	self->ly1 = self->fp4[idx1].y - self->fp4[idx2].y;
	self->aly1 = self->ly1;
	if (self->aly1 < 0) {
	    self->aly1 = -self->aly1;
	    self->lflag |= RI_Y1_NEG;
	}
	self->kx1 = self->lx1 / self->ly1;

	self->o1add = self->a1add;
	if (self->alx1 > self->aly1) {
	    t = self->aly1 * 0.5 / self->alx1;
	} else {
	    self->lflag |= W_X1_ge45;
	    t = 0.5;
	}
	self->a1add = (int)((t * 0x10000) + 0.5);
}

static void Setlxy2(TPenALD *self, int idx1, int idx2) {
  float t;

	self->lflag &= ~(RI_X2_oldNEG|RI_Y2_oldNEG|W_X2_old_ge45);
	self->lflag |= (self->lflag & (RI_X2_NEG|RI_Y2_NEG|W_X2_ge45)) << 1;
	self->lflag &= ~(RI_X2_NEG|RI_Y2_NEG|W_X2_ge45);

	self->lx2 = self->fp4[idx1].x - self->fp4[idx2].x;
	self->alx2 = self->lx2;
	if (self->alx2 < 0) {
	    self->alx2 = -self->alx2;
	    self->lflag |= RI_X2_NEG;
	}
	self->ly2 = self->fp4[idx1].y - self->fp4[idx2].y;
	self->aly2 = self->ly2;
	if (self->aly2 < 0) {
	    self->aly2 = -self->aly2;
	    self->lflag |= RI_Y2_NEG;
	}
	self->kx2 = self->lx2 / self->ly2;

	self->o2add = self->a2add;
	if (self->alx2 > self->aly2) {
	    t = self->aly2 * 0.5 / self->alx2;
	} else {
	    self->lflag |= W_X2_ge45;
	    t = 0.5;
	}
	self->a2add = (int)((t * 0x10000) + 0.5);
}



static void fill_rect4p(TPenALD *self) {
  float X1, Y1, X2, yf, yfp1, lyg, ft1, ft2, maxy, Y1e, Y2e;
  int i, i1a, i1b, i2a, i2b, it1, y;
  dword td;
  bool upd1lg, upd2lg, segend;
  
// az1flag = false;
	maxy = -99999;
	ft2 = 99999;
	i1a = -1;
	for (i=0;i<self->max4p;i++) {
	    if (maxy < self->fp4[i].y) maxy = self->fp4[i].y;
	    if (ft2 > self->fp4[i].y) {
		ft2 = self->fp4[i].y;
		i1a = i;
	    }
	}
	lyg = maxy - ft2;

	i2a = i1a;
	i1b = i1a - 1;
	if (i1b < 0) i1b = self->max4p - 1;
	i2b = i1a + 1;
	if (i2b >= self->max4p) i2b = 0;

	X1 = self->fp4[i1a].x;
	self->nX1 = X1;
	Y1 = self->fp4[i1a].y;
	X2 = self->fp4[i2a].x;
	self->nX2 = X2;
//	Y2 = self->fp4[i2a].y;

	y = (int)self->fp4[i1a].y;
	yf = (float)y;
	yfp1 = yf + 1;

	self->lflag = LINE_TOP | CORNER;

	Setlxy1(self, i1b, i1a);
	Setlxy2(self, i2b, i2a);

	self->o1lg = 0;
	self->o2lg = 0;
// printf("%s(%d):  a1lg=%04x/%04x  a2lg=%04x/%04x\n", FL, self->a1lg, self->o1lg, self->a2lg, self->o2lg);

	segend = false;
	upd1lg = true;
	upd2lg = true;
	Y1e = self->fp4[i1b].y;
	Y2e = self->fp4[i2b].y;

	self->yc1 = ((int)(self->fp4[i1a].y * 0x100)) & 0xff;

	it1 = ((int)(Y1 * 0x1000)) & 0xfff;
	if (!it1) {
	    ft1 = 1;
	    ft2 = 1;
	} else {
	    lyg += Y1 - yf;
	    ft1 = 1 - Y1 + yf;
	    self->oX1 = X1;
	    X1 += self->kx1 * ft1;
	    if (self->lflag & RI_X1_NEG) {
		if (X1 < self->fp4[i1b].x) X1 = self->fp4[i1b].x;
	    } else {
		if (X1 > self->fp4[i1b].x) X1 = self->fp4[i1b].x;
	    }
	    self->nX1 = X1;
	    self->oX2 = X2;
	    X2 += self->kx2 * ft1;
	    if (self->lflag & RI_X2_NEG) {
		if (X2 < self->fp4[i2b].x) X2 = self->fp4[i2b].x;
	    } else {
		if (X2 > self->fp4[i2b].x) X2 = self->fp4[i2b].x;
	    }
	    self->nX2 = X2;
	    Set_a1lg(self, ft1);

	    Set_a2lg(self, ft1);
	    ft1 = 1;
	    ft2 = 1;
	    goto StartL0;
	}

	while (lyg >= 0.1) {
	    self->oX1 = X1;
	    X1 += self->kx1 * ft1;
	    if (self->lflag & RI_X1_NEG) {
		if (X1 < self->fp4[i1b].x) X1 = self->fp4[i1b].x;
	    } else {
		if (X1 > self->fp4[i1b].x) X1 = self->fp4[i1b].x;
	    }
	    self->nX1 = X1;
	    self->oX2 = X2;
	    X2 += self->kx2 * ft2;
	    if (self->lflag & RI_X2_NEG) {
		if (X2 < self->fp4[i2b].x) X2 = self->fp4[i2b].x;
	    } else {
		if (X2 > self->fp4[i2b].x) X2 = self->fp4[i2b].x;
	    }
	    self->nX2 = X2;
	    ft1 = 1;
	    ft2 = 1;
	    if (lyg < 1) {
		self->lflag |= CORNER|LINE_BOTTOM;
		Y1e -= 0.001;
		Y2e -= 0.001;
		ft1 = lyg;
		ft2 = lyg;
		upd1lg = true;
		upd2lg = true;
	    }
	    if (upd1lg) {
		upd1lg = false;
		Set_a1lg(self, ft1);
	    }
	    if (upd2lg) {
		upd2lg = false;
	 	Set_a2lg(self, ft2);
	    }
StartL0:
	    if ((yfp1 > Y1e) && !segend) {
		self->lflag |= (CORNER|CORNER_X1);
		upd1lg = true;
		self->yc1 = ((int)(Y1e * 0x100)) & 0xff;
		if (self->yc1) {
		    Set_a1lg(self, Y1e - yf);
		    ft1 = 1 - Y1e + yf;
		}

		prev4p(self, &i1a, &i1b);
		segend = i1b == i2b;
		Y1e = self->fp4[i1b].y;
		X1 = self->fp4[i1a].x;
		Setlxy1(self, i1b, i1a);
		Set_a1lg(self, ft1);

		td = self->lflag ^ (self->lflag >> 1);
		if (!(td & (RI_X1_NEG|RI_Y1_NEG))) {
		    if (!(self->lflag & RI_X1_NEG)) {
			self->oX1 += self->kx1 * (1 - ft1);
		    } else {
			X1 += self->kx1 * ft1;
			if (self->lflag & RI_X1_NEG) {
			    if (X1 < self->fp4[i1b].x) X1 = self->fp4[i1b].x;
			} else {
			    if (X1 > self->fp4[i1b].x) X1 = self->fp4[i1b].x;
			}
			self->nX1 = X1;
			ft1 = 1;
		    }
		}
	    }
	    if ((yfp1 > Y2e) && !segend) {
		self->lflag |= (CORNER|CORNER_X2);

		upd2lg = true;
		self->yc2 = ((int)(Y2e * 0x100)) & 0xff;
		if (self->yc2) {
		    Set_a2lg(self, Y2e - yf);
		    ft2 = 1 - Y2e + yf;
		}

		next4p(self, &i2a, &i2b);
		segend = i1b == i2b;
		Y2e = self->fp4[i2b].y;
		X2 = self->fp4[i2a].x;
		Setlxy2(self, i2b, i2a);
		Set_a2lg(self, ft2);

		td = self->lflag ^ (self->lflag >> 1);
		if (!(td & (RI_X2_NEG|RI_Y2_NEG))) {
		    if (self->lflag & RI_X2_NEG) {
			self->oX2 += self->kx2 * (1 - ft2);
		    } else {
			X2 += self->kx2 * ft2;
			if (self->lflag & RI_X2_NEG) {
			    if (X2 < self->fp4[i2b].x) X2 = self->fp4[i2b].x;
			} else {
			    if (X2 > self->fp4[i2b].x) X2 = self->fp4[i2b].x;
			}
			self->nX2 = X2;
		    }
		}
	    }

	    line_x_w(self, y);

	    self->lflag &= ~(LINE_TOP|CORNER|CORNER_X1|CORNER_X2);
	    y++;
	    yf = yfp1;
	    yfp1 += 1.0;
	    if (yfp1 > maxy) yfp1 = maxy;
	    lyg -= 1;
	    Y1 += 1;
	}
}

static void setup_v(TPenALD *self) {
	if ((self->img->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
	    self->bpp = 2;
	    self->bpl = self->img->bytes_per_line >> 1;
	} else {
	    self->bpp = 4;
	    self->bpl = self->img->bytes_per_line >> 2;
	}
	self->pb = self->img->data;
	self->A = ((self->Color >> 24) & 0xff) + 1;
}

static void line_to(TPenALD *self, float x, float y) {
  dword A;

	if ((self->x == x) && (self->y == y)) return;
	if (!self->lineN_to || (self->line_width < 0.004)) goto end_ok;

	setup_v(self);
	self->max4p = 4;

	self->X1 = self->x;
	self->X2 = x;
	if (((self->X1 < 0) && (self->X2 < 0)) ||
	    ((self->X1 >= self->img->width) && (self->X2 >= self->img->width))) goto end_ok;
	self->Y1 = self->y;
	self->Y2 = y;
	if (((self->Y1 < 0) && (self->Y2 < 0)) ||
	    ((self->Y1 >= self->img->height) && (self->Y2 >= self->img->height))) goto end_ok;

	self->lxg = self->X2 - self->X1;
	self->alxg = self->lxg;
	if (self->alxg < 0) self->alxg = -self->alxg;
	if (self->alxg < 0.004) {
	    self->X2 = self->X1;
	    self->lxg = 0;
	    self->alxg = 0;
	    if (self->Y1 < 0) self->Y1 = 0;
	    if (self->Y2 < 0) self->Y2 = 0;
	    if (self->Y1 >= self->img->height) self->Y1 = self->img->height - 1;
	    if (self->Y2 >= self->img->height) self->Y2 = self->img->height - 1;
	}

	self->lyg = y - self->y;
	self->alyg = self->lyg;
	if (self->alyg < 0) self->alyg = -self->alyg;
	if (self->alyg < 0.004) {
	    self->Y2 = self->Y1;
	    self->lyg = 0;
	    self->alyg = 0;
	    if (self->X1 < 0) self->X1 = 0;
	    if (self->X2 < 0) self->X2 = 0;
	    if (self->X1 >= self->img->width) self->X1 = self->img->width - 1;
	    if (self->X2 >= self->img->width) self->X2 = self->img->width - 1;
	}
	self->lxyg = sqrt((self->lxg * self->lxg) + (self->lyg * self->lyg));
	if (self->lxyg < 0.004) goto end_ok;

	self->rix = self->alxg >= self->alyg;
	self->kx_xy = self->lxg / self->lxyg;
	self->ky_xy = self->lyg / self->lxyg;

	xy_min_max_korr(self);

	self->lx = self->X2 - self->X1;
	self->alx = self->lx;
	if (self->alx < 0) self->alx = -self->alx;

	self->ly = self->Y2 - self->Y1;
	self->aly = self->ly;
	if (self->aly < 0) self->aly = -self->aly;

	self->lxy = sqrt((self->lx * self->lx) + (self->ly * self->ly));
	if (self->lxy < 0.004) goto end_ok;

	if ((self->alxg == 0) || (self->alyg == 0)) {
	    if (self->alxg > 0) {
		int ix1, ix2, iy, il;
		ix1 = (int)(self->X1 * 0x100);
		ix2 = (int)(self->X2 * 0x100);
		iy = (int)((self->Y1 - (self->line_width / 2)) * 0x100);
		il = (int)(self->line_width * 0x100);
		fill_rect(self, ix1, ix2, iy, iy + il);

	    } else if (self->alyg > 0) {
		int ix1, ix2, iy1, iy2;
		ix1 = (int)((self->X1 - (self->line_width / 2)) * 0x100);
		ix2 = ix1 + (int)(self->line_width * 0x100);
		iy1 = (int)(self->Y1 * 0x100);
		iy2 = (int)(self->Y2 * 0x100);
		fill_rect(self, ix1, ix2, iy1, iy2);
	    }
	    goto end_ok;

	} else if (self->line_width > 1.5) {
	    float ox, oy;
	    ox = self->line_width * 0.5 * self->ky_xy;
	    oy = self->line_width * 0.5 * self->kx_xy;
	    self->fp4[0].x = self->X1 + ox;
	    self->fp4[0].y = self->Y1 - oy;
	    self->fp4[3].x = self->X1 - ox;
	    self->fp4[3].y = self->Y1 + oy;
	    self->fp4[1].x = self->X2 + ox;
	    self->fp4[1].y = self->Y2 - oy;
	    self->fp4[2].x = self->X2 - ox;
	    self->fp4[2].y = self->Y2 + oy;
	    fill_rect4p(self);
	    goto end_ok;
	}

	pixel05korr(self);

	float ft0 = self->kx_xy;
	if (ft0 < 0) ft0 = -ft0;
	float kxy = ft0;
	ft0 = self->ky_xy;
	if (ft0 < 0) ft0 = -ft0;
	if (kxy < ft0) kxy = ft0;

	float tlw = self->line_width;

	int kk = (int)(tlw / kxy * 0xffff);
	float ksq = sqrt(sqrt(kxy)) * tlw;
//	float ksq = sqrt(kxy);
	if (ksq > 1) ksq = 1;

	int kkq = (int)(ksq * 0x1000);
	int kkmax = (int)(ksq * 0xffff);
	int skk = (kk - kkmax) >> 1;
	int v1, v2;

	v1 = 0xffff;
	v2 = 0;
	A = ((self->Color >> 24) & 0xff) + 1;

	for (int i=0;i<16;i++) {
	    byte *bvk = self->vk + i;
	    int vkk = kk;
	    int vt0;
	    vt0 = skk - v2;
	    if (vt0 < 0) vt0 = 0;
	    bvk[0] = (vt0 * A) >> 16;
	    vkk -= vt0;

	    vt0 = ((v1 * kkq) >> 12) + skk;
	    if (vt0 > kkmax) vt0 = kkmax;
	    bvk[VKSIZE] = (vt0 * A) >> 16;
	    vkk -= vt0;

	    vt0 = vkk;
	    if (vt0 > kkmax) vt0 = kkmax;
	    bvk[VKSIZE * 2] = (vt0 * A) >> 16;
	    vkk -= vt0;
	    if (vkk < 0) vkk = 0;
	    bvk[VKSIZE * 3] = (vkk * A) >> 16;

	    v1 -= 0x1111;
	    v2 += 0x1111; 
	}
	self->lineN_to(self, self->X1, self->Y1);
end_ok:
	self->x = x;
	self->y = y;
	self->rectpIdx = 0;
}

static void move_to(TPenALD *self, float x, float y) {
	self->x = x;
	self->y = y;
	self->rectpIdx = 0;
}

static void rect_point(TPenALD *self, float x, float y) {
	if (!self->rectpIdx) {
	    self->fp4[0].x = self->x;
	    self->fp4[0].y = self->y;
	    self->fp4[1].x = x;
	    self->fp4[1].y = y;
	    self->rectpIdx = 2;
	} else {
	    if (self->rectpIdx >= MAX_RECT_POINT) {
		 printf("%s(%d): ERROR: max rect point =%d\n", FL, MAX_RECT_POINT);
	    } else {
		self->fp4[self->rectpIdx].x = x;
		self->fp4[self->rectpIdx].y = y;
		self->rectpIdx++;
	    }
	}
}

static void set_line_width(TPenALD *self, float width) {
	self->line_width = width;
}

static void set_color(TPenALD *self, dword Color) {
	self->Color = Color;
	if (self->rgb565) {
	    self->Color565 = ((self->Color >> 8) & 0xf800) | 
			     ((self->Color >> 5) & 0x07e0) | 
			     ((self->Color >> 3) & 0x001f);
	}
}

static void pen_flush(TPenALD *self) {
	if (self->rectpIdx > 2) {
	    setup_v(self);
	    self->max4p = self->rectpIdx;
	    fill_rect4p(self);
	    self->rectpIdx = 0;
	}
}

/*_____________________________ Destroy _________________________________*/
/* static bool TPen__Destroy(TPen *self) {
	return false;
} */

/*_____________________________ Create __________________________________*/
void TImagePen_Create(TImage *img) {
  TPenALD *self;

	if (!img) return;
	TPen_Create(img, sizeof(TPenALD), PEN_ART_ALD);
	self = (TPenALD*)img->Pen;
	if (!self) return;
	self->set_color = set_color;
	self->set_line_width = set_line_width;
	self->move_to = move_to;
	self->line_to = line_to;
	self->rect_point = rect_point;
	self->flush = pen_flush;

	if ((img->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
#if OPT_SIZE
	    self->lineN_to = line32_to;
#else
	    self->lineN_to = line16_to;
#endif
	    self->bix_rgb   = bix_rgb16;
	    self->bixn_rgb  = bixn_rgb16;
	    self->bix_argb  = bix_argb16;
	    self->bixn_argb = bixn_argb16;
	    self->Color565 = ((self->Color >> 8) & 0xf800) | ((self->Color >> 5) & 0x7e0) | ((self->Color >> 3) & 0x1f);
	    self->rgb565 = true;
	} else {
	    self->lineN_to = line32_to;
	    self->bix_rgb   = bix_rgb32;
	    self->bixn_rgb  = bixn_rgb32;
	    self->bix_argb  = bix_argb32;
	    self->bixn_argb = bixn_argb32;
	    self->rgb565 = false;
	}

}
