/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"
#include "pen_pv.h"

void TImagePen_Line(TImage *img, float x1, float y1, float x2, float y2) {
	if (img && img->Pen && img->Pen->line_to && img->Pen->move_to) {
	    img->Pen->move_to(img->Pen, x1, y1);
	    img->Pen->line_to(img->Pen, x2, y2);
	}
}

void TImagePen_LineTo(TImage *img, float x, float y) {
	if (img && img->Pen && img->Pen->line_to) {
	    img->Pen->line_to(img->Pen, x, y);
	}
}

void TImagePen_MoveTo(TImage *img, float x, float y) {
	if (img && img->Pen && img->Pen->move_to) {
	    img->Pen->move_to(img->Pen, x, y);
	}
}

void TImagePen_RectPoint(TImage *img, float x, float y) {
	if (img && img->Pen && img->Pen->rect_point) {
	    img->Pen->rect_point(img->Pen, x, y);
	}
}

void TImagePen_LineWidth(TImage *img, float width) {
	if (img && img->Pen && img->Pen->set_line_width) {
	    img->Pen->set_line_width(img->Pen, width);
	}
}

void TImagePen_Color(TImage *img, dword Color) {
	if (img && img->Pen && img->Pen->set_color) {
	    img->Pen->set_color(img->Pen, Color);
	}
}

void TImagePen_Flush(TImage *img) {
	if (img && img->Pen && img->Pen->flush) {
	    img->Pen->flush(img->Pen);
	}
}

/*_____________________________ Destroy _________________________________*/
static bool TPen__Destroy(TPen *self) {
	if (self->img) self->img->Pen = NULL;
	return false;
}

void TImagePen_Free(TImage *img) {
	if (img) ppTfree(&img->Pen);
}
/*_____________________________ Create __________________________________*/
void TPen_Create(TImage *img, int size, int art) {
  TPen *self;
  dword bpp;

	if (!img) return;
	if (img->Pen) ppTfree(&img->Pen);
	bpp = img->pxm_LD & pxmLD_bppMask;
	if ((bpp != pxm_2b_16bpp) &&
	    (bpp != pxm_4b_24bpp) && 
	    (bpp != pxm_4b_24bpp8a)) return;

	self = ttc_create(size, wtyp_Timg_TPen);
	self->destroy = (TDESTROY)TPen__Destroy;
	self->art = art;
	self->line_width = 1;
	self->img = img;
	img->Pen = self;
}
