/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

static byte sumTab[] = {
   0,1,1,2, 1,2,2,3, 1,2,2,3, 2,3,3,4, 
   1,2,2,3, 2,3,3,4, 2,3,3,4, 3,4,4,5, 
   1,2,2,3, 2,3,3,4, 2,3,3,4, 3,4,4,5, 
   2,3,3,4, 3,4,4,5, 3,4,4,5, 4,5,5,6, 

   1,2,2,3, 2,3,3,4, 2,3,3,4, 3,4,4,5, 
   2,3,3,4, 3,4,4,5, 3,4,4,5, 4,5,5,6, 
   2,3,3,4, 3,4,4,5, 3,4,4,5, 4,5,5,6, 
   3,4,4,5, 4,5,5,6, 4,5,5,6, 5,6,6,7, 

   1,2,2,3, 2,3,3,4, 2,3,3,4, 3,4,4,5, 
   2,3,3,4, 3,4,4,5, 3,4,4,5, 4,5,5,6, 
   2,3,3,4, 3,4,4,5, 3,4,4,5, 4,5,5,6, 
   3,4,4,5, 4,5,5,6, 4,5,5,6, 5,6,6,7, 

   2,3,3,4, 3,4,4,5, 3,4,4,5, 4,5,5,6, 
   3,4,4,5, 4,5,5,6, 4,5,5,6, 5,6,6,7, 
   3,4,4,5, 4,5,5,6, 4,5,5,6, 5,6,6,7, 
   4,5,5,6, 5,6,6,7, 5,6,6,7, 6,7,7,8
};

static void cmap_mem(TImage *self, int arr_size) {
	self->cmap->ColorCount = 0;
	if (self->cmap->arr_size == arr_size) {
	    if (arr_size > 0) Fill0(self->cmap->bit_arr, arr_size);
	} else {
	    ppfree(&self->cmap->bit_arr);
	    self->cmap->arr_size = arr_size;
	    if (arr_size > 0) self->cmap->bit_arr = xcalloc(arr_size);
	}
}

static void __attribute__ ((noinline)) cmap_sum(TImage *self) {
  byte *c = self->cmap->bit_arr;
  int sum = 0;
	for (int i=self->cmap->arr_size>>2;i;i--) {
	    if (*((dword*)c)) {
	        sum += sumTab[c[0]];
	        sum += sumTab[c[1]];
	        sum += sumTab[c[2]];
	        sum += sumTab[c[3]];
	    }
	    c += 4;
	}
	self->cmap->ColorCount = sum;
}

static void Set16_ColorCount(TImage *self) {
  //size = 32*64*32/8 = 8192;
  byte *bit_arr;
  word *pws;
  int eoff, h, k;
  dword v;

	cmap_mem(self, 8192);
	eoff = (self->bytes_per_line >> 1) - self->width;
	bit_arr = self->cmap->bit_arr;
	pws = (word*)self->data;
	for (h=self->height;h;h--) {
	    for (int i=self->width;i;i--) {
		v = *pws;
		pws++;
		k = 1 << (v & 0x1f);
		((dword*)bit_arr)[v >> 5] |= k;//1 << (v & 0x1f);
	    }
	    pws += eoff;
	}
	return;
}

static void __attribute__ ((noinline)) Set24_ColorCount(TImage *self) {
  byte *bit_arr;
  int eoff, h, fu;
  dword *pws, v, k;
  int size;
  int MaskAkt;
	fu = self->cmap->pxgcc & 0x07;
	size = 2097152;
	if (fu == pxgcc_RGB_565) size = 8192; else
	if (fu == pxgcc_RGB_666) size = 32768; /* else
	if (fu == pxgcc_RGB_777) size = 262144; */
	cmap_mem(self, size);



	MaskAkt = ((self->cmap->pxgcc & pxgcc_ColorCount_AddMask) != 0)
		  && alpha_akt(self);
	bit_arr = self->cmap->bit_arr;
	pws = (dword*)self->data;
	eoff = (self->bytes_per_line >> 2) - self->width;
	h = self->height;
	while (h) {
	  if (MaskAkt) {
	    if (fu == pxgcc_RGB_565) {
		for (int i=self->width;i;i--) {
		  if (*pws & 0x80000000) {
		    k = ((byte*)pws)[0]>>3;
		    v = k|((((byte*)pws)[1]&0xfc)<<3)|((((byte*)pws)[2]&0xf8)<<8);
		    ((dword*)bit_arr)[v >> 5] |= 1<<k;
		  }
		  pws++;
		}
	    } else if (fu == pxgcc_RGB_666) {
		for (int i=self->width;i;i--) {
		  if (*pws & 0x80000000) {
		    k = (((byte*)pws)[0]>>2);
		    v = k|((((byte*)pws)[1]&0xfc)<<4)|((((byte*)pws)[2]&0xfc)<<10);
		    ((dword*)bit_arr)[v >> 5] |= 1<<(k&0x1f);
		  }
		  pws++;
		}
	    } else { 
		for (int i=self->width;i;i--) {
		  if (*pws & 0x80000000) {
		    v = *pws & 0xffffff;
		    ((dword*)bit_arr)[v >> 5] |= 1<<(v & 0x1f);
		  }
		  pws++;
		}
	    }

	  } else {
	    if (fu == pxgcc_RGB_565) {
		for (int i=self->width;i;i--) {
		    k = ((byte*)pws)[0]>>3;
		    v = k|((((byte*)pws)[1]&0xfc)<<3)|((((byte*)pws)[2]&0xf8)<<8);
		    ((dword*)bit_arr)[v >> 5] |= 1<<k;
		    pws++;
		}
	    } else if (fu == pxgcc_RGB_666) {
		for (int i=self->width;i;i--) {
		    k = (((byte*)pws)[0]>>2);
		    v = k|((((byte*)pws)[1]&0xfc)<<4)|((((byte*)pws)[2]&0xfc)<<10);
		    ((dword*)bit_arr)[v >> 5] |= 1<<(k&0x1f);
		    pws++;
		}
	    } else { 
		for (int i=self->width;i;i--) {
		    v = *pws & 0xffffff;
		    ((dword*)bit_arr)[v >> 5] |= 1<<(v & 0x1f);
		    pws++;
		}
	    }
	  }
	  pws += eoff;
	  h--;
	}
	return;
}

int TImage_GetColorCount(TImage *self, int pxgcc) {
  int res;

	if (!self || !self->data || 
		(self->width <= 0) || (self->height <= 0)) return 0;

	if (!self->cmap) self->cmap = TColMap_Create();
	self->cmap->pxgcc = pxgcc;
	int k = self->pxm_LD & pxmLD_bppMask;
	if (k == pxm_2b_16bpp) {
	    Set16_ColorCount(self);
	} else if ((k == pxm_4b_24bpp) || (k == pxm_4b_24bpp8a)) {
	    Set24_ColorCount(self);
	}
	cmap_sum(self);
	res = self->cmap->ColorCount; 
	if (res > 0) {
	    self->memfu |= memfu_cmap_free;
	    if ((pxgcc & pxgcc_ColorCount_AddMask) && alpha_akt(self)) {
		self->cmap->ColorCount++;
		res = self->cmap->ColorCount;
	    }
	} else {
	    ppfree(&self->cmap);
	    self->memfu &= ~memfu_cmap_free;
	}
	return res;
}
