/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "appXbase.h"
#include "tpixmap.h"

#define NEXTL 1
/*____________________________________________________________________________*/

static void TPixMap_SetLineColor(TPixMap *pxm, int RGB) {
  XGCValues gcv;
  long gcm;

	if (!pxm->line_gc) {
	    pxm->line_gc = CreateGC_RGB(RGB);
	    pxm->LineColor = RGB;

	} else if (pxm->LineColor != RGB) {
	    bzero(&gcv, sizeof(XGCValues));
	    pxm->LineColor = RGB;
	    gcm = GCForeground;
	    if ((pxm->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
		gcv.foreground = RGB24To16(RGB);
	    } else { //if (pxm->depth == 24) {
		gcv.foreground = RGB;
	    }
	    ChangeGC(pxm->line_gc, gcm, &gcv);
	}
}

void TPixMap_DrawLinePixRGB(TPixMap *pxm, int RGB, int x1, int y1, int x2, int y2) {
//  int w = pxm->width;
//  int h = pxm->height;

//	if (((x1 < 0) && (x2 < 0)) || ((x1 >= w) && (x2 >= w)) ||
//	    ((y1 < 0) && (y2 < 0)) || ((y1 >= h) && (y2 >= h))) return;
	if (!TImage_DrawLineChk((TImage*)pxm, x1, y1, x2, y2)) return;

	TPixMap_SetLineColor(pxm, RGB);
	if (!(pxm->stat & pmxStat_NoBixImgSwitch)) TPixMap_XPutImage(pxm);
	pxm->stat &= ~pmxStat_ImgAkt;
	XDrawLine(appXDat.dpy, pxm->pmxID, pxm->line_gc, x1, y1, x2, y2);
}

#if NEXTL == 0
/*____________________________________________________________________________*/
static int _img_DrawLine(Timg_rsz *rsz) {
  int w = rsz->src_w;
  int h = rsz->src_h;
  int x1 = rsz->to_w;
  int y1 = rsz->to_h;
  int x2 = rsz->src_off_x;
  int y2 = rsz->src_off_y;
  int x, y;

	if (x1 == x2) {
	    if (y1 < 0) y1 = 0;
	    if (y2 < 0) y2 = 0;
	    if (y1 >= h) y1 = h-1;
	    if (y2 >= h) y2 = h-1;
	    if (y1 > y2) SWAP_INT(y1, y2);

	    if (rsz->src_typ == pxm_2b_16bpp) {
		word col = rsz->dest_mask_color;
		word *wx = &((word*)rsz->src_dat)[x1];
		w += rsz->src_line_eoff;
		wx += (w * y1);
		y = 1 + y2 - y1;
		while (y) {
		    *wx = col;
		    wx += w;
		    y--;
		}
	    } else if (rsz->src_typ == pxm_4b_24bpp) {
		dword col = rsz->dest_mask_color;
		dword *wx = &((dword*)rsz->src_dat)[x1];
		w += rsz->src_line_eoff;
		wx += (w * y1);
		y = 1 + y2 - y1;
		while (y) {
		    *wx = col;
		    wx += w;
		    y--;
		}
	    }
	    return 0;

	} else if (y1 == y2) {
	    if (x1 < 0) x1 = 0;
	    if (x2 < 0) x2 = 0;
	    if (x1 >= w) x1 = w-1;
	    if (x2 >= w) x2 = w-1;
	    if (x1 > x2) SWAP_INT(x1, x2);
	    if (rsz->src_typ == pxm_2b_16bpp) {
		word col = rsz->dest_mask_color;
		word *wx = &((word*)rsz->src_dat)[x1];
		w += rsz->src_line_eoff;
		wx += (w * y1);
		x = 1 + x2 - x1;
		while (x) {
		    *wx = col;
		    wx++;
		    x--;
		}
	    } else if (rsz->src_typ == pxm_4b_24bpp) {
		dword col = rsz->dest_mask_color;
		dword *wx = &((dword*)rsz->src_dat)[x1];
		w += rsz->src_line_eoff;
		wx += (w * y1);
		x = 1 + x2 - x1;
		while (x) {
		    *wx = col;
		    wx++;
		    x--;
		}
	    }
	    return 0;
	}
	    return 1;
}
#endif


int TPixMap_DrawLineImgRGB(TPixMap *pxm, dword RGB, int x1, int y1, int x2, int y2) {
  int res;

	if (!TImage_DrawLineChk((TImage*)pxm, x1, y1, x2, y2)) return 0;
	if (!(pxm->stat & pmxStat_NoBixImgSwitch)) TPixMap_XGetImage(pxm);
#if NEXTL
	res = TImage_DrawLineRGB((TImage*)pxm, RGB, x1, y1, x2, y2);
#else
  int w = pxm->width;
  int h = pxm->height;
  Timg_rsz rsz;
	TImage_SetBase_rsz((TImage*)pxm, &rsz);
	rsz.to_w = x1;
	rsz.to_h = y1;
	rsz.src_off_x = x2;
	rsz.src_off_y = y2;
	if ((pxm->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
	    rsz.dest_mask_color = RGB24To16(RGB);
	} else { //if (pxm->depth == 24) {
	    rsz.dest_mask_color = RGB;
	}
	res = _img_DrawLine(&rsz);
#endif
	if (res == 0) pxm->stat &= ~pmxStat_XPixAkt;
	return res;
}

/*____________________________________________________________________________*/
void TPixMap_DrawLineRGB(TPixMap *pxm, int RGB, int x1, int y1, int x2, int y2) {
  bool ok;

	if (!TImage_DrawLineChk((TImage*)pxm, x1, y1, x2, y2)) return;
	ok = (pxm->stat & pmxStat_ImgAkt) != 0;
#if NEXTL
	if (ok) {
	    if (!(pxm->stat & pmxStat_NoBixImgSwitch)) TPixMap_XGetImage(pxm);
	    ok = TImage_DrawLineRGB((TImage*)pxm, RGB, x1, y1, x2, y2) == 0;
	    if (ok) pxm->stat &= ~pmxStat_XPixAkt;
	}
	if (!ok) {
	    TPixMap_SetLineColor(pxm, RGB);
	    if (!(pxm->stat & pmxStat_NoBixImgSwitch)) TPixMap_XPutImage(pxm);
	    pxm->stat &= ~pmxStat_ImgAkt;
	    XDrawLine(appXDat.dpy, pxm->pmxID, pxm->line_gc, x1, y1, x2, y2);
	}
#else
	if (ok) ok = (TPixMap_DrawLineImgRGB(pxm, RGB, x1, y1, x2, y2) == 0);
	if (!ok) TPixMap_DrawLinePixRGB(pxm, RGB, x1, y1, x2, y2);
#endif
}
