/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include <errno.h>
#include <unistd.h>

Texec_pipe *Texec_pipe_Create() {
  Texec_pipe *expip;
  int i;

	expip = ttc_create(sizeof(Texec_pipe), wtyp_Texec_pipe);
	for (i=0;i<5;i++) expip->pp[i].fd = -1;
	return expip;
}

static void add_pp(Texec_pipe *expip, int fd, int nr) { 
  Tprocpipe *pp;
  Tfd_list *fd_list;
  int m;
	pp = &expip->pp[nr];
	if (pp->Proc) {
	    pp->fd = fd; //in_fd[0];
	    if (pp->Proc != (void*)1) {
		m = InPipe;
		if ((nr == 0)||(nr == 3)) m = OutPipe;
		fd_list = AddPipeProc_TmpRes(pp->fd, m, pp->Proc, expip);
		fd_list->m = nr;
	    }
	}
}

static void init_fork_pp(Texec_pipe *expip, int fd, int nr) { 
	if (expip->pp[nr].Proc) {close(nr); dup(fd);}
}

int exec_pipe_argv(char **argv, int fu, Texec_pipe *expip) {
  int pid;
  int i;
  char *argv0;
  xchar tmp;
  int x5_fd[10];
#ifdef exec_pipe_USE_VFORK
  int status, *pstatus;
#endif

	if (!argv) return 0;

	argv0 = argv[0];
#if 1
	tmp.s_const = GetFileName_pc(argv0);	//enum -Wcast-qual
	argv[0] = tmp.s;
#else
	argv[0] = (char*)GetFileName_pc(argv0);;
#endif
	for (i=0;i<10;i++) x5_fd[i] = -1;

	pid = 0;
	if (expip) {
	    for (i=0;i<5;i++) {
		if (expip->pp[i].Proc && pipe(&x5_fd[i<<1])) goto Ende; //Error
	    }
	}

#ifdef exec_pipe_USE_VFORK
	status = 0;
	pstatus = &status;
	pid = vfork();
#else
	pid = fork();
#endif
	if (pid == 0) {
	    if (expip) {
		if (expip->cd_path) {
		    if (chdir(expip->cd_path) < 0) {
#ifdef exec_pipe_USE_VFORK
			*pstatus = 2;
#endif
			fprintf(stderr, "ERROR: exec failed \"%s\" chdir=\"%s\" (\"%s\")\n", strerror(errno), expip->cd_path, argv[0]);
			_exit(EXIT_FAILURE);
		    }
		}

		ClosePipe(x5_fd[1]);
		ClosePipe(x5_fd[2]);
		ClosePipe(x5_fd[4]);
		ClosePipe(x5_fd[7]);
		ClosePipe(x5_fd[8]);
		init_fork_pp(expip, x5_fd[0], 0);
		init_fork_pp(expip, x5_fd[3], 1);
		init_fork_pp(expip, x5_fd[5], 2);
		init_fork_pp(expip, x5_fd[6], 3);
		init_fork_pp(expip, x5_fd[9], 4);
	    }
// signal (SIGPIPE, SIG_DFL);
	    execvp(argv0, argv);
#ifdef exec_pipe_USE_VFORK
	    *pstatus = 1;
#endif
	    fprintf(stderr, "ERROR: exec failed \"%s\" \"%s\"\n", strerror(errno), argv[0]);
	    _exit(EXIT_FAILURE);
	} 

#ifdef exec_pipe_USE_VFORK
	if (status) pid = 0;
#endif

Ende:
	if (pid > 0) {
	    pClosePipe(&x5_fd[0]); //out_fd
	    pClosePipe(&x5_fd[3]); //in_fd
	    pClosePipe(&x5_fd[5]); //err_fd
	    pClosePipe(&x5_fd[6]); //Dout_fd
	    pClosePipe(&x5_fd[9]); //Din_fd
	    if (expip) {
		if (expip->On_pid_close || (fu & expipfu_addPID_H)) {
		    expip->pida = Add_pid_Handle(pid, expip->On_pid_close);
		    expip->pida->obj = expip;
		}
		add_pp(expip, x5_fd[1], 0);
		add_pp(expip, x5_fd[2], 1);
		add_pp(expip, x5_fd[4], 2);
		add_pp(expip, x5_fd[7], 3);
		add_pp(expip, x5_fd[8], 4);
	    } else {
	        if (fu & expipfu_addPID_H) Add_pid_Handle(pid, NULL);
	    }
	} else {
	    int fd = x5_fd[4];
	    if ((fd > 0) && expip->pp[2].Proc) {
		add_pp(expip, fd, 2);
		Tfd_list *ftl = &appDat.fd_list[fd];
		if (ftl->Proc) {
		    if (fd_in_wait_sigle(fd, 10000)) ftl->Proc(fd, ftl);
		    ClosePipeProc(fd, 0);
		}
	    }
	    for (i=0;i<10;i++) ClosePipe(x5_fd[i]);
	}

	argv[0] = argv0;
	if (fu & expipfu_args_free) args_free(argv);

	return pid;
}
