/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#ifndef _OS_WIN32
#include <sys/wait.h>
#endif

#define MAX_APP_HZ 60
#define MIN_TIME 16667
#define MAX_TIME 500000


dword min_time = MIN_TIME;
dword max_time = MAX_TIME;

static dword gTimerEv_ms, gTimer_ms_old;
static int akt_us;
static TgTimer *gTimers;

static void Ev_gTimer(int nonsense) {
	appDat.gTimerNewEv++;
}

static void gTimer_SetUp() {
  float f;
  int us;
  int64_t k;
  TgTimer *gTimer;
  struct itimerval timeout;
  struct sigaction action;

	us = MAX_TIME;
	gTimer = gTimers;
	while (gTimer) {
	    f = (float)gTimer->OrgInterval * MAX_APP_HZ / 1000;
	    k = f + 0.3;
// printf("%s(%d): %lld %f\n", FL, k, f);

	    if (k < 1) k = 1;
	    k = k * 1000000 / MAX_APP_HZ;
	    gTimer->RunInterval = k / 1000;
// printf("%s(%d): %d\n", FL, gTimer->RunInterval);
	    if (k > MAX_TIME) k = MAX_TIME;
	    if (us > k) us = k;
	    gTimer = gTimer->next;
	}

	if (akt_us != us) {

// printf("%s(%d): set timer=%dus\n", FL, us);

//	    timeout.it_interval.tv_sec = 0;
//	    timeout.it_interval.tv_usec = us;
//	    timeout.it_value.tv_sec = 0;
//	    timeout.it_value.tv_usec = us;
	    set_timeval(&timeout.it_interval, us);
	    timeout.it_value.tv_sec  = timeout.it_interval.tv_sec;
	    timeout.it_value.tv_usec = timeout.it_interval.tv_usec;
	
	    if (!akt_us) {
		action.sa_handler = Ev_gTimer;
		sigemptyset(&action.sa_mask);
		action.sa_flags = SA_RESTART;
		signal(SIGALRM, Ev_gTimer);
	    }
	    akt_us = us;
	    setitimer(ITIMER_REAL, &timeout, NULL);
	}
}

static void gTimer_Dispatch() {
  int m, d; 
  dword gTimer_ms;
  TgTimer *gTimer;
  TgTimer *gTFree = NULL;
  bool LoopAkt;

	if (!appDat.gTimerNewEv) return;
	appDat.gTimerNewEv = 0;
	gTimer_ms = appDat.app_ms;
loop:
	d = gTimer_ms - gTimer_ms_old;
	if ((d > 3000) || (d < 0)) {
	    gTimer = gTimers;
	    while (gTimer) {
		gTimer->Tmp_ms = gTimer_ms + gTimer->RunInterval;
		gTimer = gTimer->next;
	    }
	}

	LoopAkt = False;
	gTimer_ms_old = gTimer_ms;
	gTimer = gTimers;
	while (gTimer) {

	    if (gTimer->func & Timer_Enabled) {
		m = gTimer_ms - (gTimer->Tmp_ms + gTimer->RunInterval);
		if (m >= 0) {
		    if (gTimer->func & Timer_Single) gTimer->func &= ~Timer_Enabled;
		    if (gTimer->func & Timer_NoLostOnTimer) {
			gTimer->Tmp_ms += gTimer->RunInterval;
			LoopAkt = True;
		    } else {
			gTimer->Tmp_ms = gTimer_ms;
		    }
		    if (gTimer->OnTimer && ((gTimer->intern_func & 0x01) == 0)) {
			gTimer->intern_func |= 0x01;
			gTimer->OnTimer(gTimer);
			gTimer->intern_func &= ~0x01;
		    }
		    if (gTimer->func & Timer_OnTimerAndFree) gTFree = gTimer;
		}
	    }
	    gTimer = gTimer->next;
	    if (gTFree) {
		Tfree(gTFree);
		gTFree = NULL;
	    }
	}
	if (LoopAkt) {
	    gTimer_ms = GetTime_ms();
	    goto loop;
	}
}

void TgTimer_NewInterval(TgTimer *self, dword Interval_ms) {
	if (!ttc_chk_is_wtyp(self, wtyp_TgTimer)) return;
	if (self->OrgInterval != Interval_ms) {
	    self->OrgInterval = Interval_ms;
	    gTimer_SetUp();
	}
}

void TgTimer_SetMaxTime(dword time_ms) {
	if (time_ms > MAX_TIME) time_ms = MAX_TIME;
	if (time_ms < MIN_TIME) time_ms = MIN_TIME;
	if (max_time != time_ms) {
	    max_time = time_ms;
	    gTimer_SetUp();
	}
}

void TgTimer_SetMinTime(dword time_ms) {
	if (time_ms > MAX_TIME) time_ms = MAX_TIME;
	if (time_ms < MIN_TIME) time_ms = MIN_TIME;
	if (min_time != time_ms) {
	    min_time = time_ms;
	    gTimer_SetUp();
	}
}

/*____________________________ Destroy __________________________________*/
static bool TgTimer__Destroy(TgTimer *self) {
	if (self->intern_func & 0x01) {
// Test timer del on AppHandleMessage()
	    self->func |= Timer_OnTimerAndFree;
	    return True;
	}
	TItem__Destroy((TItem*)self);
	gTimer_SetUp();
	return False;
}

/*____________________________ Create ___________________________________*/
void gTimer_Init() {
	appDat.TimerDispatch = gTimer_Dispatch;
	gTimerEv_ms = GetTime_ms();
	gTimer_ms_old = gTimerEv_ms;
}

TgTimer *gTimerX_Add(dword func, dword Interval_ms, void (*OnTimer)(struct TgTimer*)) {
  TgTimer *self;

	if (!appDat.TimerDispatch) gTimer_Init();
	self = TItemAdd((void*)&gTimers, sizeof(TgTimer), wtyp_TgTimer);
	self->destroy = (TDESTROY)TgTimer__Destroy;

	self->func = func;
	self->OrgInterval = Interval_ms;
	self->OnTimer = OnTimer;
	self->Tmp_ms = GetTime_ms();
	gTimer_SetUp();
	return self;
}
