/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include "tsearch.h"
#include <fcntl.h>

#ifdef _OS_WIN32

bool Find_Match(TSearch_Rec *sr, const char *s) {
  char *fname;
  bool res = False;

	if (((s[0] != '.') || s[1]) &&
	    (!sr->isRoot || !((s[0] == '.') && (s[1] == '.') && !s[2]))) {

	    fname = pcadd(sr->path, s); //sr->ret_entry->d_name);
	    sr->st = file_stat_fn(fname, sr->statfu);
	    xfree(fname);
	    res = File_stat_Match(sr->findAtr, sr->patt, s, sr->st);
	}
	return res;
}

static bool Set_Find_Dat(TSearch_Rec *sr) {
	if (sr->st) {xfree(sr->st); sr->st = NULL;}
	return Find_Match(sr, sr->fid.cFileName);
}



bool Find_First(TSearch_Rec *sr) {
  char *fname;

	sr->st = NULL;
	sr->isRoot = ((sr->path[0] == '\\') && (sr->path[1] == 0)) ||
		     ((sr->path[1] == ':') && (sr->path[2] == '\\') && (sr->path[3] == 0)) ||
		     (sr->findAtr & findIsRootPath);

	fname = pcadd(sr->path, "*");
	sr->dh = FindFirstFile(fname, &sr->fid);
	xfree(fname);
	if (!sr->dh) return False;

	while (!Set_Find_Dat(sr)) {
	    if (!FindNextFile(sr->dh, &sr->fid)) {
		Find_Close(sr);
		return False;
	    }
	}
	return True;
}

bool Find_Next(TSearch_Rec *sr) {
	while (FindNextFile(sr->dh, &sr->fid)) {
	    if (Set_Find_Dat(sr)) return True;
	}
	Find_Close(sr);
	return False;

}

void Find_Close(TSearch_Rec *sr) {
	if (sr->st) {xfree(sr->st); sr->st = NULL;}
	if (sr->dh) FindClose(sr->dh);
	sr->dh = 0;
}

#else /* ! _OS_WIN32 */

bool Find_First(TSearch_Rec *sr) {
  bool res = False;
	sr->st = NULL;
	sr->ret_entry = NULL;
	sr->isRoot = ((sr->path[0] == '/') && (sr->path[1] == 0)) || (sr->findAtr & findIsRootPath);
	sr->dh = opendir(sr->path);
	if (sr->dh) {
	    res = Find_Next(sr);
	} 
	return res;
}


bool Find_Match(TSearch_Rec *sr, const char *s) {
  char *fname;
  bool res = False;

	if (((s[0] != '.') || s[1]) &&
	    (!sr->isRoot || !((s[0] == '.') && (s[1] == '.') && !s[2]))) {
	    fname = pcadd(sr->path, s); //sr->ret_entry->d_name);
	    sr->st = file_stat_fn(fname, sr->statfu);
	    xfree(fname);
	    res = File_stat_Match(sr->findAtr, sr->patt, s, sr->st);
	}
	return res;
}

bool Find_Next(TSearch_Rec *sr) {
  bool res = False;

	while (!res) {
	    if (sr->st) {xfree(sr->st); sr->st = NULL;}
	    readdir64_r(sr->dh, &sr->entry, &sr->ret_entry);
	    if (!sr->ret_entry) break;
	    res = Find_Match(sr, sr->ret_entry->d_name);
	}
	if (!res) Find_Close(sr);
	return res;
}

void Find_Close(TSearch_Rec *sr) {
	if (sr->st) {xfree(sr->st); sr->st = NULL;}
	if (sr->dh) closedir(sr->dh);
	sr->dh = NULL;
}

/*_____________________________ Destroy _________________________________*/
static bool TSearch__Destroy(TSearch_Rec *self) {
	Tfree(self->patt);
	Find_Close(self);
	xfree(self->path);
	return False;
}

/*_____________________________ Create __________________________________*/
TSearch_Rec *TSearch_Create(const char *fn, int findAtr, int statfu) {
  TSearch_Rec *self;

	self = ttc_create(sizeof(TSearch_Rec), wtyp_TSearch_Rec);
	self->destroy = (TDESTROY)TSearch__Destroy;
	self->path = pcdup(fn);
	self->findAtr = findAtr; //zB.: findDIR | findIsRootPath
	self->statfu = statfu; //zB.: fsFollowLink
	return self;
}

#endif /* ! _OS_WIN32 */
