/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <errno.h>
#include <sys/stat.h>

bool file_mkdir_OS(const charOS *fn, TNetErrRet *ErrRet, int mode) {
#ifdef _OS_WIN32
	if (CreateDirectoryW(fn, NULL)) return True;
#else
	if (!mode) mode = 0755;
	if (!mkdir(fn, mode)) return True;
#endif
	if (ErrRet) {
	    ErrRet->sys_errno = errno;
	    ErrRet->bef = "mkdir";
#ifdef _OS_WIN32
	    ErrRet->fn = pwctoUTF8(fn);
#else
	    ErrRet->fn = pcdup(fn);
#endif
	}
	return False;
}

bool file_mkdir(const char *fn, TNetErrRet *ErrRet, int mode) {
  charOS *tmp;
  bool result;

	tmp = path_to_local_OS_err(fn, ErrRet);
	if (!tmp) return False;
	result = file_mkdir_OS(tmp, ErrRet, mode);
	xfree(tmp);
	return result;
}

bool file_mkdir_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode) {
  char *fn;
  bool result;
	fn = gptopc(gp, gptfu_del_slash|gptfu_noRootArt);
	result = file_mkdir(fn, ErrRet, mode);
	xfree(fn);
	return result;
}
