/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <unistd.h>
#include <errno.h>

#ifndef _OS_WIN32
#include <fcntl.h>
#endif

bool file_open_rd_OS(const charOS *fn, TNetErrRet *ErrRet, FileHandle *fd) {
#ifdef _OS_WIN32
	*fd = CreateFileW(fn, GENERIC_READ, FILE_SHARE_READ, NULL, 
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
#else
	*fd = open64(fn, O_RDONLY);
#endif
	if (ChkFileHandleOK(*fd)) return True;
	if (ErrRet) {
	    ErrRet->sys_errno = errno;
	    ErrRet->bef = "open read";
#ifdef _OS_WIN32
	    ErrRet->fn = pwctoUTF8(fn);
#else
	    ErrRet->fn = pcdup(fn);
#endif
	}
	return False;
}

bool file_open_rd(const char *fn, TNetErrRet *ErrRet, FileHandle *fd) {
  bool result;
  charOS *tmp;

	if (!fd) return False;
	tmp = path_to_local_OS_err(fn, ErrRet);
	if (!tmp) return False;
	result = file_open_rd_OS(tmp, ErrRet, fd);
	xfree(tmp);
	return result;
}

bool file_open_rd_gp(TglobPath *gp, TNetErrRet *ErrRet, FileHandle *fd) {
  char *fn;
  bool result;
	fn = gptopc(gp, gptfu_del_slash|gptfu_noRootArt);
	result = file_open_rd(fn, ErrRet, fd);
	xfree(fn);
	return result;
}
//------------------------ WR ---------------------------------------
bool file_open_wr_OS(const charOS *fn, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode) {
#ifdef _OS_WIN32
	*fd = CreateFileW(fn, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, 
			CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
#else
	if (!mode) mode = 0644;
	*fd = open64(fn, O_CREAT | O_TRUNC | O_WRONLY, mode);
#endif
	if (ChkFileHandleOK(*fd)) return True;
	if (ErrRet) {
	    ErrRet->sys_errno = errno;
	    ErrRet->bef = "open write";
#ifdef _OS_WIN32
	    ErrRet->fn = pwctoUTF8(fn);
#else
	    ErrRet->fn = pcdup(fn);
#endif
	}
	return False;
}

bool file_open_wr(const char *fn, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode) {
  bool result;
  charOS *tmp;

	if (!fd) return False;
	tmp = path_to_local_OS_err(fn, ErrRet);
	if (!tmp) return False;
	result = file_open_wr_OS(tmp, ErrRet, fd, mode);
	xfree(tmp);
	return result;
}

bool file_open_wr_gp(TglobPath *gp, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode) {
  char *fn;
  bool result;
	fn = gptopc(gp, gptfu_del_slash|gptfu_noRootArt);
	result = file_open_wr(fn, ErrRet, fd, mode);
	xfree(fn);
	return result;
}

//------------------------ Close ---------------------------------------
bool file_close(FileHandle fd, TNetErrRet *ErrRet) {
	if (!ChkFileHandleAkt(fd)) return True; 
#ifdef _OS_WIN32
	if (CloseHandle(fd)) return True;
#else
	if (!close(fd)) return True; 
#endif
	if (ErrRet) {
	    ErrRet->sys_errno = errno;
	    ErrRet->bef = "close";
	}
	return False;
}
