/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <unistd.h>
#include <errno.h>

#ifndef _OS_WIN32
#include <fcntl.h>
#endif

bool file_open_rdwr_OS(const charOS *fn, TNetErrRet *ErrRet, FileHandle *fd) {
#ifdef _OS_WIN32
	*fd = CreateFileW(fn, GENERIC_READ|GENERIC_WRITE, 0, NULL, 
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
#else
	*fd = open64(fn, O_RDWR);
#endif
	if (ChkFileHandleOK(*fd)) return True;
	if (ErrRet) {
	    ErrRet->sys_errno = errno;
	    ErrRet->bef = "open read";
#ifdef _OS_WIN32
	    ErrRet->fn = pwctoUTF8(fn);
#else
	    ErrRet->fn = pcdup(fn);
#endif
	}
	return False;
}

bool file_open_rdwr(const char *fn, TNetErrRet *ErrRet, FileHandle *fd) {
  bool result;
  charOS *tmp;

	if (!fd) return False;
	tmp = path_to_local_OS_err(fn, ErrRet);
	if (!tmp) return False;
	result = file_open_rdwr_OS(tmp, ErrRet, fd);
	xfree(tmp);
	return result;
}

bool file_open_rdwr_gp(TglobPath *gp, TNetErrRet *ErrRet, FileHandle *fd) {
  char *fn;
  bool result;
	fn = gptopc(gp, gptfu_del_slash|gptfu_noRootArt);
	result = file_open_rdwr(fn, ErrRet, fd);
	xfree(fn);
	return result;
}
