/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <utime.h>
#include <errno.h>

bool file_utime_OS(const charOS *fn, TNetErrRet *ErrRet, time_t atime, time_t mtime) {
  bool result;

#ifdef _OS_WIN32
	struct _utimbuf utb;
	utb.actime = atime;
	utb.modtime = mtime;
	result = _wutime(fn, &utb) >= 0;
#else
	struct utimbuf utb;
	utb.actime = atime;
	utb.modtime = mtime;
	result = utime(fn, &utb) >= 0;
#endif
	if (!result && ErrRet) ErrRet->sys_errno = errno;
	return result;
}

bool file_utime(const char *fn, TNetErrRet *ErrRet, time_t mtime) {
  charOS *tmp;
  bool result;

	tmp = path_to_local_OS_err(fn, ErrRet);
	if (!tmp) return False;
	result = file_utime_OS(tmp, ErrRet, mtime, mtime);
	xfree(tmp);
	return result;
}

bool file_utime_gp(TglobPath *gp, TNetErrRet *ErrRet, time_t mtime) {
  char *fn;
  bool result;
	fn = gptopc(gp, gptfu_del_slash|gptfu_noRootArt);
	result = file_utime(fn, ErrRet, mtime);
	xfree(fn);
	return result;
}
