/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "ftp_cache.h"
#include "tdirlist.h"
#include "../tdirlist/tdirlist_pv.h"
#include "net_io.h"
#include <errno.h>

typedef struct TDirList_private {
  int schr;	/* b0=Ev b1=open */
  dword open_us;
  TglobPath gp;
  TEv_Array *appEv;
} Tprivate;

/*_____________________________ net Catche Ev ___________________________*/
static void TFtpCacheEv(TDirListNotify *self, TNetDataCache *dc, int fu) {
  Tprivate *pv = self->private;
  TDirFileList *dfl;

	dfl = self->dl->dfl;
	while (dfl && !pccmp(dfl->fn, dc->name)) dfl = dfl->next;

	if (fu == ndcfu_ADD) {
	    if (dfl || !chk_file_ok(dc->st, pv->gp.path)) return;
	    dfl = TDirListNotify_Add_st(self, dc->st);
	    TDirListNotify_DodllEv(self, &dfl, dlbef_Add);

	} else if (fu == ndcfu_CHANGE) {
	    if (!dfl) return;
  Tstat64 *org = dfl->st;
  Tstat64 *new = dc->st;

	    dword chstat = chk_new_stat(org, new);
	    if (!chstat) return;
	    dfl->chstat = chstat;
	    dl_copy_new_stat(org, new);
	    TDirListNotify_DodllEv(self, &dfl, dlbef_Change);

	} else if (fu == ndcfu_DEL) {
	    if (!dfl) return;
	    TDirListNotify_Del(self, dfl);

	} else return;

	pv->schr = 0x03;
}

/*_____________________________ appEv ___________________________________*/
static void TDirList_ny_On_appEv(TEv_Array *appEv) {
  TDirListNotify *self = appEv->obj;
  Tprivate *pv = self->private;

	if (!cpTime(pv->open_us, appDat.app_us)) return;
	pv->open_us = appDat.app_us + 500000;

	if (pv->schr & 0x01) {
	    pv->schr &= ~0x01;
	    TDirListNotify_DodllEv(self, NULL, dlbef_Sync);

	} else if (pv->schr & 0x02) {
	    pv->schr = 0;
	    TDirListNotify_ScanEnd(self);
	}
}

/*_____________________________ ReScan __________________________________*/
static void TDirListNotify_Ftp_ReScan(TDirListNotify *self) {
  Tprivate *pv = self->private;
  TNetErrRet ErrRet;
  Tstat64 *st;

	TDirListNotify_DodllEv(self, NULL, dlbef_Start);
	net_err_init(&ErrRet);
	st = ftp_stat_gp(&pv->gp, &ErrRet, fsGetFileAll_UGT|fsForceReScan);
	if (!st) {

	    char *fn = get_DisplayFileStr(NULL, &pv->gp);
	    if (ErrRet.sys_errno) {
		errno = ErrRet.sys_errno;
		appError(EMSG_ERR1|EMSG_SYS, "  %s \n %s\n", "ReScan", fn);

	    } else if (ErrRet.err) {
		appError(EMSG_ERR1, "  %s \n %s\n %s\n", "ReScan", fn, ErrRet.err);

//	    } else {
//		appError(EMSG_ERR1, "  %s ??? %s\n", "ReScan", fn);
	    }
	    xfree(fn);
	}
	xfree(st);
	net_err_clr(&ErrRet);
	if (!pv->schr) TDirListNotify_DodllEv(self, NULL, dlbef_ListEnd);
}

/*_____________________________ Start ___________________________________*/
static void TDirListNotify_Ftp_Start(TDirListNotify *self) {
  Tprivate *pv = self->private;

	if (pv->appEv) return;
	pv->open_us = GetTime_us() + 500000;
	pv->appEv = Add_appEv(TDirList_ny_On_appEv);
	pv->appEv->obj = self;
	FtpCache_AddEv(&pv->gp, self, (TNetCacheEv)TFtpCacheEv);
}

/*_____________________________ Destroy _________________________________*/
static bool TDirListNotify_Ftp__Destroy(TDirListNotify *self) {
  Tprivate *pv = self->private;

	FtpCache_DelEv(&pv->gp, self);
	if (pv->appEv) Free_appEv(pv->appEv);
	gp_pc_free(&pv->gp);
	xfree(self->private);
	return False;
}

/*_____________________________ Create __________________________________*/
static TDirListNotify *TDirListNotify_Ftp_Create(TDirList *dl, TOnFileMsg OnFileMsg) {
  TDirListNotify *self;
  Tprivate *pv;

	if (!ftp_chk_RootArt_URL(&dl->gp)) return NULL;

	self = ttc_create(sizeof(TDirListNotify), wtyp_TDirListNotify);
	self->destroy = (TDESTROY)TDirListNotify_Ftp__Destroy;
	self->OnFileMsg = OnFileMsg;
	self->dl = dl;
	pv = xcalloc(sizeof(Tprivate));
	self->private = pv;
	gp_dup_gpdata_path(&pv->gp, &dl->gp);

	self->start  = TDirListNotify_Ftp_Start;
	self->ReScan = TDirListNotify_Ftp_ReScan;
	self->UseRescan = True;
	return self;
}

const ThDirListNotify hdirlist_ny_Ftp = {
  ftp_chk_RootArt_URL,
  TDirListNotify_Ftp_Create,
  0
};
