/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "ftp_net.h"


static void TNetFtpLineBuf_OnPipeWrite(int fd, Tfd_list *fdl) {
  int blg;
  TNetFtpLineBuf *self = fdl->obj;
  char *pc;

	TNetLineBuf_sync_wr((TNetLineBuf*)self);
	blg = TLineBuf_Get(self->lb, BUFSIZ - sizeof(TdData), &pc);
	self->shd->lg = sizeof(TdData);

	if (!self->lb->eof && (blg < (BUFSIZ - sizeof(TdData)))) {
	    appDat.AktPipes[fd] = OutPipe + 1;
	    return;
	}

	if (blg > 0) {
	    self->shd->lg += blg;
	    pnmove(self->shd->dat, pc, blg);
	    self->shd->sh_DataNowSize += blg;
	    self->shd->sh_LoadTime_us = GetCurTimeDiv_us64(self->ref_us);
	    blg = TLineBuf_Sub(self->lb, blg);
	}

	if (self->lb->eof && !blg) self->shd->fu = hSHfu_EOF;
//	FileWrite(fd, self->shd, self->shd->lg, NULL);
	file_write(fd, NULL, self->shd, self->shd->lg, NULL);
	if (self->shd->fu & hSHfu_EOF) ClosePipeProc(fd, 0);
}

static void TNetFtpLineBuf_OnFileMsg(TFileMsg *fmsg) {
  TFtpClientCMD *ftpcmd;
  TNetFtpLineBuf *self;

	if (!TNetFtp_fmsg_GetSelf(fmsg, &ftpcmd, (void*)&self)) return;

	if (TNetFtp_Chk_ftpappclient((TNetFtp*)self, fmsg)) {

	    if (fmsg->D.m == facSendPipe) {
		self->ref_us = GetTime_us64();
		self->iofd = fmsg->F.m;
		AddPipeProc(fmsg->F.m, OutPipe, TNetFtpLineBuf_OnPipeWrite, self);
		fmsg->res = 1;
	    }

	} else if ((fmsg->art == fmArt_TdServ) &&
		   (fmsg->typF == fmTyp_shd)) {

	    TdData *shd = fmsg->F.shd;
	    self->ioinfo.DataNowSize = shd->sh_DataNowSize;
	    self->ioinfo.Time_us = shd->sh_LoadTime_us;

	    self->FREE_LOCK++;
	    if (ftpcmd->stat & ftpcmdmk_Put) {
		if (shd->fu & hSHfu_EOF) self->ioinfo.stat |= fmIOInfo_EOF;
		TNet_send_ioinfo((TNet*)self);

	    } else {
		if (!self->error && ((shd->fu & hSHfu_onlyInfo) == 0)) {
		    int dlg = shd->lg - sizeof(TdData);
		    if (self->OnDataAdd) self->OnDataAdd((TNetLineBuf*)self, False);
		    TLineBuf_Add(self->lb, shd->dat, dlg);
		    if (self->OnDataAdd) self->OnDataAdd((TNetLineBuf*)self, True);
		    TNetLineBuf_sync_rd((TNetLineBuf*)self);
		    TNet_send_ioinfo((TNet*)self);
		}
	    }
	    self->FREE_LOCK--;
	    if (self->FREE_LOCK_Stat) Tfree(self);
	    fmsg->Ok = True;
	    fmsg->ev_Stop = True;

	} else {

	    TNet_msg_next((TNet*)self, fmsg);
	}
}

static bool TNetFtpLineBuf_wait(TNetFtpLineBuf *self) {
  bool result = TNetFtp_wait((TNetFtp*)self);
	if (result && (self->omod == NetOpenMod_Write)) {
	    Tstat64 *st = ftp_stat_gp(&self->gp, NULL, fsSingleReScan);
	    xfree(st);
	}
	return result;
}

static bool TNetFtpLineBuf_stop(TNetFtpLineBuf *self) {
	return TNetFtp_stop((TNetFtp*)self);
}

static bool TNetFtpLineBuf_open(TNetFtpLineBuf *self, int omod) {
  const char *bef;
  char *cmd = NULL;
  bool result;

	if (!self->lb) return False;
	self->omod = omod;

	if (omod == NetOpenMod_Read) {
	    TLineBuf_Clr(self->lb);
	    bef = "get";

	} else if (omod == NetOpenMod_Write) {
	    if (self->lb->wr_memlook) {
		self->lb->lg = self->lb->datalg;
		self->lb->pc_rdidx = NULL;
	    }
	    bef = "put";

	} else {
	    return False;
	}

	cmd = pcadd_n(4, "\"", self->gp.path, self->gp.name, "\"");
	result = TNetFtp_ftpac_add((TNetFtp*)self, bef, cmd, TNetFtpLineBuf_OnFileMsg);
	xfree(cmd);
	return result;
}

/*_____________________________ Destroy _________________________________*/
static bool TNetFtpLineBuf__Destroy(TNetFtpLineBuf *self) {
//	if (self->FREE_LOCK) {
//	    self->FREE_LOCK_Stat = 1;
//	    return True;
//	}
	ClosePipeProc(self->iofd, 0);
	return TNetFtp__Destroy((TNetFtp*)self);
}

/*_____________________________ Create __________________________________*/
static TNetLineBuf *TNetFtpLineBuf_Create(void *owner, TglobPath *gp) {
  TNetFtpLineBuf *self;

	self = TNetFtp_Create(sizeof(TNetFtpLineBuf), wtyp_TNetLineBuf, owner, gp);
	self->destroy = (TDESTROY)TNetFtpLineBuf__Destroy;
	self->open = (void*)TNetFtpLineBuf_open;
	self->wait = (void*)TNetFtpLineBuf_wait; //(void*)TNetFtp_wait;
	self->stop = (void*)TNetFtpLineBuf_stop; //(void*)TNetFtp_wait;
	self->shd = xcalloc(BUFSIZ);
	self->shd->HeadArt = hServData;
	self->shd->bef = hSHbd_Dat;
	return (TNetLineBuf*)self;
}


const ThNetLineBufLoader hNetLineBufLd_ftp = {
  ftp_chk_RootArt,
  .Create.lb = TNetFtpLineBuf_Create
};
