/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "ftp_net.h"

bool TNetFtp_OnFileMsg_ChkAlgErr(TNetFtp *self, TFileMsg *fmsg) {

	if (fmsg->art == fmArt_NewLogin) {
	    if ((fmsg->typR == fmTyp_gp) && (self->ErrRet)) {
		self->ErrRet->errFlag |= NetErrFlag_NewLogin;
		self->ErrRet->new_user = pcdup(fmsg->R.gp->user);
		self->ErrRet->new_passwd = pcdup(fmsg->R.gp->passwd);
	    }
	    return False;
	}

	if ((fmsg->art != fmArt_AlgErr) && 
	    (fmsg->art != fmArt_errConnect)) return False;

	self->retCode = fmsg->F.sh->m;

	if (self->ErrRet) {
	    self->ErrRet->err = pcdup(fmsg->D.s);
	    if (fmsg->art == fmArt_errConnect) 
			    self->ErrRet->errFlag |= NetErrFlag_NoDev;
	}
//					     "350 = Requested file action"
	if ((fmsg->art == fmArt_errConnect) || (self->retCode == 350)) {
	    fmsg->Ok = True;
	    fmsg->ev_Stop = True;
	}
	return True;
}

void TNetFtp_OnFileMsg(TFileMsg *fmsg) {
  TFtpClientCMD *ftpcmd;
  TNetFtp *self;

	if (!TNetFtp_fmsg_GetSelf(fmsg, &ftpcmd, (void*)&self)) return;

	if (TNetFtp_Chk_ftpappclient((TNetFtp*)self, fmsg)) {

	} else TNetFtp_OnFileMsg_ChkAlgErr(self, fmsg);
}

bool TNetFtp_fmsg_GetSelf(TFileMsg *fmsg, TFtpClientCMD **ftpcmd, TNetFtp **self) {
	if (!file_msg_chk_wtyp(fmsg, wtyp_TFtpClientCMD, FL)) return False;
	if (!fmsg->owner) return False;
	*ftpcmd = fmsg->obj;
	*self = fmsg->owner;
	return True;
}

bool TNetFtp_Chk_ftpappclient(TNetFtp *self, TFileMsg *fmsg) {
	if (fmsg->art != fmArt_ftpappclient) return False;

	if ((fmsg->D.m & facEndCode_MASK) == facEndCode_OK) {
	    self->ende = True;
	    fmsg->owner = NULL;
	    self->ftpcmd = NULL;

	} else if ((fmsg->D.m & facEndCode_MASK) == facCMD_Free) {  //or Error
	    self->ende = self->error = True;
	    fmsg->owner = NULL;
	    self->ftpcmd = NULL;
	}

	fmsg->Ok = True;
	fmsg->ev_Stop = True;
	return True;
}

bool TNetFtp_wait(TNetFtp *self) {
	if (!self->omod) return False;
	while (AppHandleMessage() && !self->ende);
	if (self->ftpcmd) {
	    self->ftpcmd->OnFileMsg = NULL;
	    self->ftpcmd->owner = NULL;
	    self->ftpcmd = NULL;
	}
	return self->ende && !self->error;
}

bool TNetFtp_stop(TNetFtp *self) {
	if (!self->omod || !self->ftpcmd) return False;
	return TFtpClientCMD_Stop(self->ftpcmd);
}

bool TNetFtp_ftpac_add_fcnx(TNetFtp *self, const char *bef, const char *cmd, 
			TOnFileMsg ev, dword akt_fkt, TFtpClientNextCMD *fcnx) {
  TFtpAppClient *ftpac;
  bool res = False;

	if (self->run) goto end_err;
	TNet_Reset((TNet*)self);

	ftpac = TFtpAppClient_get(&self->gp);
	if (!ftpac) goto end_err;
	self->ftpcmd = TFtpClientCMD_Create(bef, cmd, &self->gp);
	self->ftpcmd->owner = self;
	self->ftpcmd->stat = ftpcmdmk_Get;
	if (self->omod == NetOpenMod_Write) self->ftpcmd->stat = ftpcmdmk_Put;
	self->ftpcmd->OnFileMsg = ev;
	self->ftpcmd->fcnx = fcnx;
	self->ftpcmd->akt_fkt = akt_fkt;
	TFtpAppClient_CMD_add(ftpac, self->ftpcmd);
	res = True;
end_err:
	if (!res) self->omod = 0;
//	if (afu & afu_CMD_free) xfree(cmd);
//	if (afu & afu_BEF_free) xfree(bef);

	return res;
}

bool TNetFtp_ftpac_add(TNetFtp *self, const char *bef,const char *cmd, TOnFileMsg ev) {
	return TNetFtp_ftpac_add_fcnx(self, bef, cmd, ev, 0, NULL);
}

static bool TNetFtp_set_gp(TNetFtp *self, TglobPath *gp) {
	if (self->run || !ftp_chk_RootArt_URL(gp)) return False;
	gp_pc_free(&self->gp);
	gp_dup_gpdata(&self->gp, gp);
	return True;
}

/*_____________________________ Destroy _________________________________*/
bool TNetFtp__Destroy(TNetFtp *self) {

	if (self->ftpcmd) {
	    if (!self->ftpcmd->ende) {
 printf("%s(%d): TNetFtp__Destroy and RUN !!!\n", FL);
		TFtpClientCMD_Stop(self->ftpcmd);
	    }
	    self->ftpcmd->OnFileMsg = NULL;
	    self->ftpcmd->owner = NULL;
	    self->ftpcmd = NULL;
	}
	xfree(self->shd);
	return TNet__Destroy((TNet*)self);
}

/*_____________________________ Create __________________________________*/
void *TNetFtp_Create(int TNet_size, int wtyp, void *owner, TglobPath *gp) {
  TNetFtp *self;

	self = TNet_Create(TNet_size, wtyp, gp);
	self->destroy = (TDESTROY)TNetFtp__Destroy;
	self->owner = owner;
	self->set_gp = TNetFtp_set_gp;
	return self;
}
