/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "ftp_net.h"
#include <errno.h>

#define USE_CustomStream 1

#if USE_CustomStream
typedef struct {
  AlgTNetFtp
  TLineBuf *lb;
  _IO_off64_t idx;
  FILE *f;
  cookie_io_functions_t cookie_io; 
} TNetFtpStream;

static __ssize_t io_read_fn(TNetFtpStream *cookie, char *buf, size_t nbytes) {
  size_t lg;
  TLineBuf *lb;

	if (!buf || (nbytes <= 0)) return 0;
	lb = cookie->lb;
	if (lb->pc_rdidx) {
	     printf("%s(%d): io_read_fn  read write ???\n", FL);
	     return 0;
	}
loop:
	if (cookie->error) return 0;
	lg = lb->lg - cookie->idx;
	if (lg < 0) return 0;

	if (lg >= nbytes) {
	    lg = nbytes;

	} else if (!lb->eof && !cookie->ende) {
	    lg = 0;
	    if (AppHandleMessage()) goto loop;
	}

	if (lg > 0) {
	    pnmove(buf, lb->pc + cookie->idx, lg);
	    cookie->idx += lg;
	}
	return lg;
}

static __ssize_t io_write_fn(TNetFtpStream *cookie, const char *buf, size_t n) {
	if (!buf || (n <= 0)) return 0;
	cookie->idx += n;
	TLineBuf_Add(cookie->lb, buf, n);
	return n;
}

static int io_seek_fn(TNetFtpStream *cookie, _IO_off64_t *pos, int w) {
  TLineBuf *lb;
  _IO_off64_t idx, min_idx, max_dat_lg;
	lb = cookie->lb;
loop:
	if (cookie->error) return ESPIPE;
	idx = *pos;

	if (w == SEEK_CUR) {
	    idx += cookie->idx;

	} else if (w == SEEK_END) {
	    if (!lb->eof && !cookie->ende && AppHandleMessage()) goto loop;
	    idx += lb->datalg;
	};

	max_dat_lg = lb->lg;
	if (lb->pc_rdidx) max_dat_lg += lb->pc_rdidx - lb->pc;

	min_idx = lb->datalg - max_dat_lg;

	if (idx >= min_idx) {
	    if (idx <= (min_idx + max_dat_lg)) {
		cookie->idx = idx;
		*pos = idx;
		return 0;

	    } else {
		if (!lb->eof && !cookie->ende && AppHandleMessage()) goto loop;
	    }
	}
#if __WORDSIZE == 64
 printf("%s(%d): io_seek_fn min_idx=%ld %ld maxlg=%ld lg=%ld\n", FL, min_idx, idx, max_dat_lg, lb->lg);
#else
 printf("%s(%d): io_seek_fn min_idx=%lld %lld maxlg=%lld lg=%ld\n", FL, min_idx, idx, max_dat_lg, lb->lg);
#endif
	return ESPIPE;
}

static int io_close_fn(TNetFtpStream *cookie) {
 int result = 0;
	if (cookie->cookie_io.write) {
	    cookie->ioinfo.DataSize = cookie->lb->datalg;
	    cookie->lb->eof = True;
	    while (!cookie->ende && !cookie->error && AppHandleMessage());
	}
	if (cookie->error) result = -1;
	Tfree(cookie);
	return result;
}

/*_____________________________ Write ___________________________________*/
static void TNetFtpStream_OnPipeWrite(int fd, Tfd_list *fdl) {
  int blg;
  TNetFtpStream *self = fdl->obj;
  char *pc;

	blg = TLineBuf_Get(self->lb, BUFSIZ - sizeof(TdData), &pc);
	if (!self->lb->eof && (blg != (BUFSIZ - sizeof(TdData)))) {
	    appDat.AktPipes[fd] = OutPipe + 1;
	    return;
	}

	self->shd->lg = sizeof(TdData);
	if (blg > 0) {
	    self->shd->lg += blg;
	    pnmove(self->shd->dat, pc, blg);
	    self->shd->sh_DataNowSize += blg;
	    self->shd->sh_LoadTime_us = GetCurTimeDiv_us64(self->ref_us);
	    blg = TLineBuf_Sub(self->lb, blg);
	}

	if (self->lb->eof && !blg) self->shd->fu = hSHfu_EOF;
//	FileWrite(fd, self->shd, self->shd->lg, NULL);
	file_write(fd, NULL, self->shd, self->shd->lg, NULL);
	if (self->shd->fu & hSHfu_EOF) ClosePipeProc(fd, 0);
}

/*_____________________________ fmsg ____________________________________*/
static void OnFileMsg(TFileMsg *fmsg) {
  TFtpClientCMD *ftpcmd;
  TNetFtpStream *self;
	if (!TNetFtp_fmsg_GetSelf(fmsg, &ftpcmd, (void*)&self)) return;

	if (TNetFtp_Chk_ftpappclient((TNetFtp*)self, fmsg)) {
	    if (fmsg->D.m == facSendPipe) {
		self->ref_us = GetTime_us64();
		AddPipeProc(fmsg->F.m, OutPipe, TNetFtpStream_OnPipeWrite, self);
		fmsg->res = 1;
	    }

	} else if ((fmsg->art == fmArt_TdServ) &&
		   (fmsg->typF == fmTyp_shd)) {

	    TdData *shd = fmsg->F.shd;
	    self->ioinfo.DataNowSize = shd->sh_DataNowSize;
	    self->ioinfo.Time_us = shd->sh_LoadTime_us;

	    if (ftpcmd->stat & ftpcmdmk_Put) {
		if (shd->fu & hSHfu_EOF) self->ioinfo.stat |= fmIOInfo_EOF;
		TNet_send_ioinfo((TNet*)self);

	    } else {
		if (!self->error && ((shd->fu & hSHfu_onlyInfo) == 0)) {
		    int dlg = shd->lg - sizeof(TdData);
		    TLineBuf_Add(self->lb, shd->dat, dlg);
		    if (self->lb->eof) {
			self->ende = True;
			self->ioinfo.stat |= fmIOInfo_EOF;
		    }
		    TNet_send_ioinfo((TNet*)self);
		}
	    }
	    fmsg->Ok = True;
	    fmsg->ev_Stop = True;
	} else {

	    TNet_msg_next((TNet*)self, fmsg);
	}
}

/*_____________________________ Destroy _________________________________*/
static bool TNetFtpStream__Destroy(TNetFtpStream *self) {
	Tfree(self->lb);
	return TNetFtp__Destroy((TNetFtp*)self);
}

/*_____________________________ Create __________________________________*/
static TNetFtpStream *TNetFtpStream_Create(TglobPath *gp) {
  TNetFtpStream *self;

	self = TNetFtp_Create(sizeof(TNetFtpStream), wtyp_TNetStream, NULL, gp);
	self->destroy = (TDESTROY)TNetFtpStream__Destroy;
	self->lb = TLineBuf_Create();
	self->cookie_io.seek = (cookie_seek_function_t*)io_seek_fn;
	self->cookie_io.close = (cookie_close_function_t*)io_close_fn;
	return self;
}

static FILE *TNetFtpStream_ftpac_add(TNetFtpStream *self) {
  const char *bef, *opentype;
  char *cmd;
  bool Ok;

	bef = "get";
	opentype = "rb";
	if (self->omod == NetOpenMod_Write) {
	    bef = "put";
	    opentype = "wb";
	}
	self->f = fopencookie(self, opentype, self->cookie_io);

	cmd = pcadd_n(4, "\"", self->gp.path, self->gp.name, "\"");
	Ok = self->f && TNetFtp_ftpac_add((TNetFtp*)self, bef, cmd, OnFileMsg);
	xfree(cmd);
	if (Ok) return self->f;
// error
	self->error = True;
	if (self->f) {
	    fclose(self->f);  //+Tfree(self);
//	    self->f = NULL;
	} else {
	    Tfree(self);
	}
	return NULL;
}

static FILE *ftp_stream_open_rd_gp(TglobPath *gp) {
  TNetFtpStream *self;
  Tstat64 *st;

	st = ftp_stat_gp(gp, NULL, fsFollowLink | fsGetFileType | fsNoFileName | fsNoLinkName);
	if (!st || (st == PTR_FAILED)) return NULL;

	self = TNetFtpStream_Create(gp);
	self->ioinfo.DataSize = st->size;
	xfree(st);
	self->cookie_io.read = (cookie_read_function_t*)io_read_fn;
	self->omod = NetOpenMod_Read;

	return TNetFtpStream_ftpac_add(self);
}

static FILE *ftp_stream_open_wr_gp(TglobPath *gp) {
  TNetFtpStream *self;

	self = TNetFtpStream_Create(gp);
	self->cookie_io.write = (cookie_write_function_t*)io_write_fn;
	self->shd = xcalloc(BUFSIZ);
	self->shd->HeadArt = hServData;
	self->shd->bef = hSHbd_Dat;
	self->omod = NetOpenMod_Write;

	return TNetFtpStream_ftpac_add(self);
}


#else /* USE_CustomStream */
typedef struct {
  AlgTNetFtp
  FILE *f;
} Tprivate, TNetFtpStream;

static char *x_bp;
static size_t x_size;

static void OnFileMsg(TFileMsg *fmsg) {
  TFtpClientCMD *ftpcmd;
  TNetFtpStream *self;
	if (!TNetFtp_fmsg_GetSelf(fmsg, &ftpcmd, (void*)&self)) return;

	if (TNetFtp_Chk_ftpappclient((TNetFtp*)self, fmsg)) {
	    if (self->error && self->f) {
		fclose(self->f);
		self->f = NULL;
	    }

	} else if ((fmsg->art == fmArt_TdServ) &&
		   (fmsg->typF == fmTyp_shd)) {

	    TdData *shd = fmsg->F.shd;
	    self->ioinfo.DataNowSize = shd->sh_DataNowSize;
	    self->ioinfo.Time_us = shd->sh_LoadTime_us;

	    if (!self->error && ((shd->fu & hSHfu_onlyInfo) == 0)) {
		int dlg = shd->lg - sizeof(TdData);
		if (dlg > 0) {
		    if (!self->f) {
			self->f = open_memstream(&x_bp, &x_size);
		    }
		    if (fwrite(shd->dat, 1, dlg, self->f) != dlg) {
			int err = errno;
			fclose(self->f);
			self->f = NULL;
			self->error = True;
			errno = err;
			appError(EMSG_ERR1 | EMSG_H_LF | EMSG_tIO | EMSG_SYS, 
			    "%s(%d) : stream load\n %s\n", FL, ftpcmd->cmd);
		    }
		} else {
		    self->ioinfo.stat |= fmIOInfo_EOF;
		}
		TNet_send_ioinfo((TNet*)self);

		fmsg->Ok = True;
		fmsg->ev_Stop = True;
	    }

	}
}

FILE *ftp_stream_open_rd_gp(TglobPath *gp) {
  TNetFtpStream *self;
  char *bef;
  FILE *result = NULL;

	bef = pcadd(gp->path, gp->name);
	self = TNetFtp_Create(sizeof(TNetFtpStream), wtyp_TNetStream, NULL, gp);

	self->omod = NetOpenMod_Read;

	if (!TNetFtp_ftpac_add((TNetFtp*)self, "get ", bef, 
				OnFileMsg, afu_BEF_free)) goto end_err;

	if (TNetFtp_wait((TNetFtp*)self)) result = self->f;

end_err:
	Tfree(self);
	return result;
}

#endif /* ! USE_CustomStream */

static FILE *ftp_stream_open_gp(TglobPath *gp, const char *opentype) {
  bool mkrd, mkwr;

	if (!opentype) return NULL;
	if (!ftp_chk_RootArt_URL(gp)) return NULL;

	mkrd = mkwr = False;
	while (*opentype) {
	    if ((*opentype == 'r') || (*opentype == 'R')) mkrd = True;
	    if ((*opentype == 'w') || (*opentype == 'W')) mkwr = True;
	    opentype++;
	}

	if (mkrd && !mkwr) return ftp_stream_open_rd_gp(gp);
#if USE_CustomStream
	if (!mkrd && mkwr) return ftp_stream_open_wr_gp(gp);
#endif
	return NULL;
}

const ThNetStreamLoader hNetStreamLd_ftp = {
  ftp_chk_RootArt,
  .Create.lb = NULL,
  ftp_stream_open_gp
};
