/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "ftp_net.h"
#include "ftp_cache.h"
#include <time.h>
#include <errno.h>

static bool ftp_bef_gp_fcnx(TglobPath *gp, TNetErrRet *ErrRet, char *bef,
			dword akt_fkt, TFtpClientNextCMD *fcnx) {
  TNetFtp *self;
  int result;

	result = False;
	self = TNetFtp_Create(sizeof(TNetFtp), wtyp_TNetFtp, NULL, gp);
	self->omod = NetOpenMod_Read;
	self->ErrRet = ErrRet;
	if (!TNetFtp_ftpac_add_fcnx((TNetFtp*)self, "cmd", bef, 
			TNetFtp_OnFileMsg, akt_fkt, fcnx)) goto end_err;

	result = TNetFtp_wait((TNetFtp*)self);
end_err:
	xfree(bef);
	Tfree(self);
	return result;
}

static bool ftp_bef_gp(TglobPath *gp, TNetErrRet *ErrRet, char *bef) {
	return ftp_bef_gp_fcnx(gp, ErrRet, bef, 0, NULL);
}

static char *get_ftp_filename(TglobPath *gp) {
	return path_add(gp->path, gp->name);
}

/* _______________________ chmod ________________________________________ */
static char *get_site_chmod(char *fn, int mode) {
  char cmode[8];

	cmode[0] = '0' + ((mode >> 6) & 0x07);
	cmode[1] = '0' + ((mode >> 3) & 0x07);
	cmode[2] = '0' + (mode & 0x07);
	cmode[3] = ' ';
	cmode[4] = 0;

	return pcadd_n(4, "\"SITE CHMOD ", cmode, fn, "\"");
}

bool ftp_chmod_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode) {
  char *fn, *bef;
  bool result;

	fn = get_ftp_filename(gp);
	bef = get_site_chmod(fn, mode);
	xfree(fn);

	result = ftp_bef_gp_fcnx(gp, ErrRet, bef, facfkt_CHMOD, NULL);
	if (result) FtpCache_chmod(gp, mode, 0x01ff);
	return result;
}

/* _______________________ utime ________________________________________ */
bool ftp_utime_gp(TglobPath *gp, TNetErrRet *ErrRet, time_t mtime) {
  char *bef, *pc;
  char buf[32];
  bool result;
  struct tm ut;

	localtime_r(&mtime, &ut);

	pc = buf;
	ppcmove_v32(&pc, ut.tm_year + 1900, 4);
	ppcmove_v32(&pc, ut.tm_mon + 1, 2);
	ppcmove_v32(&pc, ut.tm_mday, 2);
	ppcmove_v32(&pc, ut.tm_hour, 2);
	ppcmove_v32(&pc, ut.tm_min, 2);
	ppcmove_v32(&pc, ut.tm_sec, 2);
	pc = buf;
	while (*pc) {
	    if (*pc == ' ') *pc = '0';
	    pc++;
	}

	bef = get_ftp_filename(gp);
//	ppcchange_add_n(&bef, 5, "\"SITE UTIME ", ctime, " ", bef, "\"");
	ppcchange_add_n(&bef, 9, "\"SITE UTIME ", bef, " ", buf, " ", 
						buf, " ", buf, " UTC\"");

	result = ftp_bef_gp_fcnx(gp, ErrRet, bef, facfkt_UTIME, NULL);
	if (result) FtpCache_mtime(gp, mtime);
	return result;
}

/* _______________________ rename _______________________________________ */
static bool mv_rename(TglobPath *src_gp, TNetErrRet *ErrRet, TglobPath *dest_gp) {
  TNetFtp *self;
  char *bef;
  Tstat64 *st;
  bool result;

	self = NULL;
	bef = NULL;
	result = False;
	st = ftp_stat_gp(dest_gp, NULL, 0);
	if (st) {
	    if (ErrRet) {
		ErrRet->err = pcadd("dest exists ", dest_gp->name);
	    }
	    goto end_err;
	}
	self = TNetFtp_Create(sizeof(TNetFtp), wtyp_TNetFtp, NULL, src_gp);
	self->omod = NetOpenMod_Read;
	self->ErrRet = ErrRet;

	bef = get_ftp_filename(src_gp);
	ppcchange_add_n(&bef,  3, "\"RNFR ", bef, "\"");
	if (!TNetFtp_ftpac_add((TNetFtp*)self, "cmd", bef, 
				TNetFtp_OnFileMsg)) goto end_err;

	result = TNetFtp_wait((TNetFtp*)self);
	if (!result && (self->retCode != 350)) goto end_err;

	if (ErrRet) ppcfree(&ErrRet->err);
	xfree(bef);
	bef = get_ftp_filename(dest_gp);
	ppcchange_add_n(&bef, 3, "\"RNTO ", bef, "\"");
	if (!TNetFtp_ftpac_add((TNetFtp*)self, "cmd", bef, 
				TNetFtp_OnFileMsg)) goto end_err;
	result = TNetFtp_wait((TNetFtp*)self);
end_err:
	xfree(bef);
	Tfree(self);
	return result;
}

static void ftp_gp_setup(TglobPath *dup_gp, TglobPath *gp) {
	if (gp->name) {
	    gp_dup_gpdata(dup_gp, gp);
	} else {
	    gp_dup_gpdata_path(dup_gp, gp);
	    gp_PathToName(dup_gp);
	}
}

static bool ftp_rename_gp(TglobPath *gp, TNetErrRet *ErrRet, const char *new_name) {
  TglobPath src_gp, dest_gp;
  bool result;

	ftp_gp_setup(&src_gp, gp);
	gp_dup_gpdata_path(&dest_gp, &src_gp);
	dest_gp.name = pcdup(new_name);

	result = mv_rename(&src_gp, ErrRet, &dest_gp);
	if (result) FtpCache_rename(&src_gp, new_name, True);
	gp_pc_free(&dest_gp);
	gp_pc_free(&src_gp);
	return result;
}

/* _______________________ mv ___________________________________________ */
static bool ftp_mv_gp(TglobPath *gp, TNetErrRet *ErrRet, TglobPath *new_gp) {
  TglobPath src_gp, dest_gp;
  bool result;

	ftp_gp_setup(&src_gp, gp);
	ftp_gp_setup(&dest_gp, new_gp);

	result = mv_rename(&src_gp, ErrRet, &dest_gp);
	if (result && !FtpCache_mv(&src_gp, &dest_gp, True)) {
	    Tstat64 *st = ftp_stat_gp(&dest_gp, NULL, fsGetFileAll_UGT|fsForceReScan);
	    xfree(st);
	}
	gp_pc_free(&dest_gp);
	gp_pc_free(&src_gp);
	return result;
}

/* _______________________ unlink _______________________________________ */
bool ftp_unlink_gp(TglobPath *gp, TNetErrRet *ErrRet) {
  char *bef;
  bool result;

	bef = get_ftp_filename(gp);
	ppcchange_add_n(&bef, 3, "\"DELE ", bef, "\"");

	result = ftp_bef_gp(gp, ErrRet, bef);
	if (result) FtpCache_Clr_name(gp, True);
	return result;
}

/* _______________________ rmdir ________________________________________ */
bool ftp_rmdir_gp(TglobPath *gp, TNetErrRet *ErrRet) {
  char *bef;
  bool result;

	bef = get_ftp_filename(gp);
	ppcchange_add_n(&bef, 3, "\"RMD ", bef, "\"");

	result = ftp_bef_gp(gp, ErrRet, bef);
	if (result) {
	    if (gp->name) {
		FtpCache_Clr_name(gp, True);
	    } else {
		FtpCache_Clr(gp, True);
	    }
	}
	return result;
}

/* _______________________ mkdir ________________________________________ */
// ftp_mkdir_gp_NoReScan is used in xcy_ftp.c !!!
bool ftp_mkdir_gp_NoReScan(TglobPath *gp, TNetErrRet *ErrRet, int mode) {
  char *bef;
  TFtpClientNextCMD *fcnx;

	bef = get_ftp_filename(gp);

	fcnx = NULL;
	int fkt = TFtpAppClient_get_fkt(gp);
	if (mode && (fkt & facfkt_CHMOD)) {
	    fcnx = xcalloc(sizeof(TFtpClientNextCMD));
	    fcnx->bef = pcdup("cmd");
	    fcnx->cmd = get_site_chmod(bef, mode);
	    fcnx->nxfu = fcnxfu_ErrorIsOk;
	    fcnx->akt_fkt = facfkt_CHMOD;
	}
	ppcchange_add_n(&bef, 3, "\"MKD ", bef, "\"");

	return ftp_bef_gp_fcnx(gp, ErrRet, bef, 0, fcnx);
}

bool ftp_mkdir_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode) {
  bool result = ftp_mkdir_gp_NoReScan(gp, ErrRet, mode);
	if (result) {
	    Tstat64 *st = ftp_stat_gp(gp, NULL, fsGetFileAll_UGT|fsForceReScan);
	    xfree(st);
	}
	return result;
}

const ThNetToolsLoader hNetToolsLd_ftp = {
  ftp_chk_RootArt,
  .Create.lb = NULL,
  ftp_chmod_gp,
  ftp_utime_gp,
  ftp_rename_gp,
  ftp_mv_gp,
  ftp_unlink_gp,
  ftp_rmdir_gp,
  ftp_mkdir_gp
};
