/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "../net_xxx.h"
#include <unistd.h>

static TFtpAppClient *FtpAppClient_RxD;
static TFtpAppClient *FtpAppClient_TxD2;
static TEv_Array *ftpac_appEv;
static dword ftpac_us;

int TFtpAppClient_TimeOut = 120;

static void OnDataPipeRead(Texec_pipe* expip);

static void acDoFileMsg(TFtpAppClient *ftpac, TFtpClientCMD *ftpcmd) {

	if (AppToAtExit) return;
	ftpac->fmsg.res = 0;
	ftpac->fmsg.Ok = False;
	ftpac->fmsg.ev_Stop = False;
	if (ftpcmd) {
	    int wtyp = ftpac->fmsg.wtyp;
	    void *obj = ftpac->fmsg.obj;
	    void *owner = ftpac->fmsg.owner;
	    char *path = ftpac->gp.path;
	    char *name = ftpac->gp.name;
	    if (ftpcmd->gp.path) {
		ftpac->gp.path = ftpcmd->gp.path;
		ftpac->gp.name = ftpcmd->gp.name;
	    }
	    ftpac->fmsg.wtyp = ftpcmd->wtyp;
	    ftpac->fmsg.obj = ftpcmd;
	    ftpac->fmsg.owner = ftpcmd->owner;

	    if (ftpcmd->OnFileMsg) {
		ftpcmd->OnFileMsg(&ftpac->fmsg);
		if (!ftpac->fmsg.owner) ftpcmd->owner = NULL;
	    }
	    file_msg(&ftpac->fmsg);
	    ftpac->fmsg.wtyp = wtyp;
	    ftpac->fmsg.obj = obj;
	    ftpac->fmsg.owner = owner;

	    ftpac->gp.path = path;
	    ftpac->gp.name = name;
	} else {
	    TNet_msg_next((TNet*)ftpac, &ftpac->fmsg);
	}
}

static void acDoFileMsg_appc(TFtpAppClient *ftpac, TFtpClientCMD *ftpcmd, int bef) {
	ftpac->fmsg.art = fmArt_ftpappclient;
	ftpac->fmsg.typD = fmTyp_int;
	ftpac->fmsg.D.m = bef;
	acDoFileMsg(ftpac, ftpcmd);
}

static void ac5DoFileMsg(TFtpAppClient *ftpac, TFtpClientCMD *ftpcmd, int art, 
			int typD, void *D, int typF, void *F) {
	ftpac->fmsg.art = art;
	ftpac->fmsg.typD = typD;
	ftpac->fmsg.D.v = D;
	ftpac->fmsg.typF = typF;
	ftpac->fmsg.F.v = F;
	acDoFileMsg(ftpac, ftpcmd);
}


static void ftpcmd0_free(TFtpAppClient *ftpac) {
  TFtpClientCMD *ftpcmd = ftpac->ftpcmd;

	if (ftpcmd) {
	    ftpac->ftpcmd = ftpcmd->next;
	    Tfree(ftpcmd);
	}
}

static void fcnx_free(TFtpClientNextCMD *fcnx) {
	xfree(fcnx->bef);
	xfree(fcnx->cmd);
	xfree(fcnx);
}

static void ftpcmd_fcnx_free(TFtpClientCMD *self) {
	while (self->fcnx) {
	    TFtpClientNextCMD *fcnx = self->fcnx;
	    self->fcnx = fcnx->next;
	    fcnx_free(fcnx);
	}
}

static void TFtpAppClient_CMD_send(TFtpAppClient *self, TFtpClientCMD *ftpcmd) {
  int fd;
	if (!ftpcmd) return;
	ftpcmd->stat |= ftpcmdmk_send;
	fd = self->expip->pp[0].fd;
	if (ftpcmd->bef && ftpcmd->cmd) {
	    fd_write_va(fd, "%s %s\n", ftpcmd->bef, ftpcmd->cmd);

	} else if (ftpcmd->bef) {
	    fd_write_va(fd, "%s\n", ftpcmd->bef);

	} else if (ftpcmd->cmd) {
	    fd_write_va(fd, "%s\n", ftpcmd->cmd);

	} else {
	    TdServHead sh;
	    Fill0(&sh, sizeof(TdServHead));
	    sh.lg = sizeof(TdServHead);
	    sh.HeadArt = hServHead;
	    sh.bef = hSHb_Ok;
	    TLineBuf_Add(self->expip->pp[4].lb, (char*)&sh, sh.lg);
	    OnDataPipeRead(self->expip);
	}
}

static void TFtpAppClient_Open_send(TFtpAppClient *self) {
  char *url;
  int fd = self->expip->pp[0].fd;

	if (!self->VerStr) fd_write_va(fd, "ver\n");
	self->stat |= ftpacmk_SendOpen;
	url = get_ftpURL(&self->gp);
	fd_write_va(fd, "open %s\n", url);
	xfree(url);
}

static void ftpcmd_set_fcnx(TFtpClientCMD *ftpcmd, TFtpClientNextCMD *fcnx) {
	xfree(ftpcmd->bef);
	xfree(ftpcmd->cmd);
	ftpcmd->bef = fcnx->bef;
	ftpcmd->cmd = fcnx->cmd;
	ftpcmd->stat = fcnx->stat;
	ftpcmd->nxfu = fcnx->nxfu;
	ftpcmd->akt_fkt = fcnx->akt_fkt;
	if (fcnx->OnFileMsg) ftpcmd->OnFileMsg = fcnx->OnFileMsg;
	xfree(fcnx);
}

static void OnDataPipeRead(Texec_pipe* expip) {
  TdServHead *sh;
  TFtpAppClient *ftpac = expip->owner;
  TFtpClientCMD *ftpcmd;
  const char *indata, *token;
  int tsf;
  bool SendEndCode_OK;
  bool IsOkLoop_OK;

	if (!ftpac || (ftpac->stat & ftpacmk_PipeRead)) return;
	ftpac->stat |= ftpacmk_PipeRead;
	ftpac->fmsg.typR = fmTyp_gp;
	ftpac->fmsg.R.gp = &ftpac->gp;
	tsf = 0;

	while (!(ftpac->stat & (ftpacmk_ClientClose|ftpacmk_TimeEnd)) && 
		GetSyncPipeServHead(expip->pp[4].lb, &sh, &tsf)) {

	    ftpac->dofree_sec = TFtpAppClient_TimeOut;
	    IsOkLoop_OK = False;

	    if ((sh->HeadArt == hServHead) && (sh->bef == hSHb_VerStr)) {
		ftpac->VerStr = pcdup(sh->dat);
		int toklg;
		indata = ftpac->VerStr;
		while (indata) {
		    indata = pcGetToken(indata, &token, &toklg);
		    if ((toklg == 7) && pcncmp(token, "version", 7)) {
			pcGetToken(indata, &token, &toklg);
			int m = ppctoint((const char **)&token) << 16;
			if (*token == '.') m |= ppctoint_scan((const char **)&token, True) << 8;
			if (*token == '.') m |= ppctoint_scan((const char **)&token, True);
			ftpac->VerNr = m;
			break;
		    }
		}
		if (ftpac->VerNr < 0x00510d) {
		    char *err = pcadd("aldftp vernr < 0.81.13\n", ftpac->VerStr);
		    ac5DoFileMsg(ftpac, NULL, fmArt_AlgErr, fmTyp_char, err, 
							0, NULL);
		    xfree(err);
		}
		continue;
	    }

	    ftpcmd = ftpac->ftpcmd;

	    if ((ftpac->stat & ftpacmk_Open) && ftpcmd) {
		if (ftpcmd->stat & ftpcmdmk_send) {

		    if ((sh->HeadArt == hServHead) && 
			(sh->bef == hSHb_Ok) &&
			(ftpcmd->stat & ftpcmdmk_send)) {
			ftpcmd->stat |= ftpcmdmk_OK;

		    } else if (sh->HeadArt == hServData) {
			if ((sh->bef == hSHbd_Dat) && 
			    (sh->fu & hSHfu_EOF)) {
			    ftpcmd->stat |= ftpcmdmk_End;
			}
		    }
		    if (ftpcmd->stat & ftpcmdmk_OK) {
			SendEndCode_OK = True;
loop_fcnx_OK:
			if (ftpcmd->fcnx) {
			    TFtpClientNextCMD *fcnx = ftpcmd->fcnx;
			    ftpcmd->fcnx = fcnx->next;

			    if (fcnx->nxfu & fcnxfu_EndOk) {
				ftpcmd_fcnx_free(ftpcmd);
				goto loop_fcnx_OK;
			    }

			    if (fcnx->nxfu & fcnxfu_SkipOk) {
				fcnx_free(fcnx);
				goto loop_fcnx_OK;
			    }
			    ftpcmd_set_fcnx(ftpcmd, fcnx);
			    SendEndCode_OK = False;
			}
			if (SendEndCode_OK) {
			    acDoFileMsg_appc(ftpac, ftpcmd, facEndCode_OK);
			    if (ftpcmd->stat & ftpcmdmk_send) ftpcmd0_free(ftpac);
			}
			if (IsOkLoop_OK) continue;
		    }
		}
	    }

Next_ftpcmd:
	    ftpcmd = ftpac->ftpcmd;  //refresh ftpcmd
	    if (sh->HeadArt == hServHead) {
		if (sh->bef == hSHb_Ok) {
		    if (ftpac->stat & ftpacmk_SendOpen) {
			ftpac->stat &= ~ftpacmk_SendOpen;
			ftpac->stat |= ftpacmk_Open;
		    }

		    if ((ftpac->stat & ftpacmk_Open) && ftpcmd && 
			!(ftpcmd->stat & ftpcmdmk_send)) {
			if (ftpcmd->stat & ftpcmdmk_Stop) {
			    ftpcmd0_free(ftpac);
			    goto Next_ftpcmd;
			} else {
			    TFtpAppClient_CMD_send(ftpac, ftpcmd);
			}
		    }

		} else if (sh->bef == hSHb_StrgMSG) {
		    ac5DoFileMsg(ftpac, ftpcmd, fmArt_strg, 
#ifdef POINTER64BIT
				fmTyp_int, (void*)(int64_t)sh->strg, fmTyp_sh, sh);
#else
				fmTyp_int, (void*)(int)sh->strg, fmTyp_sh, sh);
#endif
		    if (sh->strg == hSHStM_WrOpen) {
			ftpac->fmsg.typF = fmTyp_int;
			ftpac->fmsg.F.m = expip->pp[3].fd;
			acDoFileMsg_appc(ftpac, ftpcmd, facSendPipe);
			if (ftpac->fmsg.res != 1) {
			    TdData Data;
			    Fill0(&Data, sizeof(TdData));
			    Data.lg = sizeof(TdData);
			    Data.HeadArt = hServData;
			    Data.fu = hSHfu_EOF;
			    write(expip->pp[3].fd, &Data, Data.lg);
			}
		    }

		} else if (sh->bef == hSHb_SET_Login) {

		    ac5DoFileMsg(ftpac, ftpcmd, fmArt_errLogin, 
					fmTyp_char, sh->dat, fmTyp_sh, sh);

		    if ((ftpac->fmsg.res == 1) &&
			!(ftpac->stat & ftpacmk_ClientClose)) {  //msgOK
			ftpac->fmsg.art = fmArt_NewLogin;
			while (ftpcmd) {
			    acDoFileMsg(ftpac, ftpcmd);
			    ftpcmd = ftpcmd->next;
			}
			TFtpAppClient_Open_send(ftpac);

		    } else {
			ftpac->stat &= ~ftpacmk_Open;
			ftpac->stat |= ftpacmk_ErrorOpen;
			ftpac->fmsg.art = fmArt_errConnect;
			while (ftpac->ftpcmd) {
			    acDoFileMsg(ftpac, ftpac->ftpcmd);
			    ftpcmd0_free(ftpac);
			}
		    }

		} else if (sh->bef == hSHb_ERR) {

		    if (sh->m < ioErr_Unknown_host) {
			ftpac->fkt &= ~ftpcmd->akt_fkt;
			if (ftpcmd->nxfu & fcnxfu_ErrorIsOk) {
			    ftpcmd->stat |= ftpcmdmk_OK;
			    SendEndCode_OK = True;
			    IsOkLoop_OK = True;
			    goto loop_fcnx_OK;
			}
loop_fcnx_Err:
			if (ftpcmd->fcnx) {
			    TFtpClientNextCMD *fcnx = ftpcmd->fcnx;

				ftpcmd->fcnx = fcnx->next;
			    if (!(fcnx->nxfu & fcnxfu_SkipOk)) {
				fcnx_free(fcnx);
				goto loop_fcnx_Err;
			    }
			    
			    ftpcmd_set_fcnx(ftpcmd, fcnx);
			    TFtpAppClient_CMD_send(ftpac, ftpcmd);
			    continue;
			}
		    }

		    ac5DoFileMsg(ftpac, ftpcmd, fmArt_AlgErr, 
					fmTyp_char, sh->dat, fmTyp_sh, sh);

		    if ((sh->m >= ioErr_Unknown_host) && (sh->m <= ioErr_NoConnMax)) {
			ftpac->stat &= ~ftpacmk_Open;
			ftpac->stat |= ftpacmk_ErrorOpen;
			ftpac->fmsg.art = fmArt_errConnect;
			while (ftpac->ftpcmd) {
			    acDoFileMsg(ftpac, ftpac->ftpcmd);
			    ftpcmd0_free(ftpac);
			}
			Tfree(ftpac);

		    } else {
			ftpcmd0_free(ftpac);
			ftpcmd = ftpac->ftpcmd;
			if (ftpcmd) TFtpAppClient_CMD_send(ftpac, ftpcmd);
		    }
		}

	    } else if (sh->HeadArt == hServData) {
		ac5DoFileMsg(ftpac, ftpcmd, fmArt_TdServ, 
					0, NULL, fmTyp_shd, sh);

	    } else if (sh->HeadArt == hServFileStat) {
		ac5DoFileMsg(ftpac, ftpcmd, fmArt_DataIO, 
				fmTyp_stat, &((TdServFile*)sh)->st, fmTyp_shst, sh);
	    }
	}

	ftpac->stat &= ~ftpacmk_PipeRead;
	if (ftpac->stat & ftpacmk_ClientClose) Tfree(ftpac);
}

#define debug_line_IN 0

#if debug_line_IN
static void OnPipeRead(Texec_pipe* expip) {
  char *pc;
  Tlgs lgs;

	while (TLineBuf_GetLineEOL_lgs(expip->pp[fd1_infd].lb, &lgs, NULL)) {
	    pc = pcndup(lgs.s, lgs.lg);
 printf(" %s\n", pc);
	    xfree(pc);
	}
}
#endif

static void TFtpAppClient_OnExecProcEv(Texec_pipe* expip, int evtyp) {
  TFtpAppClient *ftpac;

	switch (evtyp) {

	case evtyp_infd:
#if debug_line_IN
	case evtyp_infd_Close:
	    OnPipeRead(expip);
#else
	    TLineBuf_Clr(expip->pp[1].lb);
#endif
	    break;

	case evtyp_Dinfd_Close:
	    ftpac = expip->owner;
	    if (ftpac) ftpac->stat |= (ftpacmk_GoFree | ftpacmk_ClientClose);
	case evtyp_Dinfd:
	    OnDataPipeRead(expip);
	    break;

	case evtyp_ProcClose:
	    ftpac = expip->owner;
	    if (ftpac) {
		ftpac->pid = 0;
		if (!(ftpac->stat & ftpacmk_PipeRead)) Tfree(ftpac);
	    }
	    break;
	}
}

/*_____________________________ appEv ___________________________________*/
static void TFtpAppClient_appEv(TEv_Array *appEv, TFtpAppClient *ftpac) {
  TFtpAppClient *ftpac_next;

	while (ftpac) {
	    ftpac_next = ftpac->next;
	    if (!(ftpac->stat & (ftpacmk_GoFree|ftpacmk_ClientClose|ftpacmk_PipeRead))) {
		if (ftpac->dofree_sec > 0) {
		    ftpac->dofree_sec--;

		    if (!ftpac->dofree_sec && (ftpac->stat & ftpacmk_Open)) {
			ftpac->stat |= ftpacmk_TimeEnd;
			fd_write_va(ftpac->expip->pp[0].fd, "close\n");
		    }

		} else {
		    Tfree(ftpac);
		}
	    }
	    ftpac = ftpac_next;
	}
}

static void TFtpAppClient_On_appEv(TEv_Array *appEv) {

	if (cpTime(ftpac_us, appDat.app_us)) {
	    ftpac_us += 1000000;
	    TFtpAppClient_appEv(appEv, FtpAppClient_RxD);
	    TFtpAppClient_appEv(appEv, FtpAppClient_TxD2);
	}
}

bool TFtpAppClient_CMD_add(TFtpAppClient *self, TFtpClientCMD *ftpcmd) {
  TFtpClientCMD *tmp;

	ftpcmd->ftpac = self;
	if (self->ftpcmd) {
	    tmp = self->ftpcmd;
	    while (tmp->next) tmp = tmp->next;
	    tmp->next = ftpcmd;
	} else {
	    self->ftpcmd = ftpcmd;
	    if (self->stat & ftpacmk_Open) {
		TFtpAppClient_CMD_send(self, ftpcmd);

	    } else if (self->stat & ftpacmk_ErrorOpen) {
		TFtpAppClient_Open_send(self);
	    }
	}
	return True;
}

bool TFtpClientCMD_Stop(TFtpClientCMD *self) {
	if (!self || !self->ftpac) return False;
	if (self->ende || self->error) return True;
	self->stat |= ftpcmdmk_Stop;

	fd_write_va(self->ftpac->expip->pp[0].fd, "stop\n");
	return True;
}

/*_____________________________ Destroy _________________________________*/

static bool TFtpClientCMD__Destroy(TFtpClientCMD *self) {
  TFtpClientCMD **ppftpcmd;
  TFtpAppClient *ftpac = self->ftpac;

	if (ftpac) {
	    acDoFileMsg_appc(ftpac, self, facCMD_Free);

	    ppftpcmd = &ftpac->ftpcmd;
	    while (*ppftpcmd && (*ppftpcmd != self)) ppftpcmd = &(*ppftpcmd)->next;
	    if (*ppftpcmd == self) *ppftpcmd = self->next;
	}
	xfree(self->bef);
	xfree(self->cmd);
	ftpcmd_fcnx_free(self);
	return TNet__Destroy((TNet*)self);
}

static bool TFtpAppClient__Destroy(TFtpAppClient *self) {
  TFtpAppClient **ppftpac;

	while (self->ftpcmd) ftpcmd0_free(self);

	if (self->pid) {
	    if (!(self->stat & ftpacmk_GoFree)) {
		self->stat |= ftpacmk_GoFree;
		fd_write_va(self->expip->pp[0].fd, "exit\n");
	    }
	    self->pid = 0;
	    if (!AppToAtExit) return True;
	}

	self->expip->owner = NULL;
	appMutex_lock();
	ppftpac = self->ppfac;
	while (*ppftpac && (*ppftpac != self)) ppftpac = &(*ppftpac)->next;
	if (*ppftpac != self) {
#ifdef POINTER64BIT
 printf("%s(%d): ERROR: Double TFtpAppClient__Destroy 0x%lx ???\n", FL, (int64_t)self);
#else
 printf("%s(%d): ERROR: Double TFtpAppClient__Destroy 0x%x ???\n", FL, (int)self);
#endif
	}

	*ppftpac = self->next;

	if (!FtpAppClient_RxD && !FtpAppClient_TxD2) {
	    Free_appEv(ftpac_appEv);
	    ftpac_appEv = NULL;
	}
	appMutex_unlock();
	xfree(self->VerStr);
	return TNet__Destroy((TNet*)self);
}

static void TFtpClientCMD__AppAtExit() {
	if (!FtpAppClient_RxD && !FtpAppClient_TxD2) return;
	AppToAtExit_Wait = True;
	while (FtpAppClient_RxD)  TFtpAppClient__Destroy(FtpAppClient_RxD);
	while (FtpAppClient_TxD2) TFtpAppClient__Destroy(FtpAppClient_TxD2);
}

/*_____________________________ Create __________________________________*/
TFtpClientCMD *TFtpClientCMD_Create(const char *bef, const char *cmd, TglobPath *gp) {
  TFtpClientCMD *self;

	self = TNet_Create(sizeof(TFtpClientCMD), wtyp_TFtpClientCMD, NULL);
	self->destroy = (TDESTROY)TFtpClientCMD__Destroy;
	self->bef = pcdup(bef);
	self->cmd = pcdup(cmd);
	if (gp) {
	    self->gp.path = pcdup(gp->path);
	    self->gp.name = pcdup(gp->name);
	}
	return self;
}

static TFtpAppClient *TFtpAppClient_Create(TglobPath *gp) {
  const char *fn;
  TFtpAppClient *self, **ppftpac;

	self = TNet_Create(sizeof(TFtpAppClient), wtyp_TFtpAppClient, gp);
	self->destroy = (TDESTROY)TFtpAppClient__Destroy;
	self->dofree_sec = TFtpAppClient_TimeOut;

	self->fkt = facfkt_CHMOD|facfkt_UTIME|facfkt_STAT;

	self->ppfac = &FtpAppClient_RxD;
	if (gp->fu & gpi_dest) self->ppfac = &FtpAppClient_TxD2;


	ppcfree(&self->gp.name);
	ppcchange_cmp(&self->gp.path, "/");

#if 0
  fn = "/root/aldccp-0.83.04/proc/aldftp/aldftp";
  if (!TestIsFile(fn)) fn = "aldftp";
 printf("%s(%d): %s\n", FL, fn);
#if 0
 if (gp->fu & gpi_dest) {
  fn = pcadd(fn, " -l 4000");
 } else {
  fn = pcadd(fn, " -l 4000");
 }
#endif
#else
	fn = ALDccpPREFIX"/bin/aldftp";
	if (!TestIsFile(fn)) fn = "aldftp";
#endif
	self->pid = ExecuteProcApp(fn, outfd_Open|infd_Open|Dinfd_Open|Doutfd_Open, TFtpAppClient_OnExecProcEv, self, &self->expip);
	if (!self->pid) {
	    return NULL;   //Tfree(self); in TFtpAppClient_OnExecProcEv()
	}

	appMutex_lock();
	ppftpac = self->ppfac;
	while (*ppftpac) ppftpac = &(*ppftpac)->next;
	*ppftpac = self;

	if (!ftpac_appEv) {
	    ftpac_us = appDat.app_us + 1000000;
	    ftpac_appEv = Add_appEv(TFtpAppClient_On_appEv);
	}
	pp_void_add((void*)&AppProcAtExit, 1, TFtpClientCMD__AppAtExit);
	appMutex_unlock();

	TFtpAppClient_Open_send(self);
	return self;
}


static TFtpAppClient *TFtpAppClient_get_exist(TglobPath *gp) {
  TFtpAppClient *ftpac;

	ftpac = FtpAppClient_RxD;
	if (gp->fu & gpi_dest) ftpac = FtpAppClient_TxD2;
	
	while (ftpac) {
	    if (!(ftpac->stat & (ftpacmk_TimeEnd|ftpacmk_GoFree|ftpacmk_ClientClose)) &&
		pccmp(ftpac->gp.URL, gp->URL) && 
		pccmp(ftpac->gp.user, gp->user)) {
		return ftpac;
	    }
	    ftpac = ftpac->next;
	}
	return NULL;
}

TFtpAppClient *TFtpAppClient_get(TglobPath *gp) {
  TFtpAppClient *ftpac;

	if (!ftp_chk_RootArt_URL(gp)) return NULL;
	ftpac = TFtpAppClient_get_exist(gp);
	if (ftpac) return ftpac;
	return TFtpAppClient_Create(gp);
}

dword TFtpAppClient_get_fkt(TglobPath *gp) {
  TFtpAppClient *ftpac;

	if (!ftp_chk_RootArt_URL(gp)) return 0;
	ftpac = TFtpAppClient_get_exist(gp);
	if (ftpac) return ftpac->fkt;
	return facfkt_CHMOD|facfkt_UTIME|facfkt_STAT;
}
