/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "net_io.h"
#include "ftp_io.h"
#include <errno.h>
#include <stdarg.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/socket.h>

//*********************** TFtp ************************

static void TFtp_DoError(TFtp *ftp, int ErrNr, const char *msg) {
	if (ErrNr < 1100) fprintf(stderr, msg);
}

static void TFtp_ClsReply(TFtp *ftp) {
	ftp->rpIdx = 0;
	ftp->rpD = 0;
	ftp->rpCode = 0;
	ftp->rpOcode = 0;
	ftp->rpOrgCode = 0;
	ftp->rpConti = 0;
	ftp->rpPflag = 0;
	ftp->rpPx = 0;
	ftp->rpMsg = 0;
}

void TFtp_Error(TFtp *ftp, int ErrNr, const char *msg, ...) {
  va_list args;
  char line[2048];
  char lineA[2048];

	va_start(args, msg);
	vsnprintf(line, 2048, msg, args);
	va_end(args);
	snprintf(lineA, 2048, "E %4d: %s", ErrNr, line);
	if (ftp->DoError) ftp->DoError(ftp, ErrNr, lineA);
}

static void Tftp_DoFtpLineRecv(TFtp *ftp, int code, bool last) {
	if (ftp->OnLineRecv) {
	    if (ftp->ReplyBufAkt) {
		AddToReplyBuf(ftp, code, &ftp->reply_str[0], last);
	    } else {
		ftp->OnLineRecv(ftp, code, &ftp->reply_str[0], last);
	    }
	}
}

static void TFtp_EvReply(int fd, Tfd_list *fd_list) {
  int lg;
  char *pcs, *pcd, *pc;
  TFtp *ftp = fd_list->obj;
  char rpBuf[BUFSIZ];
  unsigned char c;

	lg = read(fd, rpBuf, BUFSIZ);
	if (!lg) {
	    TFtp_Error(ftp, ioErr_ServClosCon, "remote server has closed connection\n");
	    TFtp_Close(ftp);
	    return;
	}
	rpBuf[lg] = 0;
// fprintf(stderr, "-> %s(%d): %s <-\n", FL, rpBuf);
// fflush(stdout);

	pcs = rpBuf;
loop:
	pcd = &ftp->reply_str[ftp->rpIdx];
	while ((lg > 0) && (*pcs != 10)) {
	    c = *pcs;
	    if (ftp->rpD == 0) ftp->rpHCode = c - 0x30;
//		if (c > 241) printf("IAC=%x", c);
	    ftp->rpD++;
	    if ((ftp->rpD < 4) && (c >= '0') && (c <= '9')) ftp->rpCode = ftp->rpCode * 10 + (c - 0x30);
	    if ((ftp->rpPflag == 1) && (ftp->rpD > 4) && (c >= '0') && (c <= '9')) ftp->rpPflag = 2;
	    if (ftp->rpPflag == 2) {
		if ((c != 13) && (c != ')')) {
		    if (ftp->rpPx < sizeof(ftp->pasv)) {
			ftp->pasv[ftp->rpPx] = c;
			ftp->rpPx++;
		    }
		} else {
		    ftp->pasv[ftp->rpPx] = 0;
		    ftp->rpPflag = 3;
		}
	    }
	    if (ftp->rpD == 4) {
		if ((ftp->rpPflag == 0) && (ftp->rpCode == 227)) ftp->rpPflag = 1;
		ftp->rpOcode = ftp->rpCode;
		if (c == '-') {
		    if (ftp->rpConti > 0) ftp->rpCode = 0;
		    ftp->rpConti++;
		}
	    }
	    if (ftp->rpIdx < sizeof(ftp->reply_str)-1) {
		*pcd = c;
		pcd++; ftp->rpIdx++;
	    }
	    pcs++;
	    lg--;
	}

	if (*pcs == 10) {
	    pcs++;
	    lg--;
	    *pcd = 0;
	    if (ftp->rpConti == 1) {
		if ((ftp->rpCode == 213) && ftp->pcSTAT) {
		    xfree(ftp->pcSTAT);
		    ftp->pcSTAT = NULL;
		}
	    }
	    if ((ftp->rpOrgCode == 213) && (ftp->rpCode == 0)) {
		ppcchange(&ftp->pcSTAT, ftp->reply_str);
		pc = ftp->pcSTAT;
		while (*pc) {
		    if (*pc == 13) {*pc = 10; break;}
		    pc++;
		}
		if (*pc == 0) *pc = 10;
	    }

		pc = ftp->reply_str;
		while (*pc) {
		    if (*pc == 13) {*pc = 0; break;}
		    pc++;
		}

// fprintf(stderr, ">%d< %s\n",ftp->rpConti, &ftp->reply_str[0]);
// fflush(stderr);

	    if ((ftp->rpConti > 0) && (ftp->rpCode != ftp->rpOrgCode)) {
		if (ftp->rpOrgCode == 0) ftp->rpOrgCode = ftp->rpCode;
		ftp->rpIdx = 0;
		ftp->rpD = 0; ftp->rpCode = 0; ftp->rpOcode = 0;
		Tftp_DoFtpLineRecv(ftp, ftp->rpOrgCode, False);
		goto loop;
	    }
	    Tftp_DoFtpLineRecv(ftp, ftp->rpCode, True);
	    ftp->rpD = 0;  ftp->rpIdx = 0; ftp->rpCode = 0; ftp->rpOcode = 0;
	    ftp->rpOrgCode = 0;
	    ftp->rpConti = 0;
	    ftp->rpPflag = 0;
	    ftp->rpPx = 0;
	    ftp->prCpend = 0;
	    if (lg > 0) goto loop;
	}
}

bool TFtp_Open(TFtp *ftp) {
  int lg, n, e, eloop;
  socklen_t solg;
  char *pc;
  struct hostent hostbuf, *hte;
  Tfd_list *fd_list;

	if (!ftp->gp.URL) return False;

	if (inet_aton(ftp->gp.URL, &ftp->hisctladdr.sin_addr)) {
	    ftp->hisctladdr.sin_family = AF_INET;
	} else {
	    lg = 1024;
	    pc = xmalloc(lg);
	    n = 0;
	    eloop = 2; 
loop_ghn:
	    while ((n = gethostbyname_r(ftp->gp.URL, &hostbuf, pc, lg, &hte, &e)) == ERANGE) {
		lg *= 2;
		pc = xrealloc(pc, lg);
	    }
	    eloop--;
	    if ((eloop > 0) && (hte == NULL)) goto loop_ghn;

	    if (n || hte == NULL) {
		xfree(pc);
		TFtp_Error(ftp, ioErr_Unknown_host, "Unknown host \"%s\"\n", ftp->gp.URL);
		if (ftp->DoStrgMSG) ftp->DoStrgMSG(ftp, hSHStM_Connect+hSHStM_Error);
		return False;
	    }
	    ftp->hisctladdr.sin_family = hte->h_addrtype;
	    pnmove (&ftp->hisctladdr.sin_addr, hte->h_addr_list[0], hte->h_length);
	    xfree(pc);
	}

	ftp->sock = socket(ftp->hisctladdr.sin_family, SOCK_STREAM, 0);
	if (ftp->sock < 0) {
	    TFtp_Error(ftp, ioErr_socket_failed, "socket failed \"%s\"\n", ftp->gp.URL);
	    goto Err;
	}
	ftp->hisctladdr.sin_port = htons(ftp->port);
	if (connect(ftp->sock, &ftp->hisctladdr, sizeof(struct sockaddr_in)) < 0) {
	    TFtp_Error(ftp, ioErr_connection_failed, "Failed to create a connection \"%s\"\n", ftp->gp.URL);
	    goto Err;
	}
	solg = sizeof(struct sockaddr_in);
	if (getsockname(ftp->sock, &ftp->myctladdr, &solg) < 0) {
	    TFtp_Error(ftp, ioErr_Internal, "getsockname failed \"%s\"\n", SysError_pc());
	    goto Err;
	}

	ftp->pin = fdopen(ftp->sock, "r");
	if (ftp->pin == NULL) {
	    TFtp_Error(ftp, ioErr_Internal, "fdopen r failed \"%s\"\n", SysError_pc());
	    goto Err;
	}
	ftp->pout = fdopen(ftp->sock, "w");
	if (ftp->pout == NULL) {
	    fclose(ftp->pin);
	    TFtp_Error(ftp, ioErr_Internal, "fdopen w failed \"%s\"\n", SysError_pc());
	    goto Err;
	}

	ftp->curtype = TYPE_A;
#ifdef USE_uC
	fd_list = AddPipeProc_TmpRes(fileno(ftp->pin), InPipe, TFtp_EvReply, ftp);
#else
	fd_list = AddPipeProc_TmpRes(ftp->pin->_fileno, InPipe, TFtp_EvReply, ftp);
#endif
	fd_list->io = ftp->pin;
	ftp->prCpend = 1;
	ftp->isConnect = True;

	return True;

Err:
	if (ftp->DoStrgMSG) ftp->DoStrgMSG(ftp, hSHStM_Connect+hSHStM_Error);
	if (ftp->sock > 0) close(ftp->sock);
	ftp->sock = 0;
	return False;
}

void TFtp_Close(TFtp *ftp) {
	if (ftp->pin) {
	    ClosePipeProc(fileno(ftp->pin), clpp_io);
	    ftp->pin = NULL;
	}
	if (ftp->pout) {
	    ClosePipeProc(fileno(ftp->pout), clpp_io);
	    ftp->pout = NULL;
	}
	if (ftp->sock > 0) close(ftp->sock);
	ftp->sock = 0;
	TFtp_closeconn(ftp);
	TFtp_ClsReply(ftp);
	ppcfree(&ftp->pcSTAT);
	ftp->isConnect = False;
}

bool TFtp_Command(TFtp *ftp, const char *fmt, ...) {
  va_list args;
  char line[2048];

	TFtp_ClsReply(ftp);
	va_start(args, fmt);
	vsnprintf(line, 2048, fmt, args);
	va_end(args);
//	fprintf(ftp->pout, "\r\n");
// fprintf(stderr, "ahftp: ahftp_SendErrLoginToSlave\n");
// fflush(stdout);
	printf(".%s\n", line);
	if (ftp->pout) {
	    fprintf(ftp->pout, "%s\r\n", line);
	    fflush(ftp->pout);
	    return True;
	} else {
	    TFtp_Error(ftp, ioErr_not_connect, "not connect\n");
	    return False;
	}
}

typedef struct {
  const char *s;
  const char *c;
  int t;
} TTransType;

const TTransType TransTypeTab[] = {
  {"ascii", "A", TYPE_A},
  {"binary", "I", TYPE_I},
  {"image", "I", TYPE_I},
  {"ebcdic", "E", TYPE_E},
  {NULL, NULL, 0}
};

int TFtp_changetype(TFtp *ftp, int newtype) {
  int i;
	if (!newtype) newtype = TYPE_I;
	if (newtype == ftp->curtype) return 2;
	i = 0;
	while (TransTypeTab[i].s) {
	    if (TransTypeTab[i].t == newtype) {
		ftp->newtype = newtype;
		TFtp_Command(ftp, "TYPE %s", TransTypeTab[i].c);
		return 1;
	    }
	    i++;
	}
	TFtp_Error(ftp, ioErr_Internal, "internal error: unknown type %d\n", newtype);
	return 0;
}

bool TFtp_initconn(TFtp *ftp) {
  int i, _on;
  socklen_t solg;
  int ip[4];
  int po[2];
  byte *pc;

	_on = 1;
	ftp->data_addr = ftp->myctladdr;
	if (ftp->sendport) ftp->data_addr.sin_port = 0;

	TFtp_closeconn(ftp);
	ftp->data = socket(AF_INET, SOCK_STREAM, 0);
	if (ftp->data < 0) {
	    TFtp_Error(ftp, ioErr_socket_failed, "socket failed (TFtp_initconn)");
	    return False;
	}

	if (ftp->passivemode) {
	    TFtp_Command(ftp, "PASV");
	    return True;
	}

	if (!ftp->sendport) {
	    if (setsockopt(ftp->data, SOL_SOCKET, SO_REUSEADDR, &_on, sizeof(_on)) < 0) {
		TFtp_Error(ftp, ioErr_setsockopt, "setsockopt failed (TFtp_initconn)");
		goto Err;
	    }
	}

	if (bind(ftp->data, &ftp->data_addr, sizeof(ftp->data_addr)) < 0) {
	    TFtp_Error(ftp, errno, "bind failed \"%s\"\n", SysError_pc());
	    goto Err;
	}

	solg = sizeof(struct sockaddr_in);
	if (getsockname(ftp->data, &ftp->data_addr, &solg) < 0) {
	    TFtp_Error(ftp, errno, "getsockname failed \"%s\"\n", SysError_pc());
	    goto Err;
	}

	if (listen(ftp->data, 1) < 0) {
	    TFtp_Error(ftp, errno, "listen failed \"%s\"\n", SysError_pc());
	}

	if (ftp->sendport) {
	    pc = (byte*)&ftp->data_addr.sin_addr;
	    for (i=0; i<4; i++) {ip[i] = *pc; pc++;}
	    pc = (byte*)&ftp->data_addr.sin_port;
	    for (i=0; i<2; i++) {po[i] = *pc; pc++;}
	    TFtp_Command(ftp, "PORT %d,%d,%d,%d,%d,%d",
		    ip[0], ip[1], ip[2], ip[3], po[0], po[1]);
	}

	return True;
Err:
	TFtp_closeconn(ftp);
	return False;
}

bool TFtp_initconn_pasv(TFtp *ftp) {
  int i, m, sf = 0;
  dword iadr = 0;
  const char *pc;

	Fill0(&ftp->data_addr, sizeof (ftp->data_addr));
	ftp->data_addr.sin_family = AF_INET;
	pc = ftp->pasv;
	for (i=4;i;i--) {
	    m = ppctoint_scan(&pc, True);
	    iadr |= ((m & 0xff) << sf);
	    sf += 8;
	}
	ftp->data_addr.sin_addr.s_addr = iadr;
	iadr = sf = 0;
	for (i=2;i;i--) {
	    m = ppctoint_scan(&pc, True);
	    iadr |= ((m & 0xff) << sf);
	    sf += 8;
	}
	ftp->data_addr.sin_port = iadr;
	if (connect(ftp->data, &ftp->data_addr, sizeof(struct sockaddr_in)) < 0) {
	    TFtp_Error(ftp, ioErr_connection_failed, "Failed to create a connection \"%s\"\n", ftp->gp.URL);
	    TFtp_closeconn(ftp);
	    return False;
	}
	return True;
}

void TFtp_closeconn(TFtp *ftp) {
// fprintf(stderr, "TFtp_closeconn: %d 0x%x 0x%x\n", ftp->data, (int)ftp->datain, (int)ftp->dataout);

	if (ftp->data > 0) {
	    close(ftp->data);
	    ClosePipeProc(ftp->data, 0);
	}
	ftp->data = -1;
}

bool TFtp_dataconn(TFtp *ftp) {
  int m;
  socklen_t solg;
  struct sockaddr_in from;

	ftp->TransfStart_us = GetTime_us64();
	ftp->TransfDatLg = 0;
	ftp->TransfNxEv_us = 50000;

	if (ftp->passivemode) return True;

	solg = sizeof(struct sockaddr_in);
	m = accept(ftp->data, &from, &solg);
	if (m <= 0) TFtp_Error(ftp, errno, "accept failed \"%s\"\n", SysError_pc());
	TFtp_closeconn(ftp);
	if (m <= 0) return False;
	ftp->data = m;
	return True;
}

/*_____________________________ Destroy _________________________________*/
static bool TFtp__Destroy(TFtp *self) {
	TFtp_Close(self);
	xfree(self->lastErrStr);
	return False;
}

/*_____________________________ Create __________________________________*/
TFtp *TFtp_Create() {
  TFtp *self;

	self = ttc_create(sizeof(TFtp), wtyp_TFtp);
	self->port = 21;
	self->data = -1;
	self->sendport = True;
	self->DoError = TFtp_DoError;
	self->passivemode = True;
	self->destroy = (TDESTROY)TFtp__Destroy;
	return self;
}
