/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"

#include "tmime_pv.h"

#ifdef _OS_WIN32
static bool gp_path_cmp(TglobPath *gp, const char *path) {
  char *pc;
  bool result;
// printf("%s(%d): %s\n", FL, path);
	pc = gptopc(gp, gptfu_noName);
// printf("%s(%d): %s\n", FL, pc);
	result = path_cmp(pc, path);
	xfree(pc);
	return result;
}
#else
static bool gp_path_cmp(TglobPath *gp, const char *path) {
	return pccmp(gp->path, path);
}
#endif


static TMimeDat_x **md2hash_pp(TMime *self, void *md) {
  TMime_private *pv;
  int idx;
	idx = (intptr_t)md + ((intptr_t)md >> 16);
	idx += idx >> 8;
	idx = (idx + (idx >> 4)) & 0x0f;
	pv = self->private;
	return &pv->md_hashtable[idx];
}

static void md_add_to_hashtable(TMime *self, void *md) {
  TMimeDat_x **ppmd;

	if (!md) return;
	ppmd = md2hash_pp(self, md);
	while (*ppmd) ppmd = &(*ppmd)->next_hash;
	*ppmd = md;
}

static void md_clr_from_hashtable(TMime *self, void *md) {
  TMimeDat_x **ppmd;

	ppmd = md2hash_pp(self, md);
	while (*ppmd && (*ppmd != md)) ppmd = &(*ppmd)->next_hash;
	if (!*ppmd) return;
	*ppmd = (*ppmd)->next_hash;
}

TMimeDat_x *TMime_get_md(TMime *self, void *md) {
  TMimeDat_x *pmd;
	if (!md) return NULL;
	pmd = *md2hash_pp(self, md);
	while (pmd != md) pmd = pmd->next_hash;
	return pmd;
}

static TMimeObj* TMimeObj_find_path(TMime *self, char *path) {
  TMimeObj *mo, **ppmo;
  TMime_private *pv = self->private;

	ppmo = pv->mos;
	while (*ppmo) {
	    mo = *ppmo;
	    while (mo) {
	      TDirListObjNy *ObjNy = mo->ObjNy;
	      if (ttc_chk_struct(ObjNy)) {
		if (!ObjNy->dl) {
 printf("%s(%d): ERROR mo->ObjNy->dl == NULL ???\n", FL);
		} else {
		    if (gp_path_cmp(&ObjNy->dl->gp, path)) return mo;
		}
	      }
	      mo = mo->next;
	    }
	    ppmo++;
	}
	return NULL;
}
/*_____________________________ free ____________________________________*/
static void free_md_str(char *pc, char *arr) {
	if (pc && (pc != arr)) xfree(pc);
}

static void TMimeDat_free(TMime *self, TMimeDat_x *md) {

	md_clr_from_hashtable(self, md);
	free_md_str(md->fn, md->bfn);
	free_md_str(md->icon, md->bicon);
	free_md_str(md->mimetype, md->bmimetype);

	if (md->type == ddkt_Application) {
  TMimeDat_app *mda = (TMimeDat_app*)md;
	  free_md_str(mda->exec, mda->bexec);
	} else if (md->type == ddkt_Mime) {
  TMimeDat_type *mdt = (TMimeDat_type*)md;
	    free_md_str(mdt->pattern, mdt->bpattern);
	}
	xfree(md);
}

static void TMimeObj_free(TMimeObj *mo) {
  TMimeDat_x *md;
	while (mo->mda) {
	    md = (TMimeDat_x*)mo->mda;
	    mo->mda = (TMimeDat_app*)md->next;
	    TMimeDat_free(mo->mime_self, md);
	}
	while (mo->mdt) {
	    md = (TMimeDat_x*)mo->mdt;
	    mo->mdt = (TMimeDat_type*)md->next;
	    TMimeDat_free(mo->mime_self, md);
	}
	xfree(mo);
}

/*_____________________________ dflEv __________________________________*/
static void set_md_str(char **ppc, char *pcarr, const char *str, int max) {
  int lg;

	free_md_str(*ppc, pcarr);
	*ppc = NULL;
	lg = pclen(str);
	if (!lg) return;

	if (lg < max) {
	    *ppc = pcarr;
	    pcnmove(pcarr, str, lg);
	} else {
	    *ppc = pcdup(str);
	}
}

static void set_mimedat_x(TMimeDat_x *mdx, TpcIni *ini, TDesktopHead *ddkh) {
  const char *pc;
	mdx->stat &= ~(ME_Pic16_NoFound|ME_Pic32_NoFound);
	mdx->pic16 = NULL;
	mdx->pic32 = NULL;
	set_md_str(&mdx->icon, mdx->bicon, ddkh->icon, bicon_max_size);
	pc = TpcIni_RD(ini, "Desktop Entry", "MimeType");
	set_md_str(&mdx->mimetype, mdx->bmimetype, pc, bmimetype_max_size);
}

static void set_mimedat_app(TMimeDat_app *mda, TpcIni *ini, TDesktopHead *ddkh) {
	set_mimedat_x((TMimeDat_x*)mda, ini, ddkh);
	set_md_str(&mda->exec, mda->bexec, ddkh->exec, bexec_max_size);
}

static void set_mimedat_type(TMimeDat_type *mdt, TpcIni *ini, TDesktopHead *ddkh) {
  const char *pc;
	set_mimedat_x((TMimeDat_x*)mdt, ini, ddkh);
	pc = TpcIni_RD(ini, "Desktop Entry", "Patterns");
// printf("%s(%d): %s %s\n", FL, ini->FileName, pc);
	set_md_str(&mdt->pattern, mdt->bpattern, pc, bpattern_max_size);
}

static void add_mime(TDirListObjNy *self, TDirFileList *dfl) {
  char *fn, *deskfn;
  TpcIni *ini;
  TMimeObj *mo = self->tagp;
  TMimeDat_x *mdx;
  TDesktopHead ddkh;

	deskfn = ChangeFileExt(dfl->fn, NULL);
	fn = gptofnpc(&self->dl->gp, gptfu_noName, dfl->fn);

	ini = TpcIni_Create(fn);
	xfree(fn);
	dotdesktop_get_head(ini, &ddkh);

	mdx = NULL;
	if (ddkh.ddkt == ddkt_Application) {
	    if (ddkh.exec) {
#if 0
		const char *pc = ddkh.exec;
		while (*pc) {
		    if (*pc == '%') {
			pc++;
			if ((*pc == 'f') || (*pc == 'F') ||
			    (*pc == 'u') || (*pc == 'U')) break;
		    }
		    pc++;
		}
		if (*pc) {
#endif
		    mdx = xcalloc(sizeof(TMimeDat_app));
		    mdx->next = (TMimeDat_x*)mo->mda;
		    mo->mda = (TMimeDat_app*)mdx;
		    mdx->type = ddkt_Application; //mdtype_app;
		    set_mimedat_app((TMimeDat_app*)mdx, ini, &ddkh);
		}
//	    }

	} else if (ddkh.ddkt == ddkt_Mime) {
	    mdx = xcalloc(sizeof(TMimeDat_type));
	    mdx->next = (TMimeDat_x*)mo->mdt;
	    mo->mdt = (TMimeDat_type*)mdx;
	    mdx->type = ddkt_Mime; //mdtype_type;
	    set_mimedat_type((TMimeDat_type*)mdx, ini, &ddkh);
	}

	if (mdx) {
	    mdx->dfl = dfl;
	    set_md_str(&mdx->fn, mdx->bfn, deskfn, bfn_max_size);
	    md_add_to_hashtable(self->owner, mdx);
	}

	Tfree(ini);
	xfree(deskfn);
}

static void TMime_OnFileMsg(TFileMsg *fmsg);

static bool TMime_add_mime_path(TMime *self, TMimeObj **ppmo, const char *path1,
				const char *path2, TDirFileList *dfl) {
  TDirListObjNy *ObjNy;
  char *path;
  word pattft[4] = {stft_desktop, stft_Txt, 0, 0};
  TMimeObj *mo;
  int lg;
  bool res = False;

	path = pcadd(path1, path2);
	if (!path) return False;
	lg = pclen(path);
	if (!lg) return False;
	if (TestNoSlash(path[lg-1])) ppcadd(&path, "/");
#ifdef _OS_WIN32
	slash_korr(path, True);
#endif

	if (!TMimeObj_find_path(self, path) && TestIsPath(path)) {
	    ObjNy = TDirListObjNy_Create(self);
	    ObjNy->findAtr = findIsRootPath | findDIR_NoPattern;
	    ObjNy->OnFileMsg = TMime_OnFileMsg;
	    TPattern_set(&ObjNy->patt, NULL, pattft);
	    mo = xcalloc(sizeof(TMimeObj));
	    ObjNy->tagp = mo;
	    mo->mime_self = self;
	    mo->ObjNy = ObjNy;
	    mo->dfl = dfl;
	    mo->ppmo_root = ppmo;
	    while (*ppmo) ppmo = &(*ppmo)->next;
	    *ppmo = mo;
	    res = TDirListObjNy_Open(ObjNy, path);
	    if (!res) {
		*ppmo = NULL;
		Tfree(ObjNy);
		xfree(mo);
	    }
	}
	xfree(path);
	return res;
}

static void add_mime_path(TDirListObjNy *self, TDirFileList *dfl) {
  char *pc;
  const char *cpc;
  TMimeObj *mo, *nmo;

	mo = self->tagp;
//**************  TODO chk link Loop ***********************
	if (dfl->st->st & st_Link) {
	    cpc = Get_ExFileDat_pc(dfl->st, exfdt_LinkName);
	    pc = path_rel2abs(cpc, self->dl->gp.path);
//  printf("%s(%d): %s\n", FL, pc);
	    nmo = *mo->ppmo_root;
	    while (nmo) {
//  printf("%s(%d): %s\n", FL, nmo->ObjNy->dl->gp.path);
		if (gp_path_cmp(&nmo->ObjNy->dl->gp, pc)) {
		    xfree(pc);
		    return;
		}
		nmo = nmo->next;
	    }
	    xfree(pc);
	}
#ifdef _OS_WIN32
	pc = gptopc(&self->dl->gp, gptfu_noName);
	TMime_add_mime_path(self->owner, mo->ppmo_root, 
			    pc, dfl->fn, dfl);
	xfree(pc);
#else
	TMime_add_mime_path(self->owner, mo->ppmo_root, 
			    self->dl->gp.path, dfl->fn, dfl);
#endif
}


static void add_mime_overwrite(TDirListObjNy *self) {
  TMime *mime_self = self->owner;
  TMime_private *pv = mime_self->private;

	if (!gp_path_cmp(&self->dl->gp, pv->path_overwrite)) return;
// printf("%s(%d): dflEv_default Add Mime %s\n", FL, dfl->fn);

	Tfree(pv->overwrite_ini);
	pv->overwrite_ini = TpcIni_Create(pv->fn_mime_overwrite);
}

/*_____________________________ dflEv Change ____________________________*/
static void change_ObjNy_mime(TDirListObjNy *self, TDirFileList *dfl) {
  char *fn;
  TMimeObj *mo = self->tagp;
  TMime *mime_self = self->owner;
  TMimeDat_x *md;
  TpcIni *ini;
  TDesktopHead ddkh;
  bool mk = False;

	md = (TMimeDat_x*)mo->mda;
	while (md && (md->dfl != dfl)) md = md->next;

	if (!md) {
	    md = (TMimeDat_x*)mo->mdt;
	    while (md && (md->dfl != dfl)) md = md->next;
	}
	if (!md) return;

	fn = gptofnpc(&self->dl->gp, gptfu_noName, dfl->fn);
// printf("%s(%d): %s\n", FL, fn);
	ini = TpcIni_Create(fn);
	xfree(fn);
	dotdesktop_get_head(ini, &ddkh);

	if (ddkh.ddkt == ddkt_Application) {
	    if (md->type == ddkt_Application) {
		set_mimedat_app((TMimeDat_app*)md, ini, &ddkh);
		mk = True;
	    }

	} else if (ddkh.ddkt == ddkt_Mime) {
	    if (md->type == ddkt_Mime) {
		set_mimedat_type((TMimeDat_type*)md, ini, &ddkh);
		mk = True;
	    }
	}
	Tfree(ini);
	if (mk && mime_self->OnChange) mime_self->OnChange(mime_self, MimeOnChange_item);
}

/*_____________________________ dflEv Del _______________________________*/
static TMimeDat_x **get_ppmd_dfl(TMimeDat_x **ppmd, TDirFileList *dfl) {
	while (*ppmd && ((*ppmd)->dfl != dfl)) ppmd = &(*ppmd)->next;
	return ppmd;
}

static void del_ObjNy_mime(TDirListObjNy *self, TDirFileList *dfl) {
  TMimeObj *mo = self->tagp;
  TMimeDat_x *md, **ppmd;

	ppmd = get_ppmd_dfl((void*)&mo->mda, dfl);
	if (!*ppmd) ppmd = get_ppmd_dfl((void*)&mo->mdt, dfl);
	if (!*ppmd) {
#ifdef POINTER64BIT
	    printf("%s(%d): ERROR: del_mime  TMimeDat 0x%lx not found ???\n", FL, (int64_t)*ppmd);
#else
	    printf("%s(%d): ERROR: del_mime  TMimeDat 0x%x not found ???\n", FL, (int)*ppmd);
#endif
	    return;
	}
	md = *ppmd;
	*ppmd = md->next;
	TMimeDat_free(self->owner, md);
}

static void del_ObjNy_mimeobj(TDirListObjNy *self) {
  TMimeObj *mo, **ppmo;

	mo = self->tagp;
	ppmo = mo->ppmo_root;
	while (*ppmo && (*ppmo != mo)) ppmo = &(*ppmo)->next;
	if (!*ppmo) {
#ifdef POINTER64BIT
	    printf("%s(%d): ERROR: del_mimeobj  TMimeObj 0x%lx not found ???\n", FL, (int64_t)*ppmo);
#else
	    printf("%s(%d): ERROR: del_mimeobj  TMimeObj 0x%x not found ???\n", FL, (int)*ppmo);
#endif
	    return;
	}
	*ppmo = mo->next;
	TMimeObj_free(mo);
}


static void del_mime_overwrite(TDirListObjNy *self) {
  TMime *mime_self = self->owner;
  TMime_private *pv = mime_self->private;

	if (!pv->overwrite_ini || !gp_path_cmp(&self->dl->gp, pv->path_overwrite)) return;
	Tfree(pv->overwrite_ini);
	pv->overwrite_ini = NULL;
}

/*_____________________________ OnFileMsg _______________________________*/
static void TMime_OnFileMsg(TFileMsg *fmsg) {
  TDirListObjNy *self;
  TDirFileList *dfl;
  int bef, befn;
  TMime *mime_self;
  TMime_private *pv;
  int ev_bef = 0;

	if (!file_msg_chk_wtyp(fmsg, wtyp_TDirListObjNy, FL)) return;

	if (fmsg->art != fmArt_DirList) return;
	if (fmsg->typD != fmTyp_int) return;

	self = fmsg->obj;
	mime_self = self->owner;
	pv = mime_self->private;

	bef = fmsg->D.m;
	befn = bef & dlbef_BefMask;

// if (dfl) printf("%s(%d): dflEv_mo_exec bef=0x%x %s\n", FL, bef, dfl->fn); else printf("%s(%d): dflEv_mo_exec bef=0x%x\n", FL, bef);
	if (befn == dlbef_DelSelf) {
// printf("%s(%d): dflEv_default DelSelf  %s\n", FL, self->dl->gp.path);
	    del_ObjNy_mimeobj(self);
	    ev_bef = MimeOnChange_all;

	} else if (fmsg->typF == fmTyp_dfl) {
	    fmsg->Ok = True;
	    dfl = fmsg->F.dfl;

	    switch (befn) {

	    case dlbef_Add:
		if (dfl->st->st & st_Dir) {
// printf("%s(%d): dflEv_default Add Dir  %s%s\n", FL, self->dl->gp.path, dfl->fn);
		    bool Ok = !gp_path_cmp(&self->dl->gp, pv->path_overwrite) ||
			      pccmp (dfl->fn, "applications") ||
			      pccmp (dfl->fn, "mimelnk");
		    if (Ok) {
// printf("%s(%d): dflEv_default Add Dir  %s%s\n", FL, self->dl->gp.path, dfl->fn);
			add_mime_path(self, dfl);
			if (bef & dlbef_NotifyRun) ev_bef = MimeOnChange_all;
		    }
		} else if (dfl->st->FileType == stft_desktop) {
// printf("%s(%d): dflEv_default Add Mime %s\n", FL, dfl->fn);
		    add_mime(self, dfl);
		    if (bef & dlbef_NotifyRun) ev_bef = MimeOnChange_all;
		} else if (pccmp(dfl->fn, "mime_overwrite")) {
		    if (bef & dlbef_NotifyRun) {
// printf("%s(%d): dflEv_default Add Mime add_mime_overwrite %s\n", FL, dfl->fn);
			add_mime_overwrite(self);
			ev_bef = MimeOnChange_all;
		    }
		}
		break;

	    case dlbef_Change:
		if (dfl->st->FileType == stft_desktop) {
		    change_ObjNy_mime(self, dfl);

		} else if (pccmp(dfl->fn, "mime_overwrite")) {
		    add_mime_overwrite(self);
		    ev_bef = MimeOnChange_all;
		}
		break;

	    case dlbef_Del:
		if (dfl->st->FileType == stft_desktop) {
// printf("%s(%d): dflEv_default Del  %s\n", FL, dfl->fn);
		    del_ObjNy_mime(self, dfl);
		    ev_bef = MimeOnChange_all;
		} else if (pccmp(dfl->fn, "mime_overwrite")) {
		    del_mime_overwrite(self);
		    ev_bef = MimeOnChange_all;
		}
		break;
	    }
	}
	if (ev_bef && mime_self->OnChange) {
	    mime_self->OnChange(mime_self, ev_bef);
	}
}

/*_____________________________ Export __________________________________*/
const char *TMime_get_fn_mime_overwrite(TMime *self) {
  TMime_private *pv = self->private;
	return pv->fn_mime_overwrite;
}

/*_____________________________ Destroy _________________________________*/
static bool TMime__Destroy(TMime *self) {
  TMime_private *pv = self->private;
  TMimeObj **ppmo, *mo;

	ppmo = pv->mos;
	while (*ppmo) {
	    while (*ppmo) {
		mo = *ppmo;
		*ppmo = mo->next;
		Tfree(mo->ObjNy);
		TMimeObj_free(mo);
	    }
	    ppmo++;
	}
	xfree(pv->mos);
	xfree(pv->path_overwrite);
	xfree(pv->fn_mime_overwrite);
	Tfree(pv->overwrite_ini);
	xfree(pv);
	return False;
}

/*_____________________________ Create __________________________________*/
/* ----------------------------------
mime find
1. ~/.ALDccp/share/mimelnk/
2. /usr/share/mimelnk/
3. /opt/aldccp/share/mimelnk/
4. ~/.ALDccp/share/applications
5. ~/.ALDccp/Menu/OpenWith/
6. ~/.ALDccp/Menu/StartMenu/
7. /usr/share/applications
8. /opt/aldccp/share/applications
---------------------------------- */

TMime *TMime_Create() {
  int i;
  TMime *self;
  TMime_private *pv;
  char *pc, *ddPath, *token;
  const char *indata;
  TMimeObj **ppmo;
#ifndef _OS_WIN32
  TpcIni *ini;
#endif

	self = ttc_create(sizeof(TMime), wtyp_TstatMime);
	self->destroy = (TDESTROY)TMime__Destroy;
	pv = xcalloc(sizeof(TMime_private));
	self->private = pv;
	self->stat_set_mime_id = TMime_stat_set_mime_id;


	pv->path_overwrite = pcadd(appDat.ALDini_dir, "/share/");
#ifdef _OS_WIN32
	slash_korr(pv->path_overwrite, True);
#endif
	pv->fn_mime_overwrite = pcadd(pv->path_overwrite, "mime_overwrite");
	if (TestIsFile(pv->fn_mime_overwrite)) {
	    pv->overwrite_ini = TpcIni_Create(pv->fn_mime_overwrite);
	}

#ifdef _OS_WIN32
	ddPath = pcadd_n(3, "\"", appDat.ALDini_dir, "/applications\"");
	slash_korr(ddPath, True);
#else
	ini = TpcIni_Create(appDat.inifile_global);
	ddPath = pcadd_n(3, 
			TpcIni_RD_pc(ini, "App", "dotdesktop_mimelnk", "/usr/share/mimelnk"),
			" ",
			TpcIni_RD_pc(ini, "App", "dotdesktop", "/usr/share/applications"));
// printf("%s(%d): %s\n", FL, ddPath);
	Tfree(ini);

	ppcadd_n(&ddPath, 5, " ", ALDccpPREFIX, "/share/mimelnk ", 
				  ALDccpPREFIX, "/share/applications");
#endif
// printf("%s(%d): %s\n", FL, ddPath);

	indata = ddPath;
	i = 2;
	while (indata) {
	    indata = GetToken(indata, NULL);
	    i++;
	}
	ppmo = xcalloc(sizeof(void*) * i);
	pv->mos = ppmo;

	if (TMime_add_mime_path(self, ppmo, pv->path_overwrite, NULL, 
				NULL)) ppmo++;

	indata = ddPath;

	while (indata) {
	    indata = GetToken(indata, &token);
//  printf("%s(%d): %s\n", FL, token);
	    if (*token == '~') {
		pc = pcadd(appDat.sHOME, token+1);
		xfree(token);
		token = pc;
//  printf("%s(%d): %s\n", FL, token);
	    }
	    if (TMime_add_mime_path(self, ppmo, token, NULL, 
				NULL)) ppmo++;

	    xfree(token);
	}
	xfree(ddPath);

	return self;
}


