/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <errno.h>
#include "net_xxx.h"

#define hIDX_chmod  2
#define hIDX_utime  3
#define hIDX_rename 4
#define hIDX_mv     5
#define hIDX_unlink 6
#define hIDX_rmdir  7
#define hIDX_mkdir  8

static ThNetToolsLoader **hNetToolsLd = NULL;

static bool net_gp_test(TglobPath *gp, int idx) {
  ThNetToolsLoader *hLD;

	if (!gp_test_local_drive(gp)) {
	    hLD = GetNet4_hLD((void*)hNetToolsLd, gp, idx, 0);
	    if (!hLD) return False;
	}
	return True;
}

static void *Get_hLD(ThNetLoader4 **hLD, TglobPath *gp, int y, TNetErrRet *ErrRet) {
  int m = 0;
	if (!ErrRet) m = hNet_appERR;
	hLD = GetNet4_hLD((void*)hNetToolsLd, gp, y, m);
	if (!hLD) {
	    if (ErrRet) ErrRet->err = pcadd_n(3, "no \"", 
				gp->RootArt, "\" net_loader found !!!");
	    errno = EPERM;
	}
	return hLD;
}

/* _______________________ mkdir ________________________________________ */
bool net_mkdir_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode) {
  ThNetToolsLoader *hLD;

	if (gp_test_local_drive(gp)) {
	    return file_mkdir_gp(gp, ErrRet, mode);

	} else {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_mkdir, ErrRet);
	    if (!hLD) return False;
	    return hLD->mkdir(gp, ErrRet, mode);
	}
}

bool net_mkdir(const char *fn, TNetErrRet *ErrRet, int mode) {
  TglobPath *gp = pctogp(fn);
	bool result = net_mkdir_gp(gp, ErrRet, mode);
	gpFree(gp);
	return result;
}

bool net_mkdir_gp_test(TglobPath *gp) {
	return net_gp_test(gp, hIDX_mkdir);
}

/* _______________________ rmdir ________________________________________ */
bool net_rmdir_gp(TglobPath *gp, TNetErrRet *ErrRet) {
  ThNetToolsLoader *hLD;

	if (gp_test_local_drive(gp)) {
	    return file_unlink_gp(gp, ErrRet);

	} else {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_rmdir, ErrRet);
	    if (!hLD) return False;
	    return hLD->rmdir(gp, ErrRet);
	}
}

bool net_rmdir(const char *fn, TNetErrRet *ErrRet) {
  TglobPath *gp = pctogp(fn);
	bool result = net_rmdir_gp(gp, ErrRet);
	gpFree(gp);
	return result;
}

bool net_rmdir_gp_test(TglobPath *gp) {
	return net_gp_test(gp, hIDX_rmdir);
}

/* _______________________ unlink _______________________________________ */
bool net_unlink_gp(TglobPath *gp, TNetErrRet *ErrRet) {
  ThNetToolsLoader *hLD;

	if (gp_test_local_drive(gp)) {
	    return file_unlink_gp(gp, ErrRet);

	} else {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_unlink, ErrRet);
	    if (!hLD) return False;
	    return hLD->unlink(gp, ErrRet);
	}
}

bool net_unlink(const char *fn, TNetErrRet *ErrRet) {
  TglobPath *gp = pctogp(fn);
	bool result = net_unlink_gp(gp, ErrRet);
	gpFree(gp);
	return result;
}

bool net_unlink_gp_test(TglobPath *gp) {
	return net_gp_test(gp, hIDX_unlink);
}

/* _______________________ rename _______________________________________ */
bool net_rename_gp(TglobPath *gp, TNetErrRet *ErrRet, const char *new_name) {
  ThNetToolsLoader *hLD;

	if (gp_test_local_drive(gp)) {
	    return file_rename_gp(gp, ErrRet, new_name);

	} else {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_rename, ErrRet);
	    if (!hLD) return False;
	    return hLD->rename(gp, ErrRet, new_name);
	}
}

bool net_rename(const char *fn, TNetErrRet *ErrRet, const char *new_name) {
  TglobPath *gp = pctogp(fn);
	bool result = net_rename_gp(gp, ErrRet, new_name);
	gpFree(gp);
	return result;
}

bool net_rename_gp_test(TglobPath *gp) {
	return net_gp_test(gp, hIDX_rename);
}

/* _______________________ mv __________________________________________ */
bool net_mv_gp(TglobPath *gp, TNetErrRet *ErrRet, TglobPath *new_gp) {
  ThNetToolsLoader *hLD;

	if (!gp || !new_gp) return False;
	if (gp_test_local_drive(gp) && gp_test_local_drive(new_gp)) {
	    return file_mv_gp(gp, ErrRet, new_gp);

	} else if (!casecmpA(gp->RootArt, new_gp->RootArt) &&
		   !casecmpA(gp->URL,     new_gp->URL)) {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_mv, ErrRet);
	    if (!hLD) return False;
	    return hLD->mv(gp, ErrRet, new_gp);

	} else {
	    if (ErrRet) ErrRet->err = pcdup("Cross-device link");
	    errno = EXDEV;
	    return False;
	}
}

bool net_mv(const char *fn, TNetErrRet *ErrRet, const char *new_fn) {
  TglobPath *gp     = pctogp(fn);
  TglobPath *new_gp = pctogp(new_fn);
	bool result = net_mv_gp(gp, ErrRet, new_gp);
	gpFree(new_gp);
	gpFree(gp);
	return result;
}

bool net_mv_gp_test(TglobPath *gp, TglobPath *new_gp) {
	return net_gp_test(gp, hIDX_mv) && net_gp_test(new_gp, hIDX_mv);
}

/* _______________________ utime ________________________________________ */
bool net_utime_gp(TglobPath *gp, TNetErrRet *ErrRet, time_t mtime) {
  ThNetToolsLoader *hLD;

	if (gp_test_local_drive(gp)) {
	    return file_utime_gp(gp, ErrRet, mtime);

	} else {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_utime, ErrRet);
	    if (!hLD) return False;
	    return hLD->utime(gp, ErrRet, mtime);
	}
}

bool net_utime(const char *fn, TNetErrRet *ErrRet, time_t mtime) {
  TglobPath *gp = pctogp(fn);
	bool result = net_utime_gp(gp, ErrRet, mtime);
	gpFree(gp);
	return result;
}

bool net_utime_gp_test(TglobPath *gp) {
	return net_gp_test(gp, hIDX_utime);
}

/* _______________________ chmod ________________________________________ */
bool net_chmod_gp(TglobPath *gp, TNetErrRet *ErrRet, int mode) {
  ThNetToolsLoader *hLD;

	if (gp_test_local_drive(gp)) {
	    return file_chmod_gp(gp, ErrRet, mode);

	} else {
	    hLD = Get_hLD((void*)hNetToolsLd, gp, hIDX_chmod, ErrRet);
	    if (!hLD) return False;
	    return hLD->chmod(gp, ErrRet, mode);
	}
}

bool net_chmod(const char *fn, TNetErrRet *ErrRet, int mode) {
  TglobPath *gp = pctogp(fn);
	bool result = net_chmod_gp(gp, ErrRet, mode);
	gpFree(gp);
	return result;
}

bool net_chmod_gp_test(TglobPath *gp) {
	return net_gp_test(gp, hIDX_chmod);
}

void NetLoaderReg_Tools(int n, ...) {
  va_list args;
	va_start(args, n);
	pp_void_va_add((void*)&hNetToolsLd, n, args);
	va_end(args);
}
