/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include <mntent.h>

typedef struct TMounts {
  struct TMounts *next;
  int stat;
  TNotifyMount_data nym;
} TMounts;

typedef struct TNotifyMount_wd {
  AlgTTC0
  struct TNotifyMount_wd *next;
  void *owner;
  ThNotifyMountEv hnymoEv;
} TNotifyMount_wd;


typedef struct {
  TNotifyMount_wd *wd;
  TEv_Array *appEv;
  dword app_us;
  TMounts *Mounts;
} TNotifyMount;

static TNotifyMount *nym;
static int ForceReScan;
static bool NotifyMountRun;
static bool NotifyMountStop;

static void SendEv(TMounts *mts, int bef) {
  TNotifyMount_wd *wd = nym->wd;
	mts->nym.bef = bef;
	while (wd) {
	    mts->nym.owner = wd->owner;
	    if (wd->hnymoEv) wd->hnymoEv(wd, &mts->nym);
	    wd = wd->next;
	}
	ForceReScan = 0;
}

static void TNotifyMount_On_appEv(TEv_Array *appEv) {
  FILE *f;
  struct mntent *MountEntry;
  TMounts *mts, **ppmts;

	if (!nym) return;
	if (ForceReScan > 0) AppSetWait_single(50000);
	if (NotifyMountStop || !cpTime(nym->app_us, appDat.app_us)) return;
	NotifyMountRun = true;

	nym->app_us = appDat.app_us + 1000000;
	if (ForceReScan > 0) {
	    ForceReScan--;
	    nym->app_us = appDat.app_us + 100000;
	}

	mts = nym->Mounts;
	while (mts) {
	    mts->stat = 0;
	    mts = mts->next;
	}

	f = setmntent("/proc/mounts", "r");
	if (f) {
LoopME:
	    MountEntry = getmntent(f);
	    if (MountEntry) {
		if (pcncmp(MountEntry->mnt_fsname, "/dev/", 5) ||
		    pccmp(MountEntry->mnt_type, "nfs") ||
		    pccmp(MountEntry->mnt_type, "nfs4") ||
		    pccmp(MountEntry->mnt_type, "cifs")) {
// printf("%s(%d): %s %s %s\n", FL, MountEntry->mnt_fsname, MountEntry->mnt_dir, MountEntry->mnt_type);
		    mts = nym->Mounts;
		    while (mts && !pccmp(MountEntry->mnt_dir, mts->nym.mnt_dir)) {
			mts = mts->next;
		    }

		    if (!mts) {
			mts = xcalloc(sizeof(TMounts));
			mts->nym.mnt_fsname = pcdup(MountEntry->mnt_fsname);
/*			int lg = pclen(mts->nym.mnt_fsname);
			if ((lg > 2) && 
			    (mts->nym.mnt_fsname[lg-1] == '/') &&
			    (mts->nym.mnt_fsname[lg-2] != ':')) {
			    mts->nym.mnt_fsname[lg-1] = 0;
			} */
			mts->nym.mnt_dir = pcdup(MountEntry->mnt_dir);
			mts->nym.mnt_type = pcdup(MountEntry->mnt_type);
			mts->next = nym->Mounts;
			nym->Mounts = mts;
// printf("%s(%d): Add    %-16s %s\n", FL, mts->nym.mnt_dir, mts->nym.mnt_fsname);
		        mts->stat = 1;
			SendEv(mts, nymobef_mount);
			
		    } else if (pccmp(MountEntry->mnt_fsname, mts->nym.mnt_fsname) ||
			       pccmp(MountEntry->mnt_type, mts->nym.mnt_type)) {
		        mts->stat = 2;
		    }
		}
		goto LoopME;
	    }
	    endmntent(f);
	}

	ppmts = &nym->Mounts;
	mts = nym->Mounts;
	while (mts) {
	    if (!mts->stat) {
// printf("%s(%d): Del    %s\n", FL, mts->mnt_dir);
		SendEv(mts, nymobef_umount);

		*ppmts = mts->next;
		xfree(mts->nym.mnt_fsname);
		xfree(mts->nym.mnt_dir);
		xfree(mts->nym.mnt_type);
		xfree(mts);

	    } else {
		ppmts = &mts->next;
	    }
	    mts = *ppmts;
	}
	NotifyMountRun = false;
}

void TNotifyMount_Replay(TNotifyMount_wd *self) {
  TMounts *mts = nym->Mounts;
	if (!self || !self->hnymoEv) return;
	while (mts) {
	    self->hnymoEv(self, &mts->nym);
	    mts = mts->next;
	}
}

void TNotifyMount_ReScan() {
	if (nym) {
	    nym->app_us = appDat.app_us + 100000;
	    ForceReScan = 5;
	    AppSetWait_single(50000);
	}
}

bool TNotifyMount_Stop(bool Stop) {
	if (Stop && NotifyMountRun) return false;
	NotifyMountStop = Stop;
	return true;
}

bool TNotifyMount_ChkMountPointFree(const char *fn) {
  TMounts *mts;

	if (!fn) return false;
	mts = nym->Mounts;
	while (mts) {
	    if (pccmp(fn, mts->nym.mnt_dir)) return false;
	    mts = mts->next;
	}
	return true;
}

char *TNotifyMount_GetMountPoint(const char *dev) {
  TMounts *mts;

	if (dev) {
	    mts = nym->Mounts;
	    while (mts) {
		if (pccmp(dev, mts->nym.mnt_fsname)) return pcdup(mts->nym.mnt_dir);
		mts = mts->next;
	    }
	}
	return NULL;
}
/*_____________________________ Destroy _________________________________*/
static bool TNotifyMount__Destroy(TNotifyMount_wd *self) {
  TNotifyMount_wd **ppwd = &nym->wd;

	while (*ppwd && (*ppwd != self)) {
	    ppwd = &(*ppwd)->next;
	}
	*ppwd = self->next;

	if (!nym->wd) {
	    Free_appEv(nym->appEv);
	    xfree(nym);
	    nym = NULL;
	}
	return False;
}

/*_____________________________ Create __________________________________*/

TNotifyMount_wd *TNotifyMount_Create(void *owner, ThNotifyMountEv hnymoEv) {
  TNotifyMount_wd *self;

	if (!nym) {
	    nym = xcalloc(sizeof(TNotifyMount));
	    nym->app_us = appDat.app_us;
	    nym->appEv = Add_appEv(TNotifyMount_On_appEv);
	    TNotifyMount_On_appEv(nym->appEv);
	}

	self = ttc_create(sizeof(TNotifyMount_wd), wtyp_TMountNy);
	self->destroy = (TDESTROY)TNotifyMount__Destroy;
	self->owner = owner;
	self->hnymoEv = hnymoEv;
	self->next = nym->wd;
	nym->wd = self;
	return self;
}
