/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#define IS_CLASS_LIB

#include "appbase.h"
#include "trs232.h"
#include <unistd.h>
//#include <sys/ioctl.h>
//#include <linux/fs.h>

#ifdef _OS_WIN32

typedef struct {
  dword value;
  dword flag;
} TFlags;


static const TFlags Tbaud_arr[] = {
  {    110, CBR_110},
  {    300, CBR_300},
  {    600, CBR_600},
  {   1200, CBR_1200},
  {   2400, CBR_2400},
  {   4800, CBR_4800},
  {   9600, CBR_9600},
  {  19200, CBR_19200},
  {  38400, CBR_38400},
  {  56000, CBR_56000},
  {  57600, CBR_57600},
  { 115200, CBR_115200},
  { 128000, CBR_128000},
  { 256000, CBR_256000},
  {0, 0}
};

static const dword Tbit_arr[4] = {8, 5, 6, 7};

#else /* _OS_WIN32 */

#include <fcntl.h>

typedef struct {
  dword value;
  dword flag;
} TFlags;

static const TFlags Tbaud_arr[] = {
  {     50, B50},
  {     75, B75},
  {    110, B110},
  {    134, B134},
  {    150, B150},
  {    200, B200},
  {    300, B300},
  {    600, B600},
  {   1200, B1200},
  {   1800, B1800},
  {   2400, B2400},
  {   4800, B4800},
  {   9600, B9600},
  {  19200, B19200},
  {  38400, B38400},
  {  57600, B57600},
  { 115200, B115200},
  { 230400, B230400},
  { 460800, B460800},
  { 500000, B500000},
  { 576000, B576000},
  { 921600, B921600},
  {1000000, B1000000},
  {1152000, B1152000},
  {1500000, B1500000},
  {2000000, B2000000},
  {2500000, B2500000},
  {3000000, B3000000},
  {3500000, B3500000},
  {4000000, B4000000},
  {0, 0}
};

static const dword Tbit_arr[4] = {CS8, CS5, CS6, CS7};
#endif /* ! _OS_WIN32 */

#define tflag_run	1
//#define tflag_tout	(1<<1)

typedef struct {
  dword tflag;
  TgTimer *Timer;
  dword tout_ev;	//in ms (min. 50ms)
  dword tout_T;		//in ms (min. 10ms)
  dword tout_DataLg;
  dword start_ms;
  dword baud;
  dword txd_count;
#ifdef _OS_WIN32
  TEv_Array *appEv;
  OVERLAPPED *OvlRxD;
  OVERLAPPED *OvlTxD;
  DWORD cevmask;
#endif /* ! _OS_WIN32 */
} ttc_private;

static void Trs232_OnTimer(TgTimer *ti) {
  Trs232 *self;
  ttc_private *pv;
  int td;

	self = ti->obj;
	pv = self->tc_private;
	if (!(pv->tflag & tflag_run)) return;
	td = (pv->start_ms + pv->tout_ev + pv->tout_DataLg) - appDat.app_ms;
	if (td > 0) return;
	pv->tflag &= ~tflag_run;
	self->TimeOutCount++;
	if (self->OnTimeOutTimer) self->OnTimeOutTimer(self);
}

bool Trs232_InitTimeOut(Trs232 *self, dword ms) {
  ttc_private *pv;
  dword tout;

	if (!self) return false;
	pv = self->tc_private;
	pv->tout_ev = ms;
	if (pv->tout_ev < 50) pv->tout_ev = 50;

	tout = (ms + 50) / 100;
	if (tout < 1) tout = 1;
	tout *= 10;

	if (!pv->Timer) {
	    pv->Timer = gTimerX_Add(Timer_Enabled, tout, Trs232_OnTimer);
	    pv->Timer->obj = self;    
	} else if (pv->tout_T != tout) {
	    TgTimer_NewInterval(pv->Timer, tout);
	}
	pv->tout_T = tout;
	return true;
}

void Trs232_ClrTimeOut(Trs232 *self) {
  ttc_private *pv;

	if (!self) return;
	pv = self->tc_private;
	pv->tflag &= ~tflag_run;
}
/*_____________________________ RxD _____________________________________*/
#ifdef _OS_WIN32
static void Trs232_EvRxD(Trs232 *self, DWORD inlg) {
  FileHandle fd;
#else
static void Trs232_EvRxD(FileHandle fd, Tfd_list *fd_list) {
  Trs232 *self;
#endif /* _OS_WIN32 */
  size_t size, lg;
  byte *buf;
  dword txd_count;
  bool errmk;
  ttc_private *pv;

#ifdef _OS_WIN32
	fd = self->fd;
#else
	self = fd_list->obj;
#endif /* _OS_WIN32 */

	size = self->RxDsize - self->RxDidx;
	if (size <= 0) {
	    errmk = false;
	    if (self->OnERROR) errmk = self->OnERROR(self, rs232ERR_RxD_OVERRUN, NULL);
	    if (!errmk) {
		appError(EMSG_ERR1, "rs232 RxD overrun (max=%d)", self->RxDsize);
	    }
	    self->RxDidx = 0;
	    size = self->RxDsize;
	}
	pv = self->tc_private;

#if RS232TIMEEXACT
	self->usRxD = GetTime_us64();
#else
	self->usRxD = appDat.app_us;
#endif
	buf = self->RxDbuf + self->RxDidx;
#ifdef _OS_WIN32
	if (size > inlg) size = inlg;
  DWORD rlg;
	if (ReadFile(fd, buf, size, &rlg, pv->OvlRxD)) {
	   lg = rlg;
#else
	if (file_read(fd, NULL, buf, size, &lg)) {
#endif
	    self->RxDidx += lg;
	    self->GesRxD_Byte += lg;
	    if (self->OnRxD) {
		TfreeLock(self);
		txd_count = pv->txd_count;
		if (self->OnRxD(self, self->RxDbuf, self->RxDidx)) {
		    self->RxDidx = 0;
		    if (txd_count == pv->txd_count) pv->tflag &= ~tflag_run;
		} else {
		}
		if (TfreeUnLock(self)) return;
	    }
	}

#ifdef _OS_WIN32
	WaitCommEvent(self->fd, &pv->cevmask, pv->OvlRxD);
#endif
}

/*_____________________________ TxD _____________________________________*/
static bool TxD_data(Trs232 *self, const byte *data, size_t lg) {
  ttc_private *pv;

	if (!data || (lg <= 0) || !Trs232_IsOpen(self)) return false;
	pv = self->tc_private;
// printf("%s(%d): %02x %02x %02x %02x\n", FL, data[0], data[1], data[2], data[3]);

#ifdef _OS_WIN32
  DWORD rlg;
	WriteFile(self->fd, data, lg, &rlg, pv->OvlTxD);
#else 
	if (!file_write(self->fd, NULL, data, lg, NULL)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "TxD_data");
	    return false;
	}
#endif
	pv->txd_count++;
	pv->start_ms = appDat.app_ms;
	pv->tout_DataLg = ((lg + 1) * 10000) / pv->baud;
	pv->tflag |= tflag_run;
	if (self->OnTxD) self->OnTxD(self, data, lg);
	return true;
}

bool Trs232_TxD(Trs232 *self, const byte *data, size_t lg) {
	if (!self) return false;
	if (!TxD_data(self, data, lg)) return false;
#if RS232TIMEEXACT
	self->usTxD = GetTime_us64();
#else
	self->usTxD = appDat.app_us;
#endif

	if (self->TxDbuf) {
	    if (self->TxDsize >= lg) {
		self->TxDDataLg = lg;
		pnmove(self->TxDbuf, data, lg);
		self->TxDRepCount = 0;
	    } else {
		self->TxDDataLg = 0;
		self->TxDRepCount = self->TxDRepMaxCount;
	    }
	}
	return true;
}

bool Trs232_TxDBuf(Trs232 *self) {
	if (!self) return false;
	if (!TxD_data(self, self->TxDbuf, self->TxDDataLg)) return false;
#if RS232TIMEEXACT
	self->usTxD = GetTime_us64();
#else
	self->usTxD = appDat.app_us;
#endif
	return true;
}

bool Trs232_TxDBuf_Rep(Trs232 *self) {
	if (!self) return false;
	return TxD_data(self, self->TxDbuf, self->TxDDataLg);
}

/*_____________________________ misc ____________________________________*/
bool Trs232_IsOpen(Trs232 *self) {
	if (!self) return false;
	return ChkFileHandleAkt(self->fd);
}

/*_____________________________ appEv ___________________________________*/
#ifdef _OS_WIN32
static void On_appEv(TEv_Array *appEv) {
  Trs232 *self;
  DWORD Errors;
  COMSTAT ComStat;

	self = appEv->obj;
	if (!ClearCommError(self->fd, &Errors, &ComStat)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "ClearCommError");
	}
// printf("%s(%d): %u\n", FL, (dword)ComStat.cbInQue);
	if (ComStat.cbInQue) Trs232_EvRxD(self, ComStat.cbInQue);

}
#endif
/*_____________________________ Close ___________________________________*/
bool Trs232_close(Trs232 *self) {
  ttc_private *pv;

	if (!self) return false;
	pv = self->tc_private;
	pv->tflag &= ~tflag_run;
#ifdef _OS_WIN32
	PurgeComm(self->fd, PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
	CloseHandle(self->fd);
	self->fd = FileHandleNULL;
	Free_appEv(pv->appEv);
	pv->appEv = NULL;
	if (pv->OvlRxD) {
	    DelAppMsgEvent(pv->OvlRxD->hEvent);
	    CloseHandle(pv->OvlRxD->hEvent);
	    ppfree(&pv->OvlRxD);
	}
	if (pv->OvlTxD) {
	    CloseHandle(pv->OvlTxD->hEvent);
	    ppfree(&pv->OvlTxD);
	}
#else
	ClosePipeProc(self->fd, clpp_fd);
#endif
	return true;
}

bool rs232_close(Trs232 *self) {
	if (self && self->close) return self->close(self);
	return false;
}

/*_____________________________ Open ____________________________________*/
bool Trs232_open(Trs232 *self, dword com, dword baud, dword rsflag) {
  int ibaud;
  char *pc;
  ttc_private *pv;
  char buf[32];
#ifdef _OS_WIN32
  DCB ios;
#else
  struct termios ios;
#endif

	if (!self) return false;
	rs232_close(self);
	if (!baud) return true;
	Fill0(&ios, sizeof(ios));
	pv = self->tc_private;

	ibaud = 0;
	while (Tbaud_arr[ibaud].value && (Tbaud_arr[ibaud].value != baud)) ibaud++;
	if (!Tbaud_arr[ibaud].value) {
	    appError(EMSG_ERR1, "rs232 baud=%d ???", baud);
	    return false;
	}
	pv->baud = Tbaud_arr[ibaud].value;

#ifdef _OS_WIN32
	if (com > 63) {	//??
	    appError(EMSG_ERR1, "rs232 max. 191 com");
	    return false;
	}
	pc = buf;
	ppcmove(&pc, "COM");
	if (com < 1) com = 1;
	inttopc(pc, com);
	self->fd = CreateFileA(buf, GENERIC_READ | GENERIC_WRITE, 0, NULL,
				OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
// printf("%s(%d): %s 0x%04x\n", FL, buf, (int)self->fd);
	if (!ChkFileHandleOK(self->fd)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "open");
	    self->fd = FileHandleNULL;
	    return false;
	}
	if (!SetupComm(self->fd, RS232MAXBUF, RS232MAXBUF)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "open SetupComm");
	    rs232_close(self);
	    return false;
	}

	ios.DCBlength = sizeof(ios);
	ios.BaudRate = Tbaud_arr[ibaud].flag;
	ios.ByteSize = Tbit_arr[rsflag & 0x03];
	ios.Parity = NOPARITY;
	if (rsflag & rsPARITY) ios.Parity = EVENPARITY;
	if (rsflag & rsPARITYODD) ios.Parity = ODDPARITY;
	ios.fBinary = 1;
	ios.StopBits = ONESTOPBIT;
	if (rsflag & rsSTOP2BIT) ios.StopBits = TWOSTOPBITS;
	if (rsflag & rsRTSCTS) {
	    ios.fOutxCtsFlow = 1;
	    ios.fRtsControl = 1;
	}
	if (!SetCommState(self->fd, &ios)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "open SetCommState");
	    rs232_close(self);
	    return false;
	}

	SetCommMask(self->fd, EV_RXCHAR | EV_ERR);

	pv->OvlRxD = xcalloc(sizeof(OVERLAPPED));
	pv->OvlRxD->hEvent = CreateEvent(NULL, true, false, NULL);
	pv->OvlTxD = xcalloc(sizeof(OVERLAPPED));
	pv->OvlTxD->hEvent = CreateEvent(NULL, true, false, NULL);

	AddAppMsgEvent(pv->OvlRxD->hEvent);
	WaitCommEvent(self->fd, &pv->cevmask, pv->OvlRxD);

#else /* _OS_WIN32 */

	if (com > 191) {
	    appError(EMSG_ERR1, "rs232 max. 191 com");
	    return false;
	}
	pc = buf;
	ppcmove(&pc, "/dev/ttyS");
	inttopc(pc, com);
	self->fd = open(buf, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (self->fd <= 0) {
	    appError(EMSG_ERR1 | EMSG_SYS, "rs232 open");
	    self->fd = FileHandleNULL;
	    return false;
	}
	ios.c_cflag = Tbit_arr[rsflag & 0x03] | CLOCAL | CREAD;
	if (rsflag & rsPARITY) ios.c_cflag |= PARENB;
	if (rsflag & rsPARITYODD) ios.c_cflag |= PARENB|PARODD;
	if (rsflag & rsSTOP2BIT) ios.c_cflag |= CSTOPB;
	if (rsflag & rsRTSCTS) ios.c_cflag |= CRTSCTS;

	if ((cfsetispeed(&ios, Tbaud_arr[ibaud].flag) < 0) ||
	    (cfsetospeed(&ios, Tbaud_arr[ibaud].flag) < 0)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "rs232 open");
	    rs232_close(self);
	    return false;
	}
	if (tcflush(self->fd, TCIOFLUSH) < 0) {
	    appError(EMSG_ERR1 | EMSG_SYS, "rs232 open");
	    rs232_close(self);
	    return false;
	}
	if (tcsetattr(self->fd, TCSANOW, &ios) < 0) {
	    appError(EMSG_ERR1 | EMSG_SYS, "rs232 open");
	    rs232_close(self);
	    return false;
	}

#endif /* ! _OS_WIN32 */
	self->RxDidx = 0;
	if (!self->RxDbuf) {
	    self->RxDsize = RS232MAXBUF;
	    self->RxDbuf = xmalloc(self->RxDsize);
	}
#ifdef _OS_WIN32
	pv->appEv = Add_appEv(On_appEv);
	pv->appEv->obj = self;
#else
	AddPipeProc(self->fd, InPipe, Trs232_EvRxD, self);
#endif /* ! _OS_WIN32 */
	self->TimeOutCount = 0;
	return true;
}

bool rs232_open(Trs232 *self, dword com, dword baud, dword rsflag) {
	if (self && self->open) return self->open(self, com, baud, rsflag);
	return false;
}

/*_____________________________ Destroy _________________________________*/
bool Trs232__Destroy(Trs232 *self) {
  ttc_private *pv;

	rs232_close(self);
	pv = self->tc_private;
	if (pv->Timer) {
	    ppTfree(&pv->Timer);
	}
	xfree(self->tc_private);
	xfree(self->TxDbuf);
	xfree(self->RxDbuf);
	return False;
}
/*_____________________________ Create __________________________________*/
void *Trs232_Create_struct_init(int tcsize, int wtyp) {
  Trs232 *self;

	self = ttc_create(tcsize, wtyp);
	self->destroy = (TDESTROY)Trs232__Destroy;
	self->open = Trs232_open;
	self->close = Trs232_close;
	self->tc_private = xcalloc(sizeof(ttc_private));
	self->RxDsize = RS232MAXBUF;
	self->RxDbuf = xmalloc(self->RxDsize);
	return self;
}

Trs232 *Trs232_Create() {
	return Trs232_Create_struct_init(sizeof(Trs232), wtyp_Trs232);
}
