/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"
#include "tdirlist_pv.h"

typedef struct {
  dword open_us;
  int count;
  int us;
#ifdef _OS_WIN32
  HANDLE dh;
  WIN32_FIND_DATAW fid;
#else
  DIR *dh;
#endif
  TEv_Array *appEv;
  TglobPath *gp;
  TDirFileList *rescan_dfl;
} Tprivate;

static void DodllEv(TDirListLoader *self, TDirFileList **dfl, int bef) {
  Tprivate *pv = self->private;
  TDirFileList *tdfl;
  TFileMsg fmsg;

	pv->us = GetCurTimeDiv_us(pv->open_us);
	tdfl = NULL;
	if (dfl) tdfl = *dfl;

	file_msg_init_Dm(&fmsg, fmArt_DirList, self->obef | bef);
	if (tdfl) {
	    fmsg.typF = fmTyp_dfl;
	    fmsg.F.dfl = tdfl;
	}
	fmsg.typR = fmTyp_gp;
	fmsg.R.gp = self->gp;

	file_msg_ev(&fmsg, (TTC0*)self->owner, self->OnFileMsg);
	if (dfl) *dfl = fmsg.F.dfl;
}

/*_____________________________ Rescan __________________________________*/

static void ld_Rescan_On_appEv(TEv_Array *appEv) {
  TDirListLoader *self = appEv->obj;
  Tprivate *pv = self->private;
  Tstat64 *st, *org;
  dword us = GetCurTimeDiv_us(pv->open_us);
  TDirFileList *dfl = pv->rescan_dfl;
  TDirFileList *dfl_next;
  dword chstat;

	while (dfl) {
	    dfl_next = dfl->next;
	    org = dfl->st;
	    if (org->FileType == stft_NoFound) {
		pv->gp->name_const = Get_ExFileDat_pc(org, exfdt_FileName);
		st = file_stat_gp(pv->gp, self->scanfu);
		pv->gp->name_const = NULL;

		if (st) {
		    chstat = chk_new_stat_type(org, st);
		    if (chstat) {
			dl_copy_new_stat_type(org, st);
			dfl->chstat = chstat;

			DodllEv(self, &dfl, dlbef_Change);
// printf("%4d %s\n", dfl->st->FileType, Get_ExFileDat_pc(dfl->st, exfdt_FileName));
		    }
		    xfree(st);
		} else {
		    dfl->chstat = chstat_Del;
		    DodllEv(self, &dfl, dlbef_Del);
		}
	    }
	    pv->us = GetCurTimeDiv_us(pv->open_us);
	    dfl = dfl_next;
	    if (dfl && ((pv->us - us) > 40000)) {
//    printf("%s(%d): On_appEv %6dus %6dus\n", FL, self->us, us);
		pv->rescan_dfl = dfl;
	        DodllEv(self, NULL, dlbef_Sync);
		return;
	    }
	}
	pv->rescan_dfl = NULL;
	DodllEv(self, NULL, dlbef_ListEnd);
	if (self->close) self->close(self);
}

static void TestRescan(TDirListLoader *self) {
  TDirFileList *dfl = NULL;
  Tprivate *pv = self->private;
  int m = self->scanfu;

	DodllEv(self, NULL, dlbef_ListEnd);
	if (self->close) self->close(self);

	self->obef = dlbef_RescanRun;
	if (self->scanfu != m) {
	    dfl = self->rescan_dfl;
	    while (dfl) {
		if (dfl->st->FileType == stft_NoFound) break;
		dfl = dfl->next;
	    }
	}
	if (!dfl) {
	    DodllEv(self, NULL, dlbef_Close);
	    return;
	}
	pv->open_us = GetTime_us();
	pv->us = 0;
	pv->rescan_dfl = dfl;
	pv->appEv = Add_appEv(ld_Rescan_On_appEv);
	pv->appEv->obj = self;
	self->scanfu |= fsNoFileName|fsNoLinkName;
	DodllEv(self, NULL, dlbef_Open);
	return;
}

/*_____________________________ appEv ___________________________________*/
static void add_dfl(TDirListLoader *self, Tprivate *pv, char *fn) {
  Tstat64 *st;
  TFileMsg fmsg;
// printf("%s(%d): %s\n", FL, fn);
	if ((fn[0] == '.') && !fn[1]) return;
	if ((fn[0] == '.') && (fn[1] == '.') && !fn[2] &&
	    (pv->gp->path[0] == '/') && !pv->gp->path[1]) return;

	if (!chk_pattern_match(self->patt, fn, NULL)) return;

	pv->gp->name = fn;
	st = file_stat_gp(pv->gp, self->scanfu);
	pv->gp->name = NULL;
	if (st) {
// printf("%s(%d): 0x%x %s\n", FL, self->scanfu, fn);
	    pv->count++;
	    file_msg_init(&fmsg, fmArt_DirList);
	    TDirListLoader_New_st(self, st, &fmsg);
	    xfree(st);
	}
}

static void TDirList_ld_On_appEv(TDirListLoader *self);

static void ld_On_appEv(TEv_Array *appEv) {
	TDirList_ld_On_appEv(appEv->obj);
}

static void TDirList_ld_On_appEv(TDirListLoader *self) {
  Tprivate *pv = self->private;
  dword us = GetCurTimeDiv_us(pv->open_us);
#ifdef _OS_WIN32
  char *tmputf8;
#else
  struct dirent64 entry;
  struct dirent64 *ret_entry;
#endif

	pv->us = us;
	if (pv->dh) {
loop:
#ifdef _OS_WIN32
	    tmputf8 = pwctoUTF8(pv->fid.cFileName);
	    add_dfl(self, pv, tmputf8);
	    xfree(tmputf8);
	    if (!FindNextFileW(pv->dh, &pv->fid)) {
		TestRescan(self);
		return;
	    }

	    if ((pv->us - us) < 40000) goto loop;
	    if (pv->count) DodllEv(self, NULL, dlbef_Sync);
	    if (!pv->appEv) {
		pv->appEv = Add_appEv(ld_On_appEv);
		pv->appEv->obj = self;
	    }
#else
	    readdir64_r(pv->dh, &entry, &ret_entry);
	    if (ret_entry) {
		add_dfl(self, pv, ret_entry->d_name);
		if ((pv->us - us) < 40000) goto loop;
		if (pv->count) DodllEv(self, NULL, dlbef_Sync);
// printf("%s(%d): TDirList_ld_On_appEv %u %u\n", FL, self->us, us);
		if (!pv->appEv) {
		    pv->appEv = Add_appEv(ld_On_appEv);
		    pv->appEv->obj = self;
		}
	    } else {
		TestRescan(self);
	    }
#endif
	}
}

/*_____________________________ Open ____________________________________*/
static bool TDirList_ld_File_Open(TDirListLoader *self) {
  Tprivate *pv = self->private;
  Tstat64 *st = NULL;

	if (self->close) self->close(self);
	gpFree(pv->gp);
	pv->gp = gp_dup(self->gp);
	xfree(pv->gp->name);
	pv->gp->name = NULL;
	pv->open_us = GetTime_us();
	pv->count = 0;
	pv->us = 0;
	st = file_stat_gp(self->gp, fsFollowLink | fsNoFileName | fsNoLinkName);
	if (!st) return False;
	xfree(st);
#ifdef _OS_WIN32
  char *fname = gptopc(self->gp, gptfu_noRootArt|gptfu_noName); // pcadd(self->gp->path, "*");
// printf("%s(%d): %s %s -> %s\n", FL, self->gp->path, self->gp->name, fname);
	word *tmpwc = get_pcw_filename(fname, add_end_asterisk);
	pv->dh = FindFirstFileW(tmpwc, &pv->fid);
	xfree(tmpwc);
	xfree(fname);
	if (pv->dh == INVALID_HANDLE_VALUE) return False;
#else
	pv->dh = opendir(self->gp->path);
	if (!pv->dh) return False;
#endif

	self->obef = dlbef_ScanRun;
	DodllEv(self, NULL, dlbef_Open);
	TDirList_ld_On_appEv(self);
	return True;
}

/*_____________________________ Close ___________________________________*/
static void TDirList_ld_File_Close(TDirListLoader *self) {
  Tprivate *pv = self->private;
  bool mk = (pv->dh || pv->appEv);

	if (pv->dh) {
#ifdef _OS_WIN32
	    if (pv->dh) FindClose(pv->dh);
#else
	    if (pv->dh) closedir(pv->dh);
#endif
	    pv->dh = 0;
	}
	if (pv->appEv) {
	    Free_appEv(pv->appEv);
	    pv->appEv = NULL;
	}
	if (mk) DodllEv(self, NULL, dlbef_Close);
}

/*_____________________________ Destroy _________________________________*/
static bool TDirListLoader__Destroy(TDirListLoader *self) {
  Tprivate *pv = self->private;
	self->OnFileMsg = NULL;
	if (self->close) self->close(self);
	gpFree(pv->gp);
	if (!(self->memfu & dlmemfu_gpNotFree)) gpFree(self->gp);
	xfree(self->private);
	return False;
}

/*_____________________________ Create __________________________________*/
static TDirListLoader *TDirList_ld_File_Create(void *owner, TDirFileList **ppdfl, TOnFileMsg OnFileMsg) {
  TDirListLoader *self;

	self = ttc_create(sizeof(TDirListLoader), wtyp_TDirListLoader);
	self->destroy = (TDESTROY)TDirListLoader__Destroy;
	self->ppdfl = ppdfl;
	self->add_ppdfl = ppdfl;
	self->OnFileMsg = OnFileMsg;
	self->owner = owner;
	self->private = xcalloc(sizeof(Tprivate));
	self->open = TDirList_ld_File_Open;
	self->close = TDirList_ld_File_Close;
	self->scanfu = fsGetFileAll_UGT;
	return self;
}

const ThDirListLoader hdirlist_ld_File = {
  gp_test_local_drive,
  TDirList_ld_File_Create,
  appFu_select_t_NULL
};
