/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tthread.h"
#include "tdirlist.h"
#include "tdirlist_pv.h"

#define SYNC_TIME_us 200000

typedef struct {
  dword open_us;
  int count;
  dword sync_us;
#ifdef _OS_WIN32
  HANDLE dh;
  WIN32_FIND_DATAW fid;
#else
  DIR *dh;
#endif
  TEv_Array *appEv;
  TglobPath *gp;
  struct TpthData *ptD;
} Tprivate;

typedef struct TpthFst {
  struct TpthFst *next;
  Tstat64 *st;
  char *fn;
  TDirFileList *dfl;
} TpthFst;


typedef struct TpthData {
  struct TpthData *next;
  TDirListLoader *dll;
  int stat;
  TpthFst *fst;
  TpthFst **ppfst;
  TpthFst *rescan_fst;
  DIR *dh;
//  bool SendListRun;
} TpthData;


static TThread *ld_pthd;

static TpthData *pthData;

static TpthData *get_pthData(TDirListLoader *self) {
  TpthData *ptD = pthData;
	while (ptD && (ptD->dll != self)) ptD = ptD->next;
	return ptD;
}

static bool __attribute__ ((noinline)) add_fst(TDirListLoader *self, char *fn) {
  TpthData *ptD;
  Tprivate *pv = self->private;
  TglobPath *gp;
  Tstat64 *st;
  int scanfu;
  TpthFst *fst;

	TThread_lock(ld_pthd);
	ptD = get_pthData(self);
	if (!ptD) goto ende;
	pv = ptD->dll->private;
	if ((fn[0] == '.') && !fn[1]) goto ende;
	if ((fn[0] == '.') && (fn[1] == '.') && !fn[2] &&
	    (pv->gp->path[0] == '/') && !pv->gp->path[1]) goto ende;

	if (!chk_pattern_match(ptD->dll->patt, fn, NULL)) goto ende;
	pv->gp->name = fn;
	gp = gp_dup(pv->gp);
	pv->gp->name = NULL;
	scanfu = ptD->dll->scanfu;
	TThread_unlock(ld_pthd);

	st = file_stat_gp(gp, scanfu);
	gpFree(gp);
	if (!st) return True;

	fst = xcalloc(sizeof(TpthFst));
	fst->st = st;

	TThread_lock(ld_pthd);
	ptD = get_pthData(self);
	if (!ptD) {
	    xfree(fst);
	    xfree(st);
	    goto ende;
	}

	if (!ptD->fst) ptD->ppfst = &ptD->fst;
	*ptD->ppfst = fst;
	ptD->ppfst = &fst->next;
ende:
	TThread_unlock(ld_pthd);
// if (!pcncmp(pv->gp->path, "/root/", 6) && !pcncmp(pv->gp->path, "/opt/", 5)) usleep(4000);
	return ptD != NULL;
}

static void free_fst(TpthFst *fst) {
	xfree(fst->fn);
	xfree(fst->st);
	xfree(fst);
}

static void clr_fst(TpthFst **ppfst) {
  TpthFst *fst;
	while (*ppfst) {
	    fst = *ppfst;
	    *ppfst = fst->next;
	    free_fst(fst);
	}
}
 
static void Thread_Rescan(TThread *pthd, void *obj) {
  TpthData *ptD;
  TglobPath *gp;
  TpthFst *fst, *fstnext;
  int scanfu;

	TThread_lock(pthd);
	ptD = get_pthData(obj);
	if (!ptD) {
	    TThread_unlock(pthd);
	    return;
	}

	scanfu = ptD->dll->scanfu;
	gp = gp_dup(ptD->dll->gp);
	fst = ptD->rescan_fst;
	ptD->rescan_fst = NULL;
	TThread_unlock(pthd);

	xfree(gp->name);

	while (fst && ptD) {
	    fstnext = fst->next;
	    fst->next = NULL;
	    gp->name = fst->fn;
	    fst->st = file_stat_gp(gp, scanfu);

	    TThread_lock(pthd);
	    ptD = get_pthData(obj);
	    if (ptD) {
		if (!ptD->fst) ptD->ppfst = &ptD->fst;
		*ptD->ppfst = fst;
		ptD->ppfst = &fst->next;
		fst = fstnext;
	    }
	    TThread_unlock(pthd);
	}

	gp->name = NULL;
	gpFree(gp);
	clr_fst(&fst);

	TThread_lock(pthd);

	ptD = get_pthData(obj);
	if (ptD) ptD->stat = 3;

	TThread_unlock(pthd);
	TThread_Sync_app();
}

static int Thread_DoExecute(TThread *pthd) {
  int w_ret, fcount;
  TpthData *ptD;
  TglobPath *gp;
  DIR *dh;
  struct dirent64 entry;
  struct dirent64 *ret_entry;
  Tstat64 *st;

// printf("%s(%d): Thread_DoExecute 0x%x\n", FL, (int)pthd);
loop:
	w_ret = TThread_Wait(pthd, 100000);
	if (w_ret == thdWait_msg) {
	    switch (pthd->msg.thdMSG & thdMSG_Mask) {
	    case thdMSG_CloseProcess:
		goto ende;

	    case thdMSG_Init:
	      if (pthd->msg.tag == 0) {
		gp = NULL;
		dh = NULL;
		fcount = 0;
		TThread_lock(pthd);
		ptD = get_pthData(pthd->msg.u.p);
		if (!ptD) goto pth_unlock;

		gp = gp_dup(ptD->dll->gp);
		TThread_unlock(pthd);

//  if (pcncmp(gp->path, "/mnt/", 5)) {
//      sleep(3);
//      st = NULL;
//  } else {
		st = file_stat_gp(gp, fsLinkDirekt | fsNoFileName | fsNoLinkName);
//  }
		if (st) {
		    xfree(st);
		    dh = opendir(gp->path);
		} 
//  else {
//    printf("%s(%d): %s stat=NULL [%s]\n", FL, gp->path, SysError_pc());
//    goto simOk;
//  }

		TThread_lock(pthd);
		ptD = get_pthData(pthd->msg.u.p);
		if (!ptD) {
//    printf("%s(%d): %s ptD=NULL\n", FL, gp->path);
		    if (dh) closedir(dh);
		    goto pth_unlock;
		}

//		ptD->dh = dh;
		ptD->stat = 1;
		if (!dh) {
		    ptD->stat = -1;	
		    goto pth_unlock;
		}
		TThread_unlock(pthd);


		if (dh) {
loop_dir:
		    readdir64_r(dh, &entry, &ret_entry);
		    if (ret_entry) {
			fcount++;
			if (add_fst(pthd->msg.u.p, ret_entry->d_name)) goto loop_dir;
		    }
		    closedir(dh);
		}
//simOk:
		TThread_lock(pthd);

		ptD = get_pthData(pthd->msg.u.p);
		if (ptD) {
		    if (fcount) ptD->stat = 2; else {
			ptD->stat = -1;
//    printf("%s(%d): %s fcount=0\n", FL, gp->path);
		    }
//		    ptD->dh = NULL;
// } else {
//    printf("%s(%d): %s ptD=NULL\n", FL, gp->path);
		}
pth_unlock:
// printf("%s(%d): ptD->stat=%d\n", FL, ptD->stat);
		TThread_unlock(pthd);
		TThread_Sync_app();

		gpFree(gp);

	      } else if (pthd->msg.tag == 1) {
		Thread_Rescan(pthd, pthd->msg.u.p);
	      }
	      break;
	    }
	}

	goto loop;
ende:
// printf("%s(%d): Thread_Ende\n", FL);

	return 0;
}

/*_____________________________ Ev ______________________________________*/

static void DodllEv(TDirListLoader *self, TDirFileList **dfl, int bef) {
  TDirFileList *tdfl;
  TFileMsg fmsg;

// printf("%s(%d): DodllEv %d(0x%x)\n", FL, bef, bef);
	if (self->memfu & dlmemfu_goDestroy) return;
	tdfl = NULL;
	if (dfl) tdfl = *dfl;

	file_msg_init_Dm(&fmsg, fmArt_DirList, self->obef | bef);
	if (tdfl) {
	    fmsg.typF = fmTyp_dfl;
	    fmsg.F.dfl = tdfl;
	}
	fmsg.typR = fmTyp_gp;
	fmsg.R.gp = self->gp;

	file_msg_ev(&fmsg, (TTC0*)self->owner, self->OnFileMsg);
	if (dfl) *dfl = fmsg.F.dfl;
}

static void ld_Rescan_On_appEv(TEv_Array *appEv) {
  TDirListLoader *self = appEv->obj;
  Tprivate *pv = self->private;
  TpthData *ptD;
  TpthFst *fst;
  int ptDstat;
  TDirFileList *dfl;
  dword chstat;

// printf("%s(%d): 0x%08x 0x%08x\n", FL, (int)self, (int)self->rescan_dfl);

	if (self->memfu & dlmemfu_LOCK) return;
	self->memfu |= dlmemfu_LOCK;

	ptD = pv->ptD;
fst_loop:
	TThread_lock(ld_pthd);
	ptDstat = ptD->stat;
	if (ptD->fst) {
	    fst = ptD->fst;
	    ptD->fst = fst->next;
	    TThread_unlock(ld_pthd);
	    dfl = self->rescan_dfl;
	    while (dfl && (dfl != fst->dfl)) dfl = dfl->next;

	    if (dfl) {
		if (fst->st) {
		    chstat = chk_new_stat_type(dfl->st, fst->st);
		    if (chstat) {
			dl_copy_new_stat_type(dfl->st, fst->st);
			dfl->chstat = chstat;
			DodllEv(self, &dfl, dlbef_Change);
			pv->count++;
// printf("%4d %s\n", dfl->st->FileType, Get_ExFileDat_pc(dfl->st, exfdt_FileName));
		    }
		} else {
// printf("%s(%d): DEL %s\n", FL, dfl->fn);
		    TDirFileList **ppdfl = get_ppdfl(&self->rescan_dfl, dfl);
		    if (*ppdfl) {
			*ppdfl = dfl->next;
			dfl->chstat = chstat_Del;
			DodllEv(self, &dfl, dlbef_Del);
			xfree(dfl);
			pv->count++;
		    }
		}
	    }
	    free_fst(fst);
	    goto fst_loop;
	}
	TThread_unlock(ld_pthd);

	if (ptDstat > 2) {
	    DodllEv(self, NULL, dlbef_ListEnd);
	    DodllEv(self, NULL, dlbef_Close);
	    Free_appEv(pv->appEv);
	    pv->appEv = NULL;
	}

	dword us = GetTime_us();
	if (pv->count && cpTime(pv->sync_us, us)) {
	    pv->count = 0;
	    pv->sync_us = us + SYNC_TIME_us;
	    DodllEv(self, NULL, dlbef_Sync);
	}

	self->memfu &= ~dlmemfu_LOCK;
	if (self->memfu & dlmemfu_goDestroy) {
	    Tfree(self);
	}
}

static void ld_On_appEv(TEv_Array *appEv);

static void DataSendToDirList(TDirListLoader *self) {
  Tprivate *pv = self->private;
  TpthData *ptD;
  int res;
  TpthFst *fst, **ppfst;
  TFileMsg fmsg;
  int m;
  TDirFileList *dfl = NULL;

	if (self->memfu & dlmemfu_LOCK) return;
	self->memfu |= dlmemfu_LOCK;

	file_msg_init(&fmsg, fmArt_DirList);
	ptD = pv->ptD;
// printf("%s(%d): 0x%08x 0x%08x\n", FL, (int)self, (int)self->rescan_dfl);
fst_loop:
	TThread_lock(ld_pthd);
	res = ptD->stat;
	if (ptD->fst) {
	    fst = ptD->fst;
	    ptD->fst = fst->next;
	    TThread_unlock(ld_pthd);

	    TDirListLoader_New_st(self, fst->st, &fmsg);
	    free_fst(fst);
	    pv->count++;
	    if (self->memfu & dlmemfu_goDestroy) goto goDestroy;
	    goto fst_loop;
	}
	TThread_unlock(ld_pthd);

	if (res < 0) {
// printf("%s(%d): \n", FL);
	    DodllEv(self, NULL, dlbef_ListEnd);
	    DodllEv(self, NULL, dlbef_Close);
	    DodllEv(self, NULL, dlbef_Destroy);

	} else if (res > 1) {

	    if (pv->appEv) {
		Free_appEv(pv->appEv);
		pv->appEv = NULL;
	    }

	    dfl = NULL;
	    m = self->scanfu;
	    DodllEv(self, NULL, dlbef_ListEnd);
	    DodllEv(self, NULL, dlbef_Close);
	    if (self->memfu & dlmemfu_goDestroy) goto goDestroy;

	    self->obef = dlbef_RescanRun;


	    fst = NULL;
	    if (self->scanfu != m) {
// printf("%s(%d): Chk Rescan 0x%x\n", FL, (int)self->rescan_dfl);
		ppfst = &fst;
		dfl = self->rescan_dfl;
		while (dfl) {
// printf("%s(%d): Rescan %d %s\n", FL, dfl->st->FileType, dfl->fn);
		    if (dfl->st->FileType == stft_NoFound) {
			TpthFst *tfst = xcalloc(sizeof(TpthFst));
			tfst->dfl = dfl;
			tfst->fn = pcdup(dfl->fn);
			*ppfst = tfst;
			ppfst = &tfst->next;
		    }
		    dfl = dfl->next;
		}
		
	    }
	    if (!fst) {
		DodllEv(self, NULL, dlbef_Close);
	    } else {
// printf("%s(%d): Start Rescan 0x%x\n", FL, (int)self->rescan_dfl);
		pv->appEv = Add_appEv(ld_Rescan_On_appEv);
		pv->appEv->obj = self;
		self->scanfu |= fsNoFileName|fsNoLinkName;
		DodllEv(self, NULL, dlbef_Open);
		ptD->rescan_fst = fst;

  TThread_MSG msg;
		msg.thdMSG = thdMSG_Init;
		msg.tag = 1;
		msg.u.type = uthd_t_p;
		msg.u.p = self;
		app2thd_pMSG(ld_pthd, True, &msg);
	    }

	} else if (!pv->appEv) {
	    pv->appEv = Add_appEv(ld_On_appEv);
	    pv->appEv->obj = self;
	}

	dword us = GetTime_us();
	if (pv->count && cpTime(pv->sync_us, us)) {
	    pv->count = 0;
	    pv->sync_us = us + SYNC_TIME_us;
	    DodllEv(self, NULL, dlbef_Sync);
	}

//	ptD->SendListRun = False;
goDestroy:
	self->memfu &= ~dlmemfu_LOCK;
	if (self->memfu & dlmemfu_goDestroy) {
	    Tfree(self);
	}
}


static void ld_On_appEv(TEv_Array *appEv) {
	DataSendToDirList(appEv->obj);
}

/*_____________________________ Open ____________________________________*/
static bool ld_File_Open(TDirListLoader *self) {
  Tprivate *pv = self->private;
  TpthData *ptD;
//  dword ms;

	if (self->memfu & dlmemfu_goDestroy) return False;
	ptD = pv->ptD;
// printf("%s(%d): ld_File_Open 0x%x %s\n", FL, (int)ptD, self->gp->path);
	if (!ptD) return False;

	TThread_lock(ld_pthd);
	gpFree(pv->gp);
	pv->gp = gp_dup(self->gp);
	xfree(pv->gp->name);
	pv->gp->name = NULL;
	pv->open_us = GetTime_us();
	pv->count = 0;
	TThread_unlock(ld_pthd);

  TThread_MSG msg;
	msg.thdMSG = thdMSG_Init;
	msg.tag = 0;
	msg.u.type = uthd_t_p;
	msg.u.p = self;
	app2thd_pMSG(ld_pthd, True, &msg);

//	ms = GetTime_ms() + 10000;
//	while (!ptD->stat && !cpTime(ms, GetTime_ms()) && AppHandleMessage());
	if (ptD->stat < 0) return False;

	self->obef = dlbef_ScanRun;
	DodllEv(self, NULL, dlbef_Open);

	pv->sync_us = GetTime_us() + 100000;

	DataSendToDirList(self);
	return True;
}

/*_____________________________ Close ___________________________________*/
static void ld_File_Close(TDirListLoader *self) {
}

/*_____________________________ Destroy _________________________________*/
static bool TDirListLoader__Destroy(TDirListLoader *self) {
  Tprivate *pv = self->private;
  TpthData **ppptD;

// printf("%s(%d): TDirListLoader__Destroy 0x%02x %s\n", FL, self->memfu, self->gp->path);

	if (pv->appEv) {
	    Free_appEv(pv->appEv);
	    pv->appEv = NULL;
	}

	ppptD = &pthData;
	while (*ppptD && (*ppptD != pv->ptD)) ppptD = &(*ppptD)->next;
	if (*ppptD) {
	    TpthData *ptD = *ppptD;
	    TThread_lock(ld_pthd);
	    *ppptD = ptD->next;
	    TThread_unlock(ld_pthd);
	    clr_fst(&ptD->fst);
	    clr_fst(&ptD->rescan_fst);
	}

	if (self->memfu & dlmemfu_LOCK) {
	    self->memfu |= dlmemfu_goDestroy;
	    return True;
	}

	xfree(pv->ptD);
	gpFree(pv->gp);
	if (!(self->memfu & dlmemfu_gpNotFree)) gpFree(self->gp);
	xfree(self->private);
	return False;
}

static void TDirListLoader__AppAtExit() {
	if (ld_pthd) {
	    Tfree(ld_pthd);
	}
}

/*_____________________________ Create __________________________________*/
static TDirListLoader *TDirList_ldPT_File_Create(void *owner, TDirFileList **ppdfl, TOnFileMsg OnFileMsg) {
  TDirListLoader *self;
  TpthData *ptD, **pppt;
  Tprivate *pv;

	self = ttc_create(sizeof(TDirListLoader), wtyp_TDirListLoader);
	self->destroy = (TDESTROY)TDirListLoader__Destroy;
	self->ppdfl = ppdfl;
	self->add_ppdfl = ppdfl;
	self->OnFileMsg = OnFileMsg;
	self->owner = owner;
	pv = xcalloc(sizeof(Tprivate));
	self->private = pv;
	self->open = ld_File_Open;
	self->close = ld_File_Close;
	self->scanfu = fsGetFileAll_UGT;

	if (!ld_pthd) {

	    ld_pthd = TThread_Create();
	    ld_pthd->DoExecute = Thread_DoExecute;
//	    ld_pthd->OnMsgEvent = pth_OnMSG;
	    TThread_Start(ld_pthd);
	    pp_void_add((void*)&AppProcAtExit, 1, TDirListLoader__AppAtExit);
	}

	ptD = xcalloc(sizeof(TpthData));
	pv->ptD = ptD;
	ptD->dll = self;

	TThread_lock(ld_pthd);
	pppt = &pthData;
	while (*pppt) pppt = &(*pppt)->next;
	*pppt = ptD;
	TThread_unlock(ld_pthd);

	return self;
}

const ThDirListLoader hdirlist_ld_File_pth = {
  gp_test_local_drive,
  TDirList_ldPT_File_Create,
  0
};
