/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"
#include <mntent.h>
#include <unistd.h>
#include "tdirlist_pv.h"

#ifdef _OS_WIN32
/*_____________________________ Create __________________________________*/
TDirListNotify *TDirListINotify_File_Create(TDirList *dl, TDirListLoaderEv dllEv) {
	return NULL;
}
#else /* _OS_WIN32 */

#ifdef USE_INOTIFY

#include <syscall.h>
#include <linux/inotify.h>

typedef struct Tny_private {
  Alg_dirlist_nypv
  TDirListNotify *next_dlny;
  dword rescan_us;
  int wd;
  TDirFileList *scan_dfl;
  TDirFileList *last_get_dfl;
  bool scan;
  bool ReScan_Run;
} Tny_private;

typedef struct TMounts {
  struct TMounts *next;
  int stat;
  char *mnt_fsname;		/* Device or server for filesystem.  */
  char *mnt_dir;		/* Directory mounted on.  */
  char *mnt_type;		/* Type of filesystem: ufs, nfs, etc.  */
} TMounts;

static int g_ny_fd = 0;
static TLineBuf *g_ny_lb;
static TDirListNotify *g_dlny;
static TEv_Array *ny_appEv;
static Tnymo_wd mount_wd;

static void iNotify_nextev_rescan_us(TDirListNotify *dlny) {
  Tny_private *pv = dlny->private;
  int us = (pv->rescan_us - appDat.app_us);
	if (us > 300000) pv->rescan_us = GetTime_us() + 300000;
}

static void iNotify_add_dfl_rescan(TDirListNotify *dlny, 
				   TDirFileList *dfl, int dfl_nymk) {

	iNotify_nextev_rescan_us(dlny);
	if (dfl_nymk & dflnymk_Del) {
	    dfl->nymk &= ~(dflnymk_Change|dflnymk_Add|dflnymk_WrClose);
	} else {
	    dfl->nymk &= ~(dflnymk_Del|dflnymk_WrClose);
	}
	dfl->nymk |= dfl_nymk;
}

static TDirFileList *iNotify_get_dfl(TDirListNotify *dlny, const char *fn) {
  TDirFileList *dfl;
  Tny_private *pv = dlny->private;

	if (pv->last_get_dfl && pccmp(pv->last_get_dfl->fn, fn)) {
	    return pv->last_get_dfl;
	}
	dfl = dlny->dl->dfl;
	while (dfl && !pccmp(dfl->fn, fn)) dfl = dfl->next;
	pv->last_get_dfl = dfl;
	return dfl;
}

static TDirFileList *iNotify_add_dfl(TDirListNotify *dlny, char *fn, int dfl_nymk, int dfl_nymask) {
  Tny_private *pv = dlny->private;
  TDirFileList *dfl;

	if (!*fn) return NULL;
	dfl = iNotify_get_dfl(dlny, fn);
//  printf("%s(%d): iNotify_add_dfl 0x%x %s\n", FL, (int)dfl, fn);
	if (dfl) {
	    dfl_nymk |= (dfl->nymk & dfl_nymask);
	} else {
	    dfl_nymk = dflnymk_Add;
	    pv->gp.name = fn;
	    dfl = TDirListNotify_Add(dlny, &pv->gp);
	    pv->gp.name = NULL;
	}
	if (dfl) iNotify_add_dfl_rescan(dlny, dfl, dfl_nymk);
	return dfl;
}

static void iNotify_free(TDirListNotify *dlny) {
  TFileMsg fmsg;
	if (dlny->OnFileMsg) {
	    file_msg_init_Dm(&fmsg, fmArt_DirList, dlbef_NotifyRun | dlbef_DelSelf);
	    file_msg_ev(&fmsg, (TTC0*)dlny->dl, dlny->OnFileMsg);
	    dlny->OnFileMsg = NULL;
	}
	Tfree(dlny);
}
/*_____________________________ iNotify_callback ________________________*/
static void iNotify_callback(int fd, Tfd_list *fdl) {
  int lg;
  struct inotify_event *in;
  char *pcs;

	lg = TLineBuf_fd_Add(g_ny_lb, g_ny_fd);
	if (AppToAtExit) return;

	if (lg > 0) {
	    while (g_ny_lb->lg >= sizeof(struct inotify_event)) {
		pcs = g_ny_lb->pc;
		if (g_ny_lb->pc_rdidx) pcs = g_ny_lb->pc_rdidx;
		in = (struct inotify_event*)pcs;
#if 0
if (in->len > 0) {
  printf("%s(%d): TiNotify_callback wd=%d %x %x len=%d %s\n", FL, 
	 in->wd, in->mask, in->cookie, in->len, in->name);
 } else {
  printf("%s(%d): TiNotify_callback wd=%d %x %x len=%d\n", FL, 
	 in->wd, in->mask, in->cookie, in->len);
 }
#endif
		lg = sizeof(struct inotify_event) + in->len;
		g_ny_lb->lg -= lg;
		g_ny_lb->pc_rdidx = pcs+lg;

  TDirListNotify *dlny;
  TDirFileList *dfl;
		dlny = g_dlny;
		while (dlny) {
		    if (((Tny_private*)dlny->private)->wd == in->wd) break;
		    dlny = ((Tny_private*)dlny->private)->next_dlny;
		}
		if (!dlny) {
		    syscall (__NR_inotify_rm_watch, g_ny_fd, in->wd);
		    continue;
		}

		if (in->len > 0) {
		    if ((in->name[0] == '.') && !in->name[1]) continue;

		    if (in->mask & (IN_DELETE|IN_MOVED_FROM)) {
			dfl = iNotify_get_dfl(dlny, in->name);
			if (dfl) {
			    ((Tny_private*)dlny->private)->last_get_dfl = NULL;
			    iNotify_nextev_rescan_us(dlny);
			    TDirListNotify_Del(dlny, dfl);
			}

		    } else if (in->mask & (IN_MODIFY|IN_MOVED_TO|IN_CREATE)) {
			iNotify_add_dfl(dlny, in->name, dflnymk_Change, 0);

		    } else if (in->mask & IN_ATTRIB) {
			iNotify_add_dfl(dlny, in->name, dflnymk_Change, dflnymk_WrClose);

		    } else if (in->mask & IN_CLOSE_WRITE) {
			iNotify_add_dfl(dlny, in->name, dflnymk_WrClose, 0);
		    }
		}

		if (in->mask & (IN_DELETE_SELF|IN_MOVE_SELF)) {
		    iNotify_free(dlny);
		}
	    }
	}
}

/*_____________________________ appEv ___________________________________*/

static void TDirList_appEv(TDirListNotify *self) {
  Tny_private *pv;
  TDirFileList *dfl;
  dword us, chstat;

	pv = self->private;
	if (pv->ReScan_Run) return;
	dfl = pv->scan_dfl;
	if (!dfl) {
	    if (cpTime(pv->rescan_us, appDat.app_us)) {
		dfl = self->dl->dfl;
		pv->change_akt = False;
	    }
	}
	if (!dfl) return;

	us = GetTime_us();
	while (dfl) {
	    if (dfl->nymk & dflnymk_Del) {
		pv->last_get_dfl = NULL;
		dfl = TDirListNotify_Del(self, dfl);
// printf("%s(%d): dflnymk_Del\n", FL);
		continue;

	    } else if ((dfl->nymk & (dflnymk_Rescan|dflnymk_Change|dflnymk_Add|dflnymk_WrClose)) ||
		       (dfl->st->st & (st_Dir|st_Link))) {

// printf("%s(%d): dflnymk_XXXX %s\n", FL, dfl->fn);
		pv->gp.name_const = dfl->fn;
		chstat = TDirListNotify_chk_dlf_Change(self, dfl, &pv->gp);
		pv->gp.name_const = NULL;
		if (dfl->nymk & dflnymk_Rescan) pv->change_akt = True;

		int nymk = dfl->nymk;
		dfl->nymk &= ~(dflnymk_Add|dflnymk_Change|dflnymk_WrClose);

		if (nymk & dflnymk_Add) {
		    TDirListNotify_DodllEv(self, &dfl, dlbef_Add);
		    if (!(dfl->st->st & st_Dir)) dfl->nymk |= dflnymk_WrClose;

		} else if (chstat || (nymk & dflnymk_Change)) {
		    TDirListNotify_DodllEv(self, &dfl, dlbef_Change);
		    if (!(dfl->st->st & st_Dir)) dfl->nymk |= dflnymk_WrClose;
		}

		if (nymk & dflnymk_WrClose) {
		    TDirListNotify_DodllEv(self, &dfl, dlbef_WrClose);
		    dfl->nymk &= ~dflnymk_WrClose;
		}


		if (GetCurTimeDiv_us(us) > 10000) {
		    TDirListNotify_SyncEv(self);
		    pv->scan_dfl = dfl->next;
		    return;
		}
	    }
	    dfl = dfl->next;
	}

	TDirListNotify_ScanEnd(self);
	pv->scan_dfl = NULL;
	pv->rescan_us = GetTime_us() + 500000;
	if (!pv->change_akt) pv->rescan_us += 2000000;
}


static void Start_INotify(TDirListNotify *self) {
  Tny_private *pv = self->private;

	pv->last_get_dfl = NULL;
	pv->wd = syscall (__NR_inotify_add_watch, g_ny_fd, pv->gp.path, 
	      IN_MODIFY|IN_ATTRIB|IN_CLOSE_WRITE|IN_MOVED_FROM|IN_MOVED_TO|IN_CREATE|IN_DELETE|IN_DELETE_SELF|IN_MOVE_SELF|IN_ONLYDIR);
	if (pv->wd < 0) {
	    appError(EMSG_ERR1 | EMSG_SYS, "%s(%d): inotify: \"%s\"", FL, pv->gp.path);
	}
}

static void Stop_INotify(TDirListNotify *self) {
  Tny_private *pv = self->private;

	if (pv->wd >= 0) syscall(__NR_inotify_rm_watch, g_ny_fd, pv->wd);
	pv->wd = -1;
	pv->last_get_dfl = NULL;
}

static TDirListNotify *dlny_GetNext(TDirListNotify *self) {
  TDirListNotify *result;

	if (!self || !ttc_chk_struct(self)) {
 printf("%s(%d): dlny_GetNext\n", FL);
	    return NULL;
	}
	result = ((Tny_private*)self->private)->next_dlny;
	while (result && TfreeChkGoFree(result)) {
 printf("%s(%d): dlny_GetNext\n", FL);
	    result = ((Tny_private*)result->private)->next_dlny;
	}
	return result;
}
static TDirListNotify *Get_dlny(TDirListNotify *self) {
  TDirListNotify *result;

	if (!self) return NULL;
	if (!ttc_chk_struct(self)) {
 printf("%s(%d): Get_dlny\n", FL);
	    return NULL;
	}
	result = self;
	while (result && TfreeChkGoFree(result)) {
 printf("%s(%d): Get_dlny\n", FL);
	    result = ((Tny_private*)result->private)->next_dlny;
	}
	return result;
}

static void NotifyMountEv(Tnymo_wd wd, TNotifyMount_data *ny_data) {
  TDirListNotify *dlny;
  Tny_private *pv;
  int lg, lg0;
  TDirFileList *dfl;
  char *path;

	if ((ny_data->bef != nymobef_umount) &&
	    (ny_data->bef != nymobef_mount)) return;

// printf("%s(%d): %d %s\n", FL, ny_data->bef, ny_data->mnt_dir);
	path = path_add_Slash(ny_data->mnt_dir, NULL);

	lg0 = pclen(path);
	dlny = Get_dlny(g_dlny);
	while (dlny) {	//Destroy sub DIR
	    pv = dlny->private;
	    lg = pclen(pv->gp.path);
	    if ((lg0 < lg) && pcncmp(pv->gp.path, path, lg0)) {
		iNotify_free(dlny);
		dlny = Get_dlny(g_dlny);
	    } else {
		dlny = dlny_GetNext(dlny);
	    }
	}

	dlny = Get_dlny(g_dlny);
	while (dlny) {	//Clr Mountpoint and rescan
	    pv = dlny->private;
	    if ((lg0 == pclen(pv->gp.path)) && pcncmp(pv->gp.path, path, lg0)) {
		Stop_INotify(dlny);
		if (dlny->dl) {
		    dfl = dlny->dl->dfl;
		    while (dfl) {
			TDirFileList *dflx = dfl->next;
			if (!pccmp(dfl->fn, ".."))
			    TDirListNotify_Del(dlny, dfl);
			dfl = dflx;
		    }
		}
		pv->last_get_dfl = NULL;

		if (nypv_OpenDir(dlny)) {
		    while (!nypv_ScanDir(dlny));
		    Start_INotify(dlny);
		}
		TDirListNotify_DodllEv(dlny, NULL, dlbef_Sync);
		break;
	    }
	    dlny = dlny_GetNext(dlny);
	}
	xfree(path);
}

static void TDirList_nyi_On_appEv(TEv_Array *appEv) {
  TDirListNotify *dlny, *nx_dlny;

	if (!TNotifyMount_Stop(true)) return;
	dlny = Get_dlny(g_dlny);
	while (dlny) {
	    TfreeLock(dlny);
	    nx_dlny = dlny_GetNext(dlny);
	    if (nx_dlny) TfreeLock(nx_dlny);
	    TDirList_appEv(dlny);
	    if (TfreeUnLock(dlny)) {
 printf("%s(%d): TDirList_nyi_On_appEv TfreeUnLock(dlny)\n", FL);
	    }
	    if (nx_dlny && TfreeUnLock(nx_dlny)) {
 printf("%s(%d): ERROR: TDirList_nyi_On_appEv TfreeUnLock\n", FL);
		break;
	    }
	    dlny = nx_dlny;
	}
	TNotifyMount_Stop(false);
}

/*_____________________________ ReScan __________________________________*/
static void TDirListNotify_File_ReScan(TDirListNotify *self) {
  Tny_private *pv;

	pv = self->private;
	if (pv->ReScan_Run) return;
	if (!nypv_OpenDir(self)) return;
	pv->ReScan_Run = true;
	while (!nypv_ScanDir(self));
	pv->ReScan_Run = false;
	pv->rescan_us = GetTime_us();
}

/*_____________________________ Start ___________________________________*/
static void TDirListINotify_File_Start(TDirListNotify *self) {
  Tny_private *pv = self->private;

	if (pv->wd >= 0) return;
	if (!ny_appEv) ny_appEv = Add_appEv(TDirList_nyi_On_appEv);

	self->nymk = 0;
	pv->rescan_us = appDat.app_us + 3000000;

	Start_INotify(self);
	return;
}

/*_____________________________ Destroy _________________________________*/
static bool TDirListINotify_File__Destroy(TDirListNotify *self) {
  Tny_private *pv;
  TDirListNotify **ppny;

// printf("%s(%d): Destroy %s\n", FL, pv->gp.path);
	Stop_INotify(self);
	pv = self->private;

	ppny = &g_dlny;
	while (*ppny && (*ppny != self)) {
	    ppny = &((Tny_private*)(*ppny)->private)->next_dlny;
	}
	if (*ppny) *ppny = ((Tny_private*)(*ppny)->private)->next_dlny;

	if (!g_dlny && ny_appEv) {
	    Free_appEv(ny_appEv);
	    ny_appEv = NULL;
	}

	gp_pc_free(&pv->gp);
	xfree(self->private);
	return false;
}

static void TDirListINotify__AppAtExit() {
  TDirListNotify *ny = g_dlny;
	while (ny) {
	    Stop_INotify(ny);
	    ny = ((Tny_private*)ny->private)->next_dlny;
	}
}

/*_____________________________ Create __________________________________*/
static bool TDirListINotify_File_ChkType(TglobPath *gp) {
	if (!gp_test_local_drive(gp)) return False;

	if (!g_ny_fd) {
	    g_ny_fd = syscall(__NR_inotify_init);
	    if (g_ny_fd > 0) {
		g_ny_lb = TLineBuf_Create();
		AddPipeProc(g_ny_fd, InPipe, iNotify_callback, NULL);
		g_dlny = NULL;
		pp_void_add((void*)&AppProcAtExit, 1, TDirListINotify__AppAtExit);
		mount_wd = TNotifyMount_Create(NULL, NotifyMountEv);
	    } else {
		g_ny_fd = -1;
	    }
	}
	if (g_ny_fd < 0) return False;
	return True;
}

static TDirListNotify *TDirListINotify_File_Create(TDirList *dl, TOnFileMsg OnFileMsg) {
  TDirListNotify *self, **ppdlny;
  Tny_private *pv, *pvn;

	if (!TDirListINotify_File_ChkType(&dl->gp)) return NULL;

	self = ttc_create(sizeof(TDirListNotify), wtyp_TDirListNotify);
	self->destroy = (TDESTROY)TDirListINotify_File__Destroy;
	self->OnFileMsg = OnFileMsg;
	self->dl = dl;
	pv = xcalloc(sizeof(Tny_private));
	self->private = pv;
	gp_dup_gpdata_path(&pv->gp, &dl->gp);

	pv->wd = -1;
	self->start = TDirListINotify_File_Start;
	self->ReScan = TDirListNotify_File_ReScan;
	self->UseRescan = True;

	ppdlny = &g_dlny;
	while (*ppdlny) {
	    pvn = (*ppdlny)->private;
	    ppdlny = &pvn->next_dlny;
	}
	*ppdlny = self;
	return self;
}

#else /* USE_INOTIFY */
/*_____________________________ Create __________________________________*/
static TDirListNotify *TDirListINotify_File_Create(TDirList *dl, TDirList_LoaderEv dllEv) {
	return NULL;
}
#endif /* USE_INOTIFY */

#endif /* _OS_WIN32 */

const ThDirListNotify hdirlist_nyi_File = {
  TDirListINotify_File_ChkType,
  TDirListINotify_File_Create,
  0
};
