/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"
#include "tdirlist_pv.h"

static TDirList *dl_obj = NULL;

static TDirFileList *get_dfl(TDirListObjNy *self, const char *fn) {
  TDirFileList *dfl = self->last_get_dfl;

	if (dfl && pccmp(dfl->fn, fn)) return dfl;
	dfl = self->dfl;
	while (dfl && !pccmp(dfl->fn, fn)) dfl = dfl->next;
	self->last_get_dfl = dfl;
	return dfl;
}

static TDirFileList *add_dfl(TDirListObjNy *self, TDirFileList *dfl) {
  TDirFileList *dfln = NULL;
  TDirFileList **ppdfl;

	if (File_stat_Match(self->findAtr, self->patt, dfl->fn, dfl->st)) {
	    ppdfl = self->ppdfl;
	    if (!ppdfl) {
		ppdfl = &self->dfl;
		while (*ppdfl) ppdfl = &(*ppdfl)->next;
	    }
	    self->count++;
	    dfln = dfl_dup(dfl);
	    dfln->chstat = chstat_Add;
	    *ppdfl = dfln;
	    self->ppdfl = &dfln->next;
// printf("%s(%d): add_dfl 0x%x %s %s\n", FL, (int)dfln, self->dl->gp.path, dfln->fn);
	}
	return dfln;
}

/*_____________________________ appEv ___________________________________*/
static void DodflEv(TDirListObjNy *self, TDirFileList *dfl, int bef) {
  TFileMsg fmsg;

	file_msg_init_Dm(&fmsg, fmArt_DirList, bef);
	if (dfl) {
	    fmsg.typF = fmTyp_dfl;
	    fmsg.F.dfl = dfl;
	}
	if (self->dl) {
	    fmsg.typR = fmTyp_gp;
	    fmsg.R.gp = &self->dl->gp;
	}
	file_msg_ev_next(&fmsg, (TTC0*)self, self->OnFileMsg);
}

static bool dfl_clr(TDirListObjNy *self) {
  TDirFileList *dfl;

	if (!self->dfl) return False;
	DodflEv(self, NULL, dlbef_Clr);
	while (self->dfl) {
	    dfl = self->dfl;
	    self->dfl = dfl->next;
	    xfree(dfl);
	}
	self->count = 0;
	return True;
}

static int dflEv(TDirList *self, TFileMsg *fmsg) {
  TDirListObjNy *ObjNy, *ObjNy_next;
  TDirFileList *dfl;
  int bef, befn, i;
  char *pc;
  TFileMsg r_fmsg;

	ObjNy = self->ObjNy;
	if ((fmsg->art != fmArt_DirList) ||
	    (fmsg->typD != fmTyp_int)) {

	    while (ObjNy) {
		ObjNy_next = ObjNy->next;
		file_msg_ev_next(fmsg, (TTC0*)ObjNy, ObjNy->OnFileMsg);
		ObjNy = ObjNy_next;
	    }
	    return 0;
	}

	pnmove(&r_fmsg, fmsg, sizeof(TFileMsg));
	
	bef = fmsg->D.m;
	befn = bef & dlbef_BefMask;
	dfl = fmsg->F.dfl;

// printf("%s(%d): dflEv 0x%x\n", FL, (int)ObjNy);

	if ((befn == dlbef_Del) && (dfl->st->st & st_Link)) {
// printf("%s(%d): TDirListObjNy_dllEv Del Link %s\n", FL, dfl->fn);

  TDirList *dl = dl_obj;
  char *xpath = pcadd_n(3, self->gp.path, dfl->fn, "/");
  int lg = pclen(xpath);
  int maxp = 0;
  TDirList *dlnext;

	    while (dl) {
		if (pcncmp(xpath, dl->gp.path, lg)) {
// printf("%s(%d): TDirListObjNy_dllEv Del Link %s\n", FL, dl->gp.path);
		    i = 0;
		    pc = dl->gp.path + (lg-1);
		    while (*pc) {
			if (*pc == '/') i++;
			pc++;
		    }
		    if (i > maxp) maxp = i;
		}
		dl = dl->next;
	    }

	    while (maxp) {
	      dl = dl_obj;
	      while (dl) {
		dlnext = dl->next;
		if (pcncmp(xpath, dl->gp.path, lg)) {
		    i = 0;
		    pc = dl->gp.path + (lg-1);
		    while (*pc) {
			if (*pc == '/') i++;
			pc++;
		    }
		    if (i == maxp) {

  TFileMsg tmp_fmsg;
			file_msg_init_Dm(&tmp_fmsg, fmArt_DirList, dlbef_DelSelf);
			tmp_fmsg.typR = fmTyp_gp;
			tmp_fmsg.R.gp = r_fmsg.R.gp;
			dflEv(dl, &tmp_fmsg);
		    }
		}
		dl = dlnext;
	      }
	      maxp--;
	    }
	    xfree(xpath);
	}

	while (ObjNy) {
	    ObjNy_next = ObjNy->next;
	    r_fmsg.F.dfl = NULL;

	    switch (befn) {

	    case dlbef_Add:
// printf("%s(%d): dlbef_Add    0x%x %s\n", FL, (int)ObjNy, dfl->fn);
		if (!dfl) goto next;
		r_fmsg.F.dfl = add_dfl(ObjNy, dfl);
		if (!r_fmsg.F.dfl) goto next;
		ObjNy->chstat |= chstat_Add;
		break;

	    case dlbef_Del:
		if (!dfl) goto next;
// printf("%s(%d): dlbef_Del    0x%x %s\n", FL, (int)ObjNy, dfl->fn);
		r_fmsg.F.dfl = get_dfl(ObjNy, dfl->fn);
		if (!r_fmsg.F.dfl) goto next;
		ObjNy->chstat |= chstat_Del;
		ObjNy->count--;
		ObjNy->ppdfl = NULL;
		ObjNy->last_get_dfl = NULL;
  TDirFileList **ppdfl = get_ppdfl(&ObjNy->dfl, r_fmsg.F.dfl);
		*ppdfl = r_fmsg.F.dfl->next;
		file_msg_ev_next(&r_fmsg, (TTC0*)ObjNy, ObjNy->OnFileMsg);
		xfree(r_fmsg.F.dfl);
		goto next;

	    case dlbef_Change:
		if (!dfl) goto next;
		if (ObjNy->chstat_Filter && !
                    (ObjNy->chstat_Filter & dfl->chstat)) goto next;
// printf("%s(%d): dlbef_Change 0x%08x ch=0x%04x %s\n", FL, (int)ObjNy, dfl->chstat, dfl->fn);
		r_fmsg.F.dfl = get_dfl(ObjNy, dfl->fn);
		if (!r_fmsg.F.dfl) {
		    r_fmsg.F.dfl = add_dfl(ObjNy, dfl);
		    if (r_fmsg.F.dfl) {
			r_fmsg.D.m = (bef & ~dlbef_BefMask) | dlbef_Add;
			file_msg_ev_next(&r_fmsg, (TTC0*)ObjNy, ObjNy->OnFileMsg);
			r_fmsg.D.m = bef;
		    }
		    goto next;
		}
// printf("%s(%d): dlbef_Change 0x%08x ch=0x%04x %s\n", FL, (int)ObjNy, dfl->chstat, dfl->fn);
		ObjNy->chstat |= dfl->chstat;
		stat_test_copy_uid_gid(r_fmsg.F.dfl->st, dfl->st);
		pnmove(r_fmsg.F.dfl->st, dfl->st, sizeof(Tstat64));
		break;

	    case dlbef_WrClose:
// printf("%s(%d): dlbef_WrClose 0x%x %s\n", FL, (int)ObjNy, dfl->fn);
		if (!dfl) goto next;
		r_fmsg.F.dfl = get_dfl(ObjNy, dfl->fn);
		break;

	    case dlbef_Clr:
// printf("%s(%d): dlbef_Clr\n", FL);
		ObjNy->chstat |= chstat_Del;
		dfl_clr(ObjNy);
		goto next;

	    case dlbef_DelSelf:
// printf("%s(%d): dlbef_DelSelf\n", FL);
		r_fmsg.Ok = False;
		file_msg_ev_next(&r_fmsg, (TTC0*)ObjNy, ObjNy->OnFileMsg);
// printf("%s(%d): dlbef_DelSelf fmsg.Ok=%d\n", FL, r_fmsg.Ok);
		if (!r_fmsg.Ok) Tfree(ObjNy);
		r_fmsg.Ok = True;
		goto next;

	    }
	    file_msg_ev_next(&r_fmsg, (TTC0*)ObjNy, ObjNy->OnFileMsg);
next:
	    ObjNy = ObjNy_next;
	}
	return 1;
}

static void TDirListObjNy_OnFileMsg(TFileMsg *fmsg) {
	if (!file_msg_chk_wtyp(fmsg, wtyp_TDirList, FL)) return;
// printf("%s(%d): %d(0x%x)\n", FL, fmsg->D.m & dlbef_BefMask, fmsg->D.m);
	dflEv(fmsg->obj, fmsg);
}

/*_____________________________ Sort ____________________________________*/
void TDirListObjNy_Sort(TDirListObjNy *self, int fu) {
	self->ppdfl = NULL;
	self->last_get_dfl = NULL;
	TDirFileList_Sort(&self->dfl, fu);
}

/*_____________________________ ReScan  _________________________________*/
void TDirListObjNy_ReScanMime() {
  TDirList *dl;

	dl = dl_obj;
	while (dl) {
	    TDirList_ReScanMime(dl);
	    dl = dl->next;
	}
}

void TDirListObjNy_ReScan(TDirListObjNy *self) {
	if (!self || !self->dl) return;
	TDirList_ReScan(self->dl);
}
/*_____________________________ misc ____________________________________*/
TglobPath *TDirListObjNy_Get_gp(TDirListObjNy *self) {
	if (!self->dl) return NULL;
	return &self->dl->gp;
	
}

TDirList *TDirListObjNy_Enum_TDirList(TDirList *dl) {
	if (!dl) return dl_obj;
	return dl->next;
}

/*_____________________________ Open ____________________________________*/
TDirList *TDirListObjNy_Get_dl_gp(TglobPath *gp) {
  TDirList *dl;
	dl = dl_obj;
	while (dl) {
	    if (!pccasecmpASCII(dl->gp.RootArt, gp->RootArt) &&
		pccmp(dl->gp.URL, gp->URL) &&
		pccmp(dl->gp.user, gp->user) &&
		pccmp(dl->gp.path, gp->path)) break;
	    dl = dl->next;
	}
	return dl;
}

static void DelObjNyFromDirList(TDirListObjNy *self, TDirList *dl) {
  TDirListObjNy **ppObjNy;

	if (ttc_chk_struct(dl)) {
	    ppObjNy = &dl->ObjNy;
	    while (*ppObjNy && (*ppObjNy != self)) ppObjNy = &(*ppObjNy)->next;
	    *ppObjNy = self->next;
	}
}

static void DelFrom_dl_obj(TDirList *dl) {
  TDirList **ppdl;
	ppdl = &dl_obj;
	while (*ppdl != dl) ppdl = &(*ppdl)->next;
	*ppdl = dl->next;
	dl->OnFileMsg = NULL;
}

bool TDirListObjNy_Open_gp(TDirListObjNy *self, TglobPath *gp) {
  TDirList *dl, **ppdl;
  TDirFileList *dfl, *dfln;
  bool AddToList;

	if (!self) return False;

	if (dfl_clr(self)) DodflEv(self, NULL, dlbef_ClrEnd);
	self->ppdfl = &self->dfl;

// printf("%s(%d): TDirListObjNy_Open_gp %s %s %s %s %s\n", FL, gp->RootArt, gp->user, gp->passwd, gp->URL, gp->path);
// printf("%s(%d): TDirListObjNy_Open_gp %s %s\n", FL, gp->RootArt, gp->path);
/* Reg. TDirListObjNy */
	AddToList = False;
	dl = self->dl;
	if (!dl) {
	    if (!gp) return False;
	    AddToList = True;

	    dl = TDirListObjNy_Get_dl_gp(gp);
	    if (dl) {
		self->dl = dl;
// printf("%s(%d): dlbef_Start 0x%x\n", FL, self->OnFileMsg);
		DodflEv(self, NULL, dlbef_ScanRun|dlbef_Start);

	    } else {
		dl = TDirList_Create_gp(gp);
		self->dl = dl;
// printf("%s(%d): dlbef_Start 0x%x\n", FL, self->OnFileMsg);
		DodflEv(self, NULL, dlbef_ScanRun|dlbef_Start);

		ppdl = &dl_obj;
		while (*ppdl) ppdl = &(*ppdl)->next;
		*ppdl = dl;
dl_restart:
		if (!TDirList_Open(dl, dlbef_RescanRun|dlbef_NotifyRun)) {

  TFileMsg fmsg;
		    DodflEv(self, NULL, dlbef_ScanRun|dlbef_Start);
		    Fill0(&fmsg, sizeof(TFileMsg));
		    fmsg.art = fmArt_errLogin;
		    fmsg.typD = fmTyp_char;
		    fmsg.D.s = pcadd("Open: ", gp->path);
		    file_msg_ev_next(&fmsg, (TTC0*)self, self->OnFileMsg);
		    xfree(fmsg.D.s);
		    if (!dl->ObjNy) {
			DodflEv(self, NULL, dlbef_DelSelf);
			DelFrom_dl_obj(dl);
			self->dl = NULL;
			Tfree(dl);
		    } else {
 printf("%s(%d): dl->ObjNy not NULL ???\n", FL);
		    }
		    return False;
		}
	    }
	}

// printf("%s(%d): dl->runstat=%d\n", FL, dl->runstat);
	if (dl->runstat > 0) {
	    DodflEv(self, NULL, dlbef_ScanRun|dlbef_Open);
	} else if (dl->runstat < 0) {
// printf("%s(%d): dl_restart\n", FL);
	    goto dl_restart;
	} 

	self->chstat |= chstat_Add;
	dfl = dl->dfl;
	while (dfl) {
#if 0
	    TDirFileList *dflx = self->dfl;
	    while (dflx && !pccmp(dflx->fn, dfl->fn)) dflx = dflx->next;
	    if (!dflx) {
#endif
		dfln = add_dfl(self, dfl);
		if (dfln) DodflEv(self, dfln, dlbef_ScanRun|dlbef_Add);
#if 0
	    } else  printf("%s(%d): dfl Exist %s\n", FL, dfl->fn);
#endif
	    dfl = dfl->next;
	}

	if (AddToList) {
	    dl->OnFileMsg = TDirListObjNy_OnFileMsg;
	    TDirListObjNy **ppdlo = &dl->ObjNy;
	    while (*ppdlo) ppdlo = &(*ppdlo)->next;
	    *ppdlo = self;
	}

	if (dl->runstat >= dlrunstat_ScanEnd) {
	    DodflEv(self, NULL, dlbef_ScanRun|dlbef_ListEnd);
	    DodflEv(self, NULL, dlbef_ScanRun|dlbef_Close);
	}
	if (dl->runstat >= dlrunstat_ReScanEnd) {
	    DodflEv(self, NULL, dlbef_RescanRun|dlbef_Close);
	}
	return True;
}

bool TDirListObjNy_Open(TDirListObjNy *self, const char *fn) {
  TglobPath *gp;
  bool res;

	gp = pctogp(fn);
	gp_NameToPath(gp);
	res = TDirListObjNy_Open_gp(self, gp);
	gpFree(gp);
	return res;
}

void TDirListObjNy_Open_Wait(TDirListObjNy *self) {
  int TempWait;
	if (!self || (self->dl->runstat < 0)) return;
	if (self->dl->runstat < dlrunstat_ScanEnd) {
	    AppSetWait0_single();
	    TempWait = 1;
	    self->pTempWait = &TempWait;
	    while (AppHandleMessage() 
		   && TempWait
		   && (self->dl->runstat >= 0) 
		   && (self->dl->runstat < dlrunstat_ScanEnd));
	    if (TempWait) self->pTempWait = NULL;
	}
}

/*_____________________________ Destroy _________________________________*/
static bool TDirListObjNy__Destroy(TDirListObjNy *self) {
  TDirList *dl;
#if 0
 printf("%s(%d): ObjNy__Destroy 0x%x", FL, (int)self);
 if (self->dl) printf(" %s", self->dl->gp.path);
 printf("\n");
#endif
	if (dfl_clr(self)) {
	    DodflEv(self, NULL, dlbef_ClrEnd);
	}

	DodflEv(self, NULL, dlbef_Destroy);

	if (self->pTempWait) *self->pTempWait = 0;

	if (self->ppObjNy && *self->ppObjNy) {
	    if (!ttc_chk_is_wtyp(*self->ppObjNy, wtyp_TDirListObjNy)) {
		printf("%s(%d): ERROR: ttc not TDirListObjNy ???", FL);
	    }
	    *self->ppObjNy = NULL;
	}

	dl = self->dl;
	if (dl) {
	    DelObjNyFromDirList(self, dl);
	    if (!dl->ObjNy) {
		DelFrom_dl_obj(dl);
		Tfree(dl);
	    }
	}

	if (!(self->fu & dlObjNyfu_pattNotFree)) Tfree(self->patt);
	return False;
}

/*_____________________________ Create __________________________________*/
TDirListObjNy *TDirListObjNy_Create(void *owner) {
  TDirListObjNy *self;

	self = ttc_create(sizeof(TDirListObjNy), wtyp_TDirListObjNy);
	self->destroy = (TDESTROY)TDirListObjNy__Destroy;
	self->owner = owner;
// printf("%s(%d): TDirListObjNy_Create 0x%x\n", FL, (int)self);
	return self;
}

bool TDirListObjNy_Create_gp(TDirListObjNy **ppObjNy, void *owner, 
				TglobPath *gp, TOnFileMsg OnFileMsg) {
  TDirListObjNy *self;

	if (!gp || !file_chk_exist_gp(gp, findALL)) return False;
	self = TDirListObjNy_Create(owner);
	self->OnFileMsg = OnFileMsg;
	self->ppObjNy = ppObjNy;
	TDirListObjNy_Open_gp(self, gp);
	if (ppObjNy) *ppObjNy = self;
	return True;
}
