/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "xtty.h"
#include <unistd.h>

// #define TTY_IN_DEBUG	1


void Ttty_NewLine(Ttty *tty, char *wrs, int fu) {
  int lg;
  char *s = NULL;

	if (fu == 2) {
	    ppcnadd(&s, "\x0a", 1);	//lf
	} else {
	    ppcnadd(&s, "\x0d\x1b[K", 4);  //cr + EraseLine
	}

	ppcadd(&s, tty->linehead);
	if (s) TTermStr_ColOut(tty->termstr, s, mCOLOR_BLUE, mCOLOR_DEFAULT);
	xfree(s);
	TTermStr_ColOut(tty->termstr, "> ", mCOLOR_DEFAULT, mCOLOR_DEFAULT);

	lg = pclen(wrs);
	if (lg > tty->KeyLineSize) lg = tty->KeyLineSize;
	TTermStr_ColOut(tty->termstr, wrs, mCOLOR_DEFAULT, mCOLOR_DEFAULT);

	tty->KeyIdx = lg;
	tty->MaxKeyIdx = lg;
	pcnmove(tty->KeyLine, wrs, lg);
}

static void ttyIn(int fd, Tfd_list *fdl) {
  int count, i;
  char buf[2048];
  Ttty *tty = fdl->obj;
  TTermscan *term = tty->term;

	count = read(fd, &buf[0], 2048);

#ifdef	TTY_IN_DEBUG
	for (i=0; i<count; i++) fprintf(stderr, "%02x", (byte)buf[i]);
	fprintf(stderr, "\n");
	for (i=0; i<count; i++) {
	    if (term_scan(buf[i], term)) {
		fprintf(stderr, "XK=0x%04x\n", term->btty_XK);
	}  }
#else
	for (i=0; i<count; i++) {
	    if (term_scan(buf[i], term) && (!term->err)) {
		if (tty->OnKey) tty->OnKey(tty);
	    }
	}
#endif
}

static bool Ttty__Destroy(Ttty *tty) {
	TTermStr_ColOut(tty->termstr, "\x0a", mCOLOR_DEFAULT, mCOLOR_DEFAULT);
	tcsetattr (tty->fdin, TCSADRAIN, &tty->org_termios);
	xfree(tty->KeyLine);
	xfree(tty->linehead);
	TTermscan_Destroy(tty->term);
	return False;
}

Ttty *Ttty_Create(int fdin, int fdout, const char *linehead) {
  struct termios buf;
  Ttty *tty;

	if (!isatty(STDIN_FILENO)) return NULL;
//      fprintf (stderr, "Not a terminal.\n");

	tty = ttc_create(sizeof(Ttty), wtyp_Ttty);
	tty->destroy = (TDESTROY)Ttty__Destroy;

	tty->fdin = fdin;
	tty->fdout = fdout;
	if (tcgetattr (fdin, &buf) >= 0) {
	    tty->KeyLine = xcalloc(2048);
	    tty->KeyLineSize = 2048;
	    tty->linehead = pcdup(linehead);

	    tcgetattr(fdin, &tty->org_termios);
	    buf.c_lflag &= ~(ECHO | ICANON);
	    tcsetattr (fdin, TCSADRAIN, &buf);

	    Tfd_list *fd_list;
	    fd_list = AddPipeProc_TmpRes(fdin, InPipe, ttyIn, tty);
	    fd_list->io = fdopen(fdin, "r");

	    tty->term = TTermscan_Create();
	    tty->termstr = TTermStr_Create(fdout, tty->term->termart);
	    Ttty_NewLine(tty, NULL, 0);
	    return tty;
	}
	Tfree(tty);
	return NULL;
}
