/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "xtty.h"
#include <unistd.h>

static void Ttty_AddLine(Ttty *tty) {
  int i;

	if (tty->MaxLastLineIdx > 0) {
	    if (pccmp(tty->LastKeyLines[tty->MaxLastLineIdx-1], tty->KeyLine)) {
		tty->LastLineIdx = tty->MaxLastLineIdx;
		return;
	    }
	}

	if (tty->MaxLastLineIdx < MaxTermLines) {
	    ppcnchange(&tty->LastKeyLines[tty->MaxLastLineIdx], tty->KeyLine, tty->MaxKeyIdx);
	    tty->MaxLastLineIdx++;
	    tty->LastLineIdx = tty->MaxLastLineIdx;
	} else {
	    xfree(tty->LastKeyLines[0]);
	    for (i=0; i<MaxTermLines; i++) tty->LastKeyLines[i] = tty->LastKeyLines[i+1];
	    ppcnchange(&tty->LastKeyLines[MaxTermLines-1], tty->KeyLine, tty->MaxKeyIdx);
	    tty->LastLineIdx = MaxTermLines;
	    tty->LastKeyLines[MaxTermLines] = NULL;
	}
}

void Ttty_LineKey(Ttty *tty) {
  TTermscan *term = tty->term;
  int x, wrl;
  char *wrs;
  char c_BackSpace[5] = {0x08,0x1b,'[','1','P'};
  char Sendbuf[64];

	if (tty->InputStop) return;

	wrl = term->reslg;
	wrs = term->res;
	if (term->btty_XK == yXK_BackSpace) {
	    if (tty->KeyIdx) {
		wrs = c_BackSpace;
		wrl = 5;
		tty->KeyIdx--;
		tty->MaxKeyIdx--;
		for (x=tty->KeyIdx; x<tty->MaxKeyIdx; x++) {
		    tty->KeyLine[x] = tty->KeyLine[x+1];
		}
	    }

	} else if (term->btty_XK == yXK_Up) {
	    if (tty->LastLineIdx) {
		if (tty->MaxLastLineIdx == tty->LastLineIdx) {
		    ppcnchange(&tty->LastKeyLines[tty->MaxLastLineIdx], tty->KeyLine, tty->MaxKeyIdx);
		}
		tty->LastLineIdx--;
		wrs = tty->LastKeyLines[tty->LastLineIdx];
		Ttty_NewLine(tty, wrs, 1);
		wrl = 0;
	    } else wrl = 0;

	} else if (term->btty_XK == yXK_Down) {
	    if (tty->LastLineIdx < tty->MaxLastLineIdx) {
		tty->LastLineIdx++;
		wrs = tty->LastKeyLines[tty->LastLineIdx];
		Ttty_NewLine(tty, wrs, 1);
		wrl = 0;
	    } else wrl = 0;

	} else if (term->btty_XK == yXK_Left) {
	    if (tty->KeyIdx) tty->KeyIdx--; else wrl = 0;

	} else if (term->btty_XK == yXK_Right) {
	    if (tty->KeyIdx < tty->MaxKeyIdx) tty->KeyIdx++; else wrl = 0;

	} else if (term->btty_XK == yXK_Return) {
	    TTermStr_lf(tty->termstr);
	    tty->KeyLine[tty->MaxKeyIdx] = 0;
	    Ttty_AddLine(tty);
	    if (tty->OnReturn)
		    tty->OnReturn(tty, tty->KeyLine, tty->MaxKeyIdx);
	    if (!appDat.gQUIT_Ev && !tty->InputStop) {
		    Ttty_NewLine(tty, NULL, 0);
	    }
	    wrl = 0;

	} else if ((term->btty_XK >= 20) && (term->btty_XK < 256)) {
	    if (tty->MaxKeyIdx < tty->KeyLineSize-1) {
		if (tty->MaxKeyIdx == tty->KeyIdx) {
		    wrs = (char*)&term->btty_XK;
		    tty->KeyLine[tty->KeyIdx] = term->btty_XK;
		    wrl = 1;
		} else {
		    for (x=tty->MaxKeyIdx; x>tty->KeyIdx; x--) {
			    tty->KeyLine[x] = tty->KeyLine[x-1];
		    }
		    tty->KeyLine[tty->KeyIdx] = term->btty_XK;
		    Sendbuf[0] = 0x1b;
		    Sendbuf[1] = '[';
		    Sendbuf[2] = '1';
		    Sendbuf[3] = '@';
		    Sendbuf[4] = term->btty_XK;
		    wrl = 5;
		    wrs = Sendbuf;
		}
		tty->KeyIdx++;
		tty->MaxKeyIdx++;
	    } else wrl = 0;
	} else {
		wrl = 0;
	}
	if (wrl) write(1, wrs, wrl);
}
