/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "xcopy_pv.h"

ThXCopy **hxcopy = NULL;

bool XCopy_Start(TXCopy_wd *wd, Tstat64 *st_src, int xcySFU) {
  int bef;

	if (!wd) return False;
	wd->stat = 0;
	wd->xcyChk = 0;
	if (wd->xcy_src->clear) wd->xcy_src->clear(wd);
	if (wd->xcy_dest && wd->xcy_dest->clear) wd->xcy_dest->clear(wd);

	if (!wd->ges_us) wd->ges_us = GetTime_us64();

	bef = xcySFU & xcySFU_befmask;
	wd->us = GetTime_us();
	wd->xcy_src->st = tstat_dup(st_src);

	if (wd->xcy_dest) {
	    if ((bef == xcySFU_Preserv) && st_src) {
		wd->xcy_src->uid = st_src->uid;
		wd->xcy_src->gid = st_src->gid;
		wd->xcy_src->atime = st_src->atime;
		wd->xcy_src->mtime = st_src->mtime;
		wd->xcy_src->stmode = st_src->mode & Mode07777;
		wd->stat = xcywdstat_wr_run;
		if (wd->xcy_dest->init(wd, MsgSt_Preserv)) return True;

	    } else if (bef == xcySFU_Del) {
		wd->stat = xcywdstat_rd_run;
		if (wd->xcy_src->init(wd, MsgSt_Del)) return True;

	    } else {
		wd->stat = xcywdstat_run;
		if (wd->xcy_src->init(wd, 0)) return True;
	    }
	} else {
	    if (bef == xcySFU_Del) {
		wd->stat = xcywdstat_rd_run;
		if (wd->xcy_src->init(wd, MsgSt_Del)) return True;
	    }
	}

	wd->stat = 0;
	return False;
}

bool XCopy_ChkWait(TXCopy_wd *wd) {
	if (!wd) return False;

	return  ((wd->stat & (xcywdstat_rd_run|xcywdstat_wr_run)) != 0) &&
		((wd->stat & (xcywdstat_rd_err | xcywdstat_wr_err)) == 0);
}

bool XCopy_ChkErr(TXCopy_wd *wd) {
	if (!wd) return False;
	return (wd->stat & (xcywdstat_rd_err | xcywdstat_wr_err)) != 0;
}

void XCopy_SetMode(TXCopy_wd *wd, dword mode) {
	wd->mode = mode;
}
dword XCopy_GetMode(TXCopy_wd *wd) {
	return wd->mode;
}

void XCopy_SetSkip(TXCopy_wd *wd, dword skip) {
	wd->skip = skip;
}
dword XCopy_GetSkip(TXCopy_wd *wd) {
	return wd->skip;
}

void XCopy_MsgSetSkip(TXCopy_wd *wd, int xcyMsg, dword skip, bool all) {
	if (xcyMsg & xpyMsg_src) {
	    if (skip) {
		wd->xcy_src->msgOk_skip = skip;
	    } else {
		skip = wd->xcy_src->msgOk_skip;
	    }
	} else {
	    if (skip) {
		wd->xcy_dest->msgOk_skip = skip;
	    } else {
		skip = wd->xcy_dest->msgOk_skip;
	    }
	}
	if (all) wd->skip |= skip;
}

const char *XCopy_GetErr_pc(TXCopy_wd *self, int xcyMsg) {
	if (!(xcyMsg & xpyErr_Msg)) return NULL;
	if (xcyMsg & xpyMsg_src) {
	    return self->xcy_src->msg_s;
	} else if (xcyMsg & xpyMsg_dest) {
	    return self->xcy_dest->msg_s;
	}
	return NULL;
}

/* char *XCopy_GetAbsFileName_pc(TXCopy_wd *self, TXCopy_ld *ld) {
	if (!ld) return NULL;
	return ld->fname;
}

char *XCopy_GetFileName_pc(TXCopy_wd *self, TXCopy_ld *ld) {
	if (!ld) return NULL;
	return ld->gp.name;
}

void XCopy_SetFileName(TXCopy_wd *self, TXCopy_ld *ld, char *fname) {
	if (!ld) return;
	ppcchange(&ld->gp.name, fname);
} */

void XCopy_NameToPath(TXCopy_wd *self) {
  TglobPath *gp = &self->xcy_src->gp;
	if (gp->name) ppcadd2(&gp->path, gp->name, "/");
	ppcfree(&gp->name);

	if (self->xcy_dest) {
	    gp = &self->xcy_dest->gp;
	    if (gp->name) ppcadd2(&gp->path, gp->name, "/");
	    ppcfree(&gp->name);
	}
}

void XCopy_PathToName(TXCopy_wd *self) {
	gp_PathToName(&self->xcy_src->gp);
	if (self->xcy_dest) gp_PathToName(&self->xcy_dest->gp);
}

void XCopy_ChangeName(TXCopy_wd *self, const char *fname) {
	ppcchange(&self->xcy_src->gp.name, fname);
	if (self->xcy_dest) ppcchange(&self->xcy_dest->gp.name, fname);
}

void XCopy_ChangeDestName(TXCopy_wd *self, const char *fname) {
	ppcchange(&self->xcy_dest->gp.name, fname);
}

Tstat64 *XCopy_GetSrcStat(TXCopy_wd *self) {
	return tstat_dup(self->xcy_src->st);
}
Tstat64 *XCopy_GetDestStat(TXCopy_wd *self) {
	if (!self->xcy_dest) return NULL;
	return tstat_dup(self->xcy_dest->st);
}

TglobPath *XCopy_GetSrc_gp(TXCopy_wd *self) {
	return gp_dup(&self->xcy_src->gp);
}
TglobPath *XCopy_GetDest_gp(TXCopy_wd *self) {
	if (!self->xcy_dest) return NULL;
	return gp_dup(&self->xcy_dest->gp);
}

char *XCopy_GetSrc_gp_fn(TXCopy_wd *self, int gptfu) {
	return gptopc(&self->xcy_src->gp, gptfu);
}

char *XCopy_GetDest_gp_fn(TXCopy_wd *self, int gptfu) {
	if (!self->xcy_dest) return NULL;
	return gptopc(&self->xcy_dest->gp, gptfu);
}

bool XCopy_ChkDestDirDouble(TXCopy_wd *self) {
  char *pc = pcadd(self->xcy_src->gp.path, self->xcy_src->gp.name);
  bool res = chk_dest_ok(self, pc);
	xfree(pc);
	return res;
}

void *XCopy_GetOwner(TXCopy_wd *self) {
	if (!self) return NULL;
	return self->owner;
}

bool XCopy_GetStats(TXCopy_wd *self, TXCopyStats *xcyStats) {

	if (!self || !self->xcy_src || !xcyStats) return False;
	*xcyStats = self->xcs;
	return True;
}

bool XCopy_DestDirIsCreate(TXCopy_wd *self) {
	return self && (!(self->xcyChk & xcyChkDest_NoDirCreate));
}

void XCopy_Stop(TXCopy_wd *self) {

	if (!self) return;
	if (self->stat & xcywdstat_rd_run) {
	    if (self->xcy_src->close) self->xcy_src->close(self, MsgSt_skip);
	}
	if (self->stat & xcywdstat_wr_run) {
	    if (self->xcy_dest->close) self->xcy_dest->close(self, MsgSt_skip);
	}
//	self->stat &= ~xcywdstat_run;
}
/*_____________________________ Chk _____________________________________*/
ThXCopy *XCopy_GetHande(TglobPath *gp, bool dest) {
  ThXCopy **hxcy;

	if (!gp || !hxcopy) return False;
	hxcy = hxcopy;
	while (*hxcy && !(*hxcy)->ChkType(gp, dest)) hxcy++;
	return *hxcy;
}

/*_____________________________ Set _____________________________________*/
bool TXCopy_SetSrc(TXCopy_wd *self, TglobPath *gp_src) {
  ThXCopy *hxcy_src;

	if (!self) return False;
	if (self->xcy_src) {
	    if (!self->xcy_src->set_gp) return False;
	    if (self->xcy_src->set_gp(self->xcy_src, gp_src)) return True;
	    Tfree(self->xcy_src);
	    self->xcy_src = NULL;
	}

	hxcy_src = XCopy_GetHande(gp_src, False);
	if (!hxcy_src) return False;
	self->xcy_src = hxcy_src->create(xcyFU_INPUT, self, gp_src);
	if (!self->xcy_src || !self->xcy_src->init) goto Err;
	if (self->xcy_src->connect && !self->xcy_src->connect(self)) goto Err;
	return True;

Err:
	ppTfree(&self->xcy_src);
	return False;
}

/*_____________________________ Destroy _________________________________*/
static bool TXCopy__Destroy(TXCopy_wd *self) {
	ppTfree(&self->xcy_src);
	ppTfree(&self->xcy_dest);

	clr_dest_dir(self);
	return False;
}

/*_____________________________ Create __________________________________*/
TXCopy_wd *XCopy_Create(void *owner, ThCopyEv hcyEv,
			TglobPath *gp_src, TglobPath *gp_dest, int skip) {

  ThXCopy *hxcy_src, *hxcy_dest;
  TXCopy_wd *self;

	hxcy_src = NULL;
	if (gp_src) {
	    hxcy_src = XCopy_GetHande(gp_src, False);
	    if (!hxcy_src) return NULL;
	}
	hxcy_dest = NULL;
	if (gp_dest) {
	    hxcy_dest = XCopy_GetHande(gp_dest, True);
	    if (!hxcy_dest) return NULL;
	}

	self = ttc_create(sizeof(TXCopy_wd), wtyp_TXcopy);
	self->destroy = (TDESTROY)TXCopy__Destroy;
	self->owner = owner;
	self->hcyEv = hcyEv;
	self->skip = skip;
	self->MaxBufSize = 0x00100000;

	if (gp_src) {
	    self->xcy_src = hxcy_src->create(xcyFU_INPUT, self, gp_src);
	    if (!self->xcy_src || !self->xcy_src->init) goto Err;
	    if (self->xcy_src->connect && !self->xcy_src->connect(self)) goto Err;
	}

	if (gp_dest) {
	  self->xcy_dest = hxcy_dest->create(xcyFU_OUTPUT, self, gp_dest);
	  if (!self->xcy_dest || !self->xcy_dest->init) goto Err;
	  if (self->xcy_dest->connect && !self->xcy_src->connect(self)) goto Err;
	}
#ifndef _OS_WIN32
	self->uid = getuid();
	self->gid = getgid();
#endif
	self->mode_dir = 0755;
	self->mode_file = 0644;
	self->mode_blk = 0640;
	return self;

Err:
	Tfree(self);
	return NULL;
}

void ThXCopy_Reg(int n, ...) {
  va_list args;
	va_start(args, n);
	pp_void_va_add((void*)&hxcopy, n, args);
	va_end(args);
}
