/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"
#include "xcopy_pv.h"
#include <errno.h>


bool MSG_Init_RD(TXCopy_wd *wd, int MsgSt, bool isOpen) {
  TXCopy_ld *ld = wd->xcy_src;

	if (!ld || (!wd->xcy_dest && (MsgSt != MsgSt_Del)) || isOpen) return False;

	if (MsgSt == MsgSt_FileStatRet) {
// printf("%s(%03d): %7d RD MSG_Init_RD MsgSt_FileStatRet\n", FL, GetCurTimeDiv_us(wd->us));
	    ld->srt = srt_ChkStart;

	} else if (MsgSt == MsgSt_Del) {
	    ld->init_send = True;
	    xcyld_SetFileName(ld);
	    ld->srt = srt_del;
	    ld->skip = wd->skip;

	} else {
	    ld->srt = srt_init;
	    ld->skip = wd->skip;
	}

	return True;
}

/* __________________________ stat ______________________________________ */
static bool chk_link_skip(Tstat64 *st, int skip) {
#ifdef _OS_WIN32
	return False;
#else
	return (((st->mode & S_IFMT) == S_IFLNK) ||
		((st->st & st_Link) && !(skip & xcyForce_followlinks)));
#endif
}

bool chk_stat_Ok_RD(TXCopy_ld *ld, int fsfu) {
	if (!ld->st && ld->stat_gp) ld->st = ld->stat_gp(&ld->gp, NULL, fsfu);
	if (ld->st) return True;
	if (ld->skip & xcySkip_SrcNotExist) {
	    ld->srt = srt_skip;
	} else {
	    ld->msgOk_skip = xcySkip_SrcNotExist;
	    ld->msg_nr = xpyErr_Alg | xpyErr_SrcNotExist;
	    ld->srt |= srt_msg_wait;
	}
	return False;
}

void step_Init_RD(void *obj, TXCopy_wd *wd) {
  TXCopy_ld *ld = wd->xcy_src;

	if (!ld) return;

	if (!chk_stat_Ok_RD(ld, fsFollowLink)) return;

	if (ld->skip & xcySkip_OldFile) {
	    if (!chk_link_skip(ld->st, ld->skip) && (ld->st->st & st_Reg)) {
		ld->srt = srt_Wait;
		ld->send_Mt(obj, thdMSG_XCopy_RunStat, MsgSt_FileStatGet);
		return;
	    }
	}
	ld->srt = srt_Start;
}

/* __________________________ Start _____________________________________ */
void step_ChkStart_RD(void *obj, TXCopy_wd *wd) {
  TXCopy_ld *src = wd->xcy_src;
  TXCopy_ld *dest = wd->xcy_dest;
  int xcyChk, tmp;

	if (!src || !dest) {
	    src->srt = srt_0;
	    return;
	}
	xcyChk = xcyChkDest(src, dest, wd->skip, &tmp, &tmp);
// printf("%s(%03d): %7d pth RD xcyChk=0x%x\n", FL, GetCurTimeDiv_us(wd->us), xcyChk);
	if ((xcyChk & xcyChkDest_BefMask) == xcyChkDest_Skip) {
	    src->send_Mt(obj, thdMSG_XCopy_RunStat, MsgSt_skip);
	    src->srt = srt_0;
	    return;
	}
	src->srt = srt_Start;
}

void step_Start_RD(void *obj, TXCopy_wd *wd) {
  TXCopy_ld *ld = wd->xcy_src;

	if (!ld) return;
	ld->uid = ld->st->uid;
	ld->gid = ld->st->gid;
	ld->atime = ld->st->atime;
	ld->mtime = ld->st->mtime;
	ld->stmode = ld->st->mode & Mode07777;
	xcyld_SetFileName(ld);
// printf("%s(%3d): %6d RD Init %s\n", FL, GetCurTimeDiv_us(wd->us), ld->fname);

	if (!chk_link_skip(ld->st, ld->skip) && (ld->st->st & st_Reg)) {
	    ld->lb->eof = False;
	    ld->lb->pc_rdidx = NULL;
	    ld->lb->lg = 0;

	    if (ld->st->size) {
		ld->open_data_us = GetTime_us64();
		if (!ld->dev_open || !ld->dev_open(wd)) {
		    if (ld->skip & xcySkip_errOpenRD) {
			ld->srt = srt_skip;
			return;
		    }
		    ld->msgOk_skip = xcySkip_errOpenRD;
		    ld->msg_nr = xpyErr_Alg | xpyErr_Msg | xpyErr_OpenRD;
		    ld->srt |= srt_msg_wait;
		    return;
		} 
		ld->srt = srt_move;

	    } else {
		ld->srt = srt_ok;
		if (wd->mode & cpwdMode_cut) ld->srt = srt_del_wait;
	    }
	    ld->init_send = True;
	    ld->send_Mt(obj, thdMSG_XCopy_RunStat, MsgSt_init);

	} else {
	    bool dir = False;
	    if (chk_link_skip(ld->st, ld->skip)) {
//		ld->data_ges_lg += ld->st->LinkSize;
	    } else if (ld->st->st & st_Dir) {
//		ld->data_ges_lg += ld->st->size;
		dir = True;
	    }
	    ld->srt = srt_ok;
	    if (wd->mode & cpwdMode_cut) {
		if (!dir) ld->srt = srt_del_wait;
	    }
	    ld->init_send = True;
	    ld->send_Mt(obj, thdMSG_XCopy_RunStat, MsgSt_init);
	}
	return;
}

void step_Del_RD(TXCopy_wd *wd) {
  TXCopy_ld *src = wd->xcy_src;
  TNetErrRet ErrRet;

	if (!src) return;
	if (!chk_stat_Ok_RD(src, 0)) return;

	net_err_init(&ErrRet);
	if (!(src->st->st & st_Link) && (src->st->st & st_Dir)) {
	    if (!(src->skip & xcySkip_Dir)) {
		if (!src->rmdir_gp) goto err_sys_skip;
		if (!src->rmdir_gp(&src->gp, &ErrRet)) goto err_sys_skip;
	    }
	} else {
	    if (!src->unlink_gp) goto err_sys_skip;
	    if (!src->unlink_gp(&src->gp, &ErrRet)) goto err_sys_skip;
	}
	if (!(wd->mode & cpwdMode_cut)) {
	    wd->xcs.FileCount++;
	    if (src->st->st & st_Link) {
		wd->xcs.GesDataSize += src->st->LinkSize;
	    } else {
		wd->xcs.GesDataSize += src->st->size;
	    }
	}

	src->srt = srt_ok;
	return;

err_sys_skip:
	if (ErrRet.sys_errno) {
	    errno = ErrRet.sys_errno;
	    ppcchange(&src->msg_s, SysError_pc());
	} else if (ErrRet.err) {
	    ppcchange(&src->msg_s, ErrRet.err);
	} else {
	    ppcchange(&src->msg_s, "can't del");
	}
	net_err_clr(&ErrRet);
	src->msgOk_skip = xcySkip_errDelFile;
	src->msg_nr = xpyErr_Alg | xpyErr_DelFile | xpyErr_Msg;
	src->srt |= srt_msg_wait;
	if (src->skip & src->msgOk_skip) src->srt = srt_skip;
}

/* __________________________ msg _______________________________________ */
int handle_Mt_RD(TXCopy_wd *wd, int MsgSt) {
//  TXCopy_ld *ld = wd->xcy_src;

	switch (MsgSt & thdMSG_Mask) {

	case MsgSt_init:
	case MsgSt_FileStatGet:
// printf("%s(%d): %7d RD MsgSt_init stat=0x%03x\n", FL, GetCurTimeDiv_us(wd->us), wd->stat);
	    if ((wd->stat & xcywdstat_wr_run) &&
		!wd->xcy_dest->init(wd, MsgSt)) {
 printf("%s(%d): %7d RD MsgSt_init ERROR\n", FL, GetCurTimeDiv_us(wd->us));
		wd->stat |= xcywdstat_wr_err;
		if (wd->xcy_src && wd->xcy_src->close) wd->xcy_src->close(wd, MsgSt_err);
	    }
	    break;

	case MsgSt_close:
	    wd->stat &= ~xcywdstat_rd_run;
	    break;

	case MsgSt_err:
// printf("%s(%d): %7d RD MsgSt_err ERROR\n", FL, GetCurTimeDiv_us(wd->us));
	    wd->stat |= xcywdstat_rd_err;
	case MsgSt_skip:
	    wd->stat &= ~xcywdstat_rd_run;
	    if (wd->stat & xcywdstat_wr_run) {
		if (wd->xcy_dest && wd->xcy_dest->close) wd->xcy_dest->close(wd, MsgSt);
	    }
	    break;

	default:
 printf("%s(%d): %7d RD Msg ??? msg->tag=%d\n", FL, GetCurTimeDiv_us(wd->us), MsgSt);
	}
	return 0;
}
