/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"


static const char ctab[32] = "abcdefghijklmnopqrstuvwxyz234569";

static const byte entab[256] = {
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0x1a,0x1b, 0x1c,0x1d,0x1e,0xff, 0xff,0x1f,0xff,0xff, 0xff,0xff,0xff,0xff,

  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0x00,0x01,0x02, 0x03,0x04,0x05,0x06, 0x07,0x08,0x09,0x0a, 0x0b,0x0c,0x0d,0x0e,
  0x0f,0x10,0x11,0x12, 0x13,0x14,0x15,0x16, 0x17,0x18,0x19,0xff, 0xff,0xff,0xff,0xff,

  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,

  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff,
  0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff
};

typedef struct {
  char *pw;
  char *crypt;
} TCryptBuf;

#define MAX_CryptBuf 16
static TCryptBuf CryptBuf[MAX_CryptBuf];
static int CryptIdx;

static void AddToBuf(const char *pw16, const char *crypt) {
	if (CryptIdx >= MAX_CryptBuf) CryptIdx = 0;
	ppcchange(&CryptBuf[CryptIdx].pw, pw16);
	ppcchange(&CryptBuf[CryptIdx].crypt, crypt);
	CryptIdx++;
}


char *CryptMini16c_dec(char *pw16) {
  int k, i, y;
  dword t, *pdw;
  char *pc, *pcd;
  char *res;
  TBitsIO BitsIO;
  char buf[20];

	if (!pw16 || !*pw16) return NULL;
	if (pclen(pw16) > 16) return NULL;

	for (i=0;i<MAX_CryptBuf;i++) {
	    char *pwt = CryptBuf[i].pw;
	    if (pwt && pccmp(pwt, pw16)) return pcdup(CryptBuf[i].crypt);
	}

	t = GetTime_us();
	pdw = (dword*)buf;
	for (i=5;i;i--) {*pdw = t; pdw++;}

	pc = pw16;
	pcd = buf;
	for (i=4;i;i--) {
	    pcd++;
	    for (y=4;y;y--) {
		*pcd = *pc;
		if (!*pc) goto cpende;
		pcd++; pc++;
	    }
	}
cpende:
	buf[15] = 0;	//crypt method

	pc = buf;
	k = *pc;
	pc++;
	for (i=19;i;i--) {
	    k = *pc ^ k;
	    k = i ^ k;
	    *pc = k;
	    pc++;
	} 

	res = xmalloc(33);

	BitsIO.io.rd_byte = NULL;
	BitsIO.dat = (byte*)buf;
	BitsIO.buf = 0;
	BitsIO.k = 0;
//	BitsIO.io_struct = NULL;

	pc = res;
	for (i=32;i;i--) {
	    k = GetBits8(&BitsIO, 5);
	    *pc = ctab[k];
	    pc++;
	}
	*pc = 0;
	AddToBuf(pw16, res);
	return res;
}

static char *CryptMini16c_enc_NoBuf(const char *crypt) {
  int k, i, y;
  char *pc, *pcd;
  char *res;
  const byte *pb;
  TBitsIO BitsIO;
  char buf[20];

	BitsIO.io.rd_byte = NULL;
	BitsIO.dat = (byte*)buf;
	BitsIO.buf = 0;
	BitsIO.k = 0;
	pb = (const byte*)crypt;
	for (i=32;i;i--) {
	    k = entab[*pb];
	    if (k >= 32) return NULL;
	    SetBits8(&BitsIO, k, 5);
	    pb++;
	}

	pc = buf;
	k = *pc;
	pc++;
	for (i=19;i;i--) {
	    *pc = *pc ^ k ^ i;
	    k = *pc ^ k ^ i;
	    pc++;
	} 

	if (buf[15] != 0) return NULL;	//crypt method

	res = xmalloc(17);
	pc = buf;
	pcd = res;
	for (i=4;i;i--) {
	    pc++;
	    for (y=4;y;y--) {
		*pcd = *pc;
		if (!*pc) goto cpende;
		pcd++; pc++;
	    }
	}
	*pcd = 0;
cpende:
	return res;
}

char *CryptMini16c_enc(const char *crypt) {
  int i;
  char *res;
  
	if (!crypt ||!*crypt) return NULL;
	if (pclen(crypt) < 32) return NULL;

	for (i=0;i<MAX_CryptBuf;i++) {
	    char *pwt = CryptBuf[i].crypt;
	    if (pwt && pccmp(pwt, crypt)) return pcdup(CryptBuf[i].pw);
	}
	res = CryptMini16c_enc_NoBuf(crypt);
	AddToBuf(res, crypt);
	return res;
}

void CryptMini16c_enc_reg(const char *crypt) {
	char *pw = CryptMini16c_enc(crypt);
	xfree(pw);
}
