/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "sysbase.h"
#include "sysbaseio.h"

char *TEnvPath_GetExec(TEnvPath *self, const char *exec) {
  char *fn, **ppath;
  TPathInfo pathi;
#ifdef _OS_WIN32
  bool hasExt;
#endif
	if (!self || !exec) return NULL;

#ifdef _OS_WIN32
	hasExt = GetFileExt_pc(exec) != NULL;
#endif

	if (self->ppath && get_PathInfo(exec, &pathi) && (pathi.fu & pathi_rel)) {
	    ppath = self->ppath;
	    while (*ppath) {
		fn = path_add(*ppath, exec);
#ifdef _OS_WIN32
		if (!hasExt) ChangeFileExt_ppc(&fn, "exe");
		slash_korr(fn, True);
#endif
// printf("%s(%d): %s\n", FL, fn);
		if (file_chk_exist(fn, findREG)) return fn;
		xfree(fn);
		ppath++;
	    }
	    return NULL;
	}
	if (!file_chk_exist(exec, findREG)) return NULL;
	return pcdup(exec);
}

/*_____________________________ Destroy _________________________________*/
static bool TEnvPath__Destroy(TEnvPath *self) {
  char **ppath = self->ppath;
	while (*ppath) {
	    xfree(*ppath);
	    ppath++;
	}
	xfree(self->ppath);
	return False;
}

/*_____________________________ Create __________________________________*/
TEnvPath *TEnvPath_Create(const char *path) {
  TEnvPath *self;
  char *env, *pc, **ppath;
  const char *indata;

	self = ttc_create(sizeof(TEnvPath), wtyp_TEnvPath);
	self->destroy = (TDESTROY)TEnvPath__Destroy;
	pc = getenv("PATH");

	env = pcadd_n(3, path, " ", pc);
	if (env && *env) {
	    pc = env;
	    while (*pc) {
#ifdef _OS_WIN32
		if (*pc == ';') *pc = ' ';
#else
		if (*pc == ':') *pc = ' ';
#endif
		pc++;
	    }
	    indata = env;
	    while (indata) {
		indata = GetToken(indata, NULL);
		self->count++;
	    }
	    self->ppath = xmalloc(sizeof(void*) * (self->count + 1));
	    ppath = self->ppath;
	    indata = env;
	    while (indata) {
		indata = GetToken(indata, ppath);
		ppath++;
	    }
	    *ppath = NULL;

	} else {
	    self->ppath = xcalloc(sizeof(void*));
	}
	xfree(env);
	return self;
}
