/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

void TpcIni_ChangeGruppeDat(TpcIni *ti, int lg, char *pc, TGruppeDat *Gpd) {

	ti->lastGpD = Gpd;
	if (Gpd->org) {
	    xfree(Gpd->org);
	    Gpd->dat = NULL;
	    Gpd->name = NULL;
	    Gpd->lgname = 0;
	}
	Gpd->org = pcndup(pc, lg);
	pc = Gpd->org;
	if (lg <= 0) return;

	while (lg && (*pc == ' ')) { lg--; pc++; }
	if (*pc == ';') return;
	Gpd->name = pc;
	while (lg && (*pc != '=')) { lg--; pc++; }
	if (*pc != '=') return;
	Gpd->lgname = pc - Gpd->name;
	if (!Gpd->lgname || !lg) return;
	pc++;
	lg--;
	while (lg && (*pc == ' ')) { lg--; pc++; }
	Gpd->dat = pc;
}

void TpcIni_AddGruppeDat(TpcIni *ti, int lg, char *pc) {
  TGruppeDat *Gpd;
	Gpd = xcalloc(sizeof(TGruppeDat));
	*ti->npGruD = Gpd;
	ti->npGruD = &Gpd->next;
	TpcIni_ChangeGruppeDat(ti, lg, pc, Gpd);
}

void TpcIni_AddGruppe(TpcIni *ti, int lg, const char *pc) {
  TGruppe *Gp;
	Gp = xcalloc(sizeof(TGruppe));
	ti->lastGp = Gp;
	Gp->name = pcndup(pc, lg);
	Gp->lgname = lg;
	*ti->npGru = Gp;
	ti->npGru = &Gp->next;
	ti->npGruD = &Gp->Dat;
};

static void TpcIni_AddItem(TpcIni *ti, char *pc, int lg) {
  char *pce;
	if (lg && (*pc == '[')) {
	    lg--;
	    pc++;
	    pce = pc;
	    while (lg && (*pce != ']')) {
		lg--;
		pce++;
	    }
	    if (lg) TpcIni_AddGruppe(ti, pce-pc, pc);
	} else if (ti->npGruD) {
	    TpcIni_AddGruppeDat(ti, lg, pc);
	}
}

static bool ti_cmp(TGruppe *Gp, const char *pc, int lg, bool cCase) {
	if (!Gp) return True;
	if (!Gp->name || (Gp->lgname != lg)) return False;
	if (cCase) return (pcncasecmpASCII(Gp->name, pc, lg) == 0);
	return pcncmp(Gp->name, pc, lg);
}

TGruppe *TpcIni_FindGruppe(TpcIni *ti, const char *Gruppe) {
  int lg;
  TGruppe *Gp;

	ti->lastGp = NULL;
	lg = pclen(Gruppe);
	if (!lg) return NULL;
	Gp = ti->Gruppe0;

	while (!ti_cmp(Gp, Gruppe, lg, ti->cCase)) Gp = Gp->next;

	ti->lastGp = Gp;
	return Gp;
}

TGruppeDat *TpcIni_FindGruppeDat(TpcIni *ti, const char *Gruppe, const char *Funkt) {
  int lg;
  TGruppe *Gp;
  TGruppeDat *Gpd;

	ti->lastGpD = NULL;
	lg = pclen(Funkt);
	if (!lg) return NULL;
	Gp = TpcIni_FindGruppe(ti, Gruppe);
	if (!Gp) return NULL;
	Gpd = Gp->Dat;
	if (!Gpd) return NULL;

	while (!ti_cmp((TGruppe*)Gpd, Funkt, lg, ti->cCase)) Gpd = Gpd->next;

	ti->lastGpD = Gpd;
	return Gpd;
}

const char *TpcIni_RD_pc(TpcIni *ti, const char *Gruppe, const char *Funkt, const char *Default) {
  TGruppeDat *Gpd;
	Gpd = TpcIni_FindGruppeDat(ti, Gruppe, Funkt);
	if (Gpd) return Gpd->dat;
	return Default;
}

const char *TpcIni_RD(TpcIni *ti, const char *Gruppe, const char *Funkt) {
  TGruppeDat *Gpd;
	Gpd = TpcIni_FindGruppeDat(ti, Gruppe, Funkt);
	if (Gpd && Gpd->dat && *Gpd->dat) return Gpd->dat;
	return NULL;
}

#define tiMaxLineLg	16384
void TpcIni_load(TpcIni *self, FILE *fd) {
  int lg;
  byte *tline;
  byte *pc;
  byte line[tiMaxLineLg];

	if (!fd) return;
	self->Exist = true;
	tline = (byte*)fgets((char*)line, tiMaxLineLg, fd);
	if (tline && (tline[0] == 0xEF)&&(tline[1] == 0xBB)&&(tline[2] == 0xBF)) {
	    tline += 3;
	    self->BOM_UTF8 = true;
	}
	goto line0;
loop:
	tline = (byte*)fgets((char*)line, tiMaxLineLg, fd);
line0:
	if (tline) {
	    pc = tline;
	    while ((*pc > 0x1f) || (*pc == 0x09)) pc++;
	    lg = pc - tline;
	    TpcIni_AddItem(self, (char*)tline, lg);
	    goto loop;
	}
}

/*____________________________ Destroy __________________________________*/
static bool TpcIni__Destroy(TpcIni *ti) {
  TGruppe *Gp, *xGp;
  TGruppeDat *Gpd, *xGpd;

	Gp = ti->Gruppe0;
	while (Gp) {
	    Gpd = Gp->Dat;
	    while (Gpd) {
		if (Gpd->org) xfree(Gpd->org);
		xGpd = Gpd;
		Gpd = Gpd->next;
		xfree(xGpd);
	    }
	    if (Gp->name) xfree(Gp->name);
	    xGp = Gp;
	    Gp = Gp->next;
	    xfree(xGp);
	}
	xfree(ti->FileName);
	return False;
}

/*____________________________ Create ___________________________________*/
TpcIni *TpcIni_Create_struct() {
  TpcIni *self;
	self = ttc_create(sizeof(TpcIni), wtyp_TpcIni);
	self->destroy = (TDESTROY)TpcIni__Destroy;
	self->cCase = True;
	self->npGru = &self->Gruppe0;
	self->npGruD = NULL;
	return self;
}

TpcIni *TpcIni_Create(const char *FileName) {
  const char *fn;
  FILE *F;
  TpcIni *self;

	self = TpcIni_Create_struct();
	fn = path_korr_file(FileName);
	if (!fn || !*fn) fn = "/tmp/ALD_pcIni_No_FileName.ini";
	self->FileName = pcdup(fn);

	F = file_stream_open(self->FileName, "r");
	if (F) {
	    TpcIni_load(self, F);
	    fclose(F);
	}
	return self;
}
