/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

void TpcIni_push_stack(TpcIni *self, const char *Gruppe, const char *Funkt, 
			const char *Data, int Max, int Mode) {
  char *line, *pc0, *fnNoPasswd, **itm, *fnt, *pw;
  const char *pc, *s;
  int idx, dupi, i;
  bool isDup;

	if (!self || !Gruppe || !Funkt || !Data || (Max <= 0) || 
	    (pclen(Data) > 4000)) return;

	pc = Data;
	while (*pc) {
	    if ((*pc == 0x0a) || (*pc == 0x0d)) return;
	    pc++;
	}

	fnNoPasswd = NULL;
	if (Mode & pcIni_PushPath) {
	    fnNoPasswd = get_DisplayFileStr(Data, NULL);
	}

	line = xmalloc(pclen(Funkt) + 16);
	itm = xcalloc(sizeof(char*) * Max);
	pc0 = line;

	ppcmove(&pc0, Funkt);
	dupi = -1;
	idx = 1;

	for (i=0; i<Max; i++) {
	    inttopc(pc0, i);
	    s = TpcIni_RD(self, Gruppe, line);
	    if (!s) break;

	    if (Mode & pcIni_PushPath) {
		TglobPath *gp = pctogp(s);
		CryptMini16c_enc_reg(gp->passwd);
		fnt = get_DisplayFileStr(NULL, gp);
		gpFree(gp);
		isDup = pccmp(fnNoPasswd, fnt);
		xfree(fnt);
	    } else {
		isDup = pccmp(s, Data);
	    }

	    if (isDup) {
		dupi = i;
	    } else {
		itm[idx] = pcdup(s);
		idx++;
		if (idx >= Max) break;
	    }
	}

	if (dupi) {
	    if (Mode & pcIni_PushPath) {
		TglobPath *gp = pctogp(Data);
		pw = CryptMini16c_dec(gp->passwd);
		if (pw) {
		    xfree(gp->passwd);
		    gp->passwd = pw;
		}
		itm[0] = gptopc(gp, gptfu_PasswdTxt|gptfu_noRootArt);
		gpFree(gp);
	    } else {
		itm[0] = pcdup(Data);
	    }

	    for (i=0; i<Max; i++) {
		if (!itm[i]) break;
		inttopc(pc0, i);
		TpcIni_WR(self, Gruppe, line, itm[i]);
	    }
	    TpcIni_Save(self);
	}

	for (i=0; i<Max; i++) xfree(itm[i]);

	xfree(itm);
	xfree(line);
	xfree(fnNoPasswd);
}

void TpcIni_get_stack(TpcIni *self, const char *Gruppe, const char *Funkt, 
			TpcIniOnGetData Callback, void *owner, int Max, int Mode) {
  int i;
  char *line, *pc0, *fn;
  const char *s;

	if (!self || !Gruppe || !Funkt || !Callback || (Max <= 0)) return;
	line = xmalloc(pclen(Funkt) + 16);
	pc0 = line;

	ppcmove(&pc0, Funkt);
	for (i=0; i<Max; i++) {
	    inttopc(pc0, i);
	    s = TpcIni_RD(self, Gruppe, line);
	    if (!s) break;

	    if (Mode & pcIni_PushPath) {
		TglobPath *gp = pctogp(s);
		gp_CryptMini16c_enc(gp);
		fn = gptopc(gp, gptfu_PasswdTxt|gptfu_noRootArt);
		Callback(self, fn, owner);
		xfree(fn);
		gpFree(gp);
	    } else {
		Callback(self, s, owner);
	    }
	}
	xfree(line);
}
