
#include "sysbase.h"
#include "lzw.h"

/*_____________________________ read ____________________________________*/
static void AddCodeIdx(TLzw *self, int Bits) {
	Bits++;
	self->CodeBits = Bits;
	self->CodeIdxMask = 0x0000ffff >> (16 - Bits);
}

static void ClrCodeIdx(TLzw *self) {
  int i;
	self->CodeIdx = self->ClearCode;
	self->LastCode = NO_CODE;
	AddCodeIdx(self, self->Bits);
	for (i=0;i<LZ_MAX_CODE;i++) self->Prefix[i] = NO_CODE;
}

static dword read_LZW_dword(TLzw *self) {
	if (!self->BufCount) {
	    self->Lzw_rd(self, &self->BufCount, 1);
	    self->ZDataLg += self->BufCount + 1;
	    if (!self->BufCount) {
		self->Error = LZW_ERR_EOF;
		return 0;
	    }
	    self->Lzw_rd(self, self->buf, self->BufCount);
	    self->BufIdx = 0;
	} else {
	    self->BufIdx++;
	}
	self->BufCount--;
	return self->buf[self->BufIdx];
}


static dword GetPrefixCode(TLzw *self, int code) {
  int i = LZ_MAX_CODE;

	while ((code > self->ClearCode) && i) {
	    if (code > LZ_MAX_CODE) return NO_CODE;
	    code = self->Prefix[code];
	    i--;
	}
	return code;
}

bool TLzw_Decompress(TLzw *self, byte *buf, int lg) {
  int bufIdx, code, j, tmp;

	if (self->Error) return False;
	bufIdx = 0;
	while (self->StackIdx && (bufIdx < lg)) {
	    self->StackIdx--;
	    buf[bufIdx++] = self->Stack[self->StackIdx];
	}

	while (bufIdx < lg) {
	    code = GetBits8(&self->BitsIO, self->CodeBits);
	    if (self->Error) return False;

	    if (self->CodeIdx < LZ_MAX_CODE) {
		self->CodeIdx++;
		if (((self->CodeIdx & self->CodeIdxMask) == self->CodeIdxMask) &&
		    (self->CodeBits < LZ_BITS)) {

		    AddCodeIdx(self, self->CodeBits);
// printf("%s(%d): ADD --> %4d %4d %d\n", FL, bufIdx, self->MaxCode, self->CodeBits);
		}
	    }

	    if (code == self->ClearCode) {
// printf("%s(%d): Clear %4d %d\n", FL, bufIdx, code);
		ClrCodeIdx(self);

	    } else if (code == self->EOFCode) {
 printf("%s(%d): EOF   count=%d lg=%d %d\n", FL, bufIdx, lg, self->BitsIO.k);
		self->lzwEOF = True;
		goto endOk;

	    } else if (code < self->ClearCode) {
		buf[bufIdx++] = code;
		goto SetPrefix;

	    } else {
		int PrefixCode = code;
		if (self->Prefix[code] == NO_CODE) {
		    PrefixCode = self->LastCode;
		    if (code == self->CodeIdx) {
			tmp = GetPrefixCode(self, PrefixCode);
			self->Suffix[self->CodeIdx] = tmp;
			self->Stack[self->StackIdx] = tmp;
			self->StackIdx++;
		    } else {
// printf("%s(%d): ERROR %4d %d %d\n", FL, bufIdx, code, self->CodeIdx);
			self->Error = LZW_ERR_DEFECT;
			return False;
		    }
		}

		j = 0;
		while ((j <= LZ_MAX_CODE) &&
		       (PrefixCode <= LZ_MAX_CODE) &&
		       (PrefixCode > self->ClearCode)) {
		    self->Stack[self->StackIdx] = self->Suffix[PrefixCode];
		    self->StackIdx++;
		    PrefixCode = self->Prefix[PrefixCode];
		    j++;
		}
		if ((j >= LZ_MAX_CODE) || (PrefixCode > LZ_MAX_CODE)) {
		    self->Error = LZW_ERR_DEFECT;
		    return False;
		}
		self->Stack[self->StackIdx] = PrefixCode;
		self->StackIdx++;

		while (self->StackIdx && (bufIdx < lg)) {
		    self->StackIdx--;
		    buf[bufIdx++] = self->Stack[self->StackIdx];
		}
SetPrefix:
		if (self->LastCode != NO_CODE) {
		    self->Prefix[self->CodeIdx] = self->LastCode;
		    tmp = code;
		    if (code == self->CodeIdx) tmp = self->LastCode;
		    self->Suffix[self->CodeIdx] = GetPrefixCode(self, tmp);
		}
		self->LastCode = code;
	    }
	}
// printf("EOF   count=%d\n", xxx);
endOk:
	self->ResLg = bufIdx;
	self->GesDataLg += bufIdx;
	return True;
}

bool TLzw_Setup_rd(TLzw *self) {

	self->Error = 0;
	self->lzwEOF = False;
	self->GesDataLg = 0;

	if (!self || !self->Lzw_rd) return False;
	if (!self->Lzw_rd(self, &self->Bits, 1)) return False;  //read DataBits
// printf("%s(%d): Bits=%d\n", FL, self->Bits);
	if ((self->Bits < 1) || (self->Bits > 8)) return False;

	self->ClearCode = (1 << self->Bits);
	self->EOFCode = self->ClearCode + 1;
	self->StackIdx = 0;

	self->BufCount = 0;
	self->ZDataLg = 1;

	self->BitsIO.buf = 0;
	self->BitsIO.k = 0;
	self->BitsIO.io.rd_byte = (TFBitsIO_rd_byte)read_LZW_dword;
	self->BitsIO.io_struct = self;

	ClrCodeIdx(self);

//	for (int i=0;i<LZ_MAX_CODE;i++) self->Prefix[i] = NO_CODE;
	return True;
}

/*_____________________________ Destroy _________________________________*/
/* bool TLzw__Destroy(TLzw *self) {
	return False;
} */

/*_____________________________ Create __________________________________*/
TLzw *TLzw_Create() {
  TLzw *self;

	self = ttc_create(sizeof(TLzw), wtyp_TLzw);
//	self->destroy = (TDESTROY)TLzw__Destroy;

	return self;
}
