/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"

const char *app_name;

TAppProcAtExit *AppProcAtExit;
bool AppToAtExit;
bool AppToAtExit_Wait;

void (*FappMutex_lock)();
void (*FappMutex_unlock)();

void appMutex_lock() {
	if (FappMutex_lock) FappMutex_lock();
}

void appMutex_unlock() {
	if (FappMutex_unlock) FappMutex_unlock();
}


#ifndef _OS_WIN32

#ifdef DEBUG_STACK

#include "sysbaseio.h"
#include <unistd.h>
#include <sys/wait.h>

static int backtrace_run;

static void stack_backtrace(int app_pid) {
  int pid, status;
  size_t lg;
  FileHandle fd;
  char *pc;
  int in_fd[2];
  int out_fd[2];
  char *args[7];
  char buf[BUFSIZ];


	if (backtrace_run || pipe(in_fd) || pipe(out_fd)) return;
	backtrace_run = 1;

	args[0] = pcdup("gdb");
	args[1] = pcdup("--quiet");
	pc = buf;
	ppcmove(&pc, "--pid=");
	inttopc(pc, app_pid);
	args[2] = pcdup(buf);
	args[3] = pcdup("--eval-command=backtrace");
	args[4] = pcdup("--eval-command=detach");
	args[5] = pcdup("--eval-command=quit");
	args[6] = NULL;

	pid = fork();
	if (!pid) {
	    close(0); dup(in_fd[0]);   // stdin pipe
	    close(1); dup(out_fd[1]);  // stdout pipe
	    close(2); dup(out_fd[1]);  // stderr pipe
	    execvp (args[0], args);
	    _exit (0);

	} else if (pid == -1) return;

	waitpid (pid, &status, 0);

	pc = buf;
	ppcmove(&pc, "/tmp/BACKTRACE-");
	if (app_name) {ppcmove(&pc, app_name); ppcmove(&pc, "-");}
	inttoppc(&pc, app_pid);
	ppcmove(&pc, ".txt");

	fprintf(stderr, "%s\n", buf);

	if (file_open_wr(buf, NULL, &fd, 0)) {
	    fd_set in_fdset;
	    struct timeval tv;
loop:
	    tv.tv_sec = 0;
	    tv.tv_usec = 500;
	    FD_ZERO(&in_fdset);
	    FD_SET (out_fd[0], &in_fdset);
	    if ((select(FD_SETSIZE, &in_fdset, NULL, NULL, &tv) > 0) &&
		FD_ISSET(out_fd[0], &in_fdset) &&
		file_read(out_fd[0], NULL, buf, BUFSIZ, &lg)) {

		file_write(fd, NULL, buf, lg, NULL);
		goto loop;
	    }
	    file_close(fd, NULL);
	}

	close(in_fd[0]);
	close(in_fd[1]);
	close(out_fd[0]);
	close(out_fd[1]);

}

void __attribute__ ((noreturn)) abort_SIGSEGV(int n) {
  int app_pid = getpid();

	if (app_name) {
	    fprintf(stderr, "Segmentation fault \"%s\"\n", app_name);
	} else {
	    fprintf(stderr, "Segmentation fault \"pid=%d\"\n", app_pid);
	}
	stack_backtrace(app_pid);
	_exit(0);
}

#endif  /* DEBUG_STACK */
#endif  /* ! _OS_WIN32 */

static void msg_memory_exhausted() {
	appError(EMSG_ERR_EXIT, "memory exhausted");
}

#ifdef MEM_CHK

#ifdef USE_PTHREAD
#include <pthread.h>

static pthread_mutex_t mem_mutex;
static bool mem_mutex_ok = False;
#endif

typedef struct {
  size_t size;
  void *p;
} TMEM_mck;

typedef struct {
  int count;
  TMEM_mck arr[];
} TMEM_mci;

static size_t memsize = 0;
static int memcount = 0;
static TMEM_mci **mc = NULL;

static int get_mci_idx(void *p) {
	if (!mc) {
	    mc = malloc(sizeof(TMEM_mci*) * 128);
	    for (int i=0;i<128;i++) {
		mc[i] = calloc(sizeof(TMEM_mci) + (sizeof(TMEM_mck) * 256), 1);
		mc[i]->count = 256;
	    }
	}
#ifdef POINTER64BIT
  long = (long)p >> 3;
	return ((int)(idx + (idx >> 7) + (idx >> 14) + (idx >> 21))) & 0x7f;
#else
  long idx = (long)p >> 2;
	return ((int)(idx + (idx >> 7) + (idx >> 14))) & 0x7f;
#endif
}

static TMEM_mci *get_mci(void *p) {
	int idx = get_mci_idx(p);
	return mc[idx];
}

void MEM_CHK_ADD(void *xm, size_t size) {
  int i;
  char *pc;
#ifdef USE_PTHREAD
	if (!mem_mutex_ok) {
	    mem_mutex_ok = True;
	    pthread_mutex_init(&mem_mutex, NULL);
	}
	pthread_mutex_lock(&mem_mutex);
#endif

  TMEM_mci *mci = get_mci(xm);
	for (i=0;i<mci->count;i++) {
	    if (mci->arr[i].p == 0) {
		mci->arr[i].size = size;
		mci->arr[i].p = xm;
		memsize += size;
		memcount++;
#ifdef USE_PTHREAD
		    goto ende;
#else
		    return;
#endif
	    } else if (mci->arr[i].p == xm) {
  printf("ERROR MEM_CHK_ADD: 0x%08x\n", (int)xm);
#ifdef USE_PTHREAD
		    goto ende;
#else
		    return;
#endif
	    }
	}
	if (mci->count < 2048) {
	    size_t old_size = sizeof(TMEM_mci) + sizeof(TMEM_mck)*mci->count;
	    mci->count += 256;
	    size_t new_size = sizeof(TMEM_mci) + sizeof(TMEM_mck)*mci->count;
	    i = get_mci_idx(xm);
	    mc[i] = realloc(mc[i], new_size);
	    pc = (char*)mc[i];
	    Fill0(&pc[old_size], new_size-old_size);

	    mci = get_mci(xm);
	    for (i=0;i<mci->count;i++) {
		if (mci->arr[i].p == 0) {
		    mci->arr[i].size = size;
		    mci->arr[i].p = xm;
		    memsize += size;
		    memcount++;
#ifdef USE_PTHREAD
		    goto ende;
#else
		    return;
#endif
		} else if (mci->arr[i].p == xm) {
      printf("ERROR MEM_CHK_ADD: 0x%08x\n", (int)xm);
#ifdef USE_PTHREAD
		    goto ende;
#else
		    return;
#endif
		}
	    }
	}
 printf("MEM_CHK_ADD: Out of MEM Buf Size=%d\n", 256);
#ifdef USE_PTHREAD
ende:
	pthread_mutex_unlock(&mem_mutex);
#endif
}

bool MEM_CHK_CLR(void *xm) {
  bool result = True;
#ifdef USE_PTHREAD
	pthread_mutex_lock(&mem_mutex);
#endif
  TMEM_mci *mci = get_mci(xm);
	for (int i=0;i<mci->count;i++) {
	    if (mci->arr[i].p == xm) {
		mci->arr[i].p = NULL;
		memsize -= mci->arr[i].size;
		memcount--;
		goto ende;
	    }
	}
 printf("ERROR MEM_CHK_CLR: 0x%08x\n", (int)xm);
	result = False;
ende:
#ifdef USE_PTHREAD
	pthread_mutex_unlock(&mem_mutex);
#endif
	return result;
}

void MEM_CHK_Count(int *count, size_t *size) {
	if (count) *count = memcount;
	if (size) *size = memsize;
}

int LastMEM_CHK_Count;
bool MEM_AZ_Disable = False;

void MEM_CHK_AZ_pc(const char *pc) {
	if (!MEM_AZ_Disable) {
  int lMCt;
  dword lMsize;
	    MEM_CHK_Count(&lMCt, &lMsize);
	    if (LastMEM_CHK_Count != lMCt) {
		    LastMEM_CHK_Count = lMCt;
    printf("Debug %s: MEM_CHK_Count=%3d  size=%u\n", pc, lMCt, lMsize);
	    }
	}
}

void MEM_CHK_AZ() {
	MEM_CHK_AZ_pc(app_name);
}
#endif /* MEM_CHK */

void *xmalloc(size_t size) {
	void *result = malloc(size);
#ifdef MEM_CHK
 MEM_CHK_ADD(result, size);
#endif
	if (!result) msg_memory_exhausted();
	return result;
}

#ifdef MEM_CHK
void *zmalloc(size_t size) {
	void *result = malloc(size);
	if (result) MEM_CHK_ADD(result, size);
	return result;
}
#endif

void *xcalloc(size_t size) {
	void *result = calloc(size, 1);
#ifdef MEM_CHK
 MEM_CHK_ADD(result, size);
#endif
	if (!result) msg_memory_exhausted();
	return result;
}

void *xrealloc(void *ptr, size_t size) {
  void *result;
	if (!ptr) return xmalloc(size);
#ifdef MEM_CHK
 MEM_CHK_CLR(ptr);
#endif
	result = realloc(ptr, size);
#ifdef MEM_CHK
 MEM_CHK_ADD(result, size);
#endif
	if (!result) msg_memory_exhausted();
	return result;
}

void *xrecalloc(void *ptr, size_t size, size_t old_size) {
  void *result;
  char *pc;

	if (!ptr) return xcalloc(size);
#ifdef MEM_CHK
 MEM_CHK_CLR(ptr);
#endif
	result = realloc(ptr, size);
#ifdef MEM_CHK
 MEM_CHK_ADD(result, size);
#endif
	if (!result) msg_memory_exhausted();
	if (size > old_size) {
	    pc = (char*)result;
	    Fill0(&pc[old_size], size-old_size);
	}
	return result;
}

void xfree(void *ptr) {
	if (ptr) {
#ifdef MEM_CHK
	    if (!MEM_CHK_CLR(ptr)) return;
#endif
	    free(ptr);
	}
}

void pp_free(void **ptr) {
  void *p;
	if (ptr) {
	    p = *ptr;
	    *ptr = NULL;
	    xfree(p);
	}
}

