/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

//*********************** TLines ************************
void TLines_DoCount(TLines *tls) {
	if (!tls->EvStop && tls->OnCount) tls->OnCount(tls);
}

void TLines_DoLineChange(TLines *tls, int idx) {
	if (!tls->EvStop && tls->OnLineChange) {
	    if (idx >= tls->Count) idx = tls->Count - 1;
	    if (idx < 0) idx = 0;
	    tls->OnLineChange(tls, idx);
	    tls->Changed = True;
	}
}

void TLines_ResLinec(TLines *tls, TColS * pcs, int idx) {
	if (pcs->c) {
	    if (tls->OnLineDelc) tls->OnLineDelc(tls, pcs, idx);
	    ppfree(&pcs->c);
	}
}

static void TLines_ResLine(TLines *tls, int idx) {
  TColS *pcs = tls->lines->cs[idx];
	if (!pcs) return;
	if (tls->OnLineDel) tls->OnLineDel(tls, pcs, idx);
	TLines_ResLinec(tls, pcs, idx);
	xfree(pcs->s);
	xfree(pcs);
	tls->lines->cs[idx] = NULL;
}

void TLines_DelN(TLines *tls, int idx, int n) {
  int i;
  TColS **acs, **ecs;

	if ((idx >= tls->Count) || (tls->Count == 0) || (n < 1)) return;
	if ((idx+n) >= tls->Count) n = tls->Count - idx;
	for (i=idx; i<(idx+n); i++) TLines_ResLine(tls, i);
	tls->Count -= n;
	if (tls->Count <= 0) {
	    TLines_Clear(tls);
	    return;
	}
	acs = &tls->lines->cs[idx];
	ecs = &tls->lines->cs[idx+n];
	for (i = idx; i < tls->Count; i++) {
	    *acs = *ecs;
	    *ecs = NULL;
	    acs++; ecs++;
	}
	tls->Changed = True;
	TLines_DoCount(tls);
}

void TLines_Del(TLines *tls, int idx) {
	TLines_DelN(tls, idx, 1);
}

void TLines_Clear(TLines *self) {
  int i;
	if (!self) return;
	if (self->lines) {
	    for (i=0; i<self->Count; i++) TLines_ResLine(self, i);
	    ppfree(&self->lines);
	}
	self->Count = 0;
	if (self->OnCount) self->OnCount(self);
//	TLines_DoCount(tls);
}

void TLines_Clear_c(TLines *tls) {
  int i;
  TColS *pcs;

	if (!tls || !tls->lines) return;
	for (i=0;i<tls->Count;i++) {
	    pcs = tls->lines->cs[i];
	    if (pcs) TLines_ResLinec(tls, pcs, i);
	}
}

static void *TLines_Get_csMem(TLines *tls) {
  int lg = sizeof(TColS);
	if (tls->csType == csType_TFileS) lg = sizeof(TFileS);
	return xcalloc(lg);
}

void *TLines_Add_(TLines *self) {
  TColS *pcs;
	if (!self->lines) set_ptr_arr((void*)&self->lines, 255);
	if (self->Count >= self->lines->items) set_ptr_arr((void*)&self->lines, self->lines->items + 256);
	pcs = TLines_Get_csMem(self);
	self->lines->cs[self->Count] = pcs;
	self->Count++;
	return pcs;
}

TColS *TLines_nAdd(TLines *self, const char *s, int lg) {
  TColS *pcs = TLines_Add_(self);
	pcs->lg = lg;
	ppcnchange(&pcs->s, s, lg);
	TLines_DoLineChange(self, self->Count-1);
	TLines_DoCount(self);
	return pcs;
}

TColS *TLines_Add(TLines *self, const char *s) {
	return TLines_nAdd(self, s, pclen(s));
}

TColS *TLines_Add_lgs(TLines *self, Tlgs *lgs, int eol_fu) {
  TColS *cs = TLines_nAdd(self, lgs->s, lgs->lg);

	cs->fu = eol_fu & fio_EOL_Mask;
	if (self->DefaultEOL == -1) {
	    self->DefaultEOL = eol_fu & fio_EOL_Mask;
	} else if (self->DefaultEOL != (eol_fu & fio_EOL_Mask)) {
	    if (!(eol_fu & fio_EOF)) self->DefaultEOL = fio_EOL_mix;
	}
	return cs;
}

static TColS *TLines_Ins_(TLines *self, int idx) {
  int i;
  TColS *pcs;
	if (idx >= self->Count) return TLines_Add_(self);
	if (self->Count >= self->lines->items) set_ptr_arr((void*)&self->lines, self->lines->items + 256);
	for (i=self->Count; i>idx; i--) self->lines->cs[i] = self->lines->cs[i-1];
	pcs = TLines_Get_csMem(self);
	self->lines->cs[idx] = pcs;
	self->Count++;
	return pcs;
}

TColS *TLines_nIns(TLines *self, int idx, const char *s, int lg) {
  TColS *pcs = TLines_Ins_(self, idx);
	pcs->lg = lg;
	ppcnchange(&pcs->s, s, lg);
	TLines_DoLineChange(self, idx);
	TLines_DoCount(self);
	return pcs;
}

TColS *TLines_Ins(TLines *self, int idx, const char *s) {
	return TLines_nIns(self, idx, s, pclen(s));
}

const char *TLines_Get_pc(TLines *self, int idx) {
	if (!self || (idx >= self->Count) || (idx < 0)) return NULL;
	return self->lines->cs[idx]->s;
}

TColS *TLines_Get_pcs(TLines *self, int idx) {
	if (!self || (idx >= self->Count) || (idx < 0)) return NULL;
	return self->lines->cs[idx];
}

int REGP1 TLines_GetLineCount(TLines *self) {
	if (!self) return 0;
	return self->Count;
}

bool TLines__Destroy(TLines *self) {
	self->OnCount = NULL;
	TLines_Clear(self);
	xfree(self->FileName);
	xfree(self->st);
	return False;
}

TLines *TLines_Create() {
  TLines *self = ttc_create(sizeof(TLines), wtyp_TLines);

	self->destroy = (TDESTROY)TLines__Destroy;
#ifdef _OS_WIN32
	self->DefaultEOL = fio_EOLwindos;
#else
	self->DefaultEOL = fio_EOLunix;
#endif
	return self;
}
