/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include "zsrc.h"
#include "zlib.h"

int ZsrcToPath(const ThZsrc *zsrc, const char *path, const char *name, int fu) {
  int result;
  char *fn;
  FILE *f;
  byte *outBuf;
  xchar xc;
  z_stream s;

	if (!zsrc) return ZsrcError;
	if (path && !TestIsPath(path) && !file_mkdir_parents(path, NULL, 0)) return ZsrcError;
	if (!name) name = zsrc->name;
	fn = path_add(path, name);
	result = ZsrcOk;
	if ((fu & ZsrcFu_OnlyNotExists) && TestIsFile(fn)) result = ZsrcExists;
	if (result == ZsrcOk) {
	    f = file_stream_open(fn, "wb");
	    if (f) {

		if (zsrc->compress == zsrc_no_comp) {
		    fwrite(zsrc->data, 1, zsrc->org_size, f);

		} else {
		    Fill0(&s, sizeof(s));

		    outBuf = xcalloc(zsrc->org_size + 256);
		    s.avail_in = zsrc->data_size;
		    s.avail_out = zsrc->org_size+256;
#if 1
		    xc.us_const = zsrc->data;
		    s.next_in = xc.us;
#else
		    s.next_in = (byte*)zsrc->data;
#endif
		    s.next_out = outBuf;

		    if (inflateInit2(&s, -15) != Z_OK) {
			appError(EMSG_ERR, " zlib_inflateInit");
			result = ZsrcError;
		    }

		    int m = inflate(&s, Z_FINISH);
		    if ((result == ZsrcOk) && (s.total_out != zsrc->org_size)) {
			appError(EMSG_ERR, " zlib_inflate %d", m);
			result = ZsrcError;
		    }

		    if (result == ZsrcOk) {
			fwrite(outBuf, 1, zsrc->org_size, f);
		    }

		    inflateEnd(&s);
		    xfree(outBuf);
		}

		fclose(f);
	    }
	}
	xfree(fn);

	return result;
}
