/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "cicons.h"

//*********************** TScrollBar ************************
static void TScrollBar_scPosSet(TScrollBar *sb);

static void sb_GetClientRect(TScrollBar *self, TWRect *rc) {
	TWins_GetClientRect((TWins*)self, rc);
	if (!self->Hsc) {
	    SWAP(rc->left, rc->top);
	    SWAP(rc->width, rc->height);
	}
}

static void sb_BorderPaint(TScrollBar *self, int fu, int x, int y, int w, int h) {
	if (!self->Hsc) {
	    SWAP_INT(x, y);
	    SWAP_INT(w, h);
	}
	StdBorderPaint(self->WinID, fu,  x, y, w, h);
}

static void sb_Fill_Rect(TScrollBar *self, GC gc, int x, int y, int w, int h) {
	if (w <= 0) return;
	if (!self->Hsc) {
	    SWAP_INT(x, y);
	    SWAP_INT(w, h);
	}
	Fill_Rect(self->WinID, gc, x, y, w, h);
}

static bool TScrollBar_AdjustSize(TScrollBar *self) {
  int l;
  TWRect rc;
  bool result;

	result = TWins_AdjustSize((TWins*)self); //+Chk Align
	sb_GetClientRect(self, &rc);
	l = 15;
	if (rc.width < 48) {
	    l = (rc.width-3) / 3;
	    if (l < 7) l = 7;
	}
	if (result || (self->bulg != l)) {
	    self->bulg = l;
	    TScrollBar_scPosSet(self);
	}
	return result;
}

static void TScrollBar_ButtonPaint(TScrollBar *self, bool pos, bool down) {
  TWRect rc;
  int ix, iy, fu;
  TPic *pic;

	if (self->Hsc) {
	    if (pos) {
		pic = get_pic9_hsb_plus(False);
	    } else {
		pic = get_pic9_hsb_minus(False);
	    }
	} else {
	    if (pos) {
		pic = get_pic9_vsb_plus(False);
	    } else {
		pic = get_pic9_vsb_minus(False);
	    }
	}

	sb_GetClientRect(self, &rc);
	if (pos) rc.left += rc.width - self->bulg;

	iy = rc.top + ((rc.height-9) >> 1);
	ix = rc.left + ((self->bulg-9) >> 1);

	if (!self->Hsc) SWAP_INT(ix, iy);
	DrawPic(self->WinID, pic, ix, iy);
//	DrawPic(self->WinID, pic, self->Hsc ? ix : iy, self->Hsc ? iy : ix);
	fu = mBorder | mBorderW2 | mBorderBu;
	if (down) fu = mBorder | mBorderW2 | mBorderBu | mBorderL;
	sb_BorderPaint(self, fu, rc.left, rc.top, self->bulg, rc.height);
}

static void TScrollBar_DoPaint(TScrollBar *self) {
  TWRect rc;
  int l1;
  GC gc;

	TScrollBar_ButtonPaint(self, False, (self->MouseDownSeg == 1));

	sb_GetClientRect(self, &rc);
	if (self->scPaint) {
	    l1 = self->scPos - self->bulg - rc.left;
	    sb_Fill_Rect(self, appXDat.SB_GC, self->bulg + rc.left, rc.top, l1, rc.height);
	    l1 = rc.left + rc.width - self->scPos - self->scLg - self->bulg;
	    sb_Fill_Rect(self, appXDat.SB_GC, self->scPos+self->scLg, rc.top, l1, rc.height);
	    gc = GetPaintGC(self->Color);
	    sb_Fill_Rect(self, gc, self->scPos, rc.top, self->scLg, rc.height);
	    sb_BorderPaint(self, mBorder | mBorderW2 | mBorderBu, self->scPos, rc.top, self->scLg, rc.height);
	} else {
	    rc.left += self->bulg;
	    rc.width -= self->bulg + self->bulg;
	    sb_Fill_Rect(self, appXDat.SB_GC, rc.left, rc.top, rc.width, rc.height);
	}

	TScrollBar_ButtonPaint(self, True, (self->MouseDownSeg == 5));
}


static void SetCount(TScrollBar *sb, int Count) {
	if (Count > (sb->MaxCount - sb->ButtonCountSize)) Count = sb->MaxCount - sb->ButtonCountSize;
	if (Count < 0) Count = 0;
	TScrollBar_DoCount(sb, Count);
}

static void TScrollBar_DoSmallChange(TScrollBar *self, int *NewCount, bool Neg) {
	if (!self->OnSmallChange || !self->OnSmallChange(self, NewCount, Neg)) {
	    int i = self->SmallChange;
	    if (Neg) i = -i;
	    *NewCount += i;
	}
}

static void TScrollBar_DoLargeChange(TScrollBar *self, int *NewCount, bool Neg) {
	if (!self->OnLargeChange || !self->OnLargeChange(self, NewCount, Neg)) {
	    int i = self->LargeChange;
	    if (Neg) i = -i;
	    *NewCount += i;
	}
}

static void TScrollBar_OnTimer(TgTimer *ti) {
  TScrollBar *sb = (TScrollBar*)ti->obj;
  int count;

	if (sb->RepTimerStartCount > 0) {
	    sb->RepTimerStartCount--;
	    return;
	}

	count = sb->Count;

	if ((sb->MouseDownSeg == 1)||(sb->MouseDownSeg == 5)) {
	    TScrollBar_DoSmallChange(sb, &count, sb->MouseDownSeg == 1);

	} else if ((sb->MouseDownSeg == 2)||(sb->MouseDownSeg == 4)) {
	    TScrollBar_DoLargeChange(sb, &count, sb->MouseDownSeg == 2);

	} else return;

/*
	switch (sb->MouseDownSeg) {
	case 1: m -= sb->SmallChange; break;
	case 2: m -= sb->LargeChange; break;
	case 4: m += sb->LargeChange; break;
	case 5: m += sb->SmallChange; break;
	default: return;
	}
*/
	SetCount(sb, count);
}

static int TScrollBar_MouseSeg(TScrollBar *self, int evx) {
  TWRect rc;

	sb_GetClientRect(self, &rc);
	if (evx <= self->bulg) return 1;
	if (evx >= (rc.width - self->bulg)) return 5;
	if (!self->scPaint) return 0;
	if (evx < self->scPos) return 2;
	if (evx > (self->scPos + self->scLg)) return 4;
	return 3;
}

static void TScrollBar_DoMouseDown(TScrollBar *self, XButtonEvent *ev) {
  int count, evx;
  bool ref;

	if ((ev->button == 1) && TestWinIsEnabled((TWins*)self)) {

	    if (self->Hsc) evx = ev->x; else evx = ev->y;
	    self->MouseDownSeg = TScrollBar_MouseSeg(self, evx);
	    ref = True;
	    count = self->Count;

	    if ((self->MouseDownSeg == 1)||(self->MouseDownSeg == 5)) {
		TScrollBar_DoSmallChange(self, &count, self->MouseDownSeg == 1);

	    } else if ((self->MouseDownSeg == 2)||(self->MouseDownSeg == 4)) {
		TScrollBar_DoLargeChange(self, &count, self->MouseDownSeg == 2);

	    } else {
		if (self->MouseDownSeg == 3) {
		    self->MouseDownPos = evx;
		    self->MDscPos = self->scPos;
		}
		ref = False;
	    }

/*
	    switch (self->MouseDownSeg) {
	    case 1: count -= self->SmallChange; break;
	    case 2: count -= self->LargeChange; break;
	    case 3:
		self->MouseDownPos = evx;
		self->MDscPos = self->scPos;
		ref = False;
		break;
	    case 4: count += self->LargeChange; break;
	    case 5: count += self->SmallChange; break;
	    default: ref = False;
	    }
*/
	    if (ref) {
		if (!self->RepTimer) {
		    self->RepTimerStartCount = 9;
		    self->RepTimer = gTimerX_Add(Timer_Enabled, 33, TScrollBar_OnTimer);
		    self->RepTimer->obj = self;
		}
		SetCount(self, count);
		TWins_Refresh((TWins*)self);
	    }
	}
	TWins_DoMouseDown((TWins*)self, ev);
}

bool TScrollBar_MouseZ_Wheel(TScrollBar *self, XButtonEvent *ev) {
  int sbc, sbx;

	if ((ev->button != Button4) && (ev->button != Button5)) return false;
	if (TestWinIsEnabled((TWins*)self)) {
	    sbc = self->Count;
	    sbx = self->MouseZ_Wheel * self->SmallChange;
	    if (sbx) {
		if (ev->button == Button4) sbc -= sbx; 
				      else sbc += sbx;
		TScrollBar_SetCount_ev(self, sbc);
	    }
	}
	return true;
}

static void TScrollBar_DoMouseUp(TScrollBar *self, XButtonEvent *ev) {

	self->MouseDownSeg = 0;
	if (self->RepTimer) {
	    Tfree(self->RepTimer);
	    self->RepTimer = NULL;
	}
#ifdef _OS_WIN32
	Std_UngrabPointer();
#endif /* ! _OS_WIN32 */
	TWins_DoMouseUp((TWins*)self, ev);
	if (!TestWinIsEnabled((TWins*)self)) return;

	if (ev->button == 1) {
	    TWins_Refresh((TWins*)self);
	} else TScrollBar_MouseZ_Wheel(self, ev);
}

static void TScrollBar_scPosSet(TScrollBar *self) {
  TWRect rc;
  int64_t l;

	sb_GetClientRect(self, &rc);
	self->scPaint = ((self->MaxCount > 0) && (self->MaxCount > self->ButtonCountSize));
	if (self->scPaint) {
	    rc.width -= self->bulg * 2;
	    if (self->ButtonCountSize > 0) {
		l = rc.width * self->ButtonCountSize;
		l /= self->MaxCount;
		if (l < 10) l = 10;
	    } else l = self->bulg-1;
	    self->scLg = l;
	    rc.width -= l;
	    rc.left += self->bulg;
	    l = (int64_t)rc.width * (int64_t)self->Count;
	    l /= (self->MaxCount - self->ButtonCountSize);
	    self->scPos = l + rc.left;
	}
	TWins_Refresh((TWins*)self);
}

static void TScrollBar_DoMouseMove(TScrollBar *self, XMotionEvent *ev) {
  TWRect rc;
  int evx;
  int64_t m;

	sb_GetClientRect(self, &rc);
	if ((ev->state & Button1MotionMask) && (self->MouseDownSeg == 3)) {
	    
	    if (self->Hsc) evx = ev->x; else evx = ev->y;
	    m = evx - self->MouseDownPos + self->MDscPos;
	    rc.width -= self->scLg;
	    if (m > (rc.width - self->bulg)) m = rc.width - self->bulg;
	    if (m < self->bulg) m = self->bulg;
	    self->scPos = m + rc.left;
	    m -= self->bulg;
	    rc.width -= self->bulg << 1;
	    if (rc.width > 0) {
		m = m * (self->MaxCount - self->ButtonCountSize) / rc.width;
		if (m > 0x7fffffff) m = 0x7fffffff;
		SetCount(self, m);
	    }
	}
	TWins_DoMouseMove((TWins*)self, ev);
}

void TScrollBar_DoCount(TScrollBar *self, int Count) {
	if (self->Count != Count) {
	    self->Count = Count;
	    if (self->OnCount) self->OnCount(self, TScrollBar_GetCount(self));
	    TScrollBar_scPosSet(self);
	}
}

static void TScrollBar_ChkCount(TScrollBar *self, int *Count) {
  int res, max;

	res = *Count;
	res -= self->OffCount;
	max = self->MaxCount - self->ButtonCountSize;
	if (self->sbfu & sbfu_inv) res = max - res;
	if (res > max) res = max;
	if (res < 0) res = 0;
	*Count = res;
}

void TScrollBar_SetCount(TScrollBar *self, int Count) {
	TScrollBar_ChkCount(self, &Count);
	if (self->Count != Count) {
	    self->Count = Count;
	    TScrollBar_scPosSet(self);
	}
}

void TScrollBar_SetCount_ev(TScrollBar *self, int Count) {
	TScrollBar_ChkCount(self, &Count);
	TScrollBar_DoCount(self, Count);
}

void TScrollBar_SetButtonCountSize(TScrollBar *self, int ButtonCountSize) {
	if (self && (self->ButtonCountSize != ButtonCountSize)) {
	    self->ButtonCountSize = ButtonCountSize;
	    TScrollBar_scPosSet(self);
	}
}

void TScrollBar_SetMaxCount(TScrollBar *self, int MaxCount) {
	if (self && (self->MaxCount != MaxCount)) {
	    self->MaxCount = MaxCount;
	    TScrollBar_scPosSet(self);
	}
}

void TScrollBar_SetMaxCount_ButtonSize(TScrollBar *self, int MaxCount, int ButtonCountSize) {
	if (!self) return;
	self->MaxCount = MaxCount;
	self->ButtonCountSize = ButtonCountSize;
	TScrollBar_Refresh(self);
}

void TScrollBar_Refresh(TScrollBar *self) {
	if (!self) return;
	if (self->ButtonCountSize > self->MaxCount) self->ButtonCountSize = self->MaxCount;
	if ((self->ButtonCountSize + self->Count) > self->MaxCount) {
	    SetCount(self, self->MaxCount - self->ButtonCountSize);
	}
	TScrollBar_scPosSet(self); //+TWins_Refresh(self);
}

int TScrollBar_GetCount(TScrollBar *self) {
  int res = self->Count;
  int max = self->MaxCount - self->ButtonCountSize;
	if (res > max) res = max;
	if (res < 0) res = 0;
	if (self->sbfu & sbfu_inv) res = max - res;
	res += self->OffCount;
	return res;
}

void TScrollBar_SetCountParam(TScrollBar *self, int Count, int buCount, 
			      int MaxCount, int OffCount, int sbfu) {
	self->ButtonCountSize = buCount;
	self->MaxCount = MaxCount;
	self->OffCount = OffCount;
	self->sbfu = sbfu;
	TScrollBar_SetCount(self, Count);
}

static void TScrollBar_SetEnabled(TScrollBar *self, bool Value) {
}

/*_____________________________ Create __________________________________*/
static void TScrollBar_CreateParams(TWins_Create_struct *wcs) {
  TScrollBar *self = wcs->wins;

	self->AdjustSize = TScrollBar_AdjustSize;

	self->Refresh = TScrollBar_Refresh;

	self->DoPaint = TScrollBar_DoPaint;
	self->DoMouseDown = TScrollBar_DoMouseDown;
	self->DoMouseUp = TScrollBar_DoMouseUp;
	self->DoMouseMove = TScrollBar_DoMouseMove;
	self->SetEnabled = TScrollBar_SetEnabled;

	self->MaxCount = 100;
	self->SmallChange = 1;
	self->LargeChange = 10;
	self->bulg = 15;
	self->MouseZ_Wheel = appXDat.Default_MouseZ_Wheel;
}

TScrollBar *TScrollBar_Create(TWins *Parent) {
  TWins_Create_struct wcs;
	TWins_Create_struct_init(&wcs, sizeof(TScrollBar), wtyp_ScrollBar);
	wcs.Parent = Parent;
	wcs.CreateParams = TScrollBar_CreateParams;
	TWins_DefaultConf(&wcs);
	TScrollBar_scPosSet((TScrollBar*)wcs.wins);
	return wcs.wins;
}


TScrollBar *TScrollBarHorz_CreateP(TWins *Parent) {
  TScrollBar *self = TScrollBar_Create(Parent);
	self->Hsc = True;
	self->width = 60;
	self->height = 15;
	Set_Parent((TWins*)self, Parent);
	return self;
}

TScrollBar *TScrollBarVert_CreateP(TWins *Parent) {
  TScrollBar *self = TScrollBar_Create(Parent);
	self->width = 15;
	self->height = 60;
	Set_Parent((TWins*)self, Parent);
	return self;
}
