/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

static void TSplitterV_DoMouseDown(TSplitterV *self, XButtonEvent *ev) {
  TWins *wins;

	TWins_DoMouseDown((TWins*)self, ev);
	if (ev->button != 1) return;
	self->leftwin = self->rightwin = NULL;
	wins = self->prev;
	if (!wins) return;
	self->leftwin = wins;
	self->md_left = self->left;
	wins = self->next;
	if (!wins) return;
	self->rightwin = wins;
	self->rightwin_right = self->rightwin->left + self->rightwin->width;
}

static void TSplitterV_DoMouseMove(TSplitterV *self, XMotionEvent *ev) {
  int xo, w, wk, wks, wr;
  bool vis;

	TWins_DoMouseMove((TWins*)self, ev);
	if (!self->leftwin || !(ev->state & Button1MotionMask) ||
	    !TestWinIsEnabled((TWins*)self)) return;

	xo = ev->x_root - appXDat.MouseDown.x_root;
	w = (self->md_left - self->leftwin->left) + xo;
	wk = 3;
	if (self->leftwin->min_width > wk) wk = self->leftwin->min_width;
	wks = wk >> 2;
	if (wks > 20) wks = 20;
	wks = wk - wks;
	vis = w >= wks;
	if (w < wk) w = wk;
	if (self->rightwin) {
	    wr = self->rightwin_right - self->md_left - self->width - xo;
	    if (self->rightwin->min_width < 3) wr -= 3; else 
					       wr -= self->rightwin->min_width;
	} else {
	    wr = self->prev_child->width - self->md_left - self->width - xo;
	}
	if (wr < 0) w += wr;

	self->rmk |= 0x01;
	SetVisible(self->leftwin, vis);
	SetWidth(self->leftwin, w);
//	AppProcessMessages();
	self->rmk &= ~0x01;
}
/*
void TSplitterV_Resize(TSplitterV *self) {
  TWins *wins;
  TWins *leftwin;
  int mxl;
  bool mk = self->ResizeStartMk;

	TWins_Resize((TWins*)self);

	if (mk && self->IsParent && !(self->rmk & 0x01)) {
	    wins = self->prev;
	    if (!wins) return;
	    leftwin = wins;

	    mxl = self->prev_child->width - self->width;
	    wins = self->next;
	    while (wins) {
		if (wins->Align == alClient) {
		    if (wins->min_width < 3) mxl -= 3; else
						mxl -= wins->min_width;
		} else {
		    mxl -= wins->width;
		}
		wins = wins->next;
	    }
	    mxl -= self->left;
	    if (mxl < 0) {
		self->rmk |= 0x01;
		SetWidth(leftwin, leftwin->width + mxl);
//		AppProcessMessages();
		self->rmk &= ~0x01;
	    }
	}
}
*/
/*_____________________________ Destroy _________________________________*/

/*_____________________________ Create __________________________________*/
static void TSplitterV_CreateParams(TWins_Create_struct *wcs) {
  TSplitterV *self = wcs->wins;

	self->cursor_nr = XC_sb_h_double_arrow;
	self->width = 4;
	self->Align = alLeft;
	self->BorderFu = mBorder;
//	self->Resize = (TWinsEv)TSplitterV_Resize;
	self->DoMouseDown = TSplitterV_DoMouseDown;
	self->DoMouseMove = TSplitterV_DoMouseMove;
}

TSplitterV *TSplitterV_CreateP(TWins *Parent) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TSplitterV), wtyp_SplitterV);
	wcs.Parent = Parent;
	wcs.CreateParams = TSplitterV_CreateParams;
	TWins_DefaultConf(&wcs);
	Set_Parent((TWins*)wcs.wins, Parent);
	return wcs.wins;
}
