/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "tfilewin.h"
#include "dlg.h"
#include "cicons.h"

static void SettingSave(TFile_Dlg *mf) {
  TpcIni *pcIni;
	pcIni = get_globalIni();
	TFileWin_ini_WR(mf->FileWin, pcIni, "Dlg_File");
	TpcIni_Save(pcIni);
	free_globalIni(pcIni);
}

static void mCaseSort_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TFile_Dlg *mf = (TFile_Dlg*)self->FormStruct;
  int m;
	m = mf->FileWin->sortFu & ~dlsortFu_Case;
	if (mi->fu & miChecked) m |= dlsortFu_Case;
	TFileWin_SetSort(mf->FileWin, m);
	SettingSave(mf);
}

static void mItemDblClick_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TFile_Dlg *mf = (TFile_Dlg*)self->FormStruct;
	mf->FileWin->DblClick_ItemEnter = (mi->fu & miChecked) != 0;
	SettingSave(mf);
}

static void mPathDblClick_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TFile_Dlg *mf = (TFile_Dlg*)self->FormStruct;
	mf->FileWin->DblClick_PathEnter = (mi->fu & miChecked) != 0;
	SettingSave(mf);
}

static void menuSetting_OnMenuOpen(TMenu *self) {
  TFile_Dlg *mf = (TFile_Dlg*)self->FormStruct;

	mf->miItemDblClick->fu &= ~miChecked;
	if (mf->FileWin->DblClick_ItemEnter) mf->miItemDblClick->fu |= miChecked;
	mf->miPathDblClick->fu &= ~miChecked;
	if (mf->FileWin->DblClick_PathEnter) mf->miPathDblClick->fu |= miChecked;
}

//------------------------------------------------------------------

static void File_Dlg_OnItemEnter(TFileWin *self, TItemEnter *ie) {
  TFile_Dlg *fwd = (TFile_Dlg*)self->FormStruct;

	if (ie->art == ie_artItemSelect) {
	    if (ie->dfl && (ie->dfl->st->st & st_Reg)) 
		SetCaption((TWins*)fwd->edFile, ie->name_node);

	} else if (ie->art == ie_artFileEnter) {
	    SetCaption((TWins*)fwd->edFile, ie->name_node);
	    SetVisible((TWins*)fwd, False);
	    if (pclen(fwd->edFile->vds.s) > 0) fwd->RetCode = 1;
	}
}

static void File_Dlg_OnLoadDir(TFileWin *fw, char **fn, int *stat) {
  TFile_Dlg *fwd;
  char *pc;

	if (*stat & 0xff) return;
	fwd = (TFile_Dlg*)fw->FormStruct;
	pc = get_DisplayFileStr(*fn, NULL);
	SetCaption((TWins*)fwd->edPath, pc);
	xfree(pc);
	if (fwd->NewFile_Dlg) SetCaption(fwd->NewFile_Dlg->lblNewPath, *fn);
}

static void File_Dlg_PanelTop_OnResize(TWins *self) {
  TFile_Dlg *fw = (TFile_Dlg*)self->FormStruct;
  int w;

	SetLeft((TWins*)fw->edPath, fw->mbuSettings->left + fw->mbuSettings->width + 2);
	w = self->width - fw->edPath->left - fw->mbuPathRet->width - fw->mbuAzArt->width - 4;
	if (fw->mbuNewPath->Visible) w -= fw->mbuNewPath->width;
	SetWidth((TWins*)fw->edPath, w);
	w += fw->edPath->left + 2;
	SetLeft((TWins*)fw->mbuPathRet, w);
	w += fw->mbuPathRet->width;
	SetLeft((TWins*)fw->mbuAzArt, w);
	if (fw->mbuNewPath->Visible) {
	    w += fw->mbuNewPath->width;
	    SetLeft((TWins*)fw->mbuNewPath, w);
	}
}

static void File_Dlg_edPath_OnKeyDown(TComboBox *self, TKeyEvent *KeyEv) {
  TFile_Dlg *fwd;

	if (KeyEv->keysym == XK_Return) {
	    fwd = (TFile_Dlg*)self->FormStruct;
	    TFileWin_LoadDir(fwd->FileWin, self->vds.s);
	}
}

#ifdef _OS_WIN32
static void File_Dlg_edPath_OnPopUp(TComboBox *cb, TMenu *me) {
  dword DriveTyp;
  char DriveList[256];
  char *pc;
  TMenuItem *mi;

	TComboBox_ClsMenu(cb);
	GetLogicalDriveStrings(256, DriveList);
	pc = DriveList;
	while (*pc) {
	    mi = TMenu_AddItem(me, pc, 0, NULL);
	    DriveTyp = GetDriveType(pc);
	    switch(DriveTyp) {
	    case DRIVE_REMOVABLE:
		if (!pcncasecmpASCII(pc, "a:", 2) || !pcncasecmpASCII(pc, "b:", 2)) {
		    mi->pic = get_pic16_floppy_umount(False);
		} else {
		    mi->pic = get_pic16_usbdrive_umount(False);
		}
//		if (GetDiskFreeSpaceExA(pc, (PULARGE_INTEGER)&FreeBytes, (PULARGE_INTEGER)&TotalBytes, (PULARGE_INTEGER)&TotalFreeBytes)) {
// printf("%s(%d): %llu %llu %llu\n", FL, FreeBytes, TotalBytes, TotalFreeBytes);
//		}
		break;
	    case DRIVE_FIXED: mi->pic = get_pic16_hdd_umount(False); break;
	    case DRIVE_CDROM: mi->pic = get_pic16_cdrom_umount(False); break;
//	    default:
// printf("%s(%d): DriveTyp=%d\n", FL, DriveTyp);
	    } 
	    while (*pc) pc++;
	    pc++;
	}
}

static void File_Dlg_edPath_OnItemSel(TComboBox *cb, TMenuItem *SelItem) {
  TFile_Dlg *fwd = (TFile_Dlg*)cb->FormStruct;

	TFileWin_LoadDir(fwd->FileWin, fwd->edPath->vds.s);
}
#endif

static void File_Dlg_mbuPathRet_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
//	if (DblClick) return;
	TFileWin_PathEnter(((TFile_Dlg*)self->FormStruct)->FileWin, NULL, "..", ie_artPathEnter);
}

static void File_Dlg_mbuAzArt_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
  TFile_Dlg *fw;
  int m, AzArt;

//	if (DblClick) return;
	fw = (TFile_Dlg*)self->FormStruct;
	AzArt = fw->FileWin->AzArt;
	m = (AzArt & 0xff) + 1; //^ 0x01;
	if (m > 2) m = 0;
	TFileWin_SetColsArt(fw->FileWin, (AzArt & 0xffffff00) + m);
	if (m == 0x01) {
	    fw->mbuAzArt->pic = get_pic16_file_LAZ1(False);
	} else {
	    fw->mbuAzArt->pic = get_pic16_file_LAZ0(False);
	}
	RePaint((TWins*)self); 
}

static void File_Dlg_mbuNewPath_OnPaint(TMenuButton *self) {
	DrawLineRGB(self->WinID, 0x800000, 3, 13, 8, 13);
	DrawLineRGB(self->WinID, 0x800000, 3, 14, 8, 14);
	DrawLineRGB(self->WinID, 0x800000, 5, 11, 5, 16);
	DrawLineRGB(self->WinID, 0x800000, 6, 11, 6, 16);
}

static void File_Dlg_mbuNewPath_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
  TFile_Dlg *fwd;
  char *fn;
  TNewFile_Dlg *NewFile_Dlg;

//	if (DblClick) return;
	fwd = (TFile_Dlg*)self->FormStruct;
	SetEnabled((TWins*)fwd->mbuNewPath, False);
	NewFile_Dlg = Create_NewFile_Dlg((TWins*)fwd, ffnNewDirectory);
	fwd->NewFile_Dlg = NewFile_Dlg;
	NewFile_Dlg->gp = gp_dup(TFileWin_Get_gp(fwd->FileWin));
	if (NewFile_Dlg_Execute(NewFile_Dlg)) {
	    ppcchange(&NewFile_Dlg->gp->name, NewFile_Dlg->edFile->vds.s);
	    fn = gptopc(NewFile_Dlg->gp, gptfu_PasswdTxt|gptfu_add_slash);
	    TFileWin_LoadDir(fwd->FileWin, fn);
	    xfree(fn);
	}
	fwd->NewFile_Dlg = NULL;
	Tfree(NewFile_Dlg);
	SetEnabled((TWins*)fwd->mbuNewPath, True);
}

/* ********************** Bottom ********************************** */
static void File_Dlg_PanelBottom_OnResize(TWins *self) {
  TFile_Dlg *fw = (TFile_Dlg*)self->FormStruct;
  int x;

	x = self->width - fw->buOk->width - 2;
	SetLeft((TWins*)fw->buOk, x);
	SetLeft((TWins*)fw->buCancel, x);
	SetWidth((TWins*)fw->edFile, x - 10 - fw->edFile->left); 

	x -= (fw->cbHiddenFiles->width + 2);
	SetLeft((TWins*)fw->cbHiddenFiles, x);
	SetWidth((TWins*)fw->cobPattern, x - 6 - fw->cobPattern->left); 
}

static void File_Dlg_edFile_OnChange(TEdit *ed) {
  TFile_Dlg *fwd = (TFile_Dlg*)ed->FormStruct;
  bool mk = (ed->vds.s && (ed->vds.lg > 0)) || fwd->DirMode;

	SetEnabled((TWins*)fwd->buOk, mk);
}

static bool chk_OpenGetDir(TFile_Dlg *self) {
	return self->DirMode && (pclen(self->edPath->vds.s) > 0);
}

static void File_Dlg_edFile_OnKeyDown(TEdit *self, TKeyEvent *KeyEv) {
  TFile_Dlg *fwd = (TFile_Dlg*)self->FormStruct;

	if (fwd->NewFile_Dlg) return;
	if (KeyEv->keysym == XK_Return) {
	    if (TestWinIsEnabled((TWins*)fwd->buOk)) {
		if (pclen(fwd->edFile->vds.s) > 0) {
		    fwd->RetCode = fwdret_True;
		} else if (chk_OpenGetDir(fwd)) { //((fwd->fu & fwdfu_OpenGetDir) && (pclen(fwd->edPath->vds.s) > 0)) {
		    fwd->RetCode = fwdret_OnlyDir;
		}
	    } else {
		fwd->RetCode = 0;
	    }
	    SetVisible(self->FormStruct, False);
	}
}

static void File_Dlg_cbHiddenFiles_Click(TChkBox *self, long KeyState, bool DblClick) {
  TFile_Dlg *fw;

	fw = (TFile_Dlg*)self->FormStruct;
	TFileWin_SetShowHidden(fw->FileWin, ((TChkBox*)self)->Checked);
}

static void File_Dlg_buOk_Click(TButton *self, long KeyState, bool DblClick) {
  TFile_Dlg *fwd;

	fwd = (TFile_Dlg*)self->FormStruct;
	if (fwd->NewFile_Dlg) return;
	SetVisible(self->FormStruct, False);
	if (pclen(fwd->edFile->vds.s) > 0) {
	    fwd->RetCode = fwdret_True;
	} else if (chk_OpenGetDir(fwd)) {
	    fwd->RetCode = fwdret_OnlyDir;
	}
}

static void File_Dlg_buCancel_Click(TButton *self, long KeyState, bool DblClick) {

	SetVisible(self->FormStruct, False);
}

static void File_Dlg_Show(TFile_Dlg *self) {
	TForm_Show((TForm*)self);
	self->RetCode = 0;
}

bool File_Dlg_Execute(TFile_Dlg *self) {
  char *fn;

	ThDirList_Reg(1, &hdirlist_ld_File);
	gpFree(self->last_gp);
	self->last_gp = NULL;

	if (self->DirMode) {
	    if (!self->DirModeMix) {
		word ftyp[2] = {stft_Folder, 0};
		File_Dlg_AddPattern(self, NULL, ftyp);
	    }
	    SetEnabled((TWins*)self->buOk, True);
	}

	if (!self->path) ppcchange(&self->path, appDat.sHOME);
	SetVisible((TWins*)self, True);
	AppHandleMessage();

	fn = pctogptopc(self->path, gptfu_add_slash|gptfu_PasswdTxt);
	TFileWin_LoadDir(self->FileWin, fn);
	xfree(fn);

	Set_Focus((TWins*)self->FileWin);
	while (self->Visible && AppHandleMessage());
	ppcchange(&self->path, self->FileWin->Akt_FileName);
	self->last_gp = pctogp(self->FileWin->Akt_FileName);

	TFileWin_Clr(self->FileWin);
	return (self->RetCode > 0);
}

void File_Dlg_SetPath(TFile_Dlg *fw, const char *s) {
	ppcchange(&fw->path, s);
}

void File_Dlg_SetNamePath(TFile_Dlg *fw, const char *fn) {
  char *pc;
  xchar fn_use;
  const char *cpc;

	fn_use.s_const = fn;
	cpc = GetFileName_pc(fn);
	if (cpc && *cpc) {
	    while (*cpc && (*cpc != '*') && (*cpc != '?')) cpc++;
	    if (!*cpc) {				//no Pattern
		Tstat64 *st = net_stat(fn, NULL, fsFollowLink|fsNoFileName|fsNoLinkName|fsGetFileType|fsGetFileTypeHead);
		if (st && (st->st & st_Dir)) {
		    pc = pcadd(fn, "/");
		    fn_use.s = dotdot_korr(pc);
		    xfree(pc);
		}
		xfree(st);
	    }
	}

	cpc = GetFileName_pc(fn_use.s);
	SetCaption((TWins*)fw->edFile, cpc);

	pc = GetFilePath(fn_use.s);
	File_Dlg_SetPath(fw, pc);
	xfree(pc);

	if (fn_use.s_const != fn) xfree(fn_use.s);
}

void File_Dlg_SetName(TFile_Dlg *fw, const char *fn) {
  const char *cpc;
	cpc = GetFileName_pc(fn);
	if (cpc) SetCaption((TWins*)fw->edFile, cpc);
}

bool File_Dlg_GetEdFileName(TFile_Dlg *fwd, char **s) {
  TglobPath *gp;
  bool ok;

	ok = chk_OpenGetDir(fwd);
	if (!ok) ok = (pclen(fwd->edFile->vds.s) > 0);
	if (!ok) return False;
	xfree(*s);
	gp = pctogp(fwd->edPath->vds.s);
	gp_NameToPath(gp);
	
#ifdef _OS_WIN32
	fwd->RetName_gp_fu |= gptfu_noRootArt;
	if (gp->RootArt && !pccmp(gp->RootArt, "file:")) fwd->RetName_gp_fu &= ~gptfu_noRootArt;
#endif /* ! _OS_WIN32 */

	ppcchange(&gp->name, fwd->edFile->vds.s);
	if (fwd->last_gp && pccmp(gp->URL, fwd->last_gp->URL)) {
	    ppcchange(&gp->passwd, fwd->last_gp->passwd);
	}
	*s = gptopc(gp, fwd->RetName_gp_fu|gptfu_PasswdTxt);
	gpFree(gp);
	return True;
}

/*_________________________ Pattern ________________________________________*/
TPattern *File_Dlg_Get_Patter(TFile_Dlg *fw) {
	return TFileWin_get_pattern((TFileWin*)fw->FileWin);
}

typedef struct {
  char *pattern;
  word *ftyp;
} TFile_Dlg_Pattern_obj;

static void File_Dlg_cobPattern_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TFile_Dlg *fwd;
  TFile_Dlg_Pattern_obj *cob_miobj;
  TPattern *patt;
  const char *ext;

	if (!mi ||!mi->obj) return;
	cob_miobj = mi->obj;
	fwd = (TFile_Dlg*)self->FormStruct;
	TFileWin_set_pattern(fwd->FileWin, cob_miobj->pattern, cob_miobj->ftyp);
	TFileWin_ReLoadDir(fwd->FileWin);

	patt = TFileWin_get_pattern(fwd->FileWin);
	if (patt && patt->ppa && patt->ppa[0] && !patt->ppa[1] && 
	    fwd->edFile->vds.s && fwd->edFile->vds.lg) {
	    ext = GetFileExt_pc(patt->ppa[0]);
	    if (ext && *ext) {
		ChangeFileExt_ppc(&fwd->edFile->vds.s, ext);
		fwd->edFile->vds.lg = pclen(fwd->edFile->vds.s);
		Refresh((TWins*)fwd->edFile);
	    }
	}

}

static bool fdlg_TMenuItem__Destroy(TMenuItem *self) {
  TFile_Dlg_Pattern_obj *cob_miobj = self->obj;
	if (cob_miobj) {
	    xfree(cob_miobj->ftyp);
	    xfree(cob_miobj->pattern);
	    xfree(cob_miobj);
	}
	return TMenuItem__Destroy(self);
}

static void File_Dlg_Add_cobPattern(TFile_Dlg *fw, const char *pattern, unsigned short *ftyp) {
  TMenuItem *mi;
  TFile_Dlg_Pattern_obj *cob_miobj;
  int z;
  word *pw, *pwd;

	cob_miobj = xcalloc(sizeof(TFile_Dlg_Pattern_obj));

	if (ftyp) {		//copy FileType
	    pw = ftyp;
	    z = 0;
	    while (*pw) {z++; pw++;}
	    if (z) {
		pwd = xmalloc(sizeof(word) * (z+1));
		cob_miobj->ftyp = pwd;
		pw = ftyp;
		while (*pw) {*pwd = *pw; pwd++; pw++;}
		*pwd = 0;
	    }
	}
	cob_miobj->pattern = pcdup(pattern);

	mi = TComboBox_AddItem(fw->cobPattern, pattern, 0, File_Dlg_cobPattern_OnClick);
	mi->obj = cob_miobj;
	mi->destroy = (TDESTROY)fdlg_TMenuItem__Destroy;
}


void File_Dlg_AddPattern(TFile_Dlg *fw, const char *pattern, unsigned short *ftyp) {
/* TODO nicht fertig zB.:ftyp or change .....*/
	if (!fw->cobPattern->popup->MenuItem) {
	    TFileWin_set_pattern(fw->FileWin, pattern, ftyp);
	    SetCaption((TWins*)fw->cobPattern, pattern);
	}
	File_Dlg_Add_cobPattern(fw, pattern, ftyp);
//	TComboBox_AddItem(fw->cobPattern, pattern, 0, File_Dlg_cobPattern_OnClick);
};

static void File_Dlg_OnCanClose(TFile_Dlg *self, int *CanClose) {
	if (self->NewFile_Dlg) {
	    self->NewFile_Dlg->RetCode = 0;
	    SetVisible((TWins*)self->NewFile_Dlg, False);
	}
}

/*_____________________________ Destroy _________________________________*/
static bool TFile_Dlg__Destroy(TFile_Dlg *self) {
	gpFree(self->last_gp);
	xfree(self->path);
	return TForm__Destroy((TForm*)self);
}

/* ____________________________ Create _______________________________ */
TFile_Dlg *Create_FileAlg_Dlg(TWins *parent) {
  TFile_Dlg *self;
  TpcIni *pcIni;

	self = TForm_CreatParamAlg(parent, sizeof(TFile_Dlg));
	self->destroy = (TDESTROY)TFile_Dlg__Destroy;
	self->Visible = False;
	if (parent) self->FormZStyle = fszTransient;
	self->FormPosition = poCenter;
	self->Show = File_Dlg_Show;
	self->OnCanClose = File_Dlg_OnCanClose;

#ifndef _OS_WIN32
	self->Icon = GetDeskIcon(ddkt_Mime, 16, "filedlg", 
				    gdifu_noZoom, zico16globus);
#endif /* ! _OS_WIN32 */

	pcIni = get_globalIni();

	self->width =  TpcIni_RD_int(pcIni, "Dlg_File", "Width",  556);
	self->height = TpcIni_RD_int(pcIni, "Dlg_File", "Height", 270);
	self->min_width = 360;
	self->min_height = 200;

	self->PanelTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder | mBorderL, alTop);
	self->PanelTop->OnResize = File_Dlg_PanelTop_OnResize;
// ******* Menu Setting *******
	self->mbuSettings = TMenuButton_CreateS(self->PanelTop, (TMenuButton*)self, "Settings");
	self->menuSetting = TMenu_Create((TWins*)self);
	self->mbuSettings->menu = self->menuSetting;
	self->menuSetting->OnMenuOpen = menuSetting_OnMenuOpen;

	self->miCaseSort = TMenu_AddItem(self->menuSetting, "Case Sort", 
				miAutoToggle, mCaseSort_OnClick);
	TMenu_AddItem(self->menuSetting, "-", 0, NULL);
	self->miItemDblClick = TMenu_AddItem(self->menuSetting, "File DblClick", 
				miAutoToggle, mItemDblClick_OnClick);
	self->miPathDblClick = TMenu_AddItem(self->menuSetting, "Path DblClick", 
				miAutoToggle, mPathDblClick_OnClick);
	
	self->edPath = TComboBox_CreateS(self->PanelTop, 88, 2, 200, 18, NULL);
	self->edPath->OnKeyDown = File_Dlg_edPath_OnKeyDown;
#ifdef _OS_WIN32
	self->edPath->OnPopUp = File_Dlg_edPath_OnPopUp;
	self->edPath->OnItemSel = File_Dlg_edPath_OnItemSel;
#endif

	self->mbuPathRet = TMenuButton_Create(self->PanelTop, self->mbuSettings);
	self->mbuPathRet->pic = get_pic16_undo(False);
	self->mbuPathRet->OnClick = File_Dlg_mbuPathRet_OnClick;
	Set_Parent((TWins*)self->mbuPathRet, self->PanelTop);

	self->mbuAzArt = TMenuButton_Create(self->PanelTop, self->mbuSettings);
	self->mbuAzArt->pic = get_pic16_file_LAZ1(False);
	self->mbuAzArt->OnClick = File_Dlg_mbuAzArt_OnClick;
	Set_Parent((TWins*)self->mbuAzArt, self->PanelTop);
	self->mbuNewPath = TMenuButton_Create(self->PanelTop, self->mbuSettings);
	self->mbuNewPath->pic = get_fileIcon(stft_Folder);
	self->mbuNewPath->dis_pic = get_fileDisIcon(stft_Folder);
	self->mbuNewPath->OnPaint = File_Dlg_mbuNewPath_OnPaint;
	self->mbuNewPath->OnClick = File_Dlg_mbuNewPath_OnClick;
	self->mbuNewPath->Visible = False;
	Set_Parent((TWins*)self->mbuNewPath, self->PanelTop);

	TMenuButton_SetPosition(self->mbuSettings); //Set Default size

	self->FileWin = TFileWin_CreateA((TWins*)self, 0, 0, 200, 200, alClient);
#ifndef _OS_WIN32
	self->FileWin->vds.UTF8on = TpcIni_RD_boolean(pcIni, "Global", "FileSystemUTF8", 0);
#endif
	TFileWin_SetColsArt(self->FileWin, 1);
	self->FileWin->OnLoadDir = File_Dlg_OnLoadDir;
	self->FileWin->OnItemEnter = File_Dlg_OnItemEnter;

	TFileWin_ini_RD(self->FileWin, pcIni, "Dlg_File");

	if (self->FileWin->sortFu & dlsortFu_Case) {
	    self->miCaseSort->fu |= miChecked;
	}

	free_globalIni(pcIni);


	self->PanelBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 48, 0, alBottom);
	self->PanelBottom->OnResize = File_Dlg_PanelBottom_OnResize;
	self->edFile = TEdit_CreateS(self->PanelBottom, 52, 4, 45, 18, NULL);
	self->edFile->OnChange = File_Dlg_edFile_OnChange;
	self->edFile->OnKeyDown = File_Dlg_edFile_OnKeyDown;

	self->cobPattern = TComboBox_CreateS(self->PanelBottom, 52, 26, 100, 18, NULL);

	self->cbHiddenFiles = TChkBox_CreateS(self->PanelBottom, 154, 28, 130, 15, "Show Hidden Files");
	self->cbHiddenFiles->OnClick = File_Dlg_cbHiddenFiles_Click;

	self->buOk = TButton_CreateS(self->PanelBottom, 100, 2, 64, 22, "OK");
	self->buOk->OnClick = File_Dlg_buOk_Click;
	self->buOk->Enabled = False;
	self->buCancel = TButton_CreateS(self->PanelBottom, 100, 24, 64, 22, "Cancel");
	self->buCancel->OnClick = File_Dlg_buCancel_Click;

	self->lblFile = TWins_CreateLabel(self->PanelBottom, 4, 5, 45, 18, "File", mCaptionLeft, 0);
	self->lblPattern = TWins_CreateLabel(self->PanelBottom, 4, 27, 45, 18, "Pattern", mCaptionLeft, 0);
	return self;
}

TFile_Dlg *Create_File_Dlg(TWins *parent) {
  TFile_Dlg *self;

	self = Create_FileAlg_Dlg(parent);
	SetCaption((TWins*)self, "File Open Dlg");
	return self;
}

TFile_Dlg *Create_FileSave_Dlg(TWins *parent) {
  TFile_Dlg *self;

	self = Create_FileAlg_Dlg(parent);
	SetVisible((TWins*)self->mbuNewPath, True);
	SetCaption((TWins*)self, "File Save Dlg");
	return self;
}

