/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "dlg.h"
#include "cicons.h"

static int get_akt_dpi(TFont_Dlg *dlg) {
  char *pc = dlg->tv->dpi89;
	if (*pc == '*') return 0;  //*-*
	if (*pc == '0') return 1;  //0-0
	if (*pc == '1') return 3;  //100-100
	return 2;		  //75-75
}

/* ______________________ MenuCharset _____________________________________ */

static void mCharset_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;
  int tag;

	tag = get_akt_dpi(dlg);
	if (tag == 1) {
	    dlg->defmi_scale = smi;
	} else if (tag == 2) {
	    dlg->defmi_75x = smi;
	} else if (tag == 3) {
	    dlg->defmi_100 = smi;
	} else {
	    dlg->defmi_x = smi;
	}

	ppcchange(&dlg->tv->Charset12, smi->s);
	SetCaption((TWins*)dlg->mbCharset, dlg->tv->Charset12);
	TreeVFont_List(dlg->tv);
}

static void mCharset_set_en(TFont_Dlg *dlg, TMenuItem *mi, int tag) {
  TMenuItem *defmi = dlg->defmi_x;
  int mitag;

	mitag = mi->tag;
	if (tag == 1) {
	    defmi = dlg->defmi_scale;
	    mitag &= 0x01;
	} else if (tag == 2) {
	    defmi = dlg->defmi_75x;
	    mitag &= 0x02;
	} else if (tag == 3) {
	    defmi = dlg->defmi_100;
	    mitag &= 0x04;
	} else {
	    defmi = dlg->defmi_x;
	    mitag &= 0x07;
	}

	mi->fu |= miNotEnabled;
	if (mi->tag & mitag) mi->fu &= ~miNotEnabled;

	mi->fu &= ~miChecked;
	if (defmi && (defmi == mi)) mi->fu |= miChecked;
}

static void mCharset_OnMenuOpen(TMenu *self) {
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;
  int tag;
  TMenuItem *mi;

	tag = get_akt_dpi(dlg); 
	mi = self->MenuItem;
	while (mi) {
	    mCharset_set_en(dlg, mi, tag);
	    mi = mi->next;
	}
	mi = NULL;
	if (dlg->menuIso8859) mi = dlg->menuIso8859->MenuItem;
	while (mi) {
	    mCharset_set_en(dlg, mi, tag);
	    mi = mi->next;
	}
}

#define MAX_iso8859 20
#define MAX_xFont 40

static void MenuCharset_Create(TFont_Dlg *dlg) {
	dlg->mbCharset = TMenuButton_Create(dlg->pnlTop, (TMenuButton*)dlg);
	SetCaption((TWins*)dlg->mbCharset, "Charset");
	dlg->mbCharset->AutoSize = False;
	SetWidth((TWins*)dlg->mbCharset, 100);
	Set_Parent((TWins*)dlg->mbCharset, dlg->pnlTop);

	dlg->menuCharset = TMenu_Create((TWins*)dlg);
	dlg->menuCharset->OnMenuOpen = mCharset_OnMenuOpen;
	dlg->mbCharset->menu = dlg->menuCharset;
}

static void mCharset_add_default(TFont_Dlg *dlg, TMenuItem *mi, int cs) {
	if ((mi->tag & 0x01)&&(!dlg->defmi_scale ||(cs == 1))) dlg->defmi_scale = mi;
	if ((mi->tag & 0x02)&&(!dlg->defmi_75x   ||(cs == 1))) dlg->defmi_75x = mi;
	if ((mi->tag & 0x04)&&(!dlg->defmi_100   ||(cs == 1))) dlg->defmi_100 = mi;
	if ((mi->tag & 0x07)&&(!dlg->defmi_x     ||(cs == 1))) dlg->defmi_x = mi;
}

static void __attribute__ ((noinline)) MenuCharset_Init(TFont_Dlg *dlg) {
  int L, x, tag;
  char **FontNameList, *pc;
  byte *iso8859;
  char **xFont;
  int *iso8859_tag, *xFont_tag;
  int iso8859_z;
  int xFont_z;
  TMenuItem *mi;
  char buf[32];
  TFontLgs fl;

	FontNameList = XListFonts(appXDat.dpy, "-*-*-*-*-*-*-*-*-*-*-*-*-*-*", 99999, &L);
	if (!FontNameList) return;
	iso8859 = xcalloc(MAX_iso8859);
	iso8859_tag = xcalloc(sizeof(int) * MAX_iso8859);
	xFont = xcalloc(sizeof(char*) * MAX_xFont);
	xFont_tag = xcalloc(sizeof(int) * MAX_xFont);
	xFont_z = 0;
	iso8859_z = 0;

	for (int i=0;i<L;i++) {
	    pc = FontNameList[i];
	    if (!FontLgsArrFill(&fl, pc)) continue;
	    tag = 0;
	    if (fl.lgs[8].lg == 0)  continue;
#ifdef _OS_WIN32
	    if (pcncmp(fl.lgs[8].s, "96", 2)) {
		if (pctoint(fl.lgs[8].s) == 0) continue;
		tag = 0x02;
#else
	    if (pcncmp(fl.lgs[8].s, "75", 2)) {
		if ((fl.lgs[11].lg == 1) && (*fl.lgs[11].s == '0')) continue;
		tag = 0x02;
	    } else if (pcncmp(fl.lgs[8].s, "100", 3)) {
		if ((fl.lgs[11].lg == 1) && (*fl.lgs[11].s == '0')) continue;
		tag = 0x04;
#endif
	    } else if (pcncmp(fl.lgs[8].s, "0", 1)) {
		if ((fl.lgs[11].lg != 1) || (*fl.lgs[11].s != '0')) continue;
		if ((fl.lgs[10].lg = 1) && (*fl.lgs[10].s == 'c')) continue;
		tag = 0x01;
	    }
	    if (!tag) continue;

	    pc = fl.lgs[12].s;
	    if (*pc) {
		if (pcncmp(pc, "iso8859-",8)) {
		    x = pctoint(pc+8);
		    if ((x >= 0) && (x < 20)) {
			if (!iso8859[x]) {
			    iso8859[x] = 1;
			    iso8859_z++;
			}
			iso8859_tag[x] |= tag;
		    }

		} else if (xFont_z < MAX_xFont) {
		    for (x=0;x<MAX_xFont;x++) {
			if (!xFont[x]) {
			    xFont[x] = pcdup(pc);
			    xFont_tag[x] |= tag;
			    xFont_z++;
			    break;
			}
			if (pccmp(xFont[x], pc)) {
			    xFont_tag[x] |= tag;
			    break;
			}
		    }
		}
	    }
	}
	XFreeFontNames(FontNameList);


	if (iso8859_z) {
  TMenu *me = dlg->menuCharset;
	    if (iso8859_z > 2) {
		mi = TMenu_AddItem(dlg->menuCharset, "iso8859-x", 0, NULL);
		mi->tag = 0x07;
		me = TMenu_Create((TWins*)dlg);
		dlg->menuIso8859 = me;
		mi->next_menu = me;
	    }
	    pcmove(buf, "iso8859-");
	    for (int i=0;i<MAX_iso8859;i++) {
		if (iso8859[i]) {
		    inttopc(buf+8, i);
		    mi = TMenu_AddItem(me, buf, miCanCheck, mCharset_OnClick);
		    mi->tag = iso8859_tag[i];
		    mCharset_add_default(dlg, mi, i);
		}
	    }
	}

	for (int i=0;i<xFont_z;i++) {
	    mi = TMenu_AddItem(dlg->menuCharset, NULL, miCanCheck, mCharset_OnClick);
	    mi->s = xFont[i];
	    mi->lg = pclen(mi->s);
	    mi->tag = xFont_tag[i];
	    mCharset_add_default(dlg, mi, 0);
	}
	xfree(xFont);
	xfree(xFont_tag);
	xfree(iso8859);
	xfree(iso8859_tag);
}

/* ______________________ Font AZ  ______________________________________ */
#ifdef _OS_WIN32
#define MAX_fsTAB 16
static int size_tab[16] = {
  11,12,13,15,16,19,21,24,
  27,29,32,35,37,48,64,96
};
#else
#define MAX_fsTAB 18
static int size_tab[MAX_fsTAB] = {
  8, 9, 10,11,12,13,14,15,
  16,18,20,22,24,28,32,40,
  48,64
};
#endif

static void cobSIZE_pc(char *pc, int fontHeight) {
	inttoppc(&pc, fontHeight);
#ifdef _OS_WIN32
	*pc = ' '; pc++;
	*pc = '('; pc++;
	inttoppc(&pc, MulDiv(fontHeight, 72, appXDat.LogPixel));
	*pc = ')'; pc++; *pc = 0;
#endif
}

static int GetFontHeight(const char *fontname) {
  int lg;
  const char *cpc;
	GetFontSXTypeIdx(6, fontname, &cpc, &lg);
	return pctoint(cpc);
}

static void __attribute__ ((noinline)) set_cobSIZE(TFont_Dlg *dlg, char *name) {
  int L;
  char *pc0, **FontNameList;
  char buf[32];
  bool scale;

	TComboBox_ClsMenu(dlg->cobSIZE);
	if (!name) return;

	scale = ChkScaleFont(name);

	pc0 = pcdup(name);
	SetFontSXVTypeIdx(6, &pc0, "*");
	SetFontSXVTypeIdx(7, &pc0, "*");
//	SetFontSXVTypeIdx(10, &pc0, "*");
	if (scale) {
	    for (int i=0;i<MAX_fsTAB;i++) {
		inttopc(buf, size_tab[i]);
		SetFontSXVTypeIdx(6, &pc0, buf);
		if (FontExists(pc0)) {
		    cobSIZE_pc(buf, size_tab[i]);
		    TComboBox_AddItem(dlg->cobSIZE, buf, 0, NULL);
		}
	    }

	} else {
//    printf("%s(%d): %s\n", FL, pc0);
	    SetFontSXVTypeIdx(11, &pc0, "*");
	    FontNameList = XListFonts(appXDat.dpy, pc0, 99999, &L);
	    if (FontNameList) {
		for (int i=0;i<L;i++) {
// printf("%s(%d): -> %s\n", FL, FontNameList[i]);
		    int m = GetFontHeight(FontNameList[i]);
		    cobSIZE_pc(buf, m);
		    TComboBox_AddItem(dlg->cobSIZE, buf, 0, NULL);
		}
		XFreeFontNames(FontNameList);
	    }
	}
	xfree(pc0);
}

static void TFont_Dlg_LE_SetFont(TFont_Dlg *self, int fu) {
  int L, size;
  char *pc, *pc0, *name;
  const char *cpc;
  char buf[32];

	TChkBox_SetChecked(self->cbN, False);
	TChkBox_SetChecked(self->cbNK, False);
	TChkBox_SetChecked(self->cbB, False);
	TChkBox_SetChecked(self->cbBK, False);

	pc0 = pcdup(GetCaption_pc((TWins*)self->edTop));
	size = GetFontHeight(pc0);
	if (size < 1) {
	    size = self->tv->size6;
	    inttopc(buf, size);
	    SetFontSXVTypeIdx(6, &pc0, buf);
	}

// printf("%s(%d): %s\n", FL, pc0);
	GetFontSXTypeIdx(2, pc0, &cpc, &L);
	if (pcncmp(cpc, "*-*-", 4)) {	//set default Medium
	    if (GetFontArtName(pc0, FONT_Medium, &pc)) {
		xfree(pc0);
		pc0 = pc;
	    } else if (GetFontArtName(pc0, FONT_Bold, &pc)) {
		xfree(pc0);
		pc0 = pc;
	    }
	}

	if (!FontExists(pc0)) {
	    SetFontSXVTypeIdx(6, &pc0, "*");
	    name = GetFont_NextSize(pc0, size);
	} else {
	   name = GetFullFontName(pc0);
	}
	xfree(pc0);
	if (!name) return;

	if (fu & 0x01) set_cobSIZE(self, name);

	size = GetFontHeight(name);
	cobSIZE_pc(buf, size);
	SetCaption((TWins*)self->cobSIZE, buf);

	SetCaption((TWins*)self->edResult, name);
	SetFont((TWins*)self->le, name);
	Refresh((TWins*)self->le);

	switch (GetFont_Art(name)) {
	case FONT_Medium:  TChkBox_SetChecked(self->cbN, True); break;
	case FONT_MediumK: TChkBox_SetChecked(self->cbNK,True); break;
	case FONT_Bold:    TChkBox_SetChecked(self->cbB, True); break;
	case FONT_BoldK:   TChkBox_SetChecked(self->cbBK,True); break;
	}

	SetEnabled((TWins*)self->cbN, GetFontArtName(name, FONT_Medium | FONT_NoResize, NULL));
	SetEnabled((TWins*)self->cbNK,GetFontArtName(name, FONT_MediumK | FONT_NoResize, NULL));
	SetEnabled((TWins*)self->cbB, GetFontArtName(name, FONT_Bold | FONT_NoResize, NULL));
	SetEnabled((TWins*)self->cbBK,GetFontArtName(name, FONT_BoldK | FONT_NoResize, NULL));

	xfree(name);
}

/* ______________________ On Ev _________________________________________ */
//-misc-fixed-bold-r-normal-*-13-120-75-75-c-70-iso10646-1
// 0    1     2    3 4      5 6  7   8  9  1 1  1        1
//                                         0 1  2        3

static void tv_OnNodeClick(TTreeV *tv, TTreeVNode *tvn, bool DblClick) {
  TFont_Dlg *dlg = (TFont_Dlg*)tv->FormStruct;

	SetCaption((TWins*)dlg->edTop, dlg->tv->selFontName);
	TFont_Dlg_LE_SetFont(dlg, 0x01);
}

static void cbDPI_OnItemSel(TComboBox *self, TMenuItem *SelItem) {
  TFont_Dlg *dlg;
  const char *tv_dpi;
  TMenuItem *mi;

	dlg = (TFont_Dlg*)self->FormStruct;
	tv_dpi = "*-*";
	mi = NULL;
#ifdef _OS_WIN32
	if (pcncmp(SelItem->s, "96", 2)) {
	    tv_dpi = "96-96";
	    mi = dlg->defmi_75x;
#else
	if (pcncmp(SelItem->s, "75", 2)) {
	    tv_dpi = "75-75";
	    mi = dlg->defmi_75x;
	} else if (pcncmp(SelItem->s, "100", 3)) {
	    tv_dpi = "100-100";
	    mi = dlg->defmi_100;
#endif
	} else if (pcncmp(SelItem->s, "scale", 5)) {
	    tv_dpi = "0-0";
	    mi = dlg->defmi_scale;
	} else {
	    mi = dlg->defmi_x;
	}
	if (ppcchange_cmp(&dlg->tv->dpi89, tv_dpi)) {
	    if (mi) {
		ppcchange(&dlg->tv->Charset12, mi->s);
		SetCaption((TWins*)dlg->mbCharset, dlg->tv->Charset12);
	    }
	    TreeVFont_List(dlg->tv);
	}
}

static void edTop_OnKeyUp(TEdit *self, TKeyEvent *KeyEv) {
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;
	if (KeyEv->keysym == XK_Return) {
	    TFont_Dlg_LE_SetFont(dlg, 0x00);
	}
}

static void mf_SetDefaultFontSize(TFont_Dlg *mf, int size) {
  char buf[64];

	if (size < 2) size = 2;
	if (size > 255) size = 255;
	if (mf->tv->size6 == size) return;
	mf->tv->size6 = size;
	snprintf(buf, sizeof(buf), "Select Font     (default=%d)", size);
	SetCaption((TWins*)mf, buf);
}

static void cbSIZE_OnItemSel(TComboBox *self, TMenuItem *SelItem) {
  int m;
  char *name;
  char buf[16];
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;

	name = pcdup(GetCaption_pc((TWins*)dlg->edResult));
	m = pctoint(SelItem->s);
	mf_SetDefaultFontSize(dlg, m);
	inttopc(buf, m);
	SetFontSXVTypeIdx(6, &name, buf);
	SetFontSXVTypeIdx(7, &name, "*");
//	SetFontSXVTypeIdx(10, &name, "*");
	SetFontSXVTypeIdx(11, &name, "*");

	if (!FontExists(name)) return;
	SetCaption((TWins*)dlg->edTop, name);
	xfree(name);
	TFont_Dlg_LE_SetFont(dlg, 0);
}

static void le_OnSetCaret(TLineEdit *le, TColS *cs) {
  TFont_Dlg *mf = (TFont_Dlg*)le->FormStruct;
  int c;
  const char *s;
  int x = le->CaretCharIdx;
  char tmps[32];

	if (cs) {
	    if (cs->s && (cs->lg > x)) {
		if (le->vds.UTF8on) {
		    c = Utf8CharToWChar(cs->s + x);
		    if (c & Char_ErrorUTF8) c = (unsigned char)cs->s[x];
		} else {
		    c = (byte)cs->s[x];
		}
		if (c > 999) {
		    snprintf(tmps, 32, "%d/0x%02x", c, c);
		} else {
		    snprintf(tmps, 32, "%d / 0x%02x", c, c);
		}
		s = tmps;
	    } else {
		s = "EOL";
	    }
	} else s = "EOF";
	SetCaption(mf->CharAzBox, s);
}

static void cbUTF8_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFont_Dlg *mf = (TFont_Dlg*)self->FormStruct;
	mf->le->vds.UTF8on = self->Checked;
	Refresh((TWins*)mf->le);
}

static void cbN_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFont_Dlg *mf = (TFont_Dlg*)self->FormStruct;
  int art;
  char *pc;

	TChkBox_SetChecked(mf->cbN, False);
	TChkBox_SetChecked(mf->cbNK, False);
	TChkBox_SetChecked(mf->cbB, False);
	TChkBox_SetChecked(mf->cbBK, False);
	TChkBox_SetChecked(self, True);

	art = FONT_Medium | FONT_NoResize;
	if (self == mf->cbNK) {
	    art = FONT_MediumK | FONT_NoResize;
	} else if (self == mf->cbB) {
	    art = FONT_Bold | FONT_NoResize;
	} else if (self == mf->cbBK) {
	    art = FONT_BoldK | FONT_NoResize;
	}

	if (GetFontArtName(GetCaption_pc((TWins*)mf->edResult), art, &pc)) {
	    SetCaption((TWins*)mf->edResult, pc);
	    SetFont((TWins*)mf->le, pc);
	    Refresh((TWins*)mf->le);
	    xfree(pc);
	}
}

static void buCancel_OnClick(TButton *self, long KeyState, bool DblClick) {
	SetVisible(self->FormStruct, False);
}

static void buOK_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;
	SetVisible(self->FormStruct, False);
	dlg->RetCode = 1;
}

/* ______________________ Resize _____________________________________ */
static void pnlTop_OnResize(TWins *self) {
  int w;
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;
  TWins *wins = (TWins*)dlg->mbCharset;

	if (dlg->mbHelp) wins = (TWins*)dlg->mbHelp;
	SetLeft((TWins*)dlg->cobDPI, wins->left + wins->width + 4);

	w = self->width - dlg->edTop->left - 2;
	if (w < 4) w = 4;
	SetWidth((TWins*)dlg->edTop, w);

}

static void pnlRightBottom_OnResize(TWins *self) {
  TFont_Dlg *dlg = (TFont_Dlg*)self->FormStruct;
  int w;
	SetWidth((TWins*)dlg->edResult, self->width - (dlg->edResult->left * 2));
	w = self->width - dlg->buOK->width - 3;
	SetLeft((TWins*)dlg->buOK, w);
	w -= dlg->buCancel->width + 8;
	SetLeft((TWins*)dlg->buCancel, w);
}

//------------------------ TLineEdit -------------------------------------

static const char *Test_txt0 = {
  "0123456789ABCDEF0123456789abcdef0123456789ABCDEF0123456789abcdef\n"
};

static const char *Test_txt2 = {
  "................................................................\n"
  "UTF-8 encoded sample file for Postscript plain-text file formatters\n"
  "-------------------------------------------------------------------\n"
  "0040 @ COMMERCIAL AT\n"
  "0041 A Latin LETTER A\n"
  "0042 B Latin LETTER B\n"
  "00C0 À Latin LETTER A WITH GRAVE\n"
  "00D2 Ò Latin LETTER O WITH GRAVE\n"
  "00D9 Ù Latin LETTER U WITH GRAVE\n"
  "20AC € EURO SIGN\n"
  "2122 ™ TRADE MARK SIGN\n"
};

static void Set_txt(TLineEdit *le) {
  Tlgs lgs;

	TLineEdit_Clr(le);
	TLineEdit_AddPaste(le, Test_txt0);
	lgs.s = xmalloc(64);
	lgs.lg = 64;

	for (int i=0;i<64;i++) lgs.s[i] = i;
	TLineEdit_AddLine_lgs(le, &lgs);
	for (int i=0;i<64;i++) lgs.s[i] = i+0x40;
	TLineEdit_AddLine_lgs(le, &lgs);
	for (int i=0;i<64;i++) lgs.s[i] = i+0x80;
	TLineEdit_AddLine_lgs(le, &lgs);
	for (int i=0;i<64;i++) lgs.s[i] = i+0xc0;
	TLineEdit_AddLine_lgs(le, &lgs);
	xfree(lgs.s);
	TLineEdit_AddLine(le, NULL);

	TLineEdit_AddPaste(le, Test_txt2);
}

static void Font_Dlg_OnShow(TFont_Dlg *self) {
  char *fontname = pcdup(GetCaption_pc((TWins*)self->edTop));
  TMenuItem *mi;

	AppHandleMessage();
	MenuCharset_Init(self);
	AppHandleMessage();
	TTreeVFont_SetDefaultParam(self->tv, fontname);

	mi = self->menuCharset->MenuItem;
	while (mi) {
	    if (pccmp(self->tv->Charset12, mi->s)) break;
	    mi = mi->next;
	}
	if (!mi) {
	    if (self->menuIso8859) mi = self->menuIso8859->MenuItem;
	    while (mi) {
		if (pccmp(self->tv->Charset12, mi->s)) break;
		mi = mi->next;
	    }
	}
	if (mi) {
	    int tag = get_akt_dpi(self);
	    if (tag == 1) {
		self->defmi_scale = mi;
	    } else if (tag == 2) {
		self->defmi_75x = mi;
	    } else if (tag == 3) {
		self->defmi_100 = mi;
	    } else {
		self->defmi_x = mi;
	    }
	}

	AppHandleMessage();

	mf_SetDefaultFontSize(self, self->tv->size6);
// printf("%s(%d): %d\n", FL, self->font_size);
	TreeVFont_List(self->tv);
	AppHandleMessage();
	TTreeVFont_ExpandDefaultNode(self->tv);
	AppHandleMessage();

	TFont_Dlg_LE_SetFont(self, 0x01);
	Set_txt(self->le);

	SetEnabled((TWins*)self, True);
	xfree(fontname);
	SetCaption((TWins*)self->mbCharset, self->tv->Charset12);
}

/*_____________________________ Destroy _________________________________*/
//bool TFont_Dlg__Destroy(TFont_Dlg *self) {
//	return TForm__Destroy((TForm*)self);
//}

/* ______________________ Create _____________________________________ */

void Init_FontAlg_Dlg(TFont_Dlg *self, const char *fontname) {
  int lg;
  const char *cpc;

	self->OnShow = Font_Dlg_OnShow;
	self->Enabled = False;

#ifndef _OS_WIN32
	self->Icon = GetDeskIcon(ddkt_Mime, 16, "font_truetype", 
				    gdifu_noZoom, zico16font_truetype);
#endif /* ! _OS_WIN32 */

	self->pnlTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder | mBorderL, alTop);
	self->pnlTop->OnResize = pnlTop_OnResize;
	self->pnlMiddle = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, 0, alClient);

// ******* Menu *******
	MenuCharset_Create(self);
#ifdef _OS_WIN32
	self->cobDPI = TComboBox_CreateS(self->pnlTop, 20, 1, 72, 19, "96 dpi");
	TComboBox_AddItem(self->cobDPI, "96 dpi", 0, NULL);
#else
	self->cobDPI = TComboBox_CreateS(self->pnlTop, 20, 1, 72, 19, "75 dpi");
	TComboBox_AddItem(self->cobDPI, "75 dpi", 0, NULL);
	TComboBox_AddItem(self->cobDPI, "100 dpi", 0, NULL);
#endif
	TComboBox_AddItem(self->cobDPI, "scale", 0, NULL);
	TComboBox_AddItem(self->cobDPI, "all", 0, NULL);
	self->cobDPI->ReadOnly = True;
	self->cobDPI->OnItemSel = cbDPI_OnItemSel;
	self->edTop = TEdit_CreateS(self->pnlTop, 180, 1, 72, 19, "-*-*-*-*-*-*-*-*-*-*-*-*-*-*");
	self->edTop->OnKeyUp = edTop_OnKeyUp;

	self->tv = TTreeVFont_CreateA(self->pnlMiddle, 0, 0, 180, 50, alLeft);
	self->tv->OnNodeClick = tv_OnNodeClick;
	self->pnlRight = TWins_CreateBA(self->pnlMiddle, 0, 0, 200, 21, mBorderL, alClient);
	TWins_SetBorderWidth(self->pnlRight, 2);
	self->pnlRight->Decor = TDecor_dup(&appXDat.DefaultWinsDecor);

	self->pnlRightTop = TWins_CreateBA(self->pnlRight, 0, 0, 200, 25, mBorder | mBorderW2, alTop);
	self->pnlRightTop->Decor = TDecor_dup(&appXDat.DefaultWinsDecor);

	self->le = TLineEdit_CreateA(self->pnlRight, 0, 0, 200, 200, alClient);
	self->le->vds.TabSpaceOn = False;
	TLineEdit_AddKeyDownPlugIn(self->le, TLineEdit_keyplugin_cursor_Create());
	self->le->OnSetCaret = le_OnSetCaret;
	self->pnlRightBottom = TWins_CreateBA(self->pnlRight, 0, 0, 200, 49, mBorder, alBottom);
	self->pnlRightBottom->OnResize = pnlRightBottom_OnResize;
	self->pnlRightBottom->Decor = TDecor_dup(&appXDat.DefaultWinsDecor);

	self->cobSIZE = TComboBox_CreateS(self->pnlRightTop, 4, 3, 62, 19, "*");
	self->cobSIZE->OnItemSel = cbSIZE_OnItemSel;
	self->cbUTF8 = TChkBox_CreateS(self->pnlRightTop, 68, 4, 54, 17, "UTF8");
	self->cbUTF8->OnClick = cbUTF8_OnClick;
	self->cbUTF8->Checked = self->le->vds.UTF8on;

	self->edResult = TEdit_CreateS(self->pnlRightBottom, 2, 2, 72, 19, "-*-*-*-*-*-*-*-*-*-*-*-*-*-*");
	self->CharAzBox = TWins_CreateLabel(self->pnlRightBottom, 4, 26, 88, 17, "255/FF", 0, mBorder | mBorderL);
	self->cbN =  TChkBox_CreateS(self->pnlRightBottom, 95, 26, 38, 17, "N ");
	self->cbNK = TChkBox_CreateS(self->pnlRightBottom,135, 26, 38, 17, "NK");
	self->cbB =  TChkBox_CreateS(self->pnlRightBottom,175, 26, 38, 17, "B ");
	self->cbBK = TChkBox_CreateS(self->pnlRightBottom,215, 26, 38, 17, "BK");
	self->cbN->OnClick = cbN_OnClick;
	self->cbNK->OnClick = cbN_OnClick;
	self->cbB->OnClick = cbN_OnClick;
	self->cbBK->OnClick = cbN_OnClick;

	self->buCancel = TButton_CreateS(self->pnlRightBottom, 2, 22, 72, 24, "CANCEL");
	self->buCancel->OnClick = buCancel_OnClick;
	self->buOK = TButton_CreateS(self->pnlRightBottom, 82, 22, 72, 24, "OK");
	self->buOK->OnClick = buOK_OnClick;


	if (!fontname) fontname = self->le->vds.Font->name;

	if (GetFontSXTypeIdx(8, fontname, &cpc, &lg)) {
	    if (lg == 1) {
		if (*cpc == '*') {
		    SetCaption((TWins*)self->cobDPI, "all");
		} else if (*cpc == '0') {
		    SetCaption((TWins*)self->cobDPI, "scale");
		}
	    }
	}

	SetCaption((TWins*)self->edTop, fontname);
}

TFont_Dlg *Create_Font_Dlg(TWins *parent, const char *fontname) {
  TFont_Dlg *self;

	self = TForm_CreatParamAlg(parent, sizeof(TFont_Dlg));

//	self->destroy = (TDESTROY)TFont_Dlg__Destroy;
	self->Visible = False;
	if (parent) self->FormZStyle = fszTransient;
	self->FormPosition = poCenter;

	SetCaption((TWins*)self, "Select Font");
	self->IconName = "AHfont";
	self->height = 350;
	self->width = 640;
	self->min_height = 160;
	self->min_width = 320;

	Init_FontAlg_Dlg(self, fontname);
	return self;
}
