/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "tfilewin.h"
#include "tpixmap.h"

typedef struct TFileIcons {
  AlgTTC0
  int size;
  TGetFilePic_rszPara *rszPara;
  TPic pic;
  struct TFileIcons *next;
  bool picOK;
} TFileIcons;

typedef struct TGetFilePic_struct {
  AlgTTC0
  TFileWin *fw;
  TFileIcons *fico;
  TEv_Array *appEv;
  int idx;
  int PaintMk;
} TGetFilePic_struct;

static TGetFilePic_rszPara rszPara;


static bool TFileIcons__Destroy(TFileIcons *self) {
  Display *dpy = appXDat.dpy;
	if (dpy) {
	    if (self->pic.pic) XFreePixmap(dpy, self->pic.pic);
	    if (self->pic.mask) XFreePixmap(dpy, self->pic.mask);
	}
	return False;
}


static void TFileWin_On_appEv(TEv_Array *appEv) {  
  TGetFilePic_struct *gfp = appEv->obj;
  TDirFileList *dfl;
  TFileIcons *fi;
  TglobPath gp;
  TImage *img;
  TPixMap *pxm;
  TGetFilePic_rszPara *rszP;
  FILE *F;
  int new_w, new_h;
  int Time_us = GetTime_us();

	gfp->PaintMk = 0;
	gp_dup_gpdata_path(&gp, TFileWin_Get_gp(gfp->fw));
NextIcoLoop:
	dfl = TFileWin_Get_dfl(gfp->fw, gfp->idx);
	if (!dfl) {
	    gfp->appEv = NULL;
	    Free_appEv(appEv);
	    goto ende;
	}
	fi = dfl->obj;
	if (fi && (fi != (void*)-1) && !fi->picOK) {
	    rszP = fi->rszPara;
	    gp.name_const = dfl->fn;

	    F = TImageIO_GetStream(NULL, &gp, pxmLD_OnlyInfo | pxmLD_NoERR, NULL, False);
	    img = TImage_Loader_F(NULL, &gp, pxmLD_OnlyInfo | pxmLD_NoERR, F);
	    if (img) {

		int pxmLD = 0;
		if (img->in_filetyp == filetyp_JPEG) {   //JPEG 1:2 1:4 1:8
		    new_w = img->width >> 2; new_h = img->height >> 2;
		    if ((new_w > rszP->max_w)||(new_h > rszP->max_h)) pxmLD = imgLD_scale_2;
		    new_w >>= 1; new_h >>= 1;
		    if ((new_w > rszP->max_w)||(new_h > rszP->max_h)) pxmLD = imgLD_scale_4;
		    new_w >>= 1; new_h >>= 1;
		    if ((new_w > rszP->max_w)||(new_h > rszP->max_h)) pxmLD = imgLD_scale_8;
	TImage *img2 = TImage_Loader_F(NULL, &gp, pxmLD_OnlyInfo | pxmLD_NoERR | pxmLD, F);
		    if (img2) {
			Tfree(img);
			img = img2;
		    }
		}

		if ((img->width <= rszP->max_ico_size) && 
		    (img->height <= rszP->max_ico_size) &&
		    (img->width >= 4) &&
		    (img->height >= 4)) {

		    pxm = TPixMap_Loader_F(NULL, &gp, pxmLD_NoERR | pxmLD, F);
		    if (pxm) {
			if ((pxm->width > rszP->max_w) || (pxm->height > rszP->max_h)) {
			    new_w = rszP->max_w;
			    new_h = (pxm->height * new_w) / pxm->width;
			    if (new_h > rszP->max_h) {
			      new_h = rszP->max_h;
			      new_w = (pxm->width * new_h) / pxm->height;
			    }
			    TPixMap_resize(&pxm, new_w, new_h);
			}

			if ((pxm->width >= 4) &&
			    (pxm->width <= rszP->max_w) &&
			    (pxm->height >= 4) &&
			    (pxm->height <= rszP->max_h)) {

			    TPixMap_XPutImage(pxm);
			    fi->picOK = True;
			    fi->pic.pic = pxm->pmxID;
			    fi->pic.mask = pxm->maskID;
			    fi->pic.width = pxm->width;
			    fi->pic.height = pxm->height;
			    pxm->pmxID = 0;
			    pxm->maskID = 0;
			    gfp->PaintMk++;
			}
			Tfree(pxm);
		    }
		}
		Tfree(img);
	    }
	    if (F) fclose(F);
	    if (!fi->picOK) dfl->obj = (void*)-1;
	}
	gfp->idx++;

	if (GetCurTimeDiv_us(Time_us) < 10000) goto NextIcoLoop;
ende:
	gp.name = NULL;
	gp_pc_free(&gp);

	if (gfp->PaintMk) {
	    gfp->PaintMk = 0;
	    Refresh((TWins*)gfp->fw);
	}
}


TPic *TFileWin_GetFilePic(TFileWin *self, TDirFileList *dfl, int size, int bef) {
  TPic *pic = NULL;
  TGetFilePic_struct *gfp = self->GetFilePic_struct;
  TFileIcons *fi, *pfi, **ppfi;

	if ((bef == fwgfp_bef_ClrAll) || (bef == fwgfp_bef_Free)) {
	    if (gfp) {
		Free_appEv(gfp->appEv);
		while (gfp->fico) {
		    fi = gfp->fico;
		    gfp->fico = fi->next;
		    Tfree(fi);
		}
		xfree(gfp);
		self->GetFilePic_struct = NULL;
	    }
	    rszPara.size = -1;

  TDirFileList *ndfl = TFileWin_Get_dfl0(self);
	    while (ndfl) {
		ndfl->obj = NULL;
		ndfl = ndfl->next;
	    }

	} else if (bef == fwgfp_bef_ClrPic) {
	    if (dfl->obj) {
		if (dfl->obj != (void*)-1) {
		    fi = (TFileIcons*)dfl->obj;
		    pfi = gfp->fico;
		    ppfi = &gfp->fico;
		    while (pfi && (pfi != fi)) {
			ppfi = &pfi->next;
			pfi = pfi->next;
		    }
		    if (pfi) {
			*ppfi = pfi->next;
			Tfree(pfi);
		    } else {
			printf("ERROR: %s(%d): pfi == NULL  ???\n", FL);
		    }
		}
		dfl->obj = NULL;
	    }
	} else if (bef == fwgfp_bef_GetPic) {
	    if ((dfl->st->FileType == stft_IMAGE) && (dfl->obj != (void*)-1)) {

		if (dfl->obj) {
		    fi = (TFileIcons*)dfl->obj;
		    if (fi->picOK) pic = &fi->pic;
		} else {
		    if (!gfp) {
			gfp = xcalloc(sizeof(TGetFilePic_struct));
			self->GetFilePic_struct = gfp;
			gfp->wtyp = wtyp_TGetFilePic;
			gfp->fw = self;
		    }
		    if (!gfp->appEv) {
			gfp->appEv = Add_appEv(TFileWin_On_appEv);
			gfp->appEv->obj = gfp;
			gfp->idx = 0;
		    }
		    fi = ttc_create(sizeof(TFileIcons), wtyp_TFileIcons);
		    fi->destroy = (TDESTROY)TFileIcons__Destroy;
		    fi->next = gfp->fico;
		    gfp->fico = fi;
		    dfl->obj = fi;
		    fi->size = size;
		    if (self->rszPara) {
			fi->rszPara = self->rszPara;
		    } else {
			if (rszPara.size != size) {
			  rszPara.size = size;
//			  rszPara.min_w = size;
			  rszPara.max_w = size + 2;
//			  rszPara.min_h = size >> 1;
			  rszPara.max_h = size;
			  rszPara.max_ico_size = 128;
//			  rszPara.max_ico_XPM_fsize = 32000;
//			  rszPara.max_ico_PNG_fsize = 32000;
//			  rszPara.max_ico_JPEG_fsize = 32000;
			}
			fi->rszPara = &rszPara;
		    }
		}
	    }
	    if (!pic) {
		if (size < 30) {
		    if (app_mime && app_mime->stat_mime_getpic16) {
			pic = app_mime->stat_mime_getpic16(app_mime, dfl->st);
		    }
		    if (!pic) pic = get_fileIcon(dfl->st->FileType);
		} else {
		    if (app_mime && app_mime->stat_mime_getpic32) {
			pic = app_mime->stat_mime_getpic32(app_mime, dfl->st);
		    }
		    if (!pic) pic = get_fileIcon32(dfl->st->FileType);
		}
	    }
	}
	return pic;
}

