/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/


#include "appXbase.h"
#include "highl.h"


static bool Thighlight__Destroy(Thighlight *self) {
  int i;

	TFont_Destroy(self->kursiv_Font);
	TFont_Destroy(self->bold_Font);
	TFont_Destroy(self->boldkursiv_Font);
	for (i=0; i<self->tls->Count; i++) {
	    ppcfree(&self->tls->lines->cs[i]->c);
	}
	xfree(self->fc0_Dat);
	xfree(self->fc1_Dat);
	xfree(self->fc2_Dat);
	xfree(self->fc3_Dat);
	return False;
}

static TFont *get_font_art(Thighlight *self, int font_art) {
	switch (font_art & FONT_SMask) {
	case FONT_MediumK: return self->kursiv_Font;
	case FONT_Bold:    return self->bold_Font;
	case FONT_BoldK:   return self->boldkursiv_Font;
	}
	return self->Font;
}

static void fcd_SetFont(Thighlight *self, Thl_keyw_arr *fcd) {
	if (fcd) fcd->fc.Font = get_font_art(self, fcd->font_art);
}

static void hlSetFont(Thighlight *self) {

	self->Font = self->wins->vds.Font;

	MakeFontArt(self->Font->name, &self->kursiv_Font, FONT_MediumK);
	MakeFontArt(self->Font->name, &self->bold_Font, FONT_Bold);
	MakeFontArt(self->Font->name, &self->boldkursiv_Font, FONT_BoldK);
	self->fc_comment.Font = self->kursiv_Font;
	self->fc_char.Font = self->Font;
	self->fc_str.Font = self->Font;
	self->fc_nr.Font = self->Font;
	self->fc_def.Font = self->Font;

	fcd_SetFont(self, self->fc0_Dat);
	fcd_SetFont(self, self->fc1_Dat);
	fcd_SetFont(self, self->fc2_Dat);
	fcd_SetFont(self, self->fc3_Dat);
}

static bool Thighlight_ctl(Thighlight *self, TWinsCtl *wctl) {
	if (wctl->bef == wctlb_SetFont) {
	    hlSetFont(self);
	    return True;
	}
	return False;
}

Thighlight *Thighlight_Create(TWins *wins, TLines *tls) {
  Thighlight *self;

	if (!wins || !tls) return NULL;
	self = ttc_create(sizeof(Thighlight), wtyp_highlight);
	self->wins = wins;
	self->tls = tls;
	self->destroy = (TDESTROY)Thighlight__Destroy;
	self->hlCtl = Thighlight_ctl;

	hlSetFont(self);

	self->fc_comment.ForeColor = 0x282860;
	self->fc_comment.BackColor = self->Font->g.BackColor;
	self->fc_char.ForeColor = 0xe04040;
	self->fc_char.BackColor = self->Font->g.BackColor;
	self->fc_str.ForeColor = 0xe00000;
	self->fc_str.BackColor = self->Font->g.BackColor;
	self->fc_nr.ForeColor = 0x0000e0;
	self->fc_nr.BackColor = self->Font->g.BackColor;
	self->fc_def.ForeColor = 0x00a000;
	self->fc_def.BackColor = self->Font->g.BackColor;

	self->OldCaretX = -1;
	self->OldCaretY = -1;
	return self;
}

void hl_Set_keyword(Thighlight *self, Thl_keyw_arr **ppkarr, const char *ds, int font_art, int ForeColor, int BackColor) {
  const char *s, *sa;
  char *pc = NULL;
  int i = 0, count = 0, idx = 0;
  Thl_keyw_arr *karr;

	xfree(*ppkarr);
	*ppkarr = NULL;
	if (!ds) return;
	s = ds;
	while (*s) {
	    i++; s++;
	    if ((*s == 0x20) || !*s) {
		if (i > 1) count++;
		i = 0;
		if (*s) s++;
	    }
	}

	i = sizeof(ThlFCol) + sizeof(int) + sizeof(Thl_keyw) * count;
	karr = xcalloc(i);
	karr->font_art = font_art;
	fcd_SetFont(self, karr);
	karr->fc.ForeColor = ForeColor;
	karr->fc.BackColor = BackColor;
	karr->items = count;

	i = 0;
	s = ds;
	while (*s == 0x20) s++;
	sa = s;
	while (*s) {
	    i++; s++;
	    if ((*s == 0x20) || !*s) {
		if (i > 1) {
		    if (i >= (hl_keyw_MAX - 1)) i = hl_keyw_MAX - 1;
		    karr->keyw[idx].lg = i;
		    pc = (char*)karr->keyw[idx].sc;
		    while (i) {*pc = *sa; pc++; sa++; i--;}
		    idx++;
		    *pc = 0;
		}
		while (*s == 0x20) s++;
		i = 0;
		sa = s;
	    }
	}
	*ppkarr = karr;
}

void add_scd(TColS *cs, char *pc, int lg, ThlFCol *fc, int fu) {
  int nr;
  TStrColorDatArr *scda;
  TStrColorDat *scd;
	if (cs->c) nr = ((TStrColorDatArr*)cs->c)->items; else nr = 0;
	scda = set_scda_arr((void*)&cs->c, nr+1);
	scd = &scda->scd[nr];
	scd->s = pc;
	scd->lg = lg;
	scd->Font = fc->Font;
	scd->ForeColor = fc->ForeColor;
	scd->BackColor = fc->BackColor;
	scd->fu = fu;
}

void add_scd_char(Thldat *hld, char *pc, int lg) {
  add_scd(hld->cs, pc, lg, &hld->hl->fc_char, 0);
}

void add_scd_def(Thldat *hid, Tstoken *t) {
  unsigned char *pc;
  unsigned char c = 0;
	hid->mk &= ~hl_defline;
	pc = t->next;
	if (!pc) return;
	while (*pc) {c = *pc; pc++;}
	t->lg = pc - t->s;
	t->next = pc;
	if (c == 0x5c) hid->mk |= hl_defline;
	add_scd(hid->cs, (char*)t->s, t->lg, &hid->hl->fc_def, 0);
}

int scd_str_len(byte *pc0, byte cc, int fu) {
  byte *pc;
	if (!pc0) return 0;
	pc = (byte*)pc0;
	if (fu & scd_Num_cc) {
	    while (*pc && (*pc > 0x20) && ((*pc < '0') || (*pc > '9'))) {
		if ((*pc == 0x5c) && pc[1]) {
		    pc += 2;
		} else pc++;
	    }
	} else if (cc == ' ') {
	    while (*pc && (*pc > 0x20)) {
		if ((*pc == 0x5c) && pc[1]) {
		    pc += 2;
		} else pc++;
	    }
	} else if (fu & scd_SkipBlank) {
	    while (*pc && (*pc != cc)) {
		if ((*pc == 0x5c) && pc[1]) {
		    pc += 2;
		} else pc++;
	    }
	} else {
	    while (*pc && (*pc > 0x20) && (*pc != cc)) {
		if ((*pc == 0x5c) && pc[1]) {
		    pc += 2;
		} else pc++;
	    }
	}
	if ((*pc != cc) && (fu & scd_Only_cc)) return 0;
	return pc - pc0;
}

void add_scd_str_fc(Thldat *hid, Tstoken *t, int mk, char cc, ThlFCol *fc) {
  byte *pc;
  byte c = 0;
	hid->mk &= ~mk;
	pc = t->next;
	if (!pc) return;
	if (cc == ' ') {
	    while (*pc && ((byte)*pc > 0x20)) {
		if ((*pc == 0x5c) && pc[1]) {
		    pc += 2; c = 0;
		} else {c = *pc; pc++;}
	    }
	} else {
	    while (*pc && (*pc != cc)) {
		if ((*pc == 0x5c) && pc[1]) {
		    pc += 2; c = 0;
		} else {c = *pc; pc++;}
	    }
	}
	if (*pc == cc) pc++;
	t->lg = pc - t->s;
	t->next = pc;
	if (c == 0x5c) hid->mk |= mk;
	add_scd(hid->cs, (char*)t->s, t->lg, fc, 0);
}

void add_scd_str(Thldat *hid, Tstoken *t) {
  byte *pc;
  byte c = 0;
	hid->mk &= ~hl_strline;
	pc = t->next;
	if (!pc) return;
	while (*pc && (*pc != 0x22)) {
	    if ((*pc == 0x5c) && pc[1]) {
		pc += 2; c = 0;
	    } else {c = *pc; pc++;}
	}
	if (*pc == 0x22) pc++;
	t->lg = pc - t->s;
	t->next = pc;
	if (c == 0x5c) hid->mk |= hl_strline;
	add_scd(hid->cs, (char*)t->s, t->lg, &hid->hl->fc_str, 0);
}

void add_scd_nr(Thldat *hld, char *pc, int lg) {
  add_scd(hld->cs, pc, lg, &hld->hl->fc_nr, 0);
}

void add_scd_comment(Thldat *hld, char *pc, int lg) {
  add_scd(hld->cs, pc, lg, &hld->hl->fc_comment, 0);
}

static bool keyword_scan(byte *s, int lg, Thl_keyw_arr *karr) {
  int count;
  byte *pc, *sa;
  Thl_keyw *keyw;

	if (!karr) return False;
	count = karr->items;
	for (int idx=0;idx<count;idx++) {
	    keyw = &karr->keyw[idx];
	    if (keyw->lg == lg) {
		sa = keyw->sc;
		int i = lg;
		pc = s;
		while ((i > 0) && (*sa == *pc)) {sa++; pc++; i--;}
		if (!i) return True;
	    }
	}
	return False;
}

bool hl_keyword_scan(byte *s, int lg, Thl_keyw_arr *karr) {
	if (!s || (lg <= 0) || !karr) return False;
	return keyword_scan(s, lg, karr);
}

static bool keyword_scan_case(byte *s, int lg, Thl_keyw_arr *karr) {
  int count;
  byte *pc, *sa;
  Thl_keyw *keyw;

	if (!karr) return False;
	count = karr->items;
	for (int idx=0;idx<count;idx++) {
	    keyw = &karr->keyw[idx];
	    if (keyw->lg == lg) {
		sa = keyw->sc;
		int i = lg;
		pc = s;
		while ((i > 0)&&(l_ascii[*sa] == l_ascii[*pc])) {sa++; pc++; i--;}
		if (!i) return True;
	    }
	}
	return False;
}

bool hl_keyword_scan_case(byte *s, int lg, Thl_keyw_arr *karr) {
	if (!s || (lg <= 0) || !karr) return False;
	return keyword_scan_case(s, lg, karr);
}

bool hl_keyword_c0(Thldat *hld, Thl_keyw_arr *karr) {
  int i;
  int lg = hld->t->lg, idx = 0, count;
  byte *pc, *sa;
  Thl_keyw *keyw;

	if (!karr) return False;
	count = karr->items;
	if (hld->hl->CharCase) {
	    while (count) {
		keyw = &karr->keyw[idx];
		if (keyw->lg == lg) {
		    sa = &keyw->sc[0];
		    i = lg;
		    pc = hld->t->s;
		    while (i > 0) {
			if (l_ascii[*sa] != l_ascii[*pc]) break;
			sa++; pc++; i--;
		    }
		    if (!i) {
			add_scd(hld->cs, (char*)hld->t->s, lg, &karr->fc, 0);
			return True;
		    }
		}
		idx++; count--;
	    }

	} else {
	    if (keyword_scan(hld->t->s, lg, karr)) {
		add_scd(hld->cs, (char*)hld->t->s, lg, &karr->fc, 0);
		return True;
	    }
/*	    while (count) {
		keyw = &karr->keyw[idx];
		if (keyw->lg == lg) {
		    sa = &keyw->sc[0];
		    i = lg;
		    pc = hld->t->s;
		    while (i > 0) {
			if (*sa != *pc) break;
			sa++; pc++; i--;
		    }
		    if (!i) {
			add_scd(hld->cs, (char*)hld->t->s, lg, &karr->fc, 0);
			return True;
		    }
		}
		idx++; count--;
	    } */
	}
	return False;
}
