/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "tlineedit.h"

static void TLineEdit_Set_vds(TLineEdit *le, int idy) {
  TColS *cs = TLines_Get_pcs(le->tls, idy);
	if (cs) {
	    le->vds.s = cs->s;
	    le->vds.lg = cs->lg;
	    le->vds.scda = (TStrColorDatArr*)cs->c;
	} else {
	    le->vds.s = NULL;
	    le->vds.lg = 0;
	    le->vds.scda = NULL;
	}
}

static bool TLineEdit_keyplugin_cursor(TLineEdit *le, TKeyEvent *KeyEv, int *res) {
  int c, m;
  TLines *tls = le->tls;
  TVDrawString *vds = &le->vds;

	le->OldX = False;
	if (KeyEv->keysym == XK_Left) {
	    if (le->CaretCharIdx <= 0) {
		if (le->CaretCharIdy > 0) {
		    le->CaretCharIdy--;
		    TLineEdit_Set_vds(le, le->CaretCharIdy);
		    le->CaretCharIdx = vds->lg;
		}
	    } else {
		if ((KeyEv->state & ControlMask) == 0) {
		    GetVdsPrev_WChar(&le->CaretCharIdx, vds);
		} else {
		    c = GetTzC(le->TzC, GetVdsPrev_WChar(&le->CaretCharIdx, vds));
		    m = le->CaretCharIdx;
		    while (le->CaretCharIdx > 0) {
			if (c != GetTzC(le->TzC, GetVdsPrev_WChar(&m, vds))) break;
			le->CaretCharIdx = m;
		    }
		}
	    }
	} else if (KeyEv->keysym == XK_Right) {
	    if (le->CaretCharIdx >= vds->lg) {
		if (le->CaretCharIdy < tls->Count) {
		    le->CaretCharIdy++;
		    TLineEdit_Set_vds(le, le->CaretCharIdy);
		    le->CaretCharIdx = 0;
		}
	    } else {
		if ((KeyEv->state & ControlMask) == 0) {
		    GetVdsNext_WChar(&le->CaretCharIdx, vds);
		} else {
		    c = GetTzC(le->TzC, GetVds_WChar(le->CaretCharIdx, vds));
		    while (le->CaretCharIdx < vds->lg) {
			if (c != GetTzC(le->TzC, GetVdsNext_WChar(&le->CaretCharIdx, vds))) break;
		    }
		}
	    }
	} else if (KeyEv->keysym == XK_Up) {
	    if (le->CaretCharIdy > 0) {
		le->CaretCharIdy--;
		TLineEdit_Set_vds(le, le->CaretCharIdy);
		le->CaretCharIdx = GetMouseX_CharPos(le->OldCaretX, vds);
		le->OldX = True;
	    }
	} else if (KeyEv->keysym == XK_Down) {
	    if (le->CaretCharIdy < tls->Count) {
		le->CaretCharIdy++;
		TLineEdit_Set_vds(le, le->CaretCharIdy);
		le->CaretCharIdx = GetMouseX_CharPos(le->OldCaretX, vds);
		le->OldX = True;
	    }
	} else if (KeyEv->keysym == XK_Page_Up) {
	    le->CaretCharIdy -= le->VisibleLines - 1;
	    if (le->CaretCharIdy < 0) le->CaretCharIdy = 0;
	    le->OldX = True;
	    TScrollBar_SetCount_ev(le->VSB, le->VSB->Count - (le->VisibleLines - 1));

	} else if (KeyEv->keysym == XK_Page_Down) {
	    le->CaretCharIdy += le->VisibleLines - 1;
	    if (le->CaretCharIdy > tls->Count) le->CaretCharIdy = tls->Count;
	    le->OldX = True;
	    TScrollBar_SetCount_ev(le->VSB, le->VSB->Count + (le->VisibleLines - 1));

	} else if (KeyEv->keysym == XK_Home) {
		le->CaretCharIdx = 0;

	} else if (KeyEv->keysym == XK_End) {
		le->CaretCharIdx = vds->lg;

	} else return False;
	TLineEdit_ChkCaret(le, 0);
	if ((KeyEv->state & ShiftMask) == 0) {
	    if ((le->SelAx != le->CaretCharIdx) || (le->SelAy != le->CaretCharIdy))
		    Refresh((TWins*)le);
	    le->SelAx = le->CaretCharIdx;
	    le->SelAy = le->CaretCharIdy;
	} else Refresh((TWins*)le);
	TLineEdit_SetCaret(le);
	*res = 1;
	return True;
}


TPlugIn_KeyDown *TLineEdit_keyplugin_cursor_Create() {
  TPlugIn_KeyDown *plug;
	plug = ttc_create(sizeof(TPlugIn_KeyDown), wtyp_LineEditKeyPlugin);
	plug->KeyDownPlugIn = (TKeyDownPlugIn)TLineEdit_keyplugin_cursor;
	return plug;
}
