/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "tfilewin.h"
#include "dlg.h"
#include <sys/stat.h>

#define SHOW_PIC_fw 1

typedef struct {
  TDirListObjNy *ObjNy;
  TDirFileList *dfl_last;
  int dfl_nr;
  TDirFileList *dfl_foc;
  bool ShowHidden;
} T_private;

//*********************** TFileWin ************************

typedef struct {
  int art;
  int fu;
  int Width;
  const char *name;
} TcFDL_ColD;

static const TcFDL_ColD FDL_DefColD_arr[] = {
  {1, 1, 140, "Name"},
  {2, 1,  64, "Size"},
  {3, 1,  90, "File Type"},
  {4, 1, 104, "Modified"},
  {5, 0, 104, "Accessed"},
  {6, 1,  68, "Permission"},
#ifdef _OS_WIN32
  {7, 1,  90, "User"},
  {8, 1,  90, "Group"},
#else
  {7, 1,  50, "User"},
  {8, 1,  50, "Group"},
#endif
  {9, 1, 400, "Link"}
};


static void TFileWin_InitDefaultCols(TFileWin *fw) {
	for (int i=0; i<9; i++) {
	    fw->FDL_ColD_arr[i].art   = FDL_DefColD_arr[i].art;
	    fw->FDL_ColD_arr[i].fu    = FDL_DefColD_arr[i].fu;
	    fw->FDL_ColD_arr[i].Width = FDL_DefColD_arr[i].Width;
	    ppcchange(&fw->FDL_ColD_arr[i].name, FDL_DefColD_arr[i].name);
	}
}

bool TFileWin_GetColsData(TFileWin *self, int idx, TFDL_ColD *ColD) {
	if (!self || !ColD || (idx < 0) || (idx > 8)) return False;
	*ColD = self->FDL_ColD_arr[idx];
	return True;
}

bool TFileWin_SetColsData(TFileWin *self, int idx, TFDL_ColD *ColD) {
	if (!self || !ColD || (idx < 0) || (idx > 8)) return False;
	self->FDL_ColD_arr[idx].fu    = ColD->fu;
	self->FDL_ColD_arr[idx].Width = ColD->Width;
	ppcchange(&self->FDL_ColD_arr[idx].name, ColD->name);
	return True;
}

void TFileWin_Resize(TFileWin *self) {
	TGrid_Resize((TGrid*)self);
	TFileWin_SetColsArt(self, self->AzArt);
}

static TPic *TFileWin_FilePic(TFileWin *fw, TDirFileList *dfl, int size) {
	if (fw->GetFilePic) {
	    return fw->GetFilePic(fw, dfl, size, fwgfp_bef_GetPic);
	} else {
	    return get_fileIcon(dfl->st->FileType);
	}
}

static int PaintCell0_16(TFileWin *self, TDirFileList *dfl, TWRect *rc, bool FocusRect) {
  int result;
  TPic *pic;
  TClassGC *CellGC;

	CellGC = self->CellGC;
	CellGC->vds.s_const = dfl->fn;
	CellGC->vds.lg = pclen(dfl->fn);
	result = TClassGC_DrawColorStringFu(CellGC, 20, 0, mStrLeft);
	CellGC->vds.s = NULL;
	CellGC->vds.lg = 0;
	if (dfl->fu & dfl_Select) {
	    TClassGC_FillSelectRect_rc(CellGC, rc);
	}
	pic = TFileWin_FilePic(self, dfl, 16);
	TClassGC_DrawPic(CellGC, pic, 1,1);
	if (dfl->st->st & st_Link) {
	  pic = get_fileIcon(stft_Link);
	  TClassGC_DrawPic(CellGC, pic, 1,11);
	}
	if (FocusRect && (self->KeyFu & mHasFocus))
		TClassGC_DrawFocusRect_rc(CellGC, rc);


	return result;
}

bool TFileWin_DoPaintCell(TFileWin *self, int ACol, int ARow, TWRect *rc, int *fu) {
  TDirFileList *dfl;
  TVDrawString *vds;
  TClassGC *CellGC;
  int idx, AzArt;
  TcrDat_Row *cr;
  TcrDat *cc;
  char buf[32];
  bool DrawStr = True;
  bool DrawStrFree = True;

	AzArt = self->AzArt & 0xff;
	cr = &self->RowDat[ARow];
	if (((AzArt == 0) && (ARow > 0)) || AzArt) {
	    cr->Color = 0x80000000;
	    if (!(ARow & 0x01)) cr->Color = 0xf0f0ff;
	}
  
	if (!TGrid_DoPaintCell((TGrid*)self, ACol, ARow, rc, fu)) return False;

	CellGC = self->CellGC;
	vds = &CellGC->vds;
	cc = &self->ColDat[ACol];
	if (AzArt == 0) {
	    int cst_x = 2;
	    int cst_fu = mStrLeft;
	    if (ARow == 0) {
		vds->s_const = cc->name;
		if (cc->tag == self->Sort_Col) {
		    int y = 3;
		    int x = rc->left + rc->width - 13;
		    if (self->Sort_Col == 9) x = rc->left + 40;

		    TClassGC_SetForeGround(CellGC, 0xffffff);
		    if (self->sortFu & dlsortFu_DOWN) {
			TClassGC_DrawLine(CellGC, x, y+10, x+4, y);
			TClassGC_SetForeGround(CellGC, 0x404040);
			TClassGC_DrawLine(CellGC, x, y+10, x+8, y+10);
			TClassGC_DrawLine(CellGC, x+8, y+10, x+4, y);
		    } else {
			TClassGC_DrawLine(CellGC, x, y, x+8, y);
			TClassGC_DrawLine(CellGC, x, y, x+4, y+10);
			TClassGC_SetForeGround(CellGC, 0x404040);
			TClassGC_DrawLine(CellGC, x+8, y, x+4, y+10);
		    }
		}
		DrawStrFree = False;

	    } else {

		dfl = TFileWin_Get_dfl(self, ARow-1);
		if (!dfl) return False;
		Tstat64 *st = dfl->st;

		switch(cc->tag) {
		case 1:  //Name
		    PaintCell0_16(self, dfl, rc, ARow == (self->SelectIdxRow+1));
		    DrawStrFree = False;
		    DrawStr = False;
		    break;
		case 2:  //Size
		    file_str_size_pc(buf, st->size, fstrSize_space);
		    vds->s = buf;
		    DrawStrFree = False;
		    cst_fu = mStrRigth;
		    cst_x = -1;
		    break;
		case 3:  //File Type
		    vds->s_const = GetArtExt_pc(st);
		    DrawStrFree = False;
		    break;
		case 4:  //Modified
		    file_str_time_HM_pc(buf, st->mtime);
		    vds->s = buf;
		    DrawStrFree = False;
		    break;
		case 5:  //Accessed
		    file_str_time_HM_pc(buf, st->atime);
		    vds->s = buf;
		    DrawStrFree = False;
		    break;
		case 6:  //Permission
		    file_str_mode_pc(buf, st, StrModT_PlusACL);
		    vds->s = buf;
		    DrawStrFree = False;
		    break;
		case 7:  //User
		    vds->s_const = Get_ExFileDat_pc(st, exfdt_owner);
		    DrawStrFree = False;
		    break;
		case 8:  //Group
		    vds->s_const = Get_ExFileDat_pc(st, exfdt_group);
		    DrawStrFree = False;
		    break;
		case 9:  //Link
		    if (st->st & st_Link) {
			vds->s_const = Get_ExFileDat_pc(st, exfdt_LinkName);
			DrawStrFree = False;
		    } else DrawStr = False;
		    break;
		}
	    }

	    if (DrawStr) {
	        vds->lg = pclen(vds->s);
		TClassGC_DrawColorStringFu(CellGC, cst_x, 0, cst_fu);
	    }

#ifdef SHOW_PIC_fw
	} else if (AzArt == 2) {

	    idx = ACol + self->ColCount * ARow;
	    dfl = TFileWin_Get_dfl(self, idx);
	    if (!dfl) return False;

	    vds_SetStr(vds, dfl->fn);
	    TClassGC_DrawColorStringFu(CellGC, 0, -2, mStrBottom);

	    if (dfl->fu & dfl_Select) TClassGC_FillSelectRect_rc(CellGC, rc);
	    TPic *pic = TFileWin_FilePic(self, dfl, 32);
	    if (pic) {
		int m = pic->height;
		int xh = (cc->w - pic->width) / 2;
		int y = ((cr->w - (CellGC->vds.Font->height + 2)) - m) / 2;
		TClassGC_DrawPic(CellGC, pic, xh, y+2);
		if (dfl->st->st & st_Link) {
		    pic = get_fileIcon32(stft_Link);
		    TClassGC_DrawPic(CellGC, pic, xh,y+m-5);
		}
	    }
	    if ((idx == self->SelectIdxRow) && (self->KeyFu & mHasFocus))
			 TClassGC_DrawFocusRect_rc(CellGC, rc);
#endif
	} else if (AzArt == 1) {

	    idx = ARow + self->RowCount * ACol;
	    dfl = TFileWin_Get_dfl(self, idx);
	    if (!dfl) return False;

	    int wx = 2 + PaintCell0_16(self, dfl, rc, idx == self->SelectIdxRow) -
		     CellGC->XOrigin;
	    if (wx > 240) wx = 240;
	    if (wx > cc->w) {
		cc->w = wx;
		TGrid_Resize((TGrid*)self);
		Refresh((TWins*)self);
	    }
	    DrawStrFree = False;
	}

	if (DrawStrFree) xfree(vds->s);
	vds->s = NULL;
	vds->lg = 0;
	return True;
}

static void TFileWin_DoPaint(TFileWin *self) {
  T_private *pv = self->private;

	pv->dfl_nr = 0;
	pv->dfl_last = TFileWin_Get_dfl0(self);
	TGrid_DoPaint((TGrid*)self);
	pv->dfl_nr = 0;
	pv->dfl_last = NULL;
}

static int get_max_x(TFileWin *self, TcrDat *cd, int end) {
  int x;
  int res = 0;
	for (x=0; x<end; x++) {
	    if (cd[x].Visible) res += cd[x].w;
	}
	return res;
}

static int sb_get_seg(TFileWin *self, TcrDat *cd, int off, int end, bool Neg) {
  int x;

	for (x=0; x<end; x++) {
	    if (cd[x].Visible) {
		if (Neg) {
		    off -= cd[x].w;
		    if (off <= 0) return x;
		} else {
		    if (off < 0) return x;
		    off -= cd[x].w;
		}
	    }
	}
	return end-1;
}

static bool HSB_OnSmallChange(TScrollBar *self, int *NewCount, bool Neg) {
  TFileWin *fw = (TFileWin*)self->owner;
  int AzArt, seg;

	AzArt = fw->AzArt & 0xff;
	if (AzArt != 1) return False;
	seg = sb_get_seg(fw, fw->ColDat, self->Count, fw->ColCount, Neg);
	*NewCount = get_max_x(fw, fw->ColDat, seg);
	return True;
}
static bool HSB_OnLargeChange(TScrollBar *self, int *NewCount, bool Neg) {
  TFileWin *fw = (TFileWin*)self->owner;
  int AzArt, seg;

	AzArt = fw->AzArt & 0xff;
	if (AzArt != 1) return False;

	seg = sb_get_seg(fw, fw->ColDat, self->Count, fw->ColCount, Neg);
	if (Neg) {
	    if (seg > 0) seg -= 1;
	} else {
	    if (seg < fw->ColCount) seg += 1;
	}
	*NewCount = get_max_x(fw, fw->ColDat, seg);
	return True;
}

static bool VSB_OnSmallChange(TScrollBar *self, int *NewCount, bool Neg) {
  TFileWin *fw = (TFileWin*)self->owner;
  int seg;

	seg = sb_get_seg(fw, fw->RowDat, self->Count, fw->RowCount, Neg);
	*NewCount = get_max_x(fw, fw->RowDat, seg);
	return True;
}

static void TFileWin_Idx2ColRow(TFileWin *self, int idx, int *ACol, int *ARow) {
  int AzArt = self->AzArt & 0xff;
  int Col, Row;

	Col = Row = 0;
	if ((self->RowCount > 0) && (self->ColCount > 0)) {
	    if (AzArt == 0) {
		Col = 0;
		Row = idx+1;
#ifdef SHOW_PIC_fw
	    } else if (AzArt == 2) {
		Row = idx / self->ColCount;
		Col = idx - (Row * self->ColCount);
#endif
	    } else {
		Col = idx / self->RowCount;
		Row = idx - (Col * self->RowCount);
	    }
	}
	*ACol = Col;
	*ARow = Row;
}

TDirFileList *TFileWin_Get_dfl(TFileWin *self, int idx) {
  T_private *pv = self->private;

	if (!pv->ObjNy) return NULL;
	if (pv->dfl_last && (pv->dfl_nr <= idx)) {
	    while (pv->dfl_last && (pv->dfl_nr != idx)) {
		pv->dfl_nr++;
		pv->dfl_last = pv->dfl_last->next;
	    }
	    if (pv->dfl_last) return pv->dfl_last;
	}
	pv->dfl_last = pv->ObjNy->dfl;
	pv->dfl_nr = 0;
	while (pv->dfl_last && (pv->dfl_nr != idx)) {
	    pv->dfl_nr++;
	    pv->dfl_last = pv->dfl_last->next;
	}
	return pv->dfl_last;
}

static void Idx_PaintCell(TFileWin *self, int Idx) {
  int m, ACol, ARow;
	TFileWin_Idx2ColRow(self, Idx, &ACol, &ARow);
	m = 0;
	TGrid_PaintCell((TGrid*)self, ACol, ARow, &m);
}

//$01=ClsOld $02=DoItemSelect $04=SelTlg $08=OldSelIdx $11=DoItemEnter
void TFileWin_NewFocusIdx(TFileWin *self, int Idx, int fu) {
  int i, old; //, ACol, ARow;
  T_private *pv = self->private;
  TDirFileList *dfl;

	pv->dfl_foc = NULL;
	if (!pv->ObjNy) return;
	old = self->SelectIdxRow;
	self->SelectIdxRow = Idx;
	i = 0;
	dfl = pv->ObjNy->dfl;
	while (dfl) {
	    if ((dfl->fu & fls_Select) || (i == old)) {
		if (fu & 0x01) dfl->fu &= ~fls_Select;
		Idx_PaintCell(self, i);
	    }
	    if (i == Idx) {
		pv->dfl_foc = dfl;
		if (fu & 0x02) {
		    if ((fu & 0x04) == 0) dfl->fu |= fls_Select; 
				     else dfl->fu ^= fls_Select;
		}
		Idx_PaintCell(self, Idx);
		if (fu & 0x02) TFileWin_PathEnter(self, dfl, dfl->fn, ie_artItemSelect);
	    }
	    i++;
	    dfl = dfl->next;
	}
	if ((fu & 0x11)==0x11) TFileWin_DoItemEnter(self, 0x02);
}

static bool TFileWin_Sync_dfl_foc(TFileWin *self) {
  T_private *pv = self->private;
  TDirFileList *dfl;
  int i;
	if (pv->dfl_foc) {
	    i = 0;
	    dfl = TFileWin_Get_dfl0(self);
	    while (dfl) {
		if (dfl == pv->dfl_foc) {
		    if (i != self->SelectIdxRow) {
			self->SelectIdxRow = i;
			return True;
		    }
		    return False;
		}
		i++;
		dfl = dfl->next;
	    }
	    pv->dfl_foc = NULL;
	}
	return False;
}

static void set_key_stat(TFileWin *self, int state, int RowIdx, int fu) {
  int i, a, e;
  TDirFileList *dfl;

	switch(state & ~0x12) {
	case 1:  //ctrl
	    TFileWin_NewFocusIdx(self, RowIdx, 0x03);
	    if (self->OldSelIdx < 0) self->OldSelIdx = RowIdx;
	    a = self->OldSelIdx;
	    e = RowIdx;
	    if (a > e) {e = a; a = RowIdx;}
	    dfl = TFileWin_Get_dfl0(self);
	    i = 0;
	    while (dfl && (i<=e)) {
		if (i>=a) dfl->fu |= fls_Select;
		i++;
		dfl = dfl->next;
	    }
	    fu = 0x02;
	    break;

	case 4:  //strg
	    self->OldSelIdx = -1;
	    self->Strg_SelMod = True;
	    fu = 0x06;
	    break;

	default:
	    self->OldSelIdx = RowIdx;
	}

	TFileWin_SetSelect(self, RowIdx, fu);
}

void TFileWin_DoMouseDown(TFileWin *self, XButtonEvent *ev) {
  T_private *pv = self->private;
  TDirFileList *dfl;
  int RowIdx, fu;
  int AzArt;

	TGrid_DoMouseDown((TGrid*)self, ev);
	if (!pv->ObjNy) return;
	if (!self->MouseCell.InCell) return;

	AzArt = self->AzArt & 0xff;
	if ((!AzArt && (self->MouseCell.Row > 0)) || AzArt) {

  //fu $01=ClsOld $02=DoItemSelect $04=SelTlg
//		dnd_MoveStart := True;
	    fu = 0x03;
	    if (AzArt == 0) {
		RowIdx = self->MouseCell.Row-1;
#ifdef SHOW_PIC_fw
	    } else if (AzArt == 2) {
		RowIdx = self->ColCount * self->MouseCell.Row + self->MouseCell.Col;
#endif
	    } else {
		RowIdx = self->MouseCell.Row + self->MouseCell.Col * self->RowCount;
	    }
	    if (ev->button == 1) {
		set_key_stat(self, ev->state, RowIdx, fu);

	    } else if (ev->button == 3) {
		dfl = TFileWin_Get_dfl(self, RowIdx);
		if (dfl && !(dfl->fu & fls_Select)) {
		    self->OldSelIdx = -1;
		    TFileWin_NewFocusIdx(self, RowIdx, 0x03);
		}
	    }
	}
}

static void TFileWin_DocrResize(TFileWin *self, int idx, bool isCol) {

	TGrid_DocrResize((TGrid*)self, idx, isCol);

	if (isCol && ((self->AzArt & 0xff) == 0)) {
	    int art = self->ColDat[idx].tag;
	    for (int i=0; i<9; i++) {
		if (self->FDL_ColD_arr[i].art == art) {
		    self->FDL_ColD_arr[i].Width = self->ColDat[idx].w;
		    break;
		}
	    }
	}
}


void TFileWin_SetSort(TFileWin *self, int sortFu) {
  T_private *pv = self->private;
  int idx;

	sortFu &= 0x1ff;
	if (((self->sortFu & 0x1ff) == sortFu) || 
	    !TDirFileList_ChkSortOk(sortFu)) return;

	idx = (sortFu & 0xff) >> 1;
	if (idx >= sizeof(self->LastSortUD)) idx = 0;
	self->LastSortUD[idx] = sortFu & 0x01;   //== dlsortFu_DOWN
	self->Sort_Col = idx + 1;

	self->sortFu &= ~0x1ff;
	self->sortFu |= sortFu;
	if (pv->ObjNy) TDirListObjNy_Sort(pv->ObjNy, self->sortFu);
	if (TFileWin_Sync_dfl_foc(self)) TFileWin_SetIdxToWin(self, self->SelectIdxRow);

	Refresh((TWins*)self);
}

void TFileWin_DoClick(TFileWin *self, long KeyState, bool DblClick) {
  TcrDat *cr;
  int m;
  int fu = 0;

	if (DblClick) fu = 0x01;
	TWins_DoClick((TWins*)self, KeyState, DblClick);
	if (((KeyState & 0x05) != 0) || !self->MouseCell.InCell) return;

	if ((self->AzArt & 0xff) == 0) {
	    if (self->MouseCell.Row > 0) {
		if (self->MouseCell.Col == 0) TFileWin_DoItemEnter(self, fu);
	    } else if (self->RowColResize) {
	    } else if (self->MouseCell.Row == 0) {
		cr = &self->ColDat[self->MouseCell.Col];
		if (cr->tag == self->Sort_Col) {
		    m = self->sortFu ^ dlsortFu_DOWN;
		} else {
		    m = cr->tag - 1;
		    m = (self->sortFu & 0xff00) | self->LastSortUD[m] | (m << 1);
		}
		TFileWin_SetSort(self, m);
	    }
	} else {

	    TFileWin_DoItemEnter(self, fu);
	}
}

void TFileWin_SetIdxToWin(TFileWin *fw, int idx) {
  int ACol, ARow;

	TFileWin_Idx2ColRow(fw, idx, &ACol, &ARow);
	TGrid_ChkSelect_RC((TGrid*)fw, ACol, ARow);
}

void TFileWin_SetSelect(TFileWin *self, int idx, int fu) {
	TFileWin_SetIdxToWin(self, idx);
	TFileWin_NewFocusIdx(self, idx, fu);
}

int TFileWin_DoKeyDown(TFileWin *self, TKeyEvent *KeyEv) {
  T_private *pv = self->private;
  TDirFileList *dfl;
  int m, fu;
  int AzArt = self->AzArt & 0xff;

	if (TWins_DoKeyDown((TWins*)self, KeyEv)) return 1;

	if (self->SelectIdxRow < 0) return 0;
	m = self->SelectIdxRow;

	if (!pv->ObjNy) return 0;

	switch(KeyEv->keysym) {
	case XK_Up:
#ifdef SHOW_PIC_fw
	    if (AzArt == 2) m -= self->ColCount; else 
#endif
	    m--;
	    break;

	case XK_Down:
#ifdef SHOW_PIC_fw
	    if (AzArt == 2) m += self->ColCount; else
#endif
	    m++;
	    break;

	case XK_Left:
#ifdef SHOW_PIC_fw
	    if (AzArt == 2) m--; else 
#endif
	    m -= self->RowCount;
	    break;

	case XK_Right:
#ifdef SHOW_PIC_fw
	    if (AzArt == 2) m++; else 
#endif
	    m += self->RowCount;
	    break;

	case XK_Control_L:
	    if (self->Strg_SelMod) self->Strg_SelMod = False; else self->Strg_SelMod = True;
	    return 1;

	case XK_Shift_L:
	    self->Strg_SelMod = False;
	    return 1;

	case XK_space:
	    if (self->Strg_SelMod) {
		dfl = TFileWin_Get_dfl(self, m);
		if (dfl) dfl->fu ^= fls_Select;
	    }
	    break;

	case XK_Return:
	    TFileWin_DoItemEnter(self, 0x02);
	    return 1;

	default:
	    return 1;
	}

	if ((m >= pv->ObjNy->count) || (m < 0)) return 1;

	if (self->Strg_SelMod) fu = 0x00; else fu = 0x03;

	set_key_stat(self, KeyEv->state, m, fu);

	return 1;
}

static void TFileWin_DoFocus(TFileWin *self, int Change) {
	TWins_DoFocus((TWins*)self, Change);
	Idx_PaintCell(self, self->SelectIdxRow);
}

void TFileWin_PathEnter(TFileWin *self, TDirFileList *dfl, const char *fn, int art) {
  int gpfu;
  TItemEnter ie;
  char *fn_passwd = NULL;

	ie.art = art;
	ie.dfl = dfl;
	ie.name_node = fn;
	ie.stat = 0;
	gpfu = gptfu_noName;
	if (pccmp(fn, "..")) {
	    if (art != ie_artPathEnter) return;
	    gpfu = gptfu_PathBack;
	    ie.stat = 0x100;
	}

	ie.gp = TFileWin_Get_gp(self);
	NetPasswdSetup_gp(ie.gp);
	ie.fn = gptopc(ie.gp, gpfu | gptfu_DisplayFileStr);
	ie.fn_passwd = gptopc(ie.gp, gpfu | gptfu_PasswdTxt);
	if (gpfu & gptfu_noName) {
	    fn_passwd = pcdup(ie.fn_passwd);
	    ppcadd(&ie.fn, fn);
	    ppcadd(&ie.fn_passwd, fn);
	    if (ie.art == ie_artPathEnter) {
		ppcadd(&ie.fn, S_SLASH);
		ppcadd(&ie.fn_passwd, S_SLASH);
	    }
	}

	if (self->OnItemEnter) self->OnItemEnter(self, &ie);
	if ((ie.art == ie_artPathEnter) && !(ie.stat & 0x10)) {
	    ie.stat |= 0x02;
	    if (self->OnItemEnter) self->OnItemEnter(self, &ie);
	    if (!TFileWin_LoadDir(self, ie.fn_passwd)) {
		TFileWin_LoadDir(self, fn_passwd);
	    }
	}
	xfree(fn_passwd);
	xfree(ie.fn_passwd);
	xfree(ie.fn);
}

void TFileWin_DoItemEnter(TFileWin *self, int fu) {
  TDirFileList *dfl;

	dfl = TFileWin_Get_dfl(self, self->SelectIdxRow);
	if (!dfl || !(dfl->fu & fls_Select)) return;

	if (dfl->st->st & st_Dir) {
	    if ((fu & 0x02) ||
		(self->DblClick_PathEnter && (fu & 0x01)) || 
		(!self->DblClick_PathEnter && !(fu & 0x01))) {
		TFileWin_PathEnter(self, dfl, dfl->fn, ie_artPathEnter);
	    }
	} else {
	    if ((fu & 0x02) ||
		(self->DblClick_ItemEnter && (fu & 0x01)) || 
		(!self->DblClick_ItemEnter && !(fu & 0x01))) {
		TFileWin_PathEnter(self, dfl, dfl->fn, ie_artFileEnter);
	    }
	}
}

static void setrowcol(TFileWin *self, int co, int ro, bool chart) {

	SetVisible((TWins*)self->HSB, (self->AzArt & 0xff) != 2);
	SetVisible((TWins*)self->VSB, (self->AzArt & 0xff) != 1);

	if (co <= 0) co = 1;
	if (ro <= 0) ro = 1;
	if (chart || (self->RowCount != ro) || (self->ColCount != co)) {

	    if (chart) {
		TGrid_SetColCount((TGrid*)self, 0);
		TGrid_SetRowCount((TGrid*)self, 0);
	    }

	    self->FixedCols = 0;
	    self->FixedRows = 0;
	    self->DblClick_ItemEnter = self->FDL_Def.Az1_ItemE_Dbl;
	    self->DblClick_PathEnter = self->FDL_Def.Az1_PathE_Dbl;

	    TGrid_SetColCount((TGrid*)self, co);
	    TGrid_SetRowCount((TGrid*)self, ro);
	}
}

void TFileWin_SetColsArt(TFileWin *self, int AzArt) {
  T_private *pv = self->private;
  int i, ro=0, co=1;
  TGetFilePic_rszPara *old_rszPara;
  bool chart;

	chart = (self->AzArt != AzArt);
	self->AzArt = AzArt;
	old_rszPara = self->rszPara;

	if ((AzArt & 0xff) == 0) {
	    self->rszPara = &self->rszPA[0];

	    self->FixedCols = 1;
	    self->FixedRows = 1;
	    self->SelectFu |= grSelFu_NoAutoHSB;
	    self->DblClick_ItemEnter = self->FDL_Def.Az0_ItemE_Dbl;
	    self->DblClick_PathEnter = self->FDL_Def.Az0_PathE_Dbl;

	    SetVisible((TWins*)self->HSB, True);
	    SetVisible((TWins*)self->VSB, True);
	    self->DefaultRowHeight = 18;
	    TGrid_SetColCount((TGrid*)self, 9);
	    if (chart) TGrid_SetRowCount((TGrid*)self, 0);

	    TcrDat *cr = self->ColDat;
	    TFDL_ColD *ColD = self->FDL_ColD_arr;
	    for (i=0; i<9; i++) {
		cr->w = ColD->Width;
		cr->tag = ColD->art;
		cr->name = ColD->name;
		cr->Visible = (ColD->fu & 0x01);
		cr++; ColD++;
	    }
	    i = 1;
	    if (pv->ObjNy) i += pv->ObjNy->count;
	    TGrid_SetRowCount((TGrid*)self, i);

#ifdef SHOW_PIC_fw
	} else if ((AzArt & 0xff) == 2) {

	    self->SelectFu &= ~grSelFu_NoAutoHSB;
	    if (AzArt & 0x100) {
		self->rszPara = &self->rszPA[3];
		self->DefaultColWidth = 102;
		self->DefaultRowHeight = 92;
	    } else {
		self->rszPara = &self->rszPA[2];
		self->DefaultColWidth = 134;
		self->DefaultRowHeight = 116;
	    }
	    self->rszPara->max_w = self->DefaultColWidth - 6;
	    self->rszPara->max_h = self->DefaultRowHeight - 20;

	    i = 0;
	    if (pv->ObjNy) i = pv->ObjNy->count;
	    if (i) {
		co = self->width / self->DefaultColWidth;
		if (co < 1) co = 1;
		if (i > co) {
		    ro = (i+(co-1)) / co;
		} else {
		    co = i;
		}
	    }

	    setrowcol(self, co, ro, chart);

	    if ((self->AzFu & AzFu_Col1_DirHeight32) && (co == 1)) {
  TDirFileList *dfl = NULL;
  int dirZ = 0;
		dfl = TFileWin_Get_dfl0(self);
		while (dfl) {
		    if (dfl->st->st & st_Dir) dirZ++;
		    dfl = dfl->next;
		}
		for (i=0;i<dirZ;i++) {
		    self->RowDat[i].w = 56;
		}
		for (i=dirZ;i<ro;i++) {
		    self->RowDat[i].w = self->DefaultRowHeight;
		}
		TGrid_Chk_VHSB((TGrid*)self);
	    }
#endif
	} else {

	    self->SelectFu &= ~grSelFu_NoAutoHSB;
	    self->rszPara = &self->rszPA[0];
	    self->DefaultColWidth = 120;
	    self->DefaultRowHeight = 18;

	    i = 0;
	    if (pv->ObjNy) i = pv->ObjNy->count;
	    if (i) {
		ro = self->height / self->DefaultRowHeight;
		if (ro < 1) ro = 1;
		if (i > ro) {
		    co = (i+(ro-1)) / ro;
		} else {
		    ro = i;
		}
	    }

	    setrowcol(self, co, ro, chart);
	}

	if (self->SelectIdxRow < 0) self->SelectIdxRow = 0;


	if (old_rszPara != self->rszPara) {
	    if (self->GetFilePic) self->GetFilePic(self, NULL, 0, fwgfp_bef_ClrAll);
	}

	if (chart) {
	    Resize(self->pw);
	    Refresh((TWins*)self);
	}
}


TglobPath *TFileWin_Get_gp(TFileWin *self) {
  T_private *pv = self->private;
	if (!pv->ObjNy) return NULL;
	return &pv->ObjNy->dl->gp;
	
}
TDirFileList *TFileWin_Get_dfl0(TFileWin *self) {
  T_private *pv = self->private;
	if (!pv->ObjNy) return NULL;
	return pv->ObjNy->dfl;
}
TDirFileList *TFileWin_Get_Select_dfl(TFileWin *self, TDirFileList *old_dfl) {
  TDirFileList *dfl;

	dfl = TFileWin_Get_dfl0(self);
	if (old_dfl) {
	    while (dfl && (dfl != old_dfl)) dfl = dfl->next;
	    if (dfl) dfl = dfl->next;
	}
	while (dfl && !(dfl->fu & dfl_Select)) dfl = dfl->next;
	return dfl;
}


char *TFileWin_GetAktDir(TFileWin *fw) {
	return gptopc(TFileWin_Get_gp(fw), gptfu_noName | gptfu_add_slash);
}

#if 0
#ifdef _OS_WIN32
static void TFileWin_SetDriveList(TFileWin *fw) {
	if (!fw->DriveList) fw->DriveList = xmalloc(256);
	GetLogicalDriveStrings(256, fw->DriveList);
}
#endif /* _OS_WIN32 */
#endif

void TFileWin_SetMouseZ_Wheel(TFileWin *self, int Value) {
	TGrid_SetMouseZ_Wheel((TGrid*)self, Value);
}

/*_____________________________ Load Ev _________________________________*/

static void TFileWin_OnFileMsg(TFileMsg *fmsg) {
  TDirListObjNy *self;
  TFileWin *fw;
  T_private *pv;
  int kbef, bef;
  TDirFileList *dfl;

	if (!file_msg_chk_wtyp(fmsg, wtyp_TDirListObjNy, FL)) return;

	self = fmsg->obj;
	fw = self->owner;
	if (!fw) return;
	if ((fmsg->art != fmArt_DirList) ||
	    (fmsg->typD != fmTyp_int)) {
	    if (fw->OnFileMsg) fw->OnFileMsg(fw, fmsg);
	    return;
	}

	fmsg->Ok = True;
	bef = fmsg->D.m;
	dfl = fmsg->F.dfl;

	pv = fw->private;

	kbef = bef & dlbef_BefMask;
	switch (kbef) {

	case dlbef_Add:
	case dlbef_NeedSort:
	    fw->NyAzStat |= 0x01;
	    break;

	case dlbef_Del:
//	    fw->NyAzStat |= 0x02;
	    pv->dfl_last = NULL;
	    break;

	case dlbef_Change:
	    if ((bef & dlbef_NotifyRun) || (bef & dlbef_RescanRun)) {
		if (fw->GetFilePic) {
		    fw->GetFilePic(fw, dfl, 0, fwgfp_bef_ClrPic);
		    fw->GetFilePic(fw, dfl, 0, fwgfp_bef_GetPic);
		}
		fw->NyAzStat |= 0x04;
	    }
	    break;

	case dlbef_ListEnd:
	case dlbef_Sync: 
	    if (fw->NyAzStat & 0x01) TDirListObjNy_Sort(self, fw->sortFu);
	    if (fw->NyAzStat != 0x04) {
		TFileWin_SetColsArt(fw, fw->AzArt);
		if (bef & dlbef_NotifyRun) {
		    TFileWin_Sync_dfl_foc(fw);
		}
	    }
	    Refresh((TWins*)fw);
	    fw->NyAzStat = 0x00;
	    break;

	case dlbef_Clr:
	    if (fw->GetFilePic) fw->GetFilePic(fw, NULL, 0, fwgfp_bef_ClrAll);
	    break;

	case dlbef_ClrEnd:
//	case dlbef_DelSelf:
	case dlbef_Destroy:
	    if (pv->ObjNy) {
		pv->dfl_foc = NULL;
		fw->SelectIdxRow = 0;
		if (kbef == dlbef_Destroy) {
		    pv->ObjNy = NULL;
		    ppcfree(&fw->Akt_FileName);
		}
		TFileWin_SetColsArt(fw, fw->AzArt);
		Refresh((TWins*)fw);
	    }
	    break;


	case dlbef_DelSelf:
	    fmsg->Ok = False;	//in TDirListObjNy   Tfree(ObjNy);
	    break;
	}
	if (fw->OnFileMsg) fw->OnFileMsg(fw, fmsg);
}

/*_____________________________ Load ____________________________________*/
static void tvn_ObjNy_Free(TFileWin *self) {
  T_private *pv;
  TDirListObjNy *ObjNy;

	pv = self->private;
	if (!pv || !pv->ObjNy) return;
	ObjNy = pv->ObjNy;
	pv->ObjNy = NULL;
	Tfree(ObjNy);
}


void TFileWin_Clr(TFileWin *self) {
  T_private *pv = self->private;

	tvn_ObjNy_Free(self);
	ppcfree(&self->Akt_FileName);
	pv->dfl_foc = NULL;
	self->SelectIdxRow = 0;
	TGrid_SetRowCount((TGrid*)self, 0);
	TScrollBar_SetCount(self->HSB, 0);
	TScrollBar_SetCount(self->VSB, 0);
	RePaint((TWins*)self);
}

bool TFileWin_LoadDir(TFileWin *self, char *FileName) {
  T_private *pv = self->private;
  TDirListObjNy *ObjNy;
  int fst;
  char *fn;

	pv->dfl_nr = 0;
	pv->dfl_last = NULL;

	fn = dotdot_korr(FileName);
	if (!fn) fn = pcdup(S_SLASH);

	fst = fwOn_stat_Start;
	if (self->OnLoadDir) self->OnLoadDir(self, &fn, &fst);
	if (fst & fwOn_stat_RetHalt) {
	    xfree(fn);
	    return False;
	}

	if (pccmp(self->Akt_FileName, fn)) {
	    xfree(fn);
	    return True;
	}
	TFileWin_Clr(self);
	AppHandleWinMessages();

	ppcchange(&self->Akt_FileName, fn);
	ObjNy = TDirListObjNy_Create(self);
	pv->ObjNy = ObjNy;
	ObjNy->OnFileMsg = TFileWin_OnFileMsg;
	ObjNy->chstat_Filter = chstat_Add|chstat_Del|chstat_mtime|chstat_size|
			chstat_mode|chstat_uid|chstat_gid|
			chstat_FileType|chstat_mime_id|chstat_sub_mime;
	ObjNy->findAtr &= ~findNoHidden;
	ObjNy->findAtr |= findDIR_NoPattern;
	ObjNy->patt = self->patt;
	ObjNy->fu = dlObjNyfu_pattNotFree;
	if (!pv->ShowHidden) ObjNy->findAtr |= findNoHidden;

	fst = fwOn_stat_End;
	if (!TDirListObjNy_Open(ObjNy, fn)) fst = fwOn_stat_Error;

	if (self->OnLoadDir) self->OnLoadDir(self, &fn, &fst);
	if ((fst & fwOn_stat_Mask) == fwOn_stat_Error) TFileWin_Clr(self);
	xfree(fn);

	return fst == fwOn_stat_End;
}

void TFileWin_SetShowHidden(TFileWin *self, bool ShowHidden) {
  T_private *pv = self->private;

	if (pv->ShowHidden == ShowHidden) return;
	pv->ShowHidden = ShowHidden;
	if (!pv->ObjNy) return;
	pv->ObjNy->findAtr &= ~findNoHidden;
	if (!ShowHidden) pv->ObjNy->findAtr |= findNoHidden;

  TDirFileList *dfl = pv->ObjNy->dl->dfl;
	while (dfl) {
	    if (dfl->st->st & st_Hidden) {
		TDirListObjNy_Open(pv->ObjNy, NULL);
		return;
	    }
	    dfl = dfl->next;
	}
}

void TFileWin_ReLoadDir(TFileWin *self) {
  T_private *pv = self->private;
	TDirListObjNy_Open(pv->ObjNy, NULL);
}

void TFileWin_ReScan(TFileWin *self) {
  T_private *pv = self->private;

	TDirListObjNy_ReScan(pv->ObjNy);
}

/*_____________________________ Destroy _________________________________*/
bool TFileWin__Destroy(TFileWin *self) {
  int i;
  T_private *pv = self->private;

	if (self->GetFilePic) self->GetFilePic(self, NULL, 0, fwgfp_bef_Free);
	for (i=0; i<9; i++) xfree(self->FDL_ColD_arr[i].name);
	Tfree(pv->ObjNy);
	xfree(pv);
	Tfree(self->patt);
	xfree(self->Akt_FileName);
	return TGrid__Destroy((TGrid*)self);
}

/*_____________________________ Create __________________________________*/
static void TFileWin_CreateParams(TWins_Create_struct *wcs) {
  TFileWin *self = wcs->wins;

	TGrid_CreateParams(wcs);

	self->destroy = (TDESTROY)TFileWin__Destroy;
	self->Resize = TFileWin_Resize;
	self->DoPaint = TFileWin_DoPaint;
	self->DoPaintCell = TFileWin_DoPaintCell;
	self->DoMouseDown = TFileWin_DoMouseDown;
	self->DocrResize = TFileWin_DocrResize;
	self->DoClick = TFileWin_DoClick;
	self->DoKeyDown = TFileWin_DoKeyDown;
	self->DoFocus = TFileWin_DoFocus;

	self->HSB->OnSmallChange = HSB_OnSmallChange;
	self->HSB->OnLargeChange = HSB_OnLargeChange;
	self->VSB->OnSmallChange = VSB_OnSmallChange;
//	self->VSB->OnLargeChange = VSB_OnLargeChange;

	self->FDL_Def.Az0_ItemE_Dbl = False;
	self->FDL_Def.Az0_PathE_Dbl = False;
	self->FDL_Def.Az1_ItemE_Dbl = True;
	self->FDL_Def.Az1_PathE_Dbl = False;

	self->FixedColsFu |= grFixResize | grFixNoBorder;
	self->FixedColsColor = 0xf0f0f0;
	self->Sort_Col = 1;

}

TFileWin *TFileWin_Create(TWins *Parent) {
  TWins_Create_struct wcs;
	TWins_Create_struct_init(&wcs, sizeof(TFileWin), wtyp_FileWin);
	wcs.Parent = Parent;
	wcs.CreateParams = TFileWin_CreateParams;
	TGrid_DefaultConf(&wcs);
	((TFileWin*)wcs.wins)->private = xcalloc(sizeof(T_private));
	TFileWin_InitDefaultCols((TFileWin*)wcs.wins);
	return wcs.wins;
}

TFileWin *TFileWin_CreateA(TWins *Parent, int left, int top, int width, int height, int Align) {
  TFileWin *self;

	self = TFileWin_Create(Parent);
	self->top = top;
	self->left = left;
	self->width = width;
	self->height = height;
	self->Align = alClient;
	self->pw->Align = Align;
	Set_Parent((TWins*)self, Parent);
	return self;
}
