/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appbase.h"
#include "media_io.h"
#include "alsa_mixer.h"
#include <alsa/asoundlib.h>
#include <math.h>

void mixer_On_appEv(TEv_Array *appEv);

typedef struct {
  snd_mixer_t *mixer_handle;
  char	*mixer_sid;
  int	sid_size;
  int mixer_changed_state;
  TEv_Array *appEv;
  struct pollfd *fds;
} TSound_AlsaMIX_private;

static void free_elems(TSound_MIXER *mix) {
  int i;
  TSound_AlsaMIX_private *pv = mix->private;
  bool ev;

	ev = ((mix->stat & iost_init) != 0) && mix->OnElemDel;
	mix->stat = 0;

	Free_appEv(pv->appEv);
	pv->appEv = NULL;

	if (mix->mixer_n_elems && mix->mix_elem) {
	    for (i=0; i<mix->mixer_n_elems;i++) {
		if (ev) mix->OnElemDel(mix, mix->mix_elem[i]);
		xfree(mix->mix_elem[i]);
	    }
	    xfree(mix->mix_elem);
	}
	mix->mixer_n_elems = 0;
	mix->mix_elem = NULL;

	xfree(pv->mixer_sid);
	pv->mixer_sid = NULL;
	xfree(pv->fds);
	pv->fds = NULL;
}

static void close_mixer(TSound_MIXER *mix) {
  TSound_AlsaMIX_private *pv = mix->private;

	pv = mix->private;
	if (pv->mixer_handle) snd_mixer_close(pv->mixer_handle);
	pv->mixer_handle = NULL;
	free_elems(mix);
}

/* mixer bar channel : left or right */
#define MIXER_CHN_LEFT		0
#define MIXER_CHN_RIGHT		1

/* mask for toggle mute and capture */
//#define MIXER_MASK_LEFT		(1 << 0)
//#define MIXER_MASK_RIGHT	(1 << 1)
//#define MIXER_MASK_STEREO	(MIXER_MASK_LEFT|MIXER_MASK_RIGHT)

/* left and right channels for each type */
static const snd_mixer_selem_channel_id_t mixer_elem_chn[][2] = {
  { SND_MIXER_SCHN_FRONT_LEFT, SND_MIXER_SCHN_FRONT_RIGHT },
  { SND_MIXER_SCHN_REAR_LEFT, SND_MIXER_SCHN_REAR_RIGHT },
  { SND_MIXER_SCHN_FRONT_CENTER, SND_MIXER_SCHN_UNKNOWN },
  { SND_MIXER_SCHN_WOOFER, SND_MIXER_SCHN_UNKNOWN },
  { SND_MIXER_SCHN_SIDE_LEFT, SND_MIXER_SCHN_SIDE_RIGHT },
  { SND_MIXER_SCHN_FRONT_LEFT, SND_MIXER_SCHN_FRONT_RIGHT },
  { SND_MIXER_SCHN_UNKNOWN, SND_MIXER_SCHN_UNKNOWN },
  { SND_MIXER_SCHN_UNKNOWN, SND_MIXER_SCHN_UNKNOWN }
};

static int te_float(float v1, float v2) {
	return ((v1 > v2+0.001) || (v1 < v2-0.001));
}

static void mixer_update_val (TSound_MIXER *mix, int elem_index) {
  int i, m;
  snd_mixer_elem_t *elem;
  long vleft, vright;
  long vdbleft, vdbright;
  int swl, swr;
  int type;
  snd_mixer_selem_id_t *sid;
  TSound_AlsaMIX_private *pv = mix->private;
  Tmix_elem *mix_elem;

	if (!pv->mixer_sid || !pv->mixer_handle) return;
	mix_elem = mix->mix_elem[elem_index];


	sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * mix_elem->mixer_grpidx);
	elem = snd_mixer_find_selem(pv->mixer_handle, sid);
	if (!elem) return;

	type = mix_elem->mixer_type & MIXER_ELEM_TYPE_MASK;
	if (type == MIXER_ELEM_ENUM || type == MIXER_ELEM_CAPTURE_ENUM) {
	  m = mix_elem->enum_nr;
	  for (i = 0; i < 2; i++) {
	    if (snd_mixer_selem_get_enum_item(elem, i, &mix_elem->enum_nr) < 0) break;
	    if (snd_mixer_selem_get_enum_item_name(elem, mix_elem->enum_nr, 
		MIX_ITEMNAME_SIZE-1, mix_elem->enum_name) < 0) break;
	  }
	  if (mix_elem->enum_nr != m) {
	      mix_elem->st |= MIX_ELEM_Change;
	  }
	  return;
	}

	if (mix_elem->mixer_type & MIXER_ELEM_HAS_VOLUME) {

	  vleft = vright = 0;
	  vdbleft = vdbright = 0;
	  if ((type != MIXER_ELEM_CAPTURE) && snd_mixer_selem_has_playback_volume(elem)) {
	    snd_mixer_selem_get_playback_volume(elem, mix_elem->c_id_left, &vleft);
	    snd_mixer_selem_get_playback_dB(elem, mix_elem->c_id_left, &vdbleft);
	    vright = vleft;
	    vdbright = vdbleft;
	    if (mix_elem->channels > 1) {
	      snd_mixer_selem_get_playback_volume(elem, mix_elem->c_id_right, &vright);
	      snd_mixer_selem_get_playback_dB(elem, mix_elem->c_id_right, &vdbright);
	    }
	  }
	  if ((type == MIXER_ELEM_CAPTURE) && snd_mixer_selem_has_capture_volume(elem)) {
	    snd_mixer_selem_get_capture_volume(elem, mix_elem->c_id_left, &vleft);
	    snd_mixer_selem_get_capture_dB(elem, mix_elem->c_id_left, &vdbleft);
	    vright = vleft;
	    if (mix_elem->channels > 1) {
	      snd_mixer_selem_get_capture_volume(elem, mix_elem->c_id_right, &vright);
	      snd_mixer_selem_get_capture_dB(elem, mix_elem->c_id_right, &vdbright);
	    }
	  }

    float tf;
	  tf = ((float)vleft - mix_elem->vmin) / mix_elem->vCount;
	  if (te_float(mix_elem->vol_left, tf)) {
	      mix_elem->vol_left = tf;
	      mix_elem->st |= MIX_ELEM_Change;
	  }
	  tf = ((float)vright - mix_elem->vmin) / mix_elem->vCount;
	  if (te_float(mix_elem->vol_right, tf)) {
	      mix_elem->vol_right = tf;
	      mix_elem->st |= MIX_ELEM_Change;
	  }
	  tf = (float)vdbleft / 100;
	  if (te_float(mix_elem->dB_left, tf)) {
	      mix_elem->dB_left = tf;
	      mix_elem->st |= MIX_ELEM_Change;
	  }
	  tf = (float)vdbright / 100;
	  if (te_float(mix_elem->dB_right, tf)) {
	      mix_elem->dB_right = tf;
	      mix_elem->st |= MIX_ELEM_Change;
	  }
	}

	if ((mix_elem->mixer_type & MIXER_ELEM_CAPTURE_SWITCH) ||
	    (mix_elem->mixer_type & MIXER_ELEM_MUTE_SWITCH)) {
	  swr = swl = 0;
	  if (mix_elem->mixer_type & MIXER_ELEM_CAPTURE_SWITCH) {
	    snd_mixer_selem_get_capture_switch(elem, mix_elem->c_id_left, &swl);
	    swr = swl;
	    if (mix_elem->channels > 1) {
	      snd_mixer_selem_get_capture_switch(elem, mix_elem->c_id_right, &swr);
	    }
	  } else {
	    snd_mixer_selem_get_playback_switch(elem, mix_elem->c_id_left, &swl);
	    if (mix_elem->channels > 1) {
	      snd_mixer_selem_get_playback_switch(elem, mix_elem->c_id_right, &swr);
	    }
	  }

	  m = 0;
	  if (swl) m = 1;
	  if (swr) m |= 2;
	  if (mix_elem->mute != m) {
	      mix_elem->mute = m;
	      mix_elem->st |= MIX_ELEM_Change;
	  }
	}
/* if (mix_elem->st & MIX_ELEM_Change) {
  printf("%ld %ld %f %f : %ld %ld  %f %f\n", vleft, vright, mix_elem->vol_left, mix_elem->vol_right,
		vdbleft, vdbright, mix_elem->db_left, mix_elem->db_right);
 } */
}

static void mixer_update (TSound_MIXER *mix) {
  int i;
  Tmix_elem *mix_elem;

	for (i = 0; i < mix->mixer_n_elems; i++) {
	    mixer_update_val(mix, i);
	    mix_elem = mix->mix_elem[i];
	    if ((mix_elem->st & MIX_ELEM_Change) && mix->OnElemChange) {
		mix->OnElemChange(mix, mix_elem);
	    }
/*	    if (mix_elem->st & MIX_ELEM_Change) {
 printf("%d %.2f %.2f : %.2f %.2f\n", mix_elem->idx, mix_elem->vol_left, mix_elem->vol_right,
		mix_elem->db_left, mix_elem->db_right);
	    } */
	    mix_elem->st &= ~MIX_ELEM_Change;
	}
}

static int mixer_reinit (TSound_MIXER *mix) {
  TSound_AlsaMIX_private *pv;
  snd_mixer_elem_t *elem;
  int idx, i, j, selem_count, mixer_n_selems, mixer_n_elems;
  snd_mixer_selem_id_t *sid;
  Tmix_elem **pp_mix_elem = NULL;
  Tmix_elem *mix_elem;

	pv = mix->private;
	if (!pv->mixer_changed_state) return True;
	mixer_n_elems = 0;
__again:
//	xfree(pv->mixer_sid);
	free_elems(mix);
	if (pp_mix_elem) {
	    for (i=0; i<mixer_n_elems;i++) xfree(pp_mix_elem[i]);
	    xfree(pp_mix_elem);
	    pp_mix_elem = NULL;
	}
	pv->mixer_changed_state = 0;

	selem_count = snd_mixer_get_count(pv->mixer_handle);
// printf("selem_count=%d\n", selem_count);
	pv->mixer_sid = xmalloc(pv->sid_size * selem_count);

	mixer_n_selems = 0;
	elem = snd_mixer_first_elem(pv->mixer_handle);
	while (elem) {
	  sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * mixer_n_selems);
	  if (pv->mixer_changed_state) goto __again;
	  if (!snd_mixer_selem_is_active(elem)) continue;
	  snd_mixer_selem_get_id(elem, sid);
	  mixer_n_selems++;
	  elem = snd_mixer_elem_next(elem);
	}

	mixer_n_elems = 0;
	for (idx = 0; idx < mixer_n_selems; idx++) {
    int nelems_added = 0;
	  sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * idx);
	  if (pv->mixer_changed_state) goto __again;
	  elem = snd_mixer_find_selem(pv->mixer_handle, sid);
	  if (!elem) goto xerr;

	  for (i = 0; i < MIXER_ELEM_CAPTURE; i++) {
	    int ok;
	    for (j = ok = 0; j < 2; j++) {
	      if (pv->mixer_changed_state) goto __again;
	      if (snd_mixer_selem_has_playback_channel(elem, mixer_elem_chn[i][j])) ok++;
	    }
	    if (ok) {
	      nelems_added++;
	      mixer_n_elems++;
	    }
	  }
	  if (snd_mixer_selem_has_capture_volume(elem) ||
	      (!nelems_added && snd_mixer_selem_has_capture_switch(elem)))
	    mixer_n_elems++;
	}
// printf("mixer_n_selems=%d mixer_n_elems=%d\n", mixer_n_selems, mixer_n_elems);

/*	xfree(mf->mixer_type);
	xfree(mf->mixer_grpidx);
	xfree(mf->mixer_item_name);
	mf->mixer_type = xcalloc(mf->mixer_n_elems * sizeof(int));
	mf->mixer_grpidx = xcalloc(mf->mixer_n_elems * sizeof(int));
	mf->mixer_item_name = xcalloc(mf->mixer_n_elems * sizeof(Tmixer_item_name));
*/
	pp_mix_elem = xcalloc(mixer_n_elems * sizeof(void*));
	for (i=0;i<mixer_n_elems;i++) {
	    pp_mix_elem[i] = xcalloc(sizeof(Tmix_elem));
	}

  int elem_index = 0;
	for (idx = 0; idx < mixer_n_selems; idx++) {
    int nelems_added = 0;
	  mix_elem = pp_mix_elem[elem_index];
	  sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * idx);
	  if (pv->mixer_changed_state) goto __again;
	  elem = snd_mixer_find_selem(pv->mixer_handle, sid);
	  if (!elem) goto xerr;

	  for (i = MIXER_ELEM_FRONT; i <= MIXER_ELEM_SIDE; i++) {
	    int ok;
	    for (j = ok = 0; j < 2; j++) {
	      if (pv->mixer_changed_state) goto __again;
	      if (snd_mixer_selem_has_playback_channel(elem, mixer_elem_chn[i][j])) ok++;
	    }
	    if (ok) {
//	      sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * idx);
	      mix_elem->mixer_grpidx = idx;
	      if (snd_mixer_selem_is_enumerated(elem)) {
//		if (mixer_view == VIEW_PLAYBACK &&
//		    snd_mixer_selem_is_enum_capture(elem))
//		  continue;
		mix_elem->mixer_type = MIXER_ELEM_ENUM;
	      } else {
		mix_elem->mixer_type = i;
		if (!i && snd_mixer_selem_has_playback_switch(elem))
		  mix_elem->mixer_type |= MIXER_ELEM_MUTE_SWITCH;
		if (snd_mixer_selem_has_playback_volume(elem))
		  mix_elem->mixer_type |= MIXER_ELEM_HAS_VOLUME;
	      }
//	      if (mixer_view == VIEW_CHANNELS) {
		if (!nelems_added &&
		    !snd_mixer_selem_has_capture_volume(elem) &&
		    snd_mixer_selem_has_capture_switch(elem))
		  mix_elem->mixer_type |= MIXER_ELEM_CAPTURE_SWITCH;
//	      }
	      elem_index++;
	      nelems_added++;
	      if (elem_index >= mixer_n_elems) goto elems_end;
	      mix_elem = pp_mix_elem[elem_index];
	    }
	  }

	  int do_add = 0;
	  if (snd_mixer_selem_has_capture_volume(elem) &&
	      !snd_mixer_selem_has_common_volume(elem))
	    do_add = 1;
	  if (!do_add &&
	      (!nelems_added && snd_mixer_selem_has_capture_switch(elem)) &&
	       !snd_mixer_selem_has_common_switch(elem))
	    do_add = 1;
	  if (!do_add && snd_mixer_selem_is_enum_capture(elem))
	    do_add = 1;

	  if (do_add) {
	    mix_elem->mixer_grpidx = idx;
	    if (snd_mixer_selem_is_enum_capture(elem)) {
	      mix_elem->mixer_type = MIXER_ELEM_CAPTURE_ENUM;
	    } else {
	      mix_elem->mixer_type = MIXER_ELEM_CAPTURE;
	      if (!nelems_added  && snd_mixer_selem_has_capture_switch(elem))
		mix_elem->mixer_type |= MIXER_ELEM_CAPTURE_SWITCH;
	      if (nelems_added)
		mix_elem->mixer_type |= MIXER_ELEM_CAPTURE_SUFFIX;
	      if (snd_mixer_selem_has_capture_volume(elem))
		mix_elem->mixer_type |= MIXER_ELEM_HAS_VOLUME;
	    }
	    elem_index++;
	    if (elem_index >= mixer_n_elems) goto elems_end;
	  }
	}
elems_end:
	mix->mixer_n_elems = mixer_n_elems;
	mix->mix_elem = pp_mix_elem;
	mix->stat = iost_init;

	for (i = 0; i < mixer_n_elems; i++) {
	    mix_elem = pp_mix_elem[i];
	    mix_elem->idx = i;
//	    mix_elem->enum_nr = -1;
    const char *suffix = "";
	    if (mix_elem->mixer_type & MIXER_ELEM_CAPTURE_SUFFIX) suffix = " Capture";
	    sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * mix_elem->mixer_grpidx);

	    if (snd_mixer_selem_id_get_index(sid) > 0)
		snprintf(mix_elem->item_name, MIX_ITEMNAME_SIZE-1, 
			"%s%s %d", snd_mixer_selem_id_get_name(sid),
			suffix, snd_mixer_selem_id_get_index(sid));
	    else
		snprintf(mix_elem->item_name, MIX_ITEMNAME_SIZE-1, 
			"%s%s", snd_mixer_selem_id_get_name(sid), suffix);

	    elem = snd_mixer_find_selem(pv->mixer_handle, sid);
	    if (!elem) goto xerr;


	int type = mix_elem->mixer_type & MIXER_ELEM_TYPE_MASK;
	    mix_elem->c_id_left = mixer_elem_chn[type][MIXER_CHN_LEFT];
	    mix_elem->c_id_right = mixer_elem_chn[type][MIXER_CHN_RIGHT];
	    if (mix_elem->c_id_right != SND_MIXER_SCHN_UNKNOWN) {
	      if (type != MIXER_ELEM_CAPTURE) {
		if (!snd_mixer_selem_has_playback_channel(elem, mix_elem->c_id_right))
		  mix_elem->c_id_right = SND_MIXER_SCHN_UNKNOWN;
	      } else {
		if (!snd_mixer_selem_has_capture_channel(elem, mix_elem->c_id_right))
		  mix_elem->c_id_right = SND_MIXER_SCHN_UNKNOWN;
	      }
	    }

	    if (mix_elem->c_id_left >= 0) {
		mix_elem->channels = 1;
		if (mix_elem->c_id_right >= 0) mix_elem->channels = 2;
	    }

	long vmin, vmax, dB_vmin, dB_vmax;
	    vmin = vmax = dB_vmin = dB_vmax = 0;
	    mix_elem->vst &= ~(VST_VMAX|VST_VMDB);
	    if ((type != MIXER_ELEM_CAPTURE) && snd_mixer_selem_has_playback_volume(elem)) {
	      if (!snd_mixer_selem_get_playback_volume_range(elem, &vmin, &vmax) &&
		  (vmin < vmax)) mix_elem->vst |= VST_VMAX;
	      if (!snd_mixer_selem_get_playback_dB_range(elem, &dB_vmin, &dB_vmax) &&
		  (dB_vmin < dB_vmax)) mix_elem->vst |= VST_VMDB;
	    }
	    if ((type == MIXER_ELEM_CAPTURE) && snd_mixer_selem_has_capture_volume(elem)) {
	      if (!snd_mixer_selem_get_capture_volume_range(elem, &vmin, &vmax) &&
		  (vmin < vmax)) mix_elem->vst |= VST_VMAX;
	      if (!snd_mixer_selem_get_capture_dB_range(elem, &dB_vmin, &dB_vmax) &&
		  (dB_vmin < dB_vmax)) mix_elem->vst |= VST_VMDB;
	    }
	    mix_elem->vmin = vmin;
	    mix_elem->vmax = vmax;
	    mix_elem->vCount = vmax - vmin;
	    mix_elem->dB_vmin = (float)dB_vmin / 100;
	    mix_elem->dB_vmax = (float)dB_vmax / 100;

// printf("%-18s 0x%04x %3ld %3ld  %5ld %5ld  %2d %2d %d  0x%04x\n", mix_elem->item_name, mix_elem->mixer_type, vmin, vmax, dB_vmin, dB_vmax,
//		mix_elem->c_id_left, mix_elem->c_id_right, mix_elem->channels, mix_elem->vst);

	    mixer_update_val(mix, i);

	    if (mix->OnElemAdd) mix->OnElemAdd(mix, mix_elem);
	    mix_elem->st &= ~MIX_ELEM_Change;
	}

	pv->appEv = Add_appEv(mixer_On_appEv);
	pv->appEv->obj = mix;

	pv->fds = xcalloc(sizeof(struct pollfd));
	snd_mixer_poll_descriptors(pv->mixer_handle, pv->fds, 1);

	return True;

xerr:
	appError(EMSG_ERR1, "%s(%d): function snd_mixer_find_selem() failed for %s: %s", 
			    FL, mix->dev_name, snd_strerror(-EINVAL));
	return False;

}

static Tmix_elem *get_mix_elem(TSound_MIXER *mix, int elem_index) {
  TSound_AlsaMIX_private *pv = mix->private;

	if (!pv || !pv->mixer_sid || !pv->mixer_handle ||
	    (elem_index < 0) || (elem_index >= mix->mixer_n_elems)) return NULL;
	return mix->mix_elem[elem_index];
}

/*_____________________________ set mute _________________________________*/
static void mixer_set_mute(TSound_MIXER *mix, int elem_index, int mute) {
  snd_mixer_elem_t *elem;
  snd_mixer_selem_id_t *sid;
  TSound_AlsaMIX_private *pv = mix->private;
  Tmix_elem *mxe;

	mxe = get_mix_elem(mix, elem_index);
	if (!mxe) return;

	mute &= 0x03;
	if ((mxe->mute & 0x03) == mute) return;
	mxe->mute &= ~0x03;
	mxe->mute |= mute;
	if (!(mxe->mixer_type & MIXER_ELEM_MUTE_SWITCH) &&
	    !(mxe->mixer_type & MIXER_ELEM_CAPTURE_SWITCH)) return;

	sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * mxe->mixer_grpidx);
	elem = snd_mixer_find_selem(pv->mixer_handle, sid);
	if (!elem) return;

	if (mxe->mixer_type & MIXER_ELEM_MUTE_SWITCH) {
	    snd_mixer_selem_set_playback_switch(elem, mxe->c_id_left, (mute & 0x01) != 0);
	    if (mxe->channels > 1)
		snd_mixer_selem_set_playback_switch(elem, mxe->c_id_right, (mute & 0x02) != 0);
	} else {
	    snd_mixer_selem_set_capture_switch(elem, mxe->c_id_left, (mute & 0x01) != 0);
	    if (mxe->channels > 1)
		snd_mixer_selem_set_capture_switch(elem, mxe->c_id_right, (mute & 0x02) != 0);
	}
}

void TSound_AlsaMIX_SetMute(TSound_MIXER *mix, int elem_index, int mute) {
	if (!Sound_MIXER_ChkType(mix)) return;
	mixer_set_mute(mix, elem_index, mute);
}

/*_____________________________ set vol __________________________________*/

static void mixer_set_vol(TSound_MIXER *mix, int elem_index, float vol_left, float vol_right) {
  snd_mixer_elem_t *elem;
  snd_mixer_selem_id_t *sid;
  TSound_AlsaMIX_private *pv = mix->private;
  Tmix_elem *mxe;
  int type, ival;

	mxe = get_mix_elem(mix, elem_index);
	if (!mxe) return;

	if (!(mxe->mixer_type & MIXER_ELEM_HAS_VOLUME)) return;

	sid = (snd_mixer_selem_id_t *)(pv->mixer_sid + pv->sid_size * mxe->mixer_grpidx);
	elem = snd_mixer_find_selem(pv->mixer_handle, sid);
	if (!elem) return;

	type = mxe->mixer_type & MIXER_ELEM_TYPE_MASK;

	if (vol_left < 0) vol_left = 0;
	if (vol_left > 1) vol_left = 1;

	if (te_float(mxe->vol_left, vol_left)) {
	    ival = round((vol_left * mxe->vCount) + mxe->vmin);
	    if (type != MIXER_ELEM_CAPTURE) {
		snd_mixer_selem_set_playback_volume(elem, mxe->c_id_left, ival);
	    } else {
		snd_mixer_selem_set_capture_volume(elem, mxe->c_id_left, ival);
	    }
	}

	if (vol_right < 0) vol_right = 0;
	if (vol_right > 1) vol_right = 1;
	if ((mxe->channels > 1) && te_float(mxe->vol_right, vol_right)) {
	    ival = round((vol_right * mxe->vCount) + mxe->vmin);
	    if (type != MIXER_ELEM_CAPTURE) {
		snd_mixer_selem_set_playback_volume(elem, mxe->c_id_right, ival);
	    } else {
		snd_mixer_selem_set_capture_volume(elem, mxe->c_id_right, ival);
	    }
	}

}

void TSound_AlsaMIX_SetVol(TSound_MIXER *mix, int elem_index, float vol_left, float vol_right) {
	if (!Sound_MIXER_ChkType(mix)) return;
	mixer_set_vol(mix, elem_index, vol_left, vol_right);
}

/*_____________________________ event ____________________________________*/

static int mixer_event (snd_mixer_t *mixer, unsigned int mask, snd_mixer_elem_t *elem) { 
  TSound_MIXER *mix = snd_mixer_get_callback_private(mixer);
// printf("mixer_event 0x%x 0x%x 0x%x\n", mask, (int)elem, (int)mix);
  TSound_AlsaMIX_private *pv = mix->private;
	pv->mixer_changed_state = 1;
// printf("mixer_event 0x%x 0x%x\n", mask, (int)elem);
//  ((TMainForm*)appDat->obj)->mixer_changed_state = 1;
  return 0;
}

void mixer_On_appEv(TEv_Array *appEv) {
  int finished;
  unsigned short revents;
  TSound_MIXER *mix = appEv->obj;
  TSound_AlsaMIX_private *pv = mix->private;

	if (mix->stat & iost_init) {

	  finished = poll(pv->fds, 1, 0);

	  if ((finished > 0) && 
	      (snd_mixer_poll_descriptors_revents(pv->mixer_handle, pv->fds, 1, &revents) >= 0)) {
	    if (revents & POLLIN) {
//  printf("UserAppEvent  %d 0x%x\n", finished, revents);
	      snd_mixer_handle_events(pv->mixer_handle);
	      mixer_update(mix);
	    }
	  }

	}
}
/*_____________________________ Init ____________________________________*/
static int TSound_AlsaMIX_Init(TSound_MIXER *mix) {
  int err;
  TSound_AlsaMIX_private *pv;
  snd_ctl_t *ctl_handle;
  snd_ctl_card_info_t *hw_info;
  const char *err_str;

	close_mixer(mix);

	pv = mix->private;
	snd_ctl_card_info_alloca(&hw_info);
	if ((err = snd_ctl_open(&ctl_handle, mix->dev_name, 0)) < 0) {
	    appError(EMSG_ERR1, "%s(%d): function snd_ctl_open failed for %s: %s", 
				FL, mix->dev_name, snd_strerror(err));
	    return False;
	}
	err_str = "snd_ctl_card_info";
	err = snd_ctl_card_info (ctl_handle, hw_info);
	snd_ctl_close (ctl_handle);
	if (err < 0) goto xerr;

	ppcchange(&mix->info_name, snd_ctl_card_info_get_name(hw_info));
	ppcchange(&mix->info_mixername, snd_ctl_card_info_get_mixername(hw_info));
// printf("%s  %s\n", mix->info_name, mix->info_mixername);

	err_str = "snd_mixer_open";
	if ((err = snd_mixer_open (&pv->mixer_handle, 0)) < 0) goto xerr;
	err_str = "snd_mixer_attach";
	if ((err = snd_mixer_attach (pv->mixer_handle, mix->dev_name)) < 0) goto xerr;
	err_str = "snd_mixer_selem_register";
	if ((err = snd_mixer_selem_register (pv->mixer_handle, NULL, NULL)) < 0) goto xerr;

	snd_mixer_set_callback (pv->mixer_handle, mixer_event);
	snd_mixer_set_callback_private (pv->mixer_handle, mix);
// printf("mix=0x%x\n", (int)mix);

	err_str = "snd_mixer_load";
	if ((err = snd_mixer_load (pv->mixer_handle)) < 0) goto xerr;

	pv->sid_size = snd_mixer_selem_id_sizeof();

	return mixer_reinit(mix);

xerr:
	if (ctl_handle) snd_ctl_close (ctl_handle);
	appError(EMSG_ERR1, "%s(%d): function %s failed for %s: %s", 
			    FL, err_str, mix->dev_name, snd_strerror(err));
	snd_ctl_card_info_free(hw_info);
	if (pv->mixer_handle) snd_mixer_close(pv->mixer_handle);
	pv->mixer_handle = NULL;
	return False;
}

/*_____________________________ ChkType _________________________________*/
bool Sound_MIXER_ChkType(TSound_MIXER *mix) {
	if (mix && (mix->wtyp == wtyp_TSound_MIXER)) return True;

	if (!mix) {
	    appError(EMSG_ERR1, "%s(%d): TSound_MIXER = NULL", FL);
	} else if (mix->wtyp != wtyp_TSound_IO) {
	    appError(EMSG_ERR1, "%s(%d): Class not TSound_MIXER (0x%x)", FL, mix->wtyp);
	}
	return False;
}
/*_____________________________ Close ___________________________________*/
/* Tmix_elem *TSound_AlsaMIX_Get_elem(TSound_MIXER *mix, int idx) {
	if ((!Sound_MIXER_ChkType(mix)) || 
	    !(mix->stat & iost_init) || !mix->mix_elem ||
	    (idx >= mix->mixer_n_elems) || (idx < 0)) return NULL;
	return mix->mix_elem[idx];
} */
/*_____________________________ Close ___________________________________*/
void TSound_AlsaMIX_Close(TSound_MIXER *mix) {
	if (!Sound_MIXER_ChkType(mix)) return;
	close_mixer(mix);
}
/*_____________________________ Free ____________________________________*/
void TSound_AlsaMIX_Free(TSound_MIXER *mix) {
//  int i;
//  TSound_AlsaMIX_private *pv;

	if (!Sound_MIXER_ChkType(mix)) return;
	close_mixer(mix);

	xfree(mix->info_name);
	xfree(mix->info_mixername);

	xfree(mix->private);
	xfree(mix);
}

/*_____________________________ Create __________________________________*/
TSound_MIXER *TSound_AlsaMIX_Create(int fu) {
  TSound_MIXER *mix;
  TSound_AlsaMIX_private *pv;

	mix = xcalloc(sizeof(TSound_MIXER));
	pv = xcalloc(sizeof(TSound_AlsaMIX_private));
	mix->private = pv;
	mix->wtyp = wtyp_TSound_MIXER;
	mix->art = MIXART_ALSA;
	mix->pid = getpid();
	mix->fu = fu;
	mix->dev_name = pcdup("default");
	mix->free = TSound_AlsaMIX_Free;
	mix->init = TSound_AlsaMIX_Init;
	return mix;
}
