/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#ifndef __SOUND_MIXER_H__
#define __SOUND_MIXER_H__    1

#define MIXART_ALSA		1

#define MIX_ITEMNAME_SIZE	24

/** Mixer simple element channel identifier */
enum {
	C_ID_UNKNOWN = -1,	/** Unknown */
	C_ID_FRONT_LEFT = 0,	/** Front left */
	C_ID_FRONT_RIGHT,	/** Front right */
	C_ID_REAR_LEFT,		/** Rear left */
	C_ID_REAR_RIGHT,	/** Rear right */
	C_ID_FRONT_CENTER,	/** Front center */
	C_ID_WOOFER,		/** Woofer */
	C_ID_SIDE_LEFT,		/** Side Left */
	C_ID_SIDE_RIGHT,	/** Side Right */
	C_ID_REAR_CENTER,	/** Rear Center */
	C_ID_LAST = 31,
	C_ID_MONO = C_ID_FRONT_LEFT	/** Mono (Front left alias) */
};

/* mixer split types */
enum {
  MIXER_ELEM_FRONT, MIXER_ELEM_REAR,
  MIXER_ELEM_CENTER, MIXER_ELEM_WOOFER,
  MIXER_ELEM_SIDE,
  MIXER_ELEM_CAPTURE,
  MIXER_ELEM_ENUM, MIXER_ELEM_CAPTURE_ENUM,
  MIXER_ELEM_END
};

/* ********* mixer_type ********* */
#define MIXER_ELEM_TYPE_MASK		0xff
#define MIXER_ELEM_CAPTURE_SWITCH	0x100 /* bit */
#define MIXER_ELEM_MUTE_SWITCH		0x200 /* bit */
#define MIXER_ELEM_CAPTURE_SUFFIX	0x400
#define MIXER_ELEM_HAS_VOLUME		0x800

#define MIX_ELEM_Change		1

#define VST_VMAX		1
#define VST_VMDB	(1<<1)


typedef struct {
  int	channels;
  float	vol_left;
  float	vol_right;
  float	dB_left;
  float	dB_right;
  float vmin;
  float vmax;
  float vCount;
  float dB_vmin;
  float dB_vmax;
  int	mute;	//b0=left b1=right
  int	vst;

  int	io_client_type;
  void	*io_client;
  void	*obj;
  int	mixer_type;
  int	mixer_grpidx;
  int	c_id_left;
  int	c_id_right;

  int	st;	//b0=Change
  int	idx;
  char	item_name[MIX_ITEMNAME_SIZE];
  unsigned int	enum_nr;	// Mic, Int Mic, CD, ...
  char	enum_name[MIX_ITEMNAME_SIZE];
} Tmix_elem;

typedef struct TSound_MIXER {
  int	wtyp;
  int	art;
  void	*obj;
  int	fu;
  int	stat;
  int	pid;
  char	*dev_name;
  int	mixer_n_elems;
  Tmix_elem **mix_elem;
  void	(*free)(struct TSound_MIXER *mix);
  int	(*init)(struct TSound_MIXER *mix);
  int	(*close)(struct TSound_MIXER *mix);
  void	(*OnElemAdd)(struct TSound_MIXER *mix, Tmix_elem *mix_elem);
  void	(*OnElemDel)(struct TSound_MIXER *mix, Tmix_elem *mix_elem);
  void	(*OnElemChange)(struct TSound_MIXER *mix, Tmix_elem *mix_elem);
  char	*info_name;
  char	*info_mixername;
  void	*private;
} TSound_MIXER;

TSound_MIXER *TSound_AlsaMIX_Create(int fu);
void TSound_AlsaMIX_Free(TSound_MIXER *mix);
void TSound_AlsaMIX_Close(TSound_MIXER *mix);
bool Sound_MIXER_ChkType(TSound_MIXER *mix);

void TSound_AlsaMIX_SetVol(TSound_MIXER *mix, int elem_index, float vol_left, float vol_right);
void TSound_AlsaMIX_SetMute(TSound_MIXER *mix, int elem_index, int mute);

//Tmix_elem *TSound_AlsaMIX_Get_elem(TSound_MIXER *mix, int idx);

#endif /* __SOUND_MIXER_H__ */
