/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldconsole.h"
#include "cicons.h"
#include "xtty.h"
#include <unistd.h>

static int console_DoKeyDown(TLineEdit *self, TKeyEvent *KeyEv) {
  TFormAHconsole *fmain = appDat.obj;
  TLineEdit *le = fmain->le;
  Tpty *pty = fmain->pty;
  int m;
  TKeyS KeyS;
  bool ChkBottom = True;

	KeyS.d[0] = 0;
	if (pty->ptyfd <= 0) return 0;

	if (KeyEv->state & 0x4) {

	    if ((KeyEv->keysym >= 0x40) && (KeyEv->keysym < 0x80)) {
		KeyS.d[0] = KeyEv->keysym & 0x1f;
		ChkBottom = KeyS.d[0] != 0;
	    }

	    switch(KeyEv->keycode) {
	    case 54:		//Strg_C
		if (!chk_sh_slave(fmain, NULL)) {
		    TLineEditBase_Strg_C((TLineEditBase*)le);
		    return 1;
		}
		break;
	    case 55:		//Strg_V
		if (!chk_sh_slave(fmain, NULL)) {
		    if (!GetPasteStr_lgs(le->vds.UTF8on)) return 1;
		    write(pty->ptyfd, appXDat.SelRet_lgs.s, appXDat.SelRet_lgs.lg);
		    SelRetClr();
		    KeyS.d[0] = 0;
		}
		break;
	    }

	} else if (!xktotty(KeyEv->keysym, pty->term_type, &KeyS)) {
	    keysymtostr(&KeyEv->keysym, KeyS.c, fmain->leterm->UTF8on);
	    ChkBottom = KeyS.d[0] != 0;
	}

	if (KeyS.d[0]) write(pty->ptyfd, KeyS.c, pclen(KeyS.c));

	if (ChkBottom) {
	    m = le->tls->Count - le->VisibleLines;
	    if ((m > 0) && (m > le->VSB->Count)) {
		TScrollBar_SetCount_ev(le->VSB, m);
	    }
	}

	return 1;
}

static void testle_OnSetCaret(TLineEdit *le, TColS *cs) {
  TFormAHconsole *fmain = appDat.obj;
  int x = le->CaretCharIdx;
  int y = le->CaretCharIdy;
  int c;
  char tmps[40];

	if (cs) {
	    if (cs->s && (cs->lg > x)) {
		if (fmain->leterm->UTF8on) {
		    c = Utf8CharToWChar(cs->s + x);
		    if (c & Char_ErrorUTF8) c = (byte)cs->s[x];
		} else {
		    c = (byte)cs->s[x];
		}
		if (c >= 1000) {
		    snprintf(tmps, 40, "x=%d y=%d c=%d/x%02x", x, y, c, c);
		} else {
		    snprintf(tmps, 40, "x=%d y=%d  c=%d/0x%02x", x, y, c, c);
		}
	    } else {
		snprintf(tmps, 40, "x=%d y=%d  EOL", x, y);
	    }
	} else {
	    snprintf(tmps, 40, "x=%d y=%d  EOF", x, y);
	}
	SetCaption(fmain->CharAzBox, tmps);
}

#if DEBUG_LE
static void AzLine(char *line, int lg) {
  TFormAHconsole *fmain = appDat.obj;
#if DEBUG_LE_LINE1
  TColS *cs;
  char *pc;
  char buf[4096];

	cs = TLines_Get_pcs(fmain->testle->tls, 0);
	if (!cs) return;
	cs->lg = lg;
	ppcnchange(&cs->s, line, cs->lg);

	cs = TLines_Get_pcs(fmain->testle->tls, 1);
	if (!cs) return;
	pc = buf;
	while (*line) {
	    ppcmove_hex(&pc, *line, kCOMMA_ADD | 2);
	    line++;
	}
	cs->lg = pclen(buf);
	ppcnchange(&cs->s, buf, cs->lg);
	TLineEdit_AllAtrNew(fmain->testle);
#else
  Tlgs lgs;
	lgs.s = line;
	lgs.lg = lg;
	TLineEdit_AddLine_lgs(fmain->testle, &lgs);
#endif
}
#endif

static void pty_OnInPipe(int fd, Tfd_list *fdl) {
  TFormAHconsole *fmain = appDat.obj;
  int lg;
  char buf[1024];

	lg = read(fd, buf, sizeof(buf)-1);
	if (lg <= 0) {
	    Close();
	} else {
	    buf[lg] = 0;
#if DEBUG_LE
 AzLine(buf, lg);
#endif
	    leAddChar(fmain, buf);
	}
}

static void PanelTop_OnResize(TWins *self) {
  TFormAHconsole *fmain = appDat.obj;
  int x;

	x = self->width - fmain->CharAzBox->width - 2;
	SetLeft(fmain->CharAzBox, x);
	x = fmain->mbHelp->left + fmain->mbHelp->width + 4;
	SetLeft(fmain->SizeAzBox, x);
}

void SetLeTermRect(TFormAHconsole *fmain, int fu) {
  int lg, w, h;
  TLineEdit *le = fmain->le;
  char line[256];

	FillChar(line, sizeof(line), ' ');
	lg = fmain->pty->wsize.ws_col;
	if (lg > sizeof(line)) lg = sizeof(line);
	le->vds.lg = lg;
	le->vds.s = line;
	w = TextWidthColorString(&le->vds);
	le->MaxStrX = w;

	le->vds.lg = 0;
	le->vds.s = NULL;
	h = le->vds.Font->height * fmain->pty->wsize.ws_row;

	w += le->VSB->width + 4;
	h += le->HSB->height + fmain->PanelTop->height+ 2;
#if DEBUG_LE
	h += fmain->testle->pw->height + fmain->Splitter->height;
#endif
	fmain->o_width = w;
	fmain->o_height = h;
	if ((fu & 0x01) || (fmain->SizeAzBox->BorderFu & mBorderL)) {
	    SetWinSize((TWins*)fmain, w, h);
	}
}

static void SizeAzBox_OnClick(TWins* self, long KeyState, bool DblClick) {
  TFormAHconsole *fmain = appDat.obj;
	SetLeTermRect(fmain, 0x01);
}

static void fmain_OnResize(TFormAHconsole* self) {

	if ((self->o_width == self->width) && (self->o_height == self->height)) {
	    self->SizeAzBox->BorderFu |= mBorderL;
	} else {
	    self->SizeAzBox->BorderFu &= ~mBorderL;
	}
	Refresh(self->SizeAzBox);
}
 
static void OnTimer1000ms(TgTimer* gTimer) {
  TFormAHconsole *fmain = appDat.obj;
  Tpty *pty = fmain->pty;
  int m;
  char *pc;
  const char *cap;
  char bufcwd[32];
  char buf[1024];

	cap = "Console";
	if (pty->TermPid > 0) {
	    if (pty->path_deleted) {
		if (TestIsPath(pty->cwd)) {
		    Tpty_close(pty);
		    fmain->leterm->idx = 0;
		    fmain->leterm->lg = 0;
		    Tpty_open(pty);
//		    gTimer_set_next_ev(gTimer, 50);
		}
		return;
	    }

	    pc = bufcwd;
	    ppcmove(&pc, "/proc/");
	    ppcmove_v32(&pc, pty->TermPid, 0);
	    ppcmove(&pc, "/cwd");
	    m = readlink(bufcwd, buf, sizeof(buf)-1);
	    if ((m > 0) && (m < sizeof(buf))) {
		buf[m] = 0;
		cap = buf;
		if (!pccmp(cap, pty->cwd)) {
		    if (pcfind(cap, " (deleted)", NULL)) {
			pty->path_deleted = 
				(fmain->miAutoReconnect->fu & miChecked) != 0;
		    } else {
			ppcchange(&pty->cwd, cap);
		    }
		}
	    }
	}
	SetCaption((TWins*)fmain, cap);
}

static char *exec_command;

static void OnAppStart() {
  TFormAHconsole *fmain = appDat.obj;
  Tpty *pty = fmain->pty;

	ChkTermUTF8(fmain, True);

	ppTfree(&appDat.appIni);
	if (exec_command) {
	    AppHandleMessage();
	    if (pty && (pty->ptyfd > 0)) {
		write(pty->ptyfd, exec_command, pclen(exec_command));
		write(pty->ptyfd, "\n", 1);
	    }
	}

	fmain->t1000ms = gTimerX_Add(Timer_Enabled, 1000, OnTimer1000ms);
}

static const char *DefaultIniDat = {
  "[AHconsole]\n"
  "MouseZ_Wheel=4\n"
  "col=80\n"
  "row=24\n"
  "Font=-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso8859-1\n"
  "FontUTF8=-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso10646-1\n"
  "MaxLines=2000\n"
};

static const char *Help_Msg = {
  "\n"
  "Usage: aldconsole [options] [workdir]\n"
  "\n"
  "  -e <command>         Execute \"command\"\n"
  "  --workdir <dir>      Change to \"dir\"\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: aldconsole version "VERSION"  (build %d  %s)\n"
};

const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;

int main (int argc, char **argv) {
  TFormAHconsole *fmain;
  TpcIni *ini;
  int i;
  char *val, *workdir;
  bool help = False;
  bool err = False;

	workdir = exec_command = NULL;
	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'e':
		    err = cmd_get_pc(argc, argv, val, &i, &exec_command);
		    break;
		case 'h':
		    help = True; break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
		    } else if (pcncmp(val, "workdir", 7)) {
			err = cmd_get_pc(argc, argv, val+6, &i, &workdir);

#ifdef Test_Root_WinId
		    } else if (pcncmp(val, "TestRootWinID", 13)) {
			i++;
			val = argv[i];
			if (val && *val != '-') {
				TestRootWinId = pctoint(val);
			} else err = True;
#endif
		    } else err = True;
		    break;
		default: 
			err = True;
		}  // switch end
	    } else {
		if (workdir) err = True; else workdir = val;
	    }
	    i++;
	}

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(EXIT_FAILURE);
	}


	initXapp(0, argc, argv);
	fmain = TForm_CreatParamAlg(NULL, sizeof(TFormAHconsole));
	appDat.obj = fmain;
#ifdef USE_DECOR
	appXDat.DefaultButtonDecor = TDecor_button_w3b3_Create();
#endif
#ifdef _OS_WIN32
	appDat.inifile_app = pcadd(appDat.ALDini_dir, "\\console.ini");
#else
	appDat.inifile_app = pcadd(appDat.ALDini_dir, "/console.ini");
	fmain->Icon = GetDeskIcon_z(ddkt_Application, 16, APPNAME, 
				    gdifu_noZoom, zico16console);
#endif
	setDefaultIniDat(DefaultIniDat);
	
	if (workdir && test_local_drive(workdir)) {
	    fmain->cwd = path_to_local(workdir);
	    while (fmain->cwd && !TestIsPath(fmain->cwd)) {
		path_sub_ppc(&fmain->cwd);
	    }
	}
	if (!fmain->cwd) fmain->cwd = xgetcwd(False);
	chdir("/");

	appDat.appIni = TpcIni_Create(appDat.inifile_app);
	ini = appDat.appIni;

	fmain->height = 400;
	fmain->width = 300;
	fmain->min_width = 380;
	fmain->min_height = 80;

	SetCaption((TWins*)fmain, "Console");
	fmain->IconName = APPNAME;
	fmain->OnResize = fmain_OnResize;

	fmain->PanelTop = TWins_CreateBA((TWins*)fmain, 0, 0, 200, 21, mBorder | mBorderL, alTop);
	fmain->PanelTop->OnResize = PanelTop_OnResize;

#if DEBUG_LE
	fmain->testle = TLineEdit_CreateA((TWins*)fmain, 0, 0, 200, 100, alTop);
	SetFont((TWins*)fmain->testle, TpcIni_RD_pc(ini, APPNAME, "Font", 
		"-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso8859-1"));
	fmain->testle->OnSetCaret = testle_OnSetCaret;
#if DEBUG_LE_LINE1
	TLineEdit_AddLine(fmain->testle, " ");
	TLineEdit_AddLine(fmain->testle, " ");
#endif
	fmain->Splitter = TSplitterH_CreateP((TWins*)fmain);
#endif
	fmain->CharAzBox = TWins_CreateLabel(fmain->PanelTop, 182, 3, 150, 16, "255/FF", 0, mBorder | mBorderL);
	TLineEdit_AddKeyDownPlugIn(NULL, TLineEdit_keyplugin_cursor_Create());

	fmain->le = TLineEdit_CreateA((TWins*)fmain, 0, 0, 200, 200, alClient);
	SetFont((TWins*)fmain->le, TpcIni_RD_pc(ini, APPNAME, "Font", 
		"-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso8859-1"));
	fmain->le->LEfu = LEfu_VSBnoVisibleLines;
	fmain->le->OnSetCaret = testle_OnSetCaret;
	fmain->le->DoKeyDown = console_DoKeyDown;
	SetHeight((TWins*)fmain->le->HSB, 10);

	MenuFile_Create(fmain);
	MenuEdit_Create(fmain);
	MenuSetting_Create(fmain);
	MenuHelp_Create(fmain);
	fmain->SizeAzBox = TWins_CreateLabel(fmain->PanelTop, 182, 3, 76, 16, NULL, 0, mBorder | mBorderL);
	fmain->SizeAzBox->OnClick = SizeAzBox_OnClick;

	fmain->le->PopupMenu = fmain->menuEdit;

  int LC_xxxx = 0; //0x01=LC_ALL 0x02=LC_CTYPE
	val = getenv("LC_ALL");
	if (!val || !*val) val = getenv("LC_CTYPE"); else LC_xxxx |= 0x01;
	if (!val || !*val) val = getenv("LANG"); else LC_xxxx |= 0x02;
	if (val) {
	    fmain->LANG = pcdup(val);
	    while (*val && (*val != '.')) val++;
	    if (*val) val++;
	    fmain->LANG_utf8 = pccasecmpASCII(val, "UTF-8") == 0;
	} else {
	    fmain->miUTF8_en_US->fu |= (miChecked | miNotEnabled);
	}

	fmain->pty = Tpty_Create();
	fmain->pty->LC_xxxx = LC_xxxx;
	fmain->pty->OnInPipe = pty_OnInPipe;
//	fmain->pty->term_type = term_xterm;
	fmain->pty->term_type = term_linux;
	fmain->pty->wsize.ws_col = TpcIni_RD_int(ini, APPNAME, "col", 80);
	fmain->pty->wsize.ws_row = TpcIni_RD_int(ini, APPNAME, "row", 24);
	fmain->pty->cwd = pcdup(fmain->cwd);

	fmain->leterm = Tleterm_Create(fmain->le);
	fmain->leterm->ws_col = fmain->pty->wsize.ws_col;
	fmain->leterm->ws_row = fmain->pty->wsize.ws_row;
	fmain->leterm->term_type = fmain->pty->term_type;

	SetTermSize(fmain, 0, 0);
	SetLeTermRect(fmain, 0x01);

	Set_Focus((TWins*)fmain->le);
	appXDat.OnAppStart = OnAppStart;

	AppRun();

	return 0;
}
