/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appXbase.h"
#include "tlineedit.h"

#include <fcntl.h>
#include <pty.h>

#define DEBUG_LE 0
#define DEBUG_LE_LINE1 0


#define APPNAME "aldconsole"

extern const int build_v;
extern const char *build_date;

/* ******* pty.c ****** */
typedef struct {
  AlgTTC
  int TermPid;
  int ptyfd;
  int idx;	//line idx
  int max;	//max line size
  char *line;
  void (*OnInPipe)(int fd, Tfd_list *fdl);
  struct winsize wsize;
  int term_type;
  char *LANG;
  char *cwd;
  bool path_deleted;
  bool ncb1, ncb2, ncb3;
  int LC_xxxx; //0x01=LC_ALL 0x02=LC_CTYPE
  char tty_name[512];
} Tpty;


Tpty *Tpty_Create();
bool Tpty_open(Tpty *self);
void Tpty_close(Tpty *self);
void Tpty_SetSize(Tpty *self, int col, int row);

#define ctfu_ColReverse 1
#define ctfu_FontBold	(1<<1)

#define exbef_cd	1

typedef struct {
  byte bc;	//BackColor
  byte fc;	//ForeColor
  byte fu;
  byte nc2;
} TCharStat;

typedef union {
  int i;
  TCharStat t;
} TUCharStat;

typedef struct {
  AlgTTC0
  dword res[16];
  int reslg;
  int btty_XK;
  int state;
  int laststate;
  int nums[8];
  int numIdx;
  int numCount;
  int time_ms;

  TLineEdit *le;
  TColS *cs;
  word *line;
  char *pcline;
  TUCharStat *uct;
  TStrColorDatArr *scda;
  int linesize;
// ------------------------------------------------------------------------
  int idx;		//ClearALL Anfang  (Tleterm_gotpars   case 'J':)
  int lg;		//  ^
  int idy;		//  |

  bool err;
  bool ques;

  bool disp_ctrl;
  bool charset;
  bool decawm;
  bool alternate_screen;
  bool insert_mod;
  bool hide_cursor;
  bool region_on;
  bool region_on_lf;
  bool region_is_cls;
  bool region_sector;
  bool ClearALL;

  int region_top;	//  |
  int region_bottom;	//  v
  TUCharStat Akt_uct;	//ClearALL Ende
// ------------------------------------------------------------------------

  TFont *AktFont;
  TFont *AktFontBold;

  TFont *Font;
  TFont *FontBold;
  TFont *FontUTF8;
  TFont *FontBoldUTF8;

  int ws_col, ws_row;

  int save_idx, save_idy;
  TUCharStat save_uct;
  bool save_disp_ctrl;
  bool save_Akt;
  bool FontHasUTF8;
  bool UTF8on;

  int term_type;
  int max_lines;
  int ex_bef;

} Tleterm;

Tleterm *Tleterm_Create(TLineEdit *le);
bool Tleterm_scan(Tleterm *term, dword c);
void Tleterm_SetSize(Tleterm *term, int col, int row);
void Tleterm_Clear(Tleterm *term, bool all);


#define TCSelf TFormAHconsole
typedef struct TFormAHconsole {
  AlgTForm
  TWins *PanelTop;
  TMenuButton *mbFile;
    TMenu *menuFile;
  TMenuButton *mbEdit;
    TMenu *menuEdit;
  TMenuButton *mbSetting;
    TMenu *menuSetting;
      TMenu *menuTermSize;
	TMenu *menuTermSizeRow;
	TMenu *menuTermSizeCol;
      TMenu *menuClipbord;
      TMenu *menuFont;
    TMenuItem *miUTF8;
    TMenuItem *miUTF8_en_US;
    TMenuItem *miAutoReconnect;

  TMenuButton *mbHelp;
    TMenu *menuHelp;

  TWins *SizeAzBox;
  TWins *CharAzBox;


  TLineEdit *le;

  Tpty *pty;
  Tleterm *leterm;
#if DEBUG_LE
  TLineEdit *testle;
  TSplitterH *Splitter;
#endif

  int o_width;
  int o_height;

  TgTimer *t1000ms;

  char *exec_command;
  char *LANG;
  char *cwd;
  bool LANG_utf8, nc_1, nc_2, nc_3;

} TFormAHconsole;
#undef TCSelf

void SetLeTermRect(TFormAHconsole *fmain, int fu);

/* Menu */
void MenuFile_Create(TFormAHconsole *fmain);

void MenuEdit_Create(TFormAHconsole *fmain);

void MenuSetting_Create(TFormAHconsole *fmain);

void SetTermSize(TFormAHconsole *fmain, int new_col, int new_row);
void ChkTermUTF8(TFormAHconsole *fmain, bool force_open);

void MenuHelp_Create(TFormAHconsole *fmain);


bool xktotty(int XK, int id, TKeyS *KeyS);


void leAddChar(TFormAHconsole *fmain, const char *pc0);

/* ****** le_color.c ****** */
void Tleterm_add_char(Tleterm *term, int c);
void Tleterm_set_CS_color(Tleterm *term);
void Tleterm_CS_color_to_line(Tleterm *term, TColS *cs);
void Tleterm_set_CS_data(Tleterm *term, TStrColorDat *scd, TUCharStat uct);

/* ****** pid.c ****** */
typedef struct TPidStat {
  struct TPidStat *next;
  struct TPidStat *next_child;
  char *name;
  int pid;
  int ppid;
} TPidStat;

bool chk_sh_slave(TFormAHconsole *fmain, TPidStat **pidst);
