/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldconsole.h"

void Tleterm_add_char(Tleterm *term, int c) {
	if (term->idx >= (term->linesize-1)) return;

	if (term->disp_ctrl) {
	    if (term->FontHasUTF8) {
//CharSet to Font=-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso10646-1
		switch (c) {
		case  24: c = '^'; break;
		case  25: c = 'v';break;
		case 177: c = 0x2592; break;
		case 179: c = 0x2502; break;
		case 180: c = 0x2524; break;
		case 191: c = 0x2510; break;
		case 192: c = 0x2514; break;
		case 195: c = 0x251c; break;
		case 196: c = 0x2500; break;
		case 217: c = 0x2518; break;
		case 218: c = 0x250c; break;
		}
	    } else {
//CharSet to Font=-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso8859-1
		switch (c) {
		case  24: c = 94; break;
		case  25: c = 118;break;
		case 177: c = 2; break;
		case 179: c = 25; break;
		case 180: c = 22; break;
		case 191: c = 12; break;
		case 192: c = 14; break;
		case 195: c = 21; break;
		case 196: c = 18; break;
		case 217: c = 11; break;
		case 218: c = 13; break;
		} 
	    }
	} else if (term->charset) {
	    if (term->FontHasUTF8) {
		switch (c) {
		case 'a': c = 0x2592; break;
		case 106: c = 0x2518; break;
		case 107: c = 0x2510; break;
		case 108: c = 0x250c; break;
		case 109: c = 0x2514; break;
		case 113: c = 0x2500; break;
		case 116: c = 0x251c; break;
		case 117: c = 0x2524; break;
		case 120: c = 0x2502; break;
		}
	    } else {
		switch (c) {
		case 'a': c = 2;  break;
		case 106: c = 11; break;
		case 107: c = 12; break;
		case 108: c = 13; break;
		case 109: c = 14; break;
		case 113: c = 18; break;
		case 116: c = 21; break;
		case 117: c = 22; break;
		case 120: c = 25; break;
		}
	    }
	}

	if (term->insert_mod && (term->idx < term->lg) && (term->lg < (term->linesize - 2))) {
	    int n = term->lg - term->idx;
	    pnmove(&term->line[term->idx + 1], &term->line[term->idx], n * 2);
	    pnmove(&term->uct[term->idx + 1], &term->uct[term->idx], n * sizeof(TUCharStat));
	    if (!term->region_on) term->lg++;
	}

	term->line[term->idx] = c;
	term->uct[term->idx].i = term->Akt_uct.i;
	term->idx++;
	if (term->idx > term->lg) {
	    term->lg = term->idx;
	    term->line[term->idx] = ' ';
	}
}

static const int tab_color[16] = {
  0x000000,0xcd0000,0x00cd00,0xcdcd00,0x0000cd,0xcd00cd,0x00cdcd,0xcdcdcd,
  0x808080,0xff0000,0x00ff00,0xffff00,0x5c5cff,0xff00ff,0x00ffff,0xffffff
};

void Tleterm_set_CS_data(Tleterm *term, TStrColorDat *scd, TUCharStat uct) {
	scd->Font = term->AktFont;
	if (uct.t.fu & ctfu_FontBold) scd->Font = term->AktFontBold;
	if (uct.t.fu & ctfu_ColReverse) {
	    scd->BackColor = tab_color[uct.t.fc & 0x0f];
	    scd->ForeColor = tab_color[15-(uct.t.bc & 0x0f)];
	} else {
	    scd->ForeColor = tab_color[uct.t.fc & 0x0f];
	    scd->BackColor = tab_color[15-(uct.t.bc & 0x0f)];
	}
}

void Tleterm_set_CS_color(Tleterm *term) {
  int ucti, stk, lg, UCfu, scdfu;
  char *pc;
  TUCharStat *uct;
  TStrColorDatArr *scda;
  TStrColorDat *scd;
  bool mk;

	UCfu = UC_force_ASCII;
	lg = term->lg;
	if (term->UTF8on) {
	    lg = WCharsToUtf8(NULL, 0, term->line, term->lg, 0);
	    UCfu = 0;
	}
	TColS *cs = term->cs;
	cs->s = xmalloc(lg+1);
	WCharsToUtf8(cs->s, lg+1, term->line, term->lg, UCfu);
	cs->lg = lg;
	cs->c = NULL;
	if (!lg) return;

	stk = 0;
	ucti = -1;
	uct = term->uct;
	for (int i=term->lg;i>0;i--) {
	    if (ucti != uct->i) {ucti = uct->i; stk++;}
	    uct++;
	}

	scdfu = scdfu_override_utf8;
	if (term->UTF8on) scdfu = scdfu_override_utf8 | scdfu_UTF8on;

	scda = set_scda_arr((void*)&cs->c, stk);
	cs->c = (char*)scda;
	scd = scda->scd;
	ucti = -1;
	uct = term->uct;
	pc = cs->s;
	mk = False;
	for (int i=term->lg;i>0;i--) {
	    if (ucti != uct->i) {
		ucti = uct->i;
		if (mk) {
		    scd->lg = pc - scd->s;
		    scd->fu = scdfu;
		    scd++;
		}
		mk = True;
		scd->s = pc;
		Tleterm_set_CS_data(term, scd, *uct);
	    }
	    if (term->UTF8on) pc += get_Utf8Chars(pc); else pc++;
	    uct++;
	}
	if (mk) {
	    scd->lg = pc - scd->s;
	    scd->fu = scdfu;
	}
}

void Tleterm_CS_color_to_line(Tleterm *term, TColS *cs) {
  int items, lg, m, clg, UCfu;
  TUCharStat uct, *puct;
  TStrColorDatArr *scda;
  TStrColorDat *scd = NULL;
  char *pc;

	term->idx = 0;
	term->lg = 0;
	Fill0(term->uct, term->linesize * sizeof(TUCharStat));
	if (!cs) return;
	if (cs->s == term->pcline) return;
	if (cs->c == (char*)term->scda) return;

	term->cs = cs;

	lg = cs->lg;
	UCfu = UC_force_ASCII;
	if (term->UTF8on) {
	    lg = Utf8ToWChars(NULL, 0, cs->s, cs->lg, 0);
	    UCfu = 0;
	}
	Utf8ToWChars(term->line, term->linesize, cs->s, cs->lg, UCfu);

	if (lg >= term->linesize) lg = term->linesize-1;

	term->line[lg] = ' ';
	term->lg = lg;

	scda = (TStrColorDatArr*)cs->c;
	if (scda && scda->items) {
	    scd = scda->scd;
	    items = scda->items;

	    puct = term->uct;
	    uct.i = 0;
	    pc = cs->s;
	    clg = 0;
	    for (int i=cs->lg;i>0;i--) {
		if (clg > 0) {
		    clg--;
		    if (!clg) uct.i = 0;
		}
		if (items && (pc == scd->s)) {
		    clg = scd->lg;
		    uct.i = 0;
		    if (scd->Font == term->AktFontBold) uct.t.fu |= ctfu_FontBold;
		    m = scd->ForeColor;
		    for (int x=0;x<16;x++) {
			if (tab_color[x] == m) {uct.t.fc = x; break;}
		    }
		    m = scd->BackColor;
		    for (int x=0;x<16;x++) {
			if (tab_color[x] == m) {uct.t.bc = 15-x; break;}
		    }
		    items--;
		    scd++;
		}
		puct->i = uct.i;
		if (term->UTF8on) pc += get_Utf8Chars(pc); else pc++;
		puct++;
	    }
	}

	if (cs->s && (cs->s != term->pcline)) xfree(cs->s);
	cs->s = NULL;
	if (cs->c && (cs->c != (char*)term->scda)) xfree(cs->c);
	cs->c = NULL;
}
