/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldconsole.h"
#include "xtty.h"

#define ttyscn_DEBUG	1


#ifdef	ttyscn_DEBUG
static void tty_err(Tleterm *term, byte c) {
  int i;
	fprintf(stderr, "Error ttyscan: l=%d %d \"%02x",term->reslg, term->laststate, term->res[0]);
	for (i=1; i<term->reslg; i++) {
	    fprintf(stderr, " %02x", term->res[i]);
	}
	fprintf(stderr, "\"\n\n");
}
#endif

#define OV_DELLINES 16
static TColS *add_cs(Tleterm *term, bool WholeDataLine) {
  int m, n, count;
  TLineEdit *le = term->le;
  TLines *tls = le->tls;  
  TColS *cs;
  bool SBV_Count;

	n = tls->Count - le->VisibleLines;
	SBV_Count = (n < 0) || (n == le->VSB->Count);

	cs = TLines_Add_(tls);
	if (WholeDataLine) {
	    cs->s = xmalloc(term->ws_col+1);
	    cs->lg = term->ws_col;
	    FillChar(cs->s, term->ws_col, ' ');
	}

	if (tls->Count > term->max_lines) {
	    term->idy -= OV_DELLINES;
	    le->SelAy -= OV_DELLINES;
	    le->CaretCharIdy -= OV_DELLINES;
	    if ((le->SelAy < 0) || (le->CaretCharIdy < 0)) {
		le->SelAx = le->CaretCharIdx = 0;
		le->SelAy = le->CaretCharIdy = 0;
	    }
	    TLines_DelN(tls, 0, OV_DELLINES);
	}

	count = tls->Count;
	m = le->VSB->Count;
	n = le->VisibleLines;
	if (SBV_Count && (count > (m + n))) {
	    le->VSB->MaxCount = count;
	    TScrollBar_SetCount(le->VSB, count - n);
	} else {
	    TScrollBar_SetMaxCount(le->VSB, count);
	}

	return cs;
}

static void set_line_idy(Tleterm *term, bool WholeDataLine, bool save_idx) {
  TColS *cs;
  int idx = term->idx;

	cs = TLines_Get_pcs(term->le->tls, term->idy);
	if (!cs) cs = add_cs(term, WholeDataLine);
	Tleterm_CS_color_to_line(term, cs);
	if (!save_idx) return;
	if (idx > term->lg) idx = term->lg;
	term->idx = idx;
}

#define NO_LINE_SET_CS 1
#define NO_CS_GET_LINE 2
//#define NO_CS_SAVE 3

static void tls_cs_move(Tleterm *term, int dest, int src, int lg, int fu) {
  TColS *scs, *dcs;
  TLines *tls = term->le->tls;
  int k;

	if ((lg <= 0) || (dest == src)) return;
	if (lg > tls->Count) lg = tls->Count;
	if (dest < src) {
	    if (lg > (tls->Count - dest)) lg = tls->Count - dest;
	    k = 1;
	} else {
	    if (lg > (tls->Count - src)) lg = tls->Count - src;
	    src += lg-1; dest += lg-1;
	    k = -1;
	}
	if (lg <= 0) return;

	if (!(fu & NO_LINE_SET_CS)) Tleterm_set_CS_color(term);
	for (int i=lg;i>0;i--) {
	    scs = TLines_Get_pcs(tls, src);
	    dcs = TLines_Get_pcs(tls, dest);
	    if (dcs) {
		dcs->lg = 0;
		xfree(dcs->s); dcs->s = NULL;
		xfree(dcs->c); dcs->c = NULL;
		if (scs) {
		    dcs->lg = scs->lg; scs->lg = 0;
		    dcs->s = scs->s;   scs->s = NULL;
		    dcs->c = scs->c;   scs->c = NULL;
		}
	    } else if (scs) {
		scs->lg = 0;
		xfree(scs->s); scs->s = NULL; 
		xfree(scs->c); scs->c = NULL;
	    }
	    src += k; dest += k;
	}
	if (!(fu & NO_CS_GET_LINE)) Tleterm_CS_color_to_line(term, TLines_Get_pcs(tls, term->idy));
}

static int get_top_row(Tleterm *term, int row_off) {
  int m;
	if (row_off < 1) row_off = 1;
	if (row_off > term->ws_row) row_off = term->ws_row;
	m = row_off + term->le->tls->Count - term->ws_row - 1;
	if (m < 0) m = 0;
	return m;
}

static void add_lf(Tleterm *term, bool save_idx) {
  int m, n;

	if (term->region_sector) {
	    m = get_top_row(term, term->region_bottom);
	    if (m == term->idy) {
	        n = get_top_row(term, term->region_top);
		m = term->region_bottom - term->region_top;
		tls_cs_move(term, n, n+1, m, NO_CS_GET_LINE);
		Tleterm_set_CS_color(term);
		return;
	    }
	}
	if (term->region_on && !term->region_on_lf) {
	    m = get_top_row(term, term->region_bottom);
	    if (m == term->idy) term->region_on_lf = true;
	}

	term->ClearALL = (term->lg == 7) && (term->line[2] == 'c') &&
					    (term->line[3] == 'l') &&
					    (term->line[4] == 'e') &&
					    (term->line[5] == 'a') &&
					    (term->line[6] == 'r');

	if ((term->lg > 5) && (term->line[0] == '>') &&
			      (term->line[1] == ' ') &&
			      (term->line[2] == 'c') &&
			      (term->line[3] == 'd') &&
			      (term->line[4] == ' ')) {
	    term->ex_bef = exbef_cd;
	}

	Tleterm_set_CS_color(term);

	term->idy++;
	set_line_idy(term, False, save_idx);
	TLines_DoCount(term->le->tls);
}

static void dec_x(Tleterm *term) {
	if (term->idx > 0) term->idx--;
}

static void Cls_line(Tleterm *term, int anf, int end) {
	if (anf > term->lg) anf = term->lg;
	if (end > term->lg) end = term->lg;
	while (anf <= end) {
	    term->line[anf] = ' ';
	    term->uct[anf].i = term->Akt_uct.i;
	    anf++;
	}
}

static void Cls_line_nr(Tleterm *term, int idy) {
  TColS *cs;
  TStrColorDat *scd;
	if (term->idy == idy) {
	    Cls_line(term, 0, term->lg);
	} else {
	    cs = TLines_Get_pcs(term->le->tls, idy);
	    if (!cs || !cs->lg) return;
	    FillChar(cs->s, cs->lg, ' ');
	    if (term->Akt_uct.i) {
		set_scda_arr((void*)&cs->c, 1);
		scd = ((TStrColorDatArr*)cs->c)->scd;
		scd->s = cs->s;
		scd->lg = cs->lg;
		Tleterm_set_CS_data(term, scd, term->Akt_uct);
	    } else {
		xfree(cs->c);
		cs->c = NULL;
	    }
	}
}

static void Set_line_nr(Tleterm *term, int idy, bool WholeDataLine, bool save_idx) {
	if (idy < 0) idy = 0;
	if (term->idy != idy) {
	    Tleterm_set_CS_color(term);
	    term->idy = idy;
	    set_line_idy(term, WholeDataLine, save_idx);
	}
}


static void Set_region(Tleterm *term, int anf, int end) {
  int m;
	if (anf <= 0) anf = 1;
	if (!end) end = term->ws_row;
	if (end > term->ws_row) end = term->ws_row;

	if (anf > end) return;
	term->region_top = anf;
	term->region_bottom = end;
	term->region_on_lf = false;

	if (!term->region_on) {

	    term->region_sector = False;
	    term->region_on = True;
	    if (term->lg < term->ws_col) {
		m = term->lg;
		term->lg = term->ws_col;
		Cls_line(term, m, term->ws_col);
	    }
	    m = end - (term->le->tls->Count - term->idy);
	    if (m > 0) {
		for (;m>0;m--) add_cs(term, True);
	    }
	    TLines_DoCount(term->le->tls);
// printf("%s(%d): Region ON\n", FL);
	} else {
	    term->region_sector = (anf > 1) && (end < term->ws_row);
	}
}


static void REGP1 csiICH(Tleterm *term) { //<n>@ Insert Characters
  int n;
	if (term->ques) {term->err = True; return;}

	if (term->nums[0] <= 0) return;
	n = term->lg - term->idx;
	if ((n > 0) && ((term->lg + term->nums[0]) < (term->linesize - 1))) {
	    pnmove(&term->line[term->idx + term->nums[0]], &term->line[term->idx], n * 2);
	    pnmove(&term->uct[term->idx + term->nums[0]], &term->uct[term->idx], n * sizeof(TUCharStat));
	    term->lg += term->nums[0];
	    term->line[term->lg] = ' ';
	}
}

static void REGP1 csiCUU(Tleterm *term) { //<n>A	cursor up n lines
	if (term->ques) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	Set_line_nr(term, term->idy - term->nums[0], False, True);
}

static void REGP1 csiCUD(Tleterm *term) { //<n>B	cursor down n lines
	if (term->ques) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	Set_line_nr(term, term->idy + term->nums[0], False, True);
}

static void REGP1 csiCUF(Tleterm *term) { //<n>C	cursor right
	if (term->ques) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	term->idx += term->nums[0];
	if (term->idx > term->lg) {
	    int i = term->idx - term->lg;
	    term->idx = term->lg;
	    for (;i>0;i--) Tleterm_add_char(term, ' ');
	}
}
static void REGP1 csiCUB(Tleterm *term) { //<n>D	cursor left
	if (term->ques) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	term->idx -= term->nums[0];
	if (term->idx < 0) term->idx = 0;
}

static void REGP1 csiHPA(Tleterm *term) { //<n>G position cursor at column n-1
  int m = term->nums[0];
	if (term->ques) {term->err = True; return;}
	if (m > 0) m--;
	if (m > (term->linesize-1)) m = term->linesize-2;
	while (m > term->lg) Tleterm_add_char(term, ' ');
	term->idx = m;
}

static void REGP1 csiCUP(Tleterm *term) { //<n>H  (cup)  cursor to line n and column m
				//<n>d	(vpa)	cursor to line n
  int m;
	if (term->ques) {term->err = True; return;}
	if (term->ClearALL) return;  //user console clear
	if (!term->region_on) {
	    Set_region(term, 0, 0);
	} else {
//	    if (!term->nums[0]) term->nums[0]++;
	    if (term->region_on_lf) term->nums[0]++;
	    m = get_top_row(term,  term->nums[0]);
	    Set_line_nr(term,  m, True, True);
	}
	m = term->nums[1];
	if (m > 0) m--;
	while (m > term->lg) Tleterm_add_char(term, ' ');
	term->idx = m;
}

static void REGP1 csiED(Tleterm *term) { //<n>J  (ed)	erase in display
  int m, n;
	if (term->ClearALL) {  //user console "clear"
	    Tleterm_Clear(term, True);
	    return;
	}

	switch (term->nums[0]) {
	case 0: //erase from cursor to end of display
	    term->region_is_cls = ((term->le->tls->Count - term->idy) == term->ws_row);

	    m = term->idy;
	    n = term->le->tls->Count;
	    while (m < n) {
		Cls_line_nr(term, m);
		m++;
	    }
	    break;
/*		case 1: //erase from start to cursor
// printf("%s(%d): gotpars: %d %d %d 0x%02x\n", FL, term->termart, term->numIdx, term->nums[0], c);
//		    Clr_line(term, 0, term->idx);
	    break; */
	case 2: //erase whole display
	    term->region_is_cls = True;
	    m = get_top_row(term, 1);
	    n = term->le->tls->Count;
	    while (m < n) {
		Cls_line_nr(term, m);
		m++;
	    }
	    break;

	default: 
// printf("%s(%d): gotpars J: %d %d\n", FL, term->numIdx, term->nums[0]);
	    term->err = True; 
	}
}

static void REGP1 csiEL(Tleterm *term) { //<n>K	(el)	erase in line
	switch (term->nums[0]) {
	case 0: //erase from cursor to end of line
	    if (term->region_on) {
		Cls_line(term, term->idx, term->lg);
	    } else {
		term->lg = term->idx;
		term->line[term->lg] = ' ';
	    }
	    break;
	case 1: //erase from start of line to cursor
	    Cls_line(term, 0, term->idx);
	    break;
	case 2: //erase whole line
	    Cls_line(term, 0, term->lg);
	    break;
	default: 
// printf("%s(%d): gotpars K: %d %d\n", FL, term->numIdx, term->nums[0]);
	    term->err = True; 
	}
}

static void REGP1 csiIL(Tleterm *term) { //<n>L	(il)	insert line
	if (term->ques) {term->err = True; return;}
	if (!term->region_on) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	tls_cs_move(term, term->idy + term->nums[0], term->idy, term->ws_row, 0);
}

static void REGP1 csiDL(Tleterm *term) { //<n>M	(dl)	delete line
	if (term->ques) {term->err = True; return;}
	if (!term->region_on) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	tls_cs_move(term, term->idy, term->idy + term->nums[0], term->ws_row, 0);
}

static void REGP1 csiDCH(Tleterm *term) { //<n>P	(dch)	delete characters
  int m, n;
	if (term->ques) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;

	m = term->idx + term->nums[0];
	n = term->lg - m;
// printf("%s(%d): lg=%d idx=%d m=%d n=%d\n", FL, term->lg, term->idx, m, n);
	if (n > 0) {
	    pnmove(&term->line[term->idx], &term->line[m], n * 2);
	    pnmove(&term->uct[term->idx],  &term->uct[m], n * sizeof(TUCharStat));
	}

	if (term->region_on) {
	    n += term->idx;
	    word *pc = &term->line[n];
	    while (n < term->lg) {
		*pc = ' ';
		pc++; n++;
	    }
	} else {
	    term->lg -= term->nums[0];
	    if (term->lg < 0) term->lg = 0;
	    if (term->lg < term->idx) term->idx = term->lg;
	    term->line[term->lg] = ' ';
	}
}

static void REGP1 csiECH(Tleterm *term) { //<n>X  erase characters
  int m;
	if (term->ques) {term->err = True; return;}
	if (!term->nums[0]) term->nums[0]++;
	m = term->idx;
	for (int i=term->nums[0];i>0;i--) {
	    Tleterm_add_char(term, ' ');
	}
	term->idx = m;
}

static void REGP1 csiDA1(Tleterm *term) { //<n>c  primary device attributes
}

static void REGP1 csiVPA(Tleterm *term) { //<n>d  cursor to line n
	Set_line_nr(term, get_top_row(term, term->nums[0]), True, True);
}

static void REGP1 csiSGR(Tleterm *term) { //<0-59>m	select graphic rendition
  int m;
	if (term->ques) {term->err = True; return;}
	for (int i=0;i<=term->numIdx;i++) {
	    m = term->nums[i];
//    printf("%s(%d): gotpars: m %d %d\n", FL, i, m);
	    if ((m >= 30) && (m <= 37)) {	//ForeColor
		if (term->Akt_uct.t.fu &= ctfu_FontBold) m += 8;
		term->Akt_uct.t.fc = m-30;
	    } else if ((m >= 40) && (m <= 47)) { //BackColor
		term->Akt_uct.t.bc = 15-(m-40);

	    } else switch (m) {
	    case 0: //all attributes off
		term->Akt_uct.i = 0;
		break;
	    case 1: //bold intensity = 2
		term->Akt_uct.t.fu |= ctfu_FontBold;
		break;
	    case 2: //intensity = 0
		term->Akt_uct.t.fu &= ~ctfu_FontBold;
		break;
	    case 4: //underline
	    case 5: //blink
		break;
	    case 7: //reverse
		term->Akt_uct.t.fu |= ctfu_ColReverse;
		break;
	    case 10: //Select primary font, don't display (don't set bit 8 on output)
		term->disp_ctrl = False;
		break;
	    case 11: //Select first alternate font (chars < 32 be displayed as ROM chars)
		term->disp_ctrl = True;
		break;
//		    case 12: ///Select second alternate font (toggle high bit before displaying as ROM char)
//		    case 21,22: //intensity = 1
	    case 24: //underline off
	    case 25: //blink off
		break;
	    case 27: //reverse off
		term->Akt_uct.t.fu &= ~ctfu_ColReverse;
		break;
	    case 38: //ForeColor
		i++;
		if (term->nums[i] != 5) return;
		i++;
		m = term->nums[i];
		if (term->Akt_uct.t.fu &= ctfu_FontBold) m += 8;
		term->Akt_uct.t.fc = m;
		break;
	    case 39: //underline of default foreground (color = (def_color & 0x0f) | background)
		term->Akt_uct.t.fc = 0;
		break;
	    case 48: //BackColor
		i++;
		if (term->nums[i] != 5) return;
		i++;
		m = term->nums[i];
		term->Akt_uct.t.bc = 15-m;
		break;
	    case 49: //(color = (def_color & 0xf0) | foreground)
		term->Akt_uct.t.bc = 0;
		break;

	    default: 
// printf("%s(%d): gotpars m: %d %d\n", FL, term->numIdx, term->nums[0]);
		term->err = True; 
		break; 
	    }
	}
}

static void REGP1 csiDECSTBM(Tleterm *term) { //<n>;<m>r	Select scrolling region
	if (term->ques) {term->err = True; return;}
	Set_region(term, term->nums[0], term->nums[1]);
}

static bool REGP1 Tleterm_set_mod(Tleterm *term, bool on_off) {
  bool err = False;

	for (int i=0;i<=term->numIdx;i++) {
	    if (term->ques) {
		switch(term->nums[i]) {
		case 1:
		    if (!on_off) term->Akt_uct.i = 0;
		    break;
		//case 4: //insert/replace
		//    term->replace_mod = (on_off == 0);
		//    break;
		case 7:  //Autowrap mode
		    term->decawm = on_off;
		    break;
		case 25:
		    term->hide_cursor = (on_off == 0);
		    break;
		case 47:   //alternate buffer
		case 1047:
//		case 1048:
		case 1049: //alternate buffer + cursor
		    term->alternate_screen = on_off;
		    break;

		case 5:     //DECSCNM light/dark background
		case 1000:  //VT200_MOUSE
		    break;

		default:
		    err = True;
		}

	    } else {
		switch(term->nums[i]) {
//		case 4: 
//		    term->smkx = on_off;	//ansi mode
//		    break;
		case 4: //replace/insert
		    term->insert_mod = on_off;
// printf("%s(%d): insert mode=%d\n", FL, on_off);
		    break;
		default:
		    err = True;
		}
	    }
	}
// if (err) printf("%s(%d): set_mod: %d %d %d ques=%d on_off=%d\n", FL, term->numIdx, term->nums[0], term->nums[1], term->ques, on_off);
	return err;
}

static void REGP1 csiSET(Tleterm *term) { //<n>;h
	Tleterm_set_mod(term, True);
}
static void REGP1 csiRST(Tleterm *term) { //<n>;l
	Tleterm_set_mod(term, False);
}

typedef void REGP1 (*TCSI)(Tleterm *term);

#define IGNORE NULL
#define CERR NULL

const TCSI csitab[64] = {
	csiICH,		csiCUU,		csiCUD,		 csiCUF,	//@ABC
	csiCUB,		CERR,		CERR,		 csiHPA,	//DEFG
	csiCUP,		CERR,		csiED,		 csiEL,		//HIJK
	csiIL,		csiDL,		NULL,		 NULL,		//LMNO

	csiDCH,		CERR,		CERR,		 NULL,		//PQRS
	NULL,		CERR,		CERR,		 IGNORE,	//TUVW
	csiECH,		NULL,		CERR,		 NULL,		//XYZ[
	NULL,		NULL,		NULL,		 NULL,		//\]^_

	CERR,		CERR,		NULL,		 csiDA1,	//`abc
	csiVPA,		CERR,		csiCUP,		 IGNORE,	//defg
	csiSET,		NULL,		NULL,		 NULL,		//hijk
	csiRST,		csiSGR,		NULL,		 NULL,		//lmno

	IGNORE,		IGNORE,		csiDECSTBM,	 CERR,		//pqrs
	CERR,		IGNORE,		CERR,		 NULL,		//tuvw
	CERR,		NULL,		CERR,		 CERR,		//xyz{
	CERR,		NULL,		NULL,		 NULL		//|}~
};

static bool Tleterm_gotpars(Tleterm *term, byte c) {
  bool err = False;

	term->state = 0;
	if ((c >= 0x40) && (c <= 0x7f) && csitab[c-0x40]) {
	    csitab[c-0x40](term);
	    err = term->err;
	} else {
// printf("%s(%d): gotpars: %d [%d,%d] 0x%02x\"%c\"\n", FL, term->numIdx, term->nums[0], term->nums[1], c, c);
	    err = True; 
	}

	return err;
}

static bool Tleterm_esc(Tleterm *term, byte c) {
  int m, n;
  bool err = False;

	term->state = 0;
	switch(c) {
	case '[': term->state = tty_square; break;
	case '(': term->state = tty_setG0; break;
	case ')': term->state = tty_setG1; break;
	case '=': break;	//DECKPAM
	case '>': break;	//DECKPNM
	case 'M':	//Del Line
// printf("%s(%d): Del Line:  idy=%d  count=%d \n", FL, term->idy, term->le->tls->Count);
	    Tleterm_set_CS_color(term);
	    if (term->region_sector) {
		m = get_top_row(term, term->region_top);
		if (m == term->idy) {
		    n = get_top_row(term, term->region_top);
		    m = term->region_bottom - term->region_top;
		    tls_cs_move(term, n+1, n, m, NO_CS_GET_LINE);
		    Tleterm_set_CS_color(term);
		    return err;
		}
	    }
	    m = term->le->tls->Count - 1;
	    if (m > 0) {
		TLines_Del(term->le->tls, m);
		if (term->idy > 0) term->idy--;
	    }
	    set_line_idy(term, False, False);
	    break;

	case '7':	//(sc) save cursor position
	    term->save_idx = term->idx;
	    term->save_idy = term->idy;
	    term->save_uct.i = term->Akt_uct.i;
	    term->save_disp_ctrl = term->disp_ctrl;
	    term->save_Akt = True;
	    break;
	case '8':	//(rc) restore cursor position to saved value
	    if (term->save_Akt) {
		Set_line_nr(term, term->save_idy, False, False);
		term->idx = term->save_idx;
		if (term->idx > term->lg) term->idx = term->lg;
		term->Akt_uct.i = term->save_uct.i;
		term->disp_ctrl = term->save_disp_ctrl;

	    } else err = True;
	    term->save_Akt = False;
	    break;

	default:
	    if (term->term_type == term_xterm) {
		switch(c) {
		case 'O': term->state = tty_esc_O; break;
		default:
		    err = True;
		    break;
		}

	    } else if (term->term_type == term_xterm) {
		err = True;

	    } else {
		err = True;
	    }
	    break;
	}
	return err;
}

static bool Tleterm_esc_O(Tleterm *term, byte c) {
  bool err = True;

	term->state = 0;

// printf("Tleterm_esc_O: %d %d\n", term->numIdx, term->nums[0]);

	return err;
}

static bool Tleterm_funckey(Tleterm *term, byte c) {
  bool err = True;

// printf("Tleterm_funckey: %d %d\n", term->numIdx, term->nums[0]);
	term->state = 0;
	return err;
}

/* 0x00=ignore 0x01=print 0x01>bef */
const char C_ASCII[192] = {
  0x00,0x00,0x00,0x00 ,0x00,0x05,0x00,0x07 ,0x08,0x01,0x0a,0x0b ,0x0c,0x0d,0x0e,0x0f,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x18,0x01,0x1a,0x1b ,0x00,0x00,0x00,0x00,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,

  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x00,
#if 0
  0x00,0x00,0x00,0x00 ,0x84,0x85,0x00,0x00 ,0x88,0x00,0x00,0x00 ,0x00,0x8d,0x8e,0x8f,
  0x90,0x00,0x00,0x00 ,0x00,0x00,0x00,0x97 ,0x98,0x00,0x9a,0x9b ,0x9c,0x9d,0x9e,0x9f,
#else
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
#endif
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01
};

bool Tleterm_scan(Tleterm *term, dword c) {
  int m;
  bool err = False;

	m = GetTime_ms();
	if (cpTime(term->time_ms, m)) term->state = 0;
	term->time_ms = m + 100;

	term->laststate = term->state;
loop:
	if (term->state) {
	    if (term->state == tty_esc) {
		for (int i=0; i<8; i++) term->nums[i] = 0;
		term->numIdx = term->numCount = 0;
		term->ques = False;
		err = Tleterm_esc(term, c);

	    } else if (term->state == tty_square) {
		term->state = tty_getnums;
		if (c == '[') {
		    term->state = tty_funckey;
		} else {
		    term->ques = (c == '?');
		    if (!term->ques) goto loop;
		}

	    } else if (term->state == tty_getnums) {
		if ((c >= '0') && (c <= '9')) {
		    term->nums[term->numIdx] *= 10;
		    term->nums[term->numIdx] += c - 0x30;
		    if (!term->numCount) term->numCount++;
		} else if (c == ';') {
		    if (term->numIdx < 7) term->numIdx++;
		    term->numCount++;
		} else {
		    term->state = tty_gotpars;
		    goto loop;
		}

	    } else if (term->state == tty_gotpars) {
		err = Tleterm_gotpars(term, c);

	    } else if (term->state == tty_funckey) {
		err = Tleterm_funckey(term, c);

	    } else if (term->state == tty_setG0) {
		term->state = 0;
		err = c != 'B';

	    } else if (term->state == tty_setG1) {
		term->state = 0;
		err = c != '0';

	    } else if (term->state == tty_esc_O) {
		err = Tleterm_esc_O(term, c);

	    } else {
		err = True;
	    }
	} else {
	    term->err = False;
	    term->reslg = 0;
	    term->btty_XK = 0;
	    if (c >= 0xc0) m = 1; else m = C_ASCII[c];
	    if (m == 1) {
		if (term->region_on && (c == ' ') && 
			(term->idx == 1) && (term->line[0] == '>')) {
// printf("%s(%d): Region OFF  is_cls=%d\n", FL, term->region_is_cls);
		    term->region_on = False;
		    term->region_sector = False;
		    if (term->region_is_cls) {
			m = term->le->tls->Count - term->ws_row - 1;
			if (m < 0) m = 0;
			for (int i=term->ws_row;i>0;i--) {
			    if (m == term->idy) m++;
			    TLines_Del(term->le->tls, m);
			    if (m < term->idy) term->idy--;
			}
			term->lg = term->idx;
		    } else {
			term->lg = 1;
			while ((term->idy + 1) < term->le->tls->Count) {
			    TLines_Del(term->le->tls, term->idy+1);
			}
		    }
		}

		if (term->idx || (c != '>')) term->region_is_cls = False;

		if (term->idx >= term->ws_col) {
		    Tleterm_scan(term, '\n');
		    term->idx = 0;
		}

		Tleterm_add_char(term, c);

	    } else if (m) {

		switch(c) {
		case 0x07: XBell(appXDat.dpy, 0); break;
		case 0x08: dec_x(term); break;
		case 0x0A: add_lf(term, True); break;
		case 0x0D: term->idx = 0;  break;
		case 0x0E: term->charset = True; break;
		case 0x0F: term->charset = False; break;
		case 0x18: 
		case 0x1A: 
		    term->state = tty_normal;
		    if (term->disp_ctrl) Tleterm_add_char(term, c);
		    break; 
		case 0x1B: term->state = tty_esc; break;
//		case 0x9B: term->state = tty_square; break;
		default:
		    err = True;
		}
	    }
	}

	term->res[term->reslg] = c;
	term->reslg++;
	term->res[term->reslg] = 0;
	if ((term->reslg > 15) || err) {
#ifdef	ttyscn_DEBUG
		tty_err(term, c);
#endif
		term->state = 0;
	}
	term->err = err;
	return (term->state == 0);
}

void Tleterm_SetSize(Tleterm *term, int col, int row) {
	if (col <= 0) col = term->ws_col;
	if (row <= 0) row = term->ws_row;

	if (term->ws_col != col) {
	    term->ws_col = col;
	    if (term->region_on) {
		
		int y = term->le->tls->Count - term->ws_row;
		for (int i=term->ws_row;i>0;i--) {
		    if (term->idy == y) {
			if (term->lg > col) term->lg = col;
		    }
		    TColS *cs = TLines_Get_pcs(term->le->tls, y);
		    y++;
		    if (cs && (cs->lg > col)) cs->lg = col;
		}
		if (term->idx > col) term->idx = col;
	    }
	}

	if (term->ws_row != row) {
	    term->ws_row = row;
	}
}

void Tleterm_Clear(Tleterm *term, bool all) {

	if (term->cs->s == term->pcline) term->cs->s = NULL;
	if (term->cs->c == (char*)term->scda) term->cs->c = NULL;

	TLineEdit_Clr(term->le);
	//Clear all var 
	if (all) {
#ifdef POINTER64BIT
	    Fill0(&term->idx, sizeof(TUCharStat) + (int64_t)&term->Akt_uct - (int64_t)&term->idx);
#else
	    Fill0(&term->idx, sizeof(TUCharStat) + (int)&term->Akt_uct - (int)&term->idx);
#endif
	} else {
	    term->idy = 0;
	}
	term->cs = TLines_Add_(term->le->tls);

}

/* ______________________ Destroy ____________________________________ */
static bool Tleterm__Destroy(Tleterm *term) {
	xfree(term->line);
	xfree(term->uct);
	xfree(term->scda);
	return False;
}

/* ______________________ Create _____________________________________ */
Tleterm *Tleterm_Create(TLineEdit *le) {
  Tleterm *term = xcalloc(sizeof(Tleterm));

	term = ttc_create(sizeof(Tleterm), wtyp_leterm);
	term->destroy = (TDESTROY)Tleterm__Destroy;
	term->le = le;
	term->linesize = 256;
	term->line = xmalloc(term->linesize * 2);
	term->pcline = xmalloc(term->linesize * 4);
	term->uct = xcalloc(term->linesize * sizeof(TUCharStat));
	set_scda_arr((void*)&term->scda, term->linesize);
	term->time_ms = GetTime_ms() + 100;

	term->Font = MakeFont(TpcIni_RD_pc(appDat.appIni, "AHconsole", "Font", NULL),
		"-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso8859-1", NULL);
//	term->FontBold = MakeFont_Bold(term->Font->name, NULL);
	MakeFontArt(term->Font->name, &term->FontBold, FONT_Bold);
	term->AktFont = term->Font;
	term->AktFontBold = term->FontBold;

	term->FontUTF8 = MakeFont(TpcIni_RD_pc(appDat.appIni, "AHconsole", "FontUTF8", NULL),
		"-misc-fixed-medium-r-semicondensed-*-13-*-75-75-*-*-iso10646-1", NULL);
	MakeFontArt(term->FontUTF8->name, &term->FontBoldUTF8, FONT_Bold);

	term->idy = le->tls->Count;
	term->cs = TLines_Add_(term->le->tls);
	term->ws_col = 80;
	term->ws_row = 24;
	term->max_lines = TpcIni_RD_int(appDat.appIni, "AHconsole", "MaxLines", 1000);
	if (term->max_lines > 10000) term->max_lines = 10000;
	if (term->max_lines < 200) term->max_lines = 200;

/*	s = getenv("TERM");
	if (s) {
	    if (!pccasecmpASCII(s, "xterm")) term->termart = term_xterm; 
	} */
	return term;
}

