/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldedit.h"
#include "cicons.h"

/* ______________________ File Ev ____________________________________ */
static void menuFile_OnMenuOpen(TMenu *self) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	if (fed->PageControl) {
	    fed->miClosePage->fu |= miNotEnabled;
	    fed->miChangePage->fu |= miNotEnabled;
	    fed->miSaveProj->fu |= miNotEnabled;
	    fed->miSaveAsProj->fu |= miNotEnabled;
	    if (TPageControl_GetPageCount(fed->PageControl, False) && fed->AktLE) {
		fed->miClosePage->fu &= ~miNotEnabled;
		fed->miChangePage->fu &= ~miNotEnabled;
		if (fed->ProjFile) fed->miSaveProj->fu &= ~miNotEnabled; 
		fed->miSaveAsProj->fu &= ~miNotEnabled;
	   }
	}
}

static void menuReopen_OnMenuOpen(TMenu *self) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TpcIni *pcIni;
  int i, fileart;
  char *fnt;
  const char *fn;
  char line[32];
  TMenuItem *mi;
  TglobPath *gp;

	TMenu_Cls(fed->menuReopen);
	pcIni = TpcIni_Create(appDat.inifile_app);
	for (i=0; i<REOPEN_ITEMS; i++) {
	    snprintf(line, 32, "LastFile%d", i);
	    fn = TpcIni_RD(pcIni, APPNAME, line);
	    if (!fn) break;
	    gp = pctogp(fn);
	    gp_CryptMini16c_enc(gp);

	    fileart = stft_ErrNoFile;
	    if ((gp->fu & gpi_Local)) {
		Tstat64 *st = file_stat_gp(gp, fsFollowLink|fsNoFileName|fsNoLinkName|fsGetFileType|fsGetFileTypeHead);
		if (st) fileart = st->FileType;
		xfree(st);
	    } else {
		fileart = get_file_stft_ext(fn, NULL, NULL);
	    }

	    if ((fileart != stft_ErrNoFile) && (fileart != stft_Folder)) {
		fnt = get_DisplayFileStr(NULL, gp);
		char *tmp = path_cut(fnt, 74);
		xfree(fnt);
		mi = TMenu_AddItem(fed->menuReopen, tmp, miDisable_s2, NULL);
		xfree(tmp);
		mi->s2 = gptopc(gp, gptfu_PasswdTxt);
		mi->pic = get_fileIcon(fileart);
	    }
	    gpFree(gp);
	}
	Tfree(pcIni);
}

static void mFile_New(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	if (fed->PageControl) {
	    Add_Page(fed, NULL, " New");
	} else {
	    TLineEdit_Clr(fed->AktLE);
	    fed->AktLE->ReadOnly = False;
	    SetCaption((TWins*)fed, " New");
	    ppcfree(&fed->AktLE->tls->FileName);
	}
}

void mFile_ClosePage(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  int m;
  TPage *pag;
	pag = GetParentPage((TWins*)fed->AktLE);
	if (fed->mbSave->Enabled) {
	    m = ChkFileIsSaved(fed->AktLE);
	    if (m == msgCANCEL) return;
	}
//	if (pag) TPageControl_FreePage(fed->PageControl, pag);
	Tfree(pag);

	if (!TPageControl_GetPageCount(fed->PageControl, False)) fed->AktLE = NULL;
}

static void File_Open(TFormEdit *fed, int fu) {
  char *s=NULL;
	if (!fed->File_Dlg) {
	    fed->File_Dlg = Create_File_Dlg((TWins*)fed);
	    if (TImageRD_AllFiletyp) file_dlg_InitAzPic(fed->File_Dlg);
	}

	if (fed->AktLE) {
	    s = GetFilePath(fed->AktLE->tls->FileName);
	    if (s) File_Dlg_SetPath(fed->File_Dlg, s);
	    ppcfree(&s);
	}

	fed->File_Dlg_Run = true;
	if (File_Dlg_Execute(fed->File_Dlg)) {
	    File_Dlg_GetEdFileName(fed->File_Dlg, &s);
	    LoadFile(fed, NULL, s, fu, NULL);
	    xfree(s);
	}
	fed->File_Dlg_Run = false;
//	SetFocus(LineEdit);
}

void mFile_ChangePage(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	File_Open(fed, 0x02);
}

void mFile_Open(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	if (fed->PageControl) {
	    File_Open(fed, 0x03);
	} else {
	    File_Open(fed, 0x02);
	}
}

static void menuReopen_OnClick(TMenu *self, long KeyState, bool DblClick) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  char *fn;
	if (fed->menuReopen->SelItem) {
	    fn = fed->menuReopen->SelItem->s2;
	    if (!fn) fn = fed->menuReopen->SelItem->s;
	    LoadFile(fed, NULL, fn, 0x03, NULL);
	}
}

void mFile_Save(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	File_Save(fed);
}

void mFile_SaveAs(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	FileSaveAs(fed->AktLE);
}

static void mFile_SaveSize(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TpcIni *Ini;

	Ini = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR_int(Ini, APPNAME, "Width", fed->width);
	TpcIni_WR_int(Ini, APPNAME, "Height", fed->height);
	TpcIni_SaveAndFree(Ini);
}

/* ------- SaveProj ------- */
void File_SaveProj(TFormEdit *self, int fu) {
  TPage *pg;
  TLineEdit *le;
  char *Gp, *tmps;
  char buf[64];

	if (!self->ProjFileIni) return;

	tmps = NULL;
	pg = TPageControl_GetFirstPage(self->PageControl, False);
	while (pg) {
	    le = (TLineEdit*)pg->owner;
	    if (le) {
		Gp = path_abs2rel(le->tls->FileName, self->ProjPath);
		if (le == self->AktLE) {
		  TpcIni_WR(self->ProjFileIni, "File", "LastPage", Gp);
		}
		char *pc = buf;
		ppcmove_v32(&pc, le->VSB->Count, kVALUE_int | kSPACE_ADD);
		ppcmove_v32(&pc, le->CaretCharIdx, kVALUE_int | kSPACE_ADD);
		ppcmove_v32(&pc, le->CaretCharIdy, kVALUE_int);
		TpcIni_WR(self->ProjFileIni, Gp, "topxy", buf);
		ppcadd_n(&tmps, 3, "\"", Gp, "\" ");
	    }
	    pg = TPage_GetNext(pg, False);
	}
	if ((fu & 0x01) == 0) TpcIni_WR(self->ProjFileIni, "File", "files", tmps);
	xfree(tmps);
	TpcIni_Save(self->ProjFileIni);
}

static void mFile_SaveProj(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	File_SaveProj(fed, 0);
}

void mFile_FileSave_Dlg_Init(TFormEdit *fed) {
	fed->FileSave_Dlg = Create_FileSave_Dlg((TWins*)fed);
	if (TImageRD_AllFiletyp) file_dlg_InitAzPic(fed->FileSave_Dlg);
}

static void mFile_SaveAsProj(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  char *pc, *s;
  const char *cpc;
  int m;
  char line[1024];

	s = NULL;
	fed->IO_Run = True;
	if (!fed->FileSave_Dlg) {
	    mFile_FileSave_Dlg_Init(fed);
	    TPage *pg = TPageControl_GetFirstPage(fed->PageControl, False);
	    if (pg && pg->owner) {
		s = GetFilePath(((TLineEdit*)pg->owner)->tls->FileName);
		if (s) File_Dlg_SetPath(fed->FileSave_Dlg, s);
		ppcfree(&s);
	    }
	}

	if (fed->ProjFile) {
	    s = path_rel2abs(fed->ProjFile, fed->ProjPath);
	    File_Dlg_SetNamePath(fed->FileSave_Dlg, s);
	    ppcfree(&s);
	} else {
	    SetCaption((TWins*)fed->FileSave_Dlg->edFile, "Default.aldapj");
	}

	if (File_Dlg_Execute(fed->FileSave_Dlg)) {
	    File_Dlg_GetEdFileName(fed->FileSave_Dlg, &s);

	    if (test_local_drive(s)) {
		cpc = GetFileExt_pc(s);
		if (!cpc || (pccasecmpASCII(cpc, "apj") &&
			     pccasecmpASCII(cpc, "aldapj"))) ppcadd(&s, ".aldapj");
		pc = s;

		if (TestIsFile(pc)) {
		    snprintf(line, sizeof(line), "Overwrite exists project file \n%s", pc);
		    m = MsgBox_Show((TWins*)fed, "Warning !!!", 
			    line, msgYES | msgNO, msgNO);
		    if (m == msgNO) pc = NULL;
		}
		if (pc) {
		    if (fed->ProjFileIni) Tfree(fed->ProjFileIni);
		    xfree(fed->ProjFile);
		    xfree(fed->ProjPath);
		    fed->ProjFileIni = TpcIni_Create(pc);
		    fed->ProjFile = GetFileName(pc);
		    fed->ProjPath = GetFilePath(pc);
		    File_SaveProj(fed, 0);
		}

	    } else {
		MsgBox_Show((TWins*)fed, "ERROR", 
			"No local data drive !!!", msgCANCEL | msgIcoERROR, msgCANCEL);
	    }
	}
	xfree(s);
	fed->IO_Run = False;
}

#ifdef USE_PRINT_DLG
static void mFile_Print(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TLineEdit *le = fed->AktLE;
  TPixMap *pxm;
  TColS *cs;
  int y, idx, maxh;
  int PageNr;
  int w, h, OffLeft, OffTop, OffRight, OffBottom;

#if 0
	w = 595; h = 842;	//A4 210x297 72dpi
	OffLeft = 30; OffRight = 20;
	OffTop = 20; OffBottom = 20;
#else
	w = 826; h = 1168;	//A4 210x297 100dpi
	OffLeft = 42; OffRight = 26;
	OffTop = 26; OffBottom = 26;
#endif


//	pxm = TPixMap_Create(590, 840, 0);
//	pxm = TPixMap_Create_depth(786, 1112, 0);
	pxm = TPixMap_Create_depth(w, h, 0);
#ifdef USE_XSHM
	TPixMap_SetXShm(pxm);
#endif
	TPixMap_CreateXimg(pxm);

 printf("%s(%d): 0x%x 0x%x\n", FL, (int)pxm, (int)pxm->Ximg);

	pxm->vds = xcalloc(sizeof(TVDrawString));
	pxm->vds->Font = le->vds.Font;
	pxm->vds->BackColor = 0xffffff;
	pxm->vds->TabSpaceWidth = le->vds.TabSpaceWidth;
	pxm->vds->TabSpaceOn = le->vds.TabSpaceOn;
	pxm->vds->TabSpaceDot = le->vds.TabSpaceDot;
	pxm->vds->UTF8on = le->vds.UTF8on;
	pxm->vds->Xoff = OffLeft;
	maxh = pxm->height - OffTop - OffBottom;
	PageNr = idx = 0;

loop_Page:
	TPixMap_ClsRGB(pxm, 0xffffff);
	TPixMap_XPutImage(pxm);
	pxm->vds->WinID = pxm->pmxID;
// printf("%s(%d): 0x%x\n", FL, (int)pxm->pmxID);
	PageNr++;
	y = OffTop;
loop_line:
	cs = TLines_Get_pcs(le->tls, idx);
	if (cs) {
	    pxm->vds->s = cs->s;
	    pxm->vds->lg = cs->lg;
	    pxm->vds->scda = (TStrColorDatArr*)cs->c;
	    DrawColorString(0, y, pxm->vds);
	}
	idx++;
	y += pxm->vds->Font->height;
	if ((y < maxh) && (idx < le->tls->Count)) goto loop_line;

	if (OffRight > 0) {
	    FillRectRGB(pxm->pmxID, 0xffffff, w - OffRight, 0, OffRight, h);
	}
//	DrawRectFocus(pxm->pmxID, OffLeft, OffTop, w - OffLeft- OffRight, h - OffTop - OffBottom);

	TPixMap_XGetImage(pxm);
 char line[60];
	snprintf(line, 60, "/tmp/.ALDccp/prn_%d.png", PageNr);
//	TImage_Write_PNG((TImage*)pxm, line, fuWR_gray|fuWR_NoMask, 0);
	TImage_Write_PNG((TImage*)pxm, line, fuWR_NoMask|fuWR_use_pallette, 0);
	if (idx < (le->tls->Count - 10)) goto loop_Page;
	Tfree(pxm);
}
#endif

static void mFile_Exit(TMenu *self, long KeyState, TMenuItem *smi) {
  Close();
}


/* ______________________ Create _____________________________________ */
void MenuFile_Create(TFormEdit *fed, bool use_pages) {
  TMenuItem *mi;

	fed->mbFile = TMenuButton_Create(fed->PanelTop, (TMenuButton*)fed);
	fed->mbFile->top = 2;
	fed->mbFile->left = 1;
	SetCaption((TWins*)fed->mbFile, "File");
	Set_Parent((TWins*)fed->mbFile, fed->PanelTop);

	fed->menuFile = TMenu_Create((TWins*)fed);
	fed->menuFile->OnMenuOpen = menuFile_OnMenuOpen;
	fed->mbFile->menu = fed->menuFile;

	TMenu_AddItem(fed->menuFile, "New", 0, mFile_New);
	TMenu_AddItem(fed->menuFile, "-", 0, NULL);

	if (use_pages) {
	    fed->miClosePage = TMenu_AddItem(fed->menuFile, "Close Page", 0, mFile_ClosePage);
	    fed->miChangePage = TMenu_AddItem(fed->menuFile, "Change Page", 0, mFile_ChangePage);
	    fed->miChangePage->pic = get_pic16_fileopen(False);
	    TMenu_AddItem(fed->menuFile, "-", 0, NULL);
	}

	mi = TMenu_AddItem(fed->menuFile, "Open...", 0, mFile_Open);
	mi->pic = get_pic16_fileopen(False);
	mi = TMenu_AddItem(fed->menuFile, "Reopen", 0, NULL);
		fed->menuReopen = TMenu_Create((TWins*)fed);
		fed->menuReopen->OnClick = menuReopen_OnClick;
		fed->menuReopen->OnMenuOpen = menuReopen_OnMenuOpen;
		mi->next_menu = fed->menuReopen;

	TMenu_AddItem(fed->menuFile, "-", 0, NULL);
	fed->miSave = TMenu_AddItem(fed->menuFile, "Save", miNotEnabled, mFile_Save);
	fed->miSave->pic = get_pic16_filesave(False);
	mi = TMenu_AddItem(fed->menuFile, "Save As...", 0, mFile_SaveAs);
	mi->pic = get_pic16_filesave(False);

	if (use_pages) {
	  TMenu_AddItem(fed->menuFile, "-", 0, NULL);
	  mi = TMenu_AddItem(fed->menuFile, "Projekt New", miNotEnabled, NULL);
	  mi = TMenu_AddItem(fed->menuFile, "Projekt Open", miNotEnabled, NULL);
	  fed->miSaveProj = TMenu_AddItem(fed->menuFile, "Projekt Save", miNotEnabled, mFile_SaveProj);
	  fed->miSaveProj->pic = get_pic16_filesave(False);
	  fed->miSaveAsProj = TMenu_AddItem(fed->menuFile, "Projekt Save As...", miNotEnabled, mFile_SaveAsProj);
	  fed->miSaveAsProj->pic = get_pic16_filesave(False);
	}

#ifdef USE_PRINT_DLG
	TMenu_AddItem(fed->menuFile, "-", 0, NULL);
	TMenu_AddItem(fed->menuFile, "Print", 0, mFile_Print);
#endif

	TMenu_AddItem(fed->menuFile, "-", 0, NULL);
	fed->miSaveSize = TMenu_AddItem(fed->menuFile, "Save Size", 0, mFile_SaveSize);
	fed->miSaveSize->pic = get_pic16_filesave(False);

	TMenu_AddItem(fed->menuFile, "-", 0, NULL);
	mi = TMenu_AddItem(fed->menuFile, "Exit", 0, mFile_Exit);
	mi->pic = GetZIcon(zico16exit);

}

