/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldedit.h"
#include "cicons.h"

/* ------------------------------------------------- */
void UpdateSettingMenu(TFormEdit *fed) {
  TLineEdit *le = fed->AktLE;
  TLines *tls;
  TMenuItem *mi;
  int m, stft;

	if (!le) return;
	tls = le->tls;
	fed->miEOL_unix->fu &= ~miChecked;
	fed->miEOL_windos->fu &= ~miChecked;
	fed->miEOL_Strg_C_Korr->fu &= ~miChecked;
	fed->miEOL_Strg_C_Korr->fu |= miNotEnabled;
	fed->miChkReload->fu &= ~miChecked;

	m = tls->DefaultEOL & fio_EOL_Mask;
	if (m == fio_EOLunix) fed->miEOL_unix->fu |= miChecked;
	if (m == fio_EOLwindos) fed->miEOL_windos->fu |= miChecked;

	if (tls->DefaultEOL & fio_EOL_mix) {
	    fed->miEOL_Strg_C_Korr->fu &= ~miNotEnabled;
	} else {
#ifdef _OS_WIN32
	    if (m & fio_EOLunix) {
		fed->miEOL_Strg_C_Korr->fu &= ~miNotEnabled;
	    }
#else
	    if (m & fio_EOLwindos) {
		fed->miEOL_Strg_C_Korr->fu &= ~miNotEnabled;
	    }
#endif
	}
	if (le->fio_Strg_C_on) fed->miEOL_Strg_C_Korr->fu |= miChecked;
	if (le->user_b1) fed->miChkReload->fu |= miChecked;

	stft = stft_None;
	if (le->highlight) stft = le->highlight->hl_Type;
	mi = fed->menuHighlight->MenuItem;
	while (mi) {
	    if (mi->tag != stft) mi->fu &= ~miChecked; else mi->fu |= miChecked;
	    mi = mi->next;
	}

  int bom = le->tls->org_load_fio_fu & fio_txtTyp_Mask;
	mi = fed->menuBOM->MenuItem;
	while (mi) {
	    mi->fu &= ~miChecked;
	    if (mi->tag == bom) mi->fu |= miChecked;
	    mi = mi->next;
	}
}

void Set_highlight(TFormEdit *fed, TLineEdit *le, int stft, int sub_mime) {

	if (!le) le = fed->AktLE;
	if (!le) return;
	Tfree(le->highlight);
	le->highlight = NULL;

	switch(stft) {
	case stft_C:
	    le->highlight = Thighlight_C_Create((TWins*)le, le->tls);
	    break;
	case stft_CXX:
	    le->highlight = Thighlight_CXX_Create((TWins*)le, le->tls);
	    break;
	case stft_Pascal:
	    le->highlight = Thighlight_Pas_Create((TWins*)le, le->tls);
	    break;
	case stft_Shell:
	    switch(sub_mime) {
	    case mShell_spec:
		le->highlight = Thighlight_Spec_Create((TWins*)le, le->tls);
		break;
	    default:
		le->highlight = Thighlight_Bash_Create((TWins*)le, le->tls);
		break;
	    }
	    break;
	case stft_Make:
	    le->highlight = Thighlight_Make_Create((TWins*)le, le->tls);
	    break;
	case stft_ASM:
	    le->highlight = Thighlight_ASMi386gcc_Create((TWins*)le, le->tls);
	    break;
	default:
	    stft = stft_None;
	}
	if (le->highlight) le->highlight->hl_Type = stft;
	TLineEdit_AllAtrNew(le);
}

/* ______________________ Menu Ev _____________________________________ */
static void menuSetting_OnMenuOpen(TMenu *self) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	if (fed->AktLE) {
	    fed->miUTF8->fu &= ~miChecked;
	    if (fed->AktLE->vds.UTF8on) fed->miUTF8->fu |= miChecked;
	}
}

static void mSetting_Highlight(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	Set_highlight(fed, NULL, smi->tag, 0);
	UpdateSettingMenu(fed);
}

static void mSetting_EOL(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TLines *tls;
  int m;

	if (!fed->AktLE) return;
	tls = fed->AktLE->tls;
	m = tls->DefaultEOL & fio_EOL_Mask;

	if (tls->DefaultEOL & fio_EOL_mix) {
	    if (smi == fed->miEOL_unix) {
		if (m == fio_EOLunix) m = 0; else m = fio_EOLunix;
	    } else if (smi == fed->miEOL_windos) {
		if (m == fio_EOLwindos) m = 0; else m = fio_EOLwindos;
	    }
	} else {
	    if (smi == fed->miEOL_unix) m = fio_EOLunix;
	    if (smi == fed->miEOL_windos) m = fio_EOLwindos;
	}

	tls->DefaultEOL = (tls->DefaultEOL & ~fio_EOL_Mask) | m;
	UpdateSettingMenu(fed);
	ChkAllFileSaved(fed);
}

static void mSetting_EOL_MissMatchKorr(TMenu *self, long KeyState, TMenuItem *smi) {
	smi->fu ^= miChecked;
}

#ifndef _OS_WIN32
static void mSetting_Clipbord(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TMenuItem *mi;
	mi = fed->menuClipbord->MenuItem;
	while (mi) {mi->fu &= ~miChecked; mi = mi->next;}
	smi->fu |= miChecked;
	if (smi->tag == 1) {
	    appXDat.CLIPBOARD_ATOM = XA_PRIMARY;
	} else if (smi->tag == 2) {
	    appXDat.CLIPBOARD_ATOM = appXDat.XA_CLIPBOARD;
	}
}
#endif /* _OS_WIN32 */
/* ________________________ Font  _______________________________________ */
static void mSetting_Font(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TLineEdit *le = fed->AktLE;
  char *fontname;

	if (!le) return;

	fontname = pcdup(le->vds.Font->name);
	Font_Dlg_Create_Execute((TWins*)fed, &fontname);

	SetFont((TWins*)le, fontname);
	xfree(fontname);
//	Refresh(le);
}

static void mSetting_FontSave(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TpcIni *ini;

	if (!fed->AktLE) return;
	ini = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR(ini, "aldedit", "Font", fed->AktLE->vds.Font->name);
	TpcIni_SaveAndFree(ini);
}

static void mUTF8_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	smi->fu ^= miChecked;
	if (!fed->AktLE) return;
	if (smi->fu & miChecked) {
	    fed->AktLE->vds.UTF8on = True;
	} else {
	    fed->AktLE->vds.UTF8on = False;
	}
	TLineEdit_AllAtrNew(fed->AktLE);
//	Refresh((TWins*)fed->AktLE);
}

static void TAHedit_mBOM(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TLineEdit *le;
  TMenuItem *mi = fed->menuBOM->MenuItem;

	while (mi) {
	    mi->fu &= ~miChecked;
	    mi = mi->next;
	}
	smi->fu |= miChecked;

	le = fed->AktLE;
	if (le) {
	    le->tls->load_fio_fu &= ~fio_txtTyp_Mask_BOM;
	    le->tls->load_fio_fu |= smi->tag;
	    if (smi->tag != fio_txtTyp_ANSI) le->tls->load_fio_fu |= fio_has_BOM;
	}
}


static void menuLineOff_OnMenuOpen(TMenu *self) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TLineEdit *le;
  TMenuItem *mi;

	le = fed->AktLE;
	if (!le) return;
	mi = fed->menuLineOff->MenuItem;
	while (mi) {
	    mi->fu &= ~miChecked;
	    if (mi->tag == le->LineOffset) mi->fu |= miChecked;
	    mi = mi->next;
	}
}

static void mLineOff_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
  TLineEdit *le;
  TpcIni *ini;
  TMenuItem *mi = fed->menuLineOff->MenuItem;

	while (mi) {
	    mi->fu &= ~miChecked;
	    mi = mi->next;
	}
	smi->fu |= miChecked;
	le = fed->AktLE;
	if (le) {
	    le->LineOffset = smi->tag;
	    TLineEdit_SetCaret(le);
	    Refresh((TWins*)le);
	}
	ini = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR_int(ini, APPNAME, "LineOffset", le->LineOffset);
	TpcIni_SaveAndFree(ini);
}

static void TAHedit_mChkTimestap(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormEdit *fed = (TFormEdit*)self->FormStruct;
	smi->fu ^= miChecked;
	if (!fed->AktLE) return;
	fed->AktLE->user_b1 = (smi->fu & miChecked) != 0;
	Chk_LE_Reload(fed, fed->AktLE);
}

/* ______________________ Create _____________________________________ */
void MenuSetting_Create(TFormEdit *fed) {
  TMenuItem *mi;
#ifndef _OS_WIN32
  int m;
#endif

	fed->mbSetting = TMenuButton_Create(fed->PanelTop, fed->mbFile);
	SetCaption((TWins*)fed->mbSetting, "Settings");
	Set_Parent((TWins*)fed->mbSetting, fed->PanelTop);

	fed->menuSetting = TMenu_Create((TWins*)fed);
	fed->menuSetting->OnMenuOpen = menuSetting_OnMenuOpen;
	fed->mbSetting->menu = fed->menuSetting;

	mi = TMenu_AddItem(fed->menuSetting, "Highlight Mode", 0, NULL);
	  fed->menuHighlight = (TMenu*)TMenu_Create((TWins*)fed);
	  mi->next_menu = fed->menuHighlight;
	  TMenu_AddItem(fed->menuHighlight, "None", miCanCheck | miChecked, mSetting_Highlight)->tag = stft_None;
	  TMenu_AddItem(fed->menuHighlight, "Bash", miCanCheck, mSetting_Highlight)->tag = stft_Shell;
	  TMenu_AddItem(fed->menuHighlight, "C", miCanCheck, mSetting_Highlight)->tag = stft_C;
	  TMenu_AddItem(fed->menuHighlight, "C++", miCanCheck, mSetting_Highlight)->tag = stft_CXX;
//	  TMenu_AddItem(fed->menuHighlight, "html", miCanCheck, mSetting_Highlight)->tag = stft_HTML;
	  TMenu_AddItem(fed->menuHighlight, "Pascal", miCanCheck, mSetting_Highlight)->tag = stft_Pascal;
	  TMenu_AddItem(fed->menuHighlight, "makefile", miCanCheck, mSetting_Highlight)->tag = stft_Make;
	  TMenu_AddItem(fed->menuHighlight, "ASM gcc i386", miCanCheck, mSetting_Highlight)->tag = stft_ASM;

	mi = TMenu_AddItem(fed->menuSetting, "End of Line", 0, NULL);
	  fed->menuEOL = (TMenu*)TMenu_Create((TWins*)fed);
	  mi->next_menu = fed->menuEOL;
#ifdef OS_WIN32
	  fed->miEOL_unix = TMenu_AddItem(fed->menuEOL, "EOL Unix", 
			miCanCheck, mSetting_EOL);
	  fed->miEOL_windos = TMenu_AddItem(fed->menuEOL, "EOL Windows", 
			miCanCheck | miChecked, mSetting_EOL);
	  fed->miEOL_Strg_C_Korr = TMenu_AddItem(fed->menuEOL, "Strg-C EOL CRLF", 
			miCanCheck | miNotEnabled, mSetting_EOL_MissMatchKorr);
#else
	  fed->miEOL_unix = TMenu_AddItem(fed->menuEOL, "EOL Unix", 
			miCanCheck | miChecked, mSetting_EOL);
	  fed->miEOL_windos = TMenu_AddItem(fed->menuEOL, "EOL Windows", 
			miCanCheck, mSetting_EOL);
	  fed->miEOL_Strg_C_Korr = TMenu_AddItem(fed->menuEOL, "Strg-C EOL LF", 
			miCanCheck | miNotEnabled, mSetting_EOL_MissMatchKorr);
#endif

#ifndef OS_WIN32
	mi = TMenu_AddItem(fed->menuSetting, "Clipbord Mode", 0, NULL);
	  fed->menuClipbord = (TMenu*)TMenu_Create((TWins*)fed);
	  mi->next_menu = fed->menuClipbord;
	  m = miCanCheck;
	  if (appXDat.CLIPBOARD_ATOM != appXDat.XA_CLIPBOARD) m |= miChecked;
	  TMenu_AddItem(fed->menuClipbord, "X11 (XA_PRIMARY)", m, mSetting_Clipbord)->tag = 1;
	  m = miCanCheck;
	  if (appXDat.CLIPBOARD_ATOM == appXDat.XA_CLIPBOARD) m |= miChecked;
	  TMenu_AddItem(fed->menuClipbord, "QT3 (XA_CLIPBOARD)", m, mSetting_Clipbord)->tag = 2;
#endif /* OS_WIN32 */
	TMenu_AddItem(fed->menuSetting, "-", 0, NULL);
	mi = TMenu_AddItem(fed->menuSetting, "Font", 0, NULL);
	mi->pic = get_fileIcon(stft_Font_A);
	  fed->menuFont = (TMenu*)TMenu_Create((TWins*)fed);
	  mi->next_menu = fed->menuFont;
	  TMenu_AddItem(fed->menuFont, "Set Font", 0, mSetting_Font)->pic = get_fileIcon(stft_Font_A);
	  TMenu_AddItem(fed->menuFont, "Save Font as Default", 0, mSetting_FontSave)->pic = get_pic16_filesave(False);
//	TMenu_AddItem(fed->menuSetting, "Font CharSet", 0, NULL);
	fed->miUTF8 = TMenu_AddItem(fed->menuSetting, "Encoding UTF-8", 
			miCanCheck, mUTF8_OnClick);

	mi = TMenu_AddItem(fed->menuSetting, "BOM", 0, NULL);
	  fed->menuBOM = TMenu_Create((TWins*)fed);
	  mi->next_menu = fed->menuBOM;
	  TMenu_AddItem(fed->menuBOM, "none", miCanCheck, TAHedit_mBOM)->tag = fio_txtTyp_ANSI;
	  TMenu_AddItem(fed->menuBOM, "UTF8", miCanCheck, TAHedit_mBOM)->tag = fio_txtTyp_UTF8;
	  TMenu_AddItem(fed->menuBOM, "Unicode", miCanCheck, TAHedit_mBOM)->tag = fio_txtTyp_Unicode;
	  TMenu_AddItem(fed->menuBOM, "Unicode BE", miCanCheck, TAHedit_mBOM)->tag = fio_txtTyp_UnicodeBE;

	TMenu_AddItem(fed->menuSetting, "-", 0, NULL);
	mi = TMenu_AddItem(fed->menuSetting, "Line Offset", 0, NULL);
	  fed->menuLineOff = TMenu_Create((TWins*)fed);
	  mi->next_menu = fed->menuLineOff;
	  TMenu_AddItem(fed->menuLineOff, "-1", miCanCheck, mLineOff_OnClick)->tag = -1;
	  TMenu_AddItem(fed->menuLineOff, "0", miCanCheck, mLineOff_OnClick)->tag = 0;
	  TMenu_AddItem(fed->menuLineOff, "1", miCanCheck, mLineOff_OnClick)->tag = 1;
	  TMenu_AddItem(fed->menuLineOff, "2", miCanCheck, mLineOff_OnClick)->tag = 2;
	  fed->menuLineOff->OnMenuOpen = menuLineOff_OnMenuOpen;

	fed->miChkReload = TMenu_AddItem(fed->menuSetting, "Chk File Timestamp", 
			miCanCheck, TAHedit_mChkTimestap);

}
