/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

static int mblastx;
static int mblasty;
static int mblastw;
static int mblasth;

static bool ChkIsFrame(TwmFrame *wmf) {
	return wmf && (wmf->wtyp == wtyp_TwmFrame);
}

static bool ChkIsFrameCurrentDesk(TwmFrame *wmf) {
	return wmf && (wmf->wtyp == wtyp_TwmFrame)
		   && wmf->Visible
		   && (wmRoot->CurrentDesk == wmf->wmWin->Desk);
}

static void TwmFrame_SetNextFocusWin(TwmFrame *self) {
  TwmFrame *wmf = (TwmFrame*)appXDat.Wins;
  Display *dpy = appXDat.dpy;
  XID wmWinID;
  XEvent tmp_ev;

	GrabXServer();
	while (wmf) {
	    if (ChkIsFrameCurrentDesk(wmf)) {
		wmWinID = wmf->wmWinID;
		if ((XCheckTypedWindowEvent(dpy, wmWinID, DestroyNotify, &tmp_ev)) || 
		    (XCheckTypedWindowEvent(dpy, wmWinID, UnmapNotify, &tmp_ev))) {
		  XPutBackEvent(dpy, &tmp_ev);
		} else {
#if NEW_FLAGS_08400
		    if (!(wmf->wmWin->tflags & tfl_NOINPUT)) goto Ende;
#else
		    goto Ende;
#endif
		}
	    }
	    wmf = (TwmFrame*)wmf->next;
	}
#if USE_NoFocusWin
	wmWinID = NoFocusWin;
#else
	wmWinID = wmRoot->WinID;
#endif
	wmRoot->Focus_wmWin = NULL;
Ende:
 Debug_msg(0, "%s(%d) XSetInputFocus w=0x%x\n", FL, wmWinID);
	SetXInputFocus(wmWinID);
	UngrabXServer();
}

void TwmFrame_GetClientRect(TwmFrame *self, TWRect *rc) {
  int m = self->Border_width;
	rc->left = m;
	rc->top = m;
	m += m;		//Border_width * 2
	rc->width = self->width - m;
	rc->height = self->height - m;
}

static void TwmFrame_VisibleStat(TwmFrame *self, bool Visible) {
  int State = NormalState;
	if (!Visible) State = IconicState;
	TWins_SetVisible((TWins*)self, Visible);
	TwmWin_SetMapState(self->wmWin, State);
}

static void TwmFrame_transientfor_SetVisible(Window win, bool Visible) {
  TwmFrame *wmf;

	wmf = (TwmFrame*)appXDat.Wins;
	while (wmf) {
	    if ((wmf->wtyp == wtyp_TwmFrame) && wmf->transientfor) {
		if (wmf->transientfor == win) {
		    TwmFrame_transientfor_SetVisible(wmf->wmWinID, Visible);
		    TwmFrame_VisibleStat(wmf, Visible);
		}
	    }
	    wmf = (TwmFrame*)wmf->next;
	}
}

static void TwmFrame_SetVisible(TwmFrame *self, bool Value) {
  TwmWin *wmw = self->wmWin;

	if (self->Visible == Value) return;
// Debug_msg(0, "%s(%d) Set Visible 0x%x %d\n", FL, (int)wmw->WinID, Value);
	if (wmw->Desk == wmRoot->CurrentDesk) {
	    TwmFrame_transientfor_SetVisible(wmw->WinID, Value);
	    if (self->Visible) {
		TwmWin_GrabFocusWin(wmw, True);
		self->Visible = False;	//skip self for SetNextFocusWin
		TwmFrame_SetNextFocusWin(self);
		self->Visible = True;
		TTaskbar_buWin_SetUpDown(self->buTaskbarWin, False);
	    }
	    TwmFrame_VisibleStat(self, Value);
	} else {
	    if (Value) {
		self->Visible = True;
	    } else {
		TWins_SetVisible((TWins*)self, Value);
	    }
	}
}

static void TwmFrame_SetCaption(TwmFrame *self, const char *s) {
  int fuEOL;
	if ((self->wmWin->tflags & tfl_TITLE) && self->wmTitle) {
	    vds_SetStr(&self->vds, s);
	    fuEOL = 0;
	    Chk_fio_UTF8(s, pclen(s), &fuEOL);
	    if ((fuEOL & (fio_has_UTF8_char | fio_err_UTF8_char)) == fio_has_UTF8_char) {
		self->wmTitle->vds.UTF8on = True;
	    } else {
		self->wmTitle->vds.UTF8on = False;
	    }
	    SetCaption((TWins*)self->wmTitle, s);

	    if (self->buTaskbarWin) {
		self->buTaskbarWin->vds.UTF8on = self->wmTitle->vds.UTF8on;
		SetCaption((TWins*)self->buTaskbarWin, s);
	    }
	}
}

static void TwmFrame_Cursor(TwmFrame *self, int x, int y) {
  int lr;
  int curs;

  if (self->min_width  && (self->min_width == self->max_width) &&
      self->min_height && (self->min_height == self->max_height)) return;

	curs = XC_top_left_arrow;
	lr = 16;
	if (x <= self->Border_width) {
	  curs = XC_left_side;
	  if (y < lr) curs = XC_top_left_corner;
	  if (y >= self->height - lr) curs = XC_bottom_left_corner;

	} else if (y <= self->Border_width) {
	  curs = XC_top_side;
	  if (x < lr) curs = XC_top_left_corner;
	  if (x >= self->width - lr) curs = XC_top_right_corner;

	} else if (x >= (self->width - self->Border_width)) {
	  curs = XC_right_side;
	  if (y < lr) curs = XC_top_right_corner;
	  if (y >= (self->height - lr)) curs = XC_bottom_right_corner;

	} else if (y >= (self->height - self->Border_width)) {
	  curs = XC_bottom_side;
	  if (x < lr) curs = XC_bottom_left_corner;
	  if (x >= (self->width - lr)) curs = XC_bottom_right_corner;
	}

	Set_Cursor((TWins*)self, curs);
} 

static bool TwmFrame_ChkSizeCursor(TwmFrame *self) {
  int cur = self->cursor_nr;
	return ((cur == XC_left_side) || 
		(cur == XC_top_left_corner) ||
		(cur == XC_top_side) || 
		(cur == XC_top_right_corner) ||
		(cur == XC_right_side) || 
		(cur == XC_bottom_right_corner) ||
		(cur == XC_bottom_side) || 
		(cur == XC_bottom_left_corner));
}

static void TwmFrame_DoMouseOver(TwmFrame *self, XCrossingEvent *ev, bool Leave) {
	if (!Leave) TwmFrame_Cursor(self, ev->x, ev->y);
}

static void DrawLastRect(){
	XDrawRectangle(appXDat.dpy, appXDat.RootW, MoveBorderGC, 
		mblastx, mblasty, mblastw, mblasth);
} 

static void MoveOutline(Window root, int x, int y, int width, int height) {

  if ((x == mblastx) && (y == mblasty) && (width == mblastw) && (height == mblasth)) return;
  Wins_PaintOnEvEnd();
  //* undraw the old one, if any */
  if (mblastw || mblasth) DrawLastRect();
  mblastx = x;
  mblasty = y;
  mblastw = width;
  mblasth = height;
  //* draw the new one, if any */
  if (mblastw || mblasth) DrawLastRect();
}

static void TwmFrame_DoMouseMove(TwmFrame *self, XMotionEvent *ev) {
  int x, y, mx, my, mw, mh, mm;

// printf("%s(%d): %d %d\n", FL, ev->x_root, ev->y_root);

	if ((ev->state & (Button1Mask | Button2Mask | Button3Mask | Button4Mask | Button5Mask)) == 0) {
	    TwmFrame_Cursor(self, ev->x, ev->y);
	} else if ((ev->state & Button1Mask) && 
		   (appXDat.MouseDown.apph == (TWins*)self)) {
	    x = ev->x_root - appXDat.MouseDown.x_root;
	    y = ev->y_root - appXDat.MouseDown.y_root;
	    mx = appXDat.MouseDown.obj_left;
	    my = appXDat.MouseDown.obj_top;
	    mw = appXDat.MouseDown.obj_width;
	    mh = appXDat.MouseDown.obj_height;

	    if (self->cursor_nr == XC_left_side) {
		mx += x; mw -= x;
	    } else if (self->cursor_nr == XC_top_left_corner) {
		mx += x; mw -= x;
		my += y; mh -= y;
	    } else if (self->cursor_nr == XC_top_side) {
		my += y; mh -= y;
	    } else if (self->cursor_nr == XC_top_right_corner) {
		my += y; mh -= y;
		mw += x;
	    } else if (self->cursor_nr == XC_right_side) {
		mw += x;
	    } else if (self->cursor_nr == XC_bottom_right_corner) {
		mw += x;
		mh += y;
	    } else if (self->cursor_nr == XC_bottom_side) {
		mh += y;
	    } else if (self->cursor_nr == XC_bottom_left_corner) {
		mx += x; mw -= x;
		mh += y;
	    } else {
		SetRC((TWins*)self, appXDat.MouseDown.obj_left + x, appXDat.MouseDown.obj_top + y, self->width, self->height);
		return;
	    }
	    mm = TwmWin_ChkSize(self->wmWin, &mw, &mh);
	    if ((mm & (cWinsMinWidth|cWinsMaxWidth)) && 
		((self->cursor_nr == XC_left_side) ||
		 (self->cursor_nr == XC_top_left_corner) ||
		 (self->cursor_nr == XC_bottom_left_corner))) {
		mx = self->left + self->width - mw;
	    }
	    if ((mm & (cWinsMinHeight|cWinsMaxHeight)) && 
		((self->cursor_nr == XC_top_side) ||
		 (self->cursor_nr == XC_top_left_corner) ||
		 (self->cursor_nr == XC_top_right_corner))) {
		my = self->top + self->height - mh;
	    }
	    self->final_x = mx;
	    self->final_y = my;
	    self->final_w = mw;
	    self->final_h = mh;
	    MoveOutline(appXDat.RootW, mx, my, mw, mh);
	}
}

static void TwmFrame_DoMouseDown(TwmFrame *self, XButtonEvent *ev) {

	if (wmRoot->Focus_wmWin != self->wmWin) {
 Debug_msg(0, "%s(%d) XSetInputFocus w=0x%x\n", FL, self->wmWinID);
	    TwmFrame_SetOnTop(self, 0x00);
	    TwmWin_GrabFocusWin(self->wmWin, false);
	    SetXInputFocus(self->wmWinID);
	    XAllowEvents(appXDat.dpy, ReplayPointer, 0);
	    return;
	}
	TWins_DoMouseDown((TWins*)self, ev);

	if (ev->button != Button1) return;
	if (TwmFrame_ChkSizeCursor(self)) {
	    self->final_x = appXDat.MouseDown.obj_left;
	    self->final_y = appXDat.MouseDown.obj_top;
	    self->final_w = appXDat.MouseDown.obj_width;
	    self->final_h = appXDat.MouseDown.obj_height;
	    MoveOutline(appXDat.RootW, self->final_x, self->final_y, self->final_w, self->final_h);
	    XChangeActivePointerGrab(appXDat.dpy, ButtonReleaseMask | PointerMotionMask, 
			appXDat.XCursors[self->cursor_nr], 0);
	    appXDat.GrabWin = (TWins*)self;
	}
}

static void TwmFrame_DoMouseUp(TwmFrame *self, XButtonEvent *ev) {

	TWins_DoMouseUp((TWins*)self, ev);
	if (ev->button != Button1) return;

	if (TwmFrame_ChkSizeCursor(self)) {
	    MoveOutline(appXDat.RootW, 0, 0, 0, 0);
	    TwmWin_SetupSize(self->wmWin, self->final_x, self->final_y, 
		    self->final_w, self->final_h, wmwPos_SendPos);
	} else if (self->cursor_nr == XC_fleur) {
	    TwmWin_SetupSize(self->wmWin, self->left, self->top, 
		    self->width, self->height, wmwPos_SendPos);
	}
	Std_UngrabPointer();
	Set_Cursor((TWins*)self, XC_top_left_arrow);
}


void TwmFrame_ClrAll_Hilite(TwmFrame *wmf) {
  TwmFrame *twmf;

	twmf = (TwmFrame*)appXDat.Wins;
	while (twmf) {
#if NEW_FLAGS_08400
	    if ((twmf->wtyp == wtyp_TwmFrame) && (twmf != wmf)) {
		if (!(twmf->wmWin->flags & fl_GrabButton)) {
		    TwmWin_GrabFocusWin(twmf->wmWin, True);
		    TwmTitle_SetHilite(twmf->wmTitle, False);
		}
#else
	    if ((twmf->wtyp == wtyp_TwmFrame) && (twmf != wmf) && twmf->wmTitle) {
		if (twmf->wmTitle->flags & wmtfl_hilite) {
		    TwmWin_GrabFocusWin(twmf->wmWin, True);
		    TwmTitle_SetHilite(twmf->wmTitle, False);
		}
#endif
		TTaskbar_buWin_SetUpDown(twmf->buTaskbarWin, False);
	    }
	    twmf = (TwmFrame*)twmf->next;
	}
}

TwmFrame *GetTop_DeskWin(TwmFrame *wmf) {
  TwmFrame *twmf;

	if (ChkIsFrameCurrentDesk(wmRoot->Top_wmf)) return wmRoot->Top_wmf;

	twmf = (TwmFrame*)appXDat.Wins->next;
	while (twmf) {
//	    if (!(ap->aos & aosOnTop)) {
//		TwmFrame *twmf = (TwmFrame*)ap->wins;
//		if ((twmf != wmf) && ChkIsFrameCurrentDesk(twmf)) return twmf;
//	    }
	    if ((twmf != wmf) && 
		ChkIsFrameCurrentDesk(twmf) && 
		!(twmf->RunStatus & runst_OnTop)) return twmf;
	    twmf = (TwmFrame*)twmf->next;
	}

	twmf = (TwmFrame*)appXDat.Wins->next;
	while (twmf) {
//	    TwmFrame *twmf = (TwmFrame*)ap->wins;
	    if ((twmf != wmf) && ChkIsFrameCurrentDesk(twmf)) return twmf;
	    twmf = (TwmFrame*)twmf->next;
	}
	return NULL;
}

typedef struct {
  TWins **ppapi;
  Window *pwini;
  int wcount;
} TAddToRestack;

static TWins *TwmFrame_Getapp(Window win) {
  TWins *wins = appXDat.Wins->next;
	while (wins) {
	    if ((wins->WinID == win) || 
		((wins->wtyp == wtyp_TwmFrame) && 
		 (((TwmFrame*)wins)->wmWinID == win))) break;
	    wins = wins->next; 
	}
	return wins;
}

static void AddToRestack(TWins *wins, TAddToRestack *atrck) {
	if (!wins || (wins->RunStatus & runst_RestackMk)) return;
	wins->RunStatus |= runst_RestackMk;
	*atrck->ppapi = wins;
	atrck->ppapi++;
	*atrck->pwini = wins->WinID;
	atrck->pwini++;
	atrck->wcount++;
// Debug_msg(0,"onTop %d 0x%x\n", atrck->wcount, ap->win);
}

static void AddTransientFor(Window win, TAddToRestack *atrck) {
  TWins *wins;

	if (!win) return;
	wins = appXDat.Wins->next;
	while (wins) {
	    if (!(wins->RunStatus & runst_RestackMk) && 
		(ChkIsFrameCurrentDesk((TwmFrame*)wins)) &&
		(((TwmFrame*)wins)->transientfor == win)) {
		AddTransientFor(((TwmFrame*)wins)->wmWinID, atrck);
		AddToRestack(wins, atrck);
	    }
	    wins = wins->next;
	}
}

static void AddTopFrame(TWins *wins, TAddToRestack *atrck, int fu) {

	if (!wins) return;
	if (!(fu & 0x01)) {
apt_loop:
	    if (ChkIsFrame((TwmFrame*)wins) &&
		((TwmFrame*)wins)->transientfor) {
		TWins *apt = TwmFrame_Getapp(((TwmFrame*)wins)->transientfor);
		if (apt) {
		    wins = apt;
		    goto apt_loop;
		}
	    }

	    if (ChkIsFrame((TwmFrame*)wins)) {
		AddTransientFor(((TwmFrame*)wins)->wmWinID, atrck);
	    }
	}
	AddToRestack(wins, atrck);
}


void TwmFrame_SetOnTop(TwmFrame *wmf, int fu) {
  TWins *ap, *apTop, *apt, **ppap;
  Window *pwin;
  TAddToRestack atrck;
  int wmaxcount;

	wmRoot->Top_wmf = wmf;
	wmaxcount = 0;
	ap = appXDat.Wins->next;
	while (ap) {ap->RunStatus &= ~runst_RestackMk; ap = ap->next; wmaxcount++;}
	if (wmaxcount < 2) return;

	ppap = xmalloc(sizeof(void*) * (wmaxcount+2));
	pwin = xmalloc(sizeof(void*) * (wmaxcount+2));
	atrck.ppapi = ppap;
	atrck.pwini = pwin;
	atrck.wcount = 0;

	appXDat.Wins->RunStatus |= runst_RestackMk;
	*atrck.ppapi = appXDat.Wins;         //add wmRoot
	atrck.ppapi++;

	if (wmRoot->Taskbar && wmRoot->Taskbar->Visible) {
	    ap = TwmFrame_Getapp(wmRoot->Taskbar->WinID);
	    if (ap) AddToRestack(ap, &atrck);
	}

	apTop = NULL;
	if (wmf) apTop = TwmFrame_Getapp(wmf->WinID);
	if (apTop && (apTop->RunStatus & runst_OnTop)) {
	    AddTopFrame(apTop, &atrck, fu);
	}

	ap = appXDat.Wins->next;
	while (ap) {
	    if ((ap->RunStatus & runst_OnTop) && !(ap->RunStatus & runst_RestackMk)) {
		if (ChkIsFrameCurrentDesk((TwmFrame*)ap)) {
		    AddToRestack(ap, &atrck);
		}
	    }
	    ap = ap->next;
	}

	AddTopFrame(apTop, &atrck, fu);

	*atrck.pwini = 0;
	XRaiseWindow(appXDat.dpy, *pwin);
	XRestackWindows(appXDat.dpy, pwin, atrck.wcount);

	ap = appXDat.Wins->next;
	while (ap) {
	    if (!(ap->RunStatus & runst_RestackMk)) {
		*atrck.ppapi = ap;
		atrck.ppapi++;
// atrck.wcount++;
	    }
	    ap = ap->next;
	}
	*atrck.ppapi = NULL;

	TWins **ppapn = ppap;
	apt = NULL;
	ap = *ppapn;
	appXDat.Wins = ap;
	while (ap) {
	    ap->prev = apt;
	    apt = ap;
	    ppapn++;
	    ap = *ppapn;
	    apt->next = ap;
	}

// Debug_msg(0,"SetOnTop %d %d\n", wmaxcount, atrck.wcount);

/*
#ifdef	WTestWin
 int i;
 char bufint[32];
 char *s;
 Window *nwin = pwin;
 
  WrInteger(atrck.wcount, &bufint[0]);
  s = pcdup(&bufint[0]);
  ppcadd(&s, " ");
  for (i=0;i<atrck.wcount; i++) {
    WrInteger((int)*nwin, &bufint[0]);
    ppcadd(&s, &bufint[0]);
    ppcadd(&s, " ");
    nwin++;
  }
 WTest_msg_Add("RS %s", s);
 xfree(s);
#endif
*/
	xfree(pwin);
	xfree(ppap);
}


/*____________________________ Destroy __________________________________*/
static bool TwmFrame__Destroy(TwmFrame *self) {
	if (wmRoot->Top_wmf == self) wmRoot->Top_wmf = NULL;
	TTaskbar_DelWin(wmRoot->Taskbar, self);
	self->Visible = False;
	TwmFrame_SetNextFocusWin(self);
	self->wmWin = NULL;
	TWins__Destroy((TWins*)self);
// Debug_msg(0, "%s(%d) Destroy TwmFrame\n", FL);
	return False;
}

/*____________________________ Create ___________________________________*/
TwmFrame *TwmFrame_Create(TwmWin *wmw, XWindowAttributes *attr) {
  TwmFrame *self;
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TwmFrame), wtyp_TwmFrame);
	wcs.CreateXWin = NULL;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;

	self->owner = (TWins*)wmw;
	self->wmWin = wmw;
	self->wmWinID = wmw->WinID;
	self->top = attr->y;
	self->left = attr->x;
	self->width = attr->width;
	self->height = attr->height;

	if (wmw->tflags & tfl_BORDER) {
	    self->Border_width = glob_Border_width;
	    self->width += glob_Border_width << 1;
	    self->height += glob_Border_width << 1;
	    self->BorderFu = mBorder | mBorderW2;
	}
	if (wmw->tflags & tfl_TITLE) {
	    self->Caption_height = glob_Caption_height;
	    self->height += glob_Caption_height;
	}

//	self->apph->aos |= aosIsForm;
	if (!(wmw->tflags & tfl_BORDER)) self->Paint = NULL;

	if (XGetTransientForHint(appXDat.dpy, wmw->WinID,  &self->transientfor) != 0) {
	    if ((self->transientfor == appXDat.RootW) || 
		(self->transientfor == wmw->WinID)) self->transientfor = 0;
	    if (self->transientfor) wmw->tflags &= ~tfl_MINIMIZE;
	}

	self->destroy = (TDESTROY)TwmFrame__Destroy;
	self->GetClientRect = TwmFrame_GetClientRect;
	self->SetVisible = TwmFrame_SetVisible;
	self->SetCaption = TwmFrame_SetCaption;

	self->DoMouseMove = TwmFrame_DoMouseMove;
	self->DoMouseOver = TwmFrame_DoMouseOver;
	self->DoMouseDown = TwmFrame_DoMouseDown;
	self->DoMouseUp = TwmFrame_DoMouseUp;


	wcs.XAttr.event_mask = (SubstructureRedirectMask | ButtonPressMask |
			   ButtonReleaseMask | EnterWindowMask |
			   LeaveWindowMask | ExposureMask | PointerMotionMask);

	if (attr->save_under != 0) {
	    wcs.XAttr.save_under = 1;
	    wcs.XMask |= CWSaveUnder;
	}
#ifdef USE_DECOR
	if (wmw->tflags & tfl_BORDER) self->Decor = TDecor_dup(&wmRoot->Default_Decor);
#endif
	self->Visible = False;
	TWins_CreateXWin(&wcs);

	Set_Parent((TWins*)self, NULL);

	if (wmw->tflags & tfl_TITLE) {
	    self->wmTitle = TwmTitle_Create(self);
	    if (wmw->tflags & tfl_MINIMIZE) TTaskbar_AddWin(wmRoot->Taskbar, self);
	    if (wmw->tcreate) SetCaption((TWins*)self, wmw->tcreate->WMName);
	}

	return self;
}
