/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"
#include <X11/extensions/shape.h>

TNetWMAtom NetWMAtom;

static const char *NetWM_AtomsTab[] = {
  "_NET_WM_STATE_ABOVE",
  "_NET_WM_STATE_BELOW",
  "_NET_WM_STATE_MAXIMIZED_VERT",
  "_NET_WM_STATE_MAXIMIZED_HORZ",
  "_NET_WM_STATE_FULLSCREEN",
  NULL
};

TNetWMnum GetNetWMnum(Atom a) {
	if (NetWMAtom.XA_NET_WM_STATE_ABOVE == a) return NET_WM_STATE_ABOVE;
	if (NetWMAtom.XA_NET_WM_STATE_BELOW == a) return NET_WM_STATE_BELOW;
	return NET_WM_NONE;
}


#ifdef DRAG_AND_DROP
static int TwmRoot_DodndMouseMove(TWins *self, XMotionEvent* ev) {
// printf("%s TwmRoot_DodndMouseMove(%d): w=0x%x\n", FL, (int)self->WinID);
	return 3;
}

static int TwmRoot_DodndDrop(TWins *self) {
//  Atom SelAtom;
  Tdnd *dnd = appXDat.dnd;

 AzAtom(FL, "Sel-Atoms",dnd->XA_XdndSelection);
 fflush(stdout);
	appXDat.SelWait = True;
//	appXDat.SelWin = ((TwmRoot*)self)->xdnd_widget->WinID;
// printf("%s(%d): \n", FL);
//	SelAtom = dnd_AisText();
//	if (SelAtom) {
		if (GetPasteDaten(dnd->XA_XdndSelection, appXDat.XA_URL_STRING)) {
 printf("%s(%d): %d %s\n", FL, appXDat.SelRet_lgs.lg, appXDat.SelRet_lgs.s);
 fflush(stdout);
		}
//	}
	return 0;
}

#endif


void TwmRoot_ClearStartNotify(int pid) {
  TPid_Array *pida;
  Texec_pipe *expipe;

	if (!pid || (pid == wmRoot->DeskStartPid)) {
	    pid = wmRoot->DeskStartPid;
	    wmRoot->DeskStartPid = 0;
	}
	if (!wmRoot->DeskStartPid) Std_UngrabPointer();
	if (pid > 0) {
	    pida = Get_pid_Handle(pid);
	    if (pida) {
		expipe = pida->obj;
		if (ttc_chk_is_wtyp(expipe, wtyp_Texec_pipe)) {
		    ppTfree(&expipe->p1);
		    if ((expipe->pp[0].fd <= 0) &&
			(expipe->pp[1].fd <= 0) &&
			(expipe->pp[2].fd <= 0) &&
			(expipe->pp[3].fd <= 0) &&
			(expipe->pp[4].fd <= 0)) {
			Free_pid_Handle(pid, 0);
		    }
		}
	    }
	}
}

/* _______________________ Notify ______________________________________ */
static void TwmRoot_global_OnFileMsg(TFileMsg *fmsg) {
  TwmRoot *wmr;
  TpcIni *Ini;
  int bef;
  TDirFileList *dfl;

	bef = get_bef_dflEv(fmsg, NULL);
	if (!bef) return;
	dfl = fmsg->F.dfl;
	wmr = ((TDirListObjNy*)fmsg->obj)->owner;
	if ((bef == dlbef_Add) || (bef == dlbef_WrClose)) {

	    Ini = NULL;
	    if (pccmp(dfl->fn, "global.ini")) {
		Ini = TpcIni_Create(appDat.inifile_global);
		 wmr->DateFormat = TpcIni_RD_int(Ini, "Global", "DateFormat", 0);

	    } else if (pccmp(dfl->fn, "aldesk.ini")) {
		Ini = TpcIni_Create(appDat.inifile_app);
		wmr->BorderSnap = TpcIni_RD_int(Ini, "Style", "BorderSnap", 4);
		if (wmr->BorderSnap > 9) wmr->BorderSnap = 9;
		if (wmr->BorderSnap < 0) wmr->BorderSnap = 0;
		TwmRoot_SetUpFont(wmr, Ini);
		TwmRoot_SetBuildinStr(wmr, Ini);
		TwmRoot_SetBackground(wmr, Ini);
#ifdef USE_DECOR
		TwmFrameBorder_Create(wmr, Ini);
#endif
		SetUpKey(Ini);
	    }
	    Tfree(Ini);
	}
}

void TwmRoot_global(TwmRoot *wmr) {
  TDirListObjNy *ObjNy;

	ObjNy = TDirListObjNy_Create(wmr);
	ObjNy->OnFileMsg = TwmRoot_global_OnFileMsg;
	ObjNy->findAtr = findREG;
	if (!TDirListObjNy_Open(ObjNy, appDat.ALDini_dir)) {
	    Tfree(ObjNy);
	} else TDirListObjNy_Open_Wait(ObjNy);
}

/*_____________________________ TwmRoot Ev ______________________________*/
static void TwmRoot_OnMouseMove(TwmRoot *self, XMotionEvent *ev) {
	if ((self->di_glob.mod & (wmDImod_HighLight | wmDImod_Transparent)) == 0) return;
	TDesktopIcon_ScanMousePos((TWins*)self, ev, self->next_child);
}

static void TwmRoot_GetClientRect(TwmRoot *self, TWRect *rc) {
	rc->left = 0;
	rc->top = 0;
	rc->width = appXDat.Root_width;
	rc->height = appXDat.Root_height;
}

/*_____________________________ Create __________________________________*/
static const char titleH_AHico[] = {
 0x04,0x08,0x10,0x00,0x12,0x00,0x00,0x00,0x48,0x00,0x12,0x21,0x94,0x31,0x16,0x42,
 0x98,0x52,0x1A,0x63,0x9C,0x73,0x1E,0x84,0x9F,0x94,0xBE,0xE8,0x0D,0xE0,0x36,0x22,
 0xC5,0x56,0x38,0x50,0xED,0x92,0x30,0x5A,0x6A,0x81,0x6F,0xD7,0x50,0xD7,0xEC,0xC1,
 0xBD,0x07,0xB3,0x77,0x0D,0x5D,0xBF,0x05,0x6A,0x69,0x09,0x4B,0xD6,0x0E,0x84,0xB3,
 0x95,0x44,0xBA,0x0D,0xA0,0x37,0xBE,0x00};
static const char titleL_AHico[] = {
 0x04,0x08,0x10,0x00,0x12,0x00,0x00,0x00,0x48,0x00,0x92,0x94,0xD3,0x9C,0x14,0xA5,
 0x55,0xAD,0x96,0xB5,0xD7,0xBD,0x18,0xC6,0x59,0xCE,0xBE,0xE8,0x0D,0xE0,0x36,0x22,
 0xC5,0x56,0x38,0x50,0xED,0x92,0x30,0x5A,0x6A,0x81,0x6F,0xD7,0x50,0xD7,0xEC,0xC1,
 0xBD,0x07,0xB3,0x77,0x0D,0x5D,0xBF,0x05,0x6A,0x69,0x09,0x4B,0xD6,0x0E,0x84,0xB3,
 0x95,0x44,0xBA,0x0D,0xA0,0x37,0xBE,0x00};

TwmRoot *TwmRoot_Create(TWins *ws) {
  TwmRoot *self;
//  Tapph *ap;
  Display *dpy = appXDat.dpy;

	self = ttc_create(sizeof(TwmRoot), wtyp_TwmRoot);
	self->WinID = appXDat.RootW;
//	ap = AddAppWin(NULL, (TWins*)self);
	TWins_Add((TWins*)self, NULL);
	self->EvProc = TWins_EvProc;
//	ap->aos |= aosIsForm;
	self->Color = appXDat.Default_Color;
	self->vds.Font = appXDat.Default_Font;

	self->width = appXDat.Root_width;
	self->height = appXDat.Root_height;
	self->Enabled = True;
	self->Visible = True;
	self->cursor_nr = -1;

	self->DoMouseDown = (void*)TWins_DoMouseDown;
	self->DoMouseUp = (void*)TWins_DoMouseUp;
	self->DoMouseMove = (void*)TWins_DoMouseMove;
	self->OnMouseMove = TwmRoot_OnMouseMove;
	self->GetClientRect = TwmRoot_GetClientRect;


	self->XA_ALDesk_DESTROY = XInternAtom(dpy, "XA_ALDesk_DESTROY", 0);

  Atom *at = &NetWMAtom.XA_NET_WM_STATE_ABOVE;
  const char **ats = NetWM_AtomsTab;
	while(*ats) {
	    *at = XInternAtom(dpy, *ats, 0);
	    at++; ats++;
	}

	XSelectInput(dpy, appXDat.RootW,
        	LeaveWindowMask | EnterWindowMask | PropertyChangeMask |
        	SubstructureNotifyMask | SubstructureRedirectMask | 
		KeyPressMask | KeyReleaseMask |
        	ButtonPressMask | ButtonReleaseMask | PointerMotionMask);

	Set_Cursor((TWins*)self, XC_X_cursor);

	self->DesktopPath = pcadd(appDat.ALDini_dir, "/Desktop/");
	file_mkdir_parents_ae(self->DesktopPath, 0750, EMSG_ERR_EXIT);
	self->IconPosFile = pcadd(self->DesktopPath, ".IconPositions");

#ifdef DRAG_AND_DROP
  Tdnd *dnd;
  THdnd *Hdnd = xcalloc(sizeof(THdnd));
	self->Hdnd = Hdnd;
#ifdef USE_SEL_ROOT_PROXY
  int proxy_wid, version;
	Tdnd_Create();
	dnd = appXDat.dnd;
	self->xdnd_widget = TWins_Create((TWins*)self);
	proxy_wid = self->xdnd_widget->WinID;
	XChangeProperty(dpy, appXDat.RootW, dnd->XdndProxy,
	    XA_WINDOW, 32, PropModeReplace, (unsigned char *)&proxy_wid, 1 );
	XChangeProperty(dpy, proxy_wid, dnd->XdndProxy,
	    XA_WINDOW, 32, PropModeReplace, (unsigned char *)&proxy_wid, 1 );

	version = dnd_version;
	XChangeProperty (dpy, proxy_wid, dnd->XA_XdndAware, 
	    XA_ATOM, 32, PropModeReplace, (unsigned char*)&version, 1);

#else
  int version;
	Tdnd_Create();
	dnd = appXDat.dnd;
	version = dnd_version;
	XChangeProperty (dpy, appXDat.RootW,
		appXDat.dnd->XA_XdndAware, XA_ATOM,
		32, PropModeReplace,
		(unsigned char*)&version, 1);

#endif /* USE_SEL_ROOT_PROXY */
	Hdnd->OndndMouseMove = TwmRoot_DodndMouseMove;
	Hdnd->OndndDrop = TwmRoot_DodndDrop;
#endif /* DRAG_AND_DROP */

	self->AutostartPath = pcadd(appDat.ALDini_dir, "/Autostart");
	file_mkdir_parents_ae(self->AutostartPath, 0750, EMSG_ERR1);
	self->MenuPath = pcadd(appDat.ALDini_dir, "/Menu");
	file_mkdir_parents_ae(self->MenuPath, 0750, EMSG_ERR_EXIT);

	menu_start(self);

	menu_desktop(self);

	menu_FormOpt(self);

	menu_DeskIcon(self);

//	ap->win = self->WinID;
	self->vds.WinID = self->WinID;
	self->vds.Wins = (TWins*)self;

  TpcIni *pcIni;
	pcIni = TpcIni_Create(appDat.inifile_global);

	self->AutoMountPath = TpcIni_RDstrdup(pcIni, "AutoMount", "MountPath", "/media");
	self->sysblockPath = TpcIni_RDstrdup(pcIni, "AutoMount", "sysblockPath", "/sys/block");

	Tfree(pcIni);

	pcIni = TpcIni_Create(appDat.inifile_app);

	TwmRoot_SetUpFont(self, pcIni);

	self->di_glob.TwColor = 0xc0c0c0;
	self->di_glob.TwHlColor = 0xd0d0d0;
	self->di_glob.TwFocColor = 0x0080d0;
	self->di_glob.TwFontColor = 0xffffff;
	self->di_glob.TwHlFontColor = 0x000000;
	self->di_glob.TwFocFontColor = 0xffffff;
	self->di_glob.DesktopIcoFontName = TpcIni_RDstrdup(pcIni, "Style", 
		"DesktopIcoFont", appXDat.Default_Font->name), 
	self->di_glob.DesktopIcoFont = MakeFont(self->di_glob.DesktopIcoFontName,
				appXDat.Default_Font->name, NULL);
	SetFont_Default(self->di_glob.DesktopIcoFont, True);

	self->TitleH = CreateAHicoPic(titleH_AHico, NULL);
	self->TitleL = CreateAHicoPic(titleL_AHico, NULL);

	Tfree(pcIni);
	TWinsRegistration((TWins*)self);
	return self;
}
