/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"
#include <X11/extensions/shape.h>

static void buWinOptMenu_OnPaint(TWins *self) {
  TwmTitle *wmt = (TwmTitle*)self->prev_child;
	DrawPic(self->WinID, wmt->pic, 0, 0);
}

static void buWinOptMenu_OnMouseDown(TWins *self, XButtonEvent *ev) {
  TwmFrame *wmf = (TwmFrame*)self->prev_child->prev_child;

	exec_Buildin((TWins*)wmf->wmWin, "menu_form_opt", NULL, cond_0);
}

static void buMinimize_OnPaint(TButton *self) {
	DrawLine(self->WinID, appXDat.GC_000000, 3, 9, 11, 9);
	DrawLine(self->WinID, appXDat.GC_000000, 3, 10, 11, 10);
}

static void buMinimize_OnClick(TButton *self, long KeyState, bool DblClick) {
	SetVisible(self->prev_child->prev_child, False);
}

static void buMaximize_OnPaint(TButton *self) {
	DrawRect(self->WinID, appXDat.GC_000000, 2, 2, 12, 10);
	DrawLine(self->WinID, appXDat.GC_000000, 3, 3, 12, 3);
}

static void buMaximize_OnClick(TButton *self, long KeyState, bool DblClick) {
  TwmFrame *wmf = (TwmFrame*)self->prev_child->prev_child;
	exec_Buildin((TWins*)wmf->wmWin, "minmax_win", NULL, cond_0);
}

static void buClose_OnPaint(TButton *self) {
	DrawLine(self->WinID, appXDat.GC_000000, 3, 3, 10, 10);
	DrawLine(self->WinID, appXDat.GC_000000, 4, 3, 11, 10);
	DrawLine(self->WinID, appXDat.GC_000000, 10, 3, 3, 10);
	DrawLine(self->WinID, appXDat.GC_000000, 11, 3, 4, 10);
}

static void buClose_OnClick(TButton *self, long KeyState, bool DblClick) {
  TwmFrame *wmf = (TwmFrame*)self->prev_child->prev_child;

	TwmWin_Close(wmf->wmWin);
}

static void TwmTitle_LoadIcoName(TwmTitle *self, char *name) {
  char *pc;
  int fu;
	if (!name || !*name) return;
	pc = name;
	while (*pc) {
	    if ((byte)*pc <= 0x2f) return;
	    pc++;
	}

	fu = gdifu_noUnknown;
	if (self->mpic.pic) fu = gdifu_noError|gdifu_noUnknown;

	self->pic = GetDeskIcon(ddkt_Application, 16, name, fu, NULL);
	if (self->pic) return;

  char *pc0 = pcdup(name);
  bool mk = False;
	pc = pc0;
	while (*pc) {
	    if ((*pc >= 'A') && (*pc <= 'Z')) {
		*pc += 0x20;
		mk = True;
	    }
	    pc++;
	}
	if (mk) self->pic = GetDeskIcon(ddkt_Application, 16, pc0, fu, NULL);
	xfree(pc0);
}

static void TwmTitle_Ico_32To16(TwmTitle *self, Pixmap pmxID, Pixmap maskID) {
  TPixMap *pxm, *pmx16;

	pxm = TPixMap_Create_XID(pmxID, maskID);

	if (!pxm) return;
	if (pxm->maskID) {
	    TPixMap_XGetImage(pxm);
	    bitmask2alpha((TImage*)pxm);
	}
	pmx16 = TPixMap_dup_new_size(pxm, 16, 16);
	pxm->pmxID = pxm->maskID = 0;
	Tfree(pxm);

	TPixMap_XPutImage(pmx16);

	self->mpic.pic = pmx16->pmxID;
	self->mpic.mask = pmx16->maskID;
	self->mpic.width = 16;
	self->mpic.height = 16;
	self->pic = &self->mpic;
	pmx16->pmxID = pmx16->maskID = 0;
	Tfree(pmx16);
}

static char *TwmTitle_vds_path_cut(TwmTitle *self, const char *s) {
	int max_w = self->width - self->CaptionOffX - 54;
	if (max_w < 120) max_w = 120;
	return vds_path_cut(&self->vds, s, max_w);
}

static void TwmTitle_OnResize(TwmTitle *self) {
  int left;

	left = self->width - self->buClose->width - 2;
	self->buClose->left = left;
	if (self->buMaximize) {
	    left -= self->buMaximize->width + 2;
	    self->buMaximize->left = left;
	}
	if (self->buMinimize) {
	    left -= self->buMinimize->width + 1;
	    self->buMinimize->left = left;
	}

	char *pc = TwmTitle_vds_path_cut(self, ((TwmFrame*)self->prev_child)->vds.s);
	if (!pccmp(pc, self->vds.s)) vds_SetStr(&self->vds, pc);
	xfree(pc);
}

static void TwmTitle_DoPaint(TwmTitle *self) {

	self->vds.Font = wmRoot->wmTitleFont;
	if ((self->vds.UTF8on) && (wmRoot->wmTitleFontUTF8)) {
	    self->vds.Font = wmRoot->wmTitleFontUTF8;
	}

	self->vds.Xoff = self->CaptionOffX;
	self->vds.Yoff = (self->height - self->vds.Font->height) >> 1;
	self->vds.Yoff += self->CaptionOffY;
	self->vds.BackColor = self->Color;
	if (self->flags & wmtfl_hilite) {
	    self->vds.ForeColor = 0x000000; //SubColor(self->Color, 0x808080);
	    DrawColorString(1, 1, &self->vds);
	    self->vds.ForeColor = 0xffffff; //AddColor(self->Color, 0xffffff);
	} else {
	    self->vds.ForeColor = 0x808080; //AddColor(self->Color, 0x606060);
	    DrawColorString(1, 1, &self->vds);
	    self->vds.ForeColor = 0x404040; //SubColor(self->Color, 0x303030);
	}
	DrawColorString(0, 0, &self->vds);
}

static void TwmTitle_DoMouseMove(TwmTitle *self, XMotionEvent *ev) {
  TwmFrame *wmf = (TwmFrame*)self->prev_child;
  int x, y, m, Bsnap;

	Bsnap = wmRoot->BorderSnap;
	if ((ev->state & Button1Mask) && 
	    (appXDat.MouseDown.apph == (TWins*)self)) {
	    x = appXDat.MouseDown.obj_left + (ev->x_root - appXDat.MouseDown.x_root);
	    y = appXDat.MouseDown.obj_top + (ev->y_root - appXDat.MouseDown.y_root);

	    m = appXDat.Root_width - wmf->width;
	    if ((x >= -Bsnap) && (x <= Bsnap)) {
		x = 0;
	    } else if ((x > m-Bsnap) && (x < m+Bsnap)) {
		x = m;
	    }

	    m = appXDat.Root_height - wmf->height;
	    if (wmRoot->Taskbar) m -= wmRoot->Taskbar->height;
	    if ((y >= -Bsnap) && (y <= Bsnap)) {
		y = 0;
	    } else if ((y > m-Bsnap) && (y < m+Bsnap)) {
		y = m;
	    }

	    TwmWin_SetupSize(wmf->wmWin, x, y, wmf->width, wmf->height, 0); 
	}
}

static void TwmTitle_DoMouseDown(TWins *self, XButtonEvent *ev) {
	TWins_DoMouseDown(self, ev);
	if (ev->button != Button1) return;
	appXDat.MouseDown.obj_left = self->prev_child->left;
	appXDat.MouseDown.obj_top = self->prev_child->top;
	appXDat.MouseDown.obj_width = self->prev_child->width;
	appXDat.MouseDown.obj_height = self->prev_child->height;
	Set_Cursor(self, XC_fleur);
}

static void TwmTitle_DoMouseUp(TWins *self, XButtonEvent *ev) {
  TwmFrame *wmf = (TwmFrame*)self->prev_child;
	TWins_DoMouseUp(self, ev);
	if (ev->button != Button1) return;
	Set_Cursor(self, XC_top_left_arrow);
	TwmWin_SetupSize(wmf->wmWin, wmf->left, wmf->top, wmf->width, wmf->height, wmwPos_SendPosAlways);
}

void TwmTitle_SetHilite(TwmTitle *wmt, bool Hilite) {
  Pixmap pic;

	if (!wmt) return;
	wmt->Enabled = Hilite;
	if (wmt->Enabled) {
	    wmt->flags |= wmtfl_hilite;
	    pic = wmRoot->TitleH->pic;
	} else {
	    wmt->flags &= ~wmtfl_hilite;
	    pic = wmRoot->TitleL->pic;
	}
	XSetWindowBackgroundPixmap(appXDat.dpy, wmt->WinID, pic);
	TWins_RePaint((TWins*)wmt);
}

static void TwmTitle_SetCaption(TwmTitle *self, const char *s) {
	char *pc = TwmTitle_vds_path_cut(self, s);
	TWins_SetCaption((TWins*)self, pc);
	xfree(pc);
	XClearWindow(appXDat.dpy, self->WinID);
}

/*____________________________ Destroy __________________________________*/
static bool TwmTitle__Destroy(TwmTitle *self) {
	self->vds.Font = NULL;
	ClrPic(&self->mpic);
	return TWins__Destroy((TWins*)self);
}

/*____________________________ Create __________________________________*/
static TButton *Create_title_Button(TwmTitle *self) {
  TButton *ws;
	ws = TButton_CreateS((TWins*)self, 24, 2, 16, 14, NULL);
	ws->BorderFu = mBorder;
	ws->KeyFu = 0;
	return ws;
}

TwmTitle *TwmTitle_Create(TwmFrame *wmf) {
  TwmTitle *self;
  TWins_Create_struct wcs;
  TwmWin *wmw, *twmw;
  TwmWin_tcreate *tcreate;
  TGeometry gm;

	TWins_Create_struct_init(&wcs, sizeof(TwmTitle), wtyp_TwmTitle);
	wcs.Parent = (TWins*)wmf;
	wcs.CreateXWin = NULL;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;

	self->Align = alTop;
	self->height = glob_Caption_height;
	self->CaptionFu = mCaptionPaint | mCaptionLeft;
	self->CaptionOffX = 19;

	self->destroy = (TDESTROY)TwmTitle__Destroy;
	self->OnResize = TwmTitle_OnResize;
	self->DoPaint = TwmTitle_DoPaint;
	self->DoMouseMove = TwmTitle_DoMouseMove;
	self->DoMouseDown = (void*)TwmTitle_DoMouseDown;
	self->DoMouseUp = (void*)TwmTitle_DoMouseUp;
	self->SetCaption = TwmTitle_SetCaption;

	TWins_CreateXWin(&wcs);
	Set_Parent((TWins*)self, (TWins*)wmf);

	self->buWinOptMenu = TWins_CreateB((TWins*)self, 1, 1, 16, 16, 0);
//	self->buWinOptMenu->KeyFu = 0;
	self->buWinOptMenu->OnMouseDown = buWinOptMenu_OnMouseDown;
	self->buWinOptMenu->OnPaint = buWinOptMenu_OnPaint;

	wmw = wmf->wmWin;
	if (wmw->tflags & tfl_MINIMIZE) {
	    self->buMinimize = Create_title_Button(self);
	    self->buMinimize->OnClick = buMinimize_OnClick;
	    self->buMinimize->OnPaint = buMinimize_OnPaint;
	}
	if (wmw->tflags & tfl_MINMAXIMIZE) {
	    self->buMaximize = Create_title_Button(self);
	    self->buMaximize->OnClick = buMaximize_OnClick;
	    self->buMaximize->OnPaint = buMaximize_OnPaint;
	}
	self->buClose = Create_title_Button(self);
	self->buClose->OnClick = buClose_OnClick;
	self->buClose->OnPaint = buClose_OnPaint;

	tcreate = wmw->tcreate;
	if (tcreate) {
	    XWMHints *WMHints = tcreate->WMHints;
	    if (WMHints) {
	      if (WMHints->flags & IconPixmapHint) {
		  if (get_geometry(WMHints->icon_pixmap, &gm)) {
		    if ((gm.depth == 1) || ((gm.depth >= 16) && (gm.depth == appXDat.RootW_depth))) {
		      if ((gm.width > 7) && (gm.height > 7) && (gm.width < 65) && (gm.height < 65)) {
			  if (!(WMHints->flags & IconMaskHint)) WMHints->icon_mask = 0;
			  TwmTitle_Ico_32To16(self, WMHints->icon_pixmap, WMHints->icon_mask);
		      }
		    }
		  }
  Debug_msg(0, "WMHints IconPixmap=0x%x w=%d h=%d Depth=%d\n", WMHints->icon_pixmap, gm.width, gm.height, gm.depth);
	      }
	    }
	    TPic *pic = self->pic;
	    self->pic = NULL;
	    TwmTitle_LoadIcoName(self, tcreate->icon_name);
	    if (!self->pic) TwmTitle_LoadIcoName(self, tcreate->WMIconName);
	    if (self->pic) {
		ClrPic(&self->mpic);
	    } else {
		self->pic = pic;
	    }
	}

	if (!self->pic && wmf->transientfor) {
	  twmw = TwmWin_Find_WinsX_Ev(wmf->transientfor);
	  if (twmw) {
	    TwmFrame *nwmf = twmw->Frame;
	    if (nwmf && nwmf->wmTitle && nwmf->wmTitle->pic) {
		TwmTitle_Ico_32To16(self, nwmf->wmTitle->pic->pic, nwmf->wmTitle->pic->mask);
	    }
	  }
	}

	if (!self->pic && wmw->window_group) {
	  TwmFrame *nwmf = (TwmFrame*)appXDat.Wins;
	  while (nwmf) {
	    if (nwmf->next_child && (nwmf->wtyp == wtyp_TwmFrame)) {
	      if (nwmf->wmWin && nwmf->wmWin->window_group &&
		  (nwmf->wmWin->window_group == wmw->window_group) && 
		  nwmf->wmTitle && nwmf->wmTitle->pic) {
		TwmTitle_Ico_32To16(self, nwmf->wmTitle->pic->pic, nwmf->wmTitle->pic->mask);
		break;
	      }
	    }
	    nwmf = (TwmFrame*)nwmf->next;
	  }
	}

	if (self->pic && self->pic->mask) {
	    CombineMask(self->buWinOptMenu->WinID, self->pic->mask);
	}

	SetColor((TWins*)self, 0xa0a0a0);
	return self;
}

