/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldbatmon.h"
#include "cicons.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/hdreg.h>

static void TAHbatmon_pnlHD_DoPaint(TAHbatmon_pnlHD *self) {
  TPic *pic;
	pic = get_pic16_hdd_umount(False);
	DrawPic(self->WinID, pic, 2, 22);
	DrawPic(self->WinID, pic, 2, 42);
	DrawPic(self->WinID, pic, 2, 62);
}

// ************** set HD standby **************
static void SetUP_HDtime(TAHbatmon_pnlHD *pnlHD, char *dev, int ti) {
  int fd;
  char line[64];

	if (pccmp(dev, "none")) {
	    pnlHD->akt_HD0_time = 0;
	    return;
	}

	snprintf(line, sizeof(line), "/dev/%s", dev);
	fd = open(line, O_RDONLY | O_NONBLOCK);
	if (fd > 0) {
  unsigned char args[4] = {WIN_SETIDLE1, ti, 0, 0};
	    if (ioctl(fd, HDIO_DRIVE_CMD, &args)) {
 printf("%s(%d): ERROR: HDIO_DRIVE_CMD(setidle) failed: %s\n", FL, SysError_pc());
	    }
	    close(fd);
	} else {
 printf("%s(%d): ERROR: %s\n", FL, SysError_pc());
	}
}

static void SetHD_OnClick(TMenu *menu, long KeyState, TMenuItem *mi) {
	AHbatmon_buApply_ChkEnabled(((TAHbatmon_pnlHD*)menu->subClass)->subClass_Prev);
}

static void HD_SetMods(TAHbatmon_pnlHD *pnlHD) {
  int di, x, fd, k;
  FILE *f;
  byte *tline;
  byte line[1024];
  char *devs[16];
  byte args[4] = {WIN_CHECKPOWERMODE1,0,0,0};
  TpcIni *pcIni = appDat.appIni;
  const char *HD0dev;
  int HD0AC = 0;
  int HD0Bat = 0;
  bool mk, Ok;

	HD0dev = "none";
	if (pcIni) {
	    HD0dev = TpcIni_RD_pc(pcIni, "acpi", "HD0_dev_standby", "none");
	    HD0AC = TpcIni_RD_int(pcIni, "acpi", "HD0_AC_standby", 0);
	    HD0Bat = TpcIni_RD_int(pcIni, "acpi", "HD0_Bat_standby", 0);
	}

	f = fopen("/proc/partitions", "r");	
	if (f) {
	    fgets((char*)line, sizeof(line)-1, f);
	    fgets((char*)line, sizeof(line)-1, f);
	} else {
	    f = fopen("/proc/diskstats", "r");
	    if (!f) return;
	}
	di = 0;
loop:
	tline = (byte*)fgets((char*)line, sizeof(line)-1, f);
	if (tline) {
	    while (*tline && (*tline <= '9')) tline++;
	    if (((tline[1] == 'd') && tline[2] && (tline[3] <= ' ')) &&
		((*tline == 'h') || (*tline == 's'))) {
		devs[di] = pcndup((char*)tline, 3);
		di++;
	    }
	    if (di < 16) goto loop;
	}
	fclose(f);

	TComboBox_AddItem(pnlHD->cb_dev_HD0, "none", miAdd_setSel, SetHD_OnClick);
	x = 0;
	Ok = False;
	while (x < di) {
	    mk = True;
	    snprintf((char*)line, sizeof(line), "/sys/block/%s/removable", devs[x]);
	    f = fopen((char*)line, "r");
	    if (f) {
		tline = (byte*)fgets((char*)line, sizeof(line)-1, f);
		mk = (*tline == '0');
		fclose(f);
	    }
	    if (mk) {	//check HD has powermode
		snprintf((char*)line, sizeof(line), "/dev/%s", devs[x]);
		fd = open((char*)line, O_RDONLY|O_NONBLOCK);
		if (fd > 0) {
		    mk = (ioctl(fd, HDIO_DRIVE_CMD, &args) == 0);
		    close(fd);
		} else mk = False;
	    }

	    if (mk) {
		k = 0;
		if (pccmp(HD0dev, devs[x])) k = miAdd_setSel;
		TComboBox_AddItem(pnlHD->cb_dev_HD0, devs[x], k, SetHD_OnClick);
		Ok = True;
	    }
	    xfree(devs[x]);
	    x++;
	}

  const char *timeS[9] = {"off","5 sec","15 sec","30 sec","1 min","2 min","5 min","10 min","20 min"};
  int   timeI[9] = {    0,      1,       3,       6,     12,     24,     60,      60,     120};
	x = 1;
	if (Ok) x = 9;
	for (int i=0;i<x;i++) {
	    k = 0;
	    if (HD0AC == timeI[i]) k = miAdd_setSel;
	    TComboBox_AddItem(pnlHD->cb_AC_HD0, timeS[i], k, SetHD_OnClick)->tag = timeI[i];
	    k = 0;
	    if (HD0Bat == timeI[i]) k = miAdd_setSel;
	    TComboBox_AddItem(pnlHD->cb_Bat_HD0, timeS[i], k, SetHD_OnClick)->tag = timeI[i];
	}
}

bool AHbatmon_Chk_HD_Ok(TBatMonForm *bmf) {
  TAHbatmon_pnlHD *pnlHD = bmf->pnlHD;
	if (!pnlHD) return True;
	return ((pnlHD->cb_dev_HD0->SelItem == pnlHD->mi_dev_HD0) &&
		(pnlHD->cb_AC_HD0->SelItem == pnlHD->mi_AC_HD0) &&
		(pnlHD->cb_Bat_HD0->SelItem == pnlHD->mi_Bat_HD0));
}

void AHbatmonPM_Set_HD(TBatMonForm *bmf) {
  TAHbatmon_pnlHD *pnlHD = bmf->pnlHD;
  int i;

 if (!pnlHD->mi_dev_HD0 || !pnlHD->mi_AC_HD0 || !pnlHD->mi_Bat_HD0) {
//  printf("%s(%d): ERROR ??? %x %x %x\n", FL, (int)p->mi_dev_HD0, (int)p->mi_AC_HD0, (int)p->mi_Bat_HD0);
  return;
 }
	i = pnlHD->mi_AC_HD0->tag;
	if (bmf->mode & 0x01) i = pnlHD->mi_Bat_HD0->tag;
				  
	if (pnlHD->akt_HD0_time != i) {
	    pnlHD->akt_HD0_time = i;
	    SetUP_HDtime(pnlHD, pnlHD->mi_dev_HD0->s, i);
	}
}

void AHbatmon_Setup_HD(TBatMonForm *bmf, bool save) {
  TAHbatmon_pnlHD *pnlHD = bmf->pnlHD;
  TpcIni *pcIni = NULL;
  TMenuItem *mi;
  bool HD_set = False;

	if (!pnlHD->cb_dev_HD0) return;
	if (save) pcIni = appDat.appIni;
	mi = pnlHD->cb_dev_HD0->popup->SelItem;
	if (pnlHD->mi_dev_HD0 != mi) {
	    if (pnlHD->mi_dev_HD0) SetUP_HDtime(pnlHD, pnlHD->mi_dev_HD0->s, 0);
	    pnlHD->mi_dev_HD0 = mi;
	    HD_set = True;
	}
	mi = pnlHD->cb_AC_HD0->popup->SelItem;
	if (pnlHD->mi_AC_HD0 != mi) {
	    pnlHD->mi_AC_HD0 = mi;
	    if ((bmf->mode & 0x01) == 0) HD_set = True;
	}
	mi = pnlHD->cb_Bat_HD0->popup->SelItem;
	if (pnlHD->mi_Bat_HD0 != mi) {
	    pnlHD->mi_Bat_HD0 = mi;
	    if (bmf->mode & 0x01) HD_set = True;
	}
	if (HD_set) AHbatmonPM_Set_HD(bmf);

	if (pcIni) {
	    TpcIni_WR( pcIni, "acpi", "HD0_dev_standby", pnlHD->mi_dev_HD0->s);
	    TpcIni_WR_int(pcIni, "acpi", "HD0_AC_standby",  pnlHD->mi_AC_HD0->tag);
	    TpcIni_WR_int(pcIni, "acpi", "HD0_Bat_standby", pnlHD->mi_Bat_HD0->tag);
	}
}

/*____________________________ Destroy __________________________________*/

/*____________________________ Create ___________________________________*/
static void TAHbatmon_pnlHD_CreateParams(TWins_Create_struct *wcs) {
  TAHbatmon_pnlHD *self = wcs->wins;

	self->subClass_Prev = wcs->Parent->subClass;
	self->subClass = self;
	self->BorderFu = mBorder | mBorderL;
	self->DoPaint = TAHbatmon_pnlHD_DoPaint;
}


TAHbatmon_pnlHD *TAHbatmon_pnlHD_Create(TWins *Parent) {
  TAHbatmon_pnlHD *self;
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TAHbatmon_pnlHD), wtyp_TAHbatmon_pnlHD);
	wcs.Parent = Parent;
	wcs.CreateParams = TAHbatmon_pnlHD_CreateParams;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;
	Set_Parent((TWins*)self, Parent);

  int y = 2;
	self->lbl_HD = TWins_CreateLabel((TWins*)self, 1, y, 200, 16, "Harddisk Power control", 0, 0);
	self->lbl_HD->vds.Font = appXDat.Default_bold_Font;

	y += 20;
	self->cb_dev_HD0 = TComboBox_CreateS((TWins*)self, 148, y, 80, 18, "none");
	self->cb_dev_HD0->ReadOnly = True;
	y++;
	self->lbl_HD0 = TWins_CreateLabel((TWins*)self, 20, y, 120, 16, "Power controlled", mCaptionLeft, 0);
	y += 19;
	self->cb_AC_HD0 = TComboBox_CreateS((TWins*)self, 148, y, 80, 18, "off");
	self->cb_AC_HD0->ReadOnly = True;
	y++;
	self->lbl_AC_HD0 = TWins_CreateLabel((TWins*)self, 20, y, 120, 16, "AC Adapter standby", mCaptionLeft, 0);
	y += 19;
	self->cb_Bat_HD0 = TComboBox_CreateS((TWins*)self, 148, y, 80, 18, "off");
	self->cb_Bat_HD0->ReadOnly = True;
	y++;
	self->lbl_Bat_HD0 = TWins_CreateLabel((TWins*)self, 20, y, 120, 16, "Battery standby", mCaptionLeft, 0);

	self->akt_HD0_time = -1;
	HD_SetMods(self);

	return self;
}
