/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldbatmon.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/hdreg.h>

typedef struct {
  int cpu;
  TCPUPage *CP;
  char *governors;
} TCpuFreqSel;


static const char *CpuFreq_governors_Tab[] = {
  "performance","performance",
  "ondemand","auto",
  "conservative","auto",
  "powersave","powersave",
  NULL, NULL
};

bool AHbatmon_Chk_CpuFreq_Ok(TBatMonForm *bmf) {
  int i;
  TCPUPage *CP;

	for (i=0;i<4;i++) {
	    CP = bmf->CPUPage[i];
	    if (CP) {
		if (CP->mi_AC_freq && (((TComboBox*)CP->cb_AC_freq)->popup->SelItem !=  CP->mi_AC_freq)) return False;
		if (CP->mi_bat_freq && (((TComboBox*)CP->cb_bat_freq)->popup->SelItem !=  CP->mi_bat_freq)) return False;
	    }
	}
	return True;
}

static void AHbatmon_CpuFreqSet_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
	AHbatmon_buApply_ChkEnabled(((TCpuFreqSel*)mi->obj)->CP->bmf);
}

static void AHbatmon_SetUp_CpuFreq(TCPUPage *CP, TMenuItem *mi) {
  int fd;
  TCpuFreqSel *cfs;
  char line[256];
  struct stat64 buf;

	if (!CP || !mi) return;
	cfs = mi->obj;
	if (!cfs || !cfs->governors || pccmp(CP->akt_CPU_governors, cfs->governors)) return;
	snprintf(line, sizeof(line), "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_governor", cfs->cpu);
	if (stat64(line, &buf)) return;
	CP->akt_CPU_governors = cfs->governors;
	fd = open(line, O_WRONLY);
	if (fd > 0) {
	    write(fd, cfs->governors, pclen(cfs->governors));
	    close(fd);
	} else printf("ERROR: %s(%d): %s\n", FL, SysError_pc());
}

static void AHbatmon_CpuFreqSetMods(TCPUPage *CP, int cpu) {
  int lg;
  char line[1024];
  FILE *f;
  char *tline;
  const char **ppc, *indata, *token;
  TMenuItem *mi;
  TCpuFreqSel *cfs;
  TMenu *menu;

	mi = TComboBox_AddItem(CP->cb_AC_freq, "off", 0, AHbatmon_CpuFreqSet_OnClick);
	cfs = xcalloc(sizeof(TCpuFreqSel));
	cfs->CP = CP;
	mi->obj = cfs;
	mi = TComboBox_AddItem(CP->cb_bat_freq, "off", 0, AHbatmon_CpuFreqSet_OnClick);
	cfs = xcalloc(sizeof(TCpuFreqSel));
	cfs->CP = CP;
	mi->obj = cfs;

	snprintf(line, sizeof(line), "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_available_governors", cpu);
	f = fopen(line, "r");
	if (f) {

	    menu = ((TComboBox*)CP->cb_AC_freq)->popup;
	    CP->mi_AC_freq = menu->MenuItem;
	    menu->SelItem = menu->MenuItem;
	    menu = ((TComboBox*)CP->cb_bat_freq)->popup;
	    CP->mi_bat_freq = menu->MenuItem;
	    menu->SelItem = menu->MenuItem;

	    tline = fgets(&line[0], sizeof(line)-1, f);
	    fclose(f);
	    if (tline && pclen(tline)) {
	      ppc = CpuFreq_governors_Tab;
	      while (*ppc) {
		indata = tline;
Loop:
		indata = pcGetToken(indata, &token, &lg);
		if (token) {
		    if (pcncmp(*ppc, token, lg)) {
			mi = TComboBox_GetItemI(CP->cb_AC_freq, 0);
			while (mi) {
			    if (pccmp(ppc[1], mi->s)) goto Next;
			    mi = mi->next;
			}

			mi = TComboBox_AddItem(CP->cb_AC_freq, ppc[1], 0, AHbatmon_CpuFreqSet_OnClick);
			cfs = xcalloc(sizeof(TCpuFreqSel));
			cfs->cpu = cpu;
			cfs->CP = CP;
			cfs->governors = pcndup(token, lg);
			mi->obj = cfs;

			mi = TComboBox_AddItem(CP->cb_bat_freq, ppc[1], 0, AHbatmon_CpuFreqSet_OnClick);
			cfs = xcalloc(sizeof(TCpuFreqSel));
			cfs->cpu = cpu;
			cfs->CP = CP;
			cfs->governors = pcndup(token, lg);
			mi->obj = cfs;
			goto Next;
		    }
		    goto Loop;
		}
Next:
		ppc++; ppc++;
	      }
	    }
	}
}

// ************** set HD standby **************

static void AHbatmon_CpuFreqAz(TCPUPage *CP, char *val) {
  char *pc;

	pc = val;
	while (*pc) {
	    if (((byte)*pc < 0x20) || (*pc == '.')) {
		*pc = ' '; pc++;
		*pc = 'M'; pc++;
		*pc = 'H'; pc++;
		*pc = 'z'; pc++;
		*pc = 0;
		break;
	    }
	    pc++;
	}
	SetCaption(CP->lbl_cpufreq_v, val);
}

void AHbatmon_CpuFreqReScan(TBatMonForm *bmf) {
  int cpu;
  FILE *f;
  TCPUPage *CP;
  char *tline;
  char *val;
  char line[1024];

	f = fopen("/proc/cpuinfo", "r");
	if (!f) return;
	CP = NULL;
loop:
	tline = fgets(&line[0], sizeof(line)-1, f);
	if (tline) {
	  if (StatStrScan("processor	", ':', tline, &val)) {
	    cpu = ppctoint((void*)&val);
	    if ((cpu >= 0) && (cpu < 4)) {
		CP = bmf->CPUPage[cpu];
	    } else CP = NULL;
	  } else if (CP) {
	    if (StatStrScan("cpu MHz		", ':', tline, &val)) {
		AHbatmon_CpuFreqAz(CP, val);
	    }
	  }
	  goto loop;
	}
	fclose(f);


}

void AHbatmon_CpuFreqPageInit(TBatMonForm *bmf) {
  int i, y, cpu;
  FILE *f;
  TCPUPage *CP;
  char *tline;
  char *pc, *pc2, *val;
  char line[1024];
  char cpustr[8] = {"CPU 0"};

	f = fopen("/proc/cpuinfo", "r");
	if (!f) return;

	for (i=0;i<4;i++) bmf->CPUPage[i] = NULL;
	CP = NULL;
loop:
	tline = fgets(&line[0], sizeof(line)-1, f);
	if (tline) {
	  if (StatStrScan("processor	", ':', tline, &val)) {
	    cpu = ppctoint((void*)&val);
	    if ((cpu >= 0) && (cpu < 4)) {
		CP = xcalloc(sizeof(TCPUPage));
		bmf->CPUPage[cpu] = CP;
		CP->bmf = bmf;
		cpustr[4] = cpu + 0x30;
		CP->Page = TPageControl_AddNewPage(bmf->PageControl, cpustr);
		y = 1;
		CP->lbl_modelname = TWins_CreateLabel((TWins*)CP->Page, 1, y, 230, 17, NULL, mCaptionLeft, mBorder|mBorderL);
		y += 19;
		CP->lbl_cpufreq = TWins_CreateLabel((TWins*)CP->Page, 4, y, 130, 16, "CPU Freq:", mCaptionLeft, 0);
		CP->lbl_cpufreq_v = TWins_CreateLabel((TWins*)CP->Page, 148, y, 80, 16, "0 MHz", mCaptionRigth, mBorder | mBorderL);

  TWins *ws = TWins_CreateBA((TWins*)CP->Page, 0, 0, 110, 110, mBorder | mBorderL, alBottom);
		CP->acPageControl = TPageControl_Create(ws);
		CP->acPageControl->height = 106;
		CP->acPageControl->Align = alBottom;
		Set_Parent((TWins*)CP->acPageControl, ws);
		CP->ACadapterPage = TPageControl_AddNewPage(CP->acPageControl, "AC Adapter");
		CP->BatteryPage = TPageControl_AddNewPage(CP->acPageControl, "Battery");

		y = 2;
		CP->lbl_AC_freq = TWins_CreateLabel((TWins*)CP->ACadapterPage, 2, y+1, 130, 16, "CPU speed", mCaptionLeft, 0);
		CP->cb_AC_freq = TComboBox_CreateS((TWins*)CP->ACadapterPage, 138, y, 90, 18, "off");
		CP->cb_AC_freq->ReadOnly = True;
		CP->lbl_bat_freq = TWins_CreateLabel((TWins*)CP->BatteryPage, 2, y+1, 130, 16, "CPU speed", mCaptionLeft, 0);
		CP->cb_bat_freq = TComboBox_CreateS((TWins*)CP->BatteryPage, 138, y, 90, 18, "off");
		CP->cb_bat_freq->ReadOnly = True;
		AHbatmon_CpuFreqSetMods(CP, cpu);

	    } else CP = NULL;
	  } else if (CP) {
	    if (StatStrScan("model name	", ':', tline, &val)) {
		pc = val;
		pc2 = val;
		i = 0;
		while (*pc) {
		    if ((byte)*pc < 0x20) break;
		    if (*pc == ' ') i++; else i = 0;
		    if (i < 3) {
			if (pc != pc2) *pc2 = *pc;
			pc2++;
		    }
		    pc++;
		}
		*pc2 = 0;
		SetCaption(CP->lbl_modelname, val);
	    } else if (StatStrScan("cpu MHz		", ':', tline, &val)) {
		AHbatmon_CpuFreqAz(CP, val);
	    }
	  }
	  goto loop;
	}
	fclose(f);
}

void AHbatmon_IniSetUp(TBatMonForm *bmf) {
  int i;
  TpcIni *pcIni;
  TMenuItem *mi;
  TCPUPage *CP;
  char *pc;
  char line[64];

	pcIni = TpcIni_Create(appDat.ALDini_dir);

	for (i=0;i<4;i++) {
	  CP = bmf->CPUPage[i];
	  if (CP) {
	    snprintf(line, sizeof(line), "CPU%d_speed", i);
	    mi = ((TComboBox*)CP->cb_AC_freq)->popup->MenuItem;
	    pc = TpcIni_RDstrdup(pcIni, "AC_power", line, "off");
	    while (mi) {
		if (pccmp(pc, mi->s)) {
		    TComboBox_SetItem_mi(CP->cb_AC_freq, mi);
		    break;
		}
		mi = mi->next;
	    }
	    xfree(pc);
	    mi = ((TComboBox*)CP->cb_bat_freq)->popup->MenuItem;
	    pc = TpcIni_RDstrdup(pcIni, "BAT_power", line, "off");
	    while (mi) {
		if (pccmp(pc, mi->s)) {
		    TComboBox_SetItem_mi(CP->cb_bat_freq, mi);
		    break;
		}
		mi = mi->next;
	    }
	    xfree(pc);
	  }
	}

	i = TpcIni_RD_int(pcIni, "power", "AC_BAT_Bell", 1);
	TChkBox_SetChecked(bmf->chb_Bell, i);


	Tfree(pcIni);

	AHbatmon_Apply(bmf, False);
}

void AHbatmon_Apply(TBatMonForm *bmf, bool Save) {
  int i;
  TpcIni *pcIni;
  TMenuItem *mi;
  TCPUPage *CP;
  char line[64];

	for (i=0;i<4;i++) {
	  CP = bmf->CPUPage[i];
	  if (CP) {
	    mi = ((TComboBox*)CP->cb_AC_freq)->popup->SelItem;
	    if (CP->mi_AC_freq && (CP->mi_AC_freq != mi)) {
		CP->mi_AC_freq = mi;
		if ((bmf->mode & 0x01) == 0) AHbatmon_SetUp_CpuFreq(CP, mi);
	    }
	    mi = ((TComboBox*)CP->cb_bat_freq)->popup->SelItem;
	    if (CP->mi_bat_freq && (CP->mi_bat_freq != mi) ) {
		CP->mi_bat_freq = mi;
		if (bmf->mode & 0x01) AHbatmon_SetUp_CpuFreq(CP, mi);
	    }
	  }
	}
	bmf->Bell_EN = ((TChkBox*)bmf->chb_Bell)->Checked;


	if (Save) {

	  appDat.appIni = TpcIni_Create(appDat.inifile_app);
	  AHbatmon_Setup_HD(bmf, Save);
	  TpcIni_SaveAndFree(appDat.appIni);
	  appDat.appIni = NULL;

	  pcIni = TpcIni_Create(appDat.ALDini_dir);

	  for (i=0;i<4;i++) {
	    CP = bmf->CPUPage[i];
	    if (CP) {
	      snprintf(line, sizeof(line), "CPU%d_speed", i);
	      if (CP->mi_AC_freq) TpcIni_WR(pcIni, "AC_power", line, CP->mi_AC_freq->s);
	      if (CP->mi_bat_freq) TpcIni_WR(pcIni, "BAT_power", line, CP->mi_bat_freq->s);
	    }
	  }
	  TpcIni_WR_int(pcIni, "power", "AC_BAT_Bell", ((TChkBox*)bmf->chb_Bell)->Checked);
	  TpcIni_SaveAndFree(pcIni);
	}
	AHbatmon_buApply_ChkEnabled(bmf);
	bmf->lastPMmode = bmf->mode;
}

void AHbatmon_PM(TBatMonForm *bmf) {
  int i;
  TCPUPage *CP;

	if ((bmf->lastPMmode & 0x01) == (bmf->mode & 0x01)) return;
	AHbatmonPM_Set_HD(bmf);

	if (bmf->mode & 0x01) {

	    for (i=0;i<4;i++) {
		CP = bmf->CPUPage[i];
		if (CP) AHbatmon_SetUp_CpuFreq(CP, CP->mi_bat_freq);
	    }

	} else {
	    for (i=0;i<4;i++) {
		CP = bmf->CPUPage[i];
		if (CP) AHbatmon_SetUp_CpuFreq(CP, CP->mi_AC_freq);
	    }
	}
	bmf->lastPMmode = bmf->mode;
	if (bmf->Bell_EN) XBell(appXDat.dpy, 25);
	bmf->LowBatMk = 5;
}
