/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#ifdef STANDALONE_BUILDIN
#include "sysbase.h"
#include "sysbaseio.h"
#else
#include "buildin_app.h"
#endif

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/fs.h>

static const char *Help_Msg = {
  "Usage: gethddevtyp [options] dev\n"
  "\n"
  "  -p,                  Get partition table \n"
  "  --noerr,             No error msg \n"
  "  -h, --help           Prints out this screen\n"
#ifdef STANDALONE_BUILDIN
  "  ALDccp: gethddevtyp version "VERSION"  (build %d  %s)\n"
#else
  "  ALDccp: aldesk.gethddevtyp version "VERSION"  (build %d  %s)\n"
#endif
};

#define CP(a,b) (buf[a]==b)

#ifdef STANDALONE_BUILDIN
int main (int argc, char **argv) {
	cmdline_slash_korr(argc, argv);
#else
int bi_gethddevtyp_main (int argc, char **argv) {
#endif
  const char *pc, *devname, *devtyp;
  int fd;
  uint64_t size = 0;
  bool help = False;
  bool err = False;
  bool noerr = False;
  bool ok;
  bool partition = False;
  bool parttyp = False;
  byte buf[1024];

	devname = NULL;
	devtyp = NULL;
	for (int i=1; i<argc; i++) {
	    pc = argv[i];
	    if (*pc == '-') {
		pc++;
		if (*pc == 'h') {
		    help = True;
		} else if (*pc == 'p') {
		    partition = True;
		} else if (*pc == '-') {
		    pc++;
		    if (pcncmp(pc, "noerr", 5)) {
			noerr = True;
		    } else if (pcncmp(pc, "help", 4)) {
			help = True;
		    } else err = True;
		} else err = True;
	    } else {
		devname = pc;
	    }
	}

	if (!help && !devname) err = True;

	if (help || err) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
	    exit(1);
	}

	struct stat st;
	if (stat(devname, &st) || !S_ISBLK(st.st_mode)) {
	    if (!noerr) appError(EMSG_ERR_EXIT, "\"%s\" is no block", devname);
	    return 1;
	}

	fd = open(devname, O_RDONLY);
	if (fd < 0) {
	    if (!noerr) appError(EMSG_ERR_EXIT | EMSG_SYS, "open \"%s\"", devname);
	    return 1;
	}

	if (!partition) goto Get_blk;

	ok = read(fd, buf, 512) == 512;
	if (ok && CP(0x1bc,0x00) && CP(0x1bd,0x00) &&
		  CP(0x1fe,0x55) && CP(0x1ff,0xaa)) {

	    printf("\"%s\" TABLE 0x%02x 0x%02x 0x%02x 0x%02x\n", devname, 
		buf[0x1c2], buf[0x1d2], buf[0x1e2], buf[0x1f2]);

	} else {
	    if (!noerr) appError(EMSG_ERR_EXIT, "\"%s\" no partition table", devname);
	    return 1;
	}
	goto end_fd;


/* ------------------------------------------------------------------ */
Get_blk:
	if (ioctl(fd, BLKGETSIZE64, &size) < 0) {
	    if (ioctl(fd, BLKGETSIZE, &size) < 0) {
		size = 0;
	    } else size >>= 1;
	} else size >>= 10;

	ok = size > 0;
	if (ok) ok = read(fd, buf, 1024) == 1024;

	if (ok) {
	    if (pcncmp((char*)&buf[0x003], "NTFS    ", 8)) devtyp = "ntfs"; else
	    if (pcncmp((char*)&buf[0x052], "MSWIN",    5) ||
		pcncmp((char*)&buf[0x052], "FAT32   ", 8) ||
		pcncmp((char*)&buf[0x036], "MSDOS",    5) ||
		pcncmp((char*)&buf[0x036], "FAT16   ", 8) ||
		pcncmp((char*)&buf[0x036], "FAT12   ", 8)) devtyp = "vfat"; else
	    if (pcncmp((char*)&buf[0x000], "XFSB",     4)) devtyp = "xfs"; else
	    if (pcncmp((char*)&buf[0x000], "-rom1fs-", 8)) devtyp = "romfs"; else
	    if (pcncmp((char*)&buf[0x000], "\xCE\xFA\x7B\x1B", 4)) devtyp = "bfs"; else
	    if (pcncmp((char*)&buf[0x000], "E=\xCD\x28",4)) devtyp = "cramfs"; else
	    if (CP(0x1bc,0x00) && CP(0x1be,0x00) &&
                CP(0x1fe,0x55) && CP(0x1ff,0xaa)) parttyp = True; //devtyp = "part";
	}

	if (ok && !devtyp && (size > 1)) {
	  ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x038], "\x53\xEF", 2)) {
		devtyp = "ext2";
		if (buf[0x05c] & 0x04) devtyp = "ext3";
		if (buf[0x060] & 0x08) devtyp = "jbd";
		if ((buf[0x05c] & 0x04) && (buf[0x060] & 0x40)) devtyp = "ext4";
	    } else
	    if (pcncmp((char*)&buf[0x010], "\x7f\x13", 2)) devtyp = "minix"; else
	    if (pcncmp((char*)&buf[0x010], "\x8f\x13", 2)) devtyp = "minix c30"; else
	    if (pcncmp((char*)&buf[0x010], "\x68\x24", 2)) devtyp = "minix v2"; else
	    if (pcncmp((char*)&buf[0x010], "\x78\x24", 2)) devtyp = "minix v2 c30";
	  }
	}
	if (ok && !devtyp && (size > 3)) {
	  ok = lseek(fd, 0xc00, SEEK_SET) >= 0;
	  if (ok) ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x3f6], "SWAP-SPACE", 10) ||
		pcncmp((char*)&buf[0x3f6], "SWAPSPACE2", 10)) devtyp = "swap";
	  }    
	}
	if (ok && !devtyp && (size > 7)) {
	  ok = lseek(fd, 0x1c00, SEEK_SET) >= 0;
	  if (ok) ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x3f6], "SWAP-SPACE", 10) ||
		pcncmp((char*)&buf[0x3f6], "SWAPSPACE2", 10)) devtyp = "swap";
	  }    
	}
	if (ok && !devtyp && (size > 8)) {
	  ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x014], "ReIsErFs", 8)) devtyp = "reiserfs"; else
	    if (pcncmp((char*)&buf[0x034], "ReIsErFs", 8)) devtyp = "reiserfs";
	  }    
	}
	if (ok && !devtyp && (size > 15)) {
	  ok = lseek(fd, 0x3c00, SEEK_SET) >= 0;
	  if (ok) ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x3f6], "SWAP-SPACE", 10) ||
		pcncmp((char*)&buf[0x3f6], "SWAPSPACE2", 10)) devtyp = "swap";
	  }    
	}
	if (ok && !devtyp && (size > 31)) {
	  ok = lseek(fd, 0x7c00, SEEK_SET) >= 0;
	  if (ok) ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x3f6], "SWAP-SPACE", 10) ||
		pcncmp((char*)&buf[0x3f6], "SWAPSPACE2", 10)) devtyp = "swap";
	  }    
	}
	if (ok && !devtyp && (size > 32)) {
	  ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x001], "CD001", 5) ||
		pcncmp((char*)&buf[0x009], "CDROM", 5)) devtyp = "iso9660"; else
	    if (pcncmp((char*)&buf[0x001], "BEA01", 5) ||
		pcncmp((char*)&buf[0x001], "BOOT2", 5) ||
		pcncmp((char*)&buf[0x001], "CD001", 5) ||
		pcncmp((char*)&buf[0x001], "CDW02", 5) ||
		pcncmp((char*)&buf[0x001], "NSR02", 5) ||
		pcncmp((char*)&buf[0x001], "NSR03", 5) ||
		pcncmp((char*)&buf[0x001], "TEA01", 5)) devtyp = "udf";
	  }    
	}


	if (ok && !devtyp && (size > 63)) {
	  ok = lseek(fd, 0xfc00, SEEK_SET) >= 0;
	  if (ok) ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x3f6], "SWAP-SPACE", 10) ||
		pcncmp((char*)&buf[0x3f6], "SWAPSPACE2", 10)) devtyp = "swap";
	  }    
	}
	if (ok && !devtyp && (size > 64)) {
	  ok = read(fd, buf, 1024) == 1024;
	  if (ok) {
	    if (pcncmp((char*)&buf[0x034], "ReIsErFs",  8)) devtyp = "reiserfs"; else
	    if (pcncmp((char*)&buf[0x034], "ReIsEr2Fs", 9)) devtyp = "reiserfs"; else
	    if (pcncmp((char*)&buf[0x034], "ReIsEr3Fs", 9)) devtyp = "reiserfs";
	  }    
	}

	if (!devtyp) {
	    devtyp = "unknown";
	    if (parttyp) devtyp = "part";
	}

	pc = "kB";
	if (size >= 0x10000) {
	    size >>= 10;
	    pc = "MB";
	    if (size >= 0x10000) {
		size >>= 10;
		pc = "GB";
	    }
	}
#if __WORDSIZE == 64
	printf("\"%s\" SIZE %lu%s TYP %s\n", devname, size, pc, devtyp);
#else
	printf("\"%s\" SIZE %llu%s TYP %s\n", devname, size, pc, devtyp);
#endif
end_fd:
	close(fd);

	return 0;
}
