/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

/* _____________________________ Mount _______________________________ */
typedef struct {
  char *fn, *cmd;
  TDesktopIcon *di;
} TMsgBox_mount_path;

static void ExecEv_mount_umount(Texec_pipe* expipe, int evtyp) {
  int fu;

	TwmDeskIcon_App_ExecEv(expipe, evtyp);
	if (evtyp == evtyp_ProcClose) {
	    TDesktopIcon *di = expipe->owner;
	    Set_Cursor((TWins*)di, XC_top_left_arrow);
	    Set_Cursor(di->WinCaption, XC_top_left_arrow);
	    if (!(expipe->CloseFlag & epCF_Error)) {
		if (di->cmd2) {
		    fu = errfd_Open;
		    if (di->difs & DIFS_CIFS) fu = infd_Open|errfd_Open;
		    ExecuteProcAppTag(di->cmd2, fu, TwmDeskIcon_App_ExecEv, NULL, 
			  NULL, exShowErrMsgBox|exNoStartPid|exGrabPointer);
		}
		TNotifyMount_ReScan();
	    }
	    ppcfree(&di->cmd2);
	}
}

static void meDeskIcon_mount_umount(TDesktopIcon *di, char *cmd) {
  int fu;

	Set_Cursor((TWins*)di, XC_watch);
	Set_Cursor(di->WinCaption, XC_watch);
	fu = errfd_Open;
	if (di->difs & DIFS_CIFS) fu = infd_Open|errfd_Open;
	ExecuteProcAppTag(cmd, fu, ExecEv_mount_umount, di, 
			  NULL, exShowErrMsgBox|exNoStartPid|exAddTimer);
}

static void MsgBox_mount_path_ev(Texec_pipe *expip, int evtyp) {
  TMsgBox_mount_path *mp;

	if (evtyp == evtyp_ProcClose) {
	    mp = expip->owner;
// MsgBox return: test create mount path (zB.: /media/sda1)
	    if (pcncmp(expip->pp[evtyp_infd].lb->pc, "YES", 3)) {
		if (file_mkdir_parents_ae(mp->fn, 0, EMSG_ERR1)) {
		    meDeskIcon_mount_umount(mp->di, mp->cmd);
		}
	    }
	    xfree(mp->fn);
	    xfree(mp->cmd);
	    xfree(mp);
	}
}

void meDeskIcon_mount_Click(TDesktopIcon *di, TpcIni *ini) {
  char *pc;
  const char *dev, *fn, *cpc, *exec;
  char buf[256];

/* ChkFSDeviceOk() TDesktopIcon.c chk NULL dev,fn */
	dev = TpcIni_RD(ini, Desktop_Entry, "Dev");
//	if (!dev) return;
	fn = TpcIni_RD(ini, Desktop_Entry, "MountPoint");
//	if (!fn) return;
	if (!TNotifyMount_ChkMountPointFree(fn)) {
	    ShowExtMsgBox("MountPoint=\"%s\"\nis mount !!!", fn);
	    return;
	}

	pc = buf;
	exec = TpcIni_RD(ini, Desktop_Entry, "Exec");
	if (!exec) exec = "mount";
	ppcmove_k(&pc, exec, kSPACE_ADD);
	cpc = TpcIni_RD(ini, Desktop_Entry, "FSType");
	if (cpc) {
	    ppcmove(&pc, "-t ");
	    ppcmove_k(&pc, cpc, kSPACE_ADD);
	}
	ppcmove_k(&pc, dev, kSPACE_ADD);
	ppcmove(&pc, fn);
	SetFSDeviceOpt(di, ini, &pc);
// printf("%s(%d): %s\n", FL, buf);
// return;

	if (!file_chk_exist(fn, findDIR)) {
	    TMsgBox_mount_path *mp = xmalloc(sizeof(TMsgBox_mount_path));
	    mp->fn = pcdup(fn);
	    mp->cmd = pcdup(buf);
	    mp->di = di;

	    pc = buf;
	    wmExcCMD_headX(&pc, "aldmessagebox");
	    ppcmove(&pc, " -fYES,NO -dYES -w0 \"mount path ");
	    ppcmove(&pc, fn);
	    ppcmove(&pc, " not exist\n do create ???\"");

	    ExecuteProcApp(buf, infd_Open, MsgBox_mount_path_ev, mp, NULL);

	} else {
	    meDeskIcon_mount_umount(di, buf);
	}
}

void meDeskIcon_umount_Click(TDesktopIcon *di, TpcIni *ini) {
  char *pc, *mp;
  const char *cpc;
//  TFSDevice_info *fs_info = NULL;
  char buf[256];

/* ChkFSDeviceOk() TDesktopIcon.c chk NULL dev,fn */
//	fn = TpcIni_RD(ini, Desktop_Entry, "Dev");
//	if (!fn) return;
/*	fs_info = get_fs_info(fn);
	if (fs_info && (fs_info->stat & FSINFO_NFS)) {
	    fn = ((TNFSDevice_info*)fs_info)->dev;
	} */

	pc = buf;
	*pc = 'u'; pc++;
	cpc = TpcIni_RD(ini, Desktop_Entry, "Exec");
	if (!cpc) cpc = "mount";
	ppcmove_k(&pc, cpc, kSPACE_ADD);
	mp = TNotifyMount_GetMountPoint(TpcIni_RD(ini, Desktop_Entry, "Dev"));
	if (!mp) return;
	ppcmove(&pc, mp);
	xfree(mp);
// printf("%s(%d): %s\n", FL, buf);
	meDeskIcon_mount_umount(di, buf);
}

/* _____________________________ Ev __________________________________ */

static void miMount_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TDesktopIcon *di;
  TpcIni *ini;
	di = get_owner_DesktopIcon((TWins*)self);
	if (!di || !di->fname) return;
	ini = TpcIni_Create(di->fname);
	meDeskIcon_mount_Click(di, ini);
	Tfree(ini);
}

static void miUmount_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TDesktopIcon *di;
  TpcIni *ini;
	di = get_owner_DesktopIcon((TWins*)self);
	if (!di || !di->fname) return;
	ini = TpcIni_Create(di->fname);
	meDeskIcon_umount_Click(di, ini);
	Tfree(ini);
}

static void miEject_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TDesktopIcon *di;

	di = get_owner_DesktopIcon((TWins*)self);
	if (di && (di->pic == di->pic2)) {
	    miUmount_OnClick(self, KeyState, mi);
	    di->cmd2 = pcdup((char*)mi->obj);
	    return;
	}
	ExecuteProcAppTag((char*)mi->obj, errfd_Open, TwmDeskIcon_App_ExecEv, NULL, 
			  NULL, exShowErrMsgBox|exNoStartPid|exGrabPointer);
}

static void miDeleteDeskIcon_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TDesktopIcon *di;

	di = get_owner_DesktopIcon((TWins*)self);
	if (!di) return;
	remove(di->fname);
}

static void menu_Default_DeskIcon_OnMenuOpen(TMenu *self) {
  TDesktopIcon *di;
  TMenuItem *mi, *mitop;
  TpcIni *ini = NULL;
  char *pc, *sdev;
  const char *cpc;
  int gdifu;
  TDesktopHead ddkh;
  char buf[200];

	di = get_owner_DesktopIcon((TWins*)self);
	if (!di) return;

	mitop = TMenu_AddItem(self, GetCaption_pc(di->WinCaption), miCanNotSelect, NULL);
	ini = TpcIni_Create(di->fname);

	if (dotdesktop_get_head(ini, &ddkh)) {
	    gdifu = gdifu_getico0;
	    if (di->ddkt == ddkt_FSDevice) {

		mi = NULL;
		TMenu_AddItem(self, "-", 0, NULL);
		if (di->pic == di->pic2) {
		    gdifu = gdifu_getico1;
		    mi = TMenu_AddItem(self, "umount", 0, miUmount_OnClick);
		} else if (di->pic == di->pic3) {
		    mi = TMenu_AddItem(self, "mount", 0, miMount_OnClick);
		} else if (di->pic == di->pic4) {
		    gdifu = gdifu_getico2;
		}

		if (mi) {
		    TFSDevice_info *fs_info = get_fs_info(di->s2);
		    if (fs_info) {
			if (fs_info->stat & FSINFO_OK) {
			    pc = buf;
			    ppcmove_k(&pc, fs_info->type, kSPACE_ADD);
			    ppcmove(&pc, fs_info->size);
			    mi = TMenu_InsItem(self, buf, 1);
			    mi->fu = miCanNotSelect;
			}
			sdev = pcadd("/dev/",  fs_info->dev0);
			if (chk_dev_removable(sdev)) {
			  cpc = "eject ";
			  if (pcncmp(fs_info->dev0, "sd", 2)) cpc = "eject -s ";
			  mi = TMenu_AddItem(self, "Eject", 0, miEject_OnClick);
			  mi->obj = pcadd(cpc, sdev);
			}
			xfree(sdev);
		    }
		}
	    }
	    gdifu |= gdifu_noUnknown;
	    mitop->pic = GetDeskIcon(ddkh.ddkt, 16, ddkh.icon, gdifu, NULL);
	}
	if (!mitop->pic) {
	    mitop->pic = GetDeskIcon(ddkt_Actions, 16, "desktop", 0, NULL);
	}

	Tfree(ini);

	TMenu_AddItem(self, "-", 0, NULL);
	TMenu_AddItem(self, "Delete", 0, miDeleteDeskIcon_OnClick);
	Menu_OnMenuOpen(self);
}

static void menu_Default_DeskIcon_OnMenuClose(TMenu *self) {
  TMenuItem *mi;

	Menu_OnMenuClose(self);
	mi = self->MenuItem;
	while (mi) {
	    ppfree(&mi->obj);
	    mi = mi->next;
	}
	TMenu_Cls(self);
}

/*_____________________________ Create __________________________________*/
void menu_DeskIcon(TwmRoot *wmr) {
  TMenu *me;
/* ****** Menu Default Desktop Icon ****** */
	me = TMenu_Create((TWins*)wmr);
	me->OnMenuOpen = menu_Default_DeskIcon_OnMenuOpen;
	me->OnMenuClose = menu_Default_DeskIcon_OnMenuClose;
	wmr->menu_Default_DeskIcon = me;
}
