/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"
#include <unistd.h>

void TwmRoot_SetUpFont(TwmRoot *self, TpcIni *Ini) {
  char *pc;
  const char *cpc;
  bool ref;

	cpc = TpcIni_RD_pc(Ini, "Style", "wmTitleFont", 
				appXDat.Default_bold_Font->name);
	
	if (!pccmp(self->TitleFont, cpc)) {
	    TFont_Destroy(self->wmTitleFontUTF8);
	    self->wmTitleFontUTF8 = NULL;
	    ref = self->TitleFont != NULL;

	    ppcchange(&self->TitleFont, cpc);
	    self->wmTitleFont = MakeFont(cpc, appXDat.Default_Font->name, self->wmTitleFont);

	    if ((self->wmTitleFont->xfont->min_byte1 | self->wmTitleFont->xfont->max_byte1) == 0) {
		bool ok = False;
		pc = pcdup(self->TitleFont);
		if (SetFontSXVTypeIdx(12, &pc, "iso10646")) {
		    ok = SetFontSXVTypeIdx(13, &pc, "1");
		}
		if (ok) {
		    self->wmTitleFontUTF8 = MakeFont(pc, NULL, NULL);
		    if ((self->wmTitleFontUTF8->xfont->min_byte1 | self->wmTitleFontUTF8->xfont->max_byte1) == 0) {
			TFont_Destroy(self->wmTitleFontUTF8);
			self->wmTitleFontUTF8 = NULL;
		    }
		}
		xfree(pc);
	    }
	    if (ref) {
	      TWins *wins = appXDat.Wins;
	      while (wins) {
		if (wins->wtyp == wtyp_TwmFrame) {
		  TWins *winsn = wins->next_child;
		  while (winsn) {
		    if (winsn->wtyp == wtyp_TwmTitle) {
			RePaint(winsn);
		    }
		    winsn = winsn->next;
		  }
		}
		wins = wins->next;
	      }
	    }
	}
}

void TwmRoot_SetBackground(TwmRoot *wmr, TpcIni *Ini) {
  int desk_Col, ico_FontCol;
  const char *desk_Image, *ico_mod, *pc;
  int mk = 0;
  Display *dpy = appXDat.dpy;
  TwmDeskIcon_Global *di_glob = &wmr->di_glob;
  TPixMap *pxm;
  TFont *font;
  bool desk_Image_Ok;

	desk_Image = TpcIni_RD(Ini, "Style", "DesktopImage");
	desk_Image_Ok = pccmp(wmr->DesktopImage, desk_Image);

	desk_Col = TpcIni_RD_int(Ini, "Style", "DesktopColor", 0x505050);
	if (desk_Image) {
	  if (!desk_Image_Ok) {
	    wmr->stat &= ~wmRst_ImgIsOk;
	    pxm = TPixMap_Loader(desk_Image, NULL, pxmLD_NoMask);
	    if (pxm) {
	      if ((pxm->width > appXDat.Root_width) &&
		  (pxm->height > appXDat.Root_height)) {
 float kw = (float)appXDat.Root_width / (float)pxm->width;
 float kh = (float)appXDat.Root_height / (float)pxm->height;
 TPixMap *npxm;
 Timg_rsz rsz;
 int tmpi;

		npxm = TPixMap_Create_whLD(appXDat.Root_width, appXDat.Root_height, pxm->pxm_LD & pxmLD_bppMask);
		if (npxm) {
		  TPixMap_CreateXimg(npxm);
		  TImage_SetBase_rsz((TImage*)pxm, &rsz);
		  TImage_SetDestBase_rsz((TImage*)npxm, &rsz);
		  rsz.to_w = appXDat.Root_width;
		  rsz.to_h = appXDat.Root_height;

		  if (kw > kh) {
    tmpi = ((float)pxm->height * kw - (float)appXDat.Root_height) / kw;
		    if (tmpi > 2) {
		      rsz.src_off_y = tmpi / 2;
		      rsz.src_lg_y = pxm->height - tmpi;
		    }
		  } else {

    tmpi = ((float)pxm->width * kh - (float)appXDat.Root_width) / kh;
		    if (tmpi > 2) {
		      rsz.src_off_x = tmpi / 2;
		      rsz.src_lg_x = pxm->width - tmpi;
		    }
		  }
		  ImgDat_Resize_XX_11(&rsz);
		  Tfree(pxm);
		  pxm = npxm;
		}
	      }
	      TPixMap_ToWinBackground(pxm, wmr->WinID);
	      XClearWindow(dpy, wmr->WinID);
	      Tfree(pxm);
	      wmr->stat |= wmRst_ImgIsOk;
	      mk |= 0x0d;
	    }
	  } else {
	      if (wmr->stat & wmRst_ImgIsOk) mk |= 0x01;
	  }
	} else {
	    if (wmr->stat & wmRst_ImgIsOk) mk |= 0x0c;
	    wmr->DesktopColor = -1;
	    wmr->stat &= ~wmRst_ImgIsOk;
	}

	if (!(mk & 0x01) && (wmr->DesktopColor != desk_Col)) {
	    wmr->DesktopColor = desk_Col;
	    XSetWindowBackground(dpy, wmr->WinID, GetColor_RGB(desk_Col));
	    XClearWindow(dpy, wmr->WinID);
	    mk |= 0x01;
	}

	ico_mod = TpcIni_RD(Ini, "Style", "DesktopIcoColorMod");
	if (pccasecmpASCII(wmr->DesktopIcoColorMod, ico_mod)) {

	    if (!ico_mod || !*ico_mod) {
		xfree(wmr->DesktopIcoColorMod);
		wmr->DesktopIcoColorMod = NULL;
	    } else ppcchange(&wmr->DesktopIcoColorMod, ico_mod);
	    mk |= 0x02;

	    wmr->di_glob.mod &= ~(wmDImod_HighLight | wmDImod_Transparent);
	    wmr->di_glob.TwColor = desk_Col;
	    if (!wmr->DesktopIcoColorMod) {
	    } else if (!pccasecmpASCII(ico_mod, "default")) {
		wmr->di_glob.mod |= wmDImod_HighLight;
	    } else if (!pccasecmpASCII(ico_mod, "auto")) {
		wmr->di_glob.mod |= (wmDImod_HighLight | wmDImod_Transparent);
	    } else {
		wmr->di_glob.TwColor = pctoint(ico_mod);
	    }
	    wmr->di_glob.TwHlColor = AddColor(wmr->di_glob.TwColor, 0x101010);
	}

	ico_FontCol = TpcIni_RD_int(Ini, "Style", "DesktopIcoFontColor", 0);
	if (di_glob->TwFontColor != ico_FontCol) {
	    di_glob->TwFontColor = ico_FontCol;
	    mk |= 0x04;
	}
	pc = TpcIni_RD_pc(Ini, "Style", "DesktopIcoFont", appXDat.Default_Font->name);
	if (pccasecmpASCII(di_glob->DesktopIcoFontName, pc)) {
	    ppcchange(&di_glob->DesktopIcoFontName, pc);
	    font = MakeFont(pc, NULL, di_glob->DesktopIcoFont);
	    SetFont_Default(font, True);
	    xfree(di_glob->DesktopIcoFont);
	    di_glob->DesktopIcoFont = font;
	    mk |= 0x08;
	}

	if (mk & 0x0e) {
	    TDesktopIcon *di = (TDesktopIcon*)wmr->next_child;
	    while (di) {
		if (di->wtyp == wtyp_TwmDeskIcon) {
		    if (mk & 0x08) {
			di->WinCaption->vds.Font = di->glob->DesktopIcoFont;
			TDesktopIcon_SetSize(di);
		    }
		    TDesktopIcon_SetFunkt(di, DIFu_Default | DIFu_Force);
		    Refresh(di->WinCaption);
		}
		di = (TDesktopIcon*)di->next;
	    }
	}
}

void TwmRoot_SetBuildinStr(TwmRoot *self, TpcIni *Ini) {
  TBuildinKeys *BKy;
  TGruppeDat *Gpd;
  TGruppe *Gp;
  int mod, lmod, min, max;

	if (self->BuildinKeys) {
	    XUngrabKey(appXDat.dpy, AnyKey, AnyModifier, self->WinID);
	}

	while (self->BuildinKeys) {
	    BKy = self->BuildinKeys;
	    self->BuildinKeys = BKy->next;
	    xfree(BKy->cmd);
	    xfree(BKy);
	}

	Gp = TpcIni_FindGruppe(Ini, "KeyBinding");
	if (Gp) {
	    Gpd = Gp->Dat;
	    while (Gpd) {
		BuildinStrScane(self, Gpd->name);
		Gpd = Gpd->next;
	    }
	}

	BKy = self->BuildinKeys;
	while (BKy) {
	    lmod = -1;
	    min = BKy->mod;
	    max = (BKy->mod | ~BKy->modmask) & 0xff;
	    while (min <= max) {
		mod = (min | BKy->mod) & max;
		if (lmod < mod) {
		    XGrabKey(appXDat.dpy, BKy->code, mod, self->WinID, 
				    1, GrabModeAsync, GrabModeAsync);
		    lmod = mod;
		}
		min++;
	    }
	    BKy = BKy->next;
	}
}

/* _______________________ Default ______________________________________ */
static const char *DefaultIniDat = {
  "[Style]\n"
  "DesktopColor=0x505050\n"
  "DesktopImage=\n"
  ";DesktopIcoColorMod default,auto,val\n"
  "DesktopIcoColorMod=auto\n"
  "DesktopIcoFontColor=0xd0d0d0\n"
  "DesktopIcoFont=-adobe-helvetica-medium-r-normal-*-12-*-75-75-*-*-*-1\n"
  "FormStyle=3\n"
  "BorderSnap=5\n"
  "wmTitleFont=-adobe-helvetica-bold-r-normal-*-12-*-75-75-*-*-*-1\n"
  "[AutoFsDev]\n"
  "ScanTime=1000\n"
  "[Key]\n"
  "Numlock=\n"
  "[KeyBinding]\n"
  ";Shift Lock Control(Strg) Mod1 Mod2 Mod3 Mod4 Mod5\n"
  "Key add Lock+Mod2+Super_L	menu_start\n"
  "Key add Lock+Mod2+Mod1-F3	menu_form_opt\n"
  "Key add Lock+Mod2+Mod1-F4	close_win\n"
  "Key add Lock+Mod2+Print		snapshot\n"
  "Key add Lock+Mod2+Control-Print	snapshot_form\n"
  "Key add Lock+Mod2+Mod4-e	exec aldfile\n"
};

void ChkSetDefault() {
  TpcIni *Ini = TpcIni_Create(appDat.inifile_global);


	InitIconPaths(Ini);

	if (!TpcIni_FindGruppeDat(Ini, "Global", "DateFormat")) {
	    TpcIni_WR(Ini, "Global", NULL, "; DateFormat 0=d-m-y 1=m-d-y 2=y-m-d");
	    TpcIni_WR_int(Ini, "Global", "DateFormat", 0);
	}
	if (!TpcIni_FindGruppeDat(Ini, "Global", "UTF8")) {
	    TpcIni_WR_int(Ini, "Global", "UTF8", 0);
	}

	if (!TpcIni_FindGruppeDat(Ini, "AutoFsDev", "MountPath")) {
	    TpcIni_WR(Ini, "AutoFsDev", "MountPath", "/media");
	}
	if (!TpcIni_FindGruppeDat(Ini, "AutoFsDev", "no_AutoFsDev")) {
	    TpcIni_WR(Ini, "AutoFsDev", "no_AutoFsDev", NULL);
	}
	if (!TpcIni_FindGruppeDat(Ini, "AutoFsDev", "no_AutoFsTyp")) {
	    TpcIni_WR(Ini, "AutoFsDev", "no_AutoFsTyp", "swap unknown");
	}


	TpcIni_SaveAndFree(Ini);

	if (access(appDat.inifile_app, F_OK)) {
	    FILE *fd = fopen(appDat.inifile_app, "w");
	    if (fd) {
		fputs(DefaultIniDat, fd);
		fclose(fd);
	    }
	}

}
