/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"
#include "dlg.h"

#define cbSelCount 5

static int LastPosX, LastPosY;

#define TCSelf TFormCopy
typedef struct TFormCopy {
  AlgTForm

  TWins *pnlTop;
  TWins *pnlMiddle;
  TWins *pnlBottom;

  TWins *pnlMiddleLeft;
  TLineEdit *le;

  TEdit *edFrom;
  TEdit *edTo;

  TChkBox *cbSel[cbSelCount];
  TChkBox *cbFollowLinks;
  TChkBox *cbPreservDate;
  TChkBox *cbPreservUID;

  TButton *buStart;
  TButton *buCancle;
  
  TglobPath *gp_src;
  TglobPath *gp_dest;
} TFormCopy;
#undef TCSelf

typedef struct {
  const char *cap;
  dword skip;
} TcbSelBIT;

static const TcbSelBIT cbSelBIT[cbSelCount] = {
  {"Overwrite ask",      0},
  {"Overwrite all",      xcyForce_ovwDir | xcyForce_ovwFile},
  {"Overwrite skip",     xcySkip_ovwDir | xcySkip_ovwFile},
  {"Overwrite old File", xcySkip_OldFile | xcyForce_ovwDir | xcyForce_ovwFile},
  {"Update Dest File", xcySkip_OldFile | xcyForce_ovwDir | xcyForce_ovwFile | xcySkip_DestNotExist}
};

/* _____________________________ Ev __________________________________ */
static void cbSel_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFormCopy *mf = (TFormCopy*)GetParentForm((TWins*)self);

	for (int i=0;i<cbSelCount;i++) {
	    TChkBox_SetChecked(mf->cbSel[i], mf->cbSel[i] == self);
	}
}

static void buStart_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormCopy *cf = GetParentForm((TWins*)self);
  int skip = 0;

	for (int i=0;i<cbSelCount;i++) {
	    if (cf->cbSel[i]->Checked) skip = cbSelBIT[i].skip;
	}

	if (cf->cbFollowLinks->Checked) skip |= xcyForce_followlinks;
	if (!cf->cbPreservDate->Checked) skip |= xcySkip_Utime;
	if (!cf->cbPreservUID->Checked) skip |= (xcySkip_Chmod | xcySkip_Chown);

  TWins *Parent = cf->Parent;
  TLines *tls = cf->le->tls;
  TglobPath *gp_src = cf->gp_src;
  TglobPath *gp_dest = cf->gp_dest;

	cf->le->tls = NULL;
	cf->gp_src = NULL;
	cf->gp_dest = NULL;

	TForm_Close((TForm*)cf);
	TFormCopyRun_Show(Parent, tls, gp_src, gp_dest, skip);
}

static void buCancle_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormCopy *cf = GetParentForm((TWins*)self);
	TForm_Close((TForm*)cf);
}

/* ____________________________ Resize _______________________________ */
static void PanelTop_OnResize(TWins *self) {
  TFormCopy *cf = GetParentForm(self);
  int w;
	w = self->width - 43;
	if (w < 40) w = 40;
	SetWidth((TWins*)cf->edFrom, w);
	SetWidth((TWins*)cf->edTo, w);
}


static void PanelBottom_OnResize(TWins *self) {
  TFormCopy *cf = (TFormCopy*)GetParentForm(self);
  int w;

	w = self->width - 2 - cf->buCancle->width;
	SetLeft((TWins*)cf->buCancle, w);
}

/* ____________________________ Destroy ______________________________ */
static bool TFormCopy__Destroy(TFormCopy *self) {
	LastPosX = self->left;
	LastPosY = self->top;
	gpFree(self->gp_src);
	gpFree(self->gp_dest);
	return TWins__Destroy((TWins*)self);
}

static void mf_OnCanClose(TFormCopy *self, int *CanClose) {
	*CanClose = FormCloseAndFree;
}

/* ____________________________ Create _______________________________ */
void TFormCopy_Show(TWins *Parent, char *pcsel, TglobPath *gp_dest) {
  int y;
  TFormCopy *self;
  TglobPath *gp_src, *gpdup_dest;
  char *pc;
  TLines *tlsdup;

	if (!pcsel || !gp_dest) return;
	gpdup_dest = gp_dup(gp_dest);
	ppcfree(&gpdup_dest->name);

	tlsdup = TLines_Create();
	gp_src = pcsel_to_tls(pcsel, tlsdup);

	if (pccmp(gp_src->path, gpdup_dest->path) && 
	    ((gp_test_local_drive(gp_src) && gp_test_local_drive(gpdup_dest)) ||
	     pccmp(gp_src->RootArt, gpdup_dest->RootArt))) {

	    if (tlsdup->Count > 1) {
		MsgBox_Show(Parent, "ERROR", "Can only one file rename", msgOK|msgIcoERROR, msgOK);
		Tfree(tlsdup);
		gpFree(gp_src);
		gpFree(gpdup_dest);
		return;
	    }
	    TFormCopyRun_Show(Parent, tlsdup, gp_src, gpdup_dest, 0);
	    return;
	}

	self = TForm_CreatParamAlg(Parent, sizeof(TFormCopy));
	self->gp_src = gp_src;
	self->destroy = (TDESTROY)TFormCopy__Destroy;
	self->FormZStyle = fszTransient;
	if (LastPosX || LastPosY) {
	    self->left = LastPosX;
	    self->top = LastPosY;
	    self->FormPosition = poDesigned;
	} else {
	    self->FormPosition = poCenter;
	}
	self->width = 480;
	self->height = 231;
	self->min_width = 360;
	self->min_height = 231;
	SetCaption((TWins*)self, "Copy");
	self->OnCanClose = mf_OnCanClose;
	self->gp_dest = gpdup_dest;

	self->pnlTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 41, mBorder, alTop);
	self->pnlTop->OnResize = PanelTop_OnResize;
	self->pnlMiddle = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder, alClient);
	self->pnlBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 28, mBorder, alBottom);
	self->pnlBottom->OnResize = PanelBottom_OnResize;

	self->pnlMiddleLeft = TWins_CreateBA(self->pnlMiddle, 0, 0, 140, 21, 0, alLeft);
	self->le = TLineEdit_CreateA(self->pnlMiddle, 0, 0, 100, 100, alClient);
	self->le->ReadOnly = True;
	Tfree(self->le->tls);
	self->le->tls = tlsdup;
	TWins_SetUTF8((TWins*)self->le, appXDat.FileSystemUTF8on);
	SetVisible((TWins*)self->le->HSB, False);
	self->le->KeyFu |= mNoKeyFocus;


	TWins_CreateLabel(self->pnlTop, 2,  3, 28, 16, "from:", mCaptionLeft, 0);
	TWins_CreateLabel(self->pnlTop, 2, 22, 28, 16, "to :", mCaptionLeft, 0);

	pc = get_DisplayPathStr(NULL, self->gp_src);
	self->edFrom = TEdit_CreateS(self->pnlTop, 40, 2, 45, 18, pc);
	xfree(pc);
	self->edFrom->KeyFu |= mNoKeyFocus;
	self->edFrom->ReadOnly = True;
	SetColor((TWins*)self->edFrom, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edFrom, appXDat.FileSystemUTF8on);

	pc = get_DisplayPathStr(NULL, self->gp_dest);
	self->edTo = TEdit_CreateS(self->pnlTop, 40, 21, 45, 18, pc);
	xfree(pc);
	self->edTo->KeyFu |= mNoKeyFocus;
	self->edTo->ReadOnly = True;
	SetColor((TWins*)self->edTo, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edTo, appXDat.FileSystemUTF8on);

	y = 2;
	for (int i=0;i<cbSelCount;i++) {
	    self->cbSel[i] = TChkBox_CreateS(self->pnlMiddleLeft, 2, y, 136, 17,
						 cbSelBIT[i].cap);
	    self->cbSel[i]->AutoToggle = False;
	    self->cbSel[i]->OnClick = cbSel_OnClick;
	    y += 19;
	}
	TChkBox_SetChecked(self->cbSel[0], True);

	y += 6;
	self->cbFollowLinks = TChkBox_CreateS(self->pnlMiddleLeft, 2, y, 136, 17,
						 "follow symbolic links");

	y += 19;
	self->cbPreservDate = TChkBox_CreateS(self->pnlMiddleLeft, 2, y, 136, 17,
						 "Preserv Date");
	y += 19;
	self->cbPreservUID = TChkBox_CreateS(self->pnlMiddleLeft, 2, y, 136, 17,
						 "Preserv UID/GID");
	TChkBox_SetChecked(self->cbPreservDate, True);
	TChkBox_SetChecked(self->cbPreservUID, True);

	self->buStart = TButton_CreateS(self->pnlBottom, 2, 2, 64, 24, "START");
	self->buStart->OnClick = buStart_OnClick;
	self->buCancle = TButton_CreateS(self->pnlBottom, 2, 2, 64, 24, "CANCEL");
	self->buCancle->OnClick = buCancle_OnClick;

	Set_Focus((TWins*)self->cbSel[0]);

	TForm_Show((TForm*)self);
	return;
}
