/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"
#include "dlg.h"

static int LastPosX, LastPosY;

#define TCSelf TFormCopyRun
typedef struct TFormCopyRun {
  AlgTForm

  TWins *pnlTop;
  TWins *pnlMiddle;
  TWins *pnlBottom;

  TEdit *edFrom;
  TEdit *edTo;

  TEdit *edFile;
  TWins *lblStk;
  TWins *lblMB;
  TWins *lblMBrun;
  TWins *lblMBpsec;

  TProgressBar *pbProz;
  TWins *lblMBpsecN;

  TButton *buCancle;

  TLines *tls;
  TglobPath *gp_src;
  TglobPath *gp_dest;
  Txpy_wd xpy_wd;
  TgTimer *gTimer;
  char *timer_fn;
  char *timer_path_src;
  char *timer_path_dest;

  int FileCount, mps, ges_mps;
  uint64_t DataSize;

//  dword us;
  bool stop;
} TFormCopyRun;
#undef TCSelf

static int msg_CopyEv(Txpy_wd wd, int xcyMsg) {
  int bef = xcyMsg & xpyMsg_NrMask;
  int res = 0;
  TFormCopyRun *mf;

	mf = XCopy_GetOwner(wd);
	SetEnabled((TWins*)mf->buCancle, False);
	if (xcyMsg & xpyErr_Alg) {
	    
	    switch(bef) {
	    case xpyErr_OverWriteType:
	    case xpyErr_OvwTypeDestDir:
	    case xpyErr_SameFile:
		res = TFormOverWrite_Show(mf->Parent, wd, xcyMsg, 0x07);
		break;
	    default:
		res = xcy_ErrorMsgShow((TWins*)mf, wd, xcyMsg);
	    }
	} else {

	    switch(bef) {
	    case xpyMsg_ovwDir:
	    case xpyMsg_ovwFile:
		res = TFormOverWrite_Show(mf->Parent, wd, xcyMsg, 0x1f);
		break;
	    default:
		res = xcy_ErrorMsgShow((TWins*)mf, wd, xcyMsg);
	    }
	}
	SetEnabled((TWins*)mf->buCancle, True);
	return res;
}

static void set_txt(TFormCopyRun *self, const char *fn) {

	ppcchange(&self->timer_fn, fn);
  	int gptfu = gptfu_DisplayFileStr|gptfu_noName;
	xfree(self->timer_path_src);
	self->timer_path_src =  XCopy_GetSrc_gp_fn(self->xpy_wd,  gptfu);
	xfree(self->timer_path_dest);
	self->timer_path_dest = XCopy_GetDest_gp_fn(self->xpy_wd, gptfu);
}

static bool Copy_StartWait(TFormCopyRun *self, Tstat64 *st_src) {
	XCopy_Start(self->xpy_wd, st_src, 0);
	while (AppHandleMessage() && XCopy_ChkWait(self->xpy_wd));
	return !XCopy_ChkErr(self->xpy_wd) && !self->stop;
}

static TDirList *Copy_TDirList(TglobPath *gp, int followlinks) {
  TDirList *dl = TDirList_Create_gp_name(gp);
	if (dl) {
	    int m = dlobef_level_0;
	    if (followlinks) m = dlobef_level_1;
	    if (TDirList_Open(dl, m)) {
		TDirList_Open_Wait(dl, dlrunstat_ScanEnd);
		TDirFileList_Sort(&dl->dfl, dlsortFu_Name);
		return dl;
	    }
	    Tfree(dl);
	}
	return NULL;
}

static bool Preserv_Del_Dir(TFormCopyRun *self, Tstat64 *st_src, bool Preserv) {
  dword skip;
  bool Ok = True;

	if (Preserv) {
	    skip = XCopy_GetSkip(self->xpy_wd);
	    if (!(skip & xcySkip_Utime)) {
		dword m = (skip & xcyForce_followlinks) | xcyForce_ovwDir |
			  xcySkip_Chmod | xcySkip_Chown;
		XCopy_SetSkip(self->xpy_wd, m);
		XCopy_Start(self->xpy_wd, st_src, xcySFU_Preserv);
		while (AppHandleMessage() && XCopy_ChkWait(self->xpy_wd));
		XCopy_SetSkip(self->xpy_wd, skip);
		Ok = !XCopy_ChkErr(self->xpy_wd) && !self->stop;
	    }
	}

	if (Ok && (XCopy_GetMode(self->xpy_wd) & cpwdMode_cut)) {
	    XCopy_Start(self->xpy_wd, st_src, xcySFU_Del);
	    while (AppHandleMessage() && XCopy_ChkWait(self->xpy_wd));
	    Ok = !XCopy_ChkErr(self->xpy_wd) && !self->stop;
	}
	return Ok;
}

static bool TFormCopyRun_CopyDir(TFormCopyRun *self, TDirFileList *dfl, const char *fn) {
  TDirList *ndl;
  TDirFileList *ndfl;
  Tstat64 *src_st, *dest_st;
  TglobPath *src_gp;
  int cpwdMode;
  bool Ok = True;
  bool followlinks;

	if (self->stop) return False;
	if (dfl) fn = dfl->fn;
	if (!fn || ((fn[0] == '.')&&(fn[1] == 0)) ||
	    ((fn[0] == '.')&&(fn[1] == '.')&&(fn[2] == 0))) return True;

	if (dfl) XCopy_NameToPath(self->xpy_wd);
	set_txt(self, fn);
	XCopy_ChangeName(self->xpy_wd, fn);
	src_st = NULL;

	if (dfl) {
	    src_st = dfl->st;
	    if (!(src_st->st & st_Link) && (src_st->st & st_Dir)) {
	        if (!XCopy_ChkDestDirDouble(self->xpy_wd)) goto ende;
	    }
	}

	Ok = Copy_StartWait(self, src_st);
	if (!Ok) goto ende;

	if (!dfl) {
	    src_st = XCopy_GetSrcStat(self->xpy_wd);
	    if (!src_st) goto ende;
	}
	if (!(src_st->st & st_Dir) || 
	    !XCopy_DestDirIsCreate(self->xpy_wd)) goto ende;

	followlinks = (XCopy_GetSkip(self->xpy_wd) & xcyForce_followlinks) != 0;

	if (!followlinks && (src_st->st & st_Link)) goto ende;

	cpwdMode = XCopy_GetMode(self->xpy_wd);

	src_gp = XCopy_GetSrc_gp(self->xpy_wd);
	dest_st = XCopy_GetDestStat(self->xpy_wd);

	ndl = Copy_TDirList(src_gp, followlinks);
	if (ndl) {
	    if (src_st->st & st_Link) XCopy_SetMode(self->xpy_wd, cpwdMode & ~cpwdMode_cut);
	    ndfl = ndl->dfl;
	    while (ndfl && Ok) {
		Ok = TFormCopyRun_CopyDir(self, ndfl, NULL) && !self->stop;
		ndfl = ndfl->next;
	    }
	    Tfree(ndl);
	}
	gpFree(src_gp);

	XCopy_SetMode(self->xpy_wd, cpwdMode);
	if (Ok) Ok = Preserv_Del_Dir(self, src_st, !dest_st);
	xfree(dest_st);

ende:
	if (dfl) {
	    XCopy_PathToName(self->xpy_wd);
	    set_txt(self, NULL);
	} else {
	    xfree(src_st);
	}
	return Ok;
}

/* _____________________________ Ev __________________________________ */
static void cf_Stop(TFormCopyRun *self) {
	if (!self->stop) {
	    XCopy_Stop(self->xpy_wd);
	    self->stop = True;
// printf("%s(%d): \n", FL);
//	    if (XCopy_ChkWait(self->xpy_wd)) AppHandleMessage();
// printf("%s(%d): \n", FL);
//	    if (XCopy_ChkWait(self->xpy_wd)) AppHandleMessage();
// printf("%s(%d): \n", FL);
	}
}

static void buCancle_OnClick(TButton *self, long KeyState, bool DblClick) {
	cf_Stop(GetParentForm((TWins*)self));
}

/* ____________________________ Resize _______________________________ */
static void pnlTop_OnResize(TWins *self) {
  TFormCopyRun *cf = (TFormCopyRun*)GetParentForm(self);
  int w;
	w = self->width - 43;
	if (w < 40) w = 40;
	SetWidth((TWins*)cf->edFrom, w);
	SetWidth((TWins*)cf->edTo, w);
}

static void pnlMiddle_OnResize(TWins *self) {
  TFormCopyRun *cf = (TFormCopyRun*)GetParentForm(self);
  int w;
	w = self->width - 43;
	if (w < 40) w = 40;
	SetWidth((TWins*)cf->edFile, w);
}


static void pnlBottom_OnResize(TWins *self) {
  TFormCopyRun *cf = (TFormCopyRun*)GetParentForm(self);
  int w;

	w = self->width - 2 - cf->buCancle->width;
	SetLeft((TWins*)cf->buCancle, w);
}

/* ____________________________ Timer Ev _____________________________ */
static void TFormCopyRun_OnTimer(TgTimer* gTimer) {
  TFormCopyRun *cf = gTimer->obj;
  double fa, fb, flg;
  TXCopyStats xcyStats;
  char *pc;
  char buf[80];

	SetCaption((TWins*)cf->edFile, cf->timer_fn);
	SetCaption((TWins*)cf->edFrom, cf->timer_path_src);
	SetCaption((TWins*)cf->edTo,   cf->timer_path_dest);

	if (XCopy_GetStats(cf->xpy_wd, &xcyStats)) {

	    if (cf->FileCount != xcyStats.FileCount) {
		cf->FileCount = xcyStats.FileCount;
		inttopc(buf, xcyStats.FileCount);
		SetCaption((TWins*)cf->lblStk, buf);
	    }

	    if (cf->DataSize != xcyStats.DataSize) {
		cf->DataSize = xcyStats.DataSize;
		file_str_size_pc(buf, xcyStats.DataSize, fstrSize_space);
		SetCaption((TWins*)cf->lblMB, buf);
	    }

	    flg = xcyStats.GesDataSize + xcyStats.DataNowSize;
	    flg /= 0x100000;
	    pc = buf;
	    doubletoppc(&pc, flg, 0x03);
	    ppcmove(&pc, " MB");
	    SetCaption((TWins*)cf->lblMBrun, buf);

	    fb = xcyStats.GesDataTime + xcyStats.DataNowTime;
	    fb /= 1000000000;
	    if (fb > 0) fa = flg / fb; else fa = 0;

	    int a = fa; // * 1000;
	    if (cf->ges_mps != a) {
		cf->ges_mps = a;
		pc = buf;
		ppcmove_v32(&pc, a, kVALUE_int + 0x300);
		ppcmove(&pc, " MB/s");
		SetCaption((TWins*)cf->lblMBpsec, buf);
	    }

	    if (xcyStats.DataNowSize && xcyStats.DataNowTime) {
		flg = xcyStats.DataNowSize;
		flg /= 0x100000;
		fb = xcyStats.DataNowTime;
		fb /= 1000000000;
		fa = flg / fb;
		int b = fa; // * 1000;
		if (cf->mps != b) {
		    cf->mps = b;
		    pc = buf;
		    ppcmove_v32(&pc, b, kVALUE_int + 0x300);
		    ppcmove(&pc, " MB/s");
		    SetCaption((TWins*)cf->lblMBpsecN, buf);
		}
		uint64_t ut = (xcyStats.DataNowSize * 1000) / xcyStats.DataSize;
		TProgressBar_SetProz(cf->pbProz, ut);
		
	    } else {
		SetCaption((TWins*)cf->lblMBpsecN, "0.000 MB/s");
		TProgressBar_SetProz(cf->pbProz, 0);
	    }
	}
}

/* ____________________________ Destroy ______________________________ */
static bool TFormCopyRun__Destroy(TFormCopyRun *self) {
	Tfree(self->gTimer);
	Tfree(self->tls);
	gpFree(self->gp_src);
	gpFree(self->gp_dest);

	xfree(self->timer_fn);
	xfree(self->timer_path_src);
	xfree(self->timer_path_dest);

	return TWins__Destroy((TWins*)self);
}

static void mf_OnCanClose(TFormCopyRun *self, int *CanClose) {
	if (self->xpy_wd) {
	    if (self->buCancle->Enabled) {
		cf_Stop(self);
	    }
	    *CanClose = False;
	    
	} else {
	    *CanClose = FormCloseAndFree;
	}
}

/* ____________________________ Create _______________________________ */
void TFormCopyRun_Show(TWins *Parent, TLines *tls_src, TglobPath *gp_src, 
			TglobPath *gp_dest, int skip) {
  TMainForm *mf;
  TFormCopyRun *self;
  char *pc;
  const char *fn;
  int count;
  TDirList *dl;
  bool CutMv;

	if (!tls_src || !gp_src || !gp_dest) return;
	mf = appDat.obj;
	self = TForm_CreatParamAlg(Parent, sizeof(TFormCopyRun));
	self->destroy = (TDESTROY)TFormCopyRun__Destroy;
	self->FormZStyle = fszTransient;
	if (LastPosX || LastPosY) {
	    self->left = LastPosX;
	    self->top = LastPosY;
	    self->FormPosition = poDesigned;
	} else {
	    self->FormPosition = poCenter;
	}
	self->width = 480;
	self->height = 140;
	self->min_width = 360;
	self->min_height = 140;
	SetCaption((TWins*)self, "Copy");
	self->OnCanClose = mf_OnCanClose;
	self->tls = tls_src;
	self->gp_src = gp_src;
	self->gp_dest = gp_dest;

	self->pnlTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 41, mBorder, alTop);
	self->pnlTop->OnResize = pnlTop_OnResize;
	self->pnlMiddle = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder, alClient);
	self->pnlMiddle->OnResize = pnlMiddle_OnResize;
	self->pnlBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 28, mBorder, alBottom);
	self->pnlBottom->OnResize = pnlBottom_OnResize;

	TWins_CreateLabel(self->pnlTop, 2,  3, 28, 16, "from:", mCaptionLeft, 0);
	TWins_CreateLabel(self->pnlTop, 2, 22, 28, 16, "to :", mCaptionLeft, 0);

	pc = get_DisplayPathStr(NULL, self->gp_src);
	self->edFrom = TEdit_CreateS(self->pnlTop, 40, 2, 45, 18, pc);
	xfree(pc);
	self->edFrom->KeyFu |= mNoKeyFocus;
	self->edFrom->ReadOnly = True;
	SetColor((TWins*)self->edFrom, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edFrom, appXDat.FileSystemUTF8on);

	pc = get_DisplayPathStr(NULL, self->gp_dest);
	self->edTo = TEdit_CreateS(self->pnlTop, 40, 21, 45, 18, pc);
	xfree(pc);
	self->edTo->KeyFu |= mNoKeyFocus;
	self->edTo->ReadOnly = True;
	SetColor((TWins*)self->edTo, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edTo, appXDat.FileSystemUTF8on);


	TWins_CreateLabel(self->pnlMiddle, 2,  3, 28, 16, "file:", mCaptionLeft, 0);

	self->edFile = TEdit_CreateS(self->pnlMiddle, 40, 2, 45, 18, NULL);
	self->edFile->KeyFu |= mNoKeyFocus;
	self->edFile->ReadOnly = True;
	SetColor((TWins*)self->edFile, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edFile, appXDat.FileSystemUTF8on);

	self->lblMB   = TWins_CreateLabel(self->pnlMiddle,  40, 22, 78, 18, 
				"0.000 MB", mCaptionRigth, mBorder|mBorderL);
	self->lblStk  = TWins_CreateLabel(self->pnlMiddle,  120, 22, 48, 18, 
				"0", mCaptionRigth, mBorder|mBorderL);
	self->lblMBrun  = TWins_CreateLabel(self->pnlMiddle, 170, 22, 88, 18, 
				"0.000 MB", mCaptionRigth, mBorder|mBorderL);
	self->lblMBpsec = TWins_CreateLabel(self->pnlMiddle, 260, 22, 88, 18, 
				"0.000 MB/s", mCaptionRigth, mBorder|mBorderL);

	self->pbProz     = TProgressBar_CreateU(self->pnlMiddle,  40, 42, 228, 18);
	self->lblMBpsecN = TWins_CreateLabel(self->pnlMiddle, 260, 42, 88, 18, 
				"0.000 MB/s", mCaptionRigth, mBorder|mBorderL);

	self->buCancle = TButton_CreateS(self->pnlBottom, 2, 2, 64, 24, "CANCEL");
	self->buCancle->OnClick = buCancle_OnClick;

	Set_Focus((TWins*)self->buCancle);

	TForm_Show((TForm*)self);

/* Copy Start */
	count = TLines_GetLineCount(self->tls);
	if (count > 0) {
	    fn = TLines_Get_pc(self->tls, 0);
	    if (fn) SetCaption((TWins*)self->edFile, fn);
	}

	self->xpy_wd = XCopy_Create(self, msg_CopyEv, self->gp_src, self->gp_dest, skip);


	if (!self->xpy_wd) {
	    MsgBox_Show_Error((TWins*)mf, NULL, msgCANCEL, 
			msgCANCEL, EMSG_ERR, 
			"libaldccp can not handle copy");

	} else {
	    self->gTimer = gTimerX_Add(Timer_Enabled, 40, TFormCopyRun_OnTimer);
	    self->gTimer->obj = self;
	    CutMv = appXDat.SelCut;
	    if (CutMv) XCopy_SetMode(self->xpy_wd, cpwdMode_cut);

	    for (int i=0;i<count;i++) {
		fn = TLines_Get_pc(self->tls, i);
		if (fn) {
		    SetCaption((TWins*)self->edFile, fn);
		    if (CutMv) {
			gp_set_name(self->gp_src, fn);
			gp_set_name(self->gp_dest, fn);
			CutMv = net_mv_gp(self->gp_src, NULL, self->gp_dest);
		    }
		    if (!CutMv && !TFormCopyRun_CopyDir(self, NULL, fn)) break;
		}
	    }
	    TFormCopyRun_OnTimer(self->gTimer);
	    TProgressBar_SetProz(self->pbProz, 1000);
	    ppTfree(&self->xpy_wd);
	}

	dl = TDirListObjNy_Get_dl_gp(self->gp_dest);
	if (dl) TDirList_ReScan(dl);

	if (self->gTimer) {
	    dword t = appDat.app_us;
	    while (((appDat.app_us - t) < 100000) && AppHandleMessage());
	}

	LastPosX = self->left;
	LastPosY = self->top;
	TForm_Close((TForm*)self);
	return;
}
