/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appXbase.h"
#include "tdirlist.h"
#include "FormProperty.h"
#include "dlg.h"
#include <utime.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>

static void ObjNy_Open(TFormProperty *self) {
	TDirListObjNy_Net_Create_gp(&self->ObjNy, self, self->gp, NULL);
}

static void ObjNy_Wait(TFormProperty *self) {
	if (!self->ObjNy) ObjNy_Open(self);
	TDirListObjNy_Open_Wait(self->ObjNy);
}

#ifndef _OS_WIN32
 #include <pwd.h>
 #include <grp.h>
 #include "elf_info.h"

static dword propBIT[12]= {
	S_IRUSR, S_IWUSR, S_IXUSR, S_ISUID,
	S_IRGRP, S_IWGRP, S_IXGRP, S_ISGID,
	S_IROTH, S_IWOTH, S_IXOTH, S_ISVTX};

static dword get_propBIT_all() {
	return 0x0fff;
}

static dword get_propBIT(TFormProperty *self) {
  int i;
  dword res = 0;

	for (i=0;i<12;i++) {
	    if (self->cbPerm[i]->Checked) res |= propBIT[i];
	}
	return res;
}

#else
typedef struct {
  int fu;
  const char *name;
  dword mask;
} TpropBIT;

/* ------ TpropBIT.fu ------
b0=fix
b1=inv
b2=dwFileAttributes
---------------------------- */

static const TpropBIT propBIT[12] = {
  {0x01, "Read",     S_IRUSR},
//  {0x01, "Write",    S_IWUSR},
  {0x06, "Write",    FILE_ATTRIBUTE_READONLY},
  {0x01, "Enter",    S_IXUSR},
  {0x01, NULL,       0},

  {0x04, "hidden",   FILE_ATTRIBUTE_HIDDEN},
  {0x04, "archiv",   FILE_ATTRIBUTE_ARCHIVE},
  {0x04, "normal",   FILE_ATTRIBUTE_NORMAL},
  {0x04, "offline",  FILE_ATTRIBUTE_OFFLINE},
//  {0x04, "readonly", FILE_ATTRIBUTE_READONLY},
  {0x04, "system",   FILE_ATTRIBUTE_SYSTEM},
  {0x04, "temp",     FILE_ATTRIBUTE_TEMPORARY},
  {0x05, "compress", FILE_ATTRIBUTE_COMPRESSED},
  {0x01, NULL,       0}
};

static dword get_propBIT_all() {
  int i;
  dword res = 0;

	for (i=0;i<12;i++) {
	    if (propBIT[i].fu & 0x04) res |= propBIT[i].mask;
	}
	return res;
}

static dword get_propBIT(TFormProperty *self) {
  int i;
  dword res = 0;
  bool mk;

	for (i=0;i<12;i++) {
	    mk = self->cbPerm[i]->Checked == ((propBIT[i].fu & 0x02) == 0);
	    if ((propBIT[i].fu & 0x04) && mk) res |= propBIT[i].mask;
	}
	return res;
}
#endif

#ifndef _OS_WIN32
static bool chk_is_num(const char *pc) {
	while (*pc) {
	    if ((*pc < '0') || (*pc > '9')) return False;
	    pc++;
	}
	return True;
}
#endif

static bool ut_SetDate(TFormProperty *self) {
  const char *pc;
  bool res;

	if (!self->edDate) return False;
	pc = GetCaption_pc((TWins*)self->edDate);
	if (!pc || !*pc) return False;
	pnmove(&self->ut, &self->org_ut, sizeof(self->ut));

	if ((GlobDateFormat & 0x03) == 0x01) {
	    self->ut.tm_mon = ppctoint_scan(&pc, True);
	    self->ut.tm_mday = ppctoint_scan(&pc, True);
	    if (!*pc) return False;
	    self->ut.tm_year = ppctoint_scan(&pc, True);

	} else if ((GlobDateFormat & 0x03) == 0x02) {
	    self->ut.tm_year = ppctoint_scan(&pc, True);
	    self->ut.tm_mon = ppctoint_scan(&pc, True);
	    if (!*pc) return False;
	    self->ut.tm_mday = ppctoint_scan(&pc, True);

	} else {
	    self->ut.tm_mday = ppctoint_scan(&pc, True);
	    self->ut.tm_mon = ppctoint_scan(&pc, True);
	    if (!*pc) return False;
	    self->ut.tm_year = ppctoint_scan(&pc, True);
	}

	self->ut.tm_year -= 1900;
	self->ut.tm_mon -= 1;
	self->ut.tm_hour = ppctoint_scan(&pc, True);
	self->ut.tm_min = ppctoint_scan(&pc, True);
	self->ut.tm_sec = ppctoint_scan(&pc, True);

	res = (self->ut.tm_year >= 0) && 
	      (self->ut.tm_mon >= 0)  && (self->ut.tm_mon < 12) && 
	      (self->ut.tm_mday > 0)  && (self->ut.tm_mday < 32) && 
	      (self->ut.tm_hour >= 0) && (self->ut.tm_hour < 24) && 
	      (self->ut.tm_min >= 0)  && (self->ut.tm_min < 60) && 
	      (self->ut.tm_sec >= 0)  && (self->ut.tm_sec < 60);
	if (!res) return False;

	return  (self->ut.tm_year != self->org_ut.tm_year) ||
		(self->ut.tm_mon  != self->org_ut.tm_mon) ||
		(self->ut.tm_mday != self->org_ut.tm_mday) ||
		(self->ut.tm_hour != self->org_ut.tm_hour) ||
		(self->ut.tm_min  != self->org_ut.tm_min) ||
		(self->ut.tm_sec  != self->org_ut.tm_sec);

}

static void chk_Ok_en(TFormProperty *self) {
  const char *pc;

	self->chname = False;
	self->chuser = False;
	self->chgroup = False;

	if (!self->multi) {
	    pc = GetCaption_pc((TWins*)self->edFile);
	    if (!pccmp(self->dfl->fn, pc)) {
		ObjNy_Wait(self);
		if (self->ObjNy) {

		    TDirFileList *dfl = self->ObjNy->dfl;
		    while (dfl) {
#ifdef _OS_WIN32
			if (!pccasecmpASCII(dfl->fn, pc)) {
#else
			if (pccmp(dfl->fn, pc)) {
#endif
			    SetEnabled((TWins*)self->buOk ,False);
			    return;
			}
			dfl = dfl->next;
		    }
		    self->chname = True;
		}
	    }
	}

#ifndef _OS_WIN32
	SetFontDefaultForeColor((TWins*)self->cobUser);
	pc = GetCaption_pc((TWins*)self->cobUser);
	if (pc && *pc && !pccmp(self->org_user, pc)) {
	    if (chk_is_num(pc)) {
		self->uid = pctoint(pc);
		self->chuser = True;
	    } else {
  TUserList *ul = ul_getby_name(pc);
		if (ul) {
		    self->uid = ul->uid;
		    self->chuser = True;
		} else {
		    SetFontForeColor((TWins*)self->cobUser, 0xE00000);
		}
	    }
	}

	SetFontDefaultForeColor((TWins*)self->cobGroup);
	pc = GetCaption_pc((TWins*)self->cobGroup);
	if (pc && *pc && !pccmp(self->org_group, pc)) {
	    if (chk_is_num(pc)) {
		self->gid = pctoint(pc);
		self->chgroup = True;
	    } else {
  TGroupList *gl = gl_getby_name(pc);
		if (gl) {
		    self->gid = gl->gid;
		    self->chgroup = True;
		} else {
		    SetFontForeColor((TWins*)self->cobGroup, 0xE00000);
		}
	    }
	}
#endif
	self->chdate = ut_SetDate(self);

	dword mask = get_propBIT_all();
	bool mk = self->chname || self->chuser || self->chgroup || 
		((self->org_mode & mask) != (get_propBIT(self) & mask)) ||
		((self->mimestat_org & 0x0f) != (self->mimestat & 0x0f)) ||
		self->chdate;
	SetEnabled((TWins*)self->buOk, mk);
}
/*_____________________________ Mime ____________________________________*/
static void pgMime_OnPaint(TPage *self) {
  int y;
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
  TPic *pic;

	y = mf->cbMimeDefault->top - 18;
	DrawPic(self->WinID, mf->default_pic, mf->cbMimeDefault->left + 8, y);

	pic = mf->app_pic;
	if (!pic) pic = mf->default_pic;
	DrawPic(self->WinID, pic, mf->cbMimeApp->left + 8, y);

	pic = mf->mime_pic;
	if (!pic) pic = mf->default_pic;
	DrawPic(self->WinID, pic, mf->cbMimeMime->left + 8, y);
}

static void cbMimeSet(TFormProperty *self, int value) {
	TChkBox_SetChecked(self->cbMimeDefault, (value & 0x01) != 0);
	TChkBox_SetChecked(self->cbMimeMime, (value & 0x02) != 0);
	TChkBox_SetChecked(self->cbMimeApp,  (value & 0x04) != 0);
}

static TPic *set_lbl(TFormProperty *self, TMimeDat *md, char **ppc) {
	*ppc = pcadd_n(3, md->path, md->fn, ".desktop");
	return GetDeskIcon(ddkt_Mime, 16, md->icon, 0, NULL);
}

static void MimePicInit(TFormProperty *self) {
  int m;
  char *pc;
  TPic *pic;
  TDirFileList *dfl = self->dfl;
  TMimeDat md;
  bool cben = False;

	self->mimestat_org &= ~0x0f;
	pc = stat_get_mimetype(dfl->st, dfl->fn);
	self->pcMimeTypeDat = pc;
	SetCaption(self->lblTypeDat, pc);
	SetCaption(self->lblMimeTypeDat, pc);
	self->default_pic = get_fileIcon(dfl->st->FileType);

	if (mimetype2app_getdat(pc, &md)) {
	    self->app_pic = set_lbl(self, &md, &self->pcMimeExecDat);
	}

	if (mimetype2lnk_getdat(pc, &md)) {
	    self->mime_pic = set_lbl(self, &md, &self->pcMimeMimeDat);
	}

	m = 0x01;
	TMime *mime = self->mime;
	if (pc && mime && mime->stat_mime_getpic16) {
	    pic = mime->stat_mime_getpic16(mime, dfl->st);
	    if (!pic || (pic == self->default_pic)) m = 0x01;
	    if (pic && (pic == self->mime_pic)) m = 0x02;
	    if (pic && (pic == self->app_pic))  m = 0x04;
	    cben = True;
	}
	self->mimestat_org |= m;
	self->mimestat = self->mimestat_org;
	cbMimeSet(self, m);
	SetEnabled((TWins*)self->cbMimeDefault, cben);
	SetEnabled((TWins*)self->cbMimeApp, cben);
	SetEnabled((TWins*)self->cbMimeMime, cben);
}

static void cbMime_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
  int m = 0;

	mf->mimestat &= ~0x0f;
	if (mf->cbMimeDefault == self) m |= 0x01;
	if (mf->cbMimeMime == self) m |= 0x02;
	if (mf->cbMimeApp  == self) m |= 0x04;
	mf->mimestat |= m;
	cbMimeSet(mf, m);
	chk_Ok_en(mf);
}

/*_____________________________ Scan ____________________________________*/
static void AzScan(TFormProperty *self) {
  char *pc, *pc0;
  char buf[64];

	pc0 = buf;
	*pc0 = '*';
	if (self->ScanStop) pc0++;

	pc = pc0;
	file_str_size_ppc(&pc, self->fsize, 0);
	ppcmove(&pc, "  (");
	int64toppc(&pc, self->fsize);
	ppcmove(&pc, ")");
	SetCaption(self->lblSizeDat, buf);

	pc = pc0;
	file_str_size_ppc(&pc, self->lsize, 0);
	ppcmove(&pc, "  (");
	int64toppc(&pc, self->lsize);
	ppcmove(&pc, ")");
	SetCaption(self->lblLinkSizeDat, buf);

	pc = pc0;
	inttoppc(&pc, self->DirCount + self->FileCount);
	ppcmove(&pc, "  (");
	inttoppc(&pc, self->DirCount);
	ppcmove(&pc, "/");
	inttoppc(&pc, self->FileCount);
	ppcmove(&pc, ")");
	
	SetCaption(self->lblCountDat, buf);
}

static void ScanNextRun(TFormProperty *self, TDirFileList *dfl, TglobPath *gp) {
  TDirList *dl;
  TglobPath *gpn;

	while (dfl && !self->ScanStop) {

	    if (dfl->st->st & st_Dir) {
		if ((dfl->fn[0] == '.') && 
		    (!dfl->fn[1] || ((dfl->fn[1] == '.') && !dfl->fn[2]))) 
			goto e_next;
		self->DirCount++;
		if (gp && !(dfl->st->st & st_Link)) {
		    gpn = gp_dup(gp);
		    ppcadd2(&gpn->path, dfl->fn, "/");
		    dl = TDirList_Create_gp(gpn);
		    if (TDirList_Open(dl, dlobef_level_1)) {
			TDirList_Open_Wait(dl, dlrunstat_ScanEnd);
			ScanNextRun(self, dl->dfl, gpn);
		    }
		    gpFree(gpn);
		    Tfree(dl);
		}
	    } else {
		self->FileCount++;
	    }

	    self->fsize += dfl->st->size;
	    if (dfl->st->st & st_Link) {
		self->lsize += dfl->st->LinkSize;
	    } else {
		self->lsize += dfl->st->size;
	    }
e_next:
	    dfl = dfl->next;
	}
	if (gp && cpTime(self->ms_Az, GetTime_ms())) {
	    self->ms_Az += 50;
	    AzScan(self);
	    AppProcessMessages();
	}
}

/*________________________ edXXXX_OnKeyDown _____________________________*/
static void edDate_OnKeyDown(TEdit* self, TKeyEvent *KeyEv) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
	if (KeyEv->keysym == XK_Escape) {
	    SetCaption((TWins*)mf->edDate, mf->org_pc_ut);
	    chk_Ok_en(mf);
	}
}

/*________________________ edXXXX_OnChange ______________________________*/
static void edFile_OnChange(TEdit *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
	chk_Ok_en(mf);
}

#ifndef _OS_WIN32
static void cobUser_OnChange(TComboBox *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
	chk_Ok_en(mf);
}

static void cobGroup_OnChange(TComboBox *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
	chk_Ok_en(mf);
}
#endif
/*_____________________________ OnClick _________________________________*/
static void buScan_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormProperty *mf = GetParentForm((TWins*)self);

	mf->ScanRun = true;
	SetVisible((TWins*)mf->buStop, True);
	Set_Focus((TWins*)mf->buStop);
	SetEnabled((TWins*)mf->buScan, False);
	AppProcessMessages();
	mf->ScanStop = False;

	mf->ms_Az = GetTime_ms() + 50;
	mf->fsize = mf->lsize = 0;
	mf->DirCount = mf->FileCount = 0;
	ScanNextRun(mf, mf->dfl, mf->gp);
	AzScan(mf);

	SetEnabled((TWins*)mf->buScan, True);
	Set_Focus((TWins*)mf->buScan);
	SetVisible((TWins*)mf->buStop, False);
	mf->ScanRun = false;
}

static void buStop_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormProperty *mf = GetParentForm((TWins*)self);
	mf->ScanStop = True;
}

static int msgAz(TFormProperty *mf, int button, const char *bef, TNetErrRet *ErrRet) {

	button |= msgCANCEL|msgIcoERROR;
	return NetError_Dlg((TWins*)mf, mf->gp, button, bef, ErrRet);
}

static void buOk_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormProperty *mf = GetParentForm((TWins*)self);
  dword mode, mask, mand, mor, mxor;
  TDirFileList *dfl;
  const char *cpc;
  TpcIni *ini;
  int rete;
  TNetErrRet ErrRet;
  bool cancel, skipall, Ok;

	SetEnabled((TWins*)mf, False);

	mode = get_propBIT(mf);
	mask = get_propBIT_all();
	if (mode != (mf->org_mode & mask)) {
	    cancel = skipall = False;
	    mxor = mode ^ (mf->org_mode & mask);
	    mand = mode | ~mxor | ~mask;
	    mor  = mode & mxor & mask;
	    dfl = mf->dfl;
	    while (dfl && !cancel) {
		mode = (dfl->st->mode & mand) | mor;

		net_err_init(&ErrRet);
		mf->gp->name_const = dfl->fn;
		Ok = net_chmod_gp(mf->gp, &ErrRet, mode);

		if (!Ok && !skipall) {
		    rete = msgAz(mf, msgIGNORE|msgSKIPALL, "chmod", &ErrRet);
		    skipall = (rete == msgSKIPALL);
		    cancel = (rete == msgCANCEL);
		}
		net_err_clr(&ErrRet);
		mf->gp->name = NULL;
		dfl = dfl->next;
	    }
	}
	if (!mf->multi) {
	    mf->gp->name_const = mf->dfl->fn;
	    if (mf->chname) {
		net_err_init(&ErrRet);
		Ok = net_rename_gp(mf->gp, &ErrRet, GetCaption_pc((TWins*)mf->edFile));
		if (Ok) {
		    mf->gp->name_const = GetCaption_pc((TWins*)mf->edFile);   
		} else {
		    msgAz(mf, 0, "rename", &ErrRet);
		    mf->gp->name = NULL;
		}
		net_err_clr(&ErrRet);
	    }

	    if (mf->chdate && mf->gp->name) {
#ifdef _OS_WIN32
		time_t tm = mktime(&mf->ut);
#else
		mf->ut.tm_isdst = -1;
		time_t tm = timelocal(&mf->ut);
//		time_t tm = mktime(&mf->ut);
#endif
		net_err_init(&ErrRet);
		Ok = net_utime_gp(mf->gp, &ErrRet, tm);
		if (!Ok) {
		    msgAz(mf, 0, "utime", &ErrRet);
		}
		net_err_clr(&ErrRet);
	    }

	    mf->gp->name = NULL;

	    if (!mf->chname && mf->pcMimeTypeDat &&
		((mf->mimestat_org & 0x0f) != (mf->mimestat & 0x0f))) {
		ini = TpcIni_Create(TMime_get_fn_mime_overwrite(mf->mime));
		cpc = "USE_SYS_ICON";
		if (mf->mimestat & 0x02) cpc = "USE_MIME_ICON";
		if (mf->mimestat & 0x04) cpc = "USE_APP_ICON";
		TpcIni_WR(ini, "icon", mf->pcMimeTypeDat, cpc);
		TpcIni_SaveAndFree(ini);
	    }
	}

#ifndef _OS_WIN32
	if ((mf->chuser || mf->chgroup) && gp_test_local_drive(mf->gp)) {
  int uid, gid;
	    cancel = skipall = False;
	    dfl = mf->dfl;
	    while (dfl && !cancel) {
		uid = dfl->st->uid;
		gid = dfl->st->gid;
		if (mf->chuser) uid = mf->uid;
		if (mf->chgroup) gid = mf->gid;
		char *fn = pcadd(mf->gp->path, dfl->fn);
		if ((chown(fn, uid, gid) < 0) && !skipall) {
		    rete = MsgBox_Show_Error((TWins*)mf, NULL, 
				msgIGNORE|msgSKIPALL|msgCANCEL|msgIcoERROR, msgCANCEL, 
				EMSG_ERR2|EMSG_SYS, 
				"  chown \n %s\n", fn);
		    skipall = (rete == msgSKIPALL);
		    cancel = (rete == msgCANCEL);
		}
		xfree(fn);
		dfl = dfl->next;
	    }
	}
#endif
	TForm_Close((TForm*)mf);
}

static void buCancel_OnClick(TButton *self, long KeyState, bool DblClick) {
	TForm_Close(GetParentForm((TWins*)self));
}

static void cbPerm_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFormProperty *mf = GetParentForm((TWins*)self);
	chk_Ok_en(mf);
}

/*_____________________________ OnResize ________________________________*/
static void pgProp_OnResize(TPage *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
  int w = self->width - 4;

	SetWidth((TWins*)mf->edFile, w);
	w -= mf->lblCountDat->left-2;
	if (w < 16) w = 16;
	SetWidth(mf->lblTypeDat, w);
	SetWidth(mf->lblLocatDat, w);
	SetWidth(mf->lblLinkDat, w);
	SetWidth(mf->lblSizeDat, w);
	SetWidth(mf->lblLinkSizeDat, w);
	SetWidth(mf->lblCountDat, w);

}

static void pgPerm_OnResize(TPage *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self);
#ifdef _OS_WIN32
  int i, x, xo;

	xo = (self->width - 8) / 4;
	x = 6;
	for (i=0;i<12;i++) {
	    SetLeft((TWins*)mf->cbPerm[i], x);
	    x += xo;
	    if ((i & 0x03) == 0x03) x = 6;
	}

#else
  int i, x, xo, row, col;

	xo = (self->width - 124) / 3;
	x = 46;
	for (col=0;col<4;col++) {
	    SetLeft((TWins*)mf->lblL1[col], x);
	    x += xo;
	}
	i = 0;
	for (row=0;row<3;row++) {
	    x = 52;
	    for (col=0;col<4;col++) {
		SetLeft((TWins*)mf->cbPerm[i], x);
		x += xo;
		i++;
	    }
	    x = self->width - 52;
	    SetLeft((TWins*)mf->lblL3[row], x);
	}
#endif
}

static void lbl_set_w(TWins *lbl, int w, char *s) {
  char *pc;
	SetWidth(lbl, w);
	pc = vds_path_cut(&lbl->vds, s, w);
	SetCaption(lbl, pc);
	xfree(pc);
}

static void pgMime_OnResize(TPage *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm((TWins*)self); //->ParentForm;
  int x, xo;
  int w = self->width - mf->lblMimeMimeDat->left;

	if (w < 16) w = 16;
	SetWidth(mf->lblMimeTypeDat, w);
	lbl_set_w(mf->lblMimeExecDat, w, mf->pcMimeExecDat);
	lbl_set_w(mf->lblMimeMimeDat, w, mf->pcMimeMimeDat);

	xo = (self->width - 50) / 3;
	x = 46;
	SetLeft((TWins*)mf->cbMimeDefault, x);
	x += xo;
	SetLeft((TWins*)mf->cbMimeMime, x);
	x += xo;
	SetLeft((TWins*)mf->cbMimeApp, x);
}

static void pnlBottom_OnResize(TWins *self) {
  TFormProperty *mf = (TFormProperty*)GetParentForm(self);
  int x = self->width;
	x -= mf->buCancel->width;
	SetLeft((TWins*)mf->buCancel, x);
	x -= mf->buOk->width + 4;
	SetLeft((TWins*)mf->buOk, x);
}

/*_____________________________ Destroy _________________________________*/
static bool TFormProperty__Destroy(TFormProperty *self) {

	gpFree(self->gp);
	Tfree(self->ObjNy);

	while (self->dfl) {
	    TDirFileList *dfl = self->dfl;
	    self->dfl = dfl->next;
	    xfree(dfl);
	}

	xfree(self->pcMimeTypeDat);
	xfree(self->pcMimeMimeDat);
	xfree(self->pcMimeExecDat);
	xfree(self->org_pc_ut);

	return TWins__Destroy((TWins*)self);
}

static void mf_OnCanClose(TFormProperty *self, int *CanClose) {
	if (self->ScanRun) {
	    self->ScanStop = true;
	    *CanClose = 0;
	} else {
	    *CanClose = FormCloseAndFree;
	}
}
/*_____________________________ Create __________________________________*/
static void lbl2Create(TWins *Parent, TWins **lbl1, TWins **lbl2, int w, int y, const char *s) {
	*lbl1 = TWins_CreateLabel(Parent, 1, y, w, 16, s, mCaptionLeft, 0);
	*lbl2 = TWins_CreateLabel(Parent, w+2, y, 80, 16, NULL, mCaptionLeft, 0);
}
static TChkBox *cbMimeCreate(TWins *Parent, int y, const char *s) {
  TChkBox *cb = TChkBox_CreateS(Parent,  0, y, 60, 17, s);
	cb->AutoToggle = False;
	cb->OnClick = cbMime_OnClick;
	return cb;
}

void TFormProperty_Show(TWins *Parent, TglobPath *gp, TDirFileList *dfl, 
						TMime *mime, bool IsPath) {
  TFormProperty *self;
  int i, x, y, count;
  char *pc;
  const char *cpc;
  TDirFileList *dfln;
  bool plusDir;
  bool en, isLocal;

	if (!gp || (!dfl && !IsPath)) return;
	self = TForm_CreatParamAlg(Parent, sizeof(TFormProperty));

	self->destroy = (TDESTROY)TFormProperty__Destroy;
	self->FormZStyle = fszTransient;
	self->FormPosition = poCenter;
	self->width = 320;
	self->height = 220;
	self->min_width = 280;
	self->min_height = 200;
	SetCaption((TWins*)self, "Properties");
	self->OnCanClose = mf_OnCanClose;


	self->gp = gp_dup(gp);
	if (IsPath) {
	    gp_PathToName(self->gp);
	    if (!dfl) {
		Tstat64 *st = file_stat_gp(self->gp, fsGetFileAll_UGT);
		if (st) {
		    dfl = dfl_create(st);
		    xfree(st);
		}
	    }
	}
	ppcfree(&self->gp->name);

	self->dfl = dfl;

	self->mime = mime;
	plusDir = False;
	count = 0;
	dfln = dfl;
	while (dfln) {
	    count++;
	    if (dfln->st->st & st_Dir) plusDir = True;
	    dfln = dfln->next;
	}
	self->multi = count > 1;
	isLocal = gp_test_local_drive(self->gp);

	self->pgcProp = TPageControl_Create((TWins*)self);
	self->pgcProp->Align = alClient;
	Set_Parent((TWins*)self->pgcProp, (TWins*)self);

	self->pnlBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 25, 0, alBottom);

	self->pgProp = TPageControl_AddNewPage(self->pgcProp, "General");
	self->pgPerm = TPageControl_AddNewPage(self->pgcProp, "Permission");
	self->pgACL  = TPageControl_AddNewPage(self->pgcProp, "ACL");
	self->pgMime = TPageControl_AddNewPage(self->pgcProp, "Mime");
#ifndef _OS_WIN32
	if (!self->multi && isLocal &&
			    ((dfl->st->FileType == stft_Exec) ||
			     (dfl->st->FileType == stft_SO))) {
	    self->pgELF  = TPageControl_AddNewPage(self->pgcProp, "ELF");
	}
#endif
	
	SetVisible((TWins*)self->pgMime, !self->multi);

	self->pgMime->OnPaint = pgMime_OnPaint;
/* Page General */
	y = 2;
	cpc = NULL;
	if (!self->multi && dfl) cpc = dfl->fn;
	self->edFile = TEdit_CreateS((TWins*)self->pgProp, 2, y, 120, 20, cpc);
	self->edFile->Enabled = !self->multi;
	self->edFile->OnChange = edFile_OnChange;
	en = cpc && *cpc && ((*cpc != '.') || cpc[1]);
	TEdit_SetReadOnly(self->edFile, !en);

	y += 22;
	lbl2Create((TWins*)self->pgProp, &self->lblType,
			 &self->lblTypeDat, 80, y, "Type:");
	y += 18;
	lbl2Create((TWins*)self->pgProp, &self->lblLocat,
			 &self->lblLocatDat, 80, y, "Location:");
	y += 18;
	lbl2Create((TWins*)self->pgProp, &self->lblLink,
			 &self->lblLinkDat, 80, y, "Link to:");
	y += 18;
	lbl2Create((TWins*)self->pgProp, &self->lblSize,
			 &self->lblSizeDat, 80, y, "Size:");
	y += 18;
	lbl2Create((TWins*)self->pgProp, &self->lblLinkSize,
			 &self->lblLinkSizeDat, 80, y, "Link Size:");
	y += 18;
	lbl2Create((TWins*)self->pgProp, &self->lblCount,
			 &self->lblCountDat, 80, y, "Items Count:");

	if (self->gp) {
//	    pc = gptopc(self->gp, gptfu_noPasswd|gptfu_noName);
	    pc = get_DisplayPathStr(NULL, self->gp);
	    SetCaption(self->lblLocatDat, pc);
	    xfree(pc);
	}
	if (dfl->st->st & st_Link) {
	    cpc = Get_ExFileDat_pc(dfl->st, exfdt_LinkName);
	    SetCaption(self->lblLinkDat, cpc);
	}

/* Page Permission */
#ifdef _OS_WIN32
	self->org_mode = dfl->st->dwFileAttributes;
	y = 20;
	x = 6;
	for (i=0;i<12;i++) {
	    self->cbPerm[i] = TChkBox_CreateS((TWins*)self->pgPerm, x, y, 76, 17, propBIT[i].name);
	    SetEnabled((TWins*)self->cbPerm[i], !(propBIT[i].fu & 0x01));
	    bool mkt;
	    if (propBIT[i].fu & 0x04) { //b2=dwFileAttributes
		mkt = ((dfl->st->dwFileAttributes & propBIT[i].mask) != 0);
	    } else {
		mkt = ((dfl->st->mode & propBIT[i].mask) != 0);
	    }
	    
	    self->cbPerm[i]->Checked = mkt ^ ((propBIT[i].fu & 0x02) != 0); //b1=inv
	    self->cbPerm[i]->OnlyInBoxCheck = True;
	    self->cbPerm[i]->OnClick = cbPerm_OnClick;
	    x += 82;
	    if ((i & 0x03) == 0x03) {
		x = 6;
		y += 20;
	    }
	}
#else
  const char *Lb1_Tab[4] = {"Read", "Write", "Enter", "Special"};
  const char *Lb2_Tab[3] = {"User", "Group", "Others"};
  const char *Lb3_Tab[3] = {"UID", "GID", "Sticky"};
  int row, col, xo;
  dword mode = dfl->st->mode;
	en = net_chmod_gp_test(self->gp);
	self->org_mode = mode;
	xo = (self->width - 124) / 3;
	x = 46;
	for (col=0;col<4;col++) {
	    self->lblL1[col] = TWins_CreateLabel((TWins*)self->pgPerm, x, 2, 50, 16, Lb1_Tab[col], mCaptionLeft, 0);
	    x += xo;
	}

	y = 20;
	i = 0;
	for (row=0;row<3;row++) {
	    self->lblL2[row] = TWins_CreateLabel((TWins*)self->pgPerm, 0, y, 50, 16, Lb2_Tab[row], mCaptionLeft, 0);
	    x = 52;
	    for (col=0;col<4;col++) {
		self->cbPerm[i] = TChkBox_CreateS((TWins*)self->pgPerm, x, y, 20, 17, NULL);
		self->cbPerm[i]->Checked = ((mode & propBIT[i]) != 0);
		self->cbPerm[i]->OnClick = cbPerm_OnClick;
		SetEnabled((TWins*)self->cbPerm[i], en);
		x += xo;
		i++;
	    }
	    x = self->width - 52;
	    self->lblL3[row] = TWins_CreateLabel((TWins*)self->pgPerm, x, y, 50, 16, Lb3_Tab[row], mCaptionLeft, 0);
	    y += 20;
	}

	en = gp_test_local_drive(self->gp);
	y += 2;
	TWins_CreateLabel((TWins*)self->pgPerm, 0, y, 50, 16, "User", mCaptionLeft, 0);
	self->org_user = Get_ExFileDat_pc(dfl->st, exfdt_owner);
	self->cobUser = TComboBox_CreateS((TWins*)self->pgPerm, 52, y, 140, 18, self->org_user);
	self->cobUser->OnChange = cobUser_OnChange;
	self->cobUser->popup->max_height = 184;
	SetEnabled((TWins*)self->cobUser, en);
	y += 20;
	TWins_CreateLabel((TWins*)self->pgPerm, 0, y, 50, 16, "Group", mCaptionLeft, 0);
	self->org_group = Get_ExFileDat_pc(dfl->st, exfdt_group);
	self->cobGroup = TComboBox_CreateS((TWins*)self->pgPerm, 52, y, 140, 18, self->org_group);
	self->cobGroup->OnChange = cobGroup_OnChange;
	self->cobGroup->popup->max_height = 184;
	SetEnabled((TWins*)self->cobGroup, en);
#endif

	if (!self->multi) {
	    en = net_utime_gp_test(self->gp);
#ifdef _OS_WIN32
	    if (dfl->st->st & st_Dir) en = False;
	    struct tm *ut = localtime(&dfl->st->mtime);
	    pnmove(&self->org_ut, ut, sizeof(struct tm));
#else
	    localtime_r(&dfl->st->mtime, &self->org_ut);
#endif
	    y += 20;
	    TWins_CreateLabel((TWins*)self->pgPerm, 0, y, 50, 16, "Date", mCaptionLeft, 0);

	    self->org_pc_ut = file_str_time_HMS(dfl->st->mtime);

	    self->edDate = TEdit_CreateS((TWins*)self->pgPerm, 52, y, 140, 20, self->org_pc_ut);
	    self->edDate->OnChange = edFile_OnChange;
	    self->edDate->OnKeyDown = edDate_OnKeyDown;
	    SetEnabled((TWins*)self->edDate, en);

	}
/* Page ACL */

/* Page Mime */
	y = 2;
	lbl2Create((TWins*)self->pgMime, &self->lblMimeType,
			 &self->lblMimeTypeDat, 40, y, "Type:");
	y += 20;
	lbl2Create((TWins*)self->pgMime, &self->lblMimeMime,
			 &self->lblMimeMimeDat, 40, y, "Mime:");
	y += 20;
	lbl2Create((TWins*)self->pgMime, &self->lblMimeExec,
			 &self->lblMimeExecDat, 40, y, "Exec:");

	y += 36;
	self->lblMimeIcon = TWins_CreateLabel((TWins*)self->pgMime, 1, y, 40, 16, "Icon:", mCaptionLeft, 0);
	y += 14;
	self->cbMimeDefault = cbMimeCreate((TWins*)self->pgMime, y, "system");
	self->cbMimeApp     = cbMimeCreate((TWins*)self->pgMime, y, "app");
	self->cbMimeMime    = cbMimeCreate((TWins*)self->pgMime, y, "mime");

	if (!self->multi) MimePicInit(self);

/* Page ELF */
#ifndef _OS_WIN32
	if (self->pgELF) {
	    self->leELF = TLineEdit_CreateA((TWins*)self->pgELF, 0, 0, 16, 16, alClient);
	    self->gp->name_const = dfl->fn;
	    char *fn = gptopc(self->gp, gptfu_noRootArt);
	    TELF_Info *elf_info = elf_getinfo(fn, NULL, 0);
	    xfree(fn);
//	    TLineEdit_AddPaste(self->leELF, "1");
//	    TLineEdit_AddPaste(self->leELF, "2");
//	    TLineEdit_AddPaste(self->leELF, "3");
	    if (elf_info) {

		const char *s = "CPU-UNKNOWN  \n";
		if (elf_info->e_machine == EM_386) s = "CPU-386  \n";
		if (elf_info->e_machine == EM_X86_64) s = "CPU-x86_64  \n";
		if (elf_info->e_machine == EM_ARM) s = "CPU-ARM  \n";
		TLineEdit_AddPaste(self->leELF, s);

		s = "32Bit-";
		if (elf_info->isElf64) s = "64Bit-";
		TLineEdit_AddPaste(self->leELF, s);
		s = "\"Invalid data encoding\"  ";
		if (elf_info->e_encoding == ELFDATA2LSB) s = "LSB  ";
		if (elf_info->e_encoding == ELFDATA2MSB) s = "MSB  ";
		TLineEdit_AddPaste(self->leELF, s);

		s = "NONE";
		if (elf_info->e_type == ET_REL)  s = "REL";
		if (elf_info->e_type == ET_EXEC) s = "EXEC";
		if ((elf_info->e_type == ET_DYN) && elf_info->isSO)  s = "DYN";
		if ((elf_info->e_type == ET_DYN) && !elf_info->isSO) s = "EXEC-DYN";
		if (elf_info->e_type == ET_CORE) s = "CORE";
		TLineEdit_AddPaste(self->leELF, s);
		if (elf_info->gnuHASH) TLineEdit_AddPaste(self->leELF, "  ld(GNU_HASH)");
		TLineEdit_AddPaste(self->leELF, "\n");

		TELF_InfoItem *ii = elf_info->InfoItem;
		while (ii) {
		    if (ii->verneed) {
			TLineEdit_AddPaste(self->leELF, ii->so_name);
			TLineEdit_AddPaste(self->leELF, " (");
			TLineEdit_AddPaste(self->leELF, ii->verneed);
			TLineEdit_AddPaste(self->leELF, ")\n");
		    } else if (ii->verdef) {
			TLineEdit_AddPaste(self->leELF, ii->so_name);
			TLineEdit_AddPaste(self->leELF, " (");
			TLineEdit_AddPaste(self->leELF, ii->verdef);
			TLineEdit_AddPaste(self->leELF, ")\n");
		    } else {
			TLineEdit_AddPaste(self->leELF, ii->so_name);
			TLineEdit_AddPaste(self->leELF, "\n");
		    }
		    ii = ii->next;
		}

	    } else {
		TLineEdit_AddLine(self->leELF, "ELF Invalid data encoding");
	    }

	    elf_info_free(elf_info);
	    self->gp->name = NULL;
	}

#endif

/* pnlBottom */
	self->buScan = TButton_CreateS((TWins*)self->pnlBottom, 0, 1, 64, 24, "Scan");
	self->buScan->Enabled = plusDir;
	self->buScan->OnClick = buScan_OnClick;

	self->buStop = TButton_CreateS((TWins*)self->pnlBottom, 68, 1, 50, 24, "Stop");
	SetVisible((TWins*)self->buStop, False);
	self->buStop->OnClick = buStop_OnClick;

	self->buOk = TButton_CreateS((TWins*)self->pnlBottom, 20, 1, 64, 24, "OK");
	self->buOk->OnClick = buOk_OnClick;
	self->buOk->Enabled = False;
	self->buCancel = TButton_CreateS((TWins*)self->pnlBottom, 100, 1, 64, 24, "Cancel");
	self->buCancel->OnClick = buCancel_OnClick;


	self->pgProp->OnResize = pgProp_OnResize;
	self->pgPerm->OnResize = pgPerm_OnResize;
	self->pgMime->OnResize = pgMime_OnResize;
	self->pnlBottom->OnResize = pnlBottom_OnResize;

/* Set AzDaten */
	self->ScanRun = true;
	ScanNextRun(self, self->dfl, NULL);
	self->ScanRun = false;
	self->ScanStop = plusDir;
	AzScan(self);

/* Set Path Daten .... */
	ObjNy_Open(self);

#ifndef _OS_WIN32
  TUserList *ul = ul_get_next(NULL);
	while (ul) {
	    TComboBox_AddItem(self->cobUser, ul->name, 0, NULL);
	    ul = ul_get_next(ul);
	}
  TGroupList *gl = gl_get_next(NULL);
	while (gl) {
	    TComboBox_AddItem(self->cobGroup, gl->name, 0, NULL);
	    gl = gl_get_next(gl);
	}
#endif

	TWins_SetUTF8((TWins*)self->edFile, appXDat.FileSystemUTF8on);
	TForm_Show((TForm*)self);
	return;
}
