/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"
#include "dlg.h"
#include "tdirlist.h"
#include "net_io.h"
#include "image.h"
#include "cicons.h"
#include "buildin_app/aldfind.h"


static void mf_chk_mbTerm_En(TMainForm *mf) {
  TglobPath *gp;
  bool mk;
	gp = TFileWin_Get_gp(mf->fw);
	mk = gp_test_local_drive(gp);
	SetEnabled((TWins*)mf->mbTerm, mk);
//	SetEnabled((TWins*)mf->mbTools, mk);
#ifdef USE_BULDIN_APP
	SetEnabled((TWins*)mf->mbFind, TFormFind_ChkWorkDirOk(gp));
#endif
}

/*_______________________ edPath Ev _____________________________________*/
static void edPath_OnKeyDown(TComboBox *self, TKeyEvent *KeyEv) {
  TMainForm *mf = appDat.obj;
  char *fn;
  const char *cfn;
  TPathInfo pathi;

	if (KeyEv->keysym == XK_Return) {
	    cfn = GetCaption_pc((TWins*)mf->edPath);
	    if (!cfn || !*cfn) cfn = "/";
	    if (!get_PathInfo(cfn, &pathi)) return;

	    if (pathi.fu & (pathi_Local|pathi_rel)) {
		if (pathi.fu & pathi_rel) {
		    fn = path_rel2abs(cfn, NULL);
		} else {
		    fn = dotdot_korr(cfn);
		}
		Tstat64 *st = file_stat_fn(fn, fsNoFileName|fsNoLinkName|fsLinkDirekt);
		if (!st) {
		    return;
		}
		if (st->st & st_Dir) {
		    path_add_Slash_ppc(&fn, NULL);
		}
		xfree(st);
		if (pathi.fu & pathi_rel) SetCaption((TWins*)mf->edPath, fn);

	    } else {
		fn = dotdot_korr(cfn);
		if (!path_end_isSlash(fn)) path_sub_ppc(&fn);
	    }

	    TFileWin_LoadDir(mf->fw, fn);
	    xfree(fn);
	    mf_chk_mbTerm_En(mf);
	}
}

/*_____________________________ TFileWin Ev _____________________________*/
static TTreeVNode *get_net_URL_tvn(TTreeVNode *tvn, TglobPath *gp) {
  TTreeVNode *res = NULL;
  TMainForm *mf = appDat.obj;

	while (tvn) {
	    if (tvn->next_child) res = get_net_URL_tvn(tvn->next_child, gp);
	    if (res) {
		tvn->st |= tvnSt_Expanded;
		return res;
	    }
	    if (((tvn->art & tfvfart_MASK) == tfvfart_ROOT) && 
		(TTreeVFile_SetSel_gp(mf->tv, tvn))) {
		if (pccmp(mf->tv->sel_gp.URL, gp->URL) &&
		    pccmp(mf->tv->sel_gp.user, gp->user)) {

		    if (mf->tv->sel_gp.passwd) NetPasswdSetup_gp(&mf->tv->sel_gp);
		    return tvn;
		}
	    }
	    tvn = tvn->next;
	}
	return NULL;
}

#define GURN_NodeAdd 1
#define GURN_NewLogin (1<<1)


static TTreeVNode *tv_ftp_DummyRoodNode(TglobPath *gp) {
  TMainForm *mf = appDat.obj;
  char *node_name;
  TTreeVNode *tvn;

	if (gp->user) {
	    node_name = pcadd_n(3, gp->user, "@", gp->URL);
	} else {
	    node_name = pcdup(gp->URL);
	}
	tvn = TTreeVFile_AddDummyRoodNode(mf->tv, mf->tvftp, NULL, node_name, NULL, gp);
	xfree(node_name);
	tvn->pic = get_pic16_ftp(False);
	TTreeV_SetSelNode((TTreeV*)mf->tv, tvn);
	Refresh((TWins*)mf->tv);
	return tvn;
}

static TTreeVNode *mf_GetUrlRootNode(TMainForm *mf, TglobPath *gp, int *gurn) {
  TTreeVNode *tvn, *tmp_tvn, *run_tvn;
  TNetErrRet ErrRet;
  int dummygurn = 0;
	
	if (!gp) return NULL;
	if (!gurn) gurn = &dummygurn;

	tvn = NULL;
	if (gp_test_local_drive(gp)) {
	    tvn = mf->tvrfn;
#ifdef _OS_WIN32
	    while (tvn && pccasecmpASCII(gp->RootArt, tvn->s)) tvn = tvn->next;
#endif

	} else if (!pccasecmpASCII(gp->RootArt, "ftp:")) {
	    TTreeV_NodeExpand((TTreeV*)mf->tv, mf->tvnet);
	    if (!mf->tvftp) {   //Add Node ftp
		mf->tvftp = TTreeV_AddNode((TTreeV*)mf->tv, mf->tvnet, NULL, "ftp:");
		mf->tvftp->pic = get_fileIcon(stft_Folder);
	    }
	    TTreeV_NodeExpand((TTreeV*)mf->tv, mf->tvftp);
	    tvn = get_net_URL_tvn(mf->tvftp->next_child, gp);

	    if (tvn && tvn->next_child) return tvn;
	    if (!(*gurn & GURN_NodeAdd)) return tvn;

	    tmp_tvn = NULL;
	    if (tvn) {
		run_tvn = tvn;
		TTreeV_SetSelNode((TTreeV*)mf->tv, tvn);
	    } else {
		run_tvn = tmp_tvn = tv_ftp_DummyRoodNode(gp);
	    }
	    TRunImg_Add(run_tvn, NULL);

	    net_err_init(&ErrRet);
	    bool Ok = net_chk_open_gp(gp, &ErrRet, 0);
	    TRunImg_Del(run_tvn);

	    if (!Ok) {
		net_err_clr(&ErrRet);
		Tfree(tmp_tvn);
		Refresh((TWins*)mf->tv);
		return NULL;
	    }
	    if (ErrRet.errFlag & NetErrFlag_NewLogin) {
		ppcchange_cmp(&gp->passwd, ErrRet.new_passwd);
		if (ppcchange_cmp(&gp->user, ErrRet.new_user)) {
		    Tfree(tmp_tvn);
		    tmp_tvn = NULL;
		}
		*gurn |= GURN_NewLogin;
	    }
	    net_err_clr(&ErrRet);

	    tvn = get_net_URL_tvn(mf->tvftp->next_child, gp);
	    if (tvn) {
		TglobPath *tmp_gp = TTreeVFileNode_Get_gp(tvn);
		if (tmp_gp) ppcchange_cmp(&tmp_gp->passwd, gp->passwd);
		return tvn;
	    }
	    tvn = tv_ftp_DummyRoodNode(gp);
	}
	return tvn;
}



static void ExpandRootNode(TTreeVNode *tvn, TglobPath *gp, bool AddNode) {
  TMainForm *mf = appDat.obj;
  char *pc, *pce;
  TTreeVNode *tvnfoc;

	if (tvn) {
	    tvnfoc = tvn;
	    pc = gp->path;
	    while (*pc && tvn) {
		tvnfoc = tvn;
		tvn = tvn->next_child;
		while (*pc && (*pc == '/')) pc++;
		pce = pc;
		while (*pce && (*pce != '/')) pce++;
		if (pc != pce) {
		    TTreeV_NodeExpand((TTreeV*)mf->tv, tvnfoc);
		    tvn = tvnfoc->next_child;
		    tvn = TTreeVNode_get_by_name_lg(tvn, pc, pce-pc, True);
		    if (!tvn) {
			if (!AddNode) break;
			char *pcn = pcndup(pc, pce-pc);
			tvn = TTreeVFile_AddDummyNode(mf->tv, tvnfoc, NULL, pcn);
			xfree(pcn);
			tvn->st |= tvnSt_Expanded;
			tvn->pic = get_fileIcon(stft_Folder);
		    }
		    TTreeV_SetNodeVisible((TTreeV*)mf->tv, tvn, True);
		    tvnfoc = tvn;
		}
		pc = pce;
	    }
	    if (tvnfoc && TTreeVNode_TestNodeIsExpand(tvnfoc, tvnCF_ALLChild)) {
		 TTreeV_SetSelNode((TTreeV*)mf->tv, tvnfoc);
	    }
	    Refresh((TWins*)mf->tv);
	}
}


static void fw_OnLoadDir(TFileWin *self, char **ppfn, int *stat) {
  TMainForm *mf = appDat.obj;
  TglobPath *gp;
  TTreeVNode *tvn; //, *tvnx;
  int gurn;

	if (*stat != fwOn_stat_Start) return;
	gp = pctogp(*ppfn);
	gp_NameToPath(gp);
	gurn = GURN_NodeAdd;
	tvn = mf_GetUrlRootNode(mf, gp, &gurn);
	if (tvn) {
//	    tvnx = TTreeVFileNode_FindNode(tvn, gp->path);

	    if (gurn & GURN_NewLogin) {
		xfree(*ppfn);
		*ppfn = gptopc(gp, gptfu_PasswdTxt);

		char *tmp = get_DisplayFileStr(NULL, gp);
		SetCaption((TWins*)mf->edPath, tmp);
		xfree(tmp);
	    }
	    ExpandRootNode(tvn, gp, True);
	} else {
	    *stat = fwOn_stat_RetHalt;
	}
	gpFree(gp);
}

// ******* in aldfile_exec.c ******

static void fw_OnItemEnter(TFileWin *self, TItemEnter *ie) {
  TMainForm *mf = appDat.obj;

	if (!(ie->stat & 0x02)) SetCaption((TWins*)mf->edPath, ie->fn);
	if (ie->art == ie_artFileEnter) {
	    fw_ItemEnter((TWins*)self, ie);
	}
}

static void Set_lblStat() {
  TMainForm *mf = appDat.obj;
  TDirFileList *dfl;
  int items = 0;
  int files = 0;
  int dirs = 0;
  int64_t ges = 0;
  char *pc;
  char buf[80];

	dfl = TFileWin_Get_dfl0(mf->fw);
	while (dfl) {
	    items++;
	    if (dfl->st->st & st_Dir) {
		 if (!pccmp(dfl->fn,"..")) dirs++;
	    } else files++;
	    if (dfl->st->st & st_Link) ges += dfl->st->LinkSize; else ges += dfl->st->size;
	    dfl = dfl->next;
	}
	pc = file_str_size(ges, 0);
	snprintf(buf, sizeof(buf), "%d Items  %d Files  %d Dirs  %s", items, files, dirs, pc);
	xfree(pc);

	mf_chk_mbTerm_En(mf);

	SetCaption((TWins*)mf->lblStat, buf);
}

static void fw_OnFileMsg(TFileWin *self, TFileMsg *fmsg) {
  int bef; //, befn;
  TglobPath *gp;
  char *pc, *pce;
  TTreeVNode *tvn, *tvnfoc;
  TMainForm *mf = appDat.obj;

// file_msg_debug(fmsg, "FileWin   ");
	if (!file_msg_chk_wtyp(fmsg, wtyp_TDirListObjNy, FL)) return;
	if ((fmsg->art != fmArt_DirList)||(fmsg->typD != fmTyp_int)) return;
	fmsg->Ok = True;
	bef = fmsg->D.m;
//	befn = bef & dlbef_BefMask;

	switch (bef & dlbef_BefMask) {

	case dlbef_DelSelf:
	    fmsg->Ok = False;  //in TDirListObjNy   Tfree(ObjNy);
	case dlbef_ListEnd:
	    TRunImg_Del_ObjNy(fmsg->obj);
	case dlbef_Sync:
	    Set_lblStat();
	    break;

/*	case dlbef_Open:
	    if (!(bef & dlbef_NotifyRun)) tv_URL_open_expand(fmsg);
	    break;
*/
	case dlbef_Start:
	    gp = pctogp(self->Akt_FileName);
	    gp_NameToPath(gp);
	    tvnfoc = NULL;
	    tvn = mf_GetUrlRootNode(mf, gp, NULL);
	    if (tvn) {
		tvnfoc = tvn;
		pc = gp->path;
		while (*pc && tvn) {
		    tvnfoc = tvn;
		    tvn = tvn->next_child;
		    while (*pc && (*pc == '/')) pc++;
		    pce = pc;
		    while (*pce && (*pce != '/')) pce++;
		    if (pc != pce) {
			tvn = tvnfoc->next_child;
			tvn = TTreeVNode_get_by_name_lg(tvn, pc, pce-pc, True);
			if (tvn) tvnfoc = tvn;
		    }
		    pc = pce;
		}
	    }
	    gpFree(gp);
	    if (tvnfoc) TRunImg_Add(tvnfoc, fmsg->obj);

	    break;
	}
}

static void tv_OnFileMsg(TTreeVFile *self, TTreeVNode *tvfn, TFileMsg *fmsg) {
  int bef;

// file_msg_debug(fmsg, "TTreeV    ");
	if ((fmsg->art != fmArt_DirList)||(fmsg->typD != fmTyp_int)) return;
	fmsg->Ok = True;
	bef = fmsg->D.m;

	switch (bef & dlbef_BefMask) {

//	case dlbef_Open:
//	    break;

	case dlbef_Start:
	    TRunImg_Add(tvfn, NULL);
	    break;

	case dlbef_ListEnd:
	    TRunImg_Del(tvfn);
	    break;

	case dlbef_DelSelf:
	    fmsg->Ok = False;  //in TDirListObjNy   Tfree(ObjNy);
	    TRunImg_Del(tvfn);
	    break;
	}

}

static void fw_OnTimer(TgTimer *ti) {
  TMainForm *mf = appDat.obj;
	mf->fw_cTi--;
	if (mf->fw_cTi) return;
	ti->func |= Timer_OnTimerAndFree;
	xfree(mf->fw_sTi);
	mf->fw_sTi = NULL;
}

static void fw_OnKeyDown(TFileWin *self, TKeyEvent *KeyEv) {
  TMainForm *mf = appDat.obj;
  TDirFileList *dfl;
  int idx;
  int lg = 0;
  char tmp[7];
	if (KeyEv->keysym == XK_BackSpace) {
	    if (mf->fw_sTi) {
		mf->fw_cTi = 10;
		lg = pclen(mf->fw_sTi);
		if (lg>0) mf->fw_sTi[lg-1] = 0;
	    }
	} else if (keysymtostr(&KeyEv->keysym, &tmp[0], self->vds.UTF8on)) {
	    if (!mf->fw_cTi) gTimerX_Add(Timer_Enabled, 100, fw_OnTimer);
	    mf->fw_cTi = 10;
	    ppcadd(&mf->fw_sTi ,tmp);
	}
	if (!mf->fw_sTi) return;
	idx = 0;
	lg = pclen(mf->fw_sTi);
	if (!lg) {
	    TFileWin_SetSelect(mf->fw, 0, 0x03);
	    return;
	}
	dfl = TFileWin_Get_dfl0(mf->fw);
	while (dfl) {
	    if (pcncmp(mf->fw_sTi, dfl->fn, lg)) {
		TFileWin_SetSelect(mf->fw, idx, 0x03);
		return;
	    }
	    dfl = dfl->next;
	    idx++;
	}
	idx = 0;
	dfl = TFileWin_Get_dfl0(mf->fw);
	while (dfl) {
	    if (pcncasecmp(mf->fw_sTi, dfl->fn, lg)) {
		TFileWin_SetSelect(mf->fw, idx, 0x03);
		return;
	    }
	    dfl = dfl->next;
	    idx++;
	}
}

/*_____________________________ TTreeV Ev _______________________________*/
static void tvf_OnNodeClick(TTreeVFile *self, TTreeVNode *tvn, bool DblClick) {
  TMainForm *mf = appDat.obj;
  char *fn;

	if (tvn->ttc) return;

	fn = TTreeVFile_GetNodeFileName_passwd(self, tvn);
	if (fn) {
	    char *tmp = get_DisplayFileStr(fn, NULL);
	    SetCaption((TWins*)mf->edPath, tmp);
	    xfree(tmp);
	    if (DblClick) {
//		if (!tvn->ttc && tvn->next_child) TFileWin_ReScan(mf->fw);
		TFileWin_ReScan(mf->fw);

	    } else {
		bool mk = TFileWin_LoadDir(mf->fw, fn);
		mf_chk_mbTerm_En(mf);
		
		if (mk) TTreeVFile_ChkOpenNode(mf->tv, 
					TTreeVFile_GetSelectNode(mf->tv));
	    }
	    xfree(fn);
	}
}

static void tv_OnChkOpenNode(TTreeVFile *self, TTreeVNode **tvn) {
  TglobPath *gp_tmp;
  int gurn;
  TTreeVNode *tvnx;

	if (!pccasecmpASCII(self->sel_gp.RootArt, "Ftp:")) {

	    if ((((*tvn)->art & tfvfart_MASK) != tfvfart_ROOT) ||
	        (*tvn)->next_child) return;
	    gp_tmp = gp_dup(&self->sel_gp);
	    gurn = GURN_NodeAdd;
	    tvnx = mf_GetUrlRootNode(appDat.obj, gp_tmp, &gurn);
	    gp_free(gp_tmp);
	    if (tvnx) {
		TTreeVFile_SetSel_gp(self, tvnx);
		TTreeV_SetSelNode((TTreeV*)self, tvnx);
	    }
	    if (*tvn != tvnx) (*tvn)->st &= ~tvnSt_Expanded;
	    *tvn = tvnx;
	}
}

static bool tv_OnExpandChange(TTreeVFile *self, TTreeVNode *tvn, int ExpEv) {
	return tvn->ttc == NULL;
}

/*_____________________________ Menu Ev _________________________________*/
static void mbPathRet_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
  TMainForm *mf = appDat.obj;

	TFileWin_PathEnter(mf->fw, NULL, "..", ie_artPathEnter);
}


static void mbTerm_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
  char *pc;
  TMainForm *mf = appDat.obj;
  TglobPath *gp = TFileWin_Get_gp(mf->fw);

	if (!gp_test_local_drive(gp)) return;
	char *path = gptopc(gp, gptfu_noName);
	pc = path_to_local(path);
	xfree(path);
	if (TestIsPath(pc)) {
#ifdef _OS_WIN32
	    exec_pc_co_path("cmd", NULL, pc);
#else
	    ppcchange_add_n(&pc, 3, "aldconsole --workdir \"", pc, "\"");
	    exec_pc(pc);
#endif
	} else {
	    MsgBox_Error((TWins*)mf, "No such directory\n\"%s\"", pc);
	}
	xfree(pc);
}

#ifdef USE_BULDIN_APP
static void mbFind_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
  TMainForm *mf = appDat.obj;
  TglobPath *gp = TFileWin_Get_gp(mf->fw);
  void *FormFind;

	if (!gp_test_local_drive(gp)) return;

	char *path = gptopc(gp, gptfu_noName);
	char *pc = path_to_local(path);
	xfree(path);
	FormFind = TFormFind_Create(NULL, pc);
	xfree(pc);
	TForm_Show(FormFind);
}
#endif

static void mbAzArt_OnClick(TMenuButton *self, long KeyState, bool DblClick) {
  TMainForm *mf = appDat.obj;
  int m, AzArt;

	AzArt = mf->fw->AzArt;
	m = (AzArt & 0xff) + 1;
	if (m > 2) m = 0;
	if (m == 2) m |= 0x100;
	TFileWin_SetColsArt(mf->fw, m);
	if (m & 0x01) {
	    mf->mbAzArt->pic = get_pic16_file_LAZ1(False);
	} else {
	    mf->mbAzArt->pic = get_pic16_file_LAZ0(False);
	}
	RePaint((TWins*)self);
	mf->miDblItem->fu &= ~miChecked;
	mf->miDblPath->fu &= ~miChecked;
	if (mf->fw->DblClick_ItemEnter) mf->miDblItem->fu |= miChecked;
	if (mf->fw->DblClick_PathEnter) mf->miDblPath->fu |= miChecked;
}

/*_____________________________ Resize __________________________________*/
void aldfile_PanelTop_OnResize(TWins *self) {
  TMainForm *mf = appDat.obj;
  int w;

	w = mf->mbHelp->left + mf->mbHelp->width + 2;
	if (mf->mbTerm->Visible) w = mf->mbTerm->left + mf->mbTerm->width + 2;
#ifdef USE_BULDIN_APP
	if (mf->mbFind->Visible) w = mf->mbFind->left + mf->mbFind->width + 2;
#endif
	SetLeft((TWins*)mf->edPath, w);
	w = self->width - mf->edPath->left - mf->mbPathRet->width - mf->mbAzArt->width - 4;
	SetWidth((TWins*)mf->edPath, w);
	w += mf->edPath->left + 2;
	SetLeft((TWins*)mf->mbPathRet, w);
	w += mf->mbPathRet->width;
	SetLeft((TWins*)mf->mbAzArt, w);
}

static void PanelBottom_OnResize(TWins *self) {
  TMainForm *mf = appDat.obj;
  int w;
	w = self->width - mf->lblURLStat->left;
	SetWidth(mf->lblURLStat, w);
}

/*_____________________________ Form Ev __________________________________*/
static void TMainForm_OnKeyDown(TMainForm *self, TKeyEvent *KeyEv) {
	if (KeyEv->keysym == XK_F5) {
	    TFileWin_ReScan(self->fw);
	}
}

/*_____________________________ aldstatMimeEV ____________________________*/
static void app_mime_OnChange(TMime *self, int bef) {
  TMainForm *mf = appDat.obj;
	switch (bef) {
	case MimeOnChange_all: 
	    TDirListObjNy_ReScanMime();
	    break;
	case MimeOnChange_item: 
	    Refresh((TWins*)mf->fw);
	    break;
	}
}

/*_____________________________ ErrorMsg ________________________________*/

static void Global_OnFileMsg(TFileMsg *fmsg) {

	if ((fmsg->art == fmArt_errLogin) || (fmsg->art == fmArt_AlgErr)) {
	    default_Global_OnFileMsg(fmsg);
	}
}

/*_____________________________ Start ___________________________________*/
static const char *path_tab[] = {
  "/share/applications/",
  "/share/mimelnk/",
  "/Menu/OpenWith/",
//  "/Menu/Tools/",
  "/Net/ftp/",
#ifdef _OS_WIN32
  "/applications/",
#endif
  NULL
};
static const char *path_subtab[] = {
  "application/",
  "audio/",
  "image/",
  "text/",
  "video/",
  NULL
};

static void aldfile_OnAppStart() {
  TMainForm *mf = appDat.obj;
  char *pc, *pc2;
  const char **ppc;

	ppc = path_tab;
	while (*ppc) {
	    pc = pcadd(appDat.ALDini_dir, *ppc);
	    file_mkdir_parents_ae(pc, 0750, EMSG_ERR1);
	    xfree(pc);
	    ppc++;
	}

	pc2 = pcadd(appDat.ALDini_dir, "/share/mimelnk/");
	ppc = path_subtab;
	while (*ppc) {
	    pc = pcadd(pc2, *ppc);
	    file_mkdir_parents_ae(pc, 0750, EMSG_ERR1);
	    xfree(pc);
	    ppc++;
	}
	xfree(pc2);

	AppProcessMessages();

	app_mime = TMime_X_Create();
	app_mime->OnChange = app_mime_OnChange;

	TTreeVFile_OpenRoodNode(mf->tv, mf->tvrfn);
#ifndef _OS_WIN32
	tv_net_create();
#endif
	mf->OnKeyDown = TMainForm_OnKeyDown;
}

/*_____________________________ Create __________________________________*/
#ifndef ALD_BUILD_V
#define ALD_BUILD_V 1
#endif

static const char *DefaultIniDat = {
  "[aldfile]\n"
  "MouseZ_Wheel=2\n"
  "Width=792\n"
  "Height=420\n"
  "[AppExecFlag]\n"
  ";AppExecFlag --> chdir onlydir onlyLocal\n"
  "aldconsole=onlydir onlyLocal\n"
#ifdef _OS_WIN32
  "cmd=onlydir onlyLocal chdir\n"
#else
  "kfind=onlydir onlyLocal\n"
  "konsole=onlydir onlyLocal chdir\n"
#endif
  "[aldfind]\n"
  "ModifiedVisible=True\n"
  "PermissonVisible=False\n"
  "OnlyASCII=False\n"
  "TextCaseSens=True\n"
  };

const char *Help_Msg = {
  "\n"
  "Usage: aldfile [options]\n"
  "\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: aldfile version "VERSION"  (build %d  %s)\n"
};

const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;

int main (int argc, char **argv) {
  TMainForm *self;
  int i, sbc;
  char *val;
  bool help = False;
  bool err = False;

#ifdef _OS_WIN32
	cmdUTF8korr(&argc, &argv);
#else
	cmdline_slash_korr(argc, argv);

#ifdef USE_BULDIN_APP
  TBuildin_app *bi_app;
	bi_app = Get_Buildin_pc(argv[0]);
	if (bi_app) return bi_app(argc, argv);
	if ((argc > 1) && (*argv[1] != '-')) {
	    cmdline_del_argv0(&argc, argv);
	    bi_app = Get_Buildin_pc(argv[0]);
	    if (bi_app) return bi_app(argc, argv);
	    BuildinHelp();
	    return EXIT_FAILURE;
	}
#endif
#endif /* ! _OS_WIN32 */

	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'h':
		    help = True; break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
#ifdef Test_Root_WinId
		    } else if (pcncmp(val, "TestRootWinID", 13)) {
			i++;
			val = argv[i];
			if (val && *val != '-') {
				TestRootWinId = pctoint(val);
			} else err = True;
#endif
		    } else err = True;
		    break;
		default: 
			err = True;
		}  // switch end
	    } else {
		err = True;
	    }
	    i++;
	}

	if (help || err) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
#ifndef _OS_WIN32
#ifdef USE_BULDIN_APP
	    printf("\nor:\n");
	    BuildinHelp();
#endif
#endif
	    exit(EXIT_FAILURE);
	}

	initXapp(0, argc, argv);
	hGetFileHeadInfo = &hGetFileHeadInfo_buildin;

	hGetDeskIcon = &hGetDeskIcon_file;
#if 0
	ThImageRD_Reg(5, &hImageRD_xpm, &hImageRD_png, &hImageRD_jpeg, 
			 &hImageRD_gif, &hImageRD_pnm);
#else
	ThImageRD_Reg(6, &hImageRD_xpm, &hImageRD_png, &hImageRD_jpeg, 
			 &hImageRD_gif, &hImageRD_bmp, &hImageRD_pnm);
#endif

#ifdef _OS_WIN32
	ThDirList_Reg(1, &hdirlist_ld_File);
	ThDirList_ny_Reg(1, &hdirlist_ny_File);
	ThXCopy_Reg(1, &hXCopy_ld_File);
	appDat.inifile_app = pcadd(appDat.ALDini_dir, "\\aldfile.ini");
#else
//	ThDirList_Reg(1, &hdirlist_ld_File);
//	ThDirList_Reg(1, &hdirlist_ld_File_pth);
	ThDirList_Reg(2, &hdirlist_ld_File_pth, &hdirlist_ld_ftp);

//	ThDirList_ny_Reg(1, &hdirlist_ny_File);
	ThDirList_ny_Reg(3, &hdirlist_nyi_File, &hdirlist_ny_File, &hdirlist_ny_Ftp);

	ThXCopy_Reg(2, &hXCopy_ld_File, &hXCopy_ld_Ftp);

	ThNetLineBufLoader_Reg(1, &hNetLineBufLd_ftp);
	ThNetStatLoader_Reg(1, &hNetStatLd_ftp);
	NetLoaderReg_Tools(1, &hNetToolsLd_ftp);

	appDat.inifile_app = pcadd(appDat.ALDini_dir, "/aldfile.ini");
#endif
	setDefaultIniDat(DefaultIniDat);
	appDat.appIni = TpcIni_Create(appDat.inifile_app);

#ifdef USE_DECOR
	appXDat.DefaultButtonDecor = TDecor_button_w3b3_Create();
#endif

	self = TForm_CreatParamAlg(NULL, sizeof(TMainForm));
	appDat.obj = self;
	InitIconPaths(NULL);

	self->vds.UTF8on = appXDat.UTF8on;
#ifndef _OS_WIN32
	self->Icon = GetDeskIcon_z(ddkt_Application, 16, APPNAME, 
				    gdifu_noZoom, zico16globus);
#endif /* ! _OS_WIN32 */

	i = TpcIni_RD_int(appDat.appIni, APPNAME, "Height", 420);
	if (i<240) i = 240;
	if (i>1400) i = 1400;
	self->height = i;
	i = TpcIni_RD_int(appDat.appIni, APPNAME, "Width", 792);
	if (i<480) i = 480;
	if (i>2000) i = 2000;
	self->width = i;
	self->min_height = 240;
	self->min_width = 480;

	self->IconName = APPNAME;
	SetCaption((TWins*)self, APPNAME);


	self->pnlTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder | mBorderL, alTop);
	self->pnlTop->OnResize = aldfile_PanelTop_OnResize;
	self->pnlMiddle = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, 0, alClient);
	self->pnlBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 19, 0, alBottom);
	self->pnlBottom->OnResize = PanelBottom_OnResize;

	MenuFile_Create(self);
	MenuEdit_Create(self);
//	MenuTools_Create(self);
	MenuSetting_Create(self);
	MenuHelp_Create(self);

	self->mbTerm = TMenuButton_CreateS(self->pnlTop, self->mbFile, NULL);
	self->mbTerm->pic = GetZIcon(zico16console);
	self->mbTerm->dis_pic = GetDisZIcon(zico16console);
	self->mbTerm->OnClick = mbTerm_OnClick;
	self->mbTerm->Enabled = False;
	SetVisible((TWins*)self->mbTerm, (self->miTermVis->fu & miChecked) != 0);
#ifdef USE_BULDIN_APP
	self->mbFind = TMenuButton_CreateS(self->pnlTop, self->mbFile, NULL);
	self->mbFind->pic = GetZIcon(zico16aldfind);
	self->mbFind->dis_pic = GetDisZIcon(zico16aldfind);
	self->mbFind->OnClick = mbFind_OnClick;
	self->mbFind->Enabled = False;
	SetVisible((TWins*)self->mbFind, (self->miFindVis->fu & miChecked) != 0);
#endif
	self->edPath = TComboBox_CreateS(self->pnlTop, 88, 2, 200, 18, NULL);
	self->edPath->OnKeyDown = edPath_OnKeyDown;
	self->mbPathRet = TMenuButton_CreateS(self->pnlTop, self->mbFile, NULL);
	self->mbPathRet->pic = get_pic16_undo(False);
	self->mbPathRet->OnClick = mbPathRet_OnClick;

	self->mbAzArt = TMenuButton_CreateS(self->pnlTop, self->mbFile, NULL);
	self->mbAzArt->pic = get_pic16_file_LAZ0(False);
	self->mbAzArt->OnClick = mbAzArt_OnClick;

	TMenuButton_SetPosition(self->mbFile); //Set Default size


// *********** TTreeVFile
	i = TpcIni_RD_int(appDat.appIni, APPNAME, "SideBarWidth", 196);
	self->tv = TTreeVFile_CreateA((TWins*)self->pnlMiddle, 0, 0, i, 50, alLeft);
	self->tv->pw->min_width = 80;
	self->tv->OnNodeClick = (TtvNodeClick)tvf_OnNodeClick;
	self->tv->OnChkOpenNode = tv_OnChkOpenNode;
	self->tv->OnExpandChange = (TOnExpandChange)tv_OnExpandChange;
	self->tv->PopupMenu = self->menuEdit;
	self->tv->OnFileMsg = tv_OnFileMsg;
	self->tvnet = TTreeV_AddNode((TTreeV*)self->tv, NULL, NULL, "Network");
	self->tvnet->pic = get_fileIcon(stft_Folder);
	  self->tvftp = TTreeVFile_AddNode(self->tv, self->tvnet, NULL, "ftp:");
	  self->tvftp->pic = get_pic16_ftp(False);

#ifdef _OS_WIN32
  char DriveList[256];

	TTreeVNode *tvnx = self->tvnet;
	GetLogicalDriveStrings(256, DriveList);
	char *pc = DriveList;
	while (*pc) {
	    char bufa[4];
	    TTreeVNode *tvn = NULL;
	    dword DriveTyp = GetDriveType(pc); 
	    pcnmove(bufa, pc, 2);

	    switch (DriveTyp) {
	    case DRIVE_REMOVABLE:
		tvn = TTreeVFile_AddRoodNode(self->tv, NULL, tvnx, bufa, pc);
		if (!pcncasecmpASCII(pc, "a:", 2) || !pcncasecmpASCII(pc, "b:", 2)) {
		    tvn->pic = get_pic16_floppy_umount(False);
		} else {
		    tvn->pic = get_pic16_usbdrive_umount(False);
		}
//		if (GetDiskFreeSpaceExA(pc, (PULARGE_INTEGER)&FreeBytes, (PULARGE_INTEGER)&TotalBytes, (PULARGE_INTEGER)&TotalFreeBytes)) {
// printf("%s(%d): %llu %llu %llu\n", FL, FreeBytes, TotalBytes, TotalFreeBytes);
//		}
		break;

	    case DRIVE_FIXED:
		tvn = TTreeVFile_AddRoodNode(self->tv, NULL, tvnx, bufa, pc);
		tvn->pic = get_pic16_hdd_umount(False);
		if (!self->tvrfn) self->tvrfn = tvn;
		break;

	    case DRIVE_CDROM:
		tvn = TTreeVFile_AddRoodNode(self->tv, NULL, tvnx, bufa, pc);
		tvn->pic = get_pic16_cdrom_umount(False);
		break;
//	    default:
// printf("%s(%d): DriveTyp=%d\n", FL, DriveTyp);
	    }
	    if (tvn) {
		tvnx = tvn;
	    }
	    while (*pc) pc++;
	    pc++;
	}
	if (!self->tvrfn) self->tvrfn = self->tvnet;
#else
	self->tvrfn = TTreeVFile_AddRoodNode(self->tv, NULL, self->tvnet, "File:", "File:/");
	self->tvrfn->pic = get_fileIcon(stft_Folder);
	self->tvrfn->st = tvnSt_Expanded | tvnSt_ForceCanExp;
#endif
	self->splitt = TSplitterV_CreateP((TWins*)self->pnlMiddle);


// *********** TFileWin
	self->fw = TFileWin_CreateA((TWins*)self->pnlMiddle, 0, 0, 200, 200, alClient);
//	self->fw->vds.UTF8on = True;
	self->fw->pw->min_width = 80;
	self->fw->OnKeyDown = fw_OnKeyDown;
	self->fw->OnItemEnter = fw_OnItemEnter;
	self->fw->OnFileMsg = fw_OnFileMsg;
	self->fw->OnLoadDir = fw_OnLoadDir;
	self->fw->PopupMenu = self->menuEdit;
	TFileWin_InitAzPic(self->fw);
	TFileWin_SetShowHidden(self->fw, True);

	TFileWin_ini_RD(self->fw, appDat.appIni, APPNAME);
	TFileWin_ini_RD_ColD(self->fw, appDat.appIni, APPNAME);

// *********** pnlBottom
	self->lblStat = TWins_CreateLabel(self->pnlBottom, 0, 2, 230, 17, "0 Items  0 Files  0 Dirs", 0, mBorder|mBorderL);
	self->lblURLStat = TWins_CreateLabel(self->pnlBottom, 231, 2, 230, 17, NULL, mCaptionLeft, mBorder|mBorderL);

	aldfile_LoadSetting(self);
	sbc = TpcIni_RD_int(appDat.appIni, APPNAME, "MouseZ_Wheel", 3);
//	self->fw->MouseZ_Wheel = TpcIni_RD_int(appDat.appIni, APPNAME, "MouseZ_Wheel", 3);
//	self->tv->MouseZ_Wheel = self->fw->MouseZ_Wheel;
	TFileWin_SetMouseZ_Wheel(self->fw, sbc);
	TTreeVFile_SetMouseZ_Wheel(self->tv, sbc);

	appXDat.OnAppStart = aldfile_OnAppStart;

	Tfree(appDat.appIni);
	appDat.appIni = NULL;

	init_tvImgRun();
	Glob_OnFileMsg = Global_OnFileMsg;

	self->tv->SelectNode = self->tvrfn;
	Set_Focus((TWins*)self->tv);
	appError_ev = default_Error_ev; //Error_ev;

	SetFileSystemUTF8(self, appXDat.FileSystemUTF8on);
	AppRun();

#ifdef MEM_CHK
  Tfree(self);
  AppHandleMessage();
  int lMCt;
  dword lMsize;
  MEM_CHK_Count(&lMCt, &lMsize);
  printf("Debug %s(%d): MEM_CHK_Count=%3d  size=%u wins=%d\n", FL, lMCt, lMsize, GetTWinsCount());
#endif
}
