/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"
#include "dlg.h"
#include "tdirlist.h"
#include "elf_info.h"
#include "FormExecWith.h"

static void ie_exec(TItemEnter *ie, TMimeDat *md) {
// printf("%s(%d): exec --> %s\n", FL, md->exec);
// printf("%s(%d): exec --> %s%s\n", FL, md->path, md->fn);
  TgpList *gpList;
  TDesktopCMD dskcmd;

	if (ie->dfl) {
	    gpList = xcalloc(sizeof(TgpList));
	    gp_dup_gpdata_path(&gpList->gp, ie->gp);
	    gpList->gp.name = pcdup(ie->dfl->fn);

	    dskcmd.gpList = gpList;
	    dskcmd.Exec = md->exec;
	    get_desktop_dskcmd(&dskcmd);
	    if (dskcmd.ret_cmd) {
// printf("%s(%d): exec --> %s\n", FL, dskcmd.ret_cmd);
		if (!exec_pc(dskcmd.ret_cmd)) {
		    appError(EMSG_ERR1|EMSG_SYS, NULL);
		}
		xfree(dskcmd.ret_cmd);
	    }
	    gpList_free(gpList);
	}
}

void fw_ItemEnter(TWins *parent, TItemEnter *ie) {
  char *pc, *pc_mime;
  TMimeDat md;
  char buf[200];

#ifdef _OS_WIN32
	if (ie->dfl->st->FileType == stft_Win32Exe) {
	    if (gp_test_local_drive(ie->gp)) {
		pc = gptofnpc(ie->gp, gptfu_noName, ie->dfl->fn);
		exec_pc(pc);
		xfree(pc);
		return;
#else
	if (ie->dfl->st->FileType == stft_Exec) {
	    if (gp_test_local_drive(ie->gp)) {
		pc = pcadd(ie->gp->path, ie->dfl->fn);
		bool X11ok = False;
		TELF_Info *elf_info;
		elf_info = elf_getinfo(pc, NULL, ELFfkt_SONEED);
		if (elf_info) {
#if defined (__arm__)
		    if (elf_info->e_machine == EM_ARM) {
#elif defined (__x86_64__)
		    if (elf_info->e_machine == EM_X86_64) {
#elif defined (__i386__)
		    if (elf_info->e_machine == EM_386) {
#else
# error "Unsupported architecture!"
#endif
			TELF_InfoItem *ii = elf_info->InfoItem;
			while (ii && !X11ok) {
			    X11ok = pcncmp(ii->so_name, "libX11.so", 9);
			    ii = ii->next;
			}
		    }
		    elf_info_free(elf_info);
		}
// printf("%s(%d): %d %s\n", FL, X11ok, pc);
		if (X11ok) exec_pc(pc);
		xfree(pc);
		if (X11ok) return;
#endif

	    } else {
		appError(EMSG_ERR1, "%s(%d): can only exec local file\n  \"%s\"", FL, ie->fn);
		return;
	    }
	}

	pc_mime = stat_get_mimetype(ie->dfl->st, ie->fn_passwd);
// printf("%s(%d): %d/%d --> %s\n", FL, ie->dfl->st->FileType, ie->dfl->st->sub_mime, pc_mime);

	if (!pc_mime) {

	    pc_mime = inFormExecWith_chk_app_mime(ie->fn, &md);
	    if (pc_mime) {
		snprintf(buf, sizeof(buf), "'Found: %s\n\nAdd mime: %s\n", 
				    md.exec, pc_mime);
		int ret = MsgBox_Show(parent, "Found App", buf,
				    msgOK | msgCANCEL|msgIcoINFO, msgCANCEL);
		if (ret == msgOK) {
		    pc = buf;
		    *pc = '*'; pc++;
		    *pc = '.'; pc++;
		    pcxmove(pc, GetFileExt_pc(ie->fn), 31);
		    pctoLcaseASCII(buf);
		    pc = pcadd_n(4, appDat.ALDini_dir, 
			    "/share/mimelnk/", pc_mime, ".desktop");
#ifdef _OS_WIN32
		    slash_korr(pc, True);
#endif
		    dotdesktop_new(pc, NULL, "", pc_mime, buf, NULL, ddkt_Mime);
		    xfree(pc);
		} else {
		    ppcfree(&pc_mime);
		}
	    }
	}

	if (mimetype2app_getdat(pc_mime, &md)) {
	    ie_exec(ie, &md);
	} else if (pcncmp(pc_mime, "text/", 5) && mimetype2app_getdat("text/plain", &md)) {
	    ie_exec(ie, &md);
	} else if (TFormExecWith_Show(parent, ie->fn_passwd, ie->dfl->st, &md)) {
	    ie_exec(ie, &md);
	    xfree(md.exec);
	}
	xfree(pc_mime);
#if 0
  char *pfn;
  TglobPath *gp;
	gp = ie->gp;
	if (gp_test_local_drive(gp)) {
	    pfn = gptopc(gp, gptfu_noName|gptfu_noRootArt);
	    ppcadd(&pfn, ie->dfl->fn);
	    pc_mime = libmagic_get_mime(pfn);
 printf("%s(%d): libmagic --> *%s*\n", FL, pc_mime);
	    xfree(pfn);
	}
#endif
}

