/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#ifndef STANDALONE_BUILDIN
#include "buildin_app.h"
#endif
#include "appXbase.h"
#include "dlg.h"
#include "image.h"
#include "cicons.h"
#include "tdirlist.h"
#include "tfind_grid.h"
#include "tfilewin.h"
#include "aldfind.h"
#include "cmpx.h"

#define APPNAME "aldfind"

#define MAX_PUSH_DATA 10

#ifdef STANDALONE_BUILDIN
const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;
#else
extern int build_v;
extern char *build_date;
#endif

extern void fw_ItemEnter(TWins *parent, TItemEnter *ie);

//void TGrid_DoSelect_110(TGrid *self, int fu);


void *pthd_find_Create();

/* ------------------------- File Data --------------------------------- */
static TExFileDat *SetUp_ExFileDat(TExFileDat *ExDat, const char *s, int lg, int exfdt) {
	ExDat->type = exfdt;
	ExDat->fu = 0;
	ExDat->lg = lg;
	if (s) pcxmove(ExDat->dat, s, lg-1);
	return (TExFileDat*)(ExDat->dat + lg);
}

#define SEGMSIZE 8192


#if stft_ListEnde != 43
#error stft_ListEnde <> 43
#endif

static const char scanfiletab[stft_ListEnde] = {
	0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
	0, 0, 0, 0, 0,  0, 1, 1, 0, 0,
	0, 0, 1, 1, 1,  1, 1, 0, 1, 0,
	0, 1, 1, 0, 0,  1, 1, 1, 0, 0,
	1, 1, 0
};


static bool  __attribute__ ((noinline)) scanfile(TFormFind *self, const char *path, TDirFileList *dfl) {
  int count, max_char, res;
  dword sgz;
  size_t sz, fsize;
  FileHandle fd;
  char *pc, *buf, *finds;
  bool ok, only_ascii;

	only_ascii = self->cbOnlyASCII->Checked;
	finds = self->cbFText->vds.s;
	max_char = self->cbFText->vds.lg;
	if (!only_ascii) max_char = cmpx_GetCount(self->cmpx);

	if (!max_char) return False;
	fsize = dfl->st->size;
	if (max_char > fsize) return False;

	if (!self->cbInclBinFile->Checked) {
	    int ft = dfl->st->FileType;
	    if ((ft >= stft_ListEnde) || !scanfiletab[ft]) return False;
	}

	pc = path_add(path, dfl->fn);
// printf("%s(%d): %s\n", FL, pc);
	ok = file_open_rd(pc, NULL, &fd);
	xfree(pc);
	if (!ok) return False;

	if (!only_ascii) max_char = cmpx_GetMaxChars(self->cmpx);
//	eof = True;
	sgz = SEGMSIZE * 3;
	if (sgz > fsize) {
	    sgz = fsize;
//	    eof = True;
	    if (!only_ascii) max_char = cmpx_GetCount(self->cmpx);
	}

	res = 0;
	buf = xmalloc(sgz);
	if (file_read_full(fd, NULL, buf, sgz, &sz) && (sz > 0)) {
	    fsize -= sz;
// printf("%s(%d): %d %u\n", FL, sz, fsize);

	    count = sz;
	    pc = buf;
	    
loop:
	    if (self->cbTextCaseSens->Checked) {
		if (only_ascii) {
		    while (count >= max_char) {
			if ((*pc == *finds) && pcncmp(pc, finds, max_char)) {
			    res = fcmp_ascii;
			    break;
			}
			pc++;
			count--;
		    }
		} else {
		    while (count >= max_char) {
			if (cmpx_c0(self->cmpx, pc)) {
			    res = cmpx_cmp(self->cmpx, pc, count);
			    if (res) break;
			}
			pc++;
			count--;
		    }
		}

	    } else {
		if (only_ascii) {
		    while (count >= max_char) {
			if ((l_iso8859_1[(byte)*pc] == l_iso8859_1[(byte)*finds]) && !pcncasecmp_fi(pc, finds, max_char, FU_CASE_8859_1)) {
			    res = fcmp_ascii;
			    break;
			}
			pc++;
			count--;
		    }
		} else {
		    while (count >= max_char) {
			if (cmpx_c0_case(self->cmpx, pc)) {
			    res = cmpx_cmp_case(self->cmpx, pc, count);
			    if (res) break;
			}
			pc++;
			count--;
		    }
		}
	    }


	    if (!res && (fsize > 0)) {
		pnmove(buf, buf + SEGMSIZE, SEGMSIZE * 2);
		if (file_read_full(fd, NULL, buf + (SEGMSIZE * 2), SEGMSIZE, &sz) && (sz > 0)) {
		    fsize -= sz;
		    count += sz;
		    pc -= SEGMSIZE;
		    if (!only_ascii && (fsize <= 0)) max_char = cmpx_GetCount(self->cmpx);
		    goto loop;
		}
	    }

	    self->FirstLineCharType = res;
	    if (res & (fcmp_ascii | fcmp_utf8)) {
		char *pca = pc;
		int m = 80;
		while  ((pca > buf) && m &&
			pca[-1] && 
			(pca[-1] != 0x0a) && 
			(pca[-1] != 0x0d)) {pca--; m--;}

		m = 200;
		while  (*pc && count && m &&
			(*pc != 0x0a) && 
			(*pc != 0x0d)) {pc++; m--; count--;}

		int l = pc - pca;
//		if (res & fcmp_ascii) {
//		    m = AsciiToUtf8(NULL, 0, pca, l) + 1;
//		    self->FirstLine = xmalloc(m);
//		    AsciiToUtf8(self->FirstLine, m, pca, l);
//		} else {
		    self->FirstLine = pcndup(pca, l);
//		}
	    }

	    if (res & fcmp_u16) {
		word *pwc  = (void*)pc;
		word *pwca = pwc;
		int m = 80;
		while (((void*)pwca > (void*)(buf + 1)) && m &&
			pwca[-1] && 
			(pwca[-1] != 0x0a) && 
			(pwca[-1] != 0x0d)) {pwca--; m--;}

		count /= 2;
		m = 200;
		while  (*pwc && count && m &&
			(*pwc != 0x0a) && 
			(*pwc != 0x0d)) {pwc++; m--; count--;}

		int l = pwc - pwca;
		m = WCharsToUtf8(NULL, 0, pwca, l, 0) + 1;
		self->FirstLine = xmalloc(m);
		WCharsToUtf8(self->FirstLine, m, pwca, l, 0);
	    }
	}

	xfree(buf);
	file_close(fd, NULL);
	return res > 0;
}


static void NewTFileItem(TFormFind *self, TDirFileList *dfl, const char *path, 
			 const char *dflpath) {
  int lg;
  TExFileDat *ExDat;
  TDirFileList *fi;
  const char *ChatType;

	self->FirstLine = NULL;
	if (self->cbFText->vds.lg) {
	    if (!scanfile(self, dflpath, dfl)) return;
	}

	ChatType = NULL;
	lg = sizeof(TDirFileList) + dfl->st->StructSize + 2;
	if (path) lg += pclen(path) + sizeof(TExFileDat) + 1;
	if (self->FirstLine) {
	    lg += pclen(self->FirstLine) + sizeof(TExFileDat) + 1;
	    if (self->FirstLineCharType & fcmp_ascii) {
		ChatType = "ascii";
	    } else if (self->FirstLineCharType & fcmp_utf8) {
		ChatType = "utf8";
	    } else if (self->FirstLineCharType & fcmp_u16) {
		ChatType = "u16";
	    } else if (self->FirstLineCharType & fcmp_u16BE) {
		ChatType = "u16BE";
	    }
	    if (ChatType) {
		lg += pclen(ChatType) + sizeof(TExFileDat) + 1;
	    }
	}
	fi = xcalloc(lg);
	pnmove(fi->st, dfl->st, dfl->st->StructSize);

	ExDat = &fi->st->ExDat;
Loop:
	if (!(ExDat->fu & exfdf_lastdat) && ExDat->type) {
	    if (ExDat->type == exfdt_FileName) {
		fi->fn = ExDat->dat;
	    }
	    ExDat = (TExFileDat*)&ExDat->dat[ExDat->lg];
	    goto Loop;
	} else {
	    if (path) {
//		fi->path = ExDat->dat;
		ExDat = SetUp_ExFileDat(ExDat, path, pclen(path)+1, exfdt_user);
	    }
	    if (self->FirstLine) {
//		fi->match = ExDat->dat;
		ExDat = SetUp_ExFileDat(ExDat, self->FirstLine, pclen(self->FirstLine)+1, exfdt_user + 1);
		xfree(self->FirstLine);
		if (ChatType) {
		    ExDat = SetUp_ExFileDat(ExDat, ChatType, pclen(ChatType)+1, exfdt_user + 2);
		}
	    }
	    ExDat->fu = exfdf_lastdat;
	}
	TFindGrid_AddItem(self->grFind, fi);

	dword ms = GetTime_ms();
	if ((ms - self->ref_ms) > 100) {
	    self->ref_ms = ms;
	    Refresh((TWins*)self->grFind);
	    AppProcessMessages();
	}

	self->FindCount++;
	if (self->FindCount >= self->FindCountMax) {
	    self->Stop = True;
	}
}

// ---------------------- pgFiles ----------------------
static void pgFiles_OnResize(TPage *self) {
  int w;
  TFormFind *mf = GetParentForm((TWins*)self);
	w = self->width - mf->cbName->left - 2;
	if (w < 20) w = 20;
	SetWidth((TWins*)mf->cbName, w);

	w -= (mf->buPath->width + 2);
	if (w < 20) w = 20;
	SetWidth((TWins*)mf->cbPath, w);
	SetLeft((TWins*)mf->buPath, mf->cbPath->left + w + 2);
}

static void pgData_OnResize(TPage *self) {
  int w;
  TFormFind *mf = GetParentForm((TWins*)self);
	w = self->width - mf->cbFText->left - 2;
	if (w < 20) w = 20;
	SetWidth((TWins*)mf->cbFText, w);
}

static void buPath_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormFind *mf = GetParentForm((TWins*)self);
  char *pc;
  const char *cpc;

	if (!mf->File_Dlg) mf->File_Dlg = Create_File_Dlg((TWins*)mf);
	cpc = GetCaption_pc((TWins*)mf->cbPath);
	if (cpc) File_Dlg_SetPath(mf->File_Dlg, cpc);
	mf->File_Dlg->DirMode = True;
	mf->File_Dlg->RetName_gp_fu |= gptfu_noRootArt;

	if (File_Dlg_Execute(mf->File_Dlg)) {
	    pc = NULL;
	    File_Dlg_GetEdFileName(mf->File_Dlg, &pc);
	    SetCaption((TWins*)mf->cbPath, pc);
	    xfree(pc);
	}
}


static void pgc_OnResize(TPageControl *self) {
  TFormFind *mf = GetParentForm((TWins*)self);
  int x = self->width - mf->mbFile->width - mf->mbSetting->width - mf->mbHelp->width - 4;

	SetLeft((TWins*)mf->mbFile, x);
	TMenuButton_SetPosition(mf->mbFile);
}

static void pnlTop2_OnResize(TWins *self) {
  TFormFind *mf = GetParentForm(self);
  int x = self->width - mf->buClose->width - 1;

	SetLeft((TWins*)mf->buClose, x);
	x = x - mf->buStop->width - 4;
	SetLeft((TWins*)mf->buStop, x);
	x = x - mf->buStart->width - 4;
	SetLeft((TWins*)mf->buStart, x);
	x = x - mf->edSelTxt->left - 4;
	if (x < 10) x = 10;
	SetWidth((TWins*)mf->edSelTxt, x);
}

// ---------------------- Start ------------------------
static bool chkPathSkip(const char *path) {
	return  pcncmp(path, "/dev", 4) || 
		pcncmp(path, "/proc", 5) || 
		pcncmp(path, "/sys", 4);
}

static void mf_findfile_next(TFormFind *self, const char *path, 
			     const char *name, char **subDir) {
  TDirList *dl;
  TDirFileList *dfl;
  char *pc;
  const char *fn;

	if (self->Stop) return;
	pc = path_add(path, name);
	if (chkPathSkip(pc)) return;
	dl = TDirList_Create_name(pc);
	if (dl) {
	    TDirList_Open(dl, dlobef_level_2_Rescan);
	    TDirList_Open_Wait(dl, dlrunstat_ReScanEnd);

	    dfl = dl->dfl;
	    while (!self->Stop && dfl) {
		fn = dfl->fn;
		if (!((fn[0] == '.') && (fn[1] == '.') && !fn[2])) {
		    if (dfl->st->st & st_Dir) {
			if (!(dfl->st->st & st_Link) && self->cbSubDir->Checked) {
			    ppcadd_n(subDir, 5, "\"", name, "/", fn, "\"");
			}
		    } else if (TPattern_Match(self->patt, fn, dfl->st)) {
			NewTFileItem(self, dfl, name, pc);
		    }
		}
		dfl = dfl->next;
	    }
	}
	Tfree(dl);
	xfree(pc);
}

static void mf_findfile(TFormFind *self, const char *path) {
  TDirList *dl;
  TDirFileList *dfl;
  char *token, *subDir_next, *subDir;
  const char *indata, *fn;

	if (chkPathSkip(path)) return;

	dl = TDirList_Create_name(path);
	if (!dl) return;

	subDir = NULL;
	if (!self->Stop) {
	    TDirList_Open(dl, dlobef_level_2_Rescan);
	    TDirList_Open_Wait(dl, dlrunstat_ReScanEnd);
	    self->ref_ms = GetTime_ms();

	    dfl = dl->dfl;
	    while (!self->Stop && dfl) {
		fn = dfl->fn;
		if (!((fn[0] == '.') && (fn[1] == '.') && !fn[2])) {
		    if (dfl->st->st & st_Dir) {
			if (!(dfl->st->st & st_Link) && self->cbSubDir->Checked) {
			    ppcadd_n(&subDir, 3, "\"", fn, "\"");
			}
		    } else if (TPattern_Match(self->patt, fn, dfl->st)) {
			NewTFileItem(self, dfl, NULL, path);
		    }
		}
		dfl = dfl->next;
	    }
	}

	Tfree(dl);

loop:
	self->ref_ms = GetTime_ms();
	if (self->grFind->fiCount_LastPaint != self->grFind->fiCount) {
	    Refresh((TWins*)self->grFind);
	    AppProcessMessages();
	}

	subDir_next = NULL;
	indata = subDir;
	while (indata && !self->Stop) {
	    indata = GetToken(indata, &token);
	    if (token) {
		mf_findfile_next(self, path, token, &subDir_next);
		xfree(token);
	    }
	}
	xfree(subDir);

	if (subDir_next) {
	    subDir = subDir_next;
	    goto loop;
	}
	if (self->grFind->fiCount_LastPaint != self->grFind->fiCount) {
	    Refresh((TWins*)self->grFind);
	}
}

static void SetCaption_edSelTxt(TFormFind *self) {
	if (self->cbSelPlusPart->Checked && self->SelPath) {
	    char *s = pcadd_n(3, self->SelPath, "/", self->SelName);
#ifdef _OS_WIN32
	    slash_korr(s, True);
#endif
	    SetCaption((TWins*)self->edSelTxt, s);
	    xfree(s);
	} else {
	    SetCaption((TWins*)self->edSelTxt, self->SelName);
	}
}

static void cbSelPlusPart_OnClick(TChkBox *self, long KeyState, bool DblClick) {
	SetCaption_edSelTxt(GetParentForm((TWins*)self));
}

static void mf_buStart_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormFind *mf = GetParentForm((TWins*)self);
  const char *pc;
  TpcIni *ini;

	mf->Stop = False;
	mf->FindCount = 0;
	ppcfree(&mf->SelPath);
	ppcfree(&mf->SelName);
	SetCaption((TWins*)mf->edSelTxt, NULL);

	SetEnabled((TWins*)mf->buStop, True);
	Set_Focus((TWins*)mf->buStop);
	SetEnabled((TWins*)mf->buStart, False);
	TFindGrid_Clr(mf->grFind);
	AppProcessMessages();
	
	ini = get_appIni();

	pc = GetCaption_pc((TWins*)mf->cbName);
	if (pc && *pc && !((*pc == '*') && !pc[1])) {
	    mf->patt = TPattern_Create(0, pc, NULL);
	    if (mf->cbFileCaseSens->Checked) mf->patt->fu_case = FU_CASE_NONE;
	    TComboBox_Push(mf->cbName, MAX_PUSH_DATA);
	    TpcIni_push_stack(ini, APPNAME, "LastPatt", pc, MAX_PUSH_DATA, 0);
	}

	ppcchange(&mf->runPath, GetCaption_pc((TWins*)mf->cbPath));
	TComboBox_Push(mf->cbPath, MAX_PUSH_DATA);
	TpcIni_push_stack(ini, APPNAME, "LastPath", mf->runPath, 
				MAX_PUSH_DATA, pcIni_PushPath);

	pc = GetCaption_pc((TWins*)mf->cbFText);
	if (pc && *pc) {
	    TComboBox_Push(mf->cbFText, MAX_PUSH_DATA);
	    TpcIni_push_stack(ini, APPNAME, "LastTxt", pc, MAX_PUSH_DATA, 0);
	}
	free_appIni(ini);

	mf->cmpx = Tcmpx_Create(mf->cbFText->vds.s, 0);


// dword t = GetTime_ms();
	mf->ref_ms = appDat.app_ms;
	mf_findfile(mf, GetCaption_pc((TWins*)mf->cbPath));
// printf("%s(%d): %dms\n", FL, GetTime_ms() - t);

	ppTfree(&mf->cmpx);

	ppTfree(&mf->patt);
	SetEnabled((TWins*)mf->buStart, True);
	Set_Focus((TWins*)mf->buStart);
	SetEnabled((TWins*)mf->buStop, False);
}

static void mf_buStop_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormFind *mf = GetParentForm((TWins*)self);

	mf->Stop = True;
}


static void mf_buClose_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormFind *mf = GetParentForm((TWins*)self);
	mf->Stop = True;
	TForm_Close((TForm*)mf);
//	Close();
}

// ---------------------- find grid ev ------------------------
static void grFind_OnItemEnter(TFindGrid* self, TDirFileList *dfl) {
  TFormFind *mf = GetParentForm((TWins*)self);
  char *pc;
  TItemEnter ie;
// printf("%s(%d): %s\n", FL, dfl->fn);

	ie.art = ie_artFileEnter;
	ie.dfl = dfl;
	ie.name_node = dfl->fn;
	ie.stat = 2;

	pc = path_add_Slash(mf->runPath, Get_ExFileDat_pc(dfl->st, exfdt_user));
	ie.fn = pcadd(pc, dfl->fn);
	ie.fn_passwd = pcadd(pc, dfl->fn);
	xfree(pc);
	ie.gp = pctogp(ie.fn_passwd);

// printf("%s(%d): %s\n", FL, ie.fn);
	fw_ItemEnter((TWins*)self, &ie);
	gp_free(ie.gp);
	xfree(ie.fn_passwd);
	xfree(ie.fn);
}

static bool grFind_OnSelect(TFindGrid* self, int ACol, int ARow, int fu) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TDirFileList *dfl;
//  char *s, *path;

	if ((fu & grBefFu_Mask) == grBefFu_Select) {
	    dfl = TFindGrid_Get_dfl(self, ARow-1);
	    if (dfl) {
		ppcchange(&mf->SelPath, Get_ExFileDat_pc(dfl->st, exfdt_user));
		ppcchange(&mf->SelName, dfl->fn);
		SetCaption_edSelTxt(mf);
//		path = Get_ExFileDat_pc(dfl->st, exfdt_user);
//		if (path) {

//		} else {
//		    SetCaption((TWins*)mf->edSelTxt, dfl->fn);
//		}
//		s = pcadd_n(3, , "/", dfl->fn);
//		xfree(s);
	    }
	}
	return False;
}

// ---------------------- Save Setting ------------------------
static void aldfile_SaveSetting(TFormFind *self) {
  TpcIni *ini = get_appIni();

	TpcIni_WR_boolean(ini, APPNAME, "ModifiedVisible", ((self->miModifiedVis->fu & miChecked) != 0));
	TpcIni_WR_boolean(ini, APPNAME, "PermissonVisible", ((self->miPermissonVis->fu & miChecked) != 0));
	TpcIni_WR_boolean(ini, APPNAME, "TextCaseSens", self->cbTextCaseSens->Checked);
	TpcIni_WR_boolean(ini, APPNAME, "OnlyASCII", self->cbOnlyASCII->Checked);
	TpcIni_Save(ini);
	free_appIni(ini);
}

// ---------------------- Menu ------------------------
static void mf_mFile_SaveSize(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TpcIni *ini = TpcIni_Create(appDat.inifile_app);
  char buf[16];

	TpcIni_WR_int(ini, APPNAME, "Width", mf->width);
	TpcIni_WR_int(ini, APPNAME, "Height",mf->height);

	pcmove(buf, "GridW_0");
	TcrDat *cr = mf->grFind->ColDat;
	for (int i=0;i<fgaColLast;i++) {
	    buf[6] = 0x30 + i;
	    TpcIni_WR_int(ini, APPNAME, buf, cr->w);
	    cr++;
	}

	TpcIni_SaveAndFree(ini);
}

static void mf_mFile_Exit(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormFind *mf = GetParentForm((TWins*)self);
	mf->Stop = True;
	TForm_Close((TForm*)mf);
//	Close();
}

static void mf_mSetting_ModifiedVis(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormFind *mf = GetParentForm((TWins*)self);
	TFindGrid_SetColVisible(mf->grFind, fgaColModi, (smi->fu & miChecked) != 0);
	aldfile_SaveSetting(mf);
}

static void mf_mSetting_PermissonVis(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormFind *mf = GetParentForm((TWins*)self);
	TFindGrid_SetColVisible(mf->grFind, fgaColPerm, (smi->fu & miChecked) != 0);
	aldfile_SaveSetting(mf);
}

static void menuEdit_OnMenuOpen(TMenu *self) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TMenuItem *mi;
  TDirFileList *dfl;

	dfl = TFindGrid_GetSelect_dfl(mf->grFind);
	mi = mf->menuEdit->MenuItem;
	while (mi) {
	    if (dfl) {
		mi->fu &= ~miNotEnabled;
	    } else {
		mi->fu |= miNotEnabled;
	    }
	    mi = mi->next;
	}
}

static bool OpenWith_OnGetDskCmd(TMenuOpenWith *self, TDesktopCMD *dskcmd, int fu) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TDirFileList *dfl;

	dfl = TFindGrid_GetSelect_dfl(mf->grFind);
	if (!dfl) return True;
	char *pc = path_add_Slash(mf->runPath, Get_ExFileDat_pc(dfl->st, exfdt_user));

	dskcmd->gpList = xcalloc(sizeof(TgpList));
	pcsetgp(&dskcmd->gpList->gp, pc);
	xfree(pc);
	if (!(fu & MOWfu_OnlyDir)) gp_set_name(&dskcmd->gpList->gp, dfl->fn);
	return True;
}

static void menuOpenWithFind_OnClick(TMenuFile *self, long KeyState, TMenuItem *mi) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TDirFileList *dfl;
  TMimeDat md;
  TDesktopCMD dskcmd;

	dfl = TFindGrid_GetSelect_dfl(mf->grFind);
	if (!dfl) return;

	if (TFormExecWith_Show((TWins*)mf, dfl->fn, dfl->st, &md)) {

	    char *pc = path_add_Slash(mf->runPath, Get_ExFileDat_pc(dfl->st, exfdt_user));
	    dskcmd.gpList = xcalloc(sizeof(TgpList));
	    pcsetgp(&dskcmd.gpList->gp, pc);
	    xfree(pc);
	    gp_set_name(&dskcmd.gpList->gp, dfl->fn);
	    dskcmd.Exec = md.exec;
	    dskcmd.fu = 0;

	    get_desktop_dskcmd(&dskcmd);
	    if (dskcmd.ret_cmd) {
		exec_pc(dskcmd.ret_cmd);
		xfree(dskcmd.ret_cmd);
	    }
	    xfree(md.exec);
	    gpList_free(dskcmd.gpList);
	}
}

static void miCopy_OnClick(TMenuFile *self, long KeyState, TMenuItem* mi) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TDirFileList *dfl;
  char *pcsel;

	dfl = TFindGrid_GetSelect_dfl(mf->grFind);
	if (!dfl) return;

	pcsel = path_add_Slash(mf->runPath, Get_ExFileDat_pc(dfl->st, exfdt_user));
	ppcadd2(&pcsel, dfl->fn, "\n");

	miCopy(pcsel, NULL, mf->grFind->vds.UTF8on);

	xfree(pcsel);
}

#if 0
static void miProperties_OnClick(TMenuFile *self, long KeyState, TMenuItem* mi) {
  TFormFind *mf = GetParentForm((TWins*)self);
  TDirFileList *dfl, *dfl0;
  TglobPath *gp;

	dfl = TFindGrid_GetSelect_dfl(mf->grFind);
	if (!dfl) return;

	char *path = path_add_Slash(mf->runPath, Get_ExFileDat_pc(dfl->st, exfdt_user));
	gp = pctogp(path);
	xfree(path);
	dfl0 = dfl_dup(dfl);
	TFormProperty_Show((TWins*)mf, gp, dfl0, app_mime, False);
	gp_free(gp);
}
#endif
// ---------------------- Misc ------------------------

static void mHelp_Info(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormFind *mf = GetParentForm((TWins*)self);
  char buf[80];
	snprintf(buf, sizeof(buf), "ALDfind %s   (build %d  %s)", 
				VERSION, build_v, build_date);
	Copyright_MsgBox_Show((TWins*)mf, buf);
}

#ifndef STANDALONE_BUILDIN
bool TFormFind_ChkWorkDirOk(TglobPath *gp) {
	if (!gp_test_local_drive(gp)) return False;
	return !chkPathSkip(gp->path);
}

void TFormFind_SetWorkDir(TFormFind *self, char *workdir) {
	SetCaption((TWins*)self->cbPath, workdir);
}
#endif

void TFormFind_UpdataUTF8(TFormFind *self) {
	if (self) {
	    TWins_SetUTF8((TWins*)self->cbName, appXDat.FileSystemUTF8on);
	    TWins_SetUTF8((TWins*)self->cbPath, appXDat.FileSystemUTF8on);
	    TWins_SetUTF8((TWins*)self->grFind, appXDat.FileSystemUTF8on);
	}
}

static void cbPath_OnChange(TComboBox *self) {
  TFormFind *mf = GetParentForm((TWins*)self);

	if (!mf->buStop->Enabled) {
	    const char *pc = GetCaption_pc((TWins*)self);
	    bool en = pc && *pc;
	    if (en) {
		TPathInfo pathi;
		en = get_PathInfo(pc, &pathi) && 
			(pathi.fu & pathi_Local) &&
			!(pathi.fu & pathi_rel) &&
			TestIsPath(pc) &&
			!chkPathSkip(pc);
	    }
	    SetEnabled((TWins*)mf->buStart, en);
	}
}


static void cbTextCaseSens_OnClick(TChkBox* self, long KeyState, bool DblClick) {
  TFormFind *mf = GetParentForm((TWins*)self);
	aldfile_SaveSetting(mf);
}

static void cbOnlyASCII_OnClick(TChkBox* self, long KeyState, bool DblClick) {
  TFormFind *mf = GetParentForm((TWins*)self);

	TWins_SetUTF8((TWins*)mf->cbFText, !self->Checked);
	TWins_SetUTF8Auto((TWins*)mf->cbFText, !self->Checked);
	aldfile_SaveSetting(mf);
}

static void Callback_Path(TpcIni *self, const char *Data, void *owner) {
	TComboBox_AddItem(owner, Data, 0, NULL);
}

// ---------------------- Form ------------------------

const char zico16aldfind[] = {
 0x06,0x0C,0x10,0x00,0x10,0x00,0x00,0x00,0x95,0x00,0x04,0x21,0xD3,0x4C,0x1D,0x67,
 0x19,0x66,0x5A,0x8E,0xFC,0xA6,0x5D,0xD7,0x08,0x42,0x0C,0x63,0x10,0x84,0x1C,0xE7,
 0xFF,0xFF,0x00,0xB5,0x99,0x00,0xD9,0x2E,0x2A,0x24,0x42,0xE4,0xB2,0x4C,0xB3,0xBA,
 0x9A,0x00,0xB9,0xC9,0x18,0x3C,0xAB,0xBB,0x9A,0x90,0xCB,0x1C,0x81,0x17,0x11,0x89,
 0x0D,0xE4,0x6E,0x08,0xDE,0x8D,0x99,0xC4,0x2E,0x42,0xEE,0x06,0x1E,0x8F,0x19,0x59,
 0xEC,0x22,0xE4,0x86,0xE1,0xF5,0x98,0x51,0x59,0x2C,0x42,0x6E,0xDC,0xA1,0xF5,0x53,
 0x54,0x17,0x8B,0x90,0x1B,0x65,0x65,0x45,0x54,0x16,0x8B,0x90,0xCB,0x61,0x85,0x44,
 0x37,0x62,0xB1,0x8B,0x90,0xCB,0x21,0x66,0x55,0x26,0x31,0x8B,0x90,0xCB,0x1C,0x61,
 0x67,0x11,0x13,0x81,0x90,0x9B,0x8C,0xF1,0xAC,0x84,0x1C,0x42,0xBE,0xDB,0x44,0x48,
 0x40,0xD2,0xA2,0x44,0x08};

static void OnCanClose(TFormFind *self, int *CanClose) {
	self->Stop = True;
	*CanClose = 0;
	if (!self->buStop->Enabled) *CanClose = FormCloseAndFree;
}

/*_____________________________ Destroy _________________________________*/
static bool TFormFind__Destroy(TFormFind *self) {
	xfree(self->SelPath);
	xfree(self->SelName);
	xfree(self->runPath);
	return TForm__Destroy((TForm*)self);
}
/*_____________________________ Create __________________________________*/
TFormFind *TFormFind_Create(TWins *Parent, char *workdir) {
  TFormFind *self;
  TpcIni *ini;
  int m;

	self = TForm_CreatParamAlg(NULL, sizeof(TFormFind));
	self->destroy = (TDESTROY)TFormFind__Destroy;
	self->wtyp = wtyp_TFormFind;

#ifndef _OS_WIN32
	self->Icon = GetZIcon(zico16aldfind);
#endif
	self->OnCanClose = OnCanClose;

	ini = get_appIni();

	SetCaption((TWins*)self, "ALDfind Files/Data");

	self->min_width = 460;
	self->min_height = 240;
	self->max_width = 1280;
	self->max_height = 800;
	self->width =  TpcIni_RD_int_mm(ini, APPNAME, "Width", 620, self->min_width, self->max_width);
	self->height = TpcIni_RD_int_mm(ini, APPNAME, "Height",300, self->min_height,self->max_height);

	self->pnlTop =  TWins_CreateBA((TWins*)self, 0, 0, 10, 104, mBorder|mBorderL, alTop);
//	self->pnlTopL = TWins_CreateBA((TWins*)self->pnlTop, 0, 0, 10, 150, 0, alClient);
//	self->pnlTop2 = TWins_CreateBA((TWins*)self->pnlTop, 0, 0, 70, 150, mBorder, alRight);
	self->pnlTop2 = TWins_CreateBA((TWins*)self, 0, 0, 70, 24, mBorder|mBorderL, alTop);
	self->pnlTop2->OnResize = pnlTop2_OnResize;

	self->pgc = TPageControl_Create((TWins*)self->pnlTop);
	self->pgc->Align = alClient;
	self->pgc->height = 80;
	self->pgc->MinTabWidth = 60;
	self->pgc->TabTxtOff = 12;
	Set_Parent((TWins*)self->pgc, (TWins*)self->pnlTop);
	self->pgc->OnResize = pgc_OnResize;

// -------- Page Files ------
	self->pgFiles = TPageControl_AddNewPage(self->pgc, "Files");
	self->pgFiles->OnResize = pgFiles_OnResize;
	  TWins_CreateLabel((TWins*)self->pgFiles, 2, 2, 50, 18, "Name:", mCaptionLeft, 0);
	  self->cbName = TComboBox_CreateS((TWins*)self->pgFiles, 54, 1, 160, 20, "*");
	  TWins_CreateLabel((TWins*)self->pgFiles, 2, 24, 50, 18, "Path:", mCaptionLeft, 0);
	  self->cbPath = TComboBox_CreateS((TWins*)self->pgFiles, 54, 23, 160, 20, NULL);
	  self->cbPath->OnChange = cbPath_OnChange;
	  self->buPath = TButton_CreateS((TWins*)self->pgFiles, 220, 23, 32, 20, "...");
	  self->buPath->OnClick =  buPath_OnClick;
	  self->cbSubDir = TChkBox_CreateS((TWins*)self->pgFiles, 54, 44, 160, 18, "Include subdirs");
	  TChkBox_SetChecked(self->cbSubDir, True);
	  self->cbFileCaseSens = TChkBox_CreateS((TWins*)self->pgFiles, 54, 62, 160, 18, "Filename case sensitive");
	  TChkBox_SetChecked(self->cbFileCaseSens, True);

// -------- Page Data ------
	self->pgData  = TPageControl_AddNewPage(self->pgc, "Data");
	self->pgData->OnResize = pgData_OnResize;
	  TWins_CreateLabel((TWins*)self->pgData, 2, 2, 60, 18, "Find text:", mCaptionLeft, 0);
	  self->cbFText = TComboBox_CreateS((TWins*)self->pgData, 64, 1, 160, 20, NULL);
	  self->cbTextCaseSens = TChkBox_CreateS((TWins*)self->pgData, 64, 26, 140, 18, "Text case sensitive");
	  self->cbTextCaseSens->Checked = TpcIni_RD_boolean(ini, APPNAME, "TextCaseSens", False);
	  self->cbTextCaseSens->OnClick = cbTextCaseSens_OnClick;
	  self->cbInclBinFile  = TChkBox_CreateS((TWins*)self->pgData, 64, 44, 140, 18, "Include binary files");
	  self->cbOnlyASCII    = TChkBox_CreateS((TWins*)self->pgData, 64, 62, 140, 18, "Only ASCII");
	  self->cbOnlyASCII->Checked = TpcIni_RD_boolean(ini, APPNAME, "OnlyASCII", False);
	  self->cbOnlyASCII->OnClick = cbOnlyASCII_OnClick;
	  TWins_SetUTF8((TWins*)self->cbFText, !self->cbOnlyASCII->Checked);


// -------- Start/Stop ------
	self->cbSelPlusPart = TChkBox_CreateS((TWins*)self->pnlTop2, 2, 3, 55, 18, "+Path");
	self->cbSelPlusPart->OnClick = cbSelPlusPart_OnClick;

	self->edSelTxt = TEdit_CreateS(self->pnlTop2, 58, 2, 200, 20, NULL);
	self->edSelTxt->ReadOnly = True;

	self->buStart = TButton_CreateS(self->pnlTop2, 1,  1, 66, 22, "Start");
	self->buStart->OnClick = mf_buStart_OnClick;
	SetColor((TWins*)self->buStart, 0x80e080);

	self->buStop  = TButton_CreateS(self->pnlTop2, 74, 1, 66, 22, "Stop");
	self->buStop->OnClick = mf_buStop_OnClick;
	SetEnabled((TWins*)self->buStop, False);
	SetColor((TWins*)self->buStop, 0xe08080);

	self->buClose = TButton_CreateS(self->pnlTop2, 160, 1, 66, 22, "Close");
	self->buClose->OnClick = mf_buClose_OnClick;

// -------- Find Grid ------
	self->grFind = TFindGrid_Create((TWins*)self);
	self->grFind->Align = alClient;
	self->grFind->pw->Align = alClient;
	Set_Parent((TWins*)self->grFind, (TWins*)self);
	
	self->grFind->OnItemEnter = grFind_OnItemEnter;
	self->grFind->OnSelect = grFind_OnSelect;

// -------- Menu File ------
  TMenuItem *mi;

	self->mbFile = TMenuButton_CreateS((TWins*)self->pgc, (TMenuButton*)self, "File");
	self->mbFile->top = 0;
//	self->mbFile->left = 200;
	self->menuFile = TMenu_Create((TWins*)self);
	self->mbFile->menu = self->menuFile;
	TMenu_AddItem(self->menuFile, "Save Size", 0, mf_mFile_SaveSize);
	TMenu_AddItem(self->menuFile, "-", 0, NULL);
	TMenu_AddItem(self->menuFile, "Exit", 0, mf_mFile_Exit);

// -------- Menu Setting ------
	self->mbSetting = TMenuButton_CreateS((TWins*)self->pgc, self->mbFile, "Setting");
	self->menuSetting = TMenu_Create((TWins*)self);
	self->mbSetting->menu = self->menuSetting;

	m = miCanCheck | miAutoToggle;
	if (TpcIni_RD_boolean(ini, APPNAME, "ModifiedVisible", True)) m = miCanCheck | miChecked | miAutoToggle;
	self->miModifiedVis = TMenu_AddItem(self->menuSetting, 
				"Modified visible", m, mf_mSetting_ModifiedVis);
	TFindGrid_SetColVisible(self->grFind, fgaColModi, (m & miChecked) != 0);

	m = miCanCheck | miAutoToggle;
	if (TpcIni_RD_boolean(ini, APPNAME, "PermissonVisible", True)) m = miCanCheck | miChecked | miAutoToggle;
	self->miPermissonVis = TMenu_AddItem(self->menuSetting, 
				"Permisson visible", m, mf_mSetting_PermissonVis);
	TFindGrid_SetColVisible(self->grFind, fgaColPerm, (m & miChecked) != 0);

// -------- Menu Help ------
	self->mbHelp = TMenuButton_CreateS((TWins*)self->pgc, self->mbSetting, "Help");
	self->menuHelp = TMenu_Create((TWins*)self);
	self->mbHelp->menu = self->menuHelp;
	mi = TMenu_AddItem(self->menuHelp, "Info", 0, mHelp_Info);
	mi->pic = get_pic16_info(False);

// -------- Menu open with ------
	self->menuEdit = TMenu_Create((TWins*)self);
	self->menuEdit->OnMenuOpen = menuEdit_OnMenuOpen;
	self->miOpenWith = TMenu_AddItem(self->menuEdit, "Open With", 0, NULL);
	  self->menuOpenWith = TMenuOpenWith_Create((TWins*)self);
	  self->miOpenWith->next_menu = (TMenu*)self->menuOpenWith;
	  self->menuOpenWith->OnGetDskCmd = OpenWith_OnGetDskCmd;
	  self->menuOpenWith->path = pcadd(appDat.ALDini_dir, "/Menu/OpenWith/");

	TMenu_AddItem(self->menuEdit, "Win Exec With", 0, 
		(TMenuClickEv)menuOpenWithFind_OnClick);
	TMenu_AddItem(self->menuEdit, "-", 0, NULL);
	TMenu_AddItem(self->menuEdit, "Copy", 0, (TMenuClickEv)miCopy_OnClick);
#if 0
	TMenu_AddItem(self->menuEdit, "-", 0, NULL);
	TMenu_AddItem(self->menuEdit, "Properties", 0, (TMenuClickEv)miProperties_OnClick);
#endif
	self->grFind->PopupMenu = self->menuEdit;

	TpcIni_get_stack(ini, APPNAME, "LastPatt", Callback_Path, self->cbName, 
			MAX_PUSH_DATA, 0);
	TpcIni_get_stack(ini, APPNAME, "LastPath", Callback_Path, self->cbPath, 
			MAX_PUSH_DATA, pcIni_PushPath);
	TpcIni_get_stack(ini, APPNAME, "LastTxt", Callback_Path, self->cbFText, 
			MAX_PUSH_DATA, 0);

	SetCaption((TWins*)self->cbPath, workdir);

	free_appIni(ini);

	TFormFind_UpdataUTF8(self);
	self->FindCountMax = 1000;
	return self;
}

static const char *Help_Msg = {
  "\n"
  "Usage: edit\n"
  "\n"
  "  --workdir <dir>      Change to \"dir\"\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: "APPNAME" version "VERSION"  (build %d  %s)\n"
};

#ifdef STANDALONE_BUILDIN
int main (int argc, char **argv) {
	cmdline_slash_korr(argc, argv);
#else
int bi_Find_main (int argc, char **argv) {
#endif

  int i;
  char *val, *workdir;
#ifdef MEM_CHK
  TFormFind *self;
#endif
  bool help = False;
  bool err = False;
//  char buf[16];

#ifdef STANDALONE_BUILDIN
#ifdef _OS_WIN32
	cmdUTF8korr(&argc, &argv);
#endif
#endif

	app_name = APPNAME;
	workdir = NULL;
	i = 1;
	while (!err && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'h':
		    help = True;
		    break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
		    } else if (pcncmp(val, "workdir", 7)) {
			err = cmd_get_pc(argc, argv, val+6, &i, &workdir);
#ifdef Test_Root_WinId
		    } else if (pcncmp(val, "TestRootWinID", 13)) {
			i++;
			val = argv[i];
			if (val && *val != '-') {
				TestRootWinId = pctoint(val);
			} else err = True;
#endif
		    } else err = True;
		    break;
		default: 
		    err = True;
		}
	    } else {
		if (!workdir) {
		    workdir = val;
		} else {
		    err = True;
		}
	    }
	    i++;
	}

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(err);
	}
	if (workdir && test_local_drive(workdir)) {
	    workdir = path_to_local(workdir);
	    while (workdir && !TestIsPath(workdir)) {
		path_sub_ppc(&workdir);
	    }
	}
	if (!workdir) workdir = xgetcwd(False);

	initXapp(0, argc, argv);
	hGetFileHeadInfo = &hGetFileHeadInfo_buildin;

	hGetDeskIcon = &hGetDeskIcon_file;
//	ThImageRD_Reg(6, &hImageRD_xpm, &hImageRD_png, &hImageRD_jpeg, 
//			 &hImageRD_gif, &hImageRD_bmp, &hImageRD_pnm);
	ThImageRD_Reg(2, &hImageRD_xpm, &hImageRD_png);


#ifdef _OS_WIN32
	val = NT4_GetLongPathName(workdir);
	if (val) {xfree(workdir); workdir = val;}

	appDat.inifile_app = pcadd(appDat.ALDini_dir, "\\aldfile.ini");
	ThDirList_Reg(1, &hdirlist_ld_File);
#else
	appDat.inifile_app = pcadd(appDat.ALDini_dir, "/aldfile.ini");
	ThDirList_Reg(1, &hdirlist_ld_File);
#endif
	appDat.appIni = TpcIni_Create(appDat.inifile_app);
#ifdef USE_DECOR
	appXDat.DefaultButtonDecor = TDecor_button_w3b3_Create();
#endif

#ifdef MEM_CHK
	self = TFormFind_Create(NULL, workdir);
#else
	TFormFind_Create(NULL, workdir);
#endif
//	appDat.obj = self;

//#ifndef _OS_WIN32
	app_mime = TMime_X_Create();
//	app_mime->OnChange = app_mime_OnChange;
//#endif
	ppTfree(&appDat.appIni);

	AppRun();

#ifdef MEM_CHK
  Tfree(app_mime);
  Tfree(self);
  AppHandleMessage();
  int lMCt; 
  dword lMsize;
  MEM_CHK_Count(&lMCt, &lMsize);
  printf("Debug %s(%d): MEM_CHK_Count=%3d  size=%d wins=%d\n", FL, lMCt, lMsize, GetTWinsCount());
#endif

	return 0;
}
