/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "cmpx.h"

#define USE_16BE 0

typedef struct {
  char a;
  char b;
} cmpx_ascii;

typedef struct {
  byte ac;	//a char count
  byte bc;	//b char count
  char a[3];
  char b[3];
} cmpx_utf8;

typedef struct {
  union {
    char a[2];
    word wa;
  };
  union {
    char b[2];
    word wb;
  };
} cmpx_u16;

typedef struct Tcmpx {
  AlgTTC0
  int count;
  int max_char;
  int fkt;

  char ascii_a;
  char ascii_b;
  char utf8_a;
  char utf8_b;
  union {
    char u16_a[2];
    word u16_wa;
  };
  union {
    char u16_b[2];
    word u16_wb;
  };

  cmpx_ascii *ascii;
  cmpx_utf8  *utf8;
  cmpx_u16   *u16;

} Tcmpx;

bool REGP2 cmpx_c0_case(Tcmpx *self, const char *a) {
	return ((self->fkt & fcmp_ascii) && ((self->ascii_a == *a) || 
					     (self->ascii_b == *a))) ||
	       ((self->fkt & fcmp_utf8)  && ((self->utf8_a == *a) ||
					     (self->utf8_b == *a))) ||
#if USE_16BE
	       ((self->fkt & fcmp_u16)   && ((self->u16_a[0] == *a) ||
					     (self->u16_b[0] == *a))) ||
	       ((self->fkt & fcmp_u16BE) && ((self->u16_a[1] == *a) ||
					     (self->u16_b[1] == *a))) ||
#else
	       ((self->fkt & fcmp_u16)   && ((self->u16_a[0] == *a) ||
					     (self->u16_b[0] == *a)));
#endif
}

int cmpx_cmp_case(Tcmpx *self, const char *a, int max) {

	if (self->fkt & fcmp_ascii) {
	    cmpx_ascii *p = self->ascii;
	    int lg = max;
	    int i = self->count;
	    const char *pc = a;
	    while (i && lg && ((*pc == p->a)||(*pc == p->b))) {
		pc++; p++; i--; lg--;
	    }
	    if (!i) return fcmp_ascii;
	}

	if (self->fkt & fcmp_utf8) {
	    cmpx_utf8 *p = self->utf8;
	    int lg = max;
	    const char *pc = a;
	    int i = self->count;

	    while (i) {
		if (p->ac == 1) {
		    if ((lg < 1) || ((pc[0] != p->a[0])&&(pc[0] != p->b[0]))) break;
		    pc++; lg--;
		} else if (p->ac == 2) {
		    if ((lg < 2) || ((pc[0] != p->a[0])&&(pc[0] != p->b[0])) || 
				    ((pc[1] != p->a[1])&&(pc[1] != p->b[1]))) break;
		    pc += 2; lg -= 2;
		} else if (p->ac == 3) {
		    if ((lg < 3) || ((pc[0] != p->a[0])&&(pc[0] != p->b[0])) || 
				    ((pc[1] != p->a[1])&&(pc[1] != p->b[1])) ||
				    ((pc[2] != p->a[2])&&(pc[2] != p->b[2]))) break;
		    pc += 3; lg -= 3;
		} else {
		    break;
		}
		p++; i--;
	    }
	    if (!i) return fcmp_utf8;
	}

	if (self->fkt & fcmp_u16) {
	    cmpx_u16 *p = self->u16;
	    int lg = max;
	    const char *pc = a;
	    int i = self->count;
	    while (i && (lg > 1) && ((pc[0] == p->a[0])||(pc[0] == p->b[0])) && 
			 	    ((pc[1] == p->a[1])||(pc[1] == p->b[1]))) {
		pc += 2; p++; i--; lg -= 2;
	    }
	    if (!i) return fcmp_u16;
	}
	return 0;
}

bool REGP2 cmpx_c0(Tcmpx *self, const char *a) {
	return ((self->fkt & fcmp_ascii) && (self->ascii->a == *a)) ||
	       ((self->fkt & fcmp_utf8)  && (self->utf8->a[0] == *a)) ||
#if USE_16BE
	       ((self->fkt & fcmp_u16)   && (self->u16->a[0] == *a)) ||
	       ((self->fkt & fcmp_u16BE) && (self->u16->a[1] == *a));
#else
	       ((self->fkt & fcmp_u16)   && (self->u16->a[0] == *a));
#endif
}

int cmpx_cmp(Tcmpx *self, const char *a, int max) {

//	if (!self || !a || (self->count <= 0)) return 0;
//	if (!f) f = fcmp_ascii|fcmp_utf8|fcmp_u16|fcmp_u16BE;

	if (self->fkt & fcmp_ascii) {
	    cmpx_ascii *p = self->ascii;
	    int lg = max;
	    int i = self->count;
	    const char *pc = a;
	    while (i && lg && (*pc == p->a)) {
		pc++; p++; i--; lg--;
	    }
	    if (!i) return fcmp_ascii;
	}

	if (self->fkt & fcmp_utf8) {
	    cmpx_utf8 *p = self->utf8;
	    int lg = max;
	    const char *pc = a;
	    int i = self->count;

	    while (i) {
		if (p->ac == 1) {
		    if ((lg < 1) || (pc[0] != p->a[0])) break;
		    pc++; lg--;
		} else if (p->ac == 2) {
		    if ((lg < 2) || (pc[0] != p->a[0]) || 
				    (pc[1] != p->a[1])) break;
		    pc += 2; lg -= 2;
		} else {
		    if ((lg < 3) || (pc[0] != p->a[0]) || 
				    (pc[1] != p->a[1]) ||
				    (pc[2] != p->a[2])) break;
		    pc += 3; lg -= 3;
		}
		p++; i--;
	    }
	    if (!i) return fcmp_utf8;
	}

	if (self->fkt & fcmp_u16) {
	    cmpx_u16 *p = self->u16;
	    int lg = max;
	    const char *pc = a;
	    int i = self->count;
	    while (i && lg && (pc[0] == p->a[0]) && 
			      (pc[1] == p->a[1])) {
		pc += 2; p++; i--; lg -= 2;
	    }
	    if (!i) return fcmp_u16;
	}
	return 0;
}

int cmpx_GetCount(Tcmpx *self) {
	if (!self) return 0;
	return self->count;
}

int cmpx_GetMaxChars(Tcmpx *self) {
	if (!self) return 0;
	return self->max_char;
}

/*____________________________ Destroy __________________________________*/
static bool Tcmpx__Destroy(Tcmpx *self) {
	xfree(self->ascii);
	xfree(self->utf8);
	xfree(self->u16);
	return False;
}

/*_____________________________ Create __________________________________*/
Tcmpx *Tcmpx_Create(const char *utf8, int f) {
  Tcmpx *self;
  int pcl;
  word *pwc0, *pwc;
  dword wc;

	self = ttc_create(sizeof(Tcmpx), wtyp_SysAlg);
	self->destroy = (TDESTROY)Tcmpx__Destroy;
	pcl = pclen(utf8);
	self->count = Chk_fio_UTF8(utf8, pcl, NULL);
	if (self->count > 0) {
	    if (!f) f = fcmp_ascii|fcmp_utf8|fcmp_u16|fcmp_u16BE;
	    self->fkt = f;
	    self->ascii = xcalloc(sizeof(cmpx_ascii) * self->count);
	    self->utf8  = xcalloc(sizeof(cmpx_utf8)  * self->count);
	    self->u16   = xcalloc(sizeof(cmpx_u16)   * self->count);

	    pwc0 = xmalloc(self->count*2);
	    Utf8ToWChars(pwc0, self->count, utf8, pcl, 0);
	    pwc = pwc0;
	    for (int i=0;i<self->count;i++) {
		wc = *pwc++;
		self->ascii[i].a = wc;
		self->u16[i].wa   = wc;
		self->utf8[i].ac = WCharsToUtf8(self->utf8[i].a, 3, (void*)&wc, 1, 0);

		if (wc < 256) {
		    if ((byte)l_iso8859_1[wc] == wc) {
			wc = (byte)u_iso8859_1[wc];
		    } else {
			wc = (byte)l_iso8859_1[wc];
		    }
		} else {
		    if (wc < 512) wc ^= 0x0001;
		    self->fkt &= ~fcmp_ascii;
		}
		self->ascii[i].b = wc;
		self->u16[i].wb   = wc;
		self->utf8[i].bc = WCharsToUtf8(self->utf8[i].b, 3, (void*)&wc, 1, 0);

		if (self->utf8[i].ac >= self->utf8[i].bc) {
		    self->max_char += self->utf8[i].ac;
		} else {
		    self->max_char += self->utf8[i].bc;
		}
	    }

	    if (self->max_char < (self->count * 2)) self->max_char = self->count * 2;
	    xfree(pwc0);
	    self->ascii_a = self->ascii->a;
	    self->ascii_b = self->ascii->b;
	    self->utf8_a = self->utf8->a[0];
	    self->utf8_b = self->utf8->b[0];
	    self->u16_wa = self->u16->wa;
	    self->u16_wb = self->u16->wb;

	}

	return self;
}
