/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"


void OpenWith_Exec(const char *path, const char *filename, void *obj, 
		   TOnGetDskCmd OnGetDskCmd) {
  int MOWfu;
  char *fn, *exec_path, *token, *appFlag;
  const char *cpc, *indata;
  TpcIni *ini, *appIni;
  TDesktopCMD dskcmd;
  TDesktopHead ddkh;
  bool ok;

	if (!OnGetDskCmd) return;
	Fill0(&dskcmd, sizeof(dskcmd));
	appFlag = NULL;
	fn = path_add(path, filename);
	ini = TpcIni_Create(fn);
	ok = dotdesktop_get_head(ini, &ddkh) && 
	     (ddkh.ddkt == ddkt_Application) && 
	     (ddkh.exec || ddkh.tryexec);

	if (ok) {
	    cpc = ddkh.exec;
	    if (!cpc) cpc = ddkh.tryexec;
	    dskcmd.Exec = pcdup(cpc);
	    GetToken(GetFileName_pc(dskcmd.Exec), &token);
	    appIni = get_appIni();
	    appFlag = pcdup(TpcIni_RD(appIni, "AppExecFlag", token));
	    free_appIni(appIni);
	    xfree(token);
	}

	Tfree(ini);
	xfree(fn);
	if (!ok) return;

	MOWfu = 0;
	indata = appFlag;
	while (indata) {
	    indata = GetToken(indata, &token);
	    if (!pccasecmpASCII(token, "chdir")) {
		MOWfu |= MOWfu_ChDir;

	    } else if (!pccasecmpASCII(token, "onlyLocal")) {
		MOWfu |= MOWfu_QnlyLocal;

	    } else if (!pccasecmpASCII(token, "onlydir")) {
		MOWfu |= MOWfu_OnlyDir;
	    }
	    xfree(token);
	}
	xfree(appFlag);

	if (!OnGetDskCmd(obj, &dskcmd, MOWfu)) return;

	exec_path = NULL;
	if (dskcmd.gpList) {
	    if (MOWfu & MOWfu_ChDir) exec_path = gptopc(&dskcmd.gpList->gp, gptfu_noName);
	    ok = !(MOWfu & MOWfu_QnlyLocal) || 
		 !gp_test_local_drive(&dskcmd.gpList->gp);
	}

// printf("%s(%d): %d\n", FL, ok);
	if (ok) {
	    get_desktop_dskcmd(&dskcmd);
// printf("%s(%d): %s\n", FL, dskcmd.ret_cmd);
	    if (dskcmd.ret_cmd) exec_pc_co_path(dskcmd.ret_cmd, NULL, exec_path);
	}

	xfree(exec_path);
	xfree(dskcmd.Exec);
	xfree(dskcmd.ret_cmd);
	gpList_free(dskcmd.gpList);
}


static void menuOpenWith_OnClick(TMenuOpenWith *self, long KeyState, bool DblClick) {
  TMenuFileItem *mfi;

	mfi = (TMenuFileItem*)self->SelItem;
	if (!mfi || !mfi->filename) return;
	OpenWith_Exec(self->path, mfi->filename, self,
		      (TOnGetDskCmd)self->OnGetDskCmd);
}

/* ______________________ Create _____________________________________ */
TMenuOpenWith *TMenuOpenWith_Create(TWins *parent) {
  TWins_Create_struct wcs;
  TMenuOpenWith *self;

	TWins_Create_struct_init(&wcs, sizeof(TMenuOpenWith), wtyp_MenuFile);
	wcs.Parent = parent;
	wcs.CreateParams = TMenuFile_CreateParams;
	TWins_DefaultConf(&wcs);

	self = wcs.wins;
	self->OnClick = menuOpenWith_OnClick;
	self->mfk = mfk_AutoFree;
	return self;
}
