/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldhex.h"
#include "tpixmap.h"
#include "cicons.h"

#if USE_SB2_COUNT
static void mSetTag(TMenu *self) {
  TMenuItem *mi;
  int tag;

	tag = self->tag;
	mi = self->MenuItem;
	while (mi) {
	    mi->fu &= ~miChecked;
	    if (mi->tag == tag) mi->fu |= miChecked;
	    mi = mi->next;
	}
}

void miSBx_Sync(TFormMain *self) {
  int tag, tagc;

	if (!appDat.AppRun) return;
	mSetTag(self->menuSB1max);
	mSetTag(self->menuSB2max);
	mSetTag(self->menuSB2count);

	THexEdit_SetMaxSegData(self->HexEdit, 1 << self->menuSB1max->tag);

	tag = self->menuSB2max->tag;
	if ((tag < 22) || (tag > 28)) tag = 24;
	self->VSB2SegShift = tag;

	tagc = self->menuSB2count->tag;
	if ((tagc < 8) || (tagc > 12)) tagc = 10;
	self->VSB2SegShift = tagc;
// printf("%s(%d): %d %d\n", FL, tag, tagc);

	int64_t max = (THexEdit_GetSizeAbs(self->HexEdit) + ((1 << tagc) - 1)) >> tagc;
	int64_t mxsft = 1 << (tag - tagc);
	if (max > mxsft) max = mxsft;

	int64_t seg = THexEdit_GetMaxSegData(self->HexEdit) >> tagc;
// printf("%s(%d): 0x%08llx 0x%08llx 0x%08llx\n", FL, max, seg, THexEdit_GetSizeAbs(self->HexEdit));
	TScrollBar_SetMaxCount_ButtonSize(self->VSB2, max, seg);

}

static void miSBx_SetTag(TMenu *self, long KeyState, TMenuItem *smi) {
	self->tag = smi->tag;
	miSBx_Sync((TFormMain*)self->FormStruct);
}
#endif /* USE_SB2_COUNT */

/* ______________________ Create _____________________________________ */
void MenuSetting_Create(TFormMain *self) {

	self->mbSetting = TMenuButton_Create(self->PanelTop, self->mbFile);
	SetCaption((TWins*)self->mbSetting, "Settings");
	Set_Parent((TWins*)self->mbSetting, self->PanelTop);

	self->menuSetting = TMenu_Create((TWins*)self);
	self->mbSetting->menu = self->menuSetting;

#if USE_SB2_COUNT
  TMenuItem *mi;
	mi = TMenu_AddItem(self->menuSetting, "SB1 max mem", 0, NULL);
	  self->menuSB1max = TMenu_Create((TWins*)self);
	  self->menuSB1max->tag = 18;
	  mi->next_menu = self->menuSB1max;
	  TMenu_AddItem(self->menuSB1max,  "32k", miCanCheck, miSBx_SetTag)->tag = 15;
	  TMenu_AddItem(self->menuSB1max,  "64k", miCanCheck, miSBx_SetTag)->tag = 16;
	  TMenu_AddItem(self->menuSB1max, "128k", miCanCheck, miSBx_SetTag)->tag = 17;
	  TMenu_AddItem(self->menuSB1max, "256k", miCanCheck, miSBx_SetTag)->tag = 18;
	  TMenu_AddItem(self->menuSB1max, "512k", miCanCheck, miSBx_SetTag)->tag = 19;
	  TMenu_AddItem(self->menuSB1max,"1024k", miCanCheck, miSBx_SetTag)->tag = 20;


	mi = TMenu_AddItem(self->menuSetting, "SB2 max mem", 0, NULL);
	  self->menuSB2max = TMenu_Create((TWins*)self);
	  self->menuSB2max->tag = 24;
	  mi->next_menu = self->menuSB2max;
	  TMenu_AddItem(self->menuSB2max,  "4M",  miCanCheck, miSBx_SetTag)->tag = 22;
	  TMenu_AddItem(self->menuSB2max,  "8M",  miCanCheck, miSBx_SetTag)->tag = 23;
	  TMenu_AddItem(self->menuSB2max, "16M",  miCanCheck, miSBx_SetTag)->tag = 24;
	  TMenu_AddItem(self->menuSB2max, "32M",  miCanCheck, miSBx_SetTag)->tag = 25;
	  TMenu_AddItem(self->menuSB2max, "64M",  miCanCheck, miSBx_SetTag)->tag = 26;


	mi = TMenu_AddItem(self->menuSetting, "SB2 count", 0, NULL);
	  self->menuSB2count = TMenu_Create((TWins*)self);
	  self->menuSB2count->tag = 10;
	  mi->next_menu = self->menuSB2count;
	  TMenu_AddItem(self->menuSB2count, "256",  miCanCheck, miSBx_SetTag)->tag = 8;
	  TMenu_AddItem(self->menuSB2count, "512",  miCanCheck, miSBx_SetTag)->tag = 9;
	  TMenu_AddItem(self->menuSB2count,  "1k",  miCanCheck, miSBx_SetTag)->tag = 10;
	  TMenu_AddItem(self->menuSB2count,  "2k",  miCanCheck, miSBx_SetTag)->tag = 11;
	  TMenu_AddItem(self->menuSB2count,  "4k",  miCanCheck, miSBx_SetTag)->tag = 12;

	miSBx_Sync(self);
#endif /* USE_SB2_COUNT */

}
