/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "thexedit.h"
#include <fcntl.h>



/*_______________________________________________________________________*/
#define GetLineFontHeight(a) TLineEditBase_GetLineFontHeight((TLineEditBase*)a)

static int get_max_lines(THexEdit *self) {
  int64_t max = FIO_GetSize(self->fio);
  
	if (max > self->MaxSegData) max = self->MaxSegData;
	return (max + self->MaxLineByts - 1) / self->MaxLineByts;
}

static dword GetCharOffI(THexEdit *self, int idx, int idy, bool Insert) {
  dword result;
  int OffX;

	OffX = self->CaretCharIdxToOffX(self, idx, idy, 0, Insert);
	result = (idy * self->MaxLineByts) + OffX;
	if (Insert) {
	    if (result > FIO_GetSize(self->fio)) result = FIO_GetSize(self->fio);
	} else {
	    if (result >= FIO_GetSize(self->fio)) result = FIO_GetSize(self->fio) - 1;
	}
	return result;
}

dword GetCharOff(THexEdit *self, int idx, int idy) {
	return GetCharOffI(self, idx, idy, self->Insert);
}

static bool chk_xy_ok(THexEdit *self, int idx, int idy, bool Insert) {
	if (Insert) {
	    return ((idy * self->MaxLineByts) + idx) <= FIO_GetSize(self->fio);
	} else {
	    return ((idy * self->MaxLineByts) + idx) < FIO_GetSize(self->fio);
	}
}


#define Res_vds(a) TLineEditBase_Res_vds((TLineEditBase*)a)
#define Set_vds(a, b) TLineEditBase_Set_vds((TLineEditBase*)a, b)

static void GetCaretIdxy(THexEdit *self, int x, int y, int *idx, int *idy) {
  TColS *cs;
  int iy;

	if (y < 0) y = 0;
	iy = (y / GetLineFontHeight(self));
	if (iy >= self->VisibleLines) iy = self->VisibleLines - 1;

	cs = TLines_Get_pcs(self->tls, iy);
	while (!(cs->fu & csfu_HasData) && (iy > 0)) {
	    iy--;
	    cs = TLines_Get_pcs(self->tls, iy);
	}
	*idy = iy + self->VSB->Count;
	Set_vds(self, iy);
	*idx = GetMouseX_CharPos(x, &self->vds);
	Res_vds(self);
}

static void fill_lines(THexEdit *self) {
  TPLine pl;

	if (!self->PrintLine) return;

	pl.Insert = self->Insert;
	pl.ASCII_Korr = True;
	pl.idx = TScrollBar_GetCount(self->VSB) * self->MaxLineByts;
	pl.lg = 0;
	pl.pb = (byte*)FIO_GetDataP(self->fio, pl.idx, self->MaxLineByts * self->VisibleLines);
	if (pl.pb && (FIO_GetSize(self->fio) > pl.idx)) {
	    pl.lg = FIO_GetSize(self->fio) - pl.idx;
	}
	
	for (int i=0; i < self->VisibleLines; i++) {
	    pl.cs = TLines_Get_pcs(self->tls, i);
	    self->PrintLine(self, &pl);
	}
}

static void set_SB(TScrollBar *SB, int max, int lew) {
	if (SB->Visible && ((SB->MaxCount != max) || 
			    (SB->ButtonCountSize != lew))) {
	    TScrollBar_SetMaxCount_ButtonSize(SB, max, lew);
	    SB->LargeChange = lew - 1;
	}
}

static void Adjust_VHSB(THexEdit *self) {
  TVDrawString *vds = &self->vds;
  int VisibleLines;

	vds->s_const = "u";
	vds->lg = 1;
	self->CharWidth = TextWidthColorString(vds);
	vds->s = NULL;
	vds->lg = 0;

	if (self->CharWidth < 1) self->CharWidth = 1;
	self->HSB->SmallChange = self->CharWidth;

	VisibleLines = self->VisibleLines;
	self->VisibleLines = (self->height / GetLineFontHeight(self)) + 1;
	if (self->VisibleLines < 1) self->VisibleLines = 1;
	if (self->VisibleLines > MAX_LINES) self->VisibleLines = MAX_LINES;

	self->VisibleAktLines = self->VisibleLines;
	if ((self->VisibleLines * GetLineFontHeight(self)) > self->height) {
	    self->VisibleAktLines--;
	}

	set_SB(self->VSB, get_max_lines(self)+1, self->VisibleLines-1);

	if (VisibleLines != self->VisibleLines) {
	    fill_lines(self);
	}
}

#define THexEdit_ChkIsSel(a) TLineEditBase_ChkIsSel((TLineEditBase*)a)

void THexEdit_DoChange(THexEdit *self, int bef, TKeyS *KeyS, Tlgs *Ins_lgs) {
  int m, xa, ya, xe, ye;
  dword oa, oe;
  TKeyS ks, *tmpKeyS;
  Tlgs Del_lgs;

	xa = self->SelAx;
	ya = self->SelAy;
	xe = self->CaretCharIdx;
	ye = self->CaretCharIdy;
	if ((bef & ur_DelSel) && !THexEdit_ChkIsSel(self)) bef &= ~ur_DelSel;

	if ((bef & ur_DelSel) && !(bef & ur_Undo)) {
	    oa = GetCharOffI(self, xa, ya, True);
	    oe = GetCharOffI(self, xe, ye, True);
	    if (oa > oe) {dword ox = oa; oa = oe; oe = ox;} 
// printf("%s(%d): %u %u\n", FL, oa, oe);
	    if (oa < oe) {
		tmpKeyS = KeyS;
	        Del_lgs.s = NULL;
	        Del_lgs.lg = 0;
		dword lg = oe - oa;
		if (self->OnChange && !(bef & (ur_Undo|ur_Redo))) {
		    if (lg < 16) {
			tmpKeyS = &ks;
			ks.c[0] = lg;
			pnmove(ks.c + 1, FIO_GetDataP(self->fio, oa, lg), lg);
			bef |= ur_LgsInKeyS;
		    } else {
			Del_lgs.s = pndup(FIO_GetDataP(self->fio, oa, lg), lg);
			Del_lgs.lg = lg;
		    }
		}
//		TLineBuf_Del(self->lb, oa, lg);
		FIO_Delete(self->fio, oa, lg);
		if (self->OnChange) {
		    m = bef & ~(ur_InsKey|ur_InsStr|ur_OverKey);
		    self->OnChange(self, m, tmpKeyS, &Del_lgs, xa, ya, xe, ye);
		}

		if (ya > ye) {
		    SWAP_INT(xa, xe);
		    SWAP_INT(ya, ye);
//		    m = xa; xa = xe; xe = m;
//		    m = ya; ya = ye; ye = m;
		} if ((ya == ye) && (xa > xe)) {
//		    m = xa; xa = xe; xe = m;
		    SWAP_INT(xa, xe);
		}

		self->SelAx = xa;
		self->SelAy = ya;
		self->CaretCharIdx = xa;
		self->CaretCharIdy = ya;
	    }
	}


	if (bef & (ur_InsKey|ur_InsStr|ur_OverKey)) {
	    if (bef & ur_InsStr) {
		oa = GetCharOffI(self, xa, ya, True);
		oe = GetCharOffI(self, xe, ye, True);
		if (oa > oe) oa = oe;

		if (bef & ur_LgsInKeyS) {
//		    TLineBuf_Ins(self->lb, oa, KeyS->c + 1, KeyS->c[0]);
		    FIO_Insert(self->fio, oa,  KeyS->c + 1, KeyS->c[0]);
		} else if (Ins_lgs) {
//		    TLineBuf_Ins(self->lb, oa, Ins_lgs->s, Ins_lgs->lg);
		    FIO_Insert(self->fio, oa, Ins_lgs->s, Ins_lgs->lg);
		}

	    } else {
		self->DoChange(self, bef, KeyS, Ins_lgs);
	    }

	    m = bef & ~ur_DelSel;
	    if (self->OnChange) {
		self->OnChange(self, m, KeyS, Ins_lgs, self->SelAx, self->SelAy, 
				self->CaretCharIdx, self->CaretCharIdy);
	    }
	}
}

/*____________________________ Undo Redo _____________________________*/
static void THexEdit_Undo_ChkCaret(THexEdit *self) {
  int idy;

	idy = self->CaretCharIdy - self->VSB->Count;
	if ((idy < 0) || (idy >= self->VisibleLines)) {
	    idy = self->CaretCharIdy - (self->VisibleLines / 2);
	    if (idy < 0) idy = 0;
	    TScrollBar_SetCount_ev(self->VSB, idy);
	}
	THexEdit_SetCaret(self);
	Refresh((TWins*)self);
}

bool THexEdit_Undo(THexEdit *self, TUnRedoLine *url) {

// printf("%s(%d): %d %d  %d %d\n", FL, url->xa, url->ya, url->xe, url->ye);
	self->SelAx = url->xe;
	self->SelAy = url->ye;
	self->CaretCharIdx = url->xa;
	self->CaretCharIdy = url->ya;

	if (url->fu & (ur_InsKey|ur_OverKey)) {
	    THexEdit_DoChange(self, url->fu, &url->KeyS, &url->lgs);
	    THexEdit_Undo_ChkCaret(self);
	    return True;

	} else if (url->fu & (ur_DelChar|ur_DelCharA)) {
	    int m = url->fu & ~(ur_DelChar|ur_DelCharA);
	    THexEdit_DoChange(self, m | ur_InsStr, &url->KeyS, &url->lgs);
	    self->SelAx = self->CaretCharIdx;
	    self->SelAy = self->CaretCharIdy;
	    THexEdit_Undo_ChkCaret(self);
	    return True;

	}
	return False;
}

bool THexEdit_Redo(THexEdit *self, TUnRedoLine *url) {
	self->SelAx = url->xa;
	self->SelAy = url->ya;
	self->CaretCharIdx = url->xe;
	self->CaretCharIdy = url->ye;

	THexEdit_DoChange(self, url->fu, &url->KeyS, &url->lgs);
	THexEdit_Undo_ChkCaret(self);
	return True;
}

/*____________________________ File IO _______________________________*/
static void fio_OnReadEnd(TFIO *self) {
  THexEdit *hxe = self->owner;
	if (hxe->OnReadEnd) hxe->OnReadEnd(hxe);
	Adjust_VHSB(hxe);
	Refresh((TWins*)hxe);
}

static void fio_OnWriteEnd(TFIO *self) {
  THexEdit *hxe = self->owner;
	if (hxe->OnWriteEnd) hxe->OnWriteEnd(hxe);
}

static void fio_OnData(TFIO *self, bool mk_eof) {
  THexEdit *hxe = self->owner;
	if (hxe->OnDataLoad) hxe->OnDataLoad(hxe, mk_eof);
	Adjust_VHSB(hxe);
	Refresh((TWins*)hxe);
}

static void fio_OnSetOffset(TFIO *self, int64_t Offset) {
  THexEdit *hxe = self->owner;
	if (hxe->OnSetOffset) hxe->OnSetOffset(hxe, Offset);
}

static void fioSetUp(THexEdit *self) {
	self->fio->owner = self;
	self->fio->OnReadEnd = fio_OnReadEnd;
	self->fio->OnWriteEnd = fio_OnWriteEnd;
	self->fio->OnData = fio_OnData;
	self->fio->OnSetOffset = fio_OnSetOffset;
}

bool THexEdit_LoadFile(THexEdit *self, const char *FileName) {
  TglobPath *gp;
  Tstat64 *st;
  const char *pcerr = NULL;
  bool result;

	ppTfree(&self->fio);

	gp = pctogp(FileName);
	st = file_stat_gp(gp, fsFollowLink | fsGetFileType | fsNoFileName | fsNoLinkName);
	if (!st || (st == PTR_FAILED)) {
	    if (!st) pcerr = "file not found";
	    st = NULL;

#ifdef _OS_WIN32
	} else if (!(st->st & st_Reg)) {
#else
	} else if (!(st->st & st_Reg) && (((st->mode & S_IFMT) != S_IFBLK) &&
		(((st->mode & S_IFMT) != S_IFCHR) || !pccmp("/dev/mem", FileName)))) {
#endif
	    pcerr = "file is not a Regular file !!!";
	}

	if (pcerr) {
	    char *fn = get_DisplayFileStr(NULL, gp);
	    TTC_netError((TTC0*)self, NULL, EMSG_ERR1, "%s \n%s", pcerr, fn);
	    xfree(fn);

	} else if (st) {
#ifdef _OS_WIN32
	    self->fio = TFIO_Create_LB(FileName);
#else
	    if (((st->mode & S_IFMT) == S_IFBLK) || 
		((st->mode & S_IFMT) == S_IFCHR) ||
		(st->size >= 0x00400000)) { 
		self->fio = TFIO_Create_FD(FileName);
	    } else {
		self->fio = TFIO_Create_LB(FileName);
	    }
#endif
	    fioSetUp(self);
//	    self->fio->owner = self;
//	    self->fio->OnReadEnd = fio_OnReadEnd;
//	    self->fio->OnWriteEnd = fio_OnWriteEnd;
//	    self->fio->OnData = fio_OnData;
	}

	xfree(st);
	gp_free(gp);


//	if (!FIO_ReOpen(self->fio, FileName)) return False;
//	if (!FIO_Open(self->fio)) return False;
	result = FIO_Open(self->fio);

	Adjust_VHSB(self);
	Refresh((TWins*)self);
	return result;
}

bool THexEdit_SaveFile(THexEdit *self, const char *FileName) {

	return FIO_Save(self->fio, FileName);
}


const char *THexEdit_GetFileName(THexEdit *self) {

	if (!self->fio) return NULL;
	return self->fio->FileName;
}

dword THexEdit_GetIOfkt(THexEdit *self) {
	if (!self) return 0;
	return FIO_GetIOfkt(self->fio);
}

int THexEdit_GetFIOArt(THexEdit *self) {
	if (!self) return 0;
	return FIO_GetArt(self->fio);
}

/*____________________________ Paint ____________________________________*/
static void SetCaretOn(THexEdit *self) {
  int idy = self->CaretCharIdy - self->VSB->Count;
	if ((idy < 0) || (idy >= self->VisibleLines)) return;
	if (self->KeyFu & mHasFocus) Caret_Draw(1);
}

static void THexEdit_DoPaint(THexEdit *self) {
  int i, y, cy, sy;
	if (self->KeyFu & mHasFocus) Caret_Draw(0);

	self->vds.Image = True;
	self->vds.BackColor = self->Color;

	cy = self->CaretCharIdy;
	sy = self->SelAy;
	self->CaretCharIdy -= self->VSB->Count;
	self->SelAy -= self->VSB->Count;

	y = 0;
	for (i = 0; i < self->VisibleLines; i++) {
	    y = TLineEditBase_PaintLine((TLineEditBase*)self, i, y);
	}

	self->CaretCharIdy = cy;
	self->SelAy = sy;


	Res_vds(self);
	SetCaretOn(self);
}

void THexEdit_SetCaret(THexEdit *self) {
  int x, y, idy;

	Caret_Draw(0);
	idy = self->CaretCharIdy - self->VSB->Count;
	if ((idy < 0) || (idy >= self->VisibleLines)) return;

	y = idy * GetLineFontHeight(self);
	Set_vds(self, idy);
	x = GetChar_XPos(self->CaretCharIdx, &self->vds);
	Res_vds(self);
	if (self->EditMode && !self->Insert && !THexEdit_ChkIsSel(self)) {
	    TWins_CaretRect_SetNewPos((TWins*)self, x, y, self->CharWidth);
	} else {
	    TWins_Caret_SetNewPos((TWins*)self, x, y);
	}
	SetCaretOn(self);
}

static void THexEdit_Refresh(THexEdit *self) {
	fill_lines(self);
	TWins_Refresh((TWins*)self);
}
/*____________________________ Ev Key ___________________________________*/
static void TmpStrg_C(THexEdit *self) {
  int xa, xe, ya, ye, lines, cx, sx, cy, sy;
  int64_t dlg, xlg;
  TLines *tls, *tmp_tls;
  TPLine pl;

	if (!THexEdit_ChkIsSel(self) || !self->PrintLine) return;
	xa = self->SelAx;
	xe = self->CaretCharIdx;
	ya = self->SelAy;
	ye = self->CaretCharIdy;

	TLines_ChkMax_xy(self->tls, &xa, &ya, &xe, &ye);
	lines = 1 + ye - ya;

	pl.Insert = self->Insert;
	pl.ASCII_Korr = !self->EditMode;
	pl.idx = ya * self->MaxLineByts;
	dlg = FIO_GetSize(self->fio);

	tls = TLines_Create();

	for (int i=lines; i>0; i--) {
	    pl.cs = TLines_Add(tls, NULL);
	    pl.cs->s = xmalloc(self->LineLength + 1);
	    pl.cs->s[self->LineLength] = 0;

	    pl.lg = 0;
	    pl.pb = (byte*)FIO_GetDataP(self->fio, pl.idx, self->MaxLineByts);
	    xlg = dlg - pl.idx;
	    if (pl.pb && (xlg > 0)) {
		if (xlg > self->MaxLineByts) xlg = self->MaxLineByts;
		pl.lg = xlg;
	    }
	    self->PrintLine(self, &pl);
// printf("%s(%d): %s\n", FL, pl.cs->s);
	}

	if (self->EditMode) {
	    if (self->Strg_C) self->Strg_C(self, tls, xa, xe, 0, lines - 1);

	} else {
	    tmp_tls = self->tls;
	    self->tls = tls;
	    sx = self->SelAx;
	    cx = self->CaretCharIdx;
	    sy = self->SelAy;
	    cy = self->CaretCharIdy;
	    self->SelAx = xa;
	    self->CaretCharIdx = xe;
	    self->SelAy = 0;
	    self->CaretCharIdy = lines - 1;
    // printf("%s(%d): %d %d %d %d\n", FL, self->SelAx, self->CaretCharIdx, self->SelAy, self->CaretCharIdy);

	    TLineEditBase_Strg_C((TLineEditBase*)self);

	    self->CaretCharIdx = cx;
	    self->SelAx = sx;
	    self->CaretCharIdy = cy;
	    self->SelAy = sy;
	    self->tls = tmp_tls;
	}
	Tfree(tls);
}

#define fu_Refresh	1
#define fu_Sel		(1<<1)
#define fu_KeyShift	(1<<2)
#define fu_IsSel	(1<<8)

static bool THexEdit_XK_Right(THexEdit *self, int *RetFu, bool NoChk) {
  int OffX;
  int fu;

	fu = 0;
	if (RetFu) fu = *RetFu;
	OffX = self->CaretCharIdxToOffX(self, self->CaretCharIdx, self->CaretCharIdy, 0, True);

	if (self->Seg == Seg_Txt) {
	    if (OffX < (self->MaxLineByts - 1)) {
		if (!chk_xy_ok(self, OffX + 1, self->CaretCharIdy, self->Insert)) return False;
		OffX++;
	    } else {
		if (!chk_xy_ok(self, 0, self->CaretCharIdy + 1, self->Insert)) return False;
		OffX = 0;
		self->CaretCharIdy++;
	    }
	} else {
	    if (NoChk || (fu & fu_KeyShift)) {
		if (OffX < (self->MaxLineByts - 1)) {
		    if (!chk_xy_ok(self, OffX + 1, self->CaretCharIdy, True)) return False;
		    self->ValChrOff = 0;
		    OffX++;
		} else {
		    if (!chk_xy_ok(self, 0, self->CaretCharIdy + 1, True))  return False;
		    self->ValChrOff = 0;
		    OffX = 0;
		    self->CaretCharIdy++;
		}
	    } else {
		if (OffX < (self->MaxLineByts - 1)) {
		    if (self->ValChrOff >= self->MaxValChrOff) {
			if (!chk_xy_ok(self, OffX + 1, self->CaretCharIdy, self->Insert)) return False;
			self->ValChrOff = 0;
			OffX++;
		    } else {
			self->ValChrOff++;
		    }
		} else if (self->ValChrOff >= self->MaxValChrOff) {
		    if (!chk_xy_ok(self, 0, self->CaretCharIdy + 1, self->Insert))  return False;
		    self->ValChrOff = 0;
		    OffX = 0;
		    self->CaretCharIdy++;
		} else {
		    self->ValChrOff++;
		}
	    }
	}
	self->CaretCharIdx = self->OffXToCaretCharIdx(self, OffX);
	if (fu & fu_KeyShift) {
	    fu |= (fu_Refresh | fu_Sel);
	    if (self->CaretKorr) { 
		self->CaretKorr(self, &self->CaretCharIdx,
				&self->CaretCharIdy, cfu_Select);
		}
	}
	if (RetFu) *RetFu = fu;
	return True;
}

static bool THexEdit_XK_Left(THexEdit *self, int *RetFu, bool NoChk) {
  int OffX;
  int fu;

	fu = 0;
	if (RetFu) fu = *RetFu;
	OffX = self->CaretCharIdxToOffX(self, self->CaretCharIdx, self->CaretCharIdy, 0, True);

	if (self->Seg == Seg_Txt) {
	    if (OffX > 0) {
		OffX--;
	    } else if (self->CaretCharIdy > 0) {
		self->CaretCharIdy--;
		OffX = self->MaxLineByts - 1;
	    } else return False;

	} else {
	    if (NoChk || (fu & fu_KeyShift)) {
		self->ValChrOff = 0; 
		if (OffX > 0) {
		    OffX--;
		} else if (self->CaretCharIdy > 0) {
		    self->CaretCharIdy--;
		    OffX = self->MaxLineByts - 1;
		} else return False;

	    } else {
		if (self->ValChrOff > 0) {
		    self->ValChrOff--;
		} else if (OffX > 0) {
		    self->ValChrOff = self->MaxValChrOff; 
		    OffX--;
		} else if (self->CaretCharIdy > 0) {
		    self->ValChrOff = self->MaxValChrOff; 
		    self->CaretCharIdy--;
		    OffX = self->MaxLineByts - 1;
		} else return False;
	    }
	}
	self->CaretCharIdx = self->OffXToCaretCharIdx(self, OffX);
	if (fu & fu_KeyShift) {
	    fu |= (fu_Refresh | fu_Sel);
	    if (self->CaretKorr) { 
		self->CaretKorr(self, &self->CaretCharIdx,
				&self->CaretCharIdy, cfu_Select);
		}
	}
	if (RetFu) *RetFu = fu;
	return True;
}


static int THexEdit_DoKeyDown(THexEdit *self, TKeyEvent *KeyEv) {
  int ValChrOff, OrgValChrOff;
  int OffX, OrgOffX;
  int idy, OrgIdy, RetFu;
  dword coff;
  bool new;

	TWins_DoKeyDown((TWins*)self, KeyEv);

	if (appXDat.SelWait) return 0;

	if ((KeyEv->keycode == 54) && (KeyEv->state & 0x4)) { //Strg_C
	    TmpStrg_C(self);
	    return 1;
	}

	if (!FIO_GetSize(self->fio) && !self->Insert) return 0;

	new = False;
	RetFu = 0;
	if (THexEdit_ChkIsSel(self)) RetFu = fu_IsSel;
	if (KeyEv->state & ShiftMask) RetFu |= fu_KeyShift;

	ValChrOff = self->ValChrOff;
	OrgValChrOff = ValChrOff;

	coff = GetCharOff(self, self->CaretCharIdx, self->CaretCharIdy);
	OffX = self->CaretCharIdxToOffX(self, self->CaretCharIdx, self->CaretCharIdy, 0, self->Insert);
	OrgOffX = OffX;
	idy = self->CaretCharIdy;
	OrgIdy = idy;

	if (KeyEv->keysym == XK_Left) {
	    new = THexEdit_XK_Left(self, &RetFu, self->Insert);

	} else if (KeyEv->keysym == XK_Right) {
	    new = THexEdit_XK_Right(self, &RetFu, self->Insert);

	} else if (KeyEv->keysym == XK_Up) {
	    if (idy > 0)  {
		idy--;
	    } else {
		int64_t ioff = FIO_GetOffset(self->fio);
		if (ioff > 0) {
		    ioff -= self->MaxLineByts;
		    if (ioff < 0) ioff = 0;
		    FIO_SetOffset(self->fio, ioff);
		    new = True;
		}
	    }
/*	    if (RetFu & fu_KeyShift) {
		RetFu |= (fu_Refresh | fu_Sel);
		if (self->CaretKorr) { 
		    self->CaretKorr(self, &self->CaretCharIdx,
				    &idy, cfu_Select);
		    }
	    } */

	} else if (KeyEv->keysym == XK_Down) {
	    if (chk_xy_ok(self, OffX, idy + 1, self->Insert)) {
		if ((idy + 1) >= self->VSB->MaxCount) {
		    int64_t ioff = FIO_GetOffset(self->fio);
		    ioff += self->MaxLineByts;
		    FIO_SetOffset(self->fio, ioff);
		    new = True;
		} else {
		    idy++;
		}
	    }
/*	    if (RetFu & fu_KeyShift) {
		RetFu |= (fu_Refresh | fu_Sel);
		if (self->CaretKorr) { 
		    self->CaretKorr(self, &self->CaretCharIdx,
				    &idy, cfu_Select);
		    }
	    } */

	} else if (KeyEv->keysym == XK_Page_Up) {

	} else if (KeyEv->keysym == XK_Page_Down) {

	} else if (KeyEv->keysym == XK_Home) {

	} else if (KeyEv->keysym == XK_End) {

	} else if (KeyEv->keysym == XK_BackSpace) {
	    if (!self->Insert) return False;
	    if (coff  > 0) {
		self->ValChrOff = 0;
		THexEdit_XK_Left(self, NULL, True);
		THexEdit_DoChange(self, ur_DelSel|ur_DelChar, NULL, NULL);
		RetFu |= fu_Refresh;
		new = True;
	    }

	} else if (KeyEv->keysym == XK_Delete) {
	    if (!self->Insert) return False;
	    if (coff  < FIO_GetSize(self->fio)) {
		THexEdit_XK_Right(self, NULL, True);
		THexEdit_DoChange(self, ur_DelSel|ur_DelCharA, NULL, NULL);
	        if (coff >= FIO_GetSize(self->fio)) THexEdit_XK_Left(self, NULL, True);
		RetFu |= fu_Refresh;
		new = True;
	    }

	} else if (self->KeyDown && self->KeyDown(self, KeyEv, &RetFu)) {
	    THexEdit_XK_Right(self, NULL, False);
	    RetFu |= fu_Refresh;
	    new = True;

	} else return False;

	if (THexEdit_ChkIsSel(self)) RetFu |= fu_Refresh;

	if ((OffX != OrgOffX) || (ValChrOff != OrgValChrOff) ||
	    (idy != OrgIdy)) {
	    self->ValChrOff = ValChrOff;
	    self->CaretCharIdx = self->OffXToCaretCharIdx(self, OffX);
	    self->CaretCharIdy = idy;
	    new = True;
	}

	if (new) {
	    if (!(RetFu & fu_Sel)) {
		self->SelAx0 = self->SelAx = self->CaretCharIdx;
		self->SelAy = self->CaretCharIdy;
	    }

	    if (self->CaretCharIdy < self->VSB->Count) {
		TScrollBar_SetCount_ev(self->VSB, self->CaretCharIdy);
	    } else if (self->CaretCharIdy >= (self->VSB->Count + self->VisibleAktLines)) {
		TScrollBar_SetCount_ev(self->VSB, self->CaretCharIdy - self->VisibleAktLines + 1);
	    }
	    THexEdit_SetCaret(self);
	}

	if (RetFu & fu_Refresh) {
	    Refresh((TWins*)self);
	}

	return True;
}

/*____________________________ Ev _______________________________________*/
static bool THexEdit_Ctl(THexEdit *self, TWinsCtl *wctl) {

	if (!TWins_Ctl((TWins*)self, wctl)) return False;
	if (wctl->bef == wctlb_SetFont) {
	    if (self->ReInit) self->ReInit(self);
	    Adjust_VHSB(self);
	    THexEdit_SetCaret(self);
	}
	return True;
}

static bool THexEdit_AdjustSize(THexEdit *self) {
  bool result;
	result = TWins_AdjustSize((TWins*)self);
	Adjust_VHSB(self);
	return result;
}

static void THexEdit_VSB_OnCount(TScrollBar* sb, int Count) {
  THexEdit *hexed = (THexEdit*)sb->owner;

	THexEdit_SetCaret(hexed);
	Refresh((TWins*)hexed);
}

static void THexEdit_DoMouseDown(THexEdit *self, XButtonEvent *ev) {
	TWins_DoMouseDown((TWins*)self, ev);
	if ((ev->button == Button1) && TestWinIsEnabled((TWins*)self)) {
	    GetCaretIdxy(self, ev->x, ev->y, &self->CaretCharIdx, &self->CaretCharIdy);
	    if (self->EditMode && self->CaretKorr) {
		self->CaretKorr(self, &self->CaretCharIdx, &self->CaretCharIdy, 0);
	    }
	    if (THexEdit_ChkIsSel(self)) {
		self->SelAx = self->CaretCharIdx;
		self->SelAy = self->CaretCharIdy;
		Refresh((TWins*)self);
	    }
	    THexEdit_SetCaret(self);
	}
}

static void THexEdit_DoMouseMove(THexEdit *self, XMotionEvent *ev) {
  int x, y, yo;

	TWins_DoMouseMove((TWins*)self, ev);
	if ((ev->state & Button1Mask) && (appXDat.MouseDown.apph == (TWins*)self)) {
//	    if (self->EditMode) {
		yo = appXDat.MouseDown.y - ev->y;
		if (yo > 0) {
		    if (yo > 4) yo = 4;
		} else {
		    if (yo < -4) yo = -4;
		}
		x = self->CaretCharIdx;
		y = self->CaretCharIdy;
		GetCaretIdxy(self, ev->x, ev->y+yo, &self->CaretCharIdx, &self->CaretCharIdy);
		if (self->EditMode && self->CaretKorr) {
		    self->CaretKorr(self, &self->CaretCharIdx, &self->CaretCharIdy, cfu_Select);
//		    self->CaretCharIdx++;
		}

		if ((x != self->CaretCharIdx) || (y != self->CaretCharIdy)) {
		    THexEdit_SetCaret(self);
		    Refresh((TWins*)self);
		}
//	    } else {

//	    }
	}
}

static void THexEdit_DoMouseUp(THexEdit *self, XButtonEvent *ev) {
	TWins_DoMouseUp((TWins*)self, ev);
	TScrollBar_MouseZ_Wheel(self->VSB, ev);
/* 	if ((ev->button == Button4) || (ev->button == Button5)) {
	    int VSBc = self->VSB->Count;
	    if (ev->button == Button4) VSBc -= self->MouseZ_Wheel; else VSBc += self->MouseZ_Wheel;
	    TScrollBar_SetCount_ev(self->VSB, VSBc);
	} */
}

static void THexEdit_DoFocus(THexEdit *self, int Change) {
	TWins_DoFocus((TWins*)self, Change);
	if (Change & WinsFocusIN) {
	    THexEdit_SetCaret(self);
	} else {
	    Caret_Draw(0);
	}
}

/*____________________________ Settings _________________________________*/
int64_t THexEdit_GetSizeAbs(THexEdit *self) {
	if (!self) return 0;
	return FIO_GetSizeAbs(self->fio);
}

bool THexEdit_SetOffset(THexEdit *self, int64_t Offset) {
	if (!self) return False;
	if (!FIO_SetOffset(self->fio, Offset)) return False;
	THexEdit_AdjustSize(self);
	Refresh((TWins*)self);
	return True;
}

int64_t THexEdit_GetOffset(THexEdit *self) {
	if (!self) return 0;
	return FIO_GetOffset(self->fio);
}

void THexEdit_SetMaxSegData(THexEdit *self, int MaxSegData) {
	if (MaxSegData < 0x1000) MaxSegData = 0x1000;
	if (MaxSegData > 0x40000000) MaxSegData = 0x40000000;
	if (self && (self->MaxSegData != MaxSegData)) {
	    self->MaxSegData = MaxSegData;
	    THexEdit_AdjustSize(self);
	    Refresh((TWins*)self);
	}
}

int THexEdit_GetMaxSegData(THexEdit *self) {
	if (!self) return 0;
	return self->MaxSegData;
}

void THexEdit_SetEditMode(THexEdit *self, bool EditMode) {
	if (self && (self->EditMode != EditMode)) {
	    self->EditMode = EditMode;
	    if (EditMode) {
		self->LEBfu |= LEBfu_SelRectXoffAkt;
	    } else {
		self->LEBfu &= ~LEBfu_SelRectXoffAkt;
	    }
	    self->SelAx = self->CaretCharIdx;
	    self->SelAy = self->CaretCharIdy;
	    Refresh((TWins*)self);
	}
}

void THexEdit_SetInsertMode(THexEdit *self, bool Insert) {
	if (self && (self->Insert != Insert)) {
	    self->Insert = Insert;
	    self->SelAx = self->CaretCharIdx;
	    self->SelAy = self->CaretCharIdy;
	    Refresh((TWins*)self);
	}
}

bool THexEdit_GetChanged(THexEdit *self) {
	if (!self) return False;
	return FIO_GetChanged(self->fio);
}
/*____________________________ Destroy __________________________________*/
static bool THexEdit__Destroy(THexEdit *self) {
	return TWins__Destroy((TWins*)self);
}

/*_____________________________ Create __________________________________*/
static void THexEdit_CreateParams(TWins_Create_struct *wcs) {
  THexEdit *self = wcs->wins;

	Caret_Init();

	TScrollWin_CreateParams(wcs);

	self->tls = TLines_Create();
	self->tls->Owner = (TWins*)self;

	self->destroy = (TDESTROY)THexEdit__Destroy;
	self->AdjustSize = THexEdit_AdjustSize;
	self->DoPaint = THexEdit_DoPaint;
	self->WinsCtl = THexEdit_Ctl;
	self->Refresh = THexEdit_Refresh;

	self->DoMouseDown = THexEdit_DoMouseDown;
	self->DoMouseUp = THexEdit_DoMouseUp;
	self->DoMouseMove = THexEdit_DoMouseMove;
	self->DoKeyDown = THexEdit_DoKeyDown;
	self->DoFocus = THexEdit_DoFocus;

	self->VSB->OnCount = THexEdit_VSB_OnCount;
	self->HSB->Visible = False;

	self->KeyFu = mCanFocus;
	self->fio_Strg_C_on = True;

	self->MaxSegData = 0x40000000;

	self->vds.Font = MakeFont(appXDat.Default_Font->name, NULL, NULL);
	ChangeFontGC(self->vds.Font, self->vds.ForeColor, self->Color);

	self->cursor_nr = XC_xterm;
//	self->EditMode = True;

	self->fio = TFIO_Create_LB(NULL);
	fioSetUp(self);

	for (int i = MAX_LINES; i > 0; i--) {
	    TColS *cs = TLines_Add(self->tls, NULL);
	    cs->s = xmalloc(MAX_CHARS);
	    cs->lg = 0;
	    TStrColorDatArr *scda = set_scda_arr((void*)&cs->c, MAX_COLORS);
	    scda->items = 0;
	}
	InitLines_8Hex(self);
}

static THexEdit *THexEdit_Create(TWins *Parent) {
  TWins_Create_struct wcs;
	TWins_Create_struct_init(&wcs, sizeof(THexEdit), wtyp_USER_Wins);
	wcs.Parent = Parent;
	wcs.CreateParams = THexEdit_CreateParams;
	TScrollWin_DefaultConf(&wcs);
	return wcs.wins;
}

THexEdit *THexEdit_CreateA(TWins *Parent, int Left, int Top, int Width, int Height, int Align) {
  THexEdit *self;
	self = THexEdit_Create(Parent);
	self->pw->top = Top;
	self->pw->left = Left;
	self->pw->width = Width;
	self->pw->height = Height;
	self->Align = alClient;
	self->pw->Align = Align;
	Set_Parent((TWins*)self, Parent);
	return self;
}
