/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "thexedit.h"

#define HEXANF 11
#define HEXLG  48
#define TXTANF (HEXANF + HEXLG + 2)

static int CaretCharToByteNr(int idx, int seg, int MaxByts, int *PValChrOff) {
  int result;
  int ValChrOff;

	result = idx;
	ValChrOff = 0;
	if (seg == Seg_Txt) {
	    result -= TXTANF;
	    if (result < 0) result = 0;
	} else {
	    result -= HEXANF;
	    if (result > 23) result--;
	    if (result < 0) result = 0;
	    ValChrOff = result - ((result / 3) * 3);
	    if (ValChrOff > 1) ValChrOff = 1;
	    result /= 3;
	}
	if (result > MaxByts) result = MaxByts;
	if (PValChrOff) *PValChrOff = ValChrOff;
	return result;
}

static bool MaxLineByteKorr(THexEdit *self, int *idx, int idy, bool Plus1) {
  int64_t off, end, max;
  int64_t x;

	x = *idx;
	end = FIO_GetSize(self->fio);
	max = idy * self->MaxLineByts;
	off = max + x;
	if (off < end) return False;
	if (Plus1) {
	    x = end - max;
	} else {
	    x = end - 1 - max;
	}
	if (x < 0) x = 0;
	*idx = x;
	return True;
}

static int CaretCharIdxToOffX(THexEdit *self, int idx, int idy, int cfu, bool Insert) {
  int x;
 // int64_t off;

	x = idx;
	if ((!(cfu & cfu_Select) && (x > (TXTANF - 2))) 
	    || ((cfu & cfu_Select) && (self->Seg == Seg_Txt))) {
	    self->SelXoffA = TXTANF;
	    self->SelXoffE = TXTANF + self->MaxLineByts;
	    self->Seg = Seg_Txt;
	} else {
	    self->SelXoffA = HEXANF;
	    self->SelXoffE = HEXANF + HEXLG;
	    self->Seg = Seg_Hex;
	}

	x = CaretCharToByteNr(idx, self->Seg, self->MaxLineByts - 1, &self->ValChrOff);

	if (!self->fio) {
	    self->ValChrOff = 0;
	    return 0;
	}

	if (MaxLineByteKorr(self, &x, idy, Insert)) {
	    if (Insert) {
		self->ValChrOff = 0;
	    } else {
		self->ValChrOff = 1;
	    }
	}
/*
	off = (idy * self->MaxLineByts) + x;
	if (off >= FIO_GetSize(self->fio)) {
	    if (Insert) {
		self->ValChrOff = 0;
		x = FIO_GetSize(self->fio) - (idy * self->MaxLineByts);
	    } else {
		self->ValChrOff = 1;
		x = FIO_GetSize(self->fio) - 1 - (idy * self->MaxLineByts);
	    }
	    if (x < 0) x = 0;
	} */
	return x;
}

static int ByteXToCaretCharIdx(THexEdit *self, int x, int ValChrOff) {
  int result;
	if (x < 0) return self->SelXoffA;
	if (self->Seg == Seg_Hex) {
	    if (x < 8) {
		result = self->SelXoffA + (x * 3) + ValChrOff;
	    } else {
		result = self->SelXoffA + 1 + (x * 3) + ValChrOff;
	    }
	} else {
	    result = self->SelXoffA + x;
	}
	if (result > self->SelXoffE) result = self->SelXoffE;
	return result;
}

static int OffXToCaretCharIdx(THexEdit *self, int x) {
	return ByteXToCaretCharIdx(self, x, self->ValChrOff);
}

static void CaretKorr_8Hex(THexEdit *self, int *idx, int *idy, int cfu) {
#if 0
	int x = CaretCharIdxToOffX(self, *idx, *idy, cfu, self->Insert);
	*idx = OffXToCaretCharIdx(self, x);
#else
  int cx, cy;
  int x, xa, rx, ValChrOff;
  bool max;

	cx = *idx;
	cy = *idy;

	x = CaretCharIdxToOffX(self, cx, cy, cfu, self->Insert && !(cfu & cfu_Select));
	max = x >= (self->MaxLineByts - 1);
	rx = OffXToCaretCharIdx(self, x);

	if ((cfu & cfu_Select) && (self->Seg == Seg_Hex)) {

	    if (max && (cx >= self->SelXoffE)) {
		 rx = self->SelXoffE;
	    } else if (cx < self->SelXoffA) {
		 rx = self->SelXoffA;
	    } else if (self->SelAy == cy) {
		if (self->SelAx0 < cx) {
		    rx = ByteXToCaretCharIdx(self, x, 2);
		} else {
		    rx = ByteXToCaretCharIdx(self, x, 0);
		}
	    } else if (self->SelAy < cy) {
		rx = ByteXToCaretCharIdx(self, x, 2);
	    } else {
		rx = ByteXToCaretCharIdx(self, x, 0);
	    }
	    *idx = rx;

	    xa = CaretCharToByteNr(self->SelAx0, self->Seg, self->MaxLineByts - 1, &ValChrOff);
	    if ((self->SelAy < cy) || 
		((self->SelAy == cy) && (self->SelAx0 < cx))) {
		self->SelAx = ByteXToCaretCharIdx(self, xa, 0);
	    } else {
		self->SelAx = ByteXToCaretCharIdx(self, xa-1, 2);
	    }

	} else if (cfu & cfu_Select) {
	    if (max && (cx >= self->SelXoffE)) {
		 rx = self->SelXoffE;
	    } else if (cx < self->SelXoffA) {
		 rx = self->SelXoffA;
	    } else if (self->SelAy == cy) {
		if (self->SelAx0 < cx) rx++;
	    } else if (self->SelAy < cy) {
		rx++;
	    }
	    *idx = rx;

	} else {
	    self->SelAx0 = OffXToCaretCharIdx(self, x);
	    *idx = self->SelAx0;
	}

// printf("%s(%d): %d %d\n", FL, OffXToCaretCharIdx(self, self->SelAx), OffXToCaretCharIdx(self, *idx));
#endif
}

static void PrintLine_8Hex(THexEdit *self, TPLine *pl) {
  TStrColorDatArr *scda;

	if (!pl->cs) return;
	pl->cs->fu &= ~csfu_HasData;
	pl->cs->nc3 = 0;
	pl->cs->lg = TXTANF + 16;
	FillChar(pl->cs->s, pl->cs->lg, ' ');

	scda = (TStrColorDatArr*)pl->cs->c;
	if (scda) scda->items = 1;

	if (pl->lg || !pl->idx || pl->Insert) {

	    if (scda) {
		scda->items = 3;
		TStrColorDat *scd = &scda->scd[1];
		if (pl->idx & 0x10) {
		    scd->s = pl->cs->s + HEXANF + 12;
		} else {
		    scd->s = pl->cs->s + HEXANF;
		}

		scd = &scda->scd[2];
		if (pl->idx & 0x10) {
		    scd->s = pl->cs->s + HEXANF + 37;
		} else {
		    scd->s = pl->cs->s + HEXANF + 25;
		}
	    }

	    pl->cs->fu |= csfu_HasData;
	    pl->cs->nc3 = pl->lg;

	    char *pc = pl->cs->s; // + 1;
	    int64_t plidx = pl->idx + FIO_GetOffset(self->fio);
//	    hextoppc(&pc, (plidx >> 32) & 0xff, 2 | kNO_NULL);
//	    *pc = '-';
//	    pc++;
	    hextoppc(&pc, (plidx >> 16) & 0xffff, 4 | kNO_NULL);
	    *pc = '-';
	    pc++;
	    hextopc(pc, plidx & 0xffff, 4 | kNO_NULL);

	    pc = pl->cs->s + HEXANF;
	    byte *p = pl->pb;
	    dword l = pl->lg;
	    for (int x = 16; x > 0; x--) {
		if (l) {
		    hextopc(pc, *p, 2 | kNO_NULL);
		    l--;
		    p++;
		} else {
		    pl->Insert = False;
		}
		pc += 3;
		if ((x & 0x07) == 1) pc++;
	    }

	    p = pl->pb;
	    l = pl->lg;
	    for (int x = 16; x > 0; x--) {
		if (l) {
		    if (!pl->ASCII_Korr
			|| ((*p >= 0x20) && (*p < 0x80)) 
			|| (*p >= 0xa0)) {
			*pc = *p;
		    } else {
			*pc = '.';
		    }
		    l--;
		    p++;
		}
		pc++;
	    }
	    pl->idx += 16;
	    pl->pb = p;
	    pl->lg = l;
	}
}

static void Strg_C(THexEdit *self, TLines *tls, int xa, int xe, int ya, int ye) {
  char *pc;
  int lg;

// printf("%s(%d): %d %d %d %d\n", FL, xa, xe, ya, ye);
	lg = 0;
	pc = NULL;
	if (ya == ye) {
	    TColS *cs = TLines_Get_pcs(tls, ya);
	    if (cs) {
		lg = xe - xa;
		pc = pcndup(cs->s + xa, lg);
	    }
	} else {

	  for (int i=ya;i<=ye;i++) {
	    TColS *cs = TLines_Get_pcs(tls, i);
	    if (cs) {
		if (i == ya) {
		    lg = self->SelXoffE - xa;
		    ppcnadd(&pc, cs->s + xa, self->SelXoffE - xa);
		} else if (i == ye) {
		    if (self->Seg == Seg_Hex) {lg++; ppcadd(&pc, "\n");}
		    int l = xe - self->SelXoffA;
		    lg += l;
		    ppcnadd(&pc, cs->s + self->SelXoffA, l);
		} else {
		    if (self->Seg == Seg_Hex) {lg++; ppcadd(&pc, "\n");}
		    int l = self->SelXoffE - self->SelXoffA;
		    lg += l;
		    ppcnadd(&pc, cs->s + self->SelXoffA, l);
		}
	    }
	  }
	}

	SelClr();
	appXDat.SelTxt.s = pc;
	appXDat.SelTxt.lg = lg;
	SetSelection(1, XA_STRING);
}

static void ReInit_8Hex(THexEdit *self) {
  TStrColorDatArr *scda;
  TStrColorDat *scd;

	for (int i = 0; i < MAX_LINES; i++) {
	    TColS *cs = TLines_Get_pcs(self->tls, i);
	    if (cs) {
		cs->fu = 0;
		cs->nc3 = 0;

		scda = (TStrColorDatArr*)cs->c;
		scda->items = 1;
		scd = &scda->scd[0];
		scd->s = cs->s; // + 1;
		scd->lg = 9;
		scd->Font = self->vds.Font;
		scd->ForeColor = 0x000000c0;
		scd->BackColor = 0x00f0f0f0;
		scd->fu = 0;

		scd = &scda->scd[1];
		scd->s = NULL;
		scd->lg = 11;
		scd->Font = self->vds.Font;
		scd->ForeColor = 0x00000000;
		scd->BackColor = 0x00e0e0e0;
		scd->fu = 0;

		scd = &scda->scd[2];
		scd->s = NULL;
		scd->lg = 11;
		scd->Font = self->vds.Font;
		scd->ForeColor = 0x00000000;
		scd->BackColor = 0x00e0e0e0;
		scd->fu = 0;
	    }
	}
}

static void DoChange(THexEdit *self, int bef, TKeyS *KeyS, Tlgs *Ins_lgs) {
  char *pc;
  dword c, off;
  char buf[4];

	off = GetCharOff(self, self->CaretCharIdx, self->CaretCharIdy);
	if (bef & ur_OverKey) {
	    pc = FIO_GetDataP(self->fio, off, 1);
	    buf[0] = *pc;
	    c = KeyS->b[0];
	    if (bef & ur_Undo) c = KeyS->b[8];
	    if (self->Seg == Seg_Hex) {
		if (c >= 'A') c -= 7;
		c &= 0x0f;
		if (self->ValChrOff) {
		    buf[0] = c | (buf[0] & 0xf0);
		} else {
		    c <<= 4;
		    buf[0] = c | (buf[0] & 0x0f);
		}
	    } else {
		buf[0] = c;
	    }
	    FIO_OverWrite(self->fio, off, buf, 1);

	} else if (bef & ur_InsKey) {
	    dword idx = GetCharOff(self, self->CaretCharIdx, self->CaretCharIdy);
	    if (bef & ur_Undo) {
		FIO_Delete(self->fio, idx, 1);
	    } else {
		if (self->Seg == Seg_Hex) {
//		    pc = FIO_GetDataP(self->fio, idx, 1);
		    c = KeyS->b[0];
		    if (c >= 'A') c -= 7;
		    c &= 0x0f;
		    c <<= 4;
		    buf[0] = c;
		    FIO_Insert(self->fio, idx, buf, 1);
		} else {
		    FIO_Insert(self->fio, idx, KeyS->c, 1);
		}
	    }
	}
}

static int KeyDown(THexEdit *self, TKeyEvent *KeyEv, int *fu) {
  char *pc;
  int c, ur_fu;
  TKeyS KeyS;
  char buf[4];

	*fu = 0;
	if (KeyEv->state & 0x4) return 0;

	if (keysymtostr(&KeyEv->keysym, KeyS.c, False)) {
	    buf[0] = 0;
	    pc = FIO_GetDataP(self->fio, GetCharOff(self, self->CaretCharIdx, self->CaretCharIdy), 1);
	    if (pc) buf[0] = *pc;
// printf("%s(%d): %d %d 0x%02x\n", FL, self->CaretCharIdx, self->CaretCharIdy, (byte)buf[0]);

	    ur_fu = ur_OverKey;
	    KeyS.c[9] = 0;
	    c = KeyS.c[0];
	    if (self->Seg == Seg_Hex) {
		if ((c >= 'a') && (c <= 'f')) c -= 0x20;
		if (((c >= '0') && (c <= '9')) ||
		    ((c >= 'A') && (c <= 'F'))) {
		    if (self->ValChrOff) {
			KeyS.c[8] = buf[0] & 0x0f;
		    } else {
			if (self->Insert) ur_fu = ur_InsKey;
			KeyS.c[8] = (buf[0] >> 4) & 0x0f;
		    }
		    if (KeyS.b[8] > 9) KeyS.b[8] += 7;
		    KeyS.b[8] += 0x30;
		    THexEdit_DoChange(self, ur_fu, &KeyS, NULL);
		    return True;
		}

	    } else {
		if (self->Insert) ur_fu = ur_InsKey;
		KeyS.c[8] = buf[0];
		THexEdit_DoChange(self, ur_fu, &KeyS, NULL);
		return True;
	    }
	}
	return 0;
}

void InitLines_8Hex(THexEdit *self) {

	self->MaxLineByts = 16;
	self->LineLength = TXTANF + 16; //77;
	self->MaxValChrOff = 1;

	for (int i = 0; i < MAX_LINES; i++) {
	    TColS *cs = TLines_Get_pcs(self->tls, i);
	    if (cs) {
		cs->lg = self->LineLength;
		FillChar(cs->s, cs->lg, ' ');
	    }
	}

	ReInit_8Hex(self);

	self->ReInit = ReInit_8Hex;
	self->PrintLine = PrintLine_8Hex;
	self->Strg_C = Strg_C;
	self->CaretKorr = CaretKorr_8Hex;
	self->CaretCharIdxToOffX = CaretCharIdxToOffX;
	self->OffXToCaretCharIdx = OffXToCaretCharIdx;
	self->KeyDown = KeyDown;
	self->DoChange = DoChange;

	self->SelXoffA = HEXANF;
	self->SelXoffE = HEXANF + HEXLG;
}
