/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldpic.h"
#include "cicons.h"
#include "filemsg.h"

void ALDpic_pxm_free(TMainForm *fpic, int fu) {
	if (fu & ib_org_pxm) {Tfree(fpic->pxm); fpic->pxm = NULL;}
	if (fu & ib_rot_pxm) {Tfree(fpic->rot_pxm); fpic->rot_pxm = NULL;}
	if (fu & ib_rsz_pxm) {Tfree(fpic->rsz_pxm); fpic->rsz_pxm = NULL;}
	if (fu & ib_vio_pxm) {Tfree(fpic->vio_pxm); fpic->vio_pxm = NULL;}
	if (fu & ib_paint_pxm) {Tfree(fpic->paint_pxm); fpic->paint_pxm = NULL;}
}

TPixMap *ALDpic_pxm_get(TMainForm *fpic, int fu) {
  TPixMap *pxm = NULL;
	if ((fu & ib_org_pxm) && fpic->pxm) pxm = fpic->pxm;
	if ((fu & ib_rot_pxm) && fpic->rot_pxm) pxm = fpic->rot_pxm;
	if ((fu & ib_rsz_pxm) && fpic->rsz_pxm) pxm = fpic->rsz_pxm;
	if ((fu & ib_vio_pxm) && fpic->vio_pxm) pxm = fpic->vio_pxm;
	if ((fu & ib_paint_pxm) && fpic->paint_pxm) pxm = fpic->paint_pxm;
	return pxm;
}

void ALDpic_resize_pxm(TMainForm *fpic) {
  int new_w, new_h;
  TPixMap *pxm;

	if (!fpic->pxm) return;
	if (fpic->rsz_run) return;
	fpic->rsz_run = True;
	fpic->new_pxm_data = True;

	if (!fpic->conv_run) {
	    SetEnabled((TWins*)fpic->cbWen, False);

	    SetEnabled((TWins*)fpic->mbZoom_no, False);
	    SetEnabled((TWins*)fpic->mbZoom_w, False);
	    SetEnabled((TWins*)fpic->mbZoom_h, False);
	    SetEnabled((TWins*)fpic->mbZoom_wh, False);

	    SetEnabled((TWins*)fpic->mbZoom_plus, False);
	    SetEnabled((TWins*)fpic->mbZoom_minus, False);
	    CutOut_SetEvEn(False);
	    AppProcessMessages();
	}
	SwapPic(fpic);
	ALDpic_pxm_free(fpic, ib_rszfree_pxm);
	pxm = ALDpic_pxm_get(fpic, ib_rszbase_pxm);

	fpic->zoom_k = fpic->new_zoom_k;
	fpic->zoom_kh = fpic->new_zoom_kh;
	if (!fpic->cbWen->Checked) {
	    fpic->no_cbWen_zoom_k = fpic->new_zoom_k;
	    fpic->no_cbWen_zoom_kh = fpic->new_zoom_kh;
	}

	fpic->zoomTime = 0;
  dword t = GetTime_us();
	fpic->old_w = pxm->width;
	fpic->old_h = pxm->height;
	if ((fpic->zoom_k != 1) || (fpic->zoom_kh != 1) || CutOut_Chk_On()) {
	    new_w = ((double)pxm->width * fpic->zoom_k) + 0.5;
	    new_h = ((double)pxm->height * fpic->zoom_kh) + 0.5;

#ifndef _OS_WIN32
	    dword freemem = get_FreeMem(fpic, (new_w * new_h) >> 8);
	    if (!freemem) {
		fpic->zoom_k = 1;
		fpic->zoom_kh = 1;

		fpic->sbZoom->Count = 100;
		TScrollBar_Refresh(fpic->sbZoom);

		SetEnabled((TWins*)fpic, False);
		MsgBox_Error((TWins*)fpic, "low mem\nDisable zoom");
		SetEnabled((TWins*)fpic, True);


	    } else {
#endif
		if (new_w < 2) new_w = 2;
		if (new_h < 2) new_h = 2;
		fpic->old_w = new_w;
		fpic->old_h = new_h;
		AzCutOut(fpic);
		pxm->pxm_LD &= ~pxmLD_FastZoom;
		if (fpic->cbFastZoom->Checked) pxm->pxm_LD |= pxmLD_FastZoom;
		fpic->rsz_pxm = CutOut_PixMap(pxm, new_w, new_h);
		pxm = fpic->rsz_pxm;
#ifndef _OS_WIN32
	    }
#endif
	} else {
	    AzCutOut(fpic);
	}
//	if (pxm->stat & pmxStat_HandleMask) fpic->AZClear = 2;

	fpic->zoomTime = GetCurTimeDiv_us(t);

  char c[64];
	snprintf(c, sizeof(c), "%d x %d  %d%%   ld=%dms  z=%dms", 
			pxm->width, pxm->height,
			((fpic->old_w * 100)+fpic->pxm->width/2) / fpic->pxm->width,
			fpic->ldTime / 1000, fpic->zoomTime / 1000);
	SetCaption(fpic->pnlZoomAZ, c);
	fpic->rsz_run = False;
}

static void ALDpic_paint_pxm(TMainForm *fpic) {
  int left, top, w;
  TScrollWin *AzGraf = fpic->AzGraf;
  TPixMap *pxm;
  double d;
  bool mk;

	pxm = ALDpic_pxm_get(fpic, ib_rszbase_pxm);
	if (!pxm) return;
	ALDpic_Chk_dup_paint_pxm(fpic);
	d = (fpic->HSB_Count - (pxm->width >> 1)) * fpic->zoom_k;
	left = fpic->cHoff + d;    
	d = (fpic->VSB_Count - (pxm->height >> 1)) * fpic->zoom_kh;
	top = fpic->cVoff + d;

	pxm = ALDpic_pxm_get(fpic, ib_all_pxm);
// printf("%s(%d): ALDpic_paint_pxm=0x%08x st=0x%08x\n", FL, (int)pxm, pxm->stat);
	ALDpic_Chk_Mask(fpic, pxm);

	fpic->pxm_left = left;
	fpic->pxm_top  = top;
//	pxm->stat &= ~pmxStat_XPixAkt;

	if (CutOut_Chk_On()) {
	    left += TDVal_GetInteger(fpic->fcutout->dvXz);
	    top += TDVal_GetInteger(fpic->fcutout->dvYz);
	}
	TPixMap_XPutImage(pxm);
	if (pxm->stat & pmxStat_HandleMask) {
//	    XClearWindow(appXDat.dpy, AzGraf->WinID);
	    FillRectRGB(AzGraf->WinID, AzGraf->Color, 
			left, top, pxm->width, pxm->height);
	}

	TPixMap_PaintToWin(pxm, AzGraf->WinID, left, top);
	fpic->new_pxm_data = False;

//	if (!fpic->AZClear) {
	    if (left > 0) FillRectRGB(AzGraf->WinID, AzGraf->Color,
			0, 0, left, AzGraf->height);
	    w = AzGraf->width - (left + pxm->width);
	     if (w > 0) FillRectRGB(AzGraf->WinID, AzGraf->Color,
			left + pxm->width, 0, w, AzGraf->height);
	    if (top > 0) FillRectRGB(AzGraf->WinID, AzGraf->Color,
			0, 0, AzGraf->width, top);
	    w = AzGraf->height - (top + pxm->height);
	    if (w > 0) FillRectRGB(AzGraf->WinID, AzGraf->Color,
			0, top + pxm->height, AzGraf->width, w);
//	}
	DrawRect(AzGraf->WinID, fpic->GC_Dot, left-1, top-1, pxm->width+2, pxm->height+2);

	mk = ((fpic->pxm->width*2 > AzGraf->width) || (fpic->pxm->width*2 > AzGraf->width));
	SetEnabled((TWins*)fpic->mbZoom_no, mk);
	SetEnabled((TWins*)fpic->mbZoom_w, mk);
	SetEnabled((TWins*)fpic->mbZoom_h, mk);
	SetEnabled((TWins*)fpic->mbZoom_wh, mk);

	SetEnabled((TWins*)fpic->mbZoom_plus, (fpic->sbZoom->Count < 200));
	SetEnabled((TWins*)fpic->mbZoom_minus, (fpic->sbZoom->Count > 0));
	SetEnabled((TWins*)fpic->cbWen, True);
	CutOut_SetEvEn(True);
	fpic->paint_run = False;
}

static void ALDpic_AzGraf_OnPaint(TScrollWin *self) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
	ALDpic_paint_pxm(fpic);
}

static void ALDpic_HSB_OnCount(TScrollBar *sb, int Count) {
  TMainForm *fpic = (TMainForm*)sb->FormStruct;
	fpic->HSB_Count = Count;
	Refresh((TWins*)fpic->AzGraf);
}

static void ALDpic_VSB_OnCount(TScrollBar *sb, int Count) {
  TMainForm *fpic = (TMainForm*)sb->FormStruct;
	fpic->VSB_Count = Count;
	Refresh((TWins*)fpic->AzGraf);
}


static void ALDpic_sbZoom_OnCount(TScrollBar *sb, int Count) {
  TMainForm *fpic = (TMainForm*)sb->FormStruct;

	if (!fpic->pxm) return;
	fpic->new_zoom_k = (double)fpic->sbZoom->Count / 100;
	if (fpic->new_zoom_k < 0.01) fpic->new_zoom_k = 0.01;
	fpic->new_zoom_kh = fpic->new_zoom_k;
	ALDpic_resize_pxm(fpic);
	Refresh((TWins*)fpic->AzGraf);
}


void ALDpic_mbXXXX_Set_en(TMainForm *fpic, int fu) {
	SetEnabled((TWins*)fpic->mbPlayleft, ((fu & 0x01) != 0));
	SetEnabled((TWins*)fpic->mbPlay2left, ((fu & 0x02) != 0));
	SetEnabled((TWins*)fpic->mbPlayrigth, ((fu & 0x04) != 0));
	SetEnabled((TWins*)fpic->mbPlay2rigth, ((fu & 0x08) != 0));
}

#ifndef _OS_WIN32
dword get_FreeMem(TMainForm *fpic, dword new_mem) {
  dword Result;
  TMemInfo MemInfo;
  Tpid_stat pid_stat;
  char buf[80];

	if (!meminfo_get(&MemInfo)) return 0;
	Result = MemInfo.Free + MemInfo.Cached;
	if (Result <= 1024) return 0;
	Result -= 1024;
// printf("%s(%d): %u %u %u free=%u add=%u\n", FL, MemInfo.Total, MemInfo.Free, MemInfo.Cached, MemInfo.Free + MemInfo.Cached, new_mem);

	snprintf(buf, sizeof(buf), "/proc/%d/stat", appDat.self_pid);
	if (get_pid_stat(&pid_stat, buf, PIDSTAT_NO_NAME)) {
	    Result += (pid_stat.vsize >> 10) - fpic->self_mem;
// printf("%s(%d): vsize=%ukB\n", FL, pid_stat.vsize >> 10);
	}

	new_mem += fpic->pxm0_mem * 2;
//	new_mem *= 2;
	if (fpic->rot_pxm) new_mem += fpic->pxm0_mem;
	if (fpic->rsz_pxm) new_mem += (fpic->rsz_pxm->width * fpic->rsz_pxm->height) >> 8;;
	if (fpic->vio_pxm) new_mem += (fpic->vio_pxm->width * fpic->vio_pxm->height) >> 8;;
	if (fpic->paint_pxm) new_mem += (fpic->paint_pxm->width * fpic->paint_pxm->height) >> 8;;
	if (fpic->fpal->cbColQuantOn->Checked) new_mem += 260;
// printf("%s(%d): add=%ukB\n", FL, new_mem);

	if (Result <= new_mem) return 0;
	Result -= new_mem;


// printf("%s(%d): freemem=%ukB ", FL, Result);
// if (Result > 750000) Result -= 750000; else Result = 0;
// printf("--> freemem=%ukB\n", Result);


	return Result;
}
#endif

void clr_vioc_pxm_para(TMainForm *fpic) {
	menuSetting_clr(fpic);

	TChkBox_SetChecked(fpic->pnlBrightness->cb, False);
	fpic->pnlBrightness->val = 0;
	TChkBox_SetChecked(fpic->pnlContrast->cb, False);
	fpic->pnlContrast->val = 0;
	TChkBox_SetChecked(fpic->pnlColorSat->cb, False);
	fpic->pnlColorSat->val = 0;
	TChkBox_SetChecked(fpic->pnlRed->cb, False);
	fpic->pnlRed->val = 0;
	TChkBox_SetChecked(fpic->pnlGreen->cb, False);
	fpic->pnlGreen->val = 0;
	TChkBox_SetChecked(fpic->pnlBlue->cb, False);
	fpic->pnlBlue->val = 0;

	TChkBox_SetChecked(fpic->fpal->cbColQuantOn, False);
	fpic->QuantOk = False;
}

void ALDpic_FW_Load_Pic(TFileWin *fw, TDirFileList *dfl, char *fn) {
  TMainForm *fpic = (TMainForm*)fw->FormStruct;
  int width, height;
  TScrollWin *sw = (TScrollWin*)fpic->AzGraf;
  double zoom_wk, zoom_hk;
  char *fnNoPasswd;
  char line[512];

	if (fpic->conv_run) return;
	TChkBox_SetChecked(fpic->cbWen, False);
	CutOut_Clr();

	fpic->conv_run = True;
	fnNoPasswd = NULL;

	if (dfl->st->st & st_Reg) {

	    fnNoPasswd = get_DisplayFileStr(fn, NULL);
	    snprintf(line, sizeof(line), "ALDpic  %s", fnNoPasswd);
	    SetCaption((TWins*)fpic, line);

	    ALDpic_mbXXXX_Set_en(fpic, 0);

	    SetEnabled((TWins*)fpic->mbZoom_no, False);
	    SetEnabled((TWins*)fpic->mbZoom_w, False);
	    SetEnabled((TWins*)fpic->mbZoom_h, False);
	    SetEnabled((TWins*)fpic->mbZoom_wh, False);
	    SetEnabled((TWins*)fpic->mbZoom_plus, False);
	    SetEnabled((TWins*)fpic->mbZoom_minus, False);

	    AppProcessMessages();

	    ppcchange(&fpic->pxmFile, fn);
	    if (fpic->pxmLastSaveAs) {
		ChangeFileName_ppc(&fpic->pxmLastSaveAs, GetFileName_pc(fpic->pxmFile));
	    }

	    ALDpic_pxm_free(fpic, ib_all_pxm);

	    width = fpic->AzGraf->width;
	    height = fpic->AzGraf->height;

 int t = GetTime_us();

#if 1
	    TPixMap *pxm = TPixMap_Loader(fn, NULL, pxmLD_OnlyInfo);
	    if (!pxm) goto Ende;
	    fpic->pxm0_mem = (pxm->width * pxm->height) >> 8;
// printf("%s(%d): %d %d %ukB\n", FL, pxm->width, pxm->height, fpic->pxm0_mem);
	    Tfree(pxm);

#ifndef _OS_WIN32
	    dword freemem = get_FreeMem(fpic, fpic->pxm0_mem);
	    if (!freemem) {
		RePaint((TWins*)fpic->AzGraf);
		MsgBox_Error((TWins*)fpic, "out of mem\n");
		goto Ende;
	    }

	    if (fpic->pxm0_mem > freemem) {
		RePaint((TWins*)fpic->AzGraf);
		menuTools_clr_rot_swap(fpic);
		clr_vioc_pxm_para(fpic);

		TChkBox_SetChecked(fpic->fpal->cbColQuantOn, False);
		MsgBox_Warning((TWins*)fpic, "low mem\n");
	    }
#endif

#endif
	    fpic->pxm = TPixMap_Loader(fn, NULL, pxm_4b_24bpp);

	    if (!fpic->pxm) goto Ende;
	    fpic->new_pxm_data = True;
	    fpic->miAutoMask->fu |= miNotEnabled;
	    fpic->miToolsMaskCLR->fu &= ~miNotEnabled;
//	    fpic->miSettingMaskCLR->fu &= ~miNotEnabled;
	    if (!(fpic->pxm->stat & pmxStat_has_alpha)) {
		fpic->miAutoMask->fu &= ~miNotEnabled;
		fpic->miToolsMaskCLR->fu |= miNotEnabled;
//		fpic->miSettingMaskCLR->fu |= miNotEnabled;
	    }
	    fpic->org_stat = fpic->pxm->stat;
	    ALDpic_Chk_AutoMask(fpic, 0);

	    if (fpic->pxm->img_info) {
		snprintf(line, sizeof(line), "ALDpic  %s     %dx%d (%s)", fnNoPasswd, fpic->pxm->width, fpic->pxm->height, fpic->pxm->img_info);
	    } else {
		snprintf(line, sizeof(line), "ALDpic  %s     %dx%d", fnNoPasswd, fpic->pxm->width, fpic->pxm->height);
	    }
	    SetCaption((TWins*)fpic, line);

	    fpic->ldTime = GetCurTimeDiv_us(t);

	    fpic->cHoff = width >> 1;
	    fpic->cVoff = height >> 1;
	    TScrollBar_SetCountParam(sw->HSB, 0, width, fpic->pxm->width + width,
				-(fpic->pxm->width>>1), sbfu_inv);
	    fpic->HSB_Count = 0;

	    TScrollBar_SetCountParam(sw->VSB, 0, height, fpic->pxm->height + height,
				-(fpic->pxm->height>>1), sbfu_inv);
	    fpic->VSB_Count = 0;

	    zoom_wk = (double)width / (double)fpic->pxm->width;
	    zoom_hk = (double)height / (double)fpic->pxm->height;
//	    if (((TChkBox*)fpic->cbWen)->Checked) {
//		w = pctoint(GetCaption_pc((TWins*)fpic->edW));
//		if (w < 2) w = 2;
//		fpic->new_zoom_k = (double)w / (double)fpic->pxm->width;
//	    } else {
		fpic->new_zoom_k = zoom_wk;
		if (zoom_wk > zoom_hk) fpic->new_zoom_k = zoom_hk;
//	    }
	    if (fpic->new_zoom_k > 1) fpic->new_zoom_k = 1;
	    if (fpic->new_zoom_k < 0.005) fpic->new_zoom_k = 0.005;

	    sw->VSB->SmallChange = (1 / fpic->new_zoom_k + 0.5);
	    sw->VSB->LargeChange = (10 / fpic->new_zoom_k + 0.5);
	    sw->HSB->SmallChange = (1 / fpic->new_zoom_k + 0.5);
	    sw->HSB->LargeChange = (10 / fpic->new_zoom_k + 0.5);

//	    if (!((TChkBox*)fpic->cbWen)->Checked) {
		switch (fpic->LoadZoomMod) {
		case 0: fpic->new_zoom_k = 1; break;
		case 1: fpic->new_zoom_k = zoom_wk; break;
		case 2: fpic->new_zoom_k = zoom_hk; break;
		}
//	    }
	    fpic->new_zoom_kh = fpic->new_zoom_k;

	    ALDpic_resize_pxm(fpic);
	    AzCutOut_0(fpic);
	    fpic->sbZoom->Count = fpic->old_w * 100 / fpic->pxm->width;

	    TScrollBar_Refresh(fpic->sbZoom);
	    Refresh((TWins*)fpic->AzGraf);
	}
Ende:
	xfree(fnNoPasswd);
	ALDpic_mbPlay_Chk_en(fpic);
	fpic->conv_run = False;
}

static void ALDpic_FW_OnItemEnter(TFileWin *fw, TItemEnter *ie) {
	if (ie->art != ie_artFileEnter) return;
	ALDpic_FW_Load_Pic(fw, ie->dfl, ie->fn_passwd);
}

static void ALDpic_FW_OnLoadDir(TFileWin *fw, char **fn, int *stat) {
  TMainForm *fpic = (TMainForm*)fw->FormStruct;
  char *pc;
  TpcIni *ini;

	if (!fn) return;
	if (((*stat & fwOn_stat_Mask) == fwOn_stat_End) && test_local_drive(*fn)) {
	    ini = TpcIni_Create(appDat.inifile_app);
	    TpcIni_WR(ini, APPNAME, "LastPath", *fn);
	    TpcIni_SaveAndFree(ini);
	}
	if ((*stat & fwOn_stat_Mask) == fwOn_stat_Start) {
	    pc = pctogptopc(*fn, gptfu_noPasswd);
	    SetCaption(fpic->pnlPath, pc);
	    xfree(pc);
	}
}

static void fw_OnFileMsg(TFileWin *self, TFileMsg *fmsg) {

	if (fmsg->art != fmArt_DirList) return;
	if (fmsg->typD != fmTyp_int) return;

	fmsg->Ok = True;

	switch (fmsg->D.m & dlbef_BefMask) {

	case dlbef_ListEnd:
	case dlbef_Sync:
	    ALDpic_mbPlay_Chk_en((TMainForm*)self->FormStruct);
	    break;

	case dlbef_DelSelf:
	    fmsg->Ok = False;  //in TDirListObjNy   Tfree(ObjNy);
	    break;
	}
}

static void ALDpic_AzGraf_OnResize(TScrollWin *self) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
  int width = self->width;
  int height = self->height;

	    if (!fpic->pxm) return;
	    fpic->cHoff = width >> 1;
	    fpic->cVoff = height >> 1;
}

static void AzGraf_OnMouseMove(TScrollWin *self, XMotionEvent *ev) {
  TMainForm *fmain;
  TPixMap *pxm;
  int x, y;

	fmain = (TMainForm*)self->FormStruct;
	pxm = ALDpic_pxm_get(fmain, ib_paintbase_pxm);
	if (!pxm) return;
	x = ev->x - fmain->pxm_left;
	y = ev->y - fmain->pxm_top;
	if (CutOut_Chk_On()) {
	    x -= TDVal_GetInteger(fmain->fcutout->dvXz);
	    y -= TDVal_GetInteger(fmain->fcutout->dvYz);
	}
	AzRGBA(fmain, x, y, (TImage*)pxm);
	AzGraf_CutOut_OnMouseMove((TWins*)self, ev);
}

static void AzGraf_OnMouseDown(TScrollWin *self, XButtonEvent *ev) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
  TPixMap *pxm = ALDpic_pxm_get(fpic, ib_paintbase_pxm);
  int x, y;

	if (!pxm) return;
	x = ev->x - fpic->pxm_left;
	y = ev->y - fpic->pxm_top;
	SetQuant_RGB(fpic, x, y, (TImage*)pxm);
	AzGraf_CutOut_OnMouseDown((TWins*)self, ev);
}

static void AzGraf_OnMouseUp(TScrollWin *self, XButtonEvent *ev) {
	AzGraf_CutOut_OnMouseUp((TWins*)self, ev);
}

static void ALDpic_cbWen_PmxResize(TMainForm *fpic) {
  char *pc;
  const char *cpc;
  int w, wq, h, hq;
	if (fpic->cbWen->Checked && fpic->pxm) {
	    SetEnabled((TWins*)fpic->sbZoom, False);

	    cpc = GetCaption_pc((TWins*)fpic->edW);
	    w = wq = ppctoint(&cpc);
	    if (w < 2) w = 2;
	    if (w > 8000) w = 8000;
	    h = hq = 0;
	    while (*cpc == ' ') cpc++;
	    if ((*cpc == 'x') || (*cpc == '*')) {
		cpc++;
		h = hq = ppctoint(&cpc);
		if (h < 2) h = 2;
		if (h > 8000) h = 8000;
	    }
	    if ((w != wq) || (h != hq)) {
  char buf[32];
		pc = buf;
		inttoppc(&pc, w);
		if (h) {
		    *pc = 'x'; pc++;
		    inttoppc(&pc, h);
		}
		SetCaption((TWins*)fpic->edW, buf);
	    }

	    fpic->new_zoom_k = (double)w / (double)fpic->pxm->width;
	    fpic->new_zoom_kh = fpic->new_zoom_k;
	    if (h) fpic->new_zoom_kh = (double)h / (double)fpic->pxm->height;

	    ALDpic_resize_pxm(fpic);
	    Refresh((TWins*)fpic->AzGraf);
	    TScrollBar_SetCount(fpic->sbZoom, fpic->old_w * 100 / fpic->pxm->width);

	} else {
	    SetEnabled((TWins*)fpic->sbZoom, True);

	    fpic->new_zoom_k = fpic->no_cbWen_zoom_k;
	    fpic->new_zoom_kh = fpic->no_cbWen_zoom_kh;

	    ALDpic_resize_pxm(fpic);
	    Refresh((TWins*)fpic->AzGraf);
	    TScrollBar_SetCount(fpic->sbZoom, fpic->old_w * 100 / fpic->pxm->width);
	}
}

static void ALDpic_cbWen_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
	ALDpic_cbWen_PmxResize(fpic);
}

static void ALDpic_cbFastZoom_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
	if (fpic->pxm) ALDpic_cbWen_PmxResize(fpic);

  TpcIni *ini = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR_boolean(ini, APPNAME, "FastZoom", self->Checked);
	TpcIni_SaveAndFree(ini);
}


static void ALDpic_edW_OnKeyDown(TEdit* self, TKeyEvent *KeyEv) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
	if ((KeyEv->keysym == XK_Return) && fpic->cbWen->Checked) {
	    ALDpic_cbWen_PmxResize(fpic);
	}
}


static void ALDpic_pnlColor_OnResize(TWins *self) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
  int H5, y;

	H5 = (self->height-2) / 6;

	y = 1;
	SetTop((TWins*)fpic->pnlBrightness, y);
	SetHeight((TWins*)fpic->pnlBrightness, H5+5);
	y += H5+5;
	SetTop((TWins*)fpic->pnlContrast, y);
	SetHeight((TWins*)fpic->pnlContrast, H5+5);
	y += H5+5;
	SetTop((TWins*)fpic->pnlColorSat, y);
	SetHeight((TWins*)fpic->pnlColorSat, H5+5);
	y += H5+5;
	SetTop((TWins*)fpic->pnlRed, y);
	SetHeight((TWins*)fpic->pnlRed, H5-5);
	y += H5-5;
	SetTop((TWins*)fpic->pnlGreen, y);
	SetHeight((TWins*)fpic->pnlGreen, H5-5);
	y += H5-5;
	SetTop((TWins*)fpic->pnlBlue, y);
	SetHeight((TWins*)fpic->pnlBlue, H5-5);
}

/* ________________________________________________________ */
static void ALDpic_Color_OnChange(TMainForm *fpic) {
	Refresh((TWins*)fpic->AzGraf);
}

static void pnlScBar_SetCaption(TpnlColorScBar *self) {
  int Count = TScrollBar_GetCount(self->sb);
  char s[16];
	inttopc(s, Count);
	SetCaption((TWins*)self, s);
}

static void pnlScBar_cb_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TpnlColorScBar *pnl = self->subClass;
  int Count = TScrollBar_GetCount(pnl->sb);
	if (Count) {
	    if (!pnl->cb->Checked) Count = 0;
	    pnl->val = Count * 256 / 100;
	    ALDpic_Color_OnChange(appDat.obj);
	}
	pnlScBar_SetCaption(pnl);
}

static void pnlScBar_OnCount(TScrollBar* self, int Count) {
  TpnlColorScBar *pnl = self->subClass;
	if (!pnl->cb->Checked) Count = 0;
	pnl->val = Count * 256 / 100;
	if (pnl->cb->Checked) ALDpic_Color_OnChange(appDat.obj);
	pnlScBar_SetCaption(pnl);
}

static void pnlScBar_OnPaint(TpnlColorScBar *self) {
  int y = (self->height / 3) + 2;
	DrawPic(self->WinID, self->pic, 3, y);
}

static void TpnlColorScBar_CreateParams(TWins_Create_struct *wcs) {
  TpnlColorScBar *self = wcs->wins;

	self->subClass = self;
	self->left = 1;
	self->width = 38;
	self->BorderFu = mBorder; // | mBorderL;
	self->CaptionFu = mCaptionPaint | mCaptionBottom | mCaptionLeft;
	vds_SetStr(&self->vds, "0");
	self->OnPaint = pnlScBar_OnPaint;
	SetFont((TWins*)self, "-adobe-helvetica-medium-r-normal-*-10-*-75-75-*-*-*-1");
}

static TpnlColorScBar *pnlScBar_Create(TWins *Parent, TPic *pic) {
  TWins_Create_struct wcs;
  TpnlColorScBar *self;

	TWins_Create_struct_init(&wcs, sizeof(TpnlColorScBar), wtyp_USER_Wins);
	wcs.Parent = Parent;
	wcs.CreateParams = TpnlColorScBar_CreateParams;

	TWins_DefaultConf(&wcs);

	self = wcs.wins;
	Set_Parent((TWins*)self, Parent);
	self->pic = pic;

	self->cb = TChkBox_CreateS((TWins*)self, 1, 1, 17, 17, NULL);
	self->cb->OnClick = pnlScBar_cb_OnClick;

	self->sb = TScrollBarVert_CreateP((TWins*)self);
	self->sb->BorderFu = mBorder | mBorderL;
	self->sb->Align = alRight;
	self->sb->OnCount = pnlScBar_OnCount;
	TScrollBar_SetCountParam(self->sb, 0, 0, 200, -100, sbfu_inv);

	return self;
}

/* ________________________________________________________ */
static void ALDpic_pnlLoadZoomMod_OnPaint(TWins *self) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
  TPic *pic = &fpic->picZoom_wh;
	switch (fpic->LoadZoomMod) {
	case 0: pic = &fpic->picZoom_no; break;
	case 1: pic = &fpic->picZoom_w; break;
	case 2: pic = &fpic->picZoom_h; break;
	}
	DrawPic(self->WinID, pic, 1, 1);
}

static void ALDpic_pnlLoadZoomMod_OnClick(TWins *self, long KeyState, bool DblClick) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
	fpic->LoadZoomMod++;
	if (fpic->LoadZoomMod > 3) fpic->LoadZoomMod = 0;
	Refresh(fpic->pnlLoadZoomMod);
}

static void ALDpic_PanelTop_OnResize(TWins *self) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
  int x = self->width - fpic->pnlLoadZoomMod->width - 1;

	SetLeft(fpic->pnlLoadZoomMod, x);
	x = x - fpic->cbWen->width - 2;
	SetLeft((TWins*)fpic->cbWen, x);
	x = x - fpic->edW->width - 2;
	SetLeft((TWins*)fpic->edW, x);
	x = x - fpic->lblW->width - 1;
	SetLeft(fpic->lblW, x);
	x = x - fpic->cbFastZoom->width - 1;
	SetLeft((TWins*)fpic->cbFastZoom, x);
}


static void ALDpic_PanelBottom_OnResize(TWins *self) {
  TMainForm *fpic = (TMainForm*)self->FormStruct;
  int w = self->width - (fpic->pnlZoomAZ->left + fpic->pnlZoomAZ->width) - 1;
	if (w < 10) w = 10;
	SetWidth(fpic->pnlPath, w);
}

static void ALDpic_OnKeyDown(TMainForm *self, TKeyEvent *KeyEv) {

	if (((KeyEv->keycode == 54)) && (KeyEv->state & 0x4)) {      //Strg_C
	    if ((self->FocusWin == (TWins*)self->edW) &&
		(self->edW->vds.SelA != self->edW->vds.SelE)) return;
	    KeyEv->keycode = 0;
	    KeyEv->keysym = 0;
	    TMainForm_Copy(self);  //in mTools.c
	    return;

	} else if ((KeyEv->keycode == 53) && (KeyEv->state & 0x4)) { //Strg_X
	    if ((self->FocusWin == (TWins*)self->edW) &&
		(self->edW->vds.SelA != self->edW->vds.SelE)) return;
	    KeyEv->keycode = 0;
	    KeyEv->keysym = 0;
	    return;

	} else if ((KeyEv->keycode == 55) && (KeyEv->state & 0x4)) { //Strg_V
      bool en;
#ifdef _OS_WIN32
	    en = Chk_GetSelection(1, CF_DIB);
#else
	    en = Chk_GetSelection(2, XA_PIXMAP, appXDat.XA_IMAGE_PNG);
#endif
	    if (en) {
		TMainForm_Paste(self);  //in mTools.c
	    }
	    if (en || (self->FocusWin != (TWins*)self->edW)) {
		KeyEv->keycode = 0;
		KeyEv->keysym = 0;
	    }
	    return;
	}

	if (self->FocusWin == (TWins*)self->edW) return;
	switch (KeyEv->keysym) {
	case XK_Right:
	    ALDpic_mbPlayrigth_OnClick(self->mbPlayrigth, KeyEv->state, False);
	    KeyEv->keysym = 0; break;
	case XK_Left:
	    ALDpic_mbPlayleft_OnClick(self->mbPlayleft, KeyEv->state, False);
	    KeyEv->keysym = 0;  break;
	}
}

static void ALDpic_OnKeyUp(TMainForm *self, TKeyEvent *KeyEv) {

	if (self->FocusWin == (TWins*)self->edW) return;
	switch (KeyEv->keysym) {
	case XK_Right: KeyEv->keysym = 0; break;
	case XK_Left: KeyEv->keysym = 0;  break;
	}
}

static void FormMain_OnMove(TMainForm *self) {
  TpcIni *ini = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR_int(ini, APPNAME, "Top", self->top);
	TpcIni_WR_int(ini, APPNAME, "Left", self->left);
	TpcIni_SaveAndFree(ini);
}

static void FormMain_OnResize(TMainForm *self) {
	if (!self->ResizeStartMk) return;
  TpcIni *ini = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR_int(ini, APPNAME, "Width", self->width);
	TpcIni_WR_int(ini, APPNAME, "Height", self->height);
	TpcIni_SaveAndFree(ini);
}

// #include <X11/extensions/Xrender.h>

static void ALDpic_OnAppStart() {
  TMainForm *fpic = appDat.obj;

	ALDpic_FileOpen(fpic, &fpic->StartFile);

#if 0
  Display *dpy = appXDat.dpy;

    bool argbVisual = False;
    int screen = DefaultScreen ( dpy );
    Colormap colormap = 0;
    Visual *visual = 0;
    int eventBase, errorBase;

    if ( XRenderQueryExtension ( dpy, &eventBase, &errorBase ) ) {
        int nvi;
        XVisualInfo templ;
        templ.screen  = screen;
        templ.depth   = 32;
        templ.class = TrueColor;
        XVisualInfo *xvi = XGetVisualInfo ( dpy, VisualScreenMask |
                                            VisualDepthMask |
                                            VisualClassMask, &templ, &nvi );
 printf("%s(%d): nvi=%d\n", FL, nvi);
        for ( int i = 0; i < nvi; ++i ) {
            XRenderPictFormat *format = XRenderFindVisualFormat ( dpy,
                                        xvi[i].visual );

 printf("%s(%d): %d %d %d R=0x%x 0x%x G=0x%x 0x%x B=0x%x 0x%x A=0x%x 0x%x\n", FL, (int)format->id, format->type, format->depth, 
		format->direct.red, format->direct.redMask,
		format->direct.green, format->direct.greenMask,
		format->direct.blue, format->direct.blueMask,
		format->direct.alpha, format->direct.alphaMask);

            if ( format->type == PictTypeDirect && format->direct.alphaMask) {
        printf ("You have ARGB support\n");
                visual = xvi[i].visual;
                colormap = XCreateColormap ( dpy, RootWindow ( dpy, screen),
                                             visual, AllocNone );
                argbVisual = True;
                break;
            }
        }
    }
#endif

	AppProcessMessages();


#if 0
	meminfo_get(&fpic->MemInfo);
	Tpid_stat pid_stat;
	char buf[80];
	snprintf(buf, sizeof(buf), "/proc/%d/stat", appDat.self_pid);
	if (get_pid_stat(&pid_stat, buf, PIDSTAT_NO_NAME)) 
					fpic->self_mem = pid_stat.vsize >> 10;

 printf("%s(%d): free=%dkB vsize=%ukB\n", FL, fpic->MemInfo.Free + fpic->MemInfo.Cached, fpic->self_mem);

// printf("%s(%d): %u %u %u\n", FL, fpic->MemInfo.Total, fpic->MemInfo.Free, fpic->MemInfo.Cached);

#endif




	Set_Focus((TWins*)fpic->FileWin);

	int idx;
	if (TFileWin_Get_dfl_by_name(fpic->FileWin, fpic->StartFile, &idx, NULL)) {
	    TFileWin_SetSelect(fpic->FileWin, idx, 0x02);
	}

	Tfree(appDat.appIni);
	appDat.appIni = NULL;
}

/*_____________________________ ErrorMsg _____________________________*/
static void Global_OnFileMsg(TFileMsg *fmsg) {
// file_msg_debug(fmsg, NULL);

	if ((fmsg->art == fmArt_errLogin) || (fmsg->art == fmArt_AlgErr)) {
	    default_Global_OnFileMsg(fmsg);
	}

	return;
}
/*_____________________________ MainForm _____________________________*/
#ifndef _OS_WIN32
static const char aldpic16x16_AHico[] = {
 0x06,0x0F,0x10,0x00,0x10,0x00,0x00,0x00,0x83,0x00,0xC2,0x21,0x84,0x3A,0x85,0x53,
 0xDF,0xFF,0xC6,0x6B,0xCB,0x73,0x48,0x83,0x0D,0x84,0x91,0x8C,0xD4,0x94,0xCB,0xA3,
 0x5A,0xA6,0xCD,0xC4,0x18,0xC6,0x91,0xE5,0xBD,0xC0,0x4B,0xA1,0xF4,0x82,0x8D,0x63,
 0x66,0x90,0xA1,0xB2,0x2F,0x37,0x35,0x16,0xD4,0x7B,0x33,0xE6,0x8A,0xC6,0x13,0xF4,
 0xC9,0xF9,0xFC,0x76,0xAE,0xE8,0x11,0xB4,0xCF,0x34,0x6E,0x2A,0x3A,0xDD,0x04,0x7D,
 0xF6,0xD6,0xF4,0xE3,0x63,0xD5,0x06,0x7D,0x6F,0xDD,0x5C,0xF7,0xA3,0x48,0x04,0x55,
 0xCD,0x54,0x0D,0x56,0xC3,0xCE,0x15,0xA4,0x93,0xE5,0x0C,0x73,0x72,0x50,0x3B,0x23,
 0x92,0x82,0x21,0x34,0xC3,0x8A,0x13,0xAA,0xA9,0x41,0x1B,0x20,0x18,0x93,0xA6,0xEA,
 0x1B,0xBD,0x00};
#endif

const char *Help_Msg = {
  "\n"
  "Usage: ALDpic [options] \"Path\"\n"
  "\n"
  "  -h, --help           Prints out this screen\n"
  "    ALDpic version "VERSION"  (build %d  date %s)\n"
};

const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;

int main (int argc, char **argv) {
  TMainForm *fmain;
  int i, x;
  char *pc, *path, *patternRD, *pattern;
  bool help = False;
  bool err = False;

#ifdef _OS_WIN32
	cmdUTF8korr(&argc, &argv);
#endif

	path = NULL;
	for (i=1; i<argc; i++) {
	    pc = argv[i];
	    if (*pc == '-') {
		pc++;
		if (*pc == 'h') {
			help = True;
		} else if (*pc == '-') {
		    pc++;
		    if (pcncmp(pc, "help", 4)) {
			help = True;
#ifdef Test_Root_WinId
		    } else if (pcncmp(pc, "TestRootWinID", 13)) {
			i++;
			pc = argv[i];
			if (pc && *pc != '-') {
				TestRootWinId = pctoint(pc);
			} else err = True;
#endif
		    } else err = True;
		} else err = True;
	    } else {
		ppcchange(&path, pc);
		cmdline_passwd_korr(pc);
	    }
	}

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(1);
	}

	initXapp(0, argc, argv);
#ifdef _OS_WIN32
	pc = pcadd(appDat.ALDini_dir, "/share/icons/16x16/apps");
	ZsrcToPath(&hZsrc_aldpic16_png, pc, NULL, ZsrcFu_OnlyNotExists);
	xfree(pc);

	pc = pcadd(appDat.ALDini_dir, "/share/icons/32x32/apps");
	ZsrcToPath(&hZsrc_aldpic32_png, pc, NULL, ZsrcFu_OnlyNotExists);
	xfree(pc);

	i = 0;
	pc = pcadd(appDat.ALDini_dir, "/applications");
	if (ZsrcToPath(&hZsrc_aldpic_desktop, pc, NULL, ZsrcFu_OnlyNotExists) == ZsrcOk) i = 1;
	xfree(pc);

	if (i) {
	    pc = pcadd(appDat.ALDini_dir, "/Menu/OpenWith");
	    ZsrcToPath(&hZsrc_aldpic_desktop, pc, NULL, ZsrcFu_OnlyNotExists);
	    xfree(pc);
	}

	ThDirList_Reg(1, &hdirlist_ld_File);
	appDat.inifile_app = pcadd(appDat.ALDini_dir, "\\aldpic.ini");
	ThDirList_ny_Reg(1, &hdirlist_ny_File);
#else
	ThDirList_Reg(2, &hdirlist_ld_File, &hdirlist_ld_ftp);
	appDat.inifile_app = pcadd(appDat.ALDini_dir, "/aldpic.ini");
	ThDirList_ny_Reg(2, &hdirlist_nyi_File, &hdirlist_ny_File);
	ThNetStatLoader_Reg(1, &hNetStatLd_ftp);
	ThNetStreamLoader_Reg(1, &hNetStreamLd_ftp);
#endif
	ThImageRD_Reg(6, &hImageRD_xpm, &hImageRD_png, &hImageRD_jpeg, 
			 &hImageRD_gif, &hImageRD_bmp, &hImageRD_pnm);

	ThImageWR_Reg(4, &hImageWR_xpm, &hImageWR_png, &hImageWR_jpeg,
			 &hImageWR_pnm);

	hGetDeskIcon = &hGetDeskIcon_file;
	hGetFileHeadInfo = &hGetFileHeadInfo_buildin;

	appDat.appIni = TpcIni_Create(appDat.inifile_app);

#ifndef _OS_WIN32
//	InitIconPaths();
#endif
#ifdef USE_DECOR
	appXDat.DefaultButtonDecor = TDecor_button_w3b3_Create();
#endif

	fmain = TForm_CreatParamAlg(NULL, sizeof(TMainForm));
#ifndef _OS_WIN32
	fmain->Icon = GetDeskIcon(ddkt_Application, 16, APPNAME, 
				    gdifu_noZoom, aldpic16x16_AHico);
#endif
	ALDpic_init_icons(fmain);
	appDat.obj = fmain;
	SetCaption((TWins*)fmain, "ALDpic");
	fmain->IconName = APPNAME;

	fmain->top =    TpcIni_RD_int(appDat.appIni, APPNAME, "Top", 0);
	fmain->left =   TpcIni_RD_int(appDat.appIni, APPNAME, "Left", 0);
	fmain->width =  TpcIni_RD_int(appDat.appIni, APPNAME, "Width", 820);
	fmain->height = TpcIni_RD_int(appDat.appIni, APPNAME, "Height", 560);

	fmain->min_width = 580;
	fmain->min_height = 366;
	fmain->FormPosition = poDesigned;
	fmain->OnMove = FormMain_OnMove;
	fmain->OnResize = FormMain_OnResize;
	fmain->OnKeyDown = ALDpic_OnKeyDown;
	fmain->OnKeyUp = ALDpic_OnKeyUp;
	fmain->File_Dlg = Create_File_Dlg((TWins*)fmain);
	file_dlg_InitAzPic(fmain->File_Dlg);

	fmain->PanelTop = TWins_CreateBA((TWins*)fmain, 0, 0, 200, 21, 0, alTop);
	fmain->PanelTop->OnResize = ALDpic_PanelTop_OnResize;
	fmain->PanelMiddle = TWins_CreateBA((TWins*)fmain, 0, 0, 200, 42, 0, alClient);
	fmain->PanelBottom = TWins_CreateBA((TWins*)fmain, 0, 0, 200, 18, 0, alBottom);

// ******* Menu File *******
	MenuFile_Create(fmain);

// ******* Menu View *******
	MenuView_Create(fmain);

	MenuTools_Create(fmain);
// ******* Menu Setting *******
	MenuSetting_Create(fmain);

	MenuHelp_Create(fmain);

// ******* Menu play *******
	MenuButtonPlay_Create(fmain);

// ******* Menu Zoom *******
	MenuButtonZoom_Create(fmain);


	TMenuButton_SetPosition(fmain->mbFile);

	fmain->cbFastZoom = TChkBox_CreateS(fmain->PanelTop, 0, 2, 80, 18, "Fast Zoom");
	fmain->cbFastZoom->OnClick = ALDpic_cbFastZoom_OnClick;
	TChkBox_SetChecked(fmain->cbFastZoom, 
		TpcIni_RD_boolean(appDat.appIni, APPNAME, "FastZoom", False));

	fmain->lblW = TWins_CreateLabel(fmain->PanelTop, 0, 2, 22, 18, "W=", mCaptionRigth, 0);
	fmain->edW = TEdit_CreateS(fmain->PanelTop, 0, 1, 80, 19, "320");
//	SetEnabled((TWins*)fmain->edW, False);
	fmain->edW->OnKeyDown = ALDpic_edW_OnKeyDown;
	fmain->cbWen = TChkBox_CreateS(fmain->PanelTop, 0, 2, 30, 18, "W");
	SetEnabled((TWins*)fmain->cbWen, False);
	fmain->cbWen->OnClick = ALDpic_cbWen_OnClick;


	fmain->pnlLoadZoomMod = TWins_CreateBA(fmain->PanelTop, 0, 2, 18, 18, mBorder, alNone);
	fmain->pnlLoadZoomMod->OnPaint = ALDpic_pnlLoadZoomMod_OnPaint;
	fmain->pnlLoadZoomMod->OnClick = ALDpic_pnlLoadZoomMod_OnClick;
	fmain->LoadZoomMod = 3;

// ******* AzGraf *******
	fmain->AzGraf = TScrollWin_CreateU(fmain->PanelMiddle, 0, 0, 200, 102, 0, alClient, True);
	fmain->AzGraf->OnPaint = ALDpic_AzGraf_OnPaint;
	fmain->AzGraf->OnResize = ALDpic_AzGraf_OnResize;
	fmain->AzGraf->OnMouseMove = AzGraf_OnMouseMove;
	fmain->AzGraf->OnMouseDown = AzGraf_OnMouseDown;
	fmain->AzGraf->OnMouseUp = AzGraf_OnMouseUp;
	fmain->AzGraf->HSB->OnCount = ALDpic_HSB_OnCount;
	fmain->AzGraf->HSB->sbfu = sbfu_inv;
	fmain->AzGraf->VSB->OnCount = ALDpic_VSB_OnCount;
	fmain->AzGraf->VSB->sbfu = sbfu_inv;
	SetColor((TWins*)fmain->AzGraf, 0xe0e0e0);

// ******* pnlColor *******
	fmain->pnlColor = TWins_CreateBA(fmain->PanelMiddle, 0, 0, 40, 42, mBorder | mBorderL, alRight);
	fmain->pnlColor->OnResize = ALDpic_pnlColor_OnResize;
	SetVisible(fmain->pnlColor, (fmain->miViewColor->fu & miChecked));

// ******* FileWin *******
	fmain->PanelList = TWins_CreateBA(fmain->PanelMiddle, 0, 0, 152, 42, 0, alRight);
	fmain->FileWin = TFileWin_CreateA(fmain->PanelList, 0, 0, 200, 200, alClient);
	fmain->FileWin->OnItemEnter = ALDpic_FW_OnItemEnter;
	fmain->FileWin->OnLoadDir = ALDpic_FW_OnLoadDir;

	fmain->FileWin->OnFileMsg = fw_OnFileMsg;

	fmain->FileWin->AzFu = AzFu_Col1_DirHeight32;
	TFileWin_InitAzPic(fmain->FileWin);


	patternRD = TImageIO_GetAllFilePattern(False);
	pattern = GetLinePattern(patternRD);
	TFileWin_set_pattern(fmain->FileWin, pattern, 0);

	TFileWin_SetColsArt(fmain->FileWin, 2);


// ******* --> pnlColor *******
	fmain->pnlBrightness = pnlScBar_Create(fmain->pnlColor, &fmain->picBrightness);
	fmain->pnlContrast = pnlScBar_Create(fmain->pnlColor, &fmain->picContrast);
	fmain->pnlColorSat = pnlScBar_Create(fmain->pnlColor, &fmain->picRGB);
	fmain->pnlRed = pnlScBar_Create(fmain->pnlColor, &fmain->picRed);
	fmain->pnlGreen = pnlScBar_Create(fmain->pnlColor, &fmain->picGreen);
	fmain->pnlBlue = pnlScBar_Create(fmain->pnlColor, &fmain->picBlue);


// ******* --> PanelBottom *******
	x = 0;
	fmain->pnlZoom = TWins_CreateBA(fmain->PanelBottom, x, 1, 280, 17, mBorder | mBorderL, 0);
	fmain->sbZoom = TScrollBarHorz_CreateP(fmain->pnlZoom);
	fmain->sbZoom->Align = alClient;
	fmain->sbZoom->OnCount = ALDpic_sbZoom_OnCount;
 TScrollBar_SetMaxCount(fmain->sbZoom, 200);

	x += 281;
	fmain->pnlZoomAZ = TWins_CreateLabel(fmain->PanelBottom, x, 1, 250, 17, "0 x 0  0%   ld=0ms  z=0ms", 0, mBorder | mBorderL);
	x += 252;
	fmain->pnlPath = TWins_CreateLabel(fmain->PanelBottom, x, 1, 250, 17, NULL, mCaptionLeft, mBorder | mBorderL);
	fmain->PanelBottom->OnResize = ALDpic_PanelBottom_OnResize;

	fmain->FileOpen_Dlg = Create_File_Dlg((TWins*)fmain);
	file_dlg_InitAzPic(fmain->FileOpen_Dlg);

	File_Dlg_AddPattern(fmain->FileOpen_Dlg, pattern, NULL);
	xfree(pattern);
	SetFileDlgSinglePattern(fmain->FileOpen_Dlg, patternRD);
	xfree(patternRD);

	fmain->FileOpen_Dlg->DirMode = True;

	if (!path) {
	    path = TpcIni_RDstrdup(appDat.appIni, APPNAME, "LastPath", NULL);
	    if (path && !file_chk_exist(path, findDIR)) ppcfree(&path);
	}

	pc = path_rel2abs(path, NULL);
	TglobPath *gp = pctogp(pc);
	xfree(pc);
	char *pw = CryptMini16c_enc(gp->passwd);
	if (pw) {
	    xfree(gp->passwd);
	    gp->passwd = pw;
	}

#ifdef _OS_WIN32
	fmain->StartFile = gptopc(gp, 0);
	pc = NT4_GetLongPathName(fmain->StartFile);
	if (pc) {xfree(fmain->StartFile); fmain->StartFile = pc;}
#else
	fmain->StartFile = gptopc(gp, gptfu_PasswdTxt);
#endif /* ! _OS_WIN32 */
	gpFree(gp);

	fmain->ioc_colorRGB = TVideo_IOC_ColorRGB_Create();

	fmain->fAlphaClr = TFormAlphaClr_Create(fmain);
	fmain->fpal = TFormPalette_Create((TWins*)fmain);
	fmain->fcutout = TFormCutOut_Create(fmain);

  long gcm;
  XGCValues gcv;
	Fill0(&gcv, sizeof(XGCValues));
	gcv.function = GXandInverted;
	gcv.line_width = 1;
	gcv.foreground = GetColor_RGB(0x000000);
	gcv.background = GetColor_RGB(0x808080);
	gcv.fill_style = FillOpaqueStippled;
	gcv.stipple = appXDat.Focus_bitmap;
	gcv.subwindow_mode = IncludeInferiors;
	gcm = GCFunction | GCGraphicsExposures | GCLineWidth | GCForeground | GCFillStyle | GCStipple | GCSubwindowMode | GCBackground;
	fmain->GC_Dot = CreateGC(gcm, &gcv);



	appXDat.OnAppStart = ALDpic_OnAppStart;

//	appError_ev = Error_ev;
//	Glob_OnFileMsg = glob_file_msg;
	appError_ev = default_Error_ev;
	Glob_OnFileMsg = Global_OnFileMsg;

	AppRun();
	return 0;
}
