/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldpic.h"

int TFormAlphaClr_GetModus(TFormAlphaClr *self) {
  int result;

	result = 0;
	if (self->cbAlphaClrOn->Checked) result = IS_ALPHA_CLR;
	if (self->cbReplaceBackColor->Checked) result |= IS_ALPHA_CLR_BACKCOLOR;
	if (self->cbRenderColor->Checked) result |= IS_ALPHA_RENDER;
	if ((result & (IS_ALPHA_CLR_BACKCOLOR|IS_ALPHA_RENDER)) == 
		      (IS_ALPHA_CLR_BACKCOLOR|IS_ALPHA_RENDER)) result = 0;
	return result;
}

int TFormAlphaClr_GetModus_IsAkt(TFormAlphaClr *self) {
  TMainForm *fmain;

	fmain = appDat.obj;
	
	if (self && self->Visible && 
	    fmain->pxm && (fmain->pxm->stat & pmxStat_has_alpha))
	{
	    return TFormAlphaClr_GetModus(self);
	}
	return 0;
}

void AlphaClr(TMainForm *fmain) {
  TFormAlphaClr *fAlpha;
  TImage *img;
  int w, h, x, modus;
  dword col, bpp, R, G, B;
  byte *pb;

	if (fmain->AlphaClrOk) return;
	fAlpha = fmain->fAlphaClr;

	modus = TFormAlphaClr_GetModus(fAlpha);
	if (!modus) return;

	img = (TImage*)fmain->vio_pxm;
	if (!img || !img->data) return;
	w = img->width;
	h = img->height;
	if ((w <= 0) || (h <= 0)) return;

	bpp = img->pxm_LD & pxmLD_bppMask;
	if (!(bpp == pxm_4b_24bpp) && !(bpp == pxm_4b_24bpp8a)) return;

	B = TDVal_GetInteger(fAlpha->dvB);
	G = TDVal_GetInteger(fAlpha->dvG);
	R = TDVal_GetInteger(fAlpha->dvR);

	col = B;
	col |= G << 8;
	col |= R << 16;

// printf("%s(%d): \n", FL);

	pb = img->data;
	if (modus == IS_ALPHA_CLR_BACKCOLOR) {
// printf("%s(%d): bBackColor\n", FL);

	    while (h > 0) {
		dword *wx = (dword*)pb;
		x = w;
		while (x > 0) {
		    if (!(*wx & 0xff000000)) *wx = col;
		    wx++; x--;
		}
		pb += img->bytes_per_line;
		h--;
	    }

	} else if (modus == (IS_ALPHA_CLR|IS_ALPHA_CLR_BACKCOLOR)) {

	    col |= 0xff000000;
// printf("%s(%d): bAlphaClr bBackColor 0x%08x\n", FL, col);

	    while (h > 0) {
		dword *wx = (dword*)pb;
		x = w;
		while (x > 0) {
		    if (*wx & 0x80000000) *wx |= 0xff000000; else  *wx = col;
		    wx++; x--;
		}
		pb += img->bytes_per_line;
		h--;
	    }

	} else if (modus == (IS_ALPHA_CLR|IS_ALPHA_RENDER)) {
// printf("%s(%d): bAlphaClr bRenderColor\n", FL);

	    col |= 0xff000000;
	    while (h > 0) {
		dword *wx = (dword*)pb;
		x = w;
		while (x > 0) {
		    if ((*wx & 0xff000000) != 0xff000000) {
		        if (!(*wx & 0xff000000)) {
			    *wx = col;
			} else {
			    dword A, cR, cG, cB;
			    A = (*wx >> 24) + 1;
			    cR = (((*wx >> 16) & 0xff) * A) >> 8;
			    cG = (((*wx >> 8) & 0xff) * A) >> 8;
			    cB = ((*wx & 0xff) * A) >> 8;
			    A = 0x100 - A;
			    cR += (R * A) >> 8;
			    cG += (G * A) >> 8;
			    cB += (B * A) >> 8;
			    *wx = 0xff000000 + (cR << 16) + (cG << 8) + cB;
			}
		    }
		    wx++; x--;
		}
		pb += img->bytes_per_line;
		h--;
	    }

	} else if (modus == IS_ALPHA_RENDER) {
// printf("%s(%d): bRenderColor\n", FL);

	    while (h > 0) {
		dword *wx = (dword*)pb;
		x = w;
		while (x > 0) {
		    if ((*wx & 0xff000000) != 0xff000000) {
		        if (!(*wx & 0x80000000)) {
			    *wx &= 0x00ffffff;
			} else {
			    dword A, cR, cG, cB;
			    A = (*wx >> 24) + 1;
			    cR = (((*wx >> 16) & 0xff) * A) >> 8;
			    cG = (((*wx >> 8) & 0xff) * A) >> 8;
			    cB = ((*wx & 0xff) * A) >> 8;
			    A = 0x100 - A;
			    cR += (R * A) >> 8;
			    cG += (G * A) >> 8;
			    cB += (B * A) >> 8;
			    *wx = 0xff000000 + (cR << 16) + (cG << 8) + cB;
			}
		    }
		    wx++; x--;
		}
		pb += img->bytes_per_line;
		h--;
	    }

	} else {
// printf("%s(%d): bAlphaClr\n", FL);

	    while (h > 0) {
		dword *wx = (dword*)pb;
		x = w;
		while (x > 0) {
		    *wx = *wx | 0xff000000;
		    wx++; x--;
		}
		pb += img->bytes_per_line;
		h--;
	    }
	}

	img->stat &= ~pmxStat_XPixAkt;
	fmain->RePaint_pxm = True;
	fmain->AlphaClrOk = true;
}

static void cbAlphaClrOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TMainForm *fmain = appDat.obj;
  TFormAlphaClr *fAlpha;

	fmain = appDat.obj;
	fAlpha = (TFormAlphaClr*)self->FormStruct;
	TChkBox_SetChecked(fAlpha->cbRenderColor, 
			   self->Checked && (fAlpha->onMk & 0x01));
	Refresh((TWins*)fmain->AzGraf);
}

static void cbReplaceBackColor_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TMainForm *fmain;
  TFormAlphaClr *fAlpha;

	fmain = appDat.obj;
	fAlpha = (TFormAlphaClr*)self->FormStruct;
	if (self->Checked) TChkBox_SetChecked(fAlpha->cbRenderColor, false); 
//	if (fAlpha->cbAlphaClrOn->Checked) 
	Refresh((TWins*)fmain->AzGraf);
	fAlpha->onMk &= ~0x01;
}

static void cbRenderColor_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TMainForm *fmain;
  TFormAlphaClr *fAlpha;

	fmain = appDat.obj;
	fAlpha = (TFormAlphaClr*)self->FormStruct;
	if (self->Checked) TChkBox_SetChecked(fAlpha->cbReplaceBackColor, false); 
	if (fAlpha->cbAlphaClrOn->Checked) {
	    fAlpha->onMk &= ~0x01;
	    if (self->Checked) fAlpha->onMk |= 0x01;
	}
	Refresh((TWins*)fmain->AzGraf);
}

static void dvRGB_OnValChange(TDVal *self, double val) {
  TMainForm *fmain;
  TFormAlphaClr *fAlpha;

	fmain = appDat.obj;
	fAlpha = GetParentForm((TWins*)self);
	if (fAlpha->cbAlphaClrOn->Checked ||
	    fAlpha->cbRenderColor->Checked) Refresh((TWins*)fmain->AzGraf);
}

static void TFormAlphaClr_OnMove(TFormAlphaClr *self) {
  TpcIni *ti = TpcIni_Create(appDat.inifile_app);
	TpcIni_WR_int(ti, "FormAlphaClr", "Top", self->top);
	TpcIni_WR_int(ti, "FormAlphaClr", "Left", self->left);
	TpcIni_SaveAndFree(ti);
}

static void TFormAlphaClr_OnVisible(TFormAlphaClr *self, bool Visible) {
  TMainForm *fmain;

	fmain = appDat.obj;
	if (self->cbAlphaClrOn->Checked ||
	    self->cbRenderColor->Checked) Refresh((TWins*)fmain->AzGraf);
}

dword TFormAlphaClr_GetColor(TFormAlphaClr *self) {
  dword result;

	result  = TDVal_GetInteger(self->dvB);
	result |= TDVal_GetInteger(self->dvG) << 8;
	result |= TDVal_GetInteger(self->dvR) << 16;
	return result;
}

/* ______________________ Create _____________________________________ */
TFormAlphaClr *TFormAlphaClr_Create(TMainForm *fmain) {
  TFormAlphaClr *self;
  TpcIni *ti = appDat.appIni;

	if (!ti) ti = TpcIni_Create(appDat.inifile_app);

	self = TForm_CreatParamAlg((TWins*)fmain, sizeof(TFormAlphaClr));
	self->FormZStyle = fszTransient;
	self->FormPosition = poDesigned;
	SetCaption((TWins*)self, "ALDpic AlphaClr");
	self->Visible = False;

	self->top = TpcIni_RD_int(ti, "FormAlphaClr", "Top", 0);
	self->left = TpcIni_RD_int(ti, "FormAlphaClr", "Left", 0);
	self->max_width = self->min_width = self->width = 200;
	self->max_height = self->min_height = self->height = 88;

	self->OnMove = TFormAlphaClr_OnMove;
	self->OnVisible = TFormAlphaClr_OnVisible;

	self->cbAlphaClrOn =   TChkBox_CreateS((TWins*)self, 2,  2, 180, 16, "Alpha Clr On");
	self->cbAlphaClrOn->OnClick = cbAlphaClrOn_OnClick;
	self->cbReplaceBackColor = TChkBox_CreateS((TWins*)self, 2, 20, 180, 16, "Replace BackColor");
	self->cbReplaceBackColor->OnClick = cbReplaceBackColor_OnClick;
	self->cbRenderColor = TChkBox_CreateS((TWins*)self, 2, 40, 180, 16,  "Alpha Render");
	self->cbRenderColor->OnClick = cbRenderColor_OnClick;

	self->lblR = TWins_CreateLabel((TWins*)self,   2, 61, 18, 16, "R=", mCaptionRigth, 0);
	self->lblG = TWins_CreateLabel((TWins*)self,  62, 61, 18, 16, "G=", mCaptionRigth, 0);
	self->lblB = TWins_CreateLabel((TWins*)self, 122, 61, 18, 16, "B=", mCaptionRigth, 0);
	self->dvR = TDVal_CreateSV4((TWins*)self, 22,  60, 36, 18, 0, 0, 255, 0);
	self->dvR->OnValChange = dvRGB_OnValChange;
//	self->dvR->OnMouseDown = dvRGB_OnMouseDown;
//	self->dvR->OnMouseMove = dvRGB_OnMouseMove;
	self->dvG = TDVal_CreateSV4((TWins*)self, 82, 60, 36, 18, 0, 0, 255, 0);
	self->dvG->OnValChange = dvRGB_OnValChange;
//	self->dvG->OnMouseDown = dvRGB_OnMouseDown;
//	self->dvG->OnMouseMove = dvRGB_OnMouseMove;
	self->dvB = TDVal_CreateSV4((TWins*)self, 142, 60, 36, 18, 0, 0, 255, 0);
	self->dvB->OnValChange = dvRGB_OnValChange;
//	self->dvB->OnMouseDown = dvRGB_OnMouseDown;
//	self->dvB->OnMouseMove = dvRGB_OnMouseMove;

	if (!appDat.appIni) Tfree(ti);

	return self;
}
