/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldpic.h"

static const char cutout_AHico[] = {
 0x06,0x03,0x10,0x00,0x10,0x00,0x00,0x00,0x4D,0x00,0x00,0x00,0x10,0x84,0xDF,0xFF,
 0xBD,0xF8,0x06,0x22,0x20,0x02,0x22,0x8D,0x14,0x22,0x20,0x02,0x22,0x8D,0x05,0xBA,
 0x0D,0x14,0x02,0xB1,0xD3,0x04,0x25,0xC4,0x4E,0x93,0x14,0x12,0x3B,0x4D,0x58,0x60,
 0xEC,0x4C,0x50,0x08,0xC4,0x4E,0x13,0x94,0x10,0xFB,0x93,0x14,0xE2,0xE6,0x60,0x81,
 0x08,0x88,0x80,0x48,0x14,0x02,0x22,0x20,0x22,0x82,0x6F,0xBD,0x00};

static const char swapX_AHico[] = {
 0x06,0x03,0x10,0x00,0x10,0x00,0x00,0x00,0x40,0x00,0x04,0x41,0x80,0x61,0x0C,0x83,
 0x4B,0x28,0x02,0xBA,0x38,0x82,0x68,0x0B,0x98,0xA0,0x2D,0x20,0x92,0xB6,0x88,0xA0,
 0xB2,0x80,0xBC,0x80,0xB5,0xE0,0x60,0xED,0x18,0x58,0x1B,0x02,0xF2,0x02,0xF6,0x22,
 0x82,0xDA,0x02,0x22,0x69,0x0B,0x98,0xA0,0x2E,0x8E,0x20,0xFA,0x12,0x8A,0x80,0x09};

static const char swapY_AHico[] = {
 0x01,0x03,0x10,0x00,0x10,0x00,0x00,0x00,0x44,0x00,0x80,0x61,0x0C,0x83,0x04,0x41,
 0x80,0x01,0x80,0x01,0x80,0x01,0x80,0x01,0x9F,0xF9,0x9E,0x79,0x9C,0x39,0x9A,0x59,
 0x91,0x89,0x81,0x81,0x82,0x41,0x9C,0x39,0xE0,0x07,0x80,0x01,0x80,0x01,0x80,0x01,
 0x99,0x99,0x00,0x24,0x00,0x00,0x09,0x00,0x24,0x00,0x24,0x00,0x09,0x24,0x24,0x40,
 0x02,0x00,0x90,0x99};

static const char rot90_AHico[] = {
 0x06,0x01,0x10,0x00,0x10,0x00,0x00,0x00,0x1D,0x00,0x00,0x00,0x57,0x0B,0xD3,0x8A,
 0x6D,0x41,0x5E,0xF0,0x16,0xB4,0xA5,0x6B,0xE1,0x5B,0xD0,0xEE,0x05};

static const char rot180_AHico[] = {
 0x06,0x01,0x10,0x00,0x10,0x00,0x00,0x00,0x27,0x00,0x00,0x00,0x57,0x0B,0xD3,0x8A,
 0x6D,0x41,0x5E,0xF0,0x16,0xE4,0x05,0x79,0xC1,0x5E,0x90,0x17,0x4C,0x90,0x56,0x33,
 0xD7,0xD2,0xB6,0xE2,0x5B,0x90,0x5B};

static const char rot270_AHico[] = {
 0x06,0x01,0x10,0x00,0x10,0x00,0x00,0x00,0x2A,0x00,0x00,0x00,0x57,0x0B,0xD3,0x8A,
 0x6D,0x41,0x5E,0xB0,0x05,0x15,0x6C,0xE1,0x04,0x59,0x1A,0x41,0x17,0x5C,0x90,0x05,
 0x59,0x90,0x05,0x5B,0x31,0x8A,0x69,0x61,0xEE,0x05};

/* ------------------------------------------------- */

#define TAG_rotX  2
#define TAG_swap  3


void TMainForm_Copy(TMainForm *self) {
  TPixMap *pxm = ALDpic_pxm_get(self, ib_paintbase_pxm);
	SetSelection_TPixMap(pxm);
}

void TMainForm_Paste(TMainForm *self) {
  TScrollWin *sw = (TScrollWin*)self->AzGraf;
  TPixMap *pxm;

	pxm = GetSelection_TPixMap();
	if (pxm) {
	    CutOut_Clr();
	    ALDpic_pxm_free(self, ib_all_pxm);
	    self->pxm = pxm;

      char line[1024];
      char *fn = GetFilePath(self->pxmFile);
	    ppcadd(&fn, "default.png");
	    xfree(self->pxmFile);
	    self->pxmFile = fn;

	    snprintf(line, sizeof(line), "ALDpic  %s     %dx%d", fn, self->pxm->width, self->pxm->height);
	    SetCaption((TWins*)self, line);

	    self->new_zoom_k = 1;
	    self->new_zoom_kh = 1;

	int width = self->AzGraf->width;
	int height = self->AzGraf->height;

	    self->cHoff = width >> 1;
	    self->cVoff = height >> 1;

	    TScrollBar_SetCountParam(sw->HSB, 0, width, self->pxm->width + width,
				-(self->pxm->width>>1), sbfu_inv);
	    self->HSB_Count = 0;

	    TScrollBar_SetCountParam(sw->VSB, 0, height, self->pxm->height + height,
				-(self->pxm->height>>1), sbfu_inv);
	    self->VSB_Count = 0;

	    ALDpic_resize_pxm(self);

	    self->sbZoom->Count = self->old_w * 100 / self->pxm->width;
	    TScrollBar_Refresh(self->sbZoom);

	    Refresh((TWins*)self->AzGraf);
	}
}


void menuTools_clr_rot_swap(TMainForm *fmain) {
TMenuItem *mi;

	if (!fmain->menuTools) return;
	mi = fmain->menuTools->MenuItem;
	while (mi) {
	    if ((mi->tag == TAG_rotX) || 
		(mi->tag == TAG_swap)) mi->fu &= ~miChecked;
	    mi = mi->next;
	}
}

/* ______________________ Menu Ev ____________________________________ */
static void menuTools_OnMenuOpen(TMenu *self) {
  TMainForm *fmain = (TMainForm*)self->FormStruct;
  bool en;
#ifdef _OS_WIN32
	en = Chk_GetSelection(1, CF_DIB);
#else
	en = Chk_GetSelection(2, XA_PIXMAP, appXDat.XA_IMAGE_PNG);
#endif
	fmain->miPaste->fu |= miNotEnabled;
	if (en) fmain->miPaste->fu &= ~miNotEnabled;
}

static void mPalette_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TMainForm *fmain = (TMainForm*)self->FormStruct;
	TForm_Show((TForm*)fmain->fpal);
}

static void miAlphaCLR_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TMainForm *fmain = (TMainForm*)self->FormStruct;
	TForm_Show((TForm*)fmain->fAlphaClr);
}

static void mCutOut_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TMainForm *fmain = (TMainForm*)self->FormStruct;
	TForm_Show((TForm*)fmain->fcutout);
}

static void mSwap_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TMainForm *fmain = (TMainForm*)self->FormStruct;
	ALDpic_resize_pxm(fmain);
	Refresh((TWins*)fmain->AzGraf);
}

static void mRotat_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  TMainForm *fmain = (TMainForm*)self->FormStruct;
  TMenuItem *mi = self->MenuItem;
	while (mi) {
	    if ((mi->tag == TAG_rotX) && (mi != smi)) mi->fu &= ~miChecked;
	    mi = mi->next;
	}
	ALDpic_resize_pxm(fmain);
	Refresh((TWins*)fmain->AzGraf);
}

static void mPalette_OnCopy(TMenu *self, long KeyState, TMenuItem *smi) {
	TMainForm_Copy((TMainForm*)self->FormStruct);
}

static void mPalette_OnPaste(TMenu *self, long KeyState, TMenuItem *smi) {
	TMainForm_Paste((TMainForm*)self->FormStruct);
}

/* ______________________ Create _____________________________________ */
void MenuTools_Create(TMainForm *fmain) {
  TMenuItem *mi;

	fmain->mbTools = TMenuButton_Create(fmain->PanelTop, fmain->mbFile);
	SetCaption((TWins*)fmain->mbTools, "Tools");
	Set_Parent((TWins*)fmain->mbTools, fmain->PanelTop);

	fmain->menuTools = TMenu_Create((TWins*)fmain);
	fmain->menuTools->OnMenuOpen = menuTools_OnMenuOpen;
	fmain->mbTools->menu = fmain->menuTools;

	mi = TMenu_AddItem(fmain->menuTools, "Cut Out", 
		0, mCutOut_OnClick);
	mi->pic = CreateAHicoPic(cutout_AHico, NULL);

	TMenu_AddItem(fmain->menuTools, "Palette", 
		0, mPalette_OnClick);
	fmain->miToolsMaskCLR = TMenu_AddItem(fmain->menuTools, "Alpha CLR", 
		0, miAlphaCLR_OnClick);


	TMenu_AddItem(fmain->menuTools, "-", 0, NULL);

	fmain->miRotat90 = TMenu_AddItem(fmain->menuTools, "Rotat 90", 
				miAutoToggle, mRotat_OnClick);
	fmain->miRotat90->tag = TAG_rotX;
	fmain->miRotat90->pic = CreateAHicoPic(rot90_AHico, NULL);

	fmain->miRotat180 = TMenu_AddItem(fmain->menuTools, "Rotat 180", 
				miAutoToggle, mRotat_OnClick);
	fmain->miRotat180->tag = TAG_rotX;
	fmain->miRotat180->pic = CreateAHicoPic(rot180_AHico, NULL);

	fmain->miRotat270 = TMenu_AddItem(fmain->menuTools, "Rotat 270", 
				miAutoToggle, mRotat_OnClick);
	fmain->miRotat270->tag = TAG_rotX;
	fmain->miRotat270->pic = CreateAHicoPic(rot270_AHico, NULL);

	fmain->miSwapX = TMenu_AddItem(fmain->menuTools, "Swap X", 
				miAutoToggle, mSwap_OnClick);
	fmain->miSwapX->tag = TAG_swap;
	fmain->miSwapX->pic = CreateAHicoPic(swapX_AHico, NULL);
	fmain->miSwapY = TMenu_AddItem(fmain->menuTools, "Swap Y", 
				miAutoToggle, mSwap_OnClick);
	fmain->miSwapX->tag = TAG_swap;
	fmain->miSwapY->pic = CreateAHicoPic(swapY_AHico, NULL);


	TMenu_AddItem(fmain->menuTools, "-", 0, NULL);

	fmain->miCopy = TMenu_AddItem(fmain->menuTools, "Copy", 
				0, mPalette_OnCopy);
	fmain->miCopy->s2 = pcdup("Strg c");

	fmain->miPaste = TMenu_AddItem(fmain->menuTools, "Paste", 
				0, mPalette_OnPaste);
	fmain->miPaste->s2 = pcdup("Strg v");
}
