/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "Form_RLC.h"
#include "ElGraf.h"
#include <math.h>

void Form_C_math(TFormC *fC, float w) {
  float uF = fC->C->dval * 0.000001;

	if (fC->cbTC->Checked) uF = fC->TC->dval * 0.000001;

	SetUp_C(&fC->rlc, w, uF);
	if (fC->cbOptOn->Checked) {
	    if (fC->cbLs->Checked && (fC->Ls->dval > 0))
		RLC_Add_Ser_L(&fC->rlc, w, fC->Ls->dval);
	    if (fC->cbRs->Checked && (fC->Rs->dval > 0))
		RLC_Add_Ser_R(&fC->rlc, w, fC->Rs->dval);
	    if (fC->cbRp->Checked && (fC->Rp->dval > 0))
		RLC_Add_Par_R(&fC->rlc, w, fC->Rp->dval * 1000);
	}
}

static void fC_math_loc_Az(TFormC *fC) {
  float w = fC->MHz->dval * M_PI * 2;
	Form_C_math(fC, w);
	if (fC->rlc.Z >= 1000) fC->Z->nk = 1;
	if (fC->rlc.Z >= 10000) fC->Z->nk = 0;
	TDVal_Set_Value(fC->Z, fC->rlc.Z);
}

static void fC_Set_Stat_exDVal(TFormC *fC) {
  int Color = 0xffffff;
	if (!fC->exDVal) return;
	if (fC->cbTC->Checked) {
	    fC->exDVal->ReadOnly = (fC->sbTC->Count != 0);
	    TDVal_Set_Value(fC->exDVal, fC->TC->dval);
	} else {
	    fC->exDVal->ReadOnly = False;
	    TDVal_Set_Value(fC->exDVal, fC->C->dval);
	}
	if (fC->exDVal->ReadOnly) Color = 0xe0e0e0;
	SetColor((TWins*)fC->exDVal, Color);
}

static void fC_math_loc(TFormC *fC) {
	fC_math_loc_Az(fC);
	fC_Set_Stat_exDVal(fC);
	if (fC->OnChange) fC->OnChange((TFormRLC*)fC);
}

static void fC_SaveDaten(TFormC *fC) {
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
  char *pc;
  int OnOff = 0;
  char buf[512];

	if (fC->cbOptOn->Checked) OnOff |= 0x01;
	if (fC->cbRs->Checked) OnOff |= 0x02;
	if (fC->cbRp->Checked) OnOff |= 0x04;
	if (fC->cbLs->Checked) OnOff |= 0x08;
	if (fC->cbTC->Checked) OnOff |= 0x10;
	if (fC->cbTCsync->Checked) OnOff |= 0x20;

	pc = buf;
	doubletoppc(&pc, fC->MHz->dval, fC->MHz->nk | kSPACE_ADD);
	doubletoppc(&pc, fC->Rs->dval, fC->Rs->nk | kSPACE_ADD);
	doubletoppc(&pc, fC->Rp->dval, fC->Rp->nk | kSPACE_ADD);
	doubletoppc(&pc, fC->Ls->dval, fC->Ls->nk | kSPACE_ADD);
	doubletoppc(&pc, fC->TCmin->dval, fC->TCmin->nk | kSPACE_ADD);
	ppcmove(&pc, "0 ");
	inttoppc(&pc, OnOff);

	TpcIni_WR(pcIni, fC->Gruppe, fC->sOpt, buf);
	TpcIni_WR_double(pcIni, fC->Gruppe, fC->sC, fC->C->dval, fC->C->nk);
	TpcIni_SaveAndFree(pcIni);
}

static void fC_cb_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFormC *fC = (TFormC*)self->FormStruct;
	fC_math_loc(fC);
	fC_SaveDaten(fC);
}

static void fC_TC_SetVal(TFormC *fC, int Count) {
  float C = fC->C->dval - (fC->C->dval - fC->TCmin->dval) * 
			    (float)(Count) / (float)(fC->sbTC->MaxCount);
	if (C < 0.01) C = 0.01;
	fC->TC->nk = 1;
	if (C < 20) fC->TC->nk = 2;
	TDVal_Set_Value(fC->TC, C);
}

static void fC_DVal_OnValChange(TDVal *dv, double val) {
  TFormC *fC = (TFormC*)dv->FormStruct;
	fC_TC_SetVal(fC, TScrollBar_GetCount(fC->sbTC));
	fC_math_loc(fC);
	fC_SaveDaten(fC);
}

static void fC_sbTC_OnCount(TScrollBar* sb, int Count) {
  TFormC *fC = (TFormC*)sb->FormStruct;
	fC_TC_SetVal(fC, Count);
	fC->val_sync = Count;
	if (fC->OnSync && fC->cbTCsync->Checked) fC->OnSync((TFormRLC*)fC, Count);
	if (fC->cbTC->Checked) fC_math_loc(fC);
}

void Form_C_sync(TFormRLC *fRLC, TFormC *fC, int val) {
	if (fRLC == (TFormRLC*)fC) return;
	fC->val_sync = val;
	if (fC->cbTCsync->Checked) {
	    TScrollBar_SetCount(fC->sbTC, val);
	    fC_TC_SetVal(fC, val);
	    fC_math_loc_Az(fC);
	    fC_Set_Stat_exDVal(fC);
	}
}

static void Form_C_SetFreq(TFormRLC *fRLC, float val) {
	TDVal_Set_Value(((TFormC*)fRLC)->MHz, val);
	fC_math_loc_Az((TFormC*)fRLC);
}

static void fC_pnlC_OnPaint(TWins *self) {
  TFormC *fC = (TFormC*)self->FormStruct;
  int x = 3;
  int y = 25;
  int y2 = y+20;

	DrawPic(self->WinID, get_Pic_pto(), x, y+2);
	DrawLineRGB(self->WinID, 0, x+7, y+5, x+9, y+5);
	DrawPic(self->WinID, get_Pic_pt(), x+10, y+3);
	DrawLineRGB(self->WinID, 0, x+15, y+5, x+19, y+5);
	DrawPic(self->WinID, get_Pic_Rh(), x+20, y);
	  DrawPic(self->WinID, get_Pic_cRs(), x+25, y+10);
	DrawLineRGB(self->WinID, 0, x+40, y+5, x+49, y+5);
	DrawPic(self->WinID, get_Pic_Lh(), x+50, y);
	  DrawPic(self->WinID, get_Pic_cLs(), x+55, y+10);
	DrawLineRGB(self->WinID, 0, x+70, y+5, x+79, y+5);
	DrawPic(self->WinID, get_Pic_Ch(), x+80, y);
	  DrawPic(self->WinID, get_Pic_cC(), x+94, y+10);
	DrawLineRGB(self->WinID, 0, x+100, y+5, x+109, y+5);
	DrawPic(self->WinID, get_Pic_pt(), x+110, y+3);
	DrawLineRGB(self->WinID, 0, x+115, y+5, x+119, y+5);
	DrawPic(self->WinID, get_Pic_pto(), x+120, y+2);

	DrawLineRGB(self->WinID, 0, x+12, y+8, x+12, y2+5);
	DrawLineRGB(self->WinID, 0, x+12, y2+5, x+49, y2+5);
	DrawPic(self->WinID, get_Pic_Rh(), x+50, y2);
	  DrawPic(self->WinID, get_Pic_cRp(), x+55, y2+10);
	DrawLineRGB(self->WinID, 0, x+70, y2+5, x+111, y2+5);
	DrawLineRGB(self->WinID, 0, x+111, y+8, x+111, y2+5);

	DrawPic(self->WinID, get_Pic_cC(), fC->C->left-11, fC->C->top+7);
	DrawPic(self->WinID, get_Pic_cpF(), fC->C->left+fC->C->width+3, fC->C->top+7);
}


static void fC_pnlOpt_OnPaint(TWins *self) {
  TFormC *fC = (TFormC*)self->FormStruct;

	DrawPic(self->WinID, get_Pic_cRs(), fC->Rs->left-15, fC->Rs->top+5);
	DrawPic(self->WinID, get_Pic_cOhm(), fC->Rs->left+fC->Rs->width+3, fC->Rs->top+5);
	DrawPic(self->WinID, get_Pic_cRp(), fC->Rp->left-15, fC->Rp->top+5);
	DrawPic(self->WinID, get_Pic_ckOhm(), fC->Rp->left+fC->Rp->width+3, fC->Rp->top+5);
	DrawPic(self->WinID, get_Pic_cLs(), fC->Ls->left-15, fC->Ls->top+5);
	DrawPic(self->WinID, get_Pic_cuH(), fC->Ls->left+fC->Ls->width+3, fC->Ls->top+5);
}

static void fC_pnlZ_OnPaint(TWins *self) {
  TFormC *fC = (TFormC*)self->FormStruct;

	DrawPic(self->WinID, get_Pic_cf(), fC->MHz->left-8, fC->MHz->top+5);
	DrawPic(self->WinID, get_Pic_cMHz(), fC->MHz->left+fC->MHz->width+3, fC->MHz->top+5);
	DrawPic(self->WinID, get_Pic_cZ(), fC->Z->left-8, fC->Z->top+5);
	DrawPic(self->WinID, get_Pic_cOhm(), fC->Z->left+fC->Z->width+3, fC->Z->top+5);

	DrawPic(self->WinID, get_Pic_cpF(), fC->TC->left+fC->TC->width+3, fC->TC->top+5);
	DrawPic(self->WinID, get_Pic_cpF(), fC->TCmin->left+fC->TCmin->width+3, fC->TCmin->top+5);
}
static void fC_pnlZ_OnResize(TWins *self) {
  TFormC *fC = (TFormC*)self->FormStruct;
	SetTop((TWins*)fC->TCmin, self->height-20);
	SetTop((TWins*)fC->TC, self->height-38);
	SetTop((TWins*)fC->cbTCsync, self->height-55);
}

static void Form_C_OnMove(TFormC *self) {
	TFormRLC_Save((TFormRLC*)self);
}

TFormC *Create_Form_C(TWins *parent, const char *Caption, const char *Gruppe, 
			const char *sC, TDVal *exDVal, float val) {
  TFormC *fC;
  TpcIni *pcIni;
  const char *sopt;

	pcIni = get_appIni();

	fC = TForm_CreatParamAlg(parent, sizeof(TFormC));
	fC->wtyp = wtyp_TFormC;
//	fC->Visible = False;
	fC->FormZStyle = fszTransient;
	fC->FormPosition = poDesigned;
	fC->Gruppe = Gruppe;
	fC->sC = sC;
	fC->sOpt = pcadd(sC, "opt");
	fC->sXY = pcadd(sC, "xy");
	fC->Save = TFormRLC_Save;
	fC->SetFreq = Form_C_SetFreq;

	SetCaption((TWins*)fC, Caption);
	fC->width = 226;
	fC->height = 143;
	fC->min_width = fC->width;
	fC->min_height = fC->height;
	fC->max_width = fC->width;
	fC->max_height = fC->height;
	sopt = TpcIni_RD_pc(pcIni, Gruppe, fC->sXY, "0 0");
	fC->left = ppctoint(&sopt);
	fC->top = ppctoint(&sopt);
	fC->tag = (ppctoint(&sopt) != 0);
	fC->Visible = False; //fC->tag;
//	fC->OnCanClose = Form_C_OnCanClose;
	fC->OnMove = Form_C_OnMove;
//	fC->SetVisible = TForm_C_SetVisible;
	fC->exDVal = exDVal;

	fC->pnl1 = TWins_CreateBA((TWins*)fC, 0, 0, 134, 21, 0, alLeft);
	fC->pnlC = TWins_CreateBA(fC->pnl1, 0, 0, 134, 67, mBorder | mBorderL, alTop);
	fC->pnlOpt = TWins_CreateBA(fC->pnl1, 0, 20, 130, 80, mBorder | mBorderL, alClient);
	fC->pnlZ = TWins_CreateBA((TWins*)fC, 136, 0, 130, 80, mBorder | mBorderL, alClient);


/* Panel C */
	fC->pnlC->OnPaint = fC_pnlC_OnPaint;
	fC->C = TDVal_CreateS(fC->pnlC, 22, 2, 70, 21);
	fC->C->nk = 1;
	fC->C->dv_fu = tdv_fu_LostFocusEnter;
	fC->C->OnValChange = fC_DVal_OnValChange;
	SetFont((TWins*)fC->C, "-*-helvetica-bold-r-*-*-14-*-75-75-*-*-*-*");
	if (fC->exDVal) fC->exDVal->owner = (TWins*)fC->C;

/* Panel Optionen */
	fC->pnlOpt->OnPaint = fC_pnlOpt_OnPaint;
	fC->cbOptOn = TChkBox_CreateS(fC->pnlOpt, 2, 2, 56, 16, "Opt On");
	fC->cbOptOn->OnClick = fC_cb_OnClick;
	fC->cbRs = TChkBox_CreateS(fC->pnlOpt, 2, 20, 16, 16, NULL);
	fC->cbRs->OnClick = fC_cb_OnClick;
	fC->cbRp = TChkBox_CreateS(fC->pnlOpt, 2, 39, 16, 16, NULL);
	fC->cbRp->OnClick = fC_cb_OnClick;
	fC->cbLs = TChkBox_CreateS(fC->pnlOpt, 2, 58, 16, 16, NULL);
	fC->cbLs->OnClick = fC_cb_OnClick;
	fC->Rs = TDVal_CreateS(fC->pnlOpt, 36, 18, 60, 18);
	fC->Rs->nk = 2;
	fC->Rs->dv_fu = tdv_fu_LostFocusEnter;
	fC->Rs->OnValChange = fC_DVal_OnValChange;
	fC->Rp = TDVal_CreateS(fC->pnlOpt, 36, 37, 60, 18);
	fC->Rp->nk = 3;
	fC->Rp->dv_fu = tdv_fu_LostFocusEnter;
	fC->Rp->OnValChange = fC_DVal_OnValChange;
	fC->Ls = TDVal_CreateS(fC->pnlOpt, 36, 56, 60, 18);
	fC->Ls->nk = 4;
	fC->Ls->dv_fu = tdv_fu_LostFocusEnter;
	fC->Ls->OnValChange = fC_DVal_OnValChange;
	fC->cbTC = TChkBox_CreateS(fC->pnlOpt, 86, 2, 32, 16, "TC");
	fC->cbTC->OnClick = fC_cb_OnClick;
	fC->sbTC = TScrollBarVert_CreateP(fC->pnlOpt);
	fC->sbTC->Align = alRight;
	fC->sbTC->OnCount = fC_sbTC_OnCount;
	TScrollBar_SetMaxCount(fC->sbTC, 1000);

/* Panel Z */
	fC->pnlZ->OnPaint = fC_pnlZ_OnPaint;
	fC->pnlZ->OnResize = fC_pnlZ_OnResize;
	fC->MHz = TDVal_CreateS(fC->pnlZ, 12, 3, 56, 18);
	fC->MHz->nk = 1;
	fC->MHz->dv_fu = tdv_fu_LostFocusEnter;
	fC->MHz->OnValChange = fC_DVal_OnValChange;

	fC->Z = TDVal_CreateS(fC->pnlZ, 12, 22, 56, 18);
	fC->Z->nk = 2;
	fC->Z->ReadOnly = True;
	fC->Z->KeyFu |= mNoKeyFocus;
	SetColor((TWins*)fC->Z, 0xe8e8e8);

	fC->cbTCsync = TChkBox_CreateS(fC->pnlZ, 2, 101, 60, 16, "TCsync");
	fC->cbTCsync->OnClick = fC_cb_OnClick;
	fC->TC = TDVal_CreateS(fC->pnlZ, 2, 121, 46, 18);
	fC->TC->nk = 1;
	fC->TC->ReadOnly = True;
	fC->TC->KeyFu |= mNoKeyFocus;
	SetColor((TWins*)fC->TC, 0xe8e8e8);
	fC->TCmin = TDVal_CreateS(fC->pnlZ, 2, 121, 46, 18);
	fC->TCmin->nk = 1;
	fC->TCmin->dv_fu = tdv_fu_LostFocusEnter;
	fC->TCmin->OnValChange = fC_DVal_OnValChange;
	TDVal_Set_Min(fC->TCmin, 0.1);


  int OnOff;
	TDVal_Set_Value(fC->C, TpcIni_RD_float(pcIni, Gruppe, sC, val));
	sopt = TpcIni_RD_pc(pcIni, Gruppe, fC->sOpt, "0 0 0 0.001 0 0 0x18");
	TDVal_Set_Value(fC->MHz, ppctodouble(&sopt));
	TDVal_Set_Value(fC->Rs, ppctodouble(&sopt));
	TDVal_Set_Value(fC->Rp, ppctodouble(&sopt));
	TDVal_Set_Value(fC->Ls, ppctodouble(&sopt));
	TDVal_Set_Value(fC->TCmin, ppctodouble(&sopt));
	ppctodouble(&sopt);
	OnOff = ppctoint(&sopt);
	TChkBox_SetChecked(fC->cbOptOn, OnOff & 0x01);
	TChkBox_SetChecked(fC->cbRs, OnOff & 0x02);
	TChkBox_SetChecked(fC->cbRp, OnOff & 0x04);
	TChkBox_SetChecked(fC->cbLs, OnOff & 0x08);
	TChkBox_SetChecked(fC->cbTC, OnOff & 0x10);
	TChkBox_SetChecked(fC->cbTCsync, OnOff & 0x20);

	free_appIni(pcIni);

	fC_TC_SetVal(fC, TScrollBar_GetCount(fC->sbTC));

	fC_math_loc(fC);

	return fC;
}

