/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "Form_RLC.h"
#include "ElGraf.h"

static void fDigA_SaveDaten(TFormDigA *mf);


static void fDigA_Set_Stat_exDVal(TFormDigA *mf) {
  int Color = 0xffffff;
	if (!mf->exDVal) return;
	if (mf->cbTX->Checked) {
	    TDVal_Set_Value(mf->exDVal, mf->dvTX->dval + mf->dvX->dval);
	    mf->exDVal->ReadOnly = (mf->exDVal->dval != mf->dvX->dval);
	} else {
	    mf->exDVal->ReadOnly = False;
	    TDVal_Set_Value(mf->exDVal, mf->dvX->dval);
	}
	if (mf->exDVal->ReadOnly) Color = 0xe0e0e0;
	SetColor((TWins*)mf->exDVal, Color);
}

static void fDigA_math_loc(TFormDigA *mf) {
	fDigA_Set_Stat_exDVal(mf);
	if (mf->OnChange) mf->OnChange((TFormRLC*)mf);
}

static void sbTX_SetVal_ml(TFormDigA *mf, double dval) {
  char *pc;
  const char *cpc;
  int i, c, line_nr = 0;
  TLineEdit *le = mf->le;
  char buf[80];
	pc = buf;
	math_line_clr(mf->mls);
	snprintf(buf, sizeof(buf), "v=%f", dval);
	if (!math_line(mf->mls, buf)) goto Error;
	c = le->tls->Count;
	for (i=0;i<c;i++) {
	    cpc = tlineedit_get_pc(le, i);
	    if (!math_line(mf->mls, cpc)) goto Error;
	}
	return;
Error:
	printf("*********************\n");
	printf("ERROR: line=%d pos=%d\n", line_nr, mf->mls->err_X);
	printf("%s\n", pc);
	printf("*********************\n");
}

static void sbTX_SetVal(TFormDigA *mf, int Count) {
  float f;
  
	f = (float)Count / 1000;
	sbTX_SetVal_ml(mf, f);
	TDVal_Set_Value(mf->dvTXval, f);
	TDVal_Set_Value(mf->dvTX, mf->mls->res);
}

static void fDigA_sbTX_OnCount(TScrollBar* sb, int Count) {
  TFormDigA *mf = (TFormDigA*)sb->FormStruct;

	sbTX_SetVal(mf, Count);
	mf->val_sync = Count;
	if (mf->OnSync && mf->cbTXsync->Checked) mf->OnSync((TFormRLC*)mf, Count);
	if (mf->cbTX->Checked) fDigA_math_loc(mf);
}

static void fDigA_cbTX_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TFormDigA *mf = (TFormDigA*)self->FormStruct;
	fDigA_math_loc(mf);
	fDigA_SaveDaten(mf);
}

void Form_DigA_sync(TFormRLC *fRLC, TFormDigA *mf, int val) {
  TFormDigA *smf = (TFormDigA*)fRLC;
	if ((smf == mf) || (smf->dvNr->dval != mf->dvNr->dval)) return;
	mf->val_sync = val;
	if (mf->cbTX->Checked && mf->cbTXsync->Checked) {
	    TScrollBar_SetCount(mf->sbTX, val);
	    sbTX_SetVal(mf, val);
	    fDigA_Set_Stat_exDVal(mf);
	}
}

static void fDigA_SaveDaten(TFormDigA *mf) {
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
  char *pc;
  int OnOff = 0;
  char buf[120];

	if (mf->cbTX->Checked) OnOff |= 0x10;
	if (mf->cbTXsync->Checked) OnOff |= 0x20;

	pc = buf;
	doubletoppc(&pc, mf->dvNr->dval, mf->dvNr->nk | kSPACE_ADD);
	ppcmove(&pc, "0 0 0 ");
	inttoppc(&pc, OnOff);
	TpcIni_WR(pcIni, mf->Gruppe, mf->sOpt, buf);

	TpcIni_WR_double(pcIni, mf->Gruppe, mf->sF, mf->dvX->dval, mf->dvX->nk);
	TpcIni_SaveAndFree(pcIni);
}

static void fDigA_DVal_OnValChange(TDVal *dv, double val) {
  TFormDigA *mf = (TFormDigA*)dv->FormStruct;
	fDigA_math_loc(mf);
	fDigA_SaveDaten(mf);

}

static void Form_DigA_OnMove(TFormDigA *self) {
	TFormRLC_Save((TFormRLC*)self);
}

TFormDigA *Create_Form_DigA(TWins *parent, const char *Caption, 
		const char *Gruppe, const char *sF, TDVal *exDVal, float val) {
  TFormDigA *mf;
  TpcIni *pcIni;
  const char *sopt;

	pcIni = get_appIni();

	mf = TForm_CreatParamAlg(parent, sizeof(TFormDigA));
	mf->wtyp = wtyp_TFormDigA;
//	fC->Visible = False;
	mf->FormZStyle = fszTransient;
	mf->FormPosition = poDesigned;
	mf->Gruppe = Gruppe;
	mf->sF = sF;
	mf->sOpt = pcadd(sF, "opt");
	mf->sXY = pcadd(sF, "xy");
	mf->Save = TFormRLC_Save;

	SetCaption((TWins*)mf, Caption);
	mf->width = 190;
	mf->height = 135;
	mf->min_width = mf->width;
	mf->min_height = mf->height;
	mf->max_width = mf->width;
	mf->max_height = mf->height;
	sopt = TpcIni_RD_pc(pcIni, Gruppe, mf->sXY, "0 0");
	mf->left = ppctoint(&sopt);
	mf->top = ppctoint(&sopt);
	mf->tag = (ppctoint(&sopt) != 0);
	mf->Visible = False; //fC->tag;
	mf->OnMove = Form_DigA_OnMove;
	mf->exDVal = exDVal;

	mf->pnl1 = TWins_CreateBA((TWins*)mf, 0, 0, 172, 21, 0, alLeft);
	mf->pnlZ = TWins_CreateBA((TWins*)mf, 116, 0, 20, 80, mBorder | mBorderL, alClient);
	mf->pnlX = TWins_CreateBA(mf->pnl1, 0, 0, 134, 60, mBorder | mBorderL, alTop);
//	mf->pnlOpt = TWins_CreateU(mf->pnl1, 0, 20, 130, 80, NULL, mBorder | mBorderL, alClient, True);
	mf->le = TLineEdit_CreateA(mf->pnl1, 0, 0, 200, 200, alClient);
	TLineEdit_AddLine(mf->le, "r=v");
	mf->mls = xcalloc(sizeof(TMathLineStruct));

/* Panel X */
	mf->dvX = TDVal_CreateSV4(mf->pnlX, 2, 2, 92, 21, val, -9.9999999, 9.9999999, 8);
	mf->dvX->dv_fu = tdv_fu_LostFocusEnter;
	mf->dvX->OnValChange = fDigA_DVal_OnValChange;
	SetFont((TWins*)mf->dvX, "-*-helvetica-bold-r-*-*-14-*-75-75-*-*-*-*");
	if (mf->exDVal) mf->exDVal->owner = (TWins*)mf->dvX;

	mf->cbTX = TChkBox_CreateS(mf->pnlX, 54, 26, 32, 16, "TR");
	mf->cbTX->OnClick = fDigA_cbTX_OnClick;
	mf->cbTXsync = TChkBox_CreateS(mf->pnlX, 54, 43, 60, 16, "TLsync");
	mf->cbTXsync->OnClick = fDigA_cbTX_OnClick;

	mf->dvTX = TDVal_CreateSV4(mf->pnlX, 96, 3, 74, 19, 0, -99999999, 99999999, 8);
	mf->dvTX->ReadOnly = True;
	SetColor((TWins*)mf->dvTX, 0xe8e8e8);
	mf->dvTXval = TDVal_CreateSV4(mf->pnlX, 118, 40, 50, 18, 0, -9, 9, 3);
	mf->dvTXval->ReadOnly = True;
	SetColor((TWins*)mf->dvTXval, 0xe8e8e8);

	mf->dvNr = TDVal_CreateSV4(mf->pnlX, 12, 26, 30, 19, 0, 0, 9, 0);
	mf->dvNr->dv_fu = tdv_fu_LostFocusEnter;
	mf->dvNr->OnValChange = fDigA_DVal_OnValChange;

/* Panel Z */
	mf->sbTX = TScrollBarVert_CreateP(mf->pnlZ);
	SetWidth((TWins*)mf->sbTX, 17);
	mf->sbTX->Align = alRight;
	mf->sbTX->BorderFu = mBorder | mBorderL | mBorderNoTop | mBorderNoRigth | mBorderNoBottom;
	mf->sbTX->OnCount = fDigA_sbTX_OnCount;
	TScrollBar_SetCountParam(mf->sbTX, 0, 0, 2000, -1000, sbfu_inv);

  int OnOff;
	TDVal_Set_Value(mf->dvX, TpcIni_RD_float(pcIni, Gruppe, sF, val));
	sopt = TpcIni_RD_pc(pcIni, Gruppe, mf->sOpt, "0 0 0 0 0x10");
	TDVal_Set_Value(mf->dvNr, ppctodouble(&sopt));
	ppctodouble(&sopt);
	ppctodouble(&sopt);
	ppctodouble(&sopt);
	OnOff = ppctoint(&sopt);
	TChkBox_SetChecked(mf->cbTX, OnOff & 0x10);
	TChkBox_SetChecked(mf->cbTXsync, OnOff & 0x20);

	free_appIni(pcIni);

	return mf;
}
