/*
 Copyright (C) 2010 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "RLC.h"
#include <math.h>

void REGP3 SetUp_R(TRLC *rlc, float w, float R) {
	rlc->typ = RLC_R;
	rlc->val = R;
	rlc->U = 1;
	rlc->I = 1/R;
	rlc->Iw = 1/R;
	rlc->Ib = 0;
	rlc->Z = R;
	rlc->Xw = R;
	rlc->Xb = 0;
}
void REGP3 SetUp_C(TRLC *rlc, float w, float C) {
	rlc->typ = RLC_C;
	rlc->val = C;
	rlc->U = 1;
	rlc->Z = 1 / (w * C);
	rlc->Xw = 0;
	rlc->Xb = -1 / (w * C);
	rlc->Ib = 1 / rlc->Xb;
	rlc->Iw = 0;
	rlc->I = -rlc->Ib;
}
void REGP3 SetUp_L(TRLC *rlc, float w, float L) {
	rlc->typ = RLC_L;
	rlc->val = L;
	rlc->U = 1;
	rlc->Z = w * L;
	rlc->Xw = 0;
	rlc->Xb = w * L;
	rlc->I = 1 / rlc->Xb;
	rlc->Iw = 0;
	rlc->Ib = rlc->I;
}

void REGP3 RLC_Add_Par(TRLC *A, TRLC *B) {
	A->Iw += B->Iw;
	A->Ib += B->Ib;
	A->I = sqrt(A->Iw*A->Iw + A->Ib*A->Ib);
	A->Z = A->U / A->I;
	A->Xw = A->Z * A->Iw / A->I;
	A->Xb = A->Z * A->Ib / A->I;
}

void REGP3 RLC_Add_Ser(TRLC *A, TRLC *B) {
	A->Xw += B->Xw;
	A->Xb += B->Xb;
	A->Z = sqrt(A->Xw*A->Xw + A->Xb*A->Xb);
	A->I = A->U / A->Z;
	A->Iw = A->I * A->Xw / A->Z;
	A->Ib = A->I * A->Xb / A->Z;
}

void REGP3 RLC_Add_Ser_R(TRLC *A, float w, float R) {
  TRLC B;
	SetUp_R(&B, w, R);
	RLC_Add_Ser(A, &B);
}
void REGP3 RLC_Add_Par_R(TRLC *A, float w, float R) {
  TRLC B;
	SetUp_R(&B, w, R);
	RLC_Add_Par(A, &B);
}

void REGP3 RLC_Add_Ser_L(TRLC *A, float w, float L) {
  TRLC B;
	SetUp_L(&B, w, L);
	RLC_Add_Ser(A, &B);
}
void REGP3 RLC_Add_Par_L(TRLC *A, float w, float L) {
  TRLC B;
	SetUp_L(&B, w, L);
	RLC_Add_Par(A, &B);
}

void REGP3 RLC_Add_Ser_C(TRLC *A, float w, float C) {
  TRLC B;
	SetUp_C(&B, w, C);
	RLC_Add_Ser(A, &B);
}
void REGP3 RLC_Add_Par_C(TRLC *A, float w, float C) {
  TRLC B;
	SetUp_C(&B, w, C);
	RLC_Add_Par(A, &B);
}

void REGP3 RLC_Copy(TRLC *dest, TRLC *src) {
	pnmove(dest, src, sizeof(TRLC));
}
