/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "ElMath.h"
#include "ElGraf.h"
#include <math.h>

#define MAX_fx_NCoef 3

typedef struct {
  int order;
  int idx;
  float ACoef[MAX_fx_NCoef+1];
  float BCoef[MAX_fx_NCoef+1];
  float x[MAX_fx_NCoef+1];
  float y[MAX_fx_NCoef+1];
} Tval_IIR;

typedef struct {
  float sampF;
  float spF;
  float phase;
  Tval_IIR ir1;
  Tval_IIR ir2;
  Tval_IIR irS;
  float ret1;
  float ret2;
  float retSum;
  int ct;
} Tval_IIR_Sum;

static float math_IIR(float val, Tval_IIR *iir) {
  int i, coi0, coi, order;
  float ret;

	val *= iir->BCoef[0];
	order = iir->order;
	if (order == 3) {
	    iir->x[3] = iir->x[2];
	    iir->x[2] = iir->x[1];
	    iir->x[1] = iir->x[0];
	    iir->y[3] = iir->y[2];
	    iir->y[2] = iir->y[1];
	    iir->y[1] = iir->y[0];
	    iir->x[0] = val;
	    ret = iir->ACoef[0] * val;
	    ret += iir->ACoef[1] * iir->x[1] - iir->BCoef[1] * iir->y[1];
	    ret += iir->ACoef[2] * iir->x[2] - iir->BCoef[2] * iir->y[2];
	    ret += iir->ACoef[3] * iir->x[3] - iir->BCoef[3] * iir->y[3];
	    iir->y[0] = ret;
	    return ret;
	} else if (order == 2) {
	    iir->x[2] = iir->x[1];
	    iir->x[1] = iir->x[0];
	    iir->y[2] = iir->y[1];
	    iir->y[1] = iir->y[0];
	    iir->x[0] = val;
	    ret = iir->ACoef[0] * val;
	    ret += iir->ACoef[1] * iir->x[1] - iir->BCoef[1] * iir->y[1];
	    ret += iir->ACoef[2] * iir->x[2] - iir->BCoef[2] * iir->y[2];
	    iir->y[0] = ret; 
	    return ret;
	} else if (order == 1) {
	    iir->x[1] = iir->x[0];
	    iir->y[1] = iir->y[0];
	    iir->x[0] = val;
	    ret = iir->ACoef[0] * val;
            ret += iir->ACoef[1] * iir->x[1] - iir->BCoef[1] * iir->y[1];
	    iir->y[0] = ret;
	    return ret;
	}

	coi0 = coi = iir->idx;
	iir->x[coi] = val;
	ret = iir->ACoef[0] * val;
	coi++;
	if (coi > order) coi = 0;
	iir->idx--;
	if (iir->idx < 0) iir->idx = order;

	for(i=1; i<=order; i++) {
            ret += iir->ACoef[i] * iir->x[coi] - iir->BCoef[i] * iir->y[coi];
	    coi++;
	    if (coi > order) coi = 0;
	}
	iir->y[coi0] = ret; 
	return ret;
}

static void math_IIR1_Ser(float fc, Tval_IIR_Sum *is) {
  float w = is->spF * fc;
  float phase = is->phase;
  int i;
  float r1, r2, rS;
  float max_phase = 2 * M_PI;
  float min1, max1, min2, max2, minS, maxS, v;
  bool mk;

	if ((w < 0.0000000001) || (w >= M_PI*4)) {
	    is->ret1 = 0;
	    is->ret2 = 0;
	    is->retSum = 0;
	    return;
	}
	for (i=4;i;i--) {
	    v = sin(phase);
	    r1 = math_IIR(v, &is->ir1);
	    math_IIR(v, &is->ir2);
	    math_IIR(r1, &is->irS);
	    phase += w;
	    while (phase >= max_phase) phase -= max_phase;
	}
	while (phase < max_phase) {
	    v = sin(phase);
	    r1 = math_IIR(v, &is->ir1);
	    math_IIR(v, &is->ir2);
	    math_IIR(r1, &is->irS);
	    phase += w;
	}
	phase -= max_phase;
	min1 = max1 = min2 = max2 = minS = maxS = 0;

	i = 8;
	mk = True;
	while (mk || (i>0)) {
	    v = sin(phase);
	    r1 = math_IIR(v, &is->ir1);
	    r2 = math_IIR(v, &is->ir2);
	    rS = math_IIR(r1, &is->irS);
	    if (r1 > max1) max1 = r1;
	    if (r1 < min1) min1 = r1;
	    if (r2 > max2) max2 = r2;
	    if (r2 < min2) min2 = r2;
	    if (rS > maxS) maxS = rS;
	    if (rS < minS) minS = rS;
	
	    phase += w;
	    while (phase >= max_phase) {phase -= max_phase; mk = False;}
	    i--;
	}
	is->phase = phase;

	is->ret1 = (max1 - min1) / 2;
	is->ret2 = (max2 - min2) / 2;
	is->retSum = (maxS - minS) / 2;
}

static void math_IIR1_Sum(float fc, Tval_IIR_Sum *is) {
  float w = is->spF * fc;
  float phase = is->phase;
  int i;
  float r1, r2;
  float max_phase = 2 * M_PI;
  float min1, max1, min2, max2, minS, maxS, v;
  bool mk;

	if ((w < 0.0000001) || (w >= M_PI*4)) {
	    is->ret1 = 0;
	    is->ret2 = 0;
	    is->retSum = 0;
	    return;
	}

	if (w > 0.1) goto PhasePI2;

	while (phase < max_phase) {
	    v = sin(phase);
	    phase += w;
	    math_IIR(v, &is->ir1);
	    math_IIR(v, &is->ir2);
	}
	phase -= max_phase;

	min1 = max1 = min2 = max2 = minS = maxS = 0;

	while (phase < max_phase) {
	    v = sin(phase);
	    phase += w;
	    r1 = math_IIR(v, &is->ir1);
	    r2 = math_IIR(v, &is->ir2);

	    if (r1 > max1) max1 = r1;
	    if (r1 < min1) min1 = r1;
	    if (r2 > max2) max2 = r2;
	    if (r2 < min2) min2 = r2;
	    r1 += r2;
	    if (r1 > maxS) maxS = r1;
	    if (r1 < minS) minS = r1;
	}
	phase -= max_phase;
	is->phase = phase;

	is->ret1 = (max1 - min1) / 2;
	is->ret2 = (max2 - min2) / 2;
	is->retSum = (maxS - minS) / 2;
	return;

PhasePI2: 
	while (phase < max_phase) {
	    v = sin(phase);
	    phase += w;
	    math_IIR(v, &is->ir1);
	    math_IIR(v, &is->ir2);
	}
	while (phase >= max_phase) phase -= max_phase;

	min1 = max1 = min2 = max2 = minS = maxS = 0;

	i = 8;
	mk = True;
	while (mk || (i>0)) {
	    i--;
	    v = sin(phase);
	    phase += w;
	    while (phase >= max_phase) {phase -= max_phase; mk = False;}

	    r1 = math_IIR(v, &is->ir1);
	    r2 = math_IIR(v, &is->ir2);
	    if (r1 > max1) max1 = r1;
	    if (r1 < min1) min1 = r1;
	    if (r2 > max2) max2 = r2;
	    if (r2 < min2) min2 = r2;
	    r1 += r2;
	    if (r1 > maxS) maxS = r1;
	    if (r1 < minS) minS = r1;
	
	}
	while (phase >= max_phase) {phase -= max_phase; mk = False;}
	is->phase = phase;

	is->ret1 = (max1 - min1) / 2;
	is->ret2 = (max2 - min2) / 2;
	is->retSum = (maxS - minS) / 2;
}

void math_Filter_IIR1(TPage_IIR1 *pg) {
  TPage_DIG_Filter *digf = pg->pgDIGF;
  TScale_pxy *pxy ,*pxy2, *pxyS, *pxyT;
  int i, x, xt;
  float fc, fcStep, fOff, fStep, f, fx, fxt;
  Tval_IIR_Sum is;
  Tval_IIR *ir1, *ir2;

	Fill0(&is, sizeof(Tval_IIR_Sum));

	is.ir1.order = pg->dv_N->dval;
	is.ir1.ACoef[0] = pg->dv_a10->dval;
	is.ir1.ACoef[1] = pg->dv_a11->dval;
	is.ir1.ACoef[2] = pg->dv_a12->dval;
	is.ir1.ACoef[3] = pg->dv_a13->dval;
	is.ir1.BCoef[0] = pg->dv_b10->dval;
	is.ir1.BCoef[1] = pg->dv_b11->dval;
	is.ir1.BCoef[2] = pg->dv_b12->dval;
	is.ir1.BCoef[3] = pg->dv_b13->dval;

	is.ir2.order = pg->dv_N2->dval;
	is.ir2.ACoef[0] = pg->dv_a20->dval;
	is.ir2.ACoef[1] = pg->dv_a21->dval;
	is.ir2.ACoef[2] = pg->dv_a22->dval;
	is.ir2.ACoef[3] = pg->dv_a23->dval;
	is.ir2.BCoef[0] = pg->dv_b20->dval;
	is.ir2.BCoef[1] = pg->dv_b21->dval;
	is.ir2.BCoef[2] = pg->dv_b22->dval;
	is.ir2.BCoef[3] = pg->dv_b23->dval;

	is.irS.order = pg->dv_N2->dval;
	is.irS.ACoef[0] = pg->dv_a20->dval;
	is.irS.ACoef[1] = pg->dv_a21->dval;
	is.irS.ACoef[2] = pg->dv_a22->dval;
	is.irS.ACoef[3] = pg->dv_a23->dval;
	is.irS.BCoef[0] = pg->dv_b20->dval;
	is.irS.BCoef[1] = pg->dv_b21->dval;
	is.irS.BCoef[2] = pg->dv_b22->dval;
	is.irS.BCoef[3] = pg->dv_b23->dval;

	is.sampF = digf->dvSampKHZ->dval;
	is.spF = 2 * M_PI / is.sampF;

	pxy = digf->xy_ind->dat->pxy;
	pxy2 = pxy->next;
	pxyS = pxy2->next;
	pxyT = digf->pxy4;

	pxy->PaintDat  = pg->cb_F1->Checked;
	pxy2->PaintDat = pg->cb_F2->Checked;
	pxyS->PaintDat = (pg->cbSumOn->Checked || pg->cbSerOn->Checked);
	pxyT->PaintDat = (pg->cb_F1T->Checked || pg->cb_F2T->Checked);

	fOff = digf->ScaleKHzoff;
	fStep = digf->ScaleKHz;
	fcStep = fStep / (pxy->count-1);

	if (pg->cbSerOn->Checked) {
	    for (i=0;i<pxy->count;i++) {
		fc = i * fcStep + fOff;
		pxy->xy[i].x = fc;
		pxy2->xy[i].x = fc;
		pxyS->xy[i].x = fc;
		math_IIR1_Ser(fc, &is);
		pxy->xy[i].y = is.ret1;
		pxy2->xy[i].y = is.ret2;
		pxyS->xy[i].y = is.retSum;
	    }
	} else {
	    for (i=0;i<pxy->count;i++) {
		fc = i * fcStep + fOff;
		pxy->xy[i].x = fc;
		pxy2->xy[i].x = fc;
		pxyS->xy[i].x = fc;
		math_IIR1_Sum(fc, &is);
		pxy->xy[i].y = is.ret1;
		pxy2->xy[i].y = is.ret2;
		pxyS->xy[i].y = is.retSum;
	    }
	}

  float txk = 1 / is.sampF;
  float tx = 0;
	if (pxyT->PaintDat) {
	    ir1 = &is.ir1;
	    ir2 = &is.ir2;
	    for (i=0;i<4;i++) {
		ir1->x[i] = 0; ir1->y[i] = 0;
		ir2->x[i] = 0; ir2->y[i] = 0;
	    }
	    x = -1;
	    fxt = 0;
	    fx = digf->dv_T->dval / (pxyT->count-1);
	    f = 0;

	    if (pg->cb_F1T->Checked && pg->cb_F2T->Checked) {
	      if (pg->cbSerOn->Checked) {
		for (i=0;i<pxyT->count;i++) {
		    fc = i * fcStep + fOff;
		    pxyT->xy[i].x = fc;
		    xt = trunc(fxt);
		    while (x != xt) {
			x++; // = xt;
			f = math_IIR(1, ir1);
			f = math_IIR(f, ir2);
			tx += txk;
		    }
		    pxyT->xy[i].v1 = tx;
		    pxyT->xy[i].y = f;
		    fxt += fx;
		}
	      } else {
		for (i=0;i<pxyT->count;i++) {
		    fc = i * fcStep + fOff;
		    pxyT->xy[i].x = fc;
		    xt = trunc(fxt);
		    while (x != xt) {
			x++; // = xt;
			f = math_IIR(1, ir1);
			f += math_IIR(1, ir2);
			tx += txk;
		    }
		    pxyT->xy[i].v1 = tx;
		    pxyT->xy[i].y = f;
		    fxt += fx;
		}
	      }

	    } else {
		if (pg->cb_F2T->Checked) ir1 = ir2;
		for (i=0;i<pxyT->count;i++) {
		    fc = i * fcStep + fOff;
		    pxyT->xy[i].x = fc;
		    xt = trunc(fxt);
		    while (x != xt) {
			x++; // = xt;
			f = math_IIR(1, ir1);
			tx += txk;
		    }
		    pxyT->xy[i].v1 = tx;
		    pxyT->xy[i].y = f;
		    fxt += fx;
		}
	    }
	}

	pxy->xy[0].y = pxy->xy[1].y;
	pxy2->xy[0].y = pxy2->xy[1].y;
	pxyS->xy[0].y = pxyS->xy[1].y;

	ELM_IIR1_DataChange(digf);
}



/* *************************************************** */
static void IIR1_fXxx_OnChange(TFormRLC* fRLC) {
  TPage_IIR1 *pg = fRLC->subClass;
	if (!(pg->c_stat & 0x01)) {
	    math_Filter_IIR1((TPage_IIR1*)fRLC->subClass);
	}
}

static void IIR1_A_OnCSync(TFormRLC* fRLC, int val) {
  TPage_IIR1 *pg = fRLC->subClass;
	Form_DigA_sync(fRLC, pg->f_a10, val);
	Form_DigA_sync(fRLC, pg->f_a11, val);
	Form_DigA_sync(fRLC, pg->f_a12, val);
	Form_DigA_sync(fRLC, pg->f_a13, val);
	Form_DigA_sync(fRLC, pg->f_b10, val);
	Form_DigA_sync(fRLC, pg->f_b11, val);
	Form_DigA_sync(fRLC, pg->f_b12, val);
	Form_DigA_sync(fRLC, pg->f_b13, val);
	Form_DigA_sync(fRLC, pg->f_a20, val);
	Form_DigA_sync(fRLC, pg->f_a21, val);
	Form_DigA_sync(fRLC, pg->f_a22, val);
	Form_DigA_sync(fRLC, pg->f_a23, val);
	Form_DigA_sync(fRLC, pg->f_b20, val);
	Form_DigA_sync(fRLC, pg->f_b21, val);
	Form_DigA_sync(fRLC, pg->f_b22, val);
	Form_DigA_sync(fRLC, pg->f_b23, val);
}

static TFormDigA *Create_f_IIR1_A(TPage_IIR1 *pg, const char *Caption, 
			const char *Gruppe, const char *s, TDVal *dv, float val) {
  TFormDigA *mfA = Create_Form_DigA(appDat.obj, Caption, Gruppe, s, dv, val);
	mfA->subClass = pg;
	TDVal_Set_Value(dv, mfA->dvX->dval);
	mfA->OnChange = IIR1_fXxx_OnChange;
	mfA->OnSync = IIR1_A_OnCSync;
	return mfA;
}

/* *************************************************** */
static void IIR1_DVal_SaveDaten(TPage_IIR1 *pg) {
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
  char *pc;
  int OnOff = 0;
  char buf[120];

	if (pg->cb_F1->Checked)    OnOff |= 0x01;
	if (pg->cb_F1T->Checked)   OnOff |= 0x02;
	if (pg->cb_F2->Checked)    OnOff |= 0x04;
	if (pg->cb_F2T->Checked)   OnOff |= 0x08;
	if (pg->cbSumOn->Checked)  OnOff |= 0x10;
	if (pg->cbSerOn->Checked)  OnOff |= 0x20;
	if (pg->cb_cToF1->Checked) OnOff |= 0x01000;
	if (pg->cb_cToF2->Checked) OnOff |= 0x02000;
	if (pg->cb_cHP->Checked)   OnOff |= 0x10000;

	pc = buf;
	doubletoppc(&pc, pg->dv_N->dval, pg->dv_N->nk | kSPACE_ADD);
	doubletoppc(&pc, pg->dv_N2->dval, pg->dv_N2->nk | kSPACE_ADD);
	doubletoppc(&pc, pg->dv_cN->dval, pg->dv_cN->nk | kSPACE_ADD);
	doubletoppc(&pc, pg->dv_cFreq->dval, pg->dv_cFreq->nk | kSPACE_ADD);

	ppcmove(&pc, "0 0 ");
	hexto0xppc(&pc, OnOff, 8);
	TpcIni_WR(pcIni, "IIR1", "Nopt", buf);
	TpcIni_SaveAndFree(pcIni);
}

static void IIR1_DVal_OnValChange(TDVal *dv, double val) {
	if (dv->owner) TDVal_DoValChange((TDVal*)dv->owner, val, tdvfu_ChangeEv);
}

static void IIR1_DVal_N_OnValChange(TDVal *dv, double val) {
	IIR1_DVal_SaveDaten(dv->subClass);
	math_Filter_IIR1(dv->subClass);
}

static void IIR1cbSumOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_IIR1 *pg = self->subClass;
	if (pg->cbSumOn == self) TChkBox_SetChecked(pg->cbSerOn, False);
	if (pg->cbSerOn == self) TChkBox_SetChecked(pg->cbSumOn, False);
	math_Filter_IIR1(pg);
	IIR1_DVal_SaveDaten(pg);
}

static void IIR1cb_cToF1_OnClick(TChkBox *self, long KeyState, bool DblClick) {
	IIR1_DVal_SaveDaten(self->subClass);
}

/* ********************** TPage_IIR1 ********************** */
typedef struct {
  float re, im, mag, phase;
  float wc;
} TComplF;

static void GetPoles(TComplF *co, int n, int i) {
  float re, im, mag, phase;
  float re2, im2, k, a, b;

	phase = M_PI/2 + (2*i+1) * M_PI/(2*n);
	mag = tan(co->wc);
	re = mag * cos(phase);
	im = mag * sin(phase);
	re2 = -re;
	im2 = -im;
	re += 1;
	re2 += 1;
	k = re2*re2 + im2*im2;
	a = re2/k;
	b = -im2/k;
	co->re = re*a - im*b;
	co->im = re*b + im*a;
	co->mag = sqrt(co->re*co->re + co->im*co->im);
	co->phase = atan(co->im/co->re);
}

static void ab_math_IIR(TPage_IIR1 *pg) {
  TPage_DIG_Filter *digf = pg->pgDIGF;
  TIRR_ab *ab = &pg->ab;
  int i;
  float ftg, W, a, b, c, f, pol;
  TComplF co;

	ab->fA = digf->dvSampKHZ->dval * 1000;
	ab->n = pg->dv_cN->dval;
	ab->fg = pg->dv_cFreq->dval;
	if ((ab->fg*2) > ab->fA) ab->fg = ab->fA * 0.5;
	ab->art = 0;			//Low Pass
	pol = 1;
	if (pg->cb_cHP->Checked) {
	    ab->art = 1;		//High Pass
	    pol = -1;
	}

	co.wc = M_PI * ab->fg / ab->fA;

	ftg = ab->fA / M_PI * tan(M_PI * ab->fg / ab->fA);
	W= ab->fA / (M_PI * ftg);

	for (i=0;i<4;i++) { ab->a[i] = 0;  ab->b[i] = 0;}
	ab->b[0] = 1;
	if (ab->n == 1) {
	    ab->a[0] = 1 / (1+W);
	    if (ab->art == 1) ab->a[0] = 1-ab->a[0];
	    ab->a[1] = pol*ab->a[0];
	    ab->b[1] = (1-W) / (1+W);
	} if (ab->n == 2) {
#ifdef M_SQRT2l
	    a = M_SQRT2l*W;
#else
	    a = sqrt(2)*W;
#endif
	    b = W*W;
	    ab->a[0] = 1/(1+a+b);
	    if (ab->art == 1) ab->a[0] = 1-2/(1+a);

	    ab->a[1] = pol*ab->a[0]*2;
	    ab->a[2] = ab->a[0];
	    ab->b[1] = 2*(1-b)/(1+a+b);
	    ab->b[2] = (1-a+b)/(1+a+b);
	} if (ab->n == 3) {
	    a = 2*W;
	    b = 2*W*W;
	    c = W*W*W;
	    ab->a[0] = 1/(1+a+b+c);
	    if (ab->art == 1) ab->a[0] = 1-2/(1+W);
	    ab->a[1] = pol*ab->a[0]*3;
	    ab->a[2] = ab->a[0]*3;
	    ab->a[3] = pol*ab->a[0];
	    ab->b[3] = (1-a+b-c)/(1+a+b+c);

	    GetPoles(&co, 3, 0);
	    ab->b[1] = -2 * co.re;
	    f = co.mag * co.mag * 2;
	    ab->b[2] = f + (f-1);
	    ab->b[3] = co.mag * co.mag;
	    GetPoles(&co, 3, 1);
	    ab->b[1] -=  co.re;
	    ab->b[3] *= -co.mag;
	}
}


static void IIR1_DVal_cN_OnValChange(TDVal *dv, double val) {
  TPage_IIR1 *pg = dv->subClass;
  TIRR_ab *ab = &pg->ab;
  const char *cpc;
  char buf[120];

	TLineEdit_Clr(pg->le);
	ab_math_IIR(pg);

	cpc = "/* Butterworth Low Pass */";
	if (ab->art == 1) cpc = "/* Butterworth High Pass */";
	TLineEdit_AddLine(pg->le, cpc);
	snprintf(buf, sizeof(buf), "/* N=%d SampF=%dHz fg=%.1fHz */", 
				ab->n, (int)ab->fA, ab->fg);
	TLineEdit_AddLine(pg->le, buf);

	if (ab->n == 1) {
	    snprintf(buf, sizeof(buf), "float A[2] = {%.8f, %.8f};", ab->a[0], ab->a[1]);
	    TLineEdit_AddLine(pg->le, buf);
	    snprintf(buf, sizeof(buf), "float B[2] = {%.8f, %.8f};", ab->b[0], ab->b[1]);
	    TLineEdit_AddLine(pg->le, buf);

	} if (ab->n == 2) {
	    snprintf(buf, sizeof(buf), "float A[3] = {%.8f, %.8f, %.8f};",
					ab->a[0], ab->a[1], ab->a[2]);
	    TLineEdit_AddLine(pg->le, buf);
	    snprintf(buf, sizeof(buf), "float B[3] = {%.8f, %.8f, %.8f};",
					ab->b[0], ab->b[1], ab->b[2]);
	    TLineEdit_AddLine(pg->le, buf);

	} if (ab->n == 3) {
	    snprintf(buf, sizeof(buf), "float A[4] = {%.8f, %.8f, %.8f, %.8f};",
					ab->a[0], ab->a[1], ab->a[2], ab->a[3]);
	    TLineEdit_AddLine(pg->le, buf);
	    snprintf(buf, sizeof(buf), "float B[4] = {%.8f, %.8f, %.8f, %.8f};",
					ab->b[0], ab->b[1], ab->b[2], ab->b[3]);
	    TLineEdit_AddLine(pg->le, buf);
	}

	TLineEdit_AddLine(pg->le, NULL);
	snprintf(buf, sizeof(buf), "float x[%d], y[%d];", ab->n+1, ab->n+1);
	TLineEdit_AddLine(pg->le, buf);
	TLineEdit_AddLine(pg->le, NULL);
	TLineEdit_AddLine(pg->le, "float IIR(float val) {");
	TLineEdit_AddLine(pg->le, "  float ret;");
	TLineEdit_AddLine(pg->le, "	val *= B[0];");
	if (ab->n > 2) TLineEdit_AddLine(pg->le, "	x[3] = x[2];");
	if (ab->n > 1) TLineEdit_AddLine(pg->le, "	x[2] = x[1];");
	TLineEdit_AddLine(pg->le, "	x[1] = x[0];");
	if (ab->n > 2) TLineEdit_AddLine(pg->le, "	y[3] = y[2];");
	if (ab->n > 1) TLineEdit_AddLine(pg->le, "	y[2] = y[1];");
	TLineEdit_AddLine(pg->le, "	y[1] = y[0];");
	TLineEdit_AddLine(pg->le, "	x[0] = val;");
	TLineEdit_AddLine(pg->le, "	ret = A[0] * val;");
	TLineEdit_AddLine(pg->le, "	ret += A[1] * x[1] - B[1] * y[1];");
	if (ab->n > 1) TLineEdit_AddLine(pg->le, "	ret += A[2] * x[2] - B[2] * y[2];");
	if (ab->n > 2) TLineEdit_AddLine(pg->le, "	ret += A[3] * x[3] - B[3] * y[3];");
	TLineEdit_AddLine(pg->le, "	y[0] = ret;");
	TLineEdit_AddLine(pg->le, "	return ret;");
	TLineEdit_AddLine(pg->le, "}");

	TLineEdit_AllAtrNew(pg->le);

	SetEnabled(pg->buToF, ((ab->n >= 1) && (ab->n <= 3)));

	IIR1_DVal_SaveDaten(pg);
}

static void IIR1_cb_cHP_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_IIR1 *pg = self->subClass;
	IIR1_DVal_cN_OnValChange(pg->dv_cN, 0);
	IIR1_DVal_SaveDaten(pg);
}

static void IIR1_buToF_OnClick(TWins *self, long KeyState, bool DblClick) {
  TPage_IIR1 *pg = self->subClass;
  TIRR_ab *ab = &pg->ab;
	pg->c_stat |= 0x01;
	if (pg->cb_cToF1->Checked) {
	    TDVal_Set_Value(pg->dv_N, ab->n);
	    IIR1_DVal_OnValChange(pg->dv_a10, ab->a[0]);
	    IIR1_DVal_OnValChange(pg->dv_a11, ab->a[1]);
	    IIR1_DVal_OnValChange(pg->dv_a12, ab->a[2]);
	    IIR1_DVal_OnValChange(pg->dv_a13, ab->a[3]);
	    IIR1_DVal_OnValChange(pg->dv_b10, ab->b[0]);
	    IIR1_DVal_OnValChange(pg->dv_b11, ab->b[1]);
	    IIR1_DVal_OnValChange(pg->dv_b12, ab->b[2]);
	    IIR1_DVal_OnValChange(pg->dv_b13, ab->b[3]);
	}
	if (pg->cb_cToF2->Checked) {
	    TDVal_Set_Value_ev(pg->dv_N2, ab->n);
	    TDVal_Set_Value_ev(pg->dv_a20, ab->a[0]);
	    TDVal_Set_Value_ev(pg->dv_a21, ab->a[1]);
	    TDVal_Set_Value_ev(pg->dv_a22, ab->a[2]);
	    TDVal_Set_Value_ev(pg->dv_a23, ab->a[3]);
	    TDVal_Set_Value_ev(pg->dv_b20, ab->b[0]);
	    TDVal_Set_Value_ev(pg->dv_b21, ab->b[1]);
	    TDVal_Set_Value_ev(pg->dv_b22, ab->b[2]);
	    TDVal_Set_Value_ev(pg->dv_b23, ab->b[3]);
	}
	pg->c_stat &= ~0x01;
	math_Filter_IIR1(pg);
	IIR1_DVal_SaveDaten(pg);
}

static void str_toGraf(TWins *Wins, const char *s, int x, int y) {
	vds_SetStr(&Wins->vds, s);
	DrawColorString(x, y, &Wins->vds);
}

static void IIR1_pnl_OnPaint(TWins *self) {
  TPage_IIR1 *pg = self->subClass;
  int x, y;

	str_toGraf(self, "N=", 2, pg->dv_N->top + 2);
	str_toGraf(self, "N=", 2, pg->dv_N2->top + 2);

	y = pg->dv_cN->top + 2;
	str_toGraf(self, "N=", 2, y);
	x = pg->dv_cN->left +  pg->dv_cN->width + 3;
	str_toGraf(self, "Freq=", x, y);
}

static void IIR1_pnl_OnResize(TWins *self) {
  TPage_IIR1 *pg = self->subClass;
  int x, y;

	y = self->height - 21;
	SetTop((TWins*)pg->dv_cN, y);

	vds_SetStr(&self->vds, "Freq=");
	x = TextWidthColorString(&self->vds) + 4;
	x += pg->dv_cN->left +  pg->dv_cN->width;
	SetPosition((TWins*)pg->dv_cFreq, x, y);
	x += pg->dv_cFreq->width + 2;
	SetPosition((TWins*)pg->cb_cHP, x, y+2);

	x = self->width - pg->buToF->width;
	SetPosition((TWins*)pg->buToF, x, y-5);
	x -= pg->cb_cToF1->width + 1;
	SetPosition((TWins*)pg->cb_cToF1, x, y-11);
	SetPosition((TWins*)pg->cb_cToF2, x, y+5);
}

void Dig_IIR1_SetVisible(TPage_IIR1 *pg, bool On) {
	SetVisible((TWins*)pg->f_a10, pg->f_a10->tag && On);
	SetVisible((TWins*)pg->f_a11, pg->f_a11->tag && On);
	SetVisible((TWins*)pg->f_a12, pg->f_a12->tag && On);
	SetVisible((TWins*)pg->f_a13, pg->f_a13->tag && On);
	SetVisible((TWins*)pg->f_b10, pg->f_b10->tag && On);
	SetVisible((TWins*)pg->f_b11, pg->f_b11->tag && On);
	SetVisible((TWins*)pg->f_b12, pg->f_b12->tag && On);
	SetVisible((TWins*)pg->f_b13, pg->f_b13->tag && On);
	SetVisible((TWins*)pg->f_a20, pg->f_a20->tag && On);
	SetVisible((TWins*)pg->f_a21, pg->f_a21->tag && On);
	SetVisible((TWins*)pg->f_a22, pg->f_a22->tag && On);
	SetVisible((TWins*)pg->f_a23, pg->f_a23->tag && On);
	SetVisible((TWins*)pg->f_b20, pg->f_b20->tag && On);
	SetVisible((TWins*)pg->f_b21, pg->f_b21->tag && On);
	SetVisible((TWins*)pg->f_b22, pg->f_b22->tag && On);
	SetVisible((TWins*)pg->f_b23, pg->f_b23->tag && On);
}

static void pg_Dig_IIR1_OnActivate(TPage_IIR1 *self) {
	if (self->RunStatus & runst_Activate1) return;
	math_Filter_IIR1(self);
	IIR1_DVal_cN_OnValChange(self->dv_cN, 0);
}

static void DVal_Create_Axx(TWins *Parent, TDVal **dv, TFormRLConoff **onoff, int *x, int y, int l2, TPic *pic) {
  int tx = *x;
  int x2 = tx;
  int y2 = y + 17;
  TDVal *tdv;
	if (l2) {y2 += 20; x2 += 24;}
	tdv = TDVal_CreateSV4(Parent, tx, y2, 84, 19, 0, -9.9999999, 9.9999999, 8);
	tdv->dv_fu = tdv_fu_LostFocusEnter;
	tdv->OnValChange = IIR1_DVal_OnValChange;
	*dv = tdv;
	*onoff = TFormRLConoff_CreateU(Parent, x2, y, 18, 16, pic);
	if (l2) *x = tx + 84 + 1;
}

static TPage *page_Add_Dig_IIR1_Create(TPageControl *Parent) {
	return TPage_CreateParamAlg(Parent, sizeof(TPage_IIR1));
}

TPage_IIR1 *page_Add_Dig_IIR1(TPageControl *pagc, TPage_DIG_Filter *pgDIGF) {
  TPage_IIR1 *pg;
  TWins *pnl;
  TpcIni *pcIni;
  int x, y;
  const char *sopt;

	pcIni = get_appIni();

	pg = (TPage_IIR1*)TPageControl_AddNewUserPage(pagc, "IIR", page_Add_Dig_IIR1_Create);
	pg->subClass = pg;
	pg->pgDIGF = pgDIGF;
	pg->OnActivate = pg_Dig_IIR1_OnActivate;

	pnl = TWins_CreateBA((TWins*)pg, 0, 0, 318, 210,
				0, alTop);
	pg->pnl = pnl;
	pnl->OnPaint = IIR1_pnl_OnPaint;
	pnl->OnResize = IIR1_pnl_OnResize;

	pg->le = TLineEdit_CreateA((TWins*)pg, 0, 210, 310, 200, alClient);
	pg->le->highlight = Thighlight_C_Create((TWins*)pg->le, pg->le->tls);
	pg->le->ReadOnly = True;

	x = 0;
	y = 2;
	pg->dv_N = TDVal_CreateSV4(pnl, 19, y, 30, 19, 0, 1, 3, 0);
	pg->dv_N->dv_fu = tdv_fu_LostFocusEnter;
	pg->dv_N->OnValChange = IIR1_DVal_N_OnValChange;
	pg->cb_F1 = TChkBox_CreateS(pnl, 52, y+1, 40, 17, "F1");
	pg->cb_F1T = TChkBox_CreateS(pnl, 94, y+1, 40, 17, "F1T");
	SetColor((TWins*)pg->cb_F1, 0x8080ff);
	SetColor((TWins*)pg->cb_F1T, 0xffff80);
	pg->cb_F1->OnClick = IIR1cbSumOn_OnClick;
	pg->cb_F1T->OnClick = IIR1cbSumOn_OnClick;


	x = 0;
	y += 20;
	DVal_Create_Axx(pnl, &pg->dv_a10, &pg->OnOff_a10, &x, y, 0, get_Pic_ca0());
	DVal_Create_Axx(pnl, &pg->dv_b10, &pg->OnOff_b10, &x, y, 1, get_Pic_cb0());

	DVal_Create_Axx(pnl, &pg->dv_a11, &pg->OnOff_a11, &x, y, 0, get_Pic_ca1());
	DVal_Create_Axx(pnl, &pg->dv_b11, &pg->OnOff_b11, &x, y, 1, get_Pic_cb1());

	DVal_Create_Axx(pnl, &pg->dv_a12, &pg->OnOff_a12, &x, y, 0, get_Pic_ca2());
	DVal_Create_Axx(pnl, &pg->dv_b12, &pg->OnOff_b12, &x, y, 1, get_Pic_cb2());

	DVal_Create_Axx(pnl, &pg->dv_a13, &pg->OnOff_a13, &x, y, 0, get_Pic_ca3());
	DVal_Create_Axx(pnl, &pg->dv_b13, &pg->OnOff_b13, &x, y, 1, get_Pic_cb3());
// Filter 2
	x = 0;
	y += 60;
	pg->dv_N2 = TDVal_CreateSV4(pnl, 19, y, 30, 19, 0, 1, 3, 0);
	pg->dv_N2->dv_fu = tdv_fu_LostFocusEnter;
	pg->dv_N2->OnValChange = IIR1_DVal_N_OnValChange;
	pg->cb_F2 = TChkBox_CreateS(pnl, 52, y+1, 40, 17, "F2");
	pg->cb_F2T = TChkBox_CreateS(pnl, 94, y+1, 40, 17, "F2T");
	SetColor((TWins*)pg->cb_F2, 0x80ff80);
	SetColor((TWins*)pg->cb_F2T, 0xffff80);
	pg->cb_F2->OnClick = IIR1cbSumOn_OnClick;
	pg->cb_F2T->OnClick = IIR1cbSumOn_OnClick;

	x = 0;
	y += 20;
	DVal_Create_Axx(pnl, &pg->dv_a20, &pg->OnOff_a20, &x, y, 0, get_Pic_ca0());
	DVal_Create_Axx(pnl, &pg->dv_b20, &pg->OnOff_b20, &x, y, 1, get_Pic_cb0());

	DVal_Create_Axx(pnl, &pg->dv_a21, &pg->OnOff_a21, &x, y, 0, get_Pic_ca1());
	DVal_Create_Axx(pnl, &pg->dv_b21, &pg->OnOff_b21, &x, y, 1, get_Pic_cb1());

	DVal_Create_Axx(pnl, &pg->dv_a22, &pg->OnOff_a22, &x, y, 0, get_Pic_ca2());
	DVal_Create_Axx(pnl, &pg->dv_b22, &pg->OnOff_b22, &x, y, 1, get_Pic_cb2());

	DVal_Create_Axx(pnl, &pg->dv_a23, &pg->OnOff_a23, &x, y, 0, get_Pic_ca3());
	DVal_Create_Axx(pnl, &pg->dv_b23, &pg->OnOff_b23, &x, y, 1, get_Pic_cb3());
// SUM
	y += 60;
	pg->cbSumOn = TChkBox_CreateS(pnl, 2, y, 60, 16, "Sum on");
	pg->cbSerOn = TChkBox_CreateS(pnl, 70, y, 60, 16, "Ser on");

	pg->cbSumOn->OnClick = IIR1cbSumOn_OnClick;
	pg->cbSerOn->OnClick = IIR1cbSumOn_OnClick;
	SetColor((TWins*)pg->cbSumOn, 0xe08080);
	SetColor((TWins*)pg->cbSerOn, 0xe08080);


// Filter 1
	pg->f_a10 = Create_f_IIR1_A(pg, "IIR A10", "IIR1", "a10", pg->dv_a10, 0.16591);
	pg->f_a11 = Create_f_IIR1_A(pg, "IIR A11", "IIR1", "a11", pg->dv_a11, 0.16591);
	pg->f_a12 = Create_f_IIR1_A(pg, "IIR A12", "IIR1", "a12", pg->dv_a12, 0);
	pg->f_a13 = Create_f_IIR1_A(pg, "IIR A13", "IIR1", "a13", pg->dv_a13, 0);
	pg->f_b10 = Create_f_IIR1_A(pg, "IIR B10", "IIR1", "b10", pg->dv_b10, 1);
	pg->f_b11 = Create_f_IIR1_A(pg, "IIR B11", "IIR1", "b11", pg->dv_b11, -0.66818);
	pg->f_b12 = Create_f_IIR1_A(pg, "IIR B12", "IIR1", "b12", pg->dv_b12, 0);
	pg->f_b13 = Create_f_IIR1_A(pg, "IIR B13", "IIR1", "b13", pg->dv_b13, 0);
	TFormRLConoff_SetForm(pg->OnOff_a10, (TForm*)pg->f_a10);
	TFormRLConoff_SetForm(pg->OnOff_a11, (TForm*)pg->f_a11);
	TFormRLConoff_SetForm(pg->OnOff_a12, (TForm*)pg->f_a12);
	TFormRLConoff_SetForm(pg->OnOff_a13, (TForm*)pg->f_a13);
	TFormRLConoff_SetForm(pg->OnOff_b10, (TForm*)pg->f_b10);
	TFormRLConoff_SetForm(pg->OnOff_b11, (TForm*)pg->f_b11);
	TFormRLConoff_SetForm(pg->OnOff_b12, (TForm*)pg->f_b12);
	TFormRLConoff_SetForm(pg->OnOff_b13, (TForm*)pg->f_b13);
// Filter 2
	pg->f_a20 = Create_f_IIR1_A(pg, "IIR A20", "IIR1", "a20", pg->dv_a20, 0.83409);
	pg->f_a21 = Create_f_IIR1_A(pg, "IIR A21", "IIR1", "a21", pg->dv_a21, -0.83409);
	pg->f_a22 = Create_f_IIR1_A(pg, "IIR A22", "IIR1", "a22", pg->dv_a22, 0);
	pg->f_a23 = Create_f_IIR1_A(pg, "IIR A23", "IIR1", "a23", pg->dv_a23, 0);
	pg->f_b20 = Create_f_IIR1_A(pg, "IIR B20", "IIR1", "b20", pg->dv_b20, 1);
	pg->f_b21 = Create_f_IIR1_A(pg, "IIR B21", "IIR1", "b21", pg->dv_b21, -0.66818);
	pg->f_b22 = Create_f_IIR1_A(pg, "IIR B22", "IIR1", "b22", pg->dv_b22, 0);
	pg->f_b23 = Create_f_IIR1_A(pg, "IIR B23", "IIR1", "b23", pg->dv_b23, 0);
	TFormRLConoff_SetForm(pg->OnOff_a20, (TForm*)pg->f_a20);
	TFormRLConoff_SetForm(pg->OnOff_a21, (TForm*)pg->f_a21);
	TFormRLConoff_SetForm(pg->OnOff_a22, (TForm*)pg->f_a22);
	TFormRLConoff_SetForm(pg->OnOff_a23, (TForm*)pg->f_a23);
	TFormRLConoff_SetForm(pg->OnOff_b20, (TForm*)pg->f_b20);
	TFormRLConoff_SetForm(pg->OnOff_b21, (TForm*)pg->f_b21);
	TFormRLConoff_SetForm(pg->OnOff_b22, (TForm*)pg->f_b22);
	TFormRLConoff_SetForm(pg->OnOff_b23, (TForm*)pg->f_b23);


	pg->dv_cN = TDVal_CreateSV4(pnl, 19, 190, 30, 19, 1, 1, 3, 0);
	pg->dv_cN->dv_fu = tdv_fu_LostFocusEnter;
	pg->dv_cN->OnValChange = IIR1_DVal_cN_OnValChange;
	pg->dv_cFreq = TDVal_CreateSV4(pnl, 80, 190, 80, 19, 1000, 1, 999999, 1);
	pg->dv_cFreq->dv_fu = tdv_fu_LostFocusEnter;
	pg->dv_cFreq->OnValChange = IIR1_DVal_cN_OnValChange;

	pg->cb_cHP = TChkBox_CreateS(pnl, 200, 190, 36, 16, "HP");
	pg->cb_cHP->OnClick = IIR1_cb_cHP_OnClick;
	

	pg->cb_cToF1 = TChkBox_CreateS(pnl, 200, 190, 46, 16, "to F1");
	pg->cb_cToF2 = TChkBox_CreateS(pnl, 250, 190, 46, 16, "to F2");
	pg->cb_cToF1->OnClick = IIR1cb_cToF1_OnClick;
	pg->cb_cToF2->OnClick = IIR1cb_cToF1_OnClick;
	pg->buToF = (TWins*)TButton_CreateS(pnl, 300, 190, 48, 20, "to F");
	pg->buToF->OnClick = IIR1_buToF_OnClick;
	SetEnabled(pg->buToF, False);

  int OnOff;
	sopt = TpcIni_RD_pc(pcIni, "IIR1", "Nopt", "0 0 1 1000 0 0");
	TDVal_Set_Value(pg->dv_N, ppctodouble(&sopt));
	TDVal_Set_Value(pg->dv_N2, ppctodouble(&sopt));

	TDVal_Set_Value(pg->dv_cN, ppctodouble(&sopt));
	TDVal_Set_Value(pg->dv_cFreq, ppctodouble(&sopt));

	ppctodouble(&sopt);
	ppctodouble(&sopt);
	OnOff = ppctoint(&sopt);
	TChkBox_SetChecked(pg->cb_F1,   OnOff & 0x01);
	TChkBox_SetChecked(pg->cb_F1T,  OnOff & 0x02);
	TChkBox_SetChecked(pg->cb_F2,   OnOff & 0x04);
	TChkBox_SetChecked(pg->cb_F2T,  OnOff & 0x08);
	TChkBox_SetChecked(pg->cbSumOn, OnOff & 0x10);
	TChkBox_SetChecked(pg->cbSerOn, OnOff & 0x20);

	TChkBox_SetChecked(pg->cb_cToF1, ((OnOff & 0x01000) != 0));
	TChkBox_SetChecked(pg->cb_cToF2, ((OnOff & 0x02000) != 0));
	TChkBox_SetChecked(pg->cb_cHP,   ((OnOff & 0x10000) != 0));

	free_appIni(pcIni);

	return pg;
}
