/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "ElMath.h"
#include "ElGraf.h"
#include <math.h>

static void str_toGraf(TWins *Wins, const char *s, int x, int y) {
	vds_SetStr(&Wins->vds, s);
	DrawColorString(x, y, &Wins->vds);
}

static void DrawBP2x(Window WinID, int x, int y, int fu) {
  int xa;
  int x2 = x;
  int y2 = y+30;

	Draw_xP1_Rg(WinID, &x2, &xa, y, y2, fu);

	if ((fu & 0x0300) == 0x000) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lv(WinID, &x2, 9, y, get_Pic_cL1());
	    DrawTP1_Ch(WinID, &x2, y, get_Pic_cC2());
	    DrawTP1_Lv(WinID, &x2, 0, y, get_Pic_cL2());
	    DrawTP1_Cv(WinID, &x2, 9, y, get_Pic_cC3());
	}

	if ((fu & 0x0300) == 0x100) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lv(WinID, &x2, 9, y, get_Pic_cL1());
	    DrawTP1_Ch(WinID, &x2, y, get_Pic_cC2());
	    DrawTP1_Lv(WinID, &x2, 0, y, get_Pic_cL2());
	    DrawTP1_Cv(WinID, &x2, 9, y, get_Pic_cC3());
	}
	if ((fu & 0x0300) == 0x200) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lv(WinID, &x2, 9, y, get_Pic_cL1());
	    DrawTP1_Ch(WinID, &x2, y, get_Pic_cC2());
	    DrawTP1_Lv(WinID, &x2, 0, y, get_Pic_cL2());
	    DrawTP1_Cv(WinID, &x2, 9, y, get_Pic_cC2());
	}

	if ((fu & 0x0300) == 0x300) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lv(WinID, &x2, 9, y, get_Pic_cL1());
	    DrawTP1_Ch(WinID, &x2, y, get_Pic_cC2());
	    DrawTP1_Lv(WinID, &x2, 0, y, get_Pic_cL2());
	    DrawTP1_Cv(WinID, &x2, 9, y, get_Pic_cC2());
	}

	Draw_xP1_Ra(WinID, x2, xa, y, y2, fu);
}

static bool ImpBP2x_ZF1(TPage_BP2x *pg, float w, TRLC *btIn, float *ret) {
  float res;
	if (w < 0.000001) {
	    if (pg->cbCxOn->Checked || pg->cbCsOn->Checked) return 0;
	    res = pg->f_Rs->R->dval;
	    *ret = pg->f_Ra->R->dval / (pg->f_Ra->R->dval + res);
	    return False;
	}
	Form_R_math(pg->f_Ra, w);

	res = 1;
	RLC_Copy(btIn, &pg->f_Ra->rlc);
	if (pg->cbCxOn->Checked) {
	    res = btIn->Z;
	    Form_C_math(pg->f_Cx, w);
	    RLC_Add_Ser(btIn, &pg->f_Cx->rlc);
	    res /= btIn->Z;
	}

	Form_C_math(pg->f_C1, w);
	Form_L_math(pg->f_L1, w);
	Form_C_math(pg->f_C2, w);
	Form_C_math(pg->f_C3, w);
	Form_L_math(pg->f_L2, w);

	RLC_Add_Par(btIn, &pg->f_C3->rlc);
	RLC_Add_Par(btIn, &pg->f_L2->rlc);
	res *= btIn->Z;

	RLC_Add_Ser(btIn, &pg->f_C2->rlc);
	res /= btIn->Z;

	RLC_Add_Par(btIn, &pg->f_C1->rlc);
	RLC_Add_Par(btIn, &pg->f_L1->rlc);
	res *= btIn->Z;

	if (pg->cbCsOn->Checked) {
	    Form_C_math(pg->f_Cs, w);
	    RLC_Add_Ser(btIn, &pg->f_Cs->rlc);
	}

	*ret = res;
	return True;
}

static float ImpBP2x_F1(TPage_BP2x *pg, float MHz, float *Z) {
  float res;
  TRLC btIn;
  float w = MHz * M_PI * 2;
	if (!ImpBP2x_ZF1(pg, w, &btIn, &res)) return res;

	*Z = btIn.Z;
	Form_R_math(pg->f_Rg, w);
	RLC_Add_Ser(&btIn, &pg->f_Rg->rlc);
	res /= btIn.Z;
	return res;
} 

static bool ImpBP2x_ZF2(TPage_BP2x *pg, float w, TRLC *btIn, float *ret) {
  float res;
	if (w < 0.000001) {
	    if (pg->cbCxOn->Checked || pg->cbCsOn->Checked) return 0;
	    res = pg->f_Rs->R->dval;
	    *ret = pg->f_Ra->R->dval / (pg->f_Ra->R->dval + res);
	    return False;
	}
	Form_R_math(pg->f_Ra, w);

	res = 1;
	RLC_Copy(btIn, &pg->f_Ra->rlc);
	if (pg->cbCxOn->Checked) {
	    res = btIn->Z;
	    Form_C_math(pg->f_Cx_F2, w);
	    RLC_Add_Ser(btIn, &pg->f_Cx_F2->rlc);
	    res /= btIn->Z;
	}

	Form_C_math(pg->f_C1_F2, w);
	Form_L_math(pg->f_L1_F2, w);
	Form_C_math(pg->f_C2_F2, w);
	Form_C_math(pg->f_C3_F2, w);
	Form_L_math(pg->f_L2_F2, w);

	RLC_Add_Par(btIn, &pg->f_C3_F2->rlc);
	RLC_Add_Par(btIn, &pg->f_L2_F2->rlc);
	res *= btIn->Z;

	RLC_Add_Ser(btIn, &pg->f_C2_F2->rlc);
	res /= btIn->Z;

	RLC_Add_Par(btIn, &pg->f_C1_F2->rlc);
	RLC_Add_Par(btIn, &pg->f_L1_F2->rlc);
	res *= btIn->Z;

	if (pg->cbCsOn->Checked) {
	    Form_C_math(pg->f_Cs_F2, w);
	    RLC_Add_Ser(btIn, &pg->f_Cs_F2->rlc);
	}

	*ret = res;
	return True;
}

static float ImpBP2x_F2(TPage_BP2x *pg, float MHz, float *Z) {
  float res;
  TRLC btIn;
  float w = MHz * M_PI * 2;
	if (!ImpBP2x_ZF2(pg, w, &btIn, &res)) return res;

	*Z = btIn.Z;
	Form_R_math(pg->f_Rg, w);
	RLC_Add_Ser(&btIn, &pg->f_Rg->rlc);
	res /= btIn.Z;
	return res;
}
static float ImpBP2x_F3(TPage_BP2x *pg, float MHz, float *Z) {
  float res, resF2;
  TRLC btIn, btInF2;
  float w = MHz * M_PI * 2;
	if (!ImpBP2x_ZF1(pg, w, &btIn, &res)) return res;
	if (!ImpBP2x_ZF2(pg, w, &btInF2, &resF2)) return res;
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &btInF2);
	res *= btIn.Z;
	*Z = btIn.Z;
	Form_R_math(pg->f_Rg, w);
	RLC_Add_Ser(&btIn, &pg->f_Rg->rlc);
	res /= btIn.Z;

//	ImpBP2x_MathIn_rc(pg, w, &btIn, &res, Z);
	return res;
}
static float ImpBP2x_F4(TPage_BP2x *pg, float MHz, float *Z) {
  float res, resF2;
  TRLC btIn, btInF2;
  float w = MHz * M_PI * 2;
	if (!ImpBP2x_ZF1(pg, w, &btIn, &res)) return res;
	if (!ImpBP2x_ZF2(pg, w, &btInF2, &resF2)) return res;
	resF2 /= btInF2.Z;
	RLC_Add_Par(&btInF2, &btIn);
	resF2 *= btInF2.Z;
	*Z = btInF2.Z;
	Form_R_math(pg->f_Rg, w);
	RLC_Add_Ser(&btInF2, &pg->f_Rg->rlc);
	resF2 /= btInF2.Z;
	return resF2;
}

void math_Filter_BP2x(TPage_BP2x *pg) {
  TPage_Filter *pgf = pg->pgfilt;
  float fcStep, fOff, fStep;
  TScale_pxy *pxy = pgf->xy_ind->dat->pxy;
  TScale_pxy *pxy2;
  TScale_pxy *pxy3 = NULL;
  float fc, Z, K;
  TScale_Point *sSP, *dSP;

	K = 1;
	if (pgf->cbRgRa->Checked)
	    K = (pg->f_Rg->rlc.val + pg->f_Ra->rlc.val) / pg->f_Ra->rlc.val;

	fOff = pgf->ScaleMHzoff;
	fStep = pgf->ScaleMHz;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpBP2x_F1(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}
	if (pg->cbF1ImpOn->Checked) pxy3 = pxy;

	pxy = pxy->next;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpBP2x_F2(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}
	if (pg->cbF2ImpOn->Checked) pxy3 = pxy;

	pxy = pxy->next;
	pxy->PaintDat = True;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpBP2x_F3(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}

	pxy = pxy->next;
	pxy->PaintDat = True;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpBP2x_F4(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}

	Z = pg->dv_Rg->dval;
	if (pxy3 && (Z > 0)) {
	    pxy2 = pgf->yInZ;
	    pxy2->PaintDat = True;
	    sSP = &pxy3->xy[0];
	    dSP = &pxy2->xy[0];
	    for (int i=0;i<pxy2->count;i++) {
		dSP->x = sSP->x;
		dSP->y = sSP->y2 / Z;
		dSP++; sSP++;
	    }
	} else {
	    pgf->yInZ->PaintDat = False;
	}

	SetCursor_Dat_txt(pgf);
	Refresh((TWins*)pgf->xy_ind);

}

static void SaveBP2xOnOff(TPage_BP2x *pg, TpcIni *pcIni) {
  int OnOff = 0;

	if (pg->cbCsOn->Checked) OnOff |= 0x01;
	if (pg->cbRsOn->Checked) OnOff |= 0x02;
	if (pg->cbCxOn->Checked) OnOff |= 0x04;
	if (pg->cbRxOn->Checked) OnOff |= 0x08;

	if (pg->cbF1On->Checked) OnOff |= 0x010;
	if (pg->cbF2On->Checked) OnOff |= 0x020;

	if (pg->cbF1ImpOn->Checked) OnOff |= 0x100;
	if (pg->cbF2ImpOn->Checked) OnOff |= 0x200;

	TpcIni_WR_int(pcIni, "BP2x", "OnOff", OnOff);
}

void BP2x_SetVisible(TPage_BP2x *pg, bool On) {
	SetVisible((TWins*)pg->f_Rg, pg->f_Rg->tag && On);
	SetVisible((TWins*)pg->f_Rs, pg->f_Rs->tag && On);
	SetVisible((TWins*)pg->f_Rx, pg->f_Rx->tag && On);
	SetVisible((TWins*)pg->f_Ra, pg->f_Ra->tag && On);
	SetVisible((TWins*)pg->f_Cs, pg->f_Cs->tag && On);
	SetVisible((TWins*)pg->f_Cx, pg->f_Cx->tag && On);
	SetVisible((TWins*)pg->f_C1, pg->f_C1->tag && On);
	SetVisible((TWins*)pg->f_C2, pg->f_C2->tag && On);
	SetVisible((TWins*)pg->f_C3, pg->f_C3->tag && On);
	SetVisible((TWins*)pg->f_L1, pg->f_L1->tag && On);
	SetVisible((TWins*)pg->f_L2, pg->f_L2->tag && On);

  TScale_pxy *pxy;
	if (On) {
	    pxy = pg->pgfilt->xy_ind->dat->pxy;
	    while (pxy) {
		pxy->PaintDat = False;
		pxy = pxy->next;
	    }
	    math_Filter_BP2x(pg);
	    ((TScale_pxy*)pg->cbF1On->tag)->PaintDat = pg->cbF1On->Checked;
	    ((TScale_pxy*)pg->cbF2On->tag)->PaintDat = pg->cbF2On->Checked;
	}
}

/* ************** BP2x Ext Ev ****************** */
bool ImpBP2x_Fx(TPage_BP2x *pg, float MHz, float *yVal, float *Z) {
	if (pg->cbF1ImpOn->Checked) {
		*yVal = ImpBP2x_F1(pg, MHz, Z);
	} else if (pg->cbF2ImpOn->Checked) {
		*yVal = ImpBP2x_F2(pg, MHz, Z);
	} else return False;
	return True;
}

/* ************** BP2x Ev ****************** */
static void cbBP2xXxOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_BP2x *pg = self->subClass;
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
	SaveBP2xOnOff(pg, pcIni);
	TpcIni_SaveAndFree(pcIni);
	RePaint((TWins*)pg->pnl);
	math_Filter_BP2x(pg);
}

static void cbBP2xFxOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_BP2x *pg = self->subClass;
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
	SaveBP2xOnOff(pg, pcIni);
	TpcIni_SaveAndFree(pcIni);
	((TScale_pxy*)self->tag)->PaintDat = ((TChkBox*)self)->Checked;
	Refresh((TWins*)pg->pgfilt->xy_ind);
}

static void cbBP2xFxImpOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_BP2x *pg = self->subClass;
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
	if (pg->cbF1ImpOn != self) TChkBox_SetChecked(pg->cbF1ImpOn, False);
	if (pg->cbF2ImpOn != self) TChkBox_SetChecked(pg->cbF2ImpOn, False);
	SaveBP2xOnOff(pg, pcIni);
	TpcIni_SaveAndFree(pcIni);
	math_Filter_BP2x(pg);
	Refresh((TWins*)pg->pgfilt->xy_ind);
}

static void pnl_BP2x_OnPaint(TWins *self) {
  TPage_BP2x *pg = self->subClass;
  int fu = 0;
  int x = pg->cbF1On->left + pg->cbF1On->width + 2;

	if (pg->cbCsOn->Checked) fu |= 0x01;
	if (pg->cbRsOn->Checked) fu |= 0x02;
	if (pg->cbCxOn->Checked) fu |= 0x04;
	if (pg->cbRxOn->Checked) fu |= 0x08;

	DrawBP2x(self->WinID, x, pg->cbF1On->top, fu);
	fu += 0x100;
	DrawBP2x(self->WinID, x, pg->cbF2On->top, fu);

	DrawPic(self->WinID, get_Pic_cOhm(), 
			pg->dv_Ra->left + pg->dv_Ra->width + 2,
			pg->dv_Ra->top + 5);

	str_toGraf(self, "pF", pg->dv_Cx->left + pg->dv_Cx->width + 1,
			       pg->dv_Cx->top + 3);

	str_toGraf(self, "pF", pg->dv_C3->left + pg->dv_C3->width + 1,
			       pg->dv_C3->top + 3);

	str_toGraf(self, "uH", pg->dv_L2->left + pg->dv_L2->width + 1,
			       pg->dv_L2->top + 3);
}

static void BP2x_DVal_OnValChange(TDVal *dv, double val) {
	if (dv->owner) TDVal_DoValChange((TDVal*)dv->owner, val, tdvfu_ChangeEv);
}

static void BP2x_fXxx_OnChange(TFormRLC* fRLC) {
	math_Filter_BP2x((TPage_BP2x*)fRLC->subClass);
}
static void BP2x_OnRSync(TFormRLC* fRLC, int val) {
  TPage_BP2x *pg = fRLC->subClass;
	Form_R_sync(fRLC, pg->f_Rg, val);
	Form_R_sync(fRLC, pg->f_Rs, val);
	Form_R_sync(fRLC, pg->f_Rx, val);
	Form_R_sync(fRLC, pg->f_Ra, val);
}
static void BP2x_OnLSync(TFormRLC* fRLC, int val) {
  TPage_BP2x *pg = fRLC->subClass;
	Form_L_sync(fRLC, pg->f_L1, val);
	Form_L_sync(fRLC, pg->f_L2, val);
}
static void BP2x_OnCSync(TFormRLC* fRLC, int val) {
  TPage_BP2x *pg = fRLC->subClass;
	Form_C_sync(fRLC, pg->f_Cs, val);
	Form_C_sync(fRLC, pg->f_Cx, val);
	Form_C_sync(fRLC, pg->f_C1, val);
	Form_C_sync(fRLC, pg->f_C2, val);
	Form_C_sync(fRLC, pg->f_C3, val);
}

/* ************** BP2x Create ****************** */
static void DVal_Create_Xxx(TWins *Parent, TDVal **dv, TFormRLConoff **onoff, int *x, int y, int w, int nk, TPic *pic) {
  int tx = *x;
  TDVal *tdv = TDVal_CreateS(Parent, tx+19, y, w, 19);
	tdv->nk = nk;
	tdv->dv_fu = tdv_fu_LostFocusEnter;
	tdv->OnValChange = BP2x_DVal_OnValChange;
	*dv = tdv;
	*onoff = TFormRLConoff_CreateU(Parent, tx, y+1, 18, 16, pic);
	*x = tx + w + 21;
}

static TFormR *Create_f_BP2x_R(TPage_BP2x *pg, const char *Caption, const char *Gruppe, 
			const char *s, TDVal *dv, float val) {
  TFormR *fR = Create_Form_R(appDat.obj, Caption, Gruppe, s, dv, val);
	fR->subClass = pg;
	TDVal_Set_Value(dv, fR->R->dval);
	fR->OnChange = BP2x_fXxx_OnChange;
	fR->OnSync = BP2x_OnRSync;
	return fR;
}
static TFormL *Create_f_BP2x_L(TPage_BP2x *pg, const char *Caption, const char *Gruppe, 
			const char *s, TDVal *dv, float val) {
  TFormL *fL = Create_Form_L(appDat.obj, Caption, Gruppe, s, dv, val);
	fL->subClass = pg;
	TDVal_Set_Value(dv, fL->L->dval);
	fL->OnChange = BP2x_fXxx_OnChange;
	fL->OnSync = BP2x_OnLSync;
	return fL;
}
static TFormC *Create_f_BP2x_C(TPage_BP2x *pg, const char *Caption, const char *Gruppe, 
			const char *s, TDVal *dv, float val) {
  TFormC *fC = Create_Form_C(appDat.obj, Caption, Gruppe, s, dv, val);
	fC->subClass = pg;
	TDVal_Set_Value(dv, fC->C->dval);
	fC->OnChange = BP2x_fXxx_OnChange;
	fC->OnSync = BP2x_OnCSync;
	return fC;
}

static TPage *page_Add_BP2x_Create(TPageControl *Parent) {
	return TPage_CreateParamAlg(Parent, sizeof(TPage_BP2x));
}

TPage_BP2x *page_Add_BP2x(TPageControl *pagc, TPage_Filter *pgfilt) {
  TPage_BP2x *hp1;
  TWins *pnl;
  TpcIni *pcIni;
  int x, y, m;

	pcIni = get_appIni();

	hp1 = (TPage_BP2x*) TPageControl_AddNewUserPage(pagc, "BP2x", page_Add_BP2x_Create);
	hp1->subClass = hp1;
	hp1->pgfilt = pgfilt;

	pnl = TWins_CreateBA((TWins*)hp1, 0, 0, 318, 21, 0, alLeft);
	hp1->pnl = pnl;
	hp1->pnl->OnPaint = pnl_BP2x_OnPaint;

/* Rg Rs Rx Ra */
	x = 0;
	DVal_Create_Xxx(pnl, &hp1->dv_Rg, &hp1->OnOff_Rg, &x, 2, 54, 1, get_Pic_cRg());
	DVal_Create_Xxx(pnl, &hp1->dv_Rs, &hp1->OnOff_Rs, &x, 2, 54, 1, get_Pic_cRs());
	DVal_Create_Xxx(pnl, &hp1->dv_Rx, &hp1->OnOff_Rx, &x, 2, 54, 1, get_Pic_cRx());
	DVal_Create_Xxx(pnl, &hp1->dv_Ra, &hp1->OnOff_Ra, &x, 2, 54, 1, get_Pic_cRa());
/* Cs Ca */
	x = 0;
	DVal_Create_Xxx(pnl, &hp1->dv_Cs, &hp1->OnOff_Cs, &x, 22, 54, 1, get_Pic_cCs());
	DVal_Create_Xxx(pnl, &hp1->dv_Cx, &hp1->OnOff_Cx, &x, 22, 54, 1, get_Pic_cCx());
/* C1 C2 C3 */
	x = 0;
	DVal_Create_Xxx(pnl, &hp1->dv_C1, &hp1->OnOff_C1, &x, 42, 54, 1, get_Pic_cC1());
	DVal_Create_Xxx(pnl, &hp1->dv_C2, &hp1->OnOff_C2, &x, 42, 54, 1, get_Pic_cC2());
	DVal_Create_Xxx(pnl, &hp1->dv_C3, &hp1->OnOff_C3, &x, 42, 54, 1, get_Pic_cC3());
/* L1 L2 L3 */
	x = 0;
	DVal_Create_Xxx(pnl, &hp1->dv_L1, &hp1->OnOff_L1, &x, 62, 54, 3, get_Pic_cL1());
	DVal_Create_Xxx(pnl, &hp1->dv_L2, &hp1->OnOff_L2, &x, 62, 54, 3, get_Pic_cL2());

	m = TpcIni_RD_int(pcIni, "BP2x", "OnOff", 0xff05);

	hp1->cbCsOn = TChkBox_CreateS(pnl, 2, 82, 60, 16, "Cs on");
	hp1->cbRsOn = TChkBox_CreateS(pnl, 72, 82, 60, 16, "Rs on");
	hp1->cbCxOn = TChkBox_CreateS(pnl, 142, 82, 60, 16, "Cx on");
	hp1->cbRxOn = TChkBox_CreateS(pnl, 212, 82, 60, 16, "Rx on");
	hp1->cbCsOn->OnClick = cbBP2xXxOn_OnClick;
	hp1->cbRsOn->OnClick = cbBP2xXxOn_OnClick;
	hp1->cbCxOn->OnClick = cbBP2xXxOn_OnClick;
	hp1->cbRxOn->OnClick = cbBP2xXxOn_OnClick;
	hp1->cbCsOn->Checked = ((m & 0x01) != 0);
	hp1->cbRsOn->Checked = ((m & 0x02) != 0);
	hp1->cbCxOn->Checked = ((m & 0x04) != 0);
	hp1->cbRxOn->Checked = ((m & 0x08) != 0);

/* Filter On/Off */
	y = 105;
	hp1->cbF1On = TChkBox_CreateS(pnl, 2, y, 38, 17, "F1");
	hp1->cbF1ImpOn = TChkBox_CreateS(pnl, 2, y+17, 38, 17, "Imp");
	SetColor((TWins*)hp1->cbF1On, 0xa0a0ff);
	SetColor((TWins*)hp1->cbF1ImpOn, 0xe0e080);

/* Filter 2 */
	y += 45;
	x = 0;
	DVal_Create_Xxx(pnl, &hp1->dv_Cs_F2, &hp1->OnOff_Cs_F2, &x, y, 54, 1, get_Pic_cCs());
	DVal_Create_Xxx(pnl, &hp1->dv_Cx_F2, &hp1->OnOff_Cx_F2, &x, y, 54, 1, get_Pic_cCx());
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &hp1->dv_C1_F2, &hp1->OnOff_C1_F2, &x, y, 54, 1, get_Pic_cC1());
	DVal_Create_Xxx(pnl, &hp1->dv_C2_F2, &hp1->OnOff_C2_F2, &x, y, 54, 1, get_Pic_cC2());
	DVal_Create_Xxx(pnl, &hp1->dv_C3_F2, &hp1->OnOff_C3_F2, &x, y, 54, 1, get_Pic_cC3());
/* L1 L2 L3 */
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &hp1->dv_L1_F2, &hp1->OnOff_L1_F2, &x, y, 54, 3, get_Pic_cL1());
	DVal_Create_Xxx(pnl, &hp1->dv_L2_F2, &hp1->OnOff_L2_F2, &x, y, 54, 3, get_Pic_cL2());

	y += 26;
	hp1->cbF2On = TChkBox_CreateS(pnl, 2, y, 38, 17, "F2");
	hp1->cbF2ImpOn = TChkBox_CreateS(pnl, 2, y+17, 38, 17, "Imp");
	SetColor((TWins*)hp1->cbF2On, 0xa0ffa0);
	SetColor((TWins*)hp1->cbF2ImpOn, 0xe0e080);

	hp1->cbF1On->OnClick = cbBP2xFxOn_OnClick;
	hp1->cbF2On->OnClick = cbBP2xFxOn_OnClick;
	hp1->cbF1On->Checked = ((m & 0x010) != 0);
	hp1->cbF2On->Checked = ((m & 0x020) != 0);

	hp1->cbF1ImpOn->OnClick = cbBP2xFxImpOn_OnClick;
	hp1->cbF2ImpOn->OnClick = cbBP2xFxImpOn_OnClick;
	hp1->cbF1ImpOn->Checked = ((m & 0x700) == 0x100);
	hp1->cbF2ImpOn->Checked = ((m & 0x700) == 0x200);

	hp1->f_Rg = Create_f_BP2x_R(hp1, "BP2x Rg", "BP2x", "Rg", hp1->dv_Rg, 50);
	hp1->f_Rs = Create_f_BP2x_R(hp1, "BP2x Rs", "BP2x", "Rs", hp1->dv_Rs, 50);
	hp1->f_Rx = Create_f_BP2x_R(hp1, "BP2x Rx", "BP2x", "Rx", hp1->dv_Rx, 50);
	hp1->f_Ra = Create_f_BP2x_R(hp1, "BP2x Ra", "BP2x", "Ra", hp1->dv_Ra, 50);

	hp1->f_Cs = Create_f_BP2x_C(hp1, "BP2x Cs", "BP2x", "Cs", hp1->dv_Cs, 4);
	hp1->f_Cx = Create_f_BP2x_C(hp1, "BP2x Cx", "BP2x", "Cx", hp1->dv_Cx, 4);
	hp1->f_C1 = Create_f_BP2x_C(hp1, "BP2x C1", "BP2x", "C1", hp1->dv_C1, 8);
	hp1->f_C2 = Create_f_BP2x_C(hp1, "BP2x C2", "BP2x", "C2", hp1->dv_C2, 1.2);
	hp1->f_C3 = Create_f_BP2x_C(hp1, "BP2x C3", "BP2x", "C3", hp1->dv_C3, 8);

	hp1->f_L1 = Create_f_BP2x_L(hp1, "BP2x L1", "BP2x", "L1", hp1->dv_L1, 0.033);
	hp1->f_L2 = Create_f_BP2x_L(hp1, "BP2x L2", "BP2x", "L2", hp1->dv_L2, 0.033);


	hp1->f_Cs_F2 = Create_f_BP2x_C(hp1, "BP2x Cs F2", "BP2x", "CsF2", hp1->dv_Cs_F2, 3);
	hp1->f_Cx_F2 = Create_f_BP2x_C(hp1, "BP2x Cx F2", "BP2x", "CxF2", hp1->dv_Cx_F2, 3);
	hp1->f_C1_F2 = Create_f_BP2x_C(hp1, "BP2x C1 F2", "BP2x", "C1F2", hp1->dv_C1_F2, 8);
	hp1->f_C2_F2 = Create_f_BP2x_C(hp1, "BP2x C2 F2", "BP2x", "C2F2", hp1->dv_C2_F2, 0.9);
	hp1->f_C3_F2 = Create_f_BP2x_C(hp1, "BP2x C3 F2", "BP2x", "C3F2", hp1->dv_C3_F2, 8);

	hp1->f_L1_F2 = Create_f_BP2x_L(hp1, "BP2x L1 F2", "BP2x", "L1F2", hp1->dv_L1_F2, 0.02);
	hp1->f_L2_F2 = Create_f_BP2x_L(hp1, "BP2x L2 F2", "BP2x", "L2F2", hp1->dv_L2_F2, 0.02);

	TFormRLConoff_SetForm(hp1->OnOff_Rg, (TForm*)hp1->f_Rg);
	TFormRLConoff_SetForm(hp1->OnOff_Rs, (TForm*)hp1->f_Rs);
	TFormRLConoff_SetForm(hp1->OnOff_Rx, (TForm*)hp1->f_Rx);
	TFormRLConoff_SetForm(hp1->OnOff_Ra, (TForm*)hp1->f_Ra);
	TFormRLConoff_SetForm(hp1->OnOff_Cs, (TForm*)hp1->f_Cs);
	TFormRLConoff_SetForm(hp1->OnOff_Cx, (TForm*)hp1->f_Cx);
	TFormRLConoff_SetForm(hp1->OnOff_C1, (TForm*)hp1->f_C1);
	TFormRLConoff_SetForm(hp1->OnOff_C2, (TForm*)hp1->f_C2);
	TFormRLConoff_SetForm(hp1->OnOff_C3, (TForm*)hp1->f_C3);
	TFormRLConoff_SetForm(hp1->OnOff_L1, (TForm*)hp1->f_L1);
	TFormRLConoff_SetForm(hp1->OnOff_L2, (TForm*)hp1->f_L2);

	TFormRLConoff_SetForm(hp1->OnOff_Cs_F2, (TForm*)hp1->f_Cs_F2);
	TFormRLConoff_SetForm(hp1->OnOff_Cx_F2, (TForm*)hp1->f_Cx_F2);
	TFormRLConoff_SetForm(hp1->OnOff_C1_F2, (TForm*)hp1->f_C1_F2);
	TFormRLConoff_SetForm(hp1->OnOff_C2_F2, (TForm*)hp1->f_C2_F2);
	TFormRLConoff_SetForm(hp1->OnOff_C3_F2, (TForm*)hp1->f_C3_F2);
	TFormRLConoff_SetForm(hp1->OnOff_L1_F2, (TForm*)hp1->f_L1_F2);
	TFormRLConoff_SetForm(hp1->OnOff_L2_F2, (TForm*)hp1->f_L2_F2);

	free_appIni(pcIni);

	return hp1;
}

