/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "ElMath.h"
#include "ElGraf.h"
#include <math.h>

static void str_toGraf(TWins *Wins, const char *s, int x, int y) {
	vds_SetStr(&Wins->vds, s);
	DrawColorString(x, y, &Wins->vds);
}

static void DrawTP1(Window WinID, int x, int y, int fu) {
  int xa;
  int x2 = x;
  int y2 = y+30;

	Draw_xP1_Rg(WinID, &x2, &xa, y, y2, fu);

	if ((fu & 0x0300) == 0x000) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	}

	if ((fu & 0x0300) == 0x100) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lh(WinID, &x2, y, get_Pic_cL1());
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC4());
	}
	if ((fu & 0x0300) == 0x200) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lh(WinID, &x2, y, get_Pic_cL1());
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC2());
	    DrawTP1_Lh(WinID, &x2, y, get_Pic_cL2());
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC4());
	}

	if ((fu & 0x0300) == 0x300) {
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC1());
	    DrawTP1_Lh(WinID, &x2, y, get_Pic_cL1());
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC2());
	    DrawTP1_Lh(WinID, &x2, y, get_Pic_cL2());
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC3());
	    DrawTP1_Lh(WinID, &x2, y, get_Pic_cL3());
	    DrawTP1_Cv(WinID, &x2, 0, y, get_Pic_cC4());
	}

	Draw_xP1_Ra(WinID, x2, xa, y, y2, fu);
}

static bool ImpTP1_MathOut_rc(TPage_TP1 *pg, float w, TRLC *btIn, float *ret) {
  float res;
	if (w < 0.000001) {
	    if (pg->cbCxOn->Checked || pg->cbCsOn->Checked) return 0;
	    res = pg->f_Rs->R->dval;
	    if (pg->cbRxOn->Checked) res += pg->f_Rx->R->dval;
	    *ret = pg->f_Ra->R->dval / (pg->f_Ra->R->dval + res);
	    return False;
	}
	Form_R_math(pg->f_Ra, w);

	res = 1;
	RLC_Copy(btIn, &pg->f_Ra->rlc);
	if (pg->cbCxOn->Checked || pg->cbRxOn->Checked) {
	    res = btIn->Z;
	    if (pg->cbCxOn->Checked) {
		Form_C_math(pg->f_Cx, w);
		RLC_Add_Ser(btIn, &pg->f_Cx->rlc);
	    }
	    if (pg->cbRxOn->Checked) {
		Form_R_math(pg->f_Rx, w);
		RLC_Add_Ser(btIn, &pg->f_Rx->rlc);
	    }
	    res /= btIn->Z;
	}
	*ret = res;
	return True;
}

static void ImpTP1_MathIn_rc(TPage_TP1 *pg, float w, TRLC *btIn, float *ret, float *Z) {
	if (pg->cbCsOn->Checked) {
	    Form_C_math(pg->f_Cs, w);
	    RLC_Add_Ser(btIn, &pg->f_Cs->rlc);
	}
	*Z = 0;
	if (pg->cbRsOn->Checked) {
	    Form_R_math(pg->f_Rs, w);
	    RLC_Add_Ser(btIn, &pg->f_Rs->rlc);
	}
	if (pg->f_Rg->R->dval > 0) {
	    *Z = btIn->Z;
	    Form_R_math(pg->f_Rg, w);
	    RLC_Add_Ser(btIn, &pg->f_Rg->rlc);
	}
  float res = *ret;
	res /= btIn->Z;
	*ret = res;
}

static float ImpTP1_F1(TPage_TP1 *pg, float MHz, float *Z) {
  float res;
  TRLC btIn;
  float w = MHz * M_PI * 2;
	if (!ImpTP1_MathOut_rc(pg, w, &btIn, &res)) return res;
	Form_C_math(pg->f_C1, w);
	RLC_Add_Par(&btIn, &pg->f_C1->rlc);
	res *= btIn.Z;

	ImpTP1_MathIn_rc(pg, w, &btIn, &res, Z);
	return res;
} 
static float ImpTP1_F2(TPage_TP1 *pg, float MHz, float *Z) {
  float res;
  TRLC btIn;
  float w = MHz * M_PI * 2;
	if (!ImpTP1_MathOut_rc(pg, w, &btIn, &res)) return res;
	Form_C_math(pg->f_C1, w);
	Form_L_math(pg->f_L1, w);
	Form_C_math(pg->f_C4, w);
	RLC_Add_Par(&btIn, &pg->f_C4->rlc);
	res *= btIn.Z;
	RLC_Add_Ser(&btIn, &pg->f_L1->rlc);
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &pg->f_C1->rlc);
	res *= btIn.Z;

	ImpTP1_MathIn_rc(pg, w, &btIn, &res, Z);
	return res;
}
static float ImpTP1_F3(TPage_TP1 *pg, float MHz, float *Z) {
  float res;
  TRLC btIn;
  float w = MHz * M_PI * 2;
	if (!ImpTP1_MathOut_rc(pg, w, &btIn, &res)) return res;
	Form_C_math(pg->f_C1, w);
	Form_C_math(pg->f_C2, w);
	Form_L_math(pg->f_L1, w);
	Form_L_math(pg->f_L2, w);
	Form_C_math(pg->f_C4, w);
	RLC_Add_Par(&btIn, &pg->f_C4->rlc);
	res *= btIn.Z;
	RLC_Add_Ser(&btIn, &pg->f_L2->rlc);
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &pg->f_C2->rlc);
	res *= btIn.Z;
	RLC_Add_Ser(&btIn, &pg->f_L1->rlc);
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &pg->f_C1->rlc);
	res *= btIn.Z;

	ImpTP1_MathIn_rc(pg, w, &btIn, &res, Z);
	return res;
}
static float ImpTP1_F4(TPage_TP1 *pg, float MHz, float *Z) {
  float res;
  TRLC btIn;
  float w = MHz * M_PI * 2;
	if (!ImpTP1_MathOut_rc(pg, w, &btIn, &res)) return res;
	Form_C_math(pg->f_C1, w);
	Form_C_math(pg->f_C2, w);
	Form_C_math(pg->f_C3, w);
	Form_L_math(pg->f_L1, w);
	Form_L_math(pg->f_L2, w);
	Form_L_math(pg->f_L3, w);
	Form_C_math(pg->f_C4, w);
	RLC_Add_Par(&btIn, &pg->f_C4->rlc);
	res *= btIn.Z;
	RLC_Add_Ser(&btIn, &pg->f_L3->rlc);
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &pg->f_C3->rlc);
	res *= btIn.Z;
	RLC_Add_Ser(&btIn, &pg->f_L2->rlc);
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &pg->f_C2->rlc);
	res *= btIn.Z;
	RLC_Add_Ser(&btIn, &pg->f_L1->rlc);
	res /= btIn.Z;
	RLC_Add_Par(&btIn, &pg->f_C1->rlc);
	res *= btIn.Z;

	ImpTP1_MathIn_rc(pg, w, &btIn, &res, Z);
	return res;
}

void math_Filter_TP1(TPage_TP1 *pg) {
  TPage_Filter *pgf = pg->pgfilt;
  float fcStep, fOff, fStep;
  TScale_pxy *pxy = pgf->xy_ind->dat->pxy;
  TScale_pxy *pxy2;
  TScale_pxy *pxy3 = NULL;
  float fc, Z, K;
  TScale_Point *sSP, *dSP;

// printf("%s(%d): math_Filter_TP1\n", FL);
	K = 1;
	if (pgf->cbRgRa->Checked)
	    K = (pg->f_Rg->rlc.val + pg->f_Ra->rlc.val) / pg->f_Ra->rlc.val;

	fOff = pgf->ScaleMHzoff;
	fStep = pgf->ScaleMHz;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpTP1_F1(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}
	if (pg->cbF1ImpOn->Checked) pxy3 = pxy;

	pxy = pxy->next;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpTP1_F2(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}
	if (pg->cbF2ImpOn->Checked) pxy3 = pxy;

	pxy = pxy->next;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpTP1_F3(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}
	if (pg->cbF3ImpOn->Checked) pxy3 = pxy;


	pxy = pxy->next;
	fcStep = fStep / (pxy->count-1);
	for (int i=0;i<pxy->count;i++) {
	    fc = i * fcStep + fOff;
	    pxy->xy[i].x = fc;
	    pxy->xy[i].y = K * ImpTP1_F4(pg, fc, &Z);
	    pxy->xy[i].y2 = Z;
	}
	if (pg->cbF4ImpOn->Checked) pxy3 = pxy;

	Z = pg->dv_Rg->dval;
	if (pxy3 && (Z > 0)) {
	    pxy2 = pgf->yInZ;
	    pxy2->PaintDat = True;
	    sSP = &pxy3->xy[0];
	    dSP = &pxy2->xy[0];
	    for (int i=0;i<pxy2->count;i++) {
		dSP->x = sSP->x;
		dSP->y = sSP->y2 / Z;
		dSP++; sSP++;
	    }
	} else {
	    pgf->yInZ->PaintDat = False;
	}

	SetCursor_Dat_txt(pgf);
	Refresh((TWins*)pgf->xy_ind);

}

static void SaveTP1OnOff(TPage_TP1 *pg, TpcIni *pcIni) {
  int OnOff = 0;

	if (pg->cbCsOn->Checked) OnOff |= 0x01;
	if (pg->cbRsOn->Checked) OnOff |= 0x02;
	if (pg->cbCxOn->Checked) OnOff |= 0x04;
	if (pg->cbRxOn->Checked) OnOff |= 0x08;

	if (pg->cbF1On->Checked) OnOff |= 0x010;
	if (pg->cbF2On->Checked) OnOff |= 0x020;
	if (pg->cbF3On->Checked) OnOff |= 0x040;
	if (pg->cbF4On->Checked) OnOff |= 0x080;

	if (pg->cbF1ImpOn->Checked) OnOff |= 0x100;
	if (pg->cbF2ImpOn->Checked) OnOff |= 0x200;
	if (pg->cbF3ImpOn->Checked) OnOff |= 0x300;
	if (pg->cbF4ImpOn->Checked) OnOff |= 0x400;

	TpcIni_WR_int(pcIni, "TP1", "OnOff", OnOff);
}

void TP1_SetVisible(TPage_TP1 *pg, bool On) {
	SetVisible((TWins*)pg->f_Rg, pg->f_Rg->tag && On);
	SetVisible((TWins*)pg->f_Rs, pg->f_Rs->tag && On);
	SetVisible((TWins*)pg->f_Rx, pg->f_Rx->tag && On);
	SetVisible((TWins*)pg->f_Ra, pg->f_Ra->tag && On);
	SetVisible((TWins*)pg->f_Cs, pg->f_Cs->tag && On);
	SetVisible((TWins*)pg->f_Cx, pg->f_Cx->tag && On);
	SetVisible((TWins*)pg->f_C1, pg->f_C1->tag && On);
	SetVisible((TWins*)pg->f_C2, pg->f_C2->tag && On);
	SetVisible((TWins*)pg->f_C3, pg->f_C3->tag && On);
	SetVisible((TWins*)pg->f_C4, pg->f_C4->tag && On);
	SetVisible((TWins*)pg->f_L1, pg->f_L1->tag && On);
	SetVisible((TWins*)pg->f_L2, pg->f_L2->tag && On);
	SetVisible((TWins*)pg->f_L3, pg->f_L3->tag && On);
  TScale_pxy *pxy;
	if (On) {
	    pxy = pg->pgfilt->xy_ind->dat->pxy;
	    while (pxy) {
		pxy->PaintDat = False;
		pxy = pxy->next;
	    }
	    math_Filter_TP1(pg);
	    ((TScale_pxy*)pg->cbF1On->tag)->PaintDat = pg->cbF1On->Checked;
	    ((TScale_pxy*)pg->cbF2On->tag)->PaintDat = pg->cbF2On->Checked;
	    ((TScale_pxy*)pg->cbF3On->tag)->PaintDat = pg->cbF3On->Checked;
	    ((TScale_pxy*)pg->cbF4On->tag)->PaintDat = pg->cbF4On->Checked;
	}
}

/* ************** TP1 Ext Ev ****************** */
bool ImpTP1_Fx(TPage_TP1 *pg, float MHz, float *yVal, float *Z) {
	if (pg->cbF1ImpOn->Checked) {
		*yVal = ImpTP1_F1(pg, MHz, Z);
	} else if (pg->cbF2ImpOn->Checked) {
		*yVal = ImpTP1_F2(pg, MHz, Z);
	} else if (pg->cbF3ImpOn->Checked) {
		*yVal = ImpTP1_F3(pg, MHz, Z);
	} else if (pg->cbF4ImpOn->Checked) {
		*yVal = ImpTP1_F4(pg, MHz, Z);
	} else return False;
	return True;
}

/* ************** TP1 Ev ****************** */
static void cbTP1XxOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_TP1 *pg = self->subClass;
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
	SaveTP1OnOff(pg, pcIni);
	TpcIni_SaveAndFree(pcIni);
	RePaint((TWins*)pg->pnl);
	math_Filter_TP1(pg);
}

static void cbTP1FxOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_TP1 *pg = self->subClass;
//  TPage_Filter *pg = self->subClass;
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
	SaveTP1OnOff(pg, pcIni);
	TpcIni_SaveAndFree(pcIni);
	((TScale_pxy*)self->tag)->PaintDat = ((TChkBox*)self)->Checked;
	Refresh((TWins*)pg->pgfilt->xy_ind);
}

static void cbTP1FxImpOn_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_TP1 *pg = self->subClass;
  TpcIni *pcIni = TpcIni_Create(appDat.inifile_app);
	if (pg->cbF1ImpOn != self) TChkBox_SetChecked(pg->cbF1ImpOn, False);
	if (pg->cbF2ImpOn != self) TChkBox_SetChecked(pg->cbF2ImpOn, False);
	if (pg->cbF3ImpOn != self) TChkBox_SetChecked(pg->cbF3ImpOn, False);
	if (pg->cbF4ImpOn != self) TChkBox_SetChecked(pg->cbF4ImpOn, False);
	SaveTP1OnOff(pg, pcIni);
	TpcIni_SaveAndFree(pcIni);
	math_Filter_TP1(pg);
	Refresh((TWins*)pg->pgfilt->xy_ind);
}

static void pnl_TP1_OnPaint(TWins *self) {
  TPage_TP1 *pg = self->subClass;
  int fu = 0;
  int x = pg->cbF1On->left + pg->cbF1On->width + 2;

	if (pg->cbCsOn->Checked) fu |= 0x01;
	if (pg->cbRsOn->Checked) fu |= 0x02;
	if (pg->cbCxOn->Checked) fu |= 0x04;
	if (pg->cbRxOn->Checked) fu |= 0x08;

	DrawTP1(self->WinID, x, pg->cbF1On->top, fu);
	fu += 0x100;
	DrawTP1(self->WinID, x, pg->cbF2On->top, fu);
	fu += 0x100;
	DrawTP1(self->WinID, x, pg->cbF3On->top, fu);
	fu += 0x100;
	DrawTP1(self->WinID, x, pg->cbF4On->top, fu);

	DrawPic(self->WinID, get_Pic_cOhm(), 
			pg->dv_Ra->left + pg->dv_Ra->width + 2,
			pg->dv_Ra->top + 5);

	str_toGraf(self, "pF", pg->dv_Cx->left + pg->dv_Cx->width + 1,
			       pg->dv_Cx->top + 3);

	str_toGraf(self, "pF", pg->dv_C4->left + pg->dv_C4->width + 1,
			       pg->dv_C4->top + 3);

	str_toGraf(self, "uH", pg->dv_L3->left + pg->dv_L3->width + 1,
			       pg->dv_L3->top + 3);
}

static void TP1_DVal_OnValChange(TDVal *dv, double val) {
	if (dv->owner) TDVal_DoValChange((TDVal*)dv->owner, val, tdvfu_ChangeEv);
}

static void TP1_fXxx_OnChange(TFormRLC* fRLC) {
	math_Filter_TP1((TPage_TP1*)fRLC->subClass);
}
static void TP1_OnRSync(TFormRLC* fRLC, int val) {
  TPage_TP1 *pg = fRLC->subClass;
	Form_R_sync(fRLC, pg->f_Rg, val);
	Form_R_sync(fRLC, pg->f_Rs, val);
	Form_R_sync(fRLC, pg->f_Rx, val);
	Form_R_sync(fRLC, pg->f_Ra, val);
}
static void TP1_OnLSync(TFormRLC* fRLC, int val) {
  TPage_TP1 *pg = fRLC->subClass;
	Form_L_sync(fRLC, pg->f_L1, val);
	Form_L_sync(fRLC, pg->f_L2, val);
	Form_L_sync(fRLC, pg->f_L3, val);
}
static void TP1_OnCSync(TFormRLC* fRLC, int val) {
  TPage_TP1 *pg = fRLC->subClass;
	Form_C_sync(fRLC, pg->f_Cs, val);
	Form_C_sync(fRLC, pg->f_Cx, val);
	Form_C_sync(fRLC, pg->f_C1, val);
	Form_C_sync(fRLC, pg->f_C2, val);
	Form_C_sync(fRLC, pg->f_C3, val);
	Form_C_sync(fRLC, pg->f_C4, val);
}

/* ************** TP1 Create ****************** */
static void DVal_Create_Xxx(TWins *Parent, TDVal **dv, TFormRLConoff **onoff, int *x, int y, int w, int nk, TPic *pic) {
  int tx = *x;
  TDVal *tdv = TDVal_CreateS(Parent, tx+19, y, w, 19);
	tdv->nk = nk;
	tdv->dv_fu = tdv_fu_LostFocusEnter;
	tdv->OnValChange = TP1_DVal_OnValChange;
	*dv = tdv;
	*onoff = TFormRLConoff_CreateU(Parent, tx, y+1, 18, 16, pic);
	*x = tx + w + 21;
}

static TFormR *Create_f_TP1_R(TPage_TP1 *pg, const char *Caption, const char *Gruppe, 
			const char *s, TDVal *dv, float val) {
  TFormR *fR = Create_Form_R(appDat.obj, Caption, Gruppe, s, dv, val);
	fR->subClass = pg;
	TDVal_Set_Value(dv, fR->R->dval);
	fR->OnChange = TP1_fXxx_OnChange;
	fR->OnSync = TP1_OnRSync;
	return fR;
}
static TFormL *Create_f_TP1_L(TPage_TP1 *pg, const char *Caption, const char *Gruppe, 
			const char *s, TDVal *dv, float val) {
  TFormL *fL = Create_Form_L(appDat.obj, Caption, Gruppe, s, dv, val);
	fL->subClass = pg;
	TDVal_Set_Value(dv, fL->L->dval);
	fL->OnChange = TP1_fXxx_OnChange;
	fL->OnSync = TP1_OnLSync;
	return fL;
}
static TFormC *Create_f_TP1_C(TPage_TP1 *pg, const char *Caption, const char *Gruppe, 
			const char *s, TDVal *dv, float val) {
  TFormC *fC = Create_Form_C(appDat.obj, Caption, Gruppe, s, dv, val);
	fC->subClass = pg;
	TDVal_Set_Value(dv, fC->C->dval);
	fC->OnChange = TP1_fXxx_OnChange;
	fC->OnSync = TP1_OnCSync;
	return fC;
}

static TPage *page_Add_TP1_Create(TPageControl *Parent) {
	return TPage_CreateParamAlg(Parent, sizeof(TPage_TP1));
}

TPage_TP1 *page_Add_TP1(TPageControl *pagc, TPage_Filter *pgfilt) {
  TPage_TP1 *self;
  TWins *pnl;
  TpcIni *pcIni;
  int x, y, m;

	pcIni = get_appIni();

	self = (TPage_TP1*) TPageControl_AddNewUserPage(pagc, "TP1", page_Add_TP1_Create);
	self->subClass = self;
	self->pgfilt = pgfilt;

	pnl = TWins_CreateBA((TWins*)self, 0, 0, 318, 21, 0, alLeft);
	self->pnl = pnl;
	self->pnl->OnPaint = pnl_TP1_OnPaint;

/* Rg Rs Rx Ra */
	x = 0;
	DVal_Create_Xxx(pnl, &self->dv_Rg, &self->OnOff_Rg, &x, 2, 54, 1, get_Pic_cRg());
	DVal_Create_Xxx(pnl, &self->dv_Rs, &self->OnOff_Rs, &x, 2, 54, 1, get_Pic_cRs());
	DVal_Create_Xxx(pnl, &self->dv_Rx, &self->OnOff_Rx, &x, 2, 54, 1, get_Pic_cRx());
	DVal_Create_Xxx(pnl, &self->dv_Ra, &self->OnOff_Ra, &x, 2, 54, 1, get_Pic_cRa());
/* Cs Ca */
	x = 0;
	DVal_Create_Xxx(pnl, &self->dv_Cs, &self->OnOff_Cs, &x, 22, 54, 1, get_Pic_cCs());
	DVal_Create_Xxx(pnl, &self->dv_Cx, &self->OnOff_Cx, &x, 22, 54, 1, get_Pic_cCx());
/* C1 C2 C3 */
	x = 0;
	DVal_Create_Xxx(pnl, &self->dv_C1, &self->OnOff_C1, &x, 42, 54, 1, get_Pic_cC1());
	DVal_Create_Xxx(pnl, &self->dv_C2, &self->OnOff_C2, &x, 42, 54, 1, get_Pic_cC2());
	DVal_Create_Xxx(pnl, &self->dv_C3, &self->OnOff_C3, &x, 42, 54, 1, get_Pic_cC3());
	DVal_Create_Xxx(pnl, &self->dv_C4, &self->OnOff_C4, &x, 42, 54, 1, get_Pic_cC4());
/* L1 L2 L3 */
	x = 0;
	DVal_Create_Xxx(pnl, &self->dv_L1, &self->OnOff_L1, &x, 62, 54, 3, get_Pic_cL1());
	DVal_Create_Xxx(pnl, &self->dv_L2, &self->OnOff_L2, &x, 62, 54, 3, get_Pic_cL2());
	DVal_Create_Xxx(pnl, &self->dv_L3, &self->OnOff_L3, &x, 62, 54, 3, get_Pic_cL3());

	m = TpcIni_RD_int(pcIni, "TP1", "OnOff", 0xffe0);

	self->cbCsOn = TChkBox_CreateS(pnl, 2, 82, 60, 16, "Cs on");
	self->cbRsOn = TChkBox_CreateS(pnl, 72, 82, 60, 16, "Rs on");
	self->cbCxOn = TChkBox_CreateS(pnl, 142, 82, 60, 16, "Cx on");
	self->cbRxOn = TChkBox_CreateS(pnl, 212, 82, 60, 16, "Rx on");
	self->cbCsOn->OnClick = cbTP1XxOn_OnClick;
	self->cbRsOn->OnClick = cbTP1XxOn_OnClick;
	self->cbCxOn->OnClick = cbTP1XxOn_OnClick;
	self->cbRxOn->OnClick = cbTP1XxOn_OnClick;
	self->cbCsOn->Checked = ((m & 0x01) != 0);
	self->cbRsOn->Checked = ((m & 0x02) != 0);
	self->cbCxOn->Checked = ((m & 0x04) != 0);
	self->cbRxOn->Checked = ((m & 0x08) != 0);

/* Filter On/Off */
	y = 105;
	self->cbF1On = TChkBox_CreateS(pnl, 2, y, 38, 17, "F1");
	self->cbF1ImpOn = TChkBox_CreateS(pnl, 2, y+17, 38, 17, "Imp");
	SetColor((TWins*)self->cbF1On, 0xa0a0ff);
	SetColor((TWins*)self->cbF1ImpOn, 0xe0e080);
	y += 45;
	self->cbF2On = TChkBox_CreateS(pnl, 2, y, 38, 17, "F2");
	self->cbF2ImpOn = TChkBox_CreateS(pnl, 2, y+17, 38, 17, "Imp");
	SetColor((TWins*)self->cbF2On, 0xa0ffa0);
	SetColor((TWins*)self->cbF2ImpOn, 0xe0e080);
	y += 45;
	self->cbF3On = TChkBox_CreateS(pnl, 2, y, 38, 17, "F3");
	self->cbF3ImpOn = TChkBox_CreateS(pnl, 2, y+17, 38, 17, "Imp");
	SetColor((TWins*)self->cbF3On, 0xff8080);
	SetColor((TWins*)self->cbF3ImpOn, 0xe0e080);
	y += 45;
	self->cbF4On = TChkBox_CreateS(pnl, 2, y, 38, 17, "F4");
	self->cbF4ImpOn = TChkBox_CreateS(pnl, 2, y+17, 38, 17, "Imp");
	SetColor((TWins*)self->cbF4On, 0xff80ff);
	SetColor((TWins*)self->cbF4ImpOn, 0xe0e080);

	self->cbF1On->OnClick = cbTP1FxOn_OnClick;
	self->cbF2On->OnClick = cbTP1FxOn_OnClick;
	self->cbF3On->OnClick = cbTP1FxOn_OnClick;
	self->cbF4On->OnClick = cbTP1FxOn_OnClick;
	self->cbF1On->Checked = ((m & 0x010) != 0);
	self->cbF2On->Checked = ((m & 0x020) != 0);
	self->cbF3On->Checked = ((m & 0x040) != 0);
	self->cbF4On->Checked = ((m & 0x080) != 0);

	self->cbF1ImpOn->OnClick = cbTP1FxImpOn_OnClick;
	self->cbF2ImpOn->OnClick = cbTP1FxImpOn_OnClick;
	self->cbF3ImpOn->OnClick = cbTP1FxImpOn_OnClick;
	self->cbF4ImpOn->OnClick = cbTP1FxImpOn_OnClick;
	self->cbF1ImpOn->Checked = ((m & 0x700) == 0x100);
	self->cbF2ImpOn->Checked = ((m & 0x700) == 0x200);
	self->cbF3ImpOn->Checked = ((m & 0x700) == 0x300);
	self->cbF4ImpOn->Checked = ((m & 0x700) == 0x400);

	self->f_Rg = Create_f_TP1_R(self, "TP1 Rg", "TP1", "Rg", self->dv_Rg, 50);
	self->f_Rs = Create_f_TP1_R(self, "TP1 Rs", "TP1", "Rs", self->dv_Rs, 50);
	self->f_Rx = Create_f_TP1_R(self, "TP1 Rx", "TP1", "Rx", self->dv_Rx, 50);
	self->f_Ra = Create_f_TP1_R(self, "TP1 Ra", "TP1", "Ra", self->dv_Ra, 50);

	self->f_Cs = Create_f_TP1_C(self, "TP1 Cs", "TP1", "Cs", self->dv_Cs, 100);
	self->f_Cx = Create_f_TP1_C(self, "TP1 Cx", "TP1", "Cx", self->dv_Cx, 100);
	self->f_C1 = Create_f_TP1_C(self, "TP1 C1", "TP1", "C1", self->dv_C1, 8);
	self->f_C2 = Create_f_TP1_C(self, "TP1 C2", "TP1", "C2", self->dv_C2, 16);
	self->f_C3 = Create_f_TP1_C(self, "TP1 C3", "TP1", "C3", self->dv_C3, 16);
	self->f_C4 = Create_f_TP1_C(self, "TP1 C4", "TP1", "C4", self->dv_C4, 8);

	self->f_L1 = Create_f_TP1_L(self, "TP1 L1", "TP1", "L1", self->dv_L1, 0.027);
	self->f_L2 = Create_f_TP1_L(self, "TP1 L2", "TP1", "L2", self->dv_L2, 0.027);
	self->f_L3 = Create_f_TP1_L(self, "TP1 L3", "TP1", "L3", self->dv_L3, 0.027);

	TFormRLConoff_SetForm(self->OnOff_Rg, (TForm*)self->f_Rg);
	TFormRLConoff_SetForm(self->OnOff_Rs, (TForm*)self->f_Rs);
	TFormRLConoff_SetForm(self->OnOff_Rx, (TForm*)self->f_Rx);
	TFormRLConoff_SetForm(self->OnOff_Ra, (TForm*)self->f_Ra);
	TFormRLConoff_SetForm(self->OnOff_Cs, (TForm*)self->f_Cs);
	TFormRLConoff_SetForm(self->OnOff_Cx, (TForm*)self->f_Cx);
	TFormRLConoff_SetForm(self->OnOff_C1, (TForm*)self->f_C1);
	TFormRLConoff_SetForm(self->OnOff_C2, (TForm*)self->f_C2);
	TFormRLConoff_SetForm(self->OnOff_C3, (TForm*)self->f_C3);
	TFormRLConoff_SetForm(self->OnOff_C4, (TForm*)self->f_C4);
	TFormRLConoff_SetForm(self->OnOff_L1, (TForm*)self->f_L1);
	TFormRLConoff_SetForm(self->OnOff_L2, (TForm*)self->f_L2);
	TFormRLConoff_SetForm(self->OnOff_L3, (TForm*)self->f_L3);

	free_appIni(pcIni);

	return self;
}
