/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "ElMath.h"
#include "ElGraf.h"
#include "Microstrip.h"
#include <math.h>

static void str_toGraf(TWins *Wins, const char *s, int x, int y) {
	vds_SetStr(&Wins->vds, s);
	DrawColorString(x, y, &Wins->vds);
}

static void math_MS1(TPage_MS1 *pg) {
  TMicrostripDat md;

	md.W = pg->dv_W->dval;
	md.G = pg->dv_G->dval;
	md.H = pg->dv_H->dval;
	md.T = pg->dv_T->dval;
	md.L = pg->dv_L->dval;
	md.Er = pg->dv_Er->dval;
	md.Er_tan = pg->dv_tan->dval;
	md.Freq = pg->dv_Freq->dval;
	md.Coplanar = pg->cbCoplan->Checked;
	md.GroundPlane = pg->cbGroundp->Checked;

	Microstrip_math(&md);
	Microstrip_math_Freq(&md);
	Microstrip_math_L(&md);

	TDVal_Set_Value(pg->dv_Imp, md.Z);
	TDVal_Set_Value(pg->dv_nSm, md.nSm);
	TDVal_Set_Value(pg->dv_nS, md.nS);
	TDVal_Set_Value(pg->dv_Cm, md.pFm);
	TDVal_Set_Value(pg->dv_C, md.pF);
	TDVal_Set_Value(pg->dv_uHm, md.uHm);
	TDVal_Set_Value(pg->dv_uH, md.uH);
	TDVal_Set_Value(pg->dv_lo_dbm, md.dbm);
	TDVal_Set_Value(pg->dv_lo_db, md.db);
	TDVal_Set_Value(pg->dv_Ohmm, md.Ohmm);
	TDVal_Set_Value(pg->dv_Ohm, md.Ohm);
  int nk;
  float f = md.OhmP / 1000;
	nk = 3;
	if (f > 999.999) nk = 2;
	if (f > 9999.999) nk = 1;
	if (f > 99999.99) nk = 0;
	pg->dv_OhmP->nk = nk;
	TDVal_Set_Value(pg->dv_OhmP, f);

}

static void MS1_DVal_OnValChange(TDVal *dv, double val) {
	math_MS1(dv->subClass);
//	if (dv->owner) TDVal_DoValChange((TDVal*)dv->owner, val, tdvfu_ChangeEv);
}

static void DrawRect4ox(Window WinID, int RGB, int x, int y, int w, int h, int t, int x0) {
  XPoint points[5];
	points[0].x = x;
	points[0].y = y;
	points[1].x = x+w;
	points[1].y = y+t;
	points[2].x = x+w+x0;
	points[2].y = y+t+h;
	points[3].x = x+x0;
	points[3].y = y+h;
	points[4].x = x;
	points[4].y = y;
	XFillPolygon(appXDat.dpy, WinID, GetPaintGC(RGB), 
			points, 4, Complex, 0);
	XDrawLines(appXDat.dpy, WinID, GetPaintGC(0x000000), 
			points, 5, 0);
}

static void DrawQ3S_45g(Window WinID, int RGB, int x, int y, int w, int l, int h) {
  int xo = l; //y >> 1;
	DrawRect4ox(WinID, RGB, x, y+h, w, -h, 0, 0);
	DrawRect4ox(WinID, RGB, x+w, y+h, 0, -l, -h, xo);
	DrawRect4ox(WinID, RGB, x, y, w, -l, 0, xo);
}

static void DrawMS1(TPage_MS1 *pg, int x, int y, int w, int l, int fu) {
  TWins *ws = pg->pnl;
  int xt;
  int xl = 20;
  int xl_05 = xl >> 1;
  int xwm = x + ((w-xl)>>1);

	if (fu & 0x01) {
	    DrawRect4ox(ws->WinID, 0xff0000, x, y+20, w, -3, 0, 0);
	    DrawRect4ox(ws->WinID, 0xff0000, x+w, y+20, 0, -l, -3, l);
	}

	DrawQ3S_45g(ws->WinID, 0xc0c000, x, y, w, l, 17);

	DrawQ3S_45g(ws->WinID, 0xff0000, xwm, y-3, xl, l, 3);
	DrawLineRGB(ws->WinID, 0x000000, xwm-2, y+2, xwm-xl_05, y+10);
	DrawLineRGB(ws->WinID, 0x000000, xwm+xl-2, y+2, xwm+xl_05, y+10);
	str_toGraf(ws, "W", xwm-4, y+5);

	xt = x+w+28;
//	DrawLineRGB(ws->WinID, 0x000000, xwm+xl+2, y-3, xt, y-3);
	DrawLineRGB(ws->WinID, 0x000000, x+w+2, y, xt, y);
	DrawLineRGB(ws->WinID, 0x000000, x+w+2, y+17, xt, y+17);
	xt += 2;
	str_toGraf(ws, "T", xt, y-8);
	str_toGraf(ws, "H", xt, y+3);

	if (fu & 0x02) {
	    DrawQ3S_45g(ws->WinID, 0xff0000, x, y-3, xwm-x-8, l, 3);
	    DrawLineRGB(ws->WinID, 0x000000, xwm-10, y+2, xwm-xl_05-8, y+10);
	    str_toGraf(ws, "G", xwm-18, y+5);
	}
	if (fu & 0x04) {
	    xt = xwm+xl+8;
	    DrawQ3S_45g(ws->WinID, 0xff0000, xt, y-3, x+w-xt, l, 3);
	    DrawLineRGB(ws->WinID, 0x000000, x+w+2, y-3, x+w+28, y-3);
	} else {
	    DrawLineRGB(ws->WinID, 0x000000, xwm+xl+2, y-3, x+w+28, y-3);
	}
}

static void pnl_MS1_OnPaint(TWins *self) {
  TPage_MS1 *pg = self->subClass;
  const char *s_mm_mil = "mm";
	switch (pg->pgMS->mm_mil) {
	    case 1: 
	        s_mm_mil = "mil";
		break;
	} 

	str_toGraf(self, "W=", pg->dv_W->left - 18, pg->dv_W->top + 3);
	str_toGraf(self, "G=", pg->dv_G->left - 17, pg->dv_G->top + 3);
	str_toGraf(self, s_mm_mil, pg->dv_G->left + pg->dv_G->width + 1,
			       pg->dv_G->top + 3);

	str_toGraf(self, "H=", pg->dv_H->left - 16, pg->dv_H->top + 3);
	str_toGraf(self, s_mm_mil, pg->dv_H->left + pg->dv_H->width + 1,
			       pg->dv_H->top + 3);

	str_toGraf(self, "L=", pg->dv_L->left - 16, pg->dv_L->top + 3);
	str_toGraf(self, s_mm_mil, pg->dv_L->left + pg->dv_L->width + 1,
			       pg->dv_L->top + 3);

	str_toGraf(self, "T=", pg->dv_T->left - 16, pg->dv_T->top + 3);
	str_toGraf(self, s_mm_mil, pg->dv_T->left + pg->dv_T->width + 1,
			       pg->dv_T->top + 3);

	str_toGraf(self, "Er", pg->dv_Er->left - 16, pg->dv_Er->top + 3);

	str_toGraf(self, "Imp", pg->dv_Imp->left - 19, pg->dv_Imp->top + 3);
	DrawPic(self->WinID, get_Pic_cOhm(), 
			pg->dv_Imp->left + pg->dv_Imp->width + 2,
			pg->dv_Imp->top + 6);
	str_toGraf(self, "(X)", pg->dv_Imp->left + pg->dv_Imp->width + 13,
			       pg->dv_Imp->top + 3);


	str_toGraf(self, "C", pg->dv_Cm->left - 16, pg->dv_Cm->top + 3);
	str_toGraf(self, "pF/m", pg->dv_Cm->left + pg->dv_Cm->width + 1,
			       pg->dv_Cm->top + 3);
	str_toGraf(self, "pF", pg->dv_C->left + pg->dv_C->width + 1,
			       pg->dv_C->top + 3);

	str_toGraf(self, "L", pg->dv_uHm->left - 16, pg->dv_uHm->top + 3);
	str_toGraf(self, "uH/m", pg->dv_uHm->left + pg->dv_uHm->width + 1,
			       pg->dv_uHm->top + 3);
	str_toGraf(self, "uH", pg->dv_uH->left + pg->dv_uH->width + 1,
			       pg->dv_uH->top + 3);

	str_toGraf(self, "t", pg->dv_nSm->left - 16, pg->dv_nSm->top + 3);
	str_toGraf(self, "nS/m", pg->dv_nSm->left + pg->dv_nSm->width + 1,
			       pg->dv_nSm->top + 3);
	str_toGraf(self, "nS", pg->dv_nS->left + pg->dv_nS->width + 1,
			       pg->dv_nS->top + 3);

	str_toGraf(self, "f=", pg->dv_Freq->left - 16, pg->dv_Freq->top + 3);
	str_toGraf(self, "MHz", pg->dv_Freq->left + pg->dv_Freq->width + 1,
			       pg->dv_Freq->top + 3);

	str_toGraf(self, "tan", pg->dv_tan->left - 19, pg->dv_tan->top + 3);
//	str_toGraf(self, "MHz", pg->dv_Freq->left + pg->dv_Freq->width + 1,
//			       pg->dv_Freq->top + 3);

	str_toGraf(self, "dB/m", pg->dv_lo_dbm->left + pg->dv_lo_dbm->width + 1,
			       pg->dv_lo_dbm->top + 3);
	str_toGraf(self, "dB", pg->dv_lo_db->left + pg->dv_lo_db->width + 1,
			       pg->dv_lo_db->top + 3);

	str_toGraf(self, "R", pg->dv_Ohmm->left - 16, pg->dv_Ohmm->top + 3);
	DrawPic(self->WinID, get_Pic_cOhm(), 
			pg->dv_Ohmm->left + pg->dv_Ohmm->width + 2,
			pg->dv_Ohmm->top + 6);
	str_toGraf(self, "/m", pg->dv_Ohmm->left + pg->dv_Ohmm->width + 11,
			       pg->dv_Ohmm->top + 3);
	DrawPic(self->WinID, get_Pic_cOhm(), 
			pg->dv_Ohm->left + pg->dv_Ohm->width + 2,
			pg->dv_Ohm->top + 5);


	str_toGraf(self, "Rp", pg->dv_OhmP->left - 16, pg->dv_OhmP->top + 3);
	DrawPic(self->WinID, get_Pic_ckOhm(), 
			pg->dv_OhmP->left + pg->dv_OhmP->width + 2,
			pg->dv_OhmP->top + 5);


  int m = 0x01;
	if (pg->cbCoplan->Checked) {
	    if (!pg->cbGroundp->Checked) m = 0x00;
	    m |= 0x06;
	}
	DrawMS1(pg, 110, 110, 120, 80, m);
}

static void MS1_cbCoplan_OnClick(TChkBox *self, long KeyState, bool DblClick) {
  TPage_MS1 *pg = self->subClass;
	SetEnabled((TWins*)pg->cbGroundp, pg->cbCoplan->Checked);
	RePaint(pg->pnl);
	math_MS1(pg);
}

/* ************** TP1 Create ****************** */
static void DVal_Create_Xxx(TWins *Parent, TDVal **dv, int *x, int y, int w, int nk) {
  int tx = *x;
  TDVal *tdv = TDVal_CreateS(Parent, tx+19, y, w, 19);
	tdv->nk = nk;
	tdv->dv_fu = tdv_fu_LostFocusEnter;
	tdv->OnValChange = MS1_DVal_OnValChange;
	*dv = tdv;
	*x = tx + w + 21;
}

static void DVal_Create_Out(TWins *Parent, TDVal **dv, int *x, int y, int w, int nk) {
  int tx = *x;
  TDVal *tdv = TDVal_CreateS(Parent, tx+19, y, w, 19);
	tdv->ReadOnly = True;
	tdv->nk = nk;
	tdv->KeyFu |= mNoKeyFocus;
	SetColor((TWins*)tdv, 0xe0e0c0);
	*dv = tdv;
	*x = tx + w + 36;
}

static TPage *page_Add_MS1_Create(TPageControl *Parent) {
	return TPage_CreateParamAlg(Parent, sizeof(TPage_MS1));
}

TPage_MS1 *page_Add_MS1(TPageControl *pagc, TPage_Microstrip *pgMS) {
  TPage_MS1 *pg;
  TWins *pnl;
  TpcIni *pcIni;
  int x, y;

	pcIni = get_appIni();

	pg = (TPage_MS1*) TPageControl_AddNewUserPage(pagc, "MS1", page_Add_MS1_Create);
	pg->subClass = pg;
	pg->pgMS = pgMS;

	pnl = TWins_CreateBA((TWins*)pg, 0, 0, 318, 21, 0, alLeft);
	pg->pnl = pnl;
	pnl->OnPaint = pnl_MS1_OnPaint;

	y = 2;
	x = 0;
	DVal_Create_Xxx(pnl, &pg->dv_W, &x, y, 60, 3);
	TDVal_Set_Value(pg->dv_W, TpcIni_RD_float(pcIni, "MS1", "W", 2.032));
	TDVal_Set_Min(pg->dv_W, 0.005);
	pg->dv_W->tag = 1;	// mm/mil
	DVal_Create_Xxx(pnl, &pg->dv_G, &x, y, 60, 3);
	TDVal_Set_Value(pg->dv_G, TpcIni_RD_float(pcIni, "MS1", "G", 0.508));
	TDVal_Set_Min(pg->dv_G, 0.005);
	pg->dv_G->tag = 1;	// mm/mil
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &pg->dv_H, &x, y, 60, 3);
	TDVal_Set_Value(pg->dv_H, TpcIni_RD_float(pcIni, "MS1", "H", 1.524));
	TDVal_Set_Min(pg->dv_H, 0.005);
	pg->dv_H->tag = 1;	// mm/mil
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &pg->dv_L, &x, y, 60, 2);
	TDVal_Set_Value(pg->dv_L, TpcIni_RD_float(pcIni, "MS1", "L", 25.4));
	TDVal_Set_Min(pg->dv_L, 0.01);
	pg->dv_L->tag = 1;	// mm/mil
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &pg->dv_T, &x, y, 60, 3);
	TDVal_Set_Value(pg->dv_T, TpcIni_RD_float(pcIni, "MS1", "T", 0.035));
	TDVal_Set_Min(pg->dv_T, 0.001);
	pg->dv_T->tag = 2;	// mm/mil
	TDVal_SetUpKorr(pg->dv_T, 0.0254, 2, False);
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &pg->dv_Er, &x, y, 60, 1);
	TDVal_Set_Value(pg->dv_Er, TpcIni_RD_float(pcIni, "MS1", "Er", 4.2));
	TDVal_Set_Min(pg->dv_Er, 1);


	y += 22;
	pg->cbCoplan = TChkBox_CreateS(pnl, 2, y, 90, 16, "Coplanar");
	pg->cbCoplan->OnClick = MS1_cbCoplan_OnClick;
	y += 18;
	pg->cbGroundp = TChkBox_CreateS(pnl, 2, y, 90, 16, "Groundplane");
	pg->cbGroundp->OnClick = MS1_cbCoplan_OnClick;
	pg->cbGroundp->Checked = True;
	pg->cbGroundp->Enabled = False;


	x = 0;
	y += 30;
	DVal_Create_Out(pnl, &pg->dv_Imp, &x, y, 60, 1);

	x = 0;
	y += 20;
	DVal_Create_Out(pnl, &pg->dv_Cm, &x, y, 60, 2);
	DVal_Create_Out(pnl, &pg->dv_C, &x, y, 60, 2);

	x = 0;
	y += 20;
	DVal_Create_Out(pnl, &pg->dv_uHm, &x, y, 60, 4);
	DVal_Create_Out(pnl, &pg->dv_uH, &x, y, 60, 4);

	x = 0;
	y += 20;
	DVal_Create_Out(pnl, &pg->dv_nSm, &x, y, 60, 3);
	DVal_Create_Out(pnl, &pg->dv_nS, &x, y, 60, 3);

	x = 0;
	y += 28;
	DVal_Create_Xxx(pnl, &pg->dv_Freq, &x, y, 60, 1);
	TDVal_Set_Value(pg->dv_Freq, TpcIni_RD_float(pcIni, "MS1", "Freq", 500));
//	DVal_Create_Out(pnl, &pg->dv_nS, &x, y, 60, 3);
	x = 0;
	y += 20;
	DVal_Create_Xxx(pnl, &pg->dv_tan, &x, y, 60, 3);
	TDVal_Set_Value(pg->dv_tan, TpcIni_RD_float(pcIni, "MS1", "tan", 0.019));
	x = 0;
	y += 20;
	DVal_Create_Out(pnl, &pg->dv_lo_dbm, &x, y, 60, 2);
	DVal_Create_Out(pnl, &pg->dv_lo_db, &x, y, 60, 2);
	x = 0;
	y += 20;
	DVal_Create_Out(pnl, &pg->dv_Ohmm, &x, y, 60, 2);
	DVal_Create_Out(pnl, &pg->dv_Ohm, &x, y, 60, 3);
	DVal_Create_Out(pnl, &pg->dv_OhmP, &x, y, 60, 3);

	free_appIni(pcIni);

	math_MS1(pg);

	return pg;
}
