/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "ElMath.h"
#include <math.h>

static void str_toGraf(TScaleXY *sxy, const char *s, int x, int y) {
	vds_SetStr(&sxy->vds, s);
	DrawColorString(x, y, &sxy->vds);
}

static void strV_toGraf(TScaleXY *sxy, const char *s, int x, int y) {
	vds_SetStr(&sxy->vds, s);
	x -= TextWidthColorString(&sxy->vds) >> 1;
	DrawColorString(x, y, &sxy->vds);
}

static void rasterV_str_toGraf(TScaleXY *sxy, int x, int y, const char *s) {
	if (s && *s) {
	    y += 2;
	    DrawLineRGB(sxy->WinID, sxy->RasterColor1, x, sxy->sy1, x, y);
	    strV_toGraf(sxy, s, x, y);
	} else {
	    DrawLineRGB(sxy->WinID, sxy->RasterColor2, x, sxy->sy1, x, y);
	}
}

static void strH_toGraf(TScaleXY *sxy, const char *s, int x, int y) {
	vds_SetStr(&sxy->vds, s);
	y -= (sxy->vds.Font->height >> 2) * 3;
	DrawColorString(x, y, &sxy->vds);
}

static void rasterH_str_toGraf(TScaleXY *sxy, int x, int y, const char *s) {
	if (s && *s) {
	    x += 3;
	    DrawLineRGB(sxy->WinID, sxy->RasterColor1, sxy->sx1, y, x, y);
	    strH_toGraf(sxy, s, x+2, y);
	} else {
	    DrawLineRGB(sxy->WinID, sxy->RasterColor2, sxy->sx1, y, x, y);
	}
}

float TScaleXY_yVal(TScaleXY_Dat *dat, float yVal) {
  float res = yVal;
	switch(dat->dy->fu & sxysetfu_scale_mask) {
	case sxysetfu_log: res = log(yVal); break;
	case sxysetfu_sqrt: res = sqrt(yVal); break;
	case sxysetfu_dB10: res = log10(yVal)*10; break;
	case sxysetfu_dB20: res = log10(yVal)*20; break;
	}
	return res;
}

static float TScaleXY_xVal(TScaleXY_Dat *dat, float xVal) {
	if ((dat->dx->fu & sxysetfu_scale_mask) == 0) return xVal;
	switch(dat->dx->fu & sxysetfu_scale_mask) {
	case sxysetfu_log: return log(xVal);
	case sxysetfu_log10: return log10(xVal);
	case sxysetfu_log_1: return log(xVal*(M_E-1)+1);
	case sxysetfu_log10_1: return log10(xVal*9+1);
	case sxysetfu_log20_1: return log10(xVal*19+1)*0.76862178684;
	case sxysetfu_log100_1: return log10(xVal*99+1)*0.5;
	}
	return xVal;
}

static void TScaleXY_DoPaint(TScaleXY *self) {
  int i;
  int gw = self->width - self->sx1 - self->sx2;
  int gh = self->height - self->sy1 - self->sy2;
	if (gw < 10) gw = 10;
	if (gh < 10) gh = 10;
  int y2 = self->sy1 + gh;
  int Y1 = self->sy1;
  int h = self->height - y2 - 1;
	FillRectRGB(self->WinID, self->Color_az, self->sx1, self->sy1, gw, gh);
	if (h < 5) h = 5;
	FillRectRGB(self->WinID, self->Color, 1, y2, self->width-2, h);
  int x = self->sx1 + gw + 1;
  int w = self->width - x - 1;
	if (w < 5) w = 5;
	FillRectRGB(self->WinID, self->Color, x, 1, w, self->height-2);

  TScaleXY_Dat *dat = self->dat;
	if (!dat || !dat->dx || !dat->dy) {
	    rasterV_str_toGraf(self, self->sx1, y2 + 2, "0");
	    rasterH_str_toGraf(self, self->sx1 + gw + 3, y2, "0");
	    return;
	}

  TScale_d *dx = dat->dx;
  TScale_d *dy = dat->dy;
  float tx, ty, tfx, tfy, minx, miny;
  float ghs = gh - 18;
  TScale_Dat *sd;

	minx = dx->min;
	tfx = dx->max - minx;
	miny = dy->min;
	tfy = dy->max - miny;

	sd = &dx->pt[0];
	for (i=0;i<dx->count;i++) {
	    sd->pos = (float)gw * TScaleXY_xVal(dat, (sd->val - minx) / tfx);
	    sd->pos += self->sx1;
	    if (!sd->s[0]) {
		rasterV_str_toGraf(self, sd->pos, y2, NULL);
	    }
	    sd++;
	}

	sd = &dy->pt[0];
	x = self->sx1 + gw; // + 3;
	for (i=0;i<dy->count;i++) {
	    sd->pos = y2;
	    sd->pos -= ghs * (sd->val - miny) / tfy;
	    if (!sd->s[0]) {
		rasterH_str_toGraf(self, x, sd->pos, NULL);
	    }
	    sd++;
	}
//*******************************
  int txt_w, xm = 0;
	sd = &dx->pt[0];
	for (i=0;i<dx->count;i++) {
	    if (sd->s[0]) {
		DrawLineRGB(self->WinID, self->RasterColor1, sd->pos, self->sy1, sd->pos, y2+2);
		vds_SetStr(&self->vds, sd->s);
		txt_w = TextWidthColorString(&self->vds);
		if ((sd->pos-xm-1) > txt_w) {
		    DrawColorString(sd->pos-(txt_w>>1), y2+2, &self->vds);
		    xm = sd->pos;
		}
	    }
	    sd++;
	}
//*******************************
	sd = &dy->pt[0];
	for (i=0;i<dy->count;i++) {
	    if (sd->s[0]) {
		rasterH_str_toGraf(self, x, sd->pos, sd->s);
	    }
	    sd++;
	}
	if (dx->s) strV_toGraf(self, dx->s, self->sx1 + (gw >> 1), y2 + 16);
	if (dy->s) strH_toGraf(self, dy->s, x+2, self->sy1+5);

/* CursorX */
  int y;
	if (self->CursFu & 0x01) {
	    tx = (float)gw * self->CursorX / tfx;
	    if (tx < 0) tx = 0;
	    if (tx > gw) tx = gw;
	    tx += self->sx1;
	    y = Y1-3;
	    FillRectRGB(self->WinID, self->Color, self->sx1-4, y, gw+7, 3);

	    DrawLineRGB(self->WinID, self->CursorX_Color, tx, self->sy1, tx, y2);
	    DrawLineRGB(self->WinID, self->CursorX_Color, tx-3, y, tx+3, y);
	    y++;
	    DrawLineRGB(self->WinID, self->CursorX_Color, tx-2, y, tx+2, y);
	    y++;
	    DrawLineRGB(self->WinID, self->CursorX_Color, tx-1, y, tx+1, y);
	    if (self->CursorX_Str) str_toGraf(self, self->CursorX_Str, self->sx1 + 2, Y1 + 2);
	    
	}

  TScale_Point *ixy;

  TScale_pxy *pxy = dat->pxy;
  float txe, tye, tw, XY_Dat_yOff;

	XY_Dat_yOff = 0;
	if (pxy->AddOff) XY_Dat_yOff = self->XY_Dat_yOff;
	while (pxy) {
	  if (!pxy->PaintDatOk && pxy->PaintDat && (pxy->count > 0)) {
	    ixy = pxy->xy;
	    for (i=pxy->count;i;i--) {
		tw = (float)gw * TScaleXY_xVal(dat, (ixy->x - minx) / tfx);
		x = tw;
		if (x < 0) x = 0;
		if (x > gw) x = gw;
		ixy->px = x + self->sx1;

		tw = TScaleXY_yVal(dat, ixy->y) + XY_Dat_yOff;
		tw = ghs * (tw - miny) / tfy;
		x = tw;
		if (x < 0) x = 0;
		if (x > gh) x = gh;
		
		ixy->py = y2 - x;
		ixy++;
	    }
	    pxy->PaintDatOk = True;
	  }
	  pxy = pxy->next;
	}

	pxy = dat->pxy;
	while (pxy) {
	    if (pxy->PaintDat && (pxy->count > 0)) {
	      ixy = pxy->xy;
	      txe = ixy->px;
	      tye = ixy->py;
	      ixy++;
	      for (i=pxy->count-1;i;i--) {
		tx = ixy->px;
		ty = ixy->py;
		if ((tx != txe) || (ty != tye)) {
		    DrawLineRGB(self->WinID, pxy->Color, tx, ty, txe, tye);
		}
		txe = tx;
		tye = ty;
		ixy++;
	      }
	    }
	    pxy = pxy->next;
	}

  TScaleXY_Txt *xy_txt = self->txt;
	while (xy_txt) {
	    if (xy_txt->s) {
		str_toGraf(self, xy_txt->s, self->sx1 + xy_txt->x, Y1 + xy_txt->y);
	    }
	    xy_txt = xy_txt->next;
	}

}

float TScaleXY_GetCursX_ScaleVal(TScaleXY *self) {
  TScaleXY_Dat *dat = self->dat;
  TScale_d *d;
  float minx, tfx, v, ret;

	if (!dat || !dat->dx) return 0;
	d = dat->dx;
	ret = self->CursorX + d->min;
	if ((d->fu & sxysetfu_scale_mask) == 0) return ret;

	minx = d->min;
	tfx = d->max - minx;
	v = self->CursorX / tfx;

	switch(d->fu & sxysetfu_scale_mask) {
	case sxysetfu_log: 
	    ret = (exp(v) * tfx) + d->min;
	    break;

	case sxysetfu_log10: 
	    ret = (exp(v*M_LN10) * tfx) + d->min;
	    break;

	case sxysetfu_log_1: 
	    ret = ((exp(v)-1)/(M_E-1) * tfx) + d->min;
	    break;

	case sxysetfu_log10_1: 
	    ret = ((exp(v*M_LN10)-1)/9 * tfx) + d->min;
	    break;

	case sxysetfu_log20_1: 
	    ret = ((exp(v*M_LN10*1.30102999566)-1)/19 * tfx) + d->min;
	    break;

	case sxysetfu_log100_1: 
	    ret = ((exp(v*M_LN10*2)-1)/99 * tfx) + d->min;
	    break;
	}

	return ret;
}

float TScaleXY_GetCursX_YVal(TScaleXY *self, TScale_pxy *pxy, TScale_Point **ppsp) {
  int i;
  float fx;
  TScale_pxy *pxyt = self->dat->pxy;
  TScale_Point *psp;

	if (ppsp) *ppsp = NULL;
	fx = TScaleXY_GetCursX_ScaleVal(self);
	while (pxyt && (pxyt != pxy)) pxyt = pxyt->next;
	if (!pxyt) return 0;

	psp = &pxy->xy[0];
	for (i=0;i<pxy->count;i++) {
	    if (psp->x >= fx) {
		if (ppsp) *ppsp = psp;
		return psp->y;
	    }
	    psp++;
	}
	return 0;
}

void TScaleXY_Set_Dat_Log(TScaleXY *self, int fu, int set_dx_or_dy) {
  TScaleXY_Dat *dat = self->dat;
  TScale_d *d;
	if (!dat) return;

	if (set_dx_or_dy & 1) {
	    d = dat->dy;
	} else {
	    d = dat->dx;
	}
	if (!d) return;
	
	d->fu &= ~sxysetfu_scale_mask;
	d->fu |= fu & sxysetfu_scale_mask;
	Refresh((TWins*)self);
}

void TScaleXY_Set_Dat(TScale_d *d, const char *s, float min, float max, int step, int fu) {
  int n, m = 0;
  float fa, fb, fc;
	if (!d) return;
	ppcchange(&d->s, s);
	d->fu = fu;
	d->min = min;
	d->max = max;
	if (fu & sxysetfu_scale_val) {
	  switch(fu & sxysetfu_scale_mask) {
	  case sxysetfu_log:
	    d->min = log(min);
	    d->max = log(max);
	    fa = log(min);
	    fb = log(max);
	    for (int i=0;i<d->count;i++) {
		d->pt[i].val = fa + (fb - fa) * (float)i / (float)(d->count-1);
		if (i == m) {
		    fc = exp(d->pt[i].val);
		    n = 0;
		    if (fc < 10.0) n = 1;
		    if (fc < 1.0) n = 2;
		    floattopc(d->pt[i].s, fc, n);
		    m += step;
		}
	    }
	    break;
	
	  case sxysetfu_sqrt:
	    d->min = sqrt(min);
	    d->max = sqrt(max);
	    for (int i=0;i<d->count;i++) {
		d->pt[i].val = d->min + (d->max - d->min) * (float)i / (float)(d->count-1);
		if (i == m) {
		    fc = d->pt[i].val * d->pt[i].val;
		    n = 0;
		    if (fc < 10.0) n = 1;
		    if (fc < 1.0) n = 2;
		    floattopc(d->pt[i].s, fc, n);
		    m += step;
		}
	    }
	    break;

	  default:
	    for (int i=0;i<d->count;i++) {
		d->pt[i].val = min + (max - min) * (float)i / (float)(d->count-1);
		if (i == m) {
		    floattopc(d->pt[i].s, d->pt[i].val, fu & sxysetfu_nk_mask);
		    m += step;
		}
	    }
	  }
	} else {
	    for (int i=0;i<d->count;i++) {
		d->pt[i].val = min + (max - min) * (float)i / (float)(d->count-1);
		if (i == m) {
		    floattopc(d->pt[i].s, d->pt[i].val, fu & sxysetfu_nk_mask);
		    m += step;
		}
	    }
	}
}

void TScaleXY_SetDat_mem(TScaleXY_Dat **dat, int cx, int cy, int cxy) {
  TScaleXY_Dat *d;
	if (!dat) return;
	d = *dat;
	if (!d) {
	    d = xcalloc(sizeof(TScaleXY_Dat));
	    *dat = d;
	}
	if (cx >= 0) {xfree(d->dx); d->dx = NULL;}
	if (cy >= 0) {xfree(d->dy); d->dy = NULL;}
	if (cxy >= 0) {xfree(d->pxy); d->pxy = NULL;}
	if (cx > 0) {
	    d->dx = xcalloc(sizeof(TScale_d) + sizeof(TScale_Dat)*cx);
	    d->dx->count = cx;
	}
	if (cy > 0) {
	    d->dy = xcalloc(sizeof(TScale_d) + sizeof(TScale_Dat)*cy);
	    d->dy->count = cy;
	}
	if (cxy > 0) {
	    d->pxy = xcalloc(sizeof(TScale_pxy) + sizeof(TScale_Point)*cxy);
	    d->pxy->count = cxy;
	}
}

static void TScaleXY_ClsPaintDatOk(TScaleXY *sxy) {
  TScale_pxy *pxy;
	if (sxy->dat) {
	    pxy = sxy->dat->pxy;
	    while (pxy) {
		pxy->PaintDatOk = False;
		pxy = pxy->next;
	    }
	}
}

static void TScaleXY_Resize(TScaleXY *self) {
	TScaleXY_ClsPaintDatOk(self);
	TWins_Resize((TWins*)self);
}


static void TScaleXY_Refresh(TScaleXY *self) {
	TScaleXY_ClsPaintDatOk(self);
	TWins_Refresh((TWins*)self);
}

/*____________________________ Create ___________________________________*/
static void TScaleXY_CreateParams(TWins_Create_struct *wcs) {
  TScaleXY *self = wcs->wins;
	self->Resize = TScaleXY_Resize;
	self->DoPaint = TScaleXY_DoPaint;
	self->Refresh = TScaleXY_Refresh;

	self->sx1 = 10;
	self->sy1 = 3;
	self->sx2 = 38;
	self->sy2 = 30;
	self->Color_az = 0xe0e0e0;
	self->RasterColor1 = 0xa0a0a0;
	self->RasterColor2 = 0xd0d0d0;
}

TScaleXY *TScaleXY_Create(TWins *Parent) {
  TWins_Create_struct wcs;
	TWins_Create_struct_init(&wcs, sizeof(TScaleXY), wtyp_Wins);
	wcs.Parent = Parent;
	wcs.CreateParams = TScaleXY_CreateParams;
	TWins_DefaultConf(&wcs);
	return wcs.wins;
}
