/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appXbase.h"
#include "cpuid.h"

#define use_SAVE_UserDat	1
#define use_PAGES		1
// #define use_FEATURE		1
// #define use_exec_Xns		1

#ifdef use_PAGES
 #define PAGES_MEM_MOVE		1
#endif

#define USE_MMAP 1

#ifdef _OS_WIN32
#undef USE_MMAP
#define USE_MMAP 0
#endif

#if defined(__arm__)
#undef USE_MMAP
#define USE_MMAP 1
#endif

#if USE_MMAP
#include <sys/mman.h>
#endif /* USE_MMAP */


#define max_cbXM 8

#define APPNAME "TeCache"

extern const int build_v;
extern const char *build_date;

typedef struct {
  float fx, fy;
} TXnsPoint;

typedef struct {
  float val;
  int fu;
} TXnsScale;
typedef struct {
  int val;
  float fx;
  int fu;
} TXmemScale;

typedef struct {
  int xa, ya;
  int gw, gh;
  int color;
} TXnsXY_Paint;

typedef struct {
  TXnsPoint *PV;
  int PV_Count;
  int xa, ya;
  int gw, gh;
  int color;
  bool enabled;
} TXnsPoints;

typedef struct {
  TXnsPoints *PVs;
  int PVs_count;
  TXmemScale *PX;
  int PXmax_Count;
  int PX_Count;
  TXnsScale *PH;
  int PHmax_Count;
  int PH_Count;
  int PH_nk;
  float PHmax_Scale;
  const char *Yname;
  int memsize;
  int fu;
  float PHmax_ns;
} TScaleXnsRaster;

#define TCSelf TScaleXns
typedef struct TScaleXns {
  AlgTWins
  int gr_x;
  int gr_y;
  int gr_x2;
  int gr_y2;
  int Color_az;
  TScaleXnsRaster *xr;
  TScaleXnsRaster *xr2;
} TScaleXns;
#undef TCSelf

typedef struct {
  int MHz;
  int sampz;
  int sampt[256];
  int sampmem[256];
  dword maxmem_ps;
  dword min_ps;
  dword max_ps;
  dword ges_time;
} TFormTeCache_Dat;

#define TCSelf TFormTeCache
typedef struct TFormTeCache {
  AlgTForm
  TWins *pnlTop;
  TWins *pnlMiddle;
  TWins *pnlBottom;

  TMenuButton *mbFile;
  TMenu *menuFile;

  TMenuButton *mbSettings;
  TMenu *menuSettings;

  TMenuButton *mbView;
  TMenu *menuView;

  TMenuButton *mbHelp;
  TMenu *menuHelp;

#ifdef use_PAGES
  TPageControl *PageControl;
  TPage *pgExec;
  TPage *pgMemMove;
  TWins *pnlAlign;
  TPic align_pic;
#ifdef use_FEATURE
  TPage *pgImageZoom;
#endif
#endif

  TWins *pnlExec;
#ifdef use_exec_Xns
  TScaleXns *ns_exec;
#endif
#ifdef PAGES_MEM_MOVE
  TScaleXns *xns_mov;
#endif
  TWins *pnlSelAz;
  int akt_MX;
  int gr_x;
  int gr_y;
  int gr_w;
  int gr_h;
  int us_off;

  TButton *buStart;
  TButton *buStop;
  TButton *buClose;

  int MHz;
  dword *ExecMem;
  dword *spI;
  int splg;
#if defined (__arm__)
  void (*spProc)(int k);
#else
  void __attribute__ ((regparm (3))) (*spProc)(int k);
#endif
  int sampk[256];
  int scale_ps;
  int color_ps;
  int color_ps_d;
  int run_time;
  bool TestRun;
  bool Stop;
  bool has_TSC;
  bool TSC_on;
  bool Start250ms;
  bool fsc_nc5;
  bool fsc_nc6;
  bool fsc_nc7;

#if defined (__i386__) || defined (__x86_64__)
  TCPUID cpuid;
#endif

  TFormTeCache_Dat *cd_view;
  TFormTeCache_Dat *cd_default_view;
  TFormTeCache_Dat *cd_default_arr;

  TFormTeCache_Dat *cdm_view;
  TFormTeCache_Dat *cdm_rd_view;

} TFormTeCache;
#undef TCSelf

extern TFormTeCache_Dat cd_Buf[max_cbXM];

extern TFormTeCache_Dat dat_CPUX_ref[max_cbXM];

#ifdef use_PAGES
extern TFormTeCache_Dat cdm_Buf[max_cbXM];	//memmove
extern TFormTeCache_Dat cdm_rd_Buf[max_cbXM];
#endif

#ifdef use_exec_Xns
extern TScaleXnsRaster exec_xr[8];
extern TScaleXnsRaster exec_rd_xr[8];
#endif

extern TScaleXnsRaster dat_xr[8];
extern TScaleXnsRaster dat_rd_xr[8];

void MenuHelp_Create(TFormTeCache *mf);

void TTeCache_pnlExec_OnPaint(TWins *self);
void TTeCache_pnlExec_OnResize(TWins *self);
void TeCache_exec_init(TFormTeCache *self);
void TTeCache_InitProc(TFormTeCache *fcs, int mc);
void TTeCache_memExec_Start(TFormTeCache *fcs);
void TTeCache_FreeExecMem(TFormTeCache *self);
void TTeCache_get_MHz(TFormTeCache *fcs);


void pnlSelAz_Create(TFormTeCache *form);

void az_MHz_line(TFormTeCache *self, TFormTeCache_Dat *cd, char *pc);
void az_str3D(TVDrawString *vds, char *s, int x, int y, int color);

void TTeCache_ImageZoom_Start(TFormTeCache *self);
void TTeCache_MemMove_Start(TFormTeCache *self);


void TeCache_ld_Head( TFormTeCache_Dat *cd, const char *s0);
void TeCache_ld_t( TFormTeCache_Dat *cd, const char *s0);
void TeCache_ld_m( TFormTeCache_Dat *cd, const char *s0);
void TeCache_mov_cd(TFormTeCache_Dat *cdd, TFormTeCache_Dat *cds);
void TTeCache_Load_User(TFormTeCache *fcs, int idx);
void TTeCache_zLoad(TFormTeCache *form, const byte *data, int size, int size_org);
void TTeCache_Load(TFormTeCache *fcs, TpcIni *pcIni, const char *Gpd);

void TeCache_add_mem(dword *mem, int shift);

extern char *m_def[max_cbXM];

void TeCache_ld_PII200(TFormTeCache *fcs);
void TeCache_ld_PIII700(TFormTeCache *fcs);
void TeCache_ld_M1870(TFormTeCache *fcs);
void TeCache_ld_C2D_2400(TFormTeCache *fcs);
void TeCache_ld_I7_870(TFormTeCache *fcs);

#ifdef use_SAVE_UserDat
extern const char *sMemSize[max_cbXM];
void TTeCache_mFile_SaveU1(TMenu *self, long KeyState, TMenuItem *smi);
#endif


#ifdef use_PAGES
void TTeCache_pnlAlign_OnPaint(TWins *self);

void TeCache_memmove_init(TFormTeCache *self);

TScaleXns *TScaleXns_Create(TWins *Parent);
void TScaleXns_DoPaint(TScaleXns *self);
void  TScaleXns_LineXY_0(TScaleXns *self, TScaleXnsRaster *xr);
void  TScaleXns_LineXY_N(TScaleXns *self, TScaleXnsRaster *xr, int idx);
void TTeCache_Set_Max_ns_Scale(TScaleXnsRaster *xr);

typedef struct {
  int *pi1;
  int *pi2;
  long size;
  int k;
} TMemMoveK;

void MemMoveK0_ASM(TMemMoveK *kkk);
void MemMoveK_ASM(TMemMoveK *kkk);

void MemMove_ScaleXns_OnPaint(TScaleXns *self);
#endif
