/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "TeCache.h"

static void strV_toGraf(TScaleXns *self, const char *s, int x, int y) {
	vds_SetStr(&self->vds, s);
	x -= TextWidthColorString(&self->vds) >> 1;
	DrawColorString(x, y, &self->vds);
}

static void rasterV_str_toGraf(TScaleXns *self, const char *s, int x, int y, bool txt) {
	if (txt) {
	    DrawLineRGB(self->WinID, 0xa0a0e0, x, self->gr_y, x, y);
	    strV_toGraf(self, s, x, y);
	} else {
	    DrawLineRGB(self->WinID, 0xc0c0e0, x, self->gr_y, x, y);
	}
}

static void strH_toGraf(TScaleXns *self, const char *s, int x, int y) {
	vds_SetStr(&self->vds, s);
	y -= (self->vds.Font->height >> 2) * 3;
	DrawColorString(x, y, &self->vds);
}

static void rasterH_str_toGraf(TScaleXns *self, const char *s, int w, int y) {
  int x = w + self->gr_x;
	DrawLineRGB(self->WinID, 0x80e080, self->gr_x, y, x, y);
	strH_toGraf(self, s, x+2, y);
}

static int Graf_y(TScaleXns *self, int gh, float val) {
  int y = 0;
	if ((self->xr->PHmax_Scale > 0) && (val > 0)) {
	    val *= (gh-20) / self->xr->PHmax_Scale;
	    y = val;
	}
	y = gh - y;
	if (y < 0) y = 0;
	y += self->gr_y;
	return y;
}

void  TScaleXns_LineXY_0(TScaleXns *self, TScaleXnsRaster *xr) {
  TXnsPoints *PVs;
  int gw, gh;

	if (!xr) return;
	gw = self->width - self->gr_x - self->gr_x2;
	gh = self->height - self->gr_y - self->gr_y2;
	
	for (int i=0;i<xr->PVs_count;i++) {
	    PVs = &xr->PVs[i];

	    PVs->gw = gw;
	    PVs->gh = gh;
	    if (PVs->PV) {
		PVs->xa = self->gr_x + gw * PVs->PV[0].fx;
		PVs->ya = Graf_y(self, gh, PVs->PV[0].fy);
	    }
	}
}

void  TScaleXns_LineXY_N(TScaleXns *self, TScaleXnsRaster *xr, int idx) {
  TXnsPoints *PVs;
  int xe, ye;

	if (!xr) return;
	for (int i=0;i<xr->PVs_count;i++) {
	    PVs = &xr->PVs[i];
	    if (PVs->enabled && PVs->PV) {
		xe = self->gr_x + PVs->gw * PVs->PV[idx].fx;
		ye = Graf_y(self, PVs->gh, PVs->PV[idx].fy);
		if (PVs->PV[idx].fy > 0)
		    DrawLineRGB(self->WinID, PVs->color, PVs->xa, PVs->ya, xe, ye);
		PVs->xa = xe; PVs->ya = ye;
	    }
	}
}

void TScaleXns_DoPaint(TScaleXns *self) {
  int gw = self->width - self->gr_x - self->gr_x2;
  int gh = self->height - self->gr_y - self->gr_y2;
	if (gw < 10) gw = 10;
	if (gh < 10) gh = 10;
  int y2 = self->gr_y + gh + 1;
  int h = self->height - y2 - 1;
	FillRectRGB(self->WinID, self->Color_az, self->gr_x, self->gr_y, gw, gh);
	if (h < 5) h = 5;
	FillRectRGB(self->WinID, self->Color, 1, y2, self->width-2, h);
  int x = self->gr_x + gw + 1;
  int w = self->width - x - 1;
	if (w < 5) w = 5;
	FillRectRGB(self->WinID, self->Color, x, 1, w, self->height-2);

  TScaleXnsRaster *xr = self->xr;

	if (!xr) xr = self->xr2;
	if (!xr || !xr->PX_Count) rasterV_str_toGraf(self, "0", self->gr_x, y2, True);
	if (!xr || !xr->PH_Count) rasterH_str_toGraf(self, "0,00", gw+3, self->gr_y + gh);
	if (!xr) return;

  char buf[16];
	if (xr->PX) {
  int n = 0;
	    for (int i=0;i<xr->PX_Count;i++) {
		if (xr->PX[i].val < 0x400) {
		    n = inttopc(buf, xr->PX[i].val);
		} else if (xr->PX[i].val < 0x100000) {
		    n = inttopc(buf, xr->PX[i].val >> 10);
		    buf[n] = 'k';
		    n++;
		} else {
		    n = inttopc(buf, xr->PX[i].val >> 20);
		    buf[n] = 'M';
		    n++;
		}
		buf[n] = 0;
		x = self->gr_x + gw * xr->PX[i].fx;
		rasterV_str_toGraf(self, buf, x, y2, (xr->PX[i].fu & 1));
	    }
	}

	for (int i=0;i<xr->PH_Count;i++) {
	    buf[0] = 0;
	    floattopc(buf, xr->PH[i].val, xr->PH_nk);
	    rasterH_str_toGraf(self, buf, gw+3, Graf_y(self, gh, xr->PH[i].val));
	}
	strH_toGraf(self, xr->Yname, self->gr_x + gw + 5, self->gr_x + 3);

	if (xr && self->xr2 && (xr != self->xr2)) {
	  if (self->xr2->fu & 0x01) {
	    TScaleXns_LineXY_0(self, self->xr2);
	    for (int i=0;i<self->xr2->PVs_count;i++) {
  TXnsPoints *PVs;
		PVs = &self->xr2->PVs[i];
		if (PVs->enabled && PVs->PV) {
		    for (x=1;x<PVs->PV_Count;x++) {
  int xe, ye;
			xe = self->gr_x + PVs->gw * PVs->PV[x].fx;
			ye = Graf_y(self, PVs->gh, PVs->PV[x].fy);
			if (PVs->PV[x].fy > 0)
			    DrawLineRGB(self->WinID, PVs->color, PVs->xa, PVs->ya, xe, ye);
			PVs->xa = xe; PVs->ya = ye;
		    }
		}
	    }
	  }
	}

	if (xr->fu & 0x01) {
	    TScaleXns_LineXY_0(self, xr);
	    for (int i=0;i<xr->PVs_count;i++) {
  TXnsPoints *PVs;
		PVs = &xr->PVs[i];
		if (PVs->enabled && PVs->PV) {
		    for (x=1;x<PVs->PV_Count;x++) {
  int xe, ye;
			xe = self->gr_x + PVs->gw * PVs->PV[x].fx;
			ye = Graf_y(self, PVs->gh, PVs->PV[x].fy);
			if (PVs->PV[x].fy > 0)
			    DrawLineRGB(self->WinID, PVs->color, PVs->xa, PVs->ya, xe, ye);
			PVs->xa = xe; PVs->ya = ye;
		    }
		}
	    }
	}
}

/*____________________________ Create ___________________________________*/
static void TScaleXns_CreateParams(TWins_Create_struct *wcs) {
  TScaleXns *self = wcs->wins;

	self->DoPaint = TScaleXns_DoPaint;

	self->gr_x = 4;
	self->gr_y = 4;
	self->gr_x2 = 34;
	self->gr_y2 = 17;
	self->Color_az = 0xe0e0e0;
}

TScaleXns *TScaleXns_Create(TWins *Parent) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TScaleXns), wtyp_Wins);
	wcs.Parent = Parent;
	wcs.CreateParams = TScaleXns_CreateParams;
	TWins_DefaultConf(&wcs);
	return wcs.wins;
}
