/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "TeCache.h"
#include <math.h>

#define KORR_GetTime_us 0

#ifdef use_exec_Xns
static void xr_PH_init(TScaleXnsRaster *xr, float fsc, int sc) {
  float Ymax = 0;
	xr->PH_Count = sc;
	for (int i=0;i<sc;i++) {
	    Ymax = fsc * i;
	    xr->PH[i].val = Ymax;
	}
	xr->PHmax_Scale = Ymax;
	xr->PH_nk = 1;
	if (Ymax < 10) xr->PH_nk = 2;
	xr->Yname = "ns";
}

static void exec_xr_init(TScaleXnsRaster *xr, int ma, int me) {
  int lg, mem, m;
  TXmemScale *PX;
  float ScaleMaxX, off;

	xr->memsize = me;
	mem = ma;
	lg = 1;
	while ((mem < me) && (lg < 256)) {
	    TeCache_add_mem((dword*)&mem, 0);
	    lg++;
	}
	xr->PX_Count = lg;
	xr->PXmax_Count = lg;
	PX = xcalloc(sizeof(TXmemScale) * lg);
	xr->PX = PX;

	mem = ma;
	PX[0].fu = 0x01;
	for (int i=1;i<lg;i++) {
	    TeCache_add_mem((dword*)&mem, 0);
	    PX[i].val = mem;
	}

	off = log((float)PX[1].val) + (log((float)PX[1].val) - log((float)PX[2].val));
	ScaleMaxX = log((float)PX[lg-1].val) - off;

	for (int i=1;i<lg;i++) {
	    PX[i].fx = (log((float)PX[i].val) - off) / ScaleMaxX;
	    m = PX[i].val >> 8;
	    while (m && ((m & 1) == 0)) m >>= 1;
	    if (m == 1) PX[i].fu |= 0x01;
	    if (m == 3) PX[i].fu |= 0x02;
	}

	xr->PHmax_Count = 16;
	xr->PH = xcalloc(sizeof(TXnsScale) * xr->PHmax_Count);

	xr_PH_init(xr, 0.5, 11);


	xr->PVs_count = 2;
	xr->PVs = xcalloc(sizeof(TXnsPoints) * xr->PVs_count);
}

void TeCache_exec_init(TFormTeCache *self) {
	exec_xr_init(&self->exec_xr[0], 0x300, 0x40000);
	exec_xr_init(&self->exec_xr[1], 0x600, 0x80000);
	exec_xr_init(&self->exec_xr[2], 0xc00, 0x100000);
	exec_xr_init(&self->exec_xr[3], 0xc00, 0x200000);
	exec_xr_init(&self->exec_xr[4], 0xc00, 0x400000);
	exec_xr_init(&self->exec_xr[5], 0x1800, 0x800000);
	exec_xr_init(&self->exec_xr[6], 0x1800, 0x1000000);
	exec_xr_init(&self->exec_xr[7], 0x1800, 0x2000000);
	self->ns_exec->xr = &self->exec_xr[0];
}
#endif

static void ClearAzPanel(TFormTeCache *self) {
  TWins *az = self->pnlExec;
  int y2 = self->gr_y + self->gr_h + 1;
  int h = az->height - y2 - 1;
	FillRectRGB(az->WinID, 0xe0e0e0, self->gr_x, self->gr_y, self->gr_w+3, self->gr_h);
	if (h < 5) h = 5;
	FillRectRGB(az->WinID, az->Color, 1, y2, az->width-2, h);
  int x = self->gr_x + self->gr_w + 1;
  int w = az->width - x - 1;
	if (w < 5) w = 5;
	FillRectRGB(az->WinID, az->Color, x, 1, w, az->height-2);
}

void az_str3D(TVDrawString *vds, char *s, int x, int y, int color) {
	vds_SetStr(vds, s);
	vds->ForeColor = 0xffffff;
	DrawColorString(x-1, y, vds);
	DrawColorString(x, y-1, vds);
	DrawColorString(x-1, y-1, vds);
	vds->ForeColor = 0xa0a0a0;
	DrawColorString(x+1, y, vds);
	DrawColorString(x, y+1, vds);
	DrawColorString(x+1, y+1, vds);
	vds->ForeColor = color;
	DrawColorString(x, y, vds);
	vds->ForeColor = 0;
}

void az_MHz_line(TFormTeCache *self, TFormTeCache_Dat *cd, char *pc) {
  char line2[32];
  const char *fs;
  double d;

	snprintf(line2, sizeof(line2), "CPU %d MHz    t = ", cd->MHz);
	ppcmove(&pc, line2);
	d = (double)cd->ges_time / 1000000;
	if (d < 0.001) {
	  ppcmove(&pc, "0");
	} else {
	  if (d < 1) {
	    fs = "%.3f s";
	  } else if (d < 10) {
	    fs = "%.2f s";
	  } else {
	    fs = "%.1f s";
	  }
	  snprintf(line2, sizeof(line2), fs, d);
	  ppcmove(&pc, line2);
	}
	ppcmove(&pc, "      ");

	d = (double)cd->min_ps / 1000;
	if (d < 0.001) {
	  ppcmove(&pc, "0 / ");
	} else {
	  if (d < 1) {
	    fs = "%.3f / ";
	  } else if (d < 10) {
	    fs = "%.2f / ";
	  } else {
	    fs = "%.1f / ";
	  }
	  snprintf(line2, sizeof(line2), fs, d);
	  ppcmove(&pc, line2);
	}

	d = (double)cd->max_ps / 1000;
	if (d < 0.001) {
	  ppcmove(&pc, "0 ns");
	} else {
	  if (d < 1) {
	    fs = "%.3f ns";
	  } else if (d < 10) {
	    fs = "%.2f ns";
	  } else {
	    fs = "%.1f ns";
	  }
	  snprintf(line2, sizeof(line2), fs, d);
	  ppcmove(&pc, line2);
	}
}

static void az_MHz(TFormTeCache *self) {
  TWins *az = self->pnlExec;
  char line[80];

	if (self->cd_view) {
	    az_MHz_line(self, self->cd_view, line);
	    az_str3D(&az->vds, line, self->gr_x+2, self->gr_y+1, 0x000000);
	}
	if (self->cd_default_view) {
	    az_MHz_line(self, self->cd_default_view, line);
	    az_str3D(&az->vds, line, self->gr_x+3, self->gr_y+18, self->color_ps_d);
	}
}

static void strV_toGraf(TFormTeCache *self, const char *s, int x, int y) {
  TWins *az = self->pnlExec;
	vds_SetStr(&az->vds, s);
	x -= TextWidthColorString(&az->vds) >> 1;
	DrawColorString(x, y, &az->vds);
}

static void rasterV_str_toGraf(TFormTeCache *self, const char *s, int x, int y, bool txt) {
	if (txt) {
	    DrawLineRGB(self->pnlExec->WinID, 0x8080c0, x, self->gr_y, x, y);
	    strV_toGraf(self, s, x, y);
	} else {
	    DrawLineRGB(self->pnlExec->WinID, 0xa0a0c0, x, self->gr_y, x, y);
	}
}


static void strH_toGraf(TFormTeCache *self, const char *s, int x, int y) {
  TWins *az = self->pnlExec;
	vds_SetStr(&az->vds, s);
	y -= (az->vds.Font->height >> 2) * 3;
	DrawColorString(x, y, &az->vds);
}

static void rasterH_str_toGraf(TFormTeCache *self, const char *s, int x, int y) {
	DrawLineRGB(self->pnlExec->WinID, 0x60c060, self->gr_x, y, x, y);
	strH_toGraf(self, s, x+2, y);
}

static int Graf_y_ps(TFormTeCache *self, int ps) {
  int y;
	y = 0;
	if (ps > 0) y = ps * self->gr_h / self->scale_ps;
	y = self->gr_h - y;
	if (y < 0) y = 0;
	y += self->gr_y;
	return y;
}

static void Set_Graf_Scale_H(TFormTeCache *self, int ps, int max_ps, int l, int k) {
  int x = self->gr_x + self->gr_w + 4;
  int xt = ps;
  char line[16];
  double d;

	self->scale_ps = max_ps;
	if (k > 1) {
	    rasterH_str_toGraf(self, "0,00", x, Graf_y_ps(self, 0));
	    for (int i=0;i<l;i++) {
		d = (double)xt / 1000;
		snprintf(line, sizeof(line), "%.2f", d);
		rasterH_str_toGraf(self, line, x, Graf_y_ps(self, xt));
		xt += ps;
	    }
	} else if (k == 1) {
	    rasterH_str_toGraf(self, "0,0", x, Graf_y_ps(self, 0));
	    for (int i=0;i<l;i++) {
		d = (double)xt / 1000;
		snprintf(line, sizeof(line), "%.1f", d);
		rasterH_str_toGraf(self, line, x, Graf_y_ps(self, xt));
		xt += ps;
	    }
	} else {
	    rasterH_str_toGraf(self, "0", x, Graf_y_ps(self, 0));
	    for (int i=0;i<l;i++) {
		snprintf(line, sizeof(line), "%d", xt / 1000);
		rasterH_str_toGraf(self, line, x, Graf_y_ps(self, xt));
		xt += ps;
	    }
	}
}

void TTeCache_pnlExec_OnPaint(TWins *self) {
  TFormTeCache *form = (TFormTeCache*)self->FormStruct;
  TWins *az = form->pnlExec;
  TFormTeCache_Dat *cdv = form->cd_view;
  TFormTeCache_Dat *cdvd = form->cd_default_view;
  TFormTeCache_Dat *cdx;
  int i, x, x2, y, y2, y2t, mem;
  int sampz = 0;
  int maxmem_ps = 0;
  const char *s;
  bool mk;

	ClearAzPanel(form);
	y2 = form->gr_y + form->gr_h;
	y2t = y2 + 2;
	rasterV_str_toGraf(form, "0", 3, y2t, True);

	if (cdvd && !cdvd->sampz) cdvd = NULL;
	if (cdv && !cdv->sampz) cdv = NULL;
	cdx = cdv;
	if (cdv) sampz = cdv->sampz;
	if (cdvd && (cdvd->sampz > sampz)) {sampz = cdvd->sampz; cdx = cdvd;}

	for (i=0; i<sampz; i++) {
	    x = 3 + form->gr_w * (i+1) / sampz;
	    mem = cdx->sampmem[i];
	    s = NULL;
	    mk = True;
	    if ((mem == 0x1000) && (form->akt_MX < 5)) {
		s = "16k"; mk = (form->akt_MX < 2);
	    } else if ((mem == 0x2000) && (form->akt_MX < 5)) {
		s = "32k"; mk = (form->akt_MX < 4);
	    } else if (mem == 0x4000) {
		s = "64k"; mk = (form->akt_MX < 4);
	    } else if (mem == 0x8000) {
		s = "128k";
	    } else if (mem == 0x10000) {
		s = "256k"; mk = (form->akt_MX < 6);
	    } else if (mem == 0x20000) {
		s = "512k"; //mk = (form->akt_MX < 7);
	    } else if (mem == 0x30000) {
		s = "768k"; mk = (form->akt_MX < 6);
	    } else if (mem == 0x40000) {
		s = "1M";
	    } else if (mem == 0x80000) {
		s = "2M";
	    } else if (mem == 0xc0000) {
		s = "3M"; mk = (form->akt_MX < 7);
	    } else if (mem == 0x100000) {
		s = "4M";
	    } else if (mem == 0x140000) {
		s = "5M"; mk = False;
	    } else if (mem == 0x180000) {
		s = "6M"; mk = (form->akt_MX < 7);
	    } else if (mem == 0x1c0000) {
		s = "7M"; mk = False;
	    } else if (mem == 0x200000) {
		s = "8M";
	    } else if (mem == 0x280000) {
		s = "10M"; mk = (form->akt_MX < 7);
	    } else if (mem == 0x300000) {
		s = "12M";
	    } else if (mem == 0x380000) {
		s = "14M"; mk = (form->akt_MX < 7);
	    } else if (mem == 0x400000) {
		s = "16M";
	    } else if (mem == 0x500000) {
		s = "20M"; mk = False;
	    } else if (mem == 0x600000) {
		s = "24M";
	    } else if (mem == 0x700000) {
		s = "28M"; mk = False;
	    } else if (mem == 0x800000) {
		s = "32M";
	    }
	    if (s) {
		rasterV_str_toGraf(form, s, x, y2t, mk);
	    } else {
		DrawLineRGB(az->WinID, 0xffffff, x, 3, x, y2);
	    }
	}

	if (cdv) maxmem_ps = cdv->maxmem_ps;
	if (cdvd && (cdvd->maxmem_ps > maxmem_ps)) maxmem_ps = cdvd->maxmem_ps;

	x = form->gr_x + form->gr_w + 4;
	strH_toGraf(form, "ns", x+4, 8);

	if (maxmem_ps < 300) {
	    Set_Graf_Scale_H(form, 50, 380, 7, 2);
	} else if (maxmem_ps < 600) {
	    Set_Graf_Scale_H(form, 100, 770, 7, 2);
	} else if (maxmem_ps < 850) {
	    Set_Graf_Scale_H(form, 200, 1100, 5, 2);
	} else if (maxmem_ps < 1100) {
	    Set_Graf_Scale_H(form, 250, 1380, 5, 2);
	} else if (maxmem_ps < 1600) {
	    Set_Graf_Scale_H(form, 250, 1950, 7, 2);
	} else if (maxmem_ps < 2200) {
	    Set_Graf_Scale_H(form, 500, 2750, 5, 1);
	} else if (maxmem_ps < 3200) {
	    Set_Graf_Scale_H(form, 500, 3800, 7, 1);
	} else if (maxmem_ps < 4500) {
	    Set_Graf_Scale_H(form, 1000, 5500, 5, 1);
	} else if (maxmem_ps < 6200) {
	    Set_Graf_Scale_H(form, 1000, 7600, 7, 1);
	} else if (maxmem_ps < 8500) {
	    Set_Graf_Scale_H(form, 2000, 11000, 5, 1);
	} else if (maxmem_ps < 11000) {
	    Set_Graf_Scale_H(form, 2500, 13500, 5, 1);
	} else if (maxmem_ps < 16000) {
	    Set_Graf_Scale_H(form, 2500, 19500, 7, 1);
	} else if (maxmem_ps < 22000) {
	    Set_Graf_Scale_H(form, 5000, 27000, 5, 0);
	} else if (maxmem_ps < 45000) {
	    Set_Graf_Scale_H(form, 10000, 55000, 5, 0);
	} else {
	    Set_Graf_Scale_H(form, 25000, 195000, 7, 0);
	}

  int color_ps = form->color_ps;
  int color_ps_d = form->color_ps_d;
  int yd, yd2;

	if (cdv && cdvd) {
	    y = Graf_y_ps(form, cdv->sampt[0]);
	    yd = Graf_y_ps(form, cdvd->sampt[0]);
	    x = form->gr_x;
	    for (i=0; i<sampz; i++) {
		x2 = form->gr_x + form->gr_w * (i+1) / sampz;
		y2 = Graf_y_ps(form, cdv->sampt[i]);
		yd2 = Graf_y_ps(form, cdvd->sampt[i]);
		DrawLineRGB(az->WinID, color_ps_d, x, yd, x2, yd2);
		DrawLineRGB(az->WinID, color_ps, x, y, x2, y2);
		x = x2; y = y2; yd = yd2;
	    }
	} else if (cdv || cdvd) {
	    cdx = cdv;
	    if (!cdx) {cdx = cdvd; color_ps = color_ps_d;}
	    y = Graf_y_ps(form, cdx->sampt[0]);
	    x = form->gr_x;
	    for (i=0; i<sampz; i++) {
		x2 = form->gr_x + form->gr_w * (i+1) / sampz;
		y2 = Graf_y_ps(form, cdx->sampt[i]);
		DrawLineRGB(az->WinID, color_ps, x, y, x2, y2);
		x = x2; y = y2;
	    }
	}

	az_MHz(form);
}

void TTeCache_pnlExec_OnResize(TWins *self) {
  TFormTeCache *form = (TFormTeCache*)self->FormStruct;
	form->gr_w = self->width - form->gr_x - 34;
	form->gr_h = self->height - 20;
}

static int TTeCache_RunMinT(TFormTeCache *self, int k, int mem) {
  dword z, zm, z0, z1, z2, z3, z4, z5, z6, z7;
#if defined (__i386__) || defined (__x86_64__)
  uint64_t za, ze0, ze1, ze2, ze3, ze4, ze5, ze6, ze7, zeX;
  int km;
#endif
	if ((k < 1) || !self->spProc) return 0;

#if defined (__i386__) || defined (__x86_64__)
	if (self->TSC_on) {
	    z = GetTime_us();

	    Get_TSC(&za);
	    self->spProc(k);
	    Get_TSC(&ze0);
	    self->spProc(k);
	    Get_TSC(&ze1);
	    self->spProc(k);
	    Get_TSC(&ze2);
	    self->spProc(k);
	    Get_TSC(&ze3);
	    self->spProc(k);
	    Get_TSC(&ze4);
	    self->spProc(k);
	    Get_TSC(&ze5);
	    self->spProc(k);
	    Get_TSC(&ze6);
	    self->spProc(k);
	    Get_TSC(&ze7);


	    zm = ze0-8-za;
	    km = ze1-8-ze0;
	    if (zm > km) zm = km;
	    km = ze2-8-ze1;
	    if (zm > km) zm = km;
	    km = ze3-8-ze2;
	    if (zm > km) zm = km;
	    km = ze4-8-ze3;
	    if (zm > km) zm = km;
	    km = ze5-8-ze4;
	    if (zm > km) zm = km;
	    km = ze6-8-ze5;
	    if (zm > km) zm = km;
	    km = ze7-8-ze6;
	    if (zm > km) zm = km;


	    z3 = GetTime_us();
	    Get_TSC(&zeX);

	    km = (zeX-8-za) / (z3-z);
	    zm *= 10;
	    zm = zm / km;
	    self->run_time += zm * 8 / 10;

	} else {
#endif
	    z = GetTime_us();
	    self->spProc(k);
	    z0 = GetTime_us();
	    self->spProc(k);
	    z1 = GetTime_us();
	    self->spProc(k);
	    z2 = GetTime_us();
	    self->spProc(k);
	    z3 = GetTime_us();
	    self->spProc(k);
	    z4 = GetTime_us();
	    self->spProc(k);
	    z5 = GetTime_us();
	    self->spProc(k);
	    z6 = GetTime_us();
	    self->spProc(k);
	    z7 = GetTime_us();
	    zm = z0-z;
	    if (zm > (z1-z0)) zm = z1-z0;
	    if (zm > (z2-z1)) zm = z2-z1;
	    if (zm > (z3-z2)) zm = z3-z2;
	    if (zm > (z4-z3)) zm = z4-z3;
	    if (zm > (z5-z4)) zm = z5-z4;
	    if (zm > (z6-z5)) zm = z6-z5;
	    if (zm > (z7-z6)) zm = z7-z6;
	    self->run_time += zm * 8;
	    zm *= 10;
#if defined (__i386__) || defined (__x86_64__)
	}
#endif

	return zm;
}

#if KORR_GetTime_us
static int t_GetTime_us(TFormTeCache *self, int k) {
  dword z, z0, z1, z2, z3, z4, z5, z6, z7;
  dword *sp;

	sp = self->spI;
	if (!sp) return 0;
#if defined (__i386__) || defined (__x86_64__)
  byte *sb;
	sb = (byte*)sp;
	*sb = 0xc3;		//ret
#elif defined(__arm__)
	*sp = 0xe12fff1e;		//bx	lr
#else
# error "Unsupported architecture!"
#endif
	    z = GetTime_us();
	    self->spProc(k);
	    z0 = GetTime_us();
	    self->spProc(k);
	    z1 = GetTime_us();
	    self->spProc(k);
	    z2 = GetTime_us();
	    self->spProc(k);
	    z3 = GetTime_us();
	    self->spProc(k);
	    z4 = GetTime_us();
	    self->spProc(k);
	    z5 = GetTime_us();
	    self->spProc(k);
	    z6 = GetTime_us();
	    self->spProc(k);
//	    z7 = GetTime_us();
// printf("%s(%d): %d\n", FL, z7 - z);
	return z7 - z;
}
#endif

#if USE_MMAP
#define EXECMEM_PLUS 0x0000
#else
#define EXECMEM_PLUS 0x2100
#endif

void TTeCache_FreeExecMem(TFormTeCache *self) {
#if USE_MMAP
	if (self->ExecMem) munmap(self->ExecMem, (self->splg * 4) + EXECMEM_PLUS);
#else
	xfree(self->ExecMem);
#endif
	self->ExecMem = NULL;
	self->spI = NULL;
	self->spProc = NULL;
	self->splg = 0;
}

static void TTeCache_SetExecMem(TFormTeCache *self, int mc) {
  dword *sp;
	if (self->splg < mc) {
	    TTeCache_FreeExecMem(self);
#if USE_MMAP
	    mc = (mc + 0x3ff) & 0xffffffc00;
	    sp = mmap(0, (mc * 4) + EXECMEM_PLUS, PROT_READ | PROT_WRITE | PROT_EXEC,
				MAP_PRIVATE | MAP_ANON, 0, 0);
	    if ((long int)sp == -1) {
		sp = NULL;
	    }
// printf("%s(%d): 0x%06x\n", FL, mc);
	    self->spI = sp;
#else
	    sp = xmalloc((mc * 4) + EXECMEM_PLUS);
	    if (sp) self->spI = (dword*)(((int)sp + 0x1fff) & 0xffffe000);
#endif
	    self->ExecMem = sp;
	    self->spProc = (void*)self->spI;
	    if (sp) self->splg = mc;
	}
}

static void ProcCodeSet(dword *sp0, int mc0) {
  dword *sp;

	if (!sp0) return;
	sp = sp0;
#if defined (__i386__) || defined (__x86_64__)
  byte *sb;
#if defined (__i386__)
  int mc = mc0 - 2;
#else
  int mc = mc0 - 3;
#endif
	
	while (mc--) {
	    *sp = 0x04244c8b;	//mov	4(%esp), %ecx;
	    sp++;
	}
	sb = (byte*)sp;
#if defined (__i386__)
	*sb++ = 0x48;		//dec	%eax
#else
	*sb++ = 0xff;		//dec	%edi
	*sb++ = 0xcf;
#endif
	*sb++ = 0x0f;		//jnz
	*sb++ = 0x85;	
	sp = (dword*)sb;
	*sp++ = ((byte*)sp0 - sb); // + 4;
	sb = (byte*)sp;
	*sb++ = 0xc3;		//ret

#elif defined(__arm__)
	mc0 -= 4;
	if (mc0 <= 0) mc0 = 1;
	if (mc0 > 0x007fffff) mc0 = 0x007fffff;
  int mc = mc0;
	while (mc--) {
//	    *sp++ = 0xe59d4000;		//ldr	r4, [sp, #0]
	    *sp++ = 0xe0011001;		//and	r1, r1, r1
	}
	*sp++ = 0xe2500001;		//subs	r0, r0, #1
	*sp++ = 0x1affffff - mc0;	//bne xxx
	*sp = 0xe12fff1e;		//bx	lr
#else
# error "Unsupported architecture!"
#endif
}

void TTeCache_InitProc(TFormTeCache *self, int mc) {
	TTeCache_SetExecMem(self, mc);
// printf("%s(%d): 0x%06x 0x%08x 0x%08x\n", FL, mc, (int)self->ExecMem, (int)self->spI);
	ProcCodeSet(self->spI, mc);
}

static int get_prog_ps(TFormTeCache *self, int k, int mem) {
  int z;
  double dz;
	TTeCache_InitProc(self, mem);
#if KORR_GetTime_us
	z = TTeCache_RunMinT(self, k, mem) - t_GetTime_us(self, 1);
#else
	z = TTeCache_RunMinT(self, k, mem);
#endif
	if (z < 1) z = 1;
	dz = (100 * (double)z) / ((double)k * (double)mem);
	return dz * 1000;
}

static void add_mem_lg(TFormTeCache *self, int *mem) {
	if (self->akt_MX == 0) {
	    *mem += 0x200;
	    if (*mem > 0x4000) *mem += 0x200;
	    if (*mem > 0x10000) *mem += 0x400;
	    if (*mem > 0x40000) *mem += 0x800;
	    if (*mem > 0x100000) *mem += 0x1000;
	} else if (self->akt_MX < 3) {
	    *mem += 0x400;
	    if (*mem > 0x4000) *mem += 0x400;
	    if (*mem > 0x10000) *mem += 0x800;
	    if (*mem > 0x40000) *mem += 0x1000;
	    if (*mem > 0x100000) *mem += 0x2000;
	} else if (self->akt_MX < 6) {
	    *mem += 0x800;
	    if (*mem > 0x4000) *mem += 0x800;
	    if (*mem > 0x10000) *mem += 0x1000;
	    if (*mem > 0x40000) *mem += 0x2000;
	    if (*mem > 0x100000) *mem += 0x4000;
	} else if (self->akt_MX < 7) {
	    *mem += 0x1000;
	    if (*mem > 0x4000) *mem += 0x1000;
	    if (*mem > 0x10000) *mem += 0x2000;
	    if (*mem > 0x40000) *mem += 0x4000;
	    if (*mem > 0x100000) *mem += 0x8000;
	} else {
	    *mem += 0x2000;
	    if (*mem > 0x4000) *mem += 0x2000;
	    if (*mem > 0x10000) *mem += 0x4000;
	    if (*mem > 0x40000) *mem += 0x8000;
	    if (*mem > 0x100000) *mem += 0x10000;
	    if (*mem > 0x400000) *mem += 0x20000;
	    if (*mem > 0x1000000) *mem += 0x40000;
	}
}

void TTeCache_get_MHz(TFormTeCache *self) {
  int z, z2;

#if defined (__i386__) || defined (__x86_64__)
  int us;
  uint64_t za, ze;

	if (self->TSC_on) {
	    us = GetTime_us();
	    Get_TSC(&za);
	    while (GetCurTimeDiv_us(us) < 5000);
	    us = GetCurTimeDiv_us(us);
	    Get_TSC(&ze);
	    self->MHz = (ze-za) / us;
	    if (self->cd_view) self->cd_view->MHz = self->MHz;

	} else {
#endif
	    z = GetTime_us();
	    GetCurTimeDiv_us(z);
	    z = GetTime_us();
	    self->us_off = GetCurTimeDiv_us(z);

#define MHzMEM 0x400

	    TTeCache_InitProc(self, MHzMEM);
	    TTeCache_RunMinT(self, 2, MHzMEM);
	    z = TTeCache_RunMinT(self, 100, MHzMEM);
	    z2 = TTeCache_RunMinT(self, 100, MHzMEM);
	    if (z2 < z) z = z2;
	    int k = 1000000 / z;

	    get_prog_ps(self, 1, MHzMEM);
	    z = get_prog_ps(self, k, MHzMEM);
	    z2 = get_prog_ps(self, k, MHzMEM);
	    if (z2 < z) z = z2;
	    z2 += get_prog_ps(self, k, MHzMEM);
	    if (z2 < z) z = z2;
	    z2 += get_prog_ps(self, k, MHzMEM);
	    if (z2 < z) z = z2;
	    if (z > 0) {
		self->MHz = 1000000 / z;
		if (self->cd_view) self->cd_view->MHz = 1000000 / z;
	    } else {
		self->MHz = 0;
		if (self->cd_view) self->cd_view->MHz = 0;
	    }
#if defined (__i386__) || defined (__x86_64__)
	}
#endif
}

#ifdef use_exec_Xns
void exec_init_PV(TScaleXnsRaster *xr, TFormTeCache_Dat *cd, int idx, int count) {
  TXnsPoints *PVs;
  TXnsPoint *PV;
  float ScaleMaxX, off;

	PVs = &xr->PVs[idx];
	if (PVs->PV_Count != count) {
	    PVs->PV_Count = count;
	    PVs->enabled = True;
	    xfree(PVs->PV);
	    PVs->PV = xcalloc(sizeof(TXnsPoints) * PVs->PV_Count);
	    PV = PVs->PV;

	off = log((float)xr->PX[1].val) + (log((float)xr->PX[1].val) - log((float)xr->PX[2].val));
	ScaleMaxX = log((float)xr->PX[xr->PX_Count-1].val) - off;

	    for (int i=0;i<cd->sampz;i++) {
		PV[i].fx = (log((float)cd->sampmem[i]*4) - off) / ScaleMaxX;
	    }
	}
}
#endif

void TTeCache_memExec_Start(TFormTeCache *self) {
  int i, z, k, mem, memsz, ms;
#ifdef use_exec_Xns
  TScaleXnsRaster *xr;
#endif
  TFormTeCache_Dat *cd;

	if (self->TestRun) return;
	cd = self->cd_view;
	if (!cd) return;
	ms = GetTime_ms() + 250;

	self->TestRun = True;
	self->Stop = False;
	SetEnabled((TWins*)self->buStop, True);

	cd->ges_time = 0;
	cd->min_ps = 0;
	cd->max_ps = 0;

	memsz = 0x10000;
	i = self->akt_MX;
	while (i--) memsz <<= 1;

	cd->sampz = 0;
	cd->ges_time = 0;
	cd->MHz = 0;
// printf("%s(%d): 0x%06x\n", FL, memsz);
//	TTeCache_InitProc(self, memsz);
	TTeCache_SetExecMem(self, memsz);
	TTeCache_pnlExec_OnPaint(self->pnlExec);

	if (self->Start250ms) {
	    while ((!cpTime(ms, GetTime_ms())) && AppProcessMessages());
	}

	z = GetTime_us();
	GetCurTimeDiv_us(z);
	z = GetTime_us();
	self->us_off = GetCurTimeDiv_us(z);

	TTeCache_get_MHz(self);
	TTeCache_pnlExec_OnPaint(self->pnlExec);
	if (!AppProcessMessages() || self->Stop) goto Stop;

	mem = 0;
	add_mem_lg(self, &mem);
	k = mem * 1024;
	for (i=0;i<256;i++) {
	    if (mem > memsz) break;
	    self->sampk[i] = k / mem;
	    cd->sampmem[i] = mem;
// printf("%s(%d): %d 0x%06x\n", FL, self->sampk[i], cd->sampmem[i]);
	    cd->sampt[i] = 0;
	    add_mem_lg(self, &mem);
	}
	cd->sampz = i;
#ifdef use_exec_Xns
	xr = self->ns_exec->xr;
	xr->fu &= ~0x01;
	exec_init_PV(xr, cd, 0, i);
#endif
	i--;

	cd->maxmem_ps = get_prog_ps(self, self->sampk[i], cd->sampmem[i]);
	z = get_prog_ps(self, self->sampk[i], cd->sampmem[i]);
	if (cd->maxmem_ps > z) cd->maxmem_ps = z;

	TTeCache_pnlExec_OnPaint(self->pnlExec);
	if (!AppProcessMessages() || self->Stop) goto Stop;
#if KORR_GetTime_us
	t_GetTime_us(self, 1);
#endif
  int x, x2, y, y2;
	cd->min_ps = 0x7fffffff;
	cd->max_ps = 0;
	x = self->gr_x;
	y = 0;
	cd->ges_time = 0;
	self->run_time = 0;
#ifdef use_exec_Xns
	TScaleXns_LineXY_0(self->ns_exec, xr);
#endif
	for (i=0;i<cd->sampz;i++) {
	    z = get_prog_ps(self, self->sampk[i], cd->sampmem[i]);
	    if (cd->min_ps > z) cd->min_ps = z;
	    if (cd->max_ps < z) cd->max_ps = z;
	    cd->sampt[i] = z;
#ifdef use_exec_Xns
	    xr->PVs[0].PV[i].fy = (float)z / 1000;
	    TScaleXns_LineXY_N(self->ns_exec, xr, i);
#endif
	    x2 = self->gr_x + self->gr_w * (i+1) / cd->sampz;
	    y2 = Graf_y_ps(self, z);
	    if (i == 0) y = y2;
	    DrawLineRGB(self->pnlExec->WinID, self->color_ps, x, y, x2, y2);
	    x = x2; y = y2;
	    if (!AppProcessMessages() || self->Stop) {cd->sampz = i; goto Stop;}
	}

#ifdef use_exec_Xns
	xr->PHmax_ns = (float)cd->max_ps / 1000;
	TTeCache_Set_Max_ns_Scale(xr);
	Refresh((TWins*)self->ns_exec);
	xr->fu |= 0x01;
#endif

Stop:
	cd->ges_time = self->run_time;
	TTeCache_FreeExecMem(self);
	self->TestRun = False;
	SetEnabled((TWins*)self->buStop, False);
	TTeCache_get_MHz(self);
	az_MHz(self);
	TTeCache_pnlExec_OnPaint(self->pnlExec);
}


