/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appXbase.h"
#include "dlg.h"
#include "cicons.h"

#define TCSelf TFormMain
typedef struct TFormMain {
  AlgTForm
  TWins *pnlTop;
  TWins *pnlMiddle;
  TWins *pnlBottom;

  TMenuButton *mbFile;
    TMenu *menuFile;
      TMenu *menuFile2;
      TMenu *menuFile3;
      TMenu *menuFile4;

  TMenuButton *mbEdit;

  TMenuButton *mbHelp;
    TMenu *menuHelp;


  TWins *pnlL;
  TSplitterV *splittV;
  TWins *pnlR;


  TButton *buTest;
  TButton *buClose;
} TFormMain;
#undef TCSelf

int build_v = ALD_BUILD_V;
static const char *build_date = __DATE__;

static void mHelp_Info(TMenu *self, long KeyState, TMenuItem *smi) {
  TFormMain *fmain = (TFormMain*)self->FormStruct;
  char buf[80];
	snprintf(buf, sizeof(buf), "ALDconsole %s   (build-%d  %s)", 
				VERSION, build_v, build_date);
	Copyright_MsgBox_Show((TWins*)fmain, buf);
}

static void mExit_OnClick(TMenu *self, long KeyState, TMenuItem *smi) {
  Close();
}

static void mFile_OnMenuOpen(TMenu *self) {
}

static void buTest_OnClick(TButton *self, long KeyState, bool DblClick) {
  TFormMain *fmain = (TFormMain*)self->FormStruct;
	if (fmain->splittV->width == 4) {
	    SetWidth((TWins*)fmain->splittV, 20);
	} else if (fmain->splittV->width == 20) {
	    SetWidth((TWins*)fmain->splittV, 40);
	} else {
	    SetWidth((TWins*)fmain->splittV, 4);
	}
}

static void buClose_OnClick(TButton *self, long KeyState, bool DblClick) {
	Close();
}

static void pnlBottom_OnResize(TWins *self) {
  TFormMain *fmain = (TFormMain*)self->FormStruct;
	fmain->buClose->left = self->width - fmain->buClose->width - 1;
}

static const char *Help_Msg = {
  "\n"
  "Usage: edit\n"
  "\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: edit version "VERSION"  (build %d  %s)\n"
};


int main (int argc, char **argv) {
  int i;
  char *val;
//  char *InFile;
  TFormMain *self;
  TMenuItem *mi;
  bool help = False;
  bool err = False;


//	InFile = NULL;
	i = 1;
	while (!err && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'h':
		    help = True;
		    break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
		    } else err = True;
		    break;
		default: 
		    err = True;
		}
	    } else {
//		InFile = val;
		err = True;
	    }
	    i++;
	}

//	err = err || !InFile;
	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(err);
	}

	initXapp(0, argc, argv);
	self = TForm_CreatParamAlg(NULL, sizeof(TFormMain));
	appDat.obj = self;

	self->pnlTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder | mBorderL, alTop);
	self->pnlMiddle = TWins_CreateBA((TWins*)self, 0, 0, 200, 42, mBorder | mBorderL, alClient);
	self->pnlBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 23, 0, alBottom);
	self->pnlBottom->OnResize = pnlBottom_OnResize;

/* Menu File */
	self->mbFile = TMenuButton_Create(self->pnlTop, (TMenuButton*)self);
	self->mbFile->top = 2;
	self->mbFile->left = 1;
	SetCaption((TWins*)self->mbFile, "File");
	Set_Parent((TWins*)self->mbFile, self->pnlTop);

	self->menuFile2 = TMenu_Create((TWins*)self);
	TMenu_AddItem(self->menuFile2, "Menu 2", miCanCheck|miAutoToggle, NULL);
	self->menuFile3 = TMenu_Create((TWins*)self);
	TMenu_AddItem(self->menuFile3, "Menu 3", miCanCheck|miChecked|miAutoToggle, NULL);
	self->menuFile4 = TMenu_Create((TWins*)self);
	TMenu_AddItem(self->menuFile4, "Menu 4", 0, NULL);

	self->menuFile = TMenu_Create((TWins*)self);
	self->mbFile->menu = self->menuFile;
	self->menuFile->OnMenuOpen = mFile_OnMenuOpen;
	TMenu_AddItem(self->menuFile, "Item 1", 0, NULL);
	TMenu_AddItem(self->menuFile, "Item 2", 0, NULL)->next_menu = self->menuFile2;
	TMenu_AddItem(self->menuFile, "Item 3", 0, NULL)->next_menu = self->menuFile3;
	TMenu_AddItem(self->menuFile, "Item 4", 0, NULL)->next_menu = self->menuFile4;
	TMenu_AddItem(self->menuFile, "-", 0, NULL);
	TMenu_AddItem(self->menuFile, "Exit", 0, mExit_OnClick);

/* Menu Edit */
	self->mbEdit = TMenuButton_Create(self->pnlTop, self->mbFile);
	SetCaption((TWins*)self->mbEdit, "Edit");
	Set_Parent((TWins*)self->mbEdit, self->pnlTop);

/* Menu Help */
	self->mbHelp = TMenuButton_Create(self->pnlTop, self->mbFile);
	SetCaption((TWins*)self->mbHelp, "Help");
	Set_Parent((TWins*)self->mbHelp, self->pnlTop);

	self->menuHelp = TMenu_Create((TWins*)self);
	self->mbHelp->menu = self->menuHelp;
	mi = TMenu_AddItem(self->menuHelp, "Info", 0, mHelp_Info);
	mi->pic = GetZIcon(zico16info);
/* pnlMiddle */
	self->pnlL = TWins_CreateBA(self->pnlMiddle, 0, 0, 150, 42, mBorder, alLeft);
	self->pnlL->min_width = 60;
	self->splittV = TSplitterV_CreateP(self->pnlMiddle);
	self->pnlR = TWins_CreateBA(self->pnlMiddle, 0, 0, 150, 42, mBorder, alClient);
	self->pnlR->min_width = 60;

/* pnlBottom */
	self->buClose = TButton_CreateS(self->pnlBottom, 1, 1, 60, 22, "Close");
	self->buClose->OnClick = buClose_OnClick;

	self->buTest = TButton_CreateS(self->pnlBottom, 1, 1, 60, 22, "Test");
	self->buTest->OnClick = buTest_OnClick;


	Set_Focus((TWins*)self->buClose);

	AppRun();

	Tfree(self);

#ifdef MEM_CHK
  int lMCt, lMsize;
  MEM_CHK_Count(&lMCt, &lMsize);
  printf("Debug %s(%d): MEM_CHK_Count=%3d  size=%d\n", FL, lMCt, lMsize);
#endif

	return 0;
}
