/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#ifdef STANDALONE_BUILDIN
#include "sysbase.h"
#include "sysbaseio.h"
#else
#include "../buildin_et.h"
#endif
#include "tsearch.h"

typedef struct {
  dword pid;
  char comm[32];
  char state[4];
  dword ppid;
  dword rss;
} TstatFindSharProc;

#define mFSP_so		1
#define mFSP_proc	(1<<1)

typedef struct {
  int num;
  dword vm_start;
  dword vm_end;
  char flags[4];
  dword vm_pgoff;
  dword MAJOR;
  dword MINOR;
  dword ino;
  char *fn;
  char *name;
  int stat;
  bool use_so;
  bool use_so_X;
  bool use_so_Cplus;
} TmapsFindSharProc;	//linux/fs/proc/task_mmu.c show_map_internal()

typedef struct Tso_maps {
  struct Tso_maps *next;
  char *fn;
} Tso_maps;

static Tso_maps *so_map;

static void ScanSpace_ppc(char **s) {
  byte *b = (byte*)*s;
	if (b) {
	    while (*b && (*b < 0x21)) b++;
	}
	*s = (char*)b;
}

static int bi_findsharprocs_maps(char *s, TmapsFindSharProc *mFSP) {
  int i;
  int res = 0;
  char *pc = s;

	mFSP->stat = 0;
	while (*pc && (*pc != '\n')) pc++;
	if (*pc == '\n') *pc = 0;

	mFSP->vm_start = ppctohex_scan((void*)&s, True);
	mFSP->vm_end = ppctohex_scan((void*)&s, True);
	ScanSpace_ppc(&s);
	i = 0;
	while (i<4) {mFSP->flags[i] = *s; if (*s) s++; i++;}
	mFSP->vm_pgoff = ppctohex_scan((void*)&s, True);
	mFSP->MAJOR = ppctohex_scan((void*)&s, True);
	mFSP->MINOR = ppctohex_scan((void*)&s, True);
	mFSP->ino = ppctoint_scan((void*)&s, True);
	ScanSpace_ppc(&s);
	mFSP->fn = NULL;
	if (*s) {
	    res = 1;
	    mFSP->num++;
	    mFSP->fn = s;
	    mFSP->name = s;
	    while (*s) {
		if (*s == '/') mFSP->name = s+1;
		s++;
	    }
	    if ((*mFSP->fn == '/') && (mFSP->flags[1] == '-')) {
		s = mFSP->name+1;
		while (*s) {
		    if ((s[0] == '.') && (s[1] == 's') && (s[2] == 'o') && 
			    (!s[3] || (s[3] == '.'))) {
			mFSP->stat |= mFSP_so;
			mFSP->use_so = True;
			if (pcncmp(mFSP->name, "libX", 4)) mFSP->use_so_X = True;
			if (pcncmp(mFSP->name, "libstdc++", 9)) mFSP->use_so_Cplus = True;
			if (pcncmp(mFSP->name, "libgcc_s", 8)) mFSP->use_so_Cplus = True;
			break;
		    }
		    s++;
		}
		if (!(mFSP->stat & mFSP_so)) mFSP->stat |= mFSP_proc;
	    }
	}
/*
  printf("%d %08x-%08x %c%c%c%c %08x %02x:%02x %u %s\n", res,
		mFSP->vm_start, mFSP->vm_end, 
		mFSP->flags[0], mFSP->flags[1], mFSP->flags[2], mFSP->flags[3], 
		mFSP->vm_pgoff, 
		mFSP->MAJOR, mFSP->MINOR, 
		mFSP->ino, mFSP->fn);
*/
	return res;
}

//#define FSP_list_all		1
#define FSP_list_name		(1<<1)
#define FSP_list_pid		(1<<2)
//#define FSP_list_maps		(1<<3)
#define FSP_list_so_procs	(1<<4)
#define FSP_list_Cplus_so_procs	(1<<5)
#define FSP_list_proc		(1<<8)
#define FSP_list_so		(1<<9)
#define FSP_list_pathfn		(1<<10)

static bool add_so_maps(TmapsFindSharProc *mFSP, int mode) {
  Tso_maps *map;
	if (mode & (FSP_list_name | FSP_list_pid | FSP_list_so_procs | FSP_list_proc)) return False;
	map = so_map;
	while (map && !pccmp(map->fn, mFSP->fn)) map = map->next;
	if (map) return True;
	map = xmalloc(sizeof(Tso_maps));
	map->fn = pcdup(mFSP->fn);
	map->next = so_map;
	so_map = map;
	return False;
}

static void bi_findsharprocs_list_maps(char *fn, int mode) {
  FILE *f;
  char line[1024];
  char *tline, *pc;
  TmapsFindSharProc mFSP;

	if (!(mode & (FSP_list_proc | FSP_list_so))) return;
	f = fopen(fn, "r");
	if (f) {
map_loop:
	    tline = fgets(line, sizeof(line)-1, f);
	    if (tline) {
		if (bi_findsharprocs_maps(tline, &mFSP)) {
		    if ((mode & FSP_list_proc) && (mFSP.stat & mFSP_proc)) {
			pc = mFSP.name;
			if (mode & FSP_list_pathfn) pc = mFSP.fn;
			printf("\"%s\"\n", pc);
		    }
		    if ((mode & FSP_list_so) && (mFSP.stat & mFSP_so) &&
			!add_so_maps(&mFSP, mode)) {
			pc = mFSP.name;
			if (mode & FSP_list_pathfn) pc = mFSP.fn;
			printf("\"%s\"\n", pc);
		    }
		}
		goto map_loop;
	    }
	    fclose(f);
	}

}

static const char *Help_Msg = {
  "\n"
  "Usage: findsharprocs [options] \n"
  "\n"
  "  -s                   chk shared libs use\n"
  "                           Output N no shared libs used\n"
  "                           Output Y shared libs used\n"
  "                           Output X shared X11 libs used\n"
  "  -s+                  chk g++ shared libs use\n"
  "                           Output Y or N shared g++ libs used\n"
  "  -n                   list name\n"
  "  -p                   list pid\n"
  "  -S                   list shared libs maps\n"
  "  -P                   list proc maps\n"
//  "  -F                   list maps with path\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: findsharprocs version "VERSION"  (build %d  %s)\n"
};

#ifdef STANDALONE_BUILDIN

const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;

int main (int argc, char **argv) {
//	cmdline_slash_korr(argc, argv);
#else

//extern const int build_v;
//extern const char *build_date;

int bi_findsharprocs_main (int argc, char **argv) {
#endif
  int i;
  int mode = 0;
  char *pc, *phat0, *pcn, *pcx;
  const char *cpc;
  TSearch_Rec sr;
  char buf[64];
  FILE *f;
  char line[1024];
  char *tline;
  TstatFindSharProc sFSP;
  TmapsFindSharProc mFSP;
  bool help = False;
  bool err = False;
  bool use_so = False;
  bool use_so_X = False;
  bool use_so_Cplus = False;

	i = 1;
	while ((i < argc) && !help && !err) {
	    pc = argv[i];
	    if (*pc == '-') {
		pc++;
OptLoop:
		if (*pc == 'h') {
		    help = True;
		} else if (*pc == 's') {
		    mode |= FSP_list_so_procs;
OptNext:
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == '+') {
		    mode |= FSP_list_Cplus_so_procs;
		    goto OptNext;
		} else if (*pc == 'n') {
		    mode |= FSP_list_name;
		    goto OptNext;
		} else if (*pc == 'p') {
		    mode |= FSP_list_pid;
		    goto OptNext;
		} else if (*pc == 'S') {
		    mode |= FSP_list_so;
		    goto OptNext;
		} else if (*pc == 'P') {
		    mode |= FSP_list_proc;
		    goto OptNext;
		} else if (*pc == 'F') {
//		    mode |= FSP_list_pathfn;
		    goto OptNext;
		} else if (*pc == '-') {
		    pc++;
		    if (pcncmp(pc, "help", 4)) {
			help = True;
		    } else err = True;
		} else err = True;
	    } else {
		err = True;
	    }
	    i++;
	}

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(err);
	}

	so_map = NULL;
	if (!mode) mode = FSP_list_so_procs | FSP_list_name | FSP_list_pid | FSP_list_so | FSP_list_pathfn;

	phat0 = buf;
	ppcmove(&phat0, "/proc/");
	
	sr.path = pcdup("/proc/");
	sr.patt = NULL;
	sr.statfu = fsNoLinkName;
	sr.findAtr = findDIR | findIsRootPath;
	Find_First(&sr);
	while (sr.ret_entry) {
	    cpc = Get_ExFileDat_pc(sr.st, exfdt_FileName);
// printf("%s(%d): %s\n", FL, pc);
	    if ((*cpc >= '0') && (*cpc <= '9')) {
		sFSP.rss = 0;
		pcn = phat0;
		ppcmove(&pcn, cpc);
		pcx = pcn;
		ppcmove(&pcx, "/stat");
// printf("%s(%d): %s\n", FL, buf);
		f = fopen(buf, "r");
		if (f) {
		    tline = fgets(line, sizeof(line)-1, f);
		    fclose(f);

		    pcx = tline;
		    if (pcx) {
			sFSP.pid = ppctoint((void*)&pcx);
			while (*pcx && (*pcx != '(')) pcx++;
			if (*pcx) pcx++;
			i = 0;
			pc = sFSP.comm;
			while (*pcx && (*pcx != ')') && (i < 31)) {
			    *pc = *pcx;
			    pc++; pcx++; i++;
			}
			*pc = 0;
		    }
		    if (pcx && (*pcx == ')')) {
			pcx++;

			i = sscanf(pcx,
				" %c %u "               /* state, ppid */
				"%*u %*u %*s %*s "       /* pgid, sid, tty, tpgid */
				"%*s %*s %*s %*s %*s " /* flags, min_flt, cmin_flt, maj_flt, cmaj_flt */
				"%*d %*d "             /* utime, stime */
				"%*s %*s %*s "         /* cutime, cstime, priority */
				"%*d "                 /* nice */
				"%*s %*s %*s "         /* timeout, it_real_value, start_time */
				"%*s "                 /* vsize */
				"%u",                 /* rss */
				sFSP.state, &sFSP.ppid,
				&sFSP.rss);
		    }
		}

//		if (sFSP.rss || (mode & FSP_list_all)) {
		if (sFSP.rss) {

// printf("%s %u\n", sFSP.comm, sFSP.pid);

		    mFSP.num = 0;
		    mFSP.use_so = False;
		    mFSP.use_so_X = False;
		    mFSP.use_so_Cplus = False;
		    pcx = pcn;
		    ppcmove(&pcx, "/maps");
		    f = fopen(buf, "r");
		    if (f) {
map_loop:
			tline = fgets(line, sizeof(line)-1, f);
			if (tline) {
			    if (bi_findsharprocs_maps(tline, &mFSP)) {
				use_so |= mFSP.use_so;
				use_so_X |= mFSP.use_so_X;
				use_so_Cplus |= mFSP.use_so_Cplus;
			    }
			    goto map_loop;
			}
			fclose(f);
		    }
		    if ((mFSP.use_so && (mode & FSP_list_so_procs) && !(mode & FSP_list_Cplus_so_procs)) ||
			(mFSP.use_so_Cplus && (mode & FSP_list_Cplus_so_procs)) ||
			!(mode & FSP_list_so_procs)) {
			if (mode & FSP_list_name) printf("%s ", sFSP.comm);
			if (mode & FSP_list_pid) printf("%d ", sFSP.pid);
			if (mode & (FSP_list_name | FSP_list_pid)) printf("\n");
//			printf("%s %u\n", sFSP.comm, sFSP.pid);
		    bi_findsharprocs_list_maps(buf, mode);
		    }
//		    FSP_list_maps
		}

	    }
	    Find_Next(&sr);
	}
	if (mode == FSP_list_so_procs) {
	    cpc = "N";
	    if (use_so) cpc = "Y";
	    if (use_so_X) cpc = "X";
	    printf("%s\n", cpc);
	} else if (mode == (FSP_list_so_procs | FSP_list_Cplus_so_procs)) {
	    cpc = "N";
	    if (use_so_Cplus) cpc = "Y";
	    printf("%s\n", cpc);
	}


	return 0;
}
