/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#ifdef STANDALONE_BUILDIN
#include "sysbase.h"
#include "sysbaseio.h"
#else
#include "../buildin_et.h"
#endif

#define APPNAME "shdata"

static const char *Help_Msg = {
  "\n"
  "Usage: "APPNAME" [options] -I \"val\" \"File\"\n"
  "\n"
  "  -r                   Read Item\n"
  "  -w                   Write/Add Item\n"
  "  -rw                  Read Item is no Item Add Item\n"
  "  -wo                  Add Item only not exists\n"
  "  -I \"val\"             Item\n"
  "  -D \"val\"             r=Default w=Daten\n"
  "  -t                   Output whit \"dat\"\n"
  "  -x                   Output whit label \"item=dat\"\n"
  "  --nolf               Output no Linefeed\n"
  "  --disable            add comment \"# item=dat\"\n"
  "  -h, --help           Prints out this screen\n"
  "\n"
  "  ALDccp: "APPNAME" version "VERSION"  (build %d  %s)\n"
};

#define MAX_IBUF 3

#ifdef STANDALONE_BUILDIN

const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;

int main (int argc, char **argv) {
#ifdef _OS_WIN32
	cmdUTF8korr(&argc, &argv);
#else
	cmdline_slash_korr(argc, argv);
#endif
#else

int bi_shdata_main (int argc, char **argv) {
#endif
  int i, itemlg, lg, dlg;
  char *pc, *result;
  char *item = NULL;
  char *data = NULL;
  char *rdata = NULL;
  char *File = NULL;
  char *ibuf[MAX_IBUF];
  int icount = 0;
  TLines *tls;
  TColS *cs;
  bool help = False;
  bool err = False;
  bool ok = False;
  bool com = False;
  bool rd = False;
  bool wr = False;
  bool wrAddOnly = False;
  bool tMk = False;
  bool xMk = False;
  bool disable = False;
  bool nolf = false;

	i = 1;
	while ((i < argc) && !help && !err) {
	    pc = argv[i];
	    if (*pc == '-') {
		pc++;
OptLoop:
		if (*pc == 'r') {
		    rd = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 'w') {
		    wr = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 'o') {
		    wrAddOnly = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 't') {
		    tMk = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 'x') {
		    xMk = true;
		    pc++;
//		    if ((*pc == '1') || (*pc == '2')) {
//			if (*pc == '1') xMk = 2; else  xMk = 3;
//			pc++;
//		    }
		    if (*pc) goto OptLoop;

		} else if (*pc == 'h') {
		    help = True;

		} else if (*pc == 'I') {
		    err = cmd_get_pc(argc, argv, pc, &i, &item);

		} else if (*pc == 'D') {
		    err = cmd_get_pc(argc, argv, pc, &i, &data);

		} else if (*pc == '-') {
		    pc++;
		    if (pccmp(pc, "disable")) {
			disable = True;

		    } else if (pccmp(pc, "nolf")) {
			nolf = True;

		    } else if (pcncmp(pc, "help", 4)) {
			help = True;

		    } else err = True;

		} else err = True;

	    } else if (icount < MAX_IBUF) {
		ibuf[icount] = pc;
		icount++;

	    } else err = True;
	    i++;
	}

	if (!help && !err) {
	    if (icount == 1) {
		File = path_rel2abs(ibuf[0], NULL);

	    } else if (icount == 2) {
		if (!item && !data) {
		    item = ibuf[0];
		    File = path_rel2abs(ibuf[1], NULL);
		} else  err = True;

	    } else if (icount == 3) {
		if (!item && !data) {
		    item = ibuf[0];
		    data = ibuf[1];
		    File = path_rel2abs(ibuf[2], NULL);
		} else  err = True;
	    }
	    if (!File || !item) err = True;
	}

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(err);
	}
	if (!rd && !wr) rd = True;

	item = pcadd(item, "=");
	tls = TLines_Create();
	if (TLines_LoadFromFile(tls, File)) {
	    itemlg = pclen(item);
	    pc = NULL;
	    i = 0;
loop0:
	    cs = TLines_Get_pcs(tls, i);
	    if (cs) {
		pc = cs->s;
		if (pc) {
		    while (*pc && ((*pc == 0x09)||(*pc == 0x20))) pc++;
		    ok = pcncmp(pc, item, itemlg);
		}
		if (!ok) {
		    i++;
		    goto loop0;
		}
	    }

	    if (!ok) {
		i = 0;
loop1:
		cs = TLines_Get_pcs(tls, i);
		if (cs) {
		    pc = cs->s;
		    if (pc) {
			while (*pc && ((*pc == 0x09)||(*pc == 0x20))) pc++;
			if (*pc && (*pc == '#')) pc++;
			while (*pc && ((*pc == 0x09)||(*pc == 0x20))) pc++;
			com = pcncmp(pc, item, itemlg);
		    }
		    if (!com) {
			i++;
			goto loop1;
		    }
		}
	    }

	    if (ok || com) rdata = pc + itemlg;

	    if ((wr && !rd) || (wr && rd && !ok)) {
		char *wrpc = NULL;
		if (wrAddOnly) {
		    if (!ok && !com) {
			wrpc = pcadd(item, data);
		    } else if ((ok && disable)||(com && !disable)) {
			wrpc = pcdup(pc);
		    }
		} else if (!pccmp(data, rdata)) {
		    wrpc = pcadd(item, data);
		}

		if (wrpc) {
		    rdata = wrpc + itemlg;
		    char *pcn = wrpc;
		    if (disable) pcn = pcadd("# ", wrpc);

		    if (cs) {
			xfree(cs->s);
			cs->s = pcn;
			cs->lg = pclen(pcn);
		    } else {
			TLines_Add(tls, pcn);
		    }
		    TLines_SaveToFile(tls, File);
		    pc = wrpc;
		    ok = !disable;
		}
	    }

	    result = NULL;
	    if (rd && ok && rdata) {
		lg = 6;
		dlg = 0;
		if (xMk) lg += pclen(item);
		if (*pc) {
		    dlg = pclen(rdata);
		    if ((dlg > 1) && (rdata[0] == '\"') && (rdata[dlg - 1] == '\"')) {
			if (tMk) {
			    tMk = false;
			} else {
			    dlg -= 2;
			    rdata++;
			}
		    }
		}
		result = xmalloc(lg + dlg);
		pc = result;
		if (xMk) ppcmove(&pc, item);
		if (tMk) {*pc = '\"'; pc++;}
		if (dlg > 0) ppcnmove(&pc, rdata, dlg);
		if (tMk) {*pc = '\"'; pc++;}
		if (!nolf) {*pc = '\n'; pc++;}
		*pc = 0;
	    } else if (rd && data) {
		lg = 6;
		dlg = 0;
		if (xMk) lg += pclen(item);
		if (*pc) {
		    dlg = pclen(data);
		    if ((dlg > 1) && (data[0] == '\"') && (data[dlg - 1] == '\"')) {
			if (tMk) {
			    tMk = false;
			} else {
			    dlg -= 2;
			    data++;
			}
		    }
		}
		result = xmalloc(lg + dlg);
		pc = result;
		if (xMk) ppcmove(&pc, item);
		if (tMk) {*pc = '\"'; pc++;}
		if (dlg > 0) ppcnmove(&pc, data, dlg);
		if (tMk) {*pc = '\"'; pc++;}
		if (!nolf) {*pc = '\n'; pc++;}
		*pc = 0;
	    }
	    if (result) printf(result);
	}

	return 0;
}
